import json
from rest_framework import serializers

from clusterCore.exceptions import APIException406
from clusterCore.models import c123_node, c123_host, c123_cluster, Task, IntegratedService, ClusterSizing
from clusterCore.enums import DBType, ClusterRoles, Status, Severity, Schedule, TaskState, Node_Size_Allowed
from clusterCore.api.serializer import c123_CreateTempClusterSerializer, object_to_json
from clusterCore.tasks import license_after_create, format_password


class c123_CreateMongoClusterSerializer(serializers.HyperlinkedModelSerializer):
    cluster = c123_CreateTempClusterSerializer()

    class Meta:
        model = c123_node
        fields = ["cluster", "version", "database_user", "database_pwd", "sizing", "ex_params"]

    @license_after_create
    def create(self, validated_data):
        current_user = self.context["request"].user
        kubernetes_api_data = IntegratedService.objects.filter(pk=validated_data["ex_params"]["service_id"]).first()

        cluster_data = validated_data.pop("cluster")
        cluster_name = cluster_data["name"]
        cluster_namespace = cluster_data["ex_params"]["namespace"]

        if c123_cluster.objects.filter(name=cluster_name, ex_params__namespace=cluster_namespace).exists():
            raise APIException406(f'Cluster "{cluster_name}" already exists in namespace "{cluster_namespace}".')

        cluster_data.update({
            "created_by": current_user,
            "last_update_by": current_user,
            "type": DBType.MONGO_K8S.value,
        })
        cluster_data.setdefault("ex_params", {}).update({
            "instances": 1,
            "rw-service": f'{cluster_data["name"]}-svc',
        })
        cluster = c123_cluster.objects.create(**cluster_data)

        host_data = {
            "name": cluster.ex_params["rw-service"],
            "created_by": current_user,
            "last_update_by": current_user,
        }
        host_data.setdefault("ex_params", {}).update({
            "cluster_id": cluster.id,
        })
        host = c123_host.objects.create(**host_data)

        validated_data.update({
            # Pods in StatefulSet are named sequentially, starting from 0.
            "name": f"{cluster.name}-{cluster.ex_params['instances']-1}",
            "created_by": current_user,
            "last_update_by": current_user,
            "status": Status.CREATING.value,
            "severity": Severity.NOT_OK.value,
            "type": DBType.MONGO_K8S.value,
            "cluster_role": ClusterRoles.PRIMARY.value,
            "host": host,
            "cluster": cluster,
            "database_pwd": format_password(validated_data["database_pwd"]),
            "ind_deleted": False,
            "port": 27017,
            "dir_123Cluster": "/data",
        })
        validated_data.setdefault("ex_params", {})
        pod = c123_node.objects.create(**validated_data)

        mongo_sys, created = ClusterSizing.objects.get_or_create(db_type=DBType.MONGO_K8S.value)
        match validated_data["sizing"]:
            case Node_Size_Allowed.SMALL.name:
                ram_gb = mongo_sys.small_memory
                cpu_cores = mongo_sys.small_cpu
                disk_gb = mongo_sys.small_disk
            case Node_Size_Allowed.MEDIUM.name:
                ram_gb = mongo_sys.medium_memory
                cpu_cores = mongo_sys.medium_cpu
                disk_gb = mongo_sys.medium_disk
            case Node_Size_Allowed.LARGE.name:
                ram_gb = mongo_sys.large_memory
                cpu_cores = mongo_sys.large_cpu
                disk_gb = mongo_sys.large_disk
            case Node_Size_Allowed.VERY_LARGE.name:
                ram_gb = mongo_sys.very_large_memory
                cpu_cores = mongo_sys.very_large_cpu
                disk_gb = mongo_sys.very_large_disk
            case _:
                ram_gb = 1
                cpu_cores = 1
                disk_gb = 10

        tmp_data = {
            "created_by": current_user.id,
            "kubernetes_api_data": object_to_json(kubernetes_api_data),
            "cluster_data": object_to_json(cluster),
            "host_data": object_to_json(host),
            "pod_data": object_to_json(pod),
            "disk_gb": f"{disk_gb}Gi",
            "ram_gb": f"{ram_gb}Gi",
            "cpu_cores": cpu_cores,
        }

        task = Task.objects.create(
            task="install_k8s_mongo_pod_task",
            displayment="creating mongo pod {}".format(pod.name),
            target_type=DBType.MONGO_K8S.value,
            target_id=pod.id,
            org_id=pod.org_id,
            status=TaskState.PENDING.value,
            schedule=Schedule.NOW.value,
            parameters=json.dumps(tmp_data),
            created_by=current_user,
        )

        return pod, task
