import json
from pprint import pprint

import pytz
from datetime import datetime
from django.utils import timezone
import re

from rest_framework import generics
from rest_framework.decorators import api_view, permission_classes
from rest_framework.permissions import IsAuthenticated
from django.http import JsonResponse

from clusterCore.scheduler import MISFIRE_GRACE_TIME_BACKUP, schedule_backup
from clusterCore.tasks import append_status_cmd, get_backup_description, get_cron_schedule, checkPathExists, \
    allow_or_403, find_real_pg_primary, get_postgres_connection, remote_command, create_ssh_connection, cleanup_logs, \
    local_command, find_tilda
from c123_kubernetes.tasks import get_k8s_postgres_connection, drop_k8s_pg_cluster, checkK8SPathExists, check_space
from clusterCore.models import c123_node, c123_host, c123_inventory, Task, ClusterSizing, SystemConfig, c123_database, \
    c123_cluster, IntegratedService
from clusterCore.enums import DBType, Inventory, ClusterRoles, Status, Schedule, TaskState, RestoreOptions, Severity, \
    ObjectType
from .serializer import c123_CreatePGClusterSerializer, c123_AddPGStandbySerializer
from clusterCore.api.serializer import get_create_node_response, c123_HostSerializer, object_to_json, \
    c123_inventory_serializer, format_size
from clusterCore.exceptions import *

from ...tasks import drop_k8s_pg_pod, export_k8s_pg_node, export_k8s_pg_database

from dateutil.parser import parse


class PgClusterCreateView(generics.CreateAPIView):
    permission_classes = [IsAuthenticated]
    queryset = c123_node.objects.all()
    serializer_class = c123_CreatePGClusterSerializer

    def create(self, request, *args, **kwargs):
        return get_create_node_response(self, request, *args, **kwargs)


class PgStandbyAddView(generics.CreateAPIView):
    permission_classes = [IsAuthenticated]
    queryset = c123_node.objects.all()
    serializer_class = c123_AddPGStandbySerializer

    def create(self, request, *args, **kwargs):
        return get_create_node_response(self, request, *args, **kwargs)


@api_view(["POST"])
@permission_classes((IsAuthenticated,))
def DropPgCluster(request):
    cluster_id = request.data.get("cluster_id")
    final_backup = request.data.get('final_backup')

    try:
        cluster = c123_cluster.objects.filter(pk=cluster_id).first()
    except c123_cluster.DoesNotExist:
        raise APIException410(f"Cluster with id {cluster_id} does not exist")

    task = drop_k8s_pg_cluster(cluster, final_backup)
    return append_status_cmd(request, task)


@api_view(["POST"])
@permission_classes((IsAuthenticated,))
def DropPgPod(request):
    pod_id = request.data.get("pod_id")

    try:
        pod = c123_node.objects.filter(pk=pod_id, type=DBType.POSTGRES_K8S.value, ind_deleted=False).first()
    except c123_node.DoesNotExist:
        raise APIException410(f"Pod with ID {pod_id} does not exist")

    task = drop_k8s_pg_pod(pod)
    return append_status_cmd(request, task)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def checkDiskSpace(request):
    """check if a directory exists for a node export operation"""
    node_id = request.data.get('node_id')
    required_space = request.data.get('required_space')
    obj_type = request.data.get("obj_type")
    path = request.data.get('path')[1] if obj_type == ClusterRoles.CDB.value or obj_type == ClusterRoles.PDB.value else request.data.get('path')
    if ObjectType.DATABASE.value in obj_type:
        db = c123_database.objects.filter(pk = node_id).first()
        if not db:
            raise APIException410(f"database id {node_id} not available")
        node = db.node

    else:
         node = c123_node.objects.filter(pk=node_id, ind_deleted = False).first()
         if not node:
            raise APIException410(f"node id {node_id} not available")

    response = check_space(node = node, path = path, required_space = required_space)
    return JsonResponse(response, status = response['status'])

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def ExportPgNode(request):
    """export (dumpall) for pg cluster"""
    data = request.data
    data['created_by'] = request.user.id
    schedule = data.get('schedule')
    api_call = data.get('api_call')
    node_id = data.get('node_id')

    if not schedule:
        task = export_k8s_pg_node(data)
        return append_status_cmd(request, task)
    else:
        from django.conf import settings
        client_timezone = pytz.timezone(data.get("client_timezone"))
        schedule_run_time_str = data.get('schedule_run_time')
        schedule_run_time = parse(schedule_run_time_str).astimezone(client_timezone)
        client_hours = schedule_run_time.hour
        client_minutes = schedule_run_time.minute
        client_seconds = schedule_run_time.second
        days_of_week = data.get("days_of_week")
        days_of_month = data.get("days_of_month")
        if days_of_month:
            int_days_of_month = [parse(d).astimezone(client_timezone).day for d in days_of_month]
        else:
            int_days_of_month = None
        description = get_backup_description(client_hours, client_minutes, client_seconds, days_of_week, int_days_of_month)
        localized_datetime = datetime.now(client_timezone).replace(hour=client_hours, minute=client_minutes, second=client_seconds).astimezone(pytz.timezone(settings.TIME_ZONE))
        hours = localized_datetime.hour
        minutes = localized_datetime.minute
        seconds = localized_datetime.second
        trigger = get_cron_schedule(hours, minutes, seconds, days_of_week, int_days_of_month)
        if api_call:
            node = c123_node.objects.filter(type = DBType.POSTGRES_K8S.value, name=node_id, ind_deleted = False).first()
        else:
            node = c123_node.objects.filter(type = DBType.POSTGRES_K8S.value, pk=node_id, ind_deleted = False).first()
        if not node:
            raise APIException410("node id {} not available".format(node_id))
        job_id =  f"export_{node.name}"
        scheduler = settings.scheduler
        export_job = next((i for i in scheduler.get_jobs() if job_id in i.id), None)
        if export_job:
            export_job.reschedule(trigger=trigger)
        else:
             settings.scheduler.add_job(export_k8s_pg_node, replace_existing=True, trigger=trigger,  kwargs={"data": data}, id=job_id, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP)
        return JsonResponse({"notification": f"export {node.name} {description} started"}, status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated,))
def ImportPgNode(request):
    """restore (import) a postgresql cluster from dump_all backup
    the call is made from client AJAX POST that contains the export id"""
    export_id = request.data.get('export_id')
    api_call = request.data.get('rest_api')
    app_host = c123_host.objects.filter(ex_params__myself=True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    if api_call:
        export = c123_inventory.objects.filter(name=export_id, type=Inventory.EXPORT.value).first()
    else:
        export = c123_inventory.objects.filter(pk=export_id, type=Inventory.EXPORT.value).first()

    if not export:
        raise APIException404(f"export id {export_id} not found")

    path = export.ex_params.get('backup_path')
    path_exists = checkPathExists(app_host, path)
    if not path_exists:
        raise APIException404(f"{path}: no such file or directory")
    node = export.node
    primary = c123_node.objects.filter(cluster=node.cluster, type=DBType.POSTGRES_K8S.value,
                                       cluster_role=ClusterRoles.PRIMARY.value, ind_deleted=False).first()
    if not primary:
        raise APIException404("primary not found")
    primary.status = Status.RESTORING.value
    primary.save()
    # serialize objects for the tasks
    backup_host = c123_HostSerializer(app_host).data
    export = object_to_json(export)
    node_data = object_to_json(primary)
    host_data = c123_HostSerializer(primary.host).data
    created_by = request.user
    tmp_data = json.dumps({'pod_data': node_data, 'cp_data': host_data, 'backup_host': backup_host,
                           'cluster_data': object_to_json(primary.cluster), 'export': export,
                           'created_by': created_by.id})
    task = Task.objects.create(task='pg_k8s_dumpall_restore_task',
                               displayment=f'restore pg cluster {primary.name}',
                               target_type=DBType.POSTGRES_K8S.value,
                               target_id=primary.id,
                               org_id=primary.org_id,
                               schedule=Schedule.NOW.value,
                               status=TaskState.PENDING.value,
                               parameters=tmp_data,
                               created_by=created_by)
    return append_status_cmd(request, task)

@api_view(['POST'])
@permission_classes((IsAuthenticated,))
def BackupPgNode(request):
    """backup (basebackup) for pg primary"""
    schedule = request.data.get('schedule')
    node_id = request.data.get('node_id')
    compress = request.data.get("compress")
    pg_sys, created = SystemConfig.objects.get_or_create(db_type=DBType.POSTGRES.value)
    backup_path = pg_sys.backups_path
    # incremental wal backup
    incremental = request.data.get('incremental')
    # incremental physical backup
    incremental_physical = request.data.get("incremental_physical")
    update_schedule = request.data.get("update_schedule")
    api_call = request.data.get('rest_api')
    from django.conf import settings
    app_host = c123_host.objects.filter(ex_params__myself=True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    if api_call:
        pg_node = c123_node.objects.filter(type=DBType.POSTGRES_K8S.value, name=node_id, ind_deleted=False).first()
    else:
        pg_node = c123_node.objects.filter(type=DBType.POSTGRES_K8S.value, pk=node_id, ind_deleted=False).first()

    if not pg_node:
        raise APIException410("node id {} not available".format(node_id))

    if incremental_physical and float(pg_node.version) < 17.0:
        raise APIException406(
            f"{pg_node.name} version ({pg_node.version}) must be >= 17 to support physical incremental backups")

    # they are different, if both present - prefer the recommended incremental physical backups
    if incremental and incremental_physical:
        incremental = False

    if schedule:
        client_timezone = pytz.timezone(request.data.get("client_timezone"))
        schedule_run_time_str = request.data.get('schedule_run_time')
        schedule_run_time = parse(schedule_run_time_str).astimezone(client_timezone)
        client_hours = schedule_run_time.hour
        client_minutes = schedule_run_time.minute
        client_seconds = schedule_run_time.second
        days_of_week = request.data.get("days_of_week")
        days_of_month = request.data.get("days_of_month")
        if days_of_month:
            int_days_of_month = [parse(d).astimezone(client_timezone).day for d in days_of_month]
        else:
            int_days_of_month = None
        description = get_backup_description(client_hours, client_minutes, client_seconds, days_of_week,
                                             int_days_of_month)
        localized_datetime = datetime.now(client_timezone).replace(hour=client_hours, minute=client_minutes,
                                                                   second=client_seconds).astimezone(
            pytz.timezone(settings.TIME_ZONE))
        hours = localized_datetime.hour
        minutes = localized_datetime.minute
        seconds = localized_datetime.second
        trigger = get_cron_schedule(hours, minutes, seconds, days_of_week, int_days_of_month)

    if update_schedule and schedule:
        allow_or_403(pg_node)
        backup = c123_inventory.objects.filter(node=pg_node, type=Inventory.BACKUP.value, is_sched=True).last()
        if backup:
            backup.ex_params['description'] = description
            backup.ex_params['schedule_run_time'] = schedule_run_time_str
            backup.ex_params['days_of_month'] = days_of_month
            backup.ex_params['days_of_week'] = days_of_week
            backup.save()
            job_id = backup.ex_params.get("job_id")
            job = settings.scheduler.get_job(job_id=job_id)
            if job:
                job.reschedule(trigger=trigger)
            else:
                settings.scheduler.add_job(schedule_backup, trigger=trigger, replace_existing=True,
                                           misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP,
                                           kwargs={"task_id": backup.ex_params["task_id"]}, id=job_id)
            notification = f"{backup.node} backup updated to '{description}'"
            return JsonResponse({"notification": notification}, status=200)

        else:
            standbys = c123_node.objects.filter(type=DBType.POSTGRES_K8S.value, cluster_role=ClusterRoles.STANDBY.value,
                                                cluster=pg_node.cluster)
            if standbys.exists():
                for standby in standbys:
                    backup = c123_inventory.objects.filter(node=standby.pk, type=Inventory.BACKUP.value,
                                                           is_sched=True).last()
                    if backup:
                        backup.ex_params['schedule_run_time'] = schedule_run_time_str
                        backup.ex_params['days_of_month'] = days_of_month
                        backup.ex_params['days_of_week'] = days_of_week
                        backup.ex_params['description'] = description
                        backup.save()
                        job_id = backup.ex_params.get("job_id")
                        job = settings.scheduler.get_job(job_id=job_id)
                        if job:
                            job.reschedule(trigger=trigger)
                        else:
                            settings.scheduler.add_job(schedule_backup, replace_existing=True, trigger=trigger,
                                                       misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP,
                                                       kwargs={"task_id": backup.ex_params["task_id"]}, id=job_id)

                        notification = f"{backup.node} backup updated to '{description}'"
                        return JsonResponse({"notification": notification}, status=200)

                return JsonResponse({"notification": "no schedule backup found"}, status=200)
            else:
                return JsonResponse({"notification": "no schedule backup found"}, status=200)

    serializer = c123_inventory_serializer(data=request.data, partial=True)
    if serializer.is_valid():
        allow_or_403(pg_node)
        # check_space(node=pg_node, path=backup_path, source_path=pg_node.dir_123Cluster)
        size_scheme, created = ClusterSizing.objects.get_or_create(db_type=DBType.POSTGRES_K8S.value)
        keep_days = getattr(size_scheme, f'{pg_node.sizing.lower()}_backup_keep_days', None)
        log_archive_path = f'/123cluster/archive_logs/{pg_node.type}/{pg_node.cluster.name}'
        # create the task signature
        if schedule:
            existing_backup = c123_inventory.objects.filter(node=pg_node, type=Inventory.BACKUP.value, is_sched=True)
            if existing_backup.exists():
                raise APIException409("Schedule backup already exists.\nCan be only one schedule backup for cluster.")

            if pg_node.cluster_role == ClusterRoles.PRIMARY.value:
                standbys = c123_node.objects.filter(type=DBType.POSTGRES_K8S.value, cluster_role=ClusterRoles.STANDBY.value,
                                                    cluster=pg_node.cluster)
            else:
                standbys = c123_node.objects.filter(type=DBType.POSTGRES_K8S.value, cluster_role=ClusterRoles.STANDBY.value,
                                                    cluster=pg_node.cluster).exclude(pk=pg_node.pk)

            if standbys.exists():
                for standby in standbys:
                    existing_backup = c123_inventory.objects.filter(node=standby, type=Inventory.BACKUP.value,
                                                                    is_sched=True)
                    if existing_backup.exists():
                        raise APIException409(
                            "Schedule backup already exists.\nCan be only one schedule backup for cluster.")
            backup = serializer.create(serializer.validated_data)
            backup.node = pg_node
            if keep_days:
                backup.keep_days = keep_days
            backup.type = Inventory.BACKUP.value
            backup.created_by = request.user
            backup.ex_params = {"size": 0}
            backup.ex_params["belong_to"] = DBType.POSTGRES_K8S.value
            backup.ex_params["node_version"] = pg_node.version
            backup.is_sched = schedule
            backup.ex_params["backup_type"] = "basebackup"
            backup.ex_params["duration"] = 0
            backup.ex_params["compress"] = compress
            backup.ex_params["modified_date"] = ''
            backup.ex_params['description'] = description
            backup.ex_params['schedule_run_time'] = schedule_run_time_str
            backup.ex_params['days_of_month'] = days_of_month
            backup.ex_params['days_of_week'] = days_of_week
            backup.ex_params['incremental'] = incremental
            backup.ex_params['incremental_physical'] = incremental_physical
            notification = "{} backup {}".format(backup.node, description)
            job_id = f"{notification} ({client_timezone})"
            backup.ex_params["job_id"] = job_id
            backup.ex_params[
                "backup_path"] = f'{backup_path}/basebackups/schedule/{pg_node.version}/{pg_node.name}'.replace(":",
                                                                                                                "_")
            backup.ex_params['log_archive_path'] = log_archive_path
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            backup.name = "{}_{}".format(pg_node.name, now)
            backup.save()
            # serialize objects for the tasks
            node_data = object_to_json(pg_node)
            host_data = c123_HostSerializer(pg_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            created_by = request.user

            tmp_data = json.dumps(
                {'pod_data': node_data, 'cp_data': host_data, 'backup_host': backup_host,
                 'cluster_data': object_to_json(pg_node.cluster), 'backup_data': object_to_json(backup),
                 'created_by': created_by.id, 'backup_type': 'basebackup',
                 'compress': compress})
            task = Task.objects.create(task='pg_k8s_basebackup_task',
                                       displayment='backup {}'.format(pg_node.name),
                                       target_type=DBType.POSTGRES_K8S.value,
                                       target_id=pg_node.id,
                                       org_id=pg_node.org_id,
                                       schedule=Schedule.SCHEDULED.value,
                                       status=TaskState.ON_DEMAND.value,
                                       parameters=tmp_data,
                                       created_by=created_by)

            backup.ex_params['task_id'] = task.pk
            backup.save()
            settings.scheduler.add_job(schedule_backup, replace_existing=True, trigger=trigger,
                                       kwargs={"task_id": task.pk}, id=job_id,
                                       misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP)
        else:
            existing_backup = Task.objects.filter(target_id=pg_node.id, task='pg_basebackup_task',
                                                  status__in=[TaskState.PENDING.value, TaskState.RUNNING.value])
            if existing_backup:
                raise APIException409(f"Already runs backup for {pg_node.name}")
            backup = serializer.create(serializer.validated_data)
            backup.node = pg_node
            if keep_days:
                backup.keep_days = keep_days
            backup.type = Inventory.BACKUP.value
            backup.created_by = request.user
            backup.ex_params = {"size": 0}
            backup.ex_params["belong_to"] = DBType.POSTGRES_K8S.value
            backup.ex_params["node_version"] = pg_node.version
            backup.is_sched = schedule
            backup.ex_params["backup_type"] = "basebackup"
            backup.ex_params["duration"] = 0
            backup.ex_params["compress"] = compress
            backup.ex_params["modified_date"] = ''
            backup.ex_params['incremental'] = incremental
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            backup.name = "{}_{}".format(pg_node.name, now)
            backup.ex_params["backup_path"] = f'{backup_path}/basebackups/{pg_node.host.name}_{pg_node.port}/{now}'
            backup.ex_params['log_archive_path'] = log_archive_path
            backup.save()
            notification = f"backup of {pg_node.name}"
            # serialize objects for the tasks
            node_data = object_to_json(pg_node)
            host_data = c123_HostSerializer(pg_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            created_by = request.user
            tmp_data = json.dumps(
                {'pod_data': node_data, 'cp_data': host_data, 'backup_host': backup_host,
                 'cluster_data': object_to_json(pg_node.cluster), 'backup_data': object_to_json(backup),
                 'created_by': created_by.id, 'backup_type': 'basebackup',
                 'compress': compress})
            task = Task.objects.create(task='pg_k8s_basebackup_task',
                                       displayment='backup {}'.format(pg_node.name),
                                       target_type=DBType.POSTGRES_K8S.value,
                                       target_id=pg_node.id,
                                       org_id=pg_node.org_id,
                                       schedule=Schedule.NOW.value,
                                       status=TaskState.PENDING.value,
                                       parameters=tmp_data,
                                       created_by=created_by)

            backup.ex_params['task_id'] = task.pk
            backup.save()
    else:
        print("invalid form")
        print(serializer.errors)
        return JsonResponse({'error': serializer.errors}, status=404)
    return append_status_cmd(request, task, notification)


@api_view(['POST'])
@permission_classes((IsAuthenticated,))
def RestoreBackupPgNode(request):
    """restore a postgresql node from backup
    the call is made from client AJAX POST that contains the backup id"""

    bu_id = request.data.get('bu_id')
    # restore_option = 'this'/'latest'/'custom'
    restore_option = request.data.get('restore_option')
    restore_time = request.data.get('restore_time')
    autostart = request.data.get('autostart')
    api_call = request.data.get('rest_api')
    from django.conf import settings
    # app_host = c123_host.objects.filter(ex_params__myself=True).first()
    # if not app_host:
    #     raise APIException404("could not determine the app host")
    if api_call:
        backup = c123_inventory.objects.filter(name=bu_id, type=Inventory.BACKUP.value).first()
    else:
        backup = c123_inventory.objects.filter(pk=bu_id, type=Inventory.BACKUP.value).first()

    if not backup:
        raise APIException404("backup id {} not found".format(bu_id))

    # path = backup.ex_params.get('backup_path')
    # path_exists = checkPathExists(app_host, path)
    # if not path_exists:
    #     raise APIException404(f"{path}: no such file or directory")

    node = backup.node
    allow_or_403(node)

    forbid_restore = Task.objects.filter(target_id=node.id, task__in=['pg_k8s_pod_restore_task', 'pg_k8s_basebackup_task'],
                                         status__in=[TaskState.PENDING.value, TaskState.RUNNING.value])
    if forbid_restore:
        raise APIException409(f"Already runs backup/restore for {node.name}")

    primary = c123_node.objects.filter(type=DBType.POSTGRES_K8S.value, cluster=node.cluster,
                                       cluster_role=ClusterRoles.PRIMARY.value, ind_deleted=False).first()
    if not primary:
        # because it can not exist, or exist as deleted
        raise APIException410(f"primary of standby {node.name} is not available")

    primary = find_real_pg_primary(primary)
    primary.status = Status.RESTORING.value
    primary.save()
    if RestoreOptions.THIS.value in restore_option:
        # as app time zone (if by a chance representation timezone will differ from internal)
        date = backup.created_date.astimezone(pytz.timezone(settings.TIME_ZONE)).strftime("%Y-%m-%d %H:%M:%S")
    elif RestoreOptions.CUSTOM.value in restore_option and restore_time:
        # example date = '2021-12-30 09:08:33'
        # date = datetime.strptime(restore_time, "%Y-%m-%dT%H:%M:%S.%f%z").astimezone()
        # date = date.strftime("%Y-%m-%d %H:%M:%S")
        date = restore_time
    else:
        date = ''

    standbys_id = []

    # serialize objects for the tasks
    # backup_host = c123_HostSerializer(app_host).data
    backup = object_to_json(backup)
    node_data = object_to_json(primary)
    host_data = c123_HostSerializer(primary.host).data

    created_by = request.user
    standbys_list = []
    standbys_set = c123_node.objects.filter(type=DBType.POSTGRES_K8S.value, cluster_role=ClusterRoles.STANDBY.value,
                                            ind_deleted=False, cluster=node.cluster)
    standbys_exists = standbys_set.exists()
    for standby in standbys_set:
        standbys_list.append({'node': object_to_json(standby), 'host': object_to_json(standby.host)})
        standbys_id.append(standby.id)
    if node.cluster_role == ClusterRoles.PRIMARY.value and c123_node.objects.filter(type=DBType.POSTGRES_K8S.value,
                                                                                    cluster=node.cluster,
                                                                                    ind_deleted=False,
                                                                                    severity=Severity.OK.value,
                                                                                    cluster_role=ClusterRoles.STANDBY.value).exists():
        autostart = True

    tmp_data = json.dumps({'node_data': node_data, 'host_data': host_data, 'backup': backup,
                           'autostart': autostart, 'standbys_exists': standbys_exists, 'standbys': standbys_list,
                           'restore_option': restore_option, 'date': date, 'app_db_timezone': settings.TIME_ZONE,
                           'standbys_id': standbys_id, 'created_by': created_by.id})
    task = Task.objects.create(task='pg_k8s_pod_restore_task',
                               displayment='restore pg node {}'.format(primary.name),
                               target_type=DBType.POSTGRES_K8S.value,
                               target_id=primary.id,
                               org_id=primary.org_id,
                               schedule=Schedule.NOW.value,
                               status=TaskState.PENDING.value,
                               parameters=tmp_data,
                               created_by=created_by)
    return append_status_cmd(request, task)

@api_view(['POST'])
@permission_classes((IsAuthenticated,))
def AddpgDB(request):
    current_user = request.user
    db_name = request.data.get('db_name')
    if not db_name or db_name.isspace():
        raise APIException406("database name cannot be blank")
    db_name = ''.join(db_name.split())
    if db_name.lower() == 'repmgr':
        raise APIException403("name repmgr is restricted")
    node_id = request.data.get('node_id')
    api_call = request.data.get('rest_api')
    if api_call:
        node = c123_node.objects.filter(type=DBType.POSTGRES_K8S.value, name=node_id, ind_deleted=False).first()
    else:
        node = c123_node.objects.filter(type=DBType.POSTGRES_K8S.value, pk=node_id, ind_deleted=False).first()

    if not node:
        raise APIException410("node id {} not available".format(node_id))
    primary = c123_node.objects.filter(cluster=node.cluster, cluster_role=ClusterRoles.PRIMARY.value,
                                       type=DBType.POSTGRES_K8S.value, ind_deleted=False).first()
    if not primary:
        raise APIException404("no found primary for node {}".format(node.name))
    allow_or_403(primary)
    auto_delete = request.data.get('auto_delete')
    auto_delete_days = request.data.get('auto_delete_days')

    existing_dbs = c123_database.objects.filter(name=db_name, node=primary)
    if existing_dbs.exists():
        undeleted = existing_dbs.filter(ind_deleted=False)
        if undeleted.exists():
            db = undeleted.first()
        else:
            db = existing_dbs.first()
            db.ind_deleted = False

        db.created_by = current_user
        db.last_update_by = current_user
        db.status = Status.CREATING.value
        db.database_user = primary.database_user
        db.database_pwd = primary.database_pwd
        db.ind_started = True
        db.type = DBType.POSTGRES_K8S.value
        db.save()
    else:
        db = c123_database.objects.create(name=db_name,
                                          type=DBType.POSTGRES_K8S.value,
                                          created_by=current_user,
                                          last_update_by=current_user,
                                          status=Status.CREATING.value,
                                          severity=Severity.NOT_OK.value,
                                          database_user=primary.database_user,
                                          database_pwd=primary.database_pwd,
                                          ind_started=True,
                                          node=primary)

    db.ex_params = {}
    db.save()

    tmp_data = json.dumps(
        {'node_data': object_to_json(primary),
         'host_data': object_to_json(primary.host),
         'db': object_to_json(db),
         'auto_delete': auto_delete,
         'created_by': current_user.id})
    task = Task.objects.create(task='create_pg_database_task',
                               displayment='creating pg db {}'.format(db_name),
                               target_type=DBType.POSTGRES_K8S.value,
                               target_id=primary.id,
                               org_id=primary.org_id,
                               status=TaskState.PENDING.value,
                               schedule=Schedule.NOW.value,
                               parameters=tmp_data,
                               created_by=current_user)
    return append_status_cmd(request, task)


@api_view(['POST'])
@permission_classes((IsAuthenticated,))
def CheckPgDatabaseName(request):
    primary_id = request.data.get('node_id')
    db_name = request.data.get('name')
    primary = c123_node.objects.filter(pk=primary_id, type=DBType.POSTGRES_K8S.value, ind_deleted=False).first()
    if not primary:
        raise APIException410(f"primary id {primary_id} not available")
    if (re.match(r'^[a-zA-Z0-9_]{1,63}$', db_name)):
        # try to connect to the source
        conn = get_postgres_connection(primary)
        if isinstance(conn, tuple):
            if hasattr(conn[1], 'strerror'):
                raise APIException401(conn[1].strerror)
            raise APIException401(conn[1])

        try:
            with conn:
                with conn.cursor() as cur:
                    cur.execute("SELECT datname FROM pg_catalog.pg_database WHERE datname = %s", (db_name.strip(),))
                    res = cur.fetchone()
                    if res:
                        return JsonResponse({'error': "already exists"}, status=409)
                    return JsonResponse({'response': "name available"}, status=200)
        except Exception as e:
            raise APIException401(e.strerror if hasattr(e, 'strerror') else str(e))

    else:
        return JsonResponse({'error': "invalid name"}, status=406)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def Exportk8sPgDatabase(request):
    """export for database"""

    data = request.data
    data['created_by'] = request.user.id
    schedule = data.get('schedule')
    api_call = data.get('api_call')
    node_id = data.get('node_id')
    db_id = data.get('db_id')

    if not schedule:
        task = export_k8s_pg_database(data)
        return append_status_cmd(request, task)
    else:
        from django.conf import settings
        client_timezone = pytz.timezone(data.get("client_timezone"))
        schedule_run_time_str = data.get('schedule_run_time')
        schedule_run_time = parse(schedule_run_time_str).astimezone(client_timezone)
        client_hours = schedule_run_time.hour
        client_minutes = schedule_run_time.minute
        client_seconds = schedule_run_time.second
        days_of_week = data.get("days_of_week")
        days_of_month = data.get("days_of_month")
        if days_of_month:
            int_days_of_month = [parse(d).astimezone(client_timezone).day for d in days_of_month]
        else:
            int_days_of_month = None
        description = get_backup_description(client_hours, client_minutes, client_seconds, days_of_week, int_days_of_month)
        localized_datetime = datetime.now(client_timezone).replace(hour=client_hours, minute=client_minutes, second=client_seconds).astimezone(pytz.timezone(settings.TIME_ZONE))
        hours = localized_datetime.hour
        minutes = localized_datetime.minute
        seconds = localized_datetime.second
        trigger = get_cron_schedule(hours, minutes, seconds, days_of_week, int_days_of_month)
        if api_call:
            db = c123_database.objects.filter(type = DBType.POSTGRES_K8S.value, name=db_id, node__name = node_id, ind_deleted = False).first()
        else:
            db = c123_database.objects.filter(type = DBType.POSTGRES_K8S.value, pk=db_id, ind_deleted = False).first()
        if not db:
            raise APIException410("database id {} not available".format(db_id))
        node = db.node
        job_id =  f"export_{db.name}_{node.name}"
        scheduler = settings.scheduler
        export_job = next((i for i in scheduler.get_jobs() if job_id in i.id), None)
        if export_job:
            export_job.reschedule(trigger=trigger)
        else:
             settings.scheduler.add_job(export_k8s_pg_database, replace_existing=True, trigger=trigger,  kwargs={"data": data}, id=job_id, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP)
        return JsonResponse({"notification": f"export db {db.name} of {node.name} {description} started"}, status = 200)


@api_view(['POST'])
@permission_classes((IsAuthenticated,))
def RestoreBackupPgDB(request):
    """restore a postgresql db from dump backup
    the call is made from client AJAX POST that contains the backup id"""
    bu_id = request.data.get('bu_id')
    include_data = request.data.get('include_data')
    include_publications = request.data.get('include_publications')
    include_subscriptions = request.data.get('include_subscriptions')
    autostart = True if request.data.get('autostart') == 'true' else False
    api_call = request.data.get('rest_api')
    app_host = c123_host.objects.filter(ex_params__myself=True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    if api_call:
        backup = c123_inventory.objects.filter(name=bu_id, type=Inventory.EXPORT.value).first()
    else:
        backup = c123_inventory.objects.filter(pk=bu_id, type=Inventory.EXPORT.value).first()

    if not backup:
        raise APIException404("export id {} not found".format(bu_id))

    path = backup.ex_params.get('backup_path')
    path_exists = checkK8SPathExists(path)
    if not path_exists:
        raise APIException404(f"{path}: no such file or directory")

    db = backup.database
    node = db.node
    primary = c123_node.objects.filter(cluster=node.cluster, type=DBType.POSTGRES_K8S.value,
                                       cluster_role=ClusterRoles.PRIMARY.value, ind_deleted=False).first()
    if not primary:
        raise APIException404("primary not found")
    kubernetes_api_data = IntegratedService.objects.filter(id=request.data["service_id"]).first()
    allow_or_403(primary)
    db.status = Status.RESTORING.value
    db.save()
    backup_host = c123_HostSerializer(app_host).data
    db_data = object_to_json(db)
    backup = object_to_json(backup)
    node_data = object_to_json(primary)
    created_by = request.user
    host_primary = c123_HostSerializer(primary.host).data

    tmp_data = json.dumps({'db_data': db_data,
                           'node_data': node_data,
                           'backup_host': backup_host,
                           'host_primary': host_primary,
                           'cluster_data': object_to_json(primary.cluster),
                           'kubernetes_api_data': object_to_json(kubernetes_api_data),
                           'backup': backup,
                           'autostart': autostart,
                           'include_data': include_data,
                           'include_publications': include_publications,
                           'include_subscriptions': include_subscriptions,
                           'created_by': created_by.id})
    task = Task.objects.create(task='pg_restore_task',
                               displayment='restore pg database {}'.format(db.name),
                               target_type=DBType.POSTGRES_K8S.value,
                               target_id=db.id,
                               org_id=primary.org_id,
                               schedule=Schedule.NOW.value,
                               status=TaskState.PENDING.value,
                               parameters=tmp_data,
                               created_by=created_by)
    return append_status_cmd(request, task)


@api_view(['POST'])
@permission_classes((IsAuthenticated,))
def DropPgDatabase(request):
    db_id = request.data.get("db_id")
    node_id = request.data.get("node_id")
    api_call = request.data.get('rest_api')
    if api_call:
        db = c123_database.objects.filter(type=DBType.POSTGRES_K8S.value, name=db_id, node__name=node_id,
                                          ind_deleted=False).first()
    else:
        db = c123_database.objects.filter(type=DBType.POSTGRES_K8S.value, pk=db_id, ind_deleted=False).first()
    if not db:
        raise APIException410("database id {} not found".format(db_id))
    node = db.node
    primary = c123_node.objects.filter(cluster=node.cluster, type=DBType.POSTGRES_K8S.value,
                                       cluster_role=ClusterRoles.PRIMARY.value, ind_deleted=False).first()
    kubernetes_api_data = IntegratedService.objects.filter(id=request.data["service_id"]).first()
    if not primary:
        raise APIException404("no found primary for node {}".format(node.name))
    db.status = Status.DROPPING.value
    db.save()
    cleanup_logs(db.pk, db.type)
    node_data = object_to_json(primary)
    db_data = object_to_json(db)
    host_data = c123_HostSerializer(primary.host).data
    created_by = request.user

    tmp_data = json.dumps(
        {'node_data': node_data,
         'cluster_data': object_to_json(primary.cluster),
         'kubernetes_api_data': object_to_json(kubernetes_api_data),
         'db_data': db_data,
         'host_data': host_data,
         'created_by': created_by.id})
    task = Task.objects.create(task='drop_pg_database_task',
                               displayment='drop database {}'.format(db.name),
                               target_type=DBType.POSTGRES_K8S.value,
                               target_id=db.pk,
                               org_id=primary.org_id,
                               schedule=Schedule.NOW.value,
                               status=TaskState.PENDING.value,
                               parameters=tmp_data,
                               created_by=created_by)

    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated,))
def CheckPgDatabaseSize(request):
    node_id = request.data.get('node_id')
    db_name = request.data.get("db_name")
    node = c123_node.objects.filter(pk=node_id, type=DBType.POSTGRES_K8S.value, ind_deleted=False).first()
    if not node:
        raise APIException410(f"node id {node_id} not available")

    conn = get_postgres_connection(node)
    if isinstance(conn, tuple):
        raise APIException401(conn[1].strerror if hasattr(conn[1], 'strerror') else str(conn[1]))
    try:
        with conn:
            with conn.cursor() as cur:
                if db_name:
                    cur.execute("SELECT pg_database_size(%s)/1024", (db_name.strip(),))
                else:
                    cur.execute("SELECT SUM(pg_catalog.pg_database_size(d.datname))/1024 FROM pg_catalog.pg_database d")
                res = cur.fetchone()
                db_size = res[0] if res and res[0] else 0.0

    except Exception as e:
        raise APIException401(e.strerror if hasattr(e, 'strerror') else str(e))

    # client = create_ssh_connection(node.host.name, node.host.username, node.host.password, node.host.private_key,
    #                                node.host.ssh_port)
    # if isinstance(client, tuple):
    #     if hasattr(client[1], 'strerror'):
    #         raise APIException401(client[1].strerror)
    #     raise APIException401(client[1])

    command = 'sudo df -P . | tail -1 | awk \'{print $4}\''
    jsn = json.loads(local_command(command, True), strict=False)
    if (not jsn['stderr'] == ''):
        print(jsn['stderr'])
        return JsonResponse({'detail': jsn['stderr']}, status=404)

    free_space = float(jsn['stdout'])

    return JsonResponse({"size": db_size, "display_size": format_size(db_size), 'free_space': free_space}, status=200)
