from enum import Enum
from rest_framework.permissions import IsAuthenticated
from django.http import JsonResponse
from django.utils import timezone
from rest_framework.decorators import api_view, permission_classes
from clusterCore.models import CustomerKey, c123_node, c123_cluster
from clusterCore.exceptions import *
from clusterCore.tasks import license_db_type, parse_license_key, generate_license_summary, revoke_license
from dateutil.relativedelta import relativedelta
from clusterCore.enums import DBType, LicenseType
from clusterCore.api.serializer import c123_NodeSerializer

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def AddLicense(request):
    current_trial_start = timezone.now() - relativedelta(months=2)
    key = request.data.get('key')
    try:
        CustomerKey.objects.create(**parse_license_key(key, False))
        summary = generate_license_summary()
        targets = c123_node.objects.filter(ind_deleted = False, cluster__isnull = False).exclude(type = DBType.HAPROXY)
        if not targets.exists():
            return JsonResponse({}, status=201)
        no_touch_licensing = []
        licensing_queries = []
        licensing_summary = [summary.oracle, summary.mariadb, summary.mysql, summary.mssql, summary.mongo, summary.redis, summary.postgresql, summary.elasticsearch]
        total_licenses = sum(licensing_summary)
        targets_filter = [
            LicenseType.ORACLE.value,
            LicenseType.MARIADB.value,
            LicenseType.MYSQL.value,
            LicenseType.MSSQL.value,
            LicenseType.MONGO.value,
            LicenseType.REDIS.value,
            LicenseType.POSTGRES.value,
            LicenseType.ELASTIC.value
        ]
        all_cluster_ids = targets.values_list('cluster__id', flat=True).distinct()
        for db_type, summary_count in zip(targets_filter, licensing_summary):
             license_db_type(targets, db_type, current_trial_start, all_cluster_ids, licensing_queries, no_touch_licensing, summary_count)
        # Perform union of all queries
        licensing = c123_node.objects.none().union(*licensing_queries).values_list("id", flat=True)
        c123_node.objects.filter(id__in=licensing).update(ind_license = True) 
        unlicensed = c123_node.objects.filter(ind_license = False, ind_deleted = False, cluster__isnull = False).exclude(type = DBType.HAPROXY)
        if unlicensed.exists():
            clusters = unlicensed.values_list("cluster__id", flat=True).distinct()
            unlicensed_clusters = ""
            for cluster in clusters:
                unlicensed_per_cluster = unlicensed.filter(cluster__id = cluster)
                cluster_names = ',\n'.join(unlicensed_per_cluster.values_list('name', flat=True))
                unlicensed_clusters = f"{unlicensed_clusters}\nCluster {cluster}:\n{cluster_names}"
            return JsonResponse({"notification": f"Apologies, there aren't enough licenses available to allocate to all nodes.\nThe licenses available can only cover licenses for up to {total_licenses} nodes ({summary.elasticsearch} Elasticsearch, {summary.mariadb} MariaDB, {summary.mysql} MySQL, {summary.mongo} Mongo, {summary.mssql} MsSQL, {summary.oracle} Oracle, {summary.postgresql} PostgreSQL, {summary.redis} Redis).\n Nodes without available licenses:\n {unlicensed_clusters}"}, status = 206)
    except Exception as e:
        raise APIException500(str(e))
    return JsonResponse({}, status = 201)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def RemoveLicense(request):
    key = request.data.get('key')
    CustomerKey.objects.delete(key = key)
    generate_license_summary()
    return JsonResponse({}, status = 204)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def GetLicensedClusters(request):
    type = request.data.get("type")
    cluster_id = request.data.get("cluster_id")
    source_clusters = c123_node.objects.filter(type = type, ind_deleted = False, cluster__isnull = False, ind_license = True).exclude(cluster__id = cluster_id).distinct()
    if not source_clusters.exists():
        raise APIException404("No clusters are available to move license from")
    return JsonResponse({"clusters": c123_NodeSerializer(source_clusters, many = True).data}, status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def MoveLicense(request):
    source_id = request.data.get("source_id")
    target_id = request.data.get("target_id")
    target = c123_cluster.objects.filter(pk = target_id).first()
    if not target:
        raise APIException404(f"target cluster {target_id} has not found")
    
    source = c123_cluster.objects.filter(pk = source_id).first()
    if not source:
        raise APIException404(f"source cluster {source_id} has not found")
    
    revoke_license(source, target, cluster = True)
    
    return JsonResponse({"notification": f"moved license from {source.name} to {target.name}"}, status = 200)
