from rest_framework.permissions import IsAuthenticated
import json
from django.http import JsonResponse
from rest_framework.decorators import api_view, permission_classes
from django.db import connection
from django.db.models import Q
from clusterCore.models import c123_node, c123_host
from clusterCore.enums import DBType, ClusterRoles, OperatingSystem, MonitorCommand
from clusterCore.tasks import fetch_data
from psycopg2.extras import RealDictCursor
from clusterCore.api.serializer import c123_dashboardSerializer

@api_view(['GET'])
@permission_classes((IsAuthenticated, ))
def GetMessage(request, id):
    try:
        with connection.connection as real_conn:
            with real_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                cursor.callproc("app_123monitor.get_text", [id])
                message = cursor.fetchone()['get_text']
    except Exception as e:
        error, = e.args
        print("Get Message: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

    return JsonResponse({"message": message}, status = 200)
@api_view(['DELETE'])
@permission_classes((IsAuthenticated, ))
def DeleteAgent(request, id):
    try:
        with connection.cursor() as cursor:
            cursor.callproc("app_123monitor.delete_agent", [id])
    except Exception as e:
        error, = e.args
        print("DeleteAgent: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

    return JsonResponse({}, status = 204)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def GetMonitors(request):
    org_id = request.data.get("org_id")
    node_id = request.data.get("node_id")
    node_type = request.data.get("node_type")
    monitor_name = request.data.get("monitor_name")
    real_conn = connection.connection
    monitors = []

    try:
        with connection.connection as real_conn:
            with real_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                # Get the monitor rows

                query = 'SELECT app_123monitor.get_monitors(%s,%s,%s,%s);FETCH ALL IN "get_monitors_cursor";'
                for monitor in fetch_data(query, cursor, [org_id, node_id, node_type, monitor_name]):
                    monitor['columns'] = []
                    # Get the monitor columns
                    col_query = 'SELECT app_123monitor.get_monitor_columns(%s,%s,%s);FETCH ALL IN "get_monitor_columns_cursor";'
                    with real_conn.cursor(cursor_factory = RealDictCursor) as col_cursor:
                        for col in fetch_data(col_query, col_cursor, [org_id, monitor["monitor_name"], monitor["db_type"]]):
                            if col["column_type"] == "numeric":
                                monitor["red_rule"] = monitor["red_rule"].replace(f'{col["column_name"]}::numeric', col["display_name"])
                                monitor["yellow_rule"] = monitor["yellow_rule"].replace(f'{col["column_name"]}::numeric', col["display_name"])
                            else:
                                monitor["red_rule"] = monitor["red_rule"].replace(col["column_name"], col["display_name"])
                                monitor["yellow_rule"] = monitor["yellow_rule"].replace(col["column_name"], col["display_name"])
                            monitor["columns"].append(col)
                            ret_cursor = real_conn.cursor('get_monitor_columns_cursor')
                            ret_cursor.close()
                    monitors.append(monitor)
    except Exception as e:
        error, = e.args
        print("Get Monitors: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

    return JsonResponse({"monitors": monitors}, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def GetMonitorChart(request):
    org_id = request.data.get("org_id")
    node_id = request.data.get("node_id")
    monitor_name = request.data.get("monitor_name")
    column_display_name = request.data.get("column_display_name")
    node_type = request.data.get("node_type")
    time_period = request.data.get("time_period")
    response = {}
    try:
        import psycopg2
        from django.conf import settings
        db = settings.DATABASES['default']
        with psycopg2.connect(database = db['NAME'], user = db['USER'], password = db['PASSWORD'], host = db['HOST'], port = db['PORT']) as psy_conn:
            with psy_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                cursor.execute(
                    """
                    SELECT severity
                    FROM app_123monitor.rp_alert
                    WHERE monitor_name = %s AND node_id = %s AND node_type = %s
                    ORDER BY date_created DESC
                    LIMIT(1);
                    """,
                    [monitor_name, node_id, node_type]
                )
                severity = cursor.fetchone()['severity']
                points = list(fetch_data( 'SELECT app_123monitor.get_graph_details(%s,%s,%s,%s,%s,%s);FETCH ALL IN "get_graph_details_cursor";', cursor, [org_id, node_id, monitor_name, node_type, time_period, column_display_name]))
                response = {"severity": severity, "points": points, "caller": request.user.id}
    except Exception as e:
        error, = e.args
        print("Get Monitor Chart: ", str(error))
        return JsonResponse({'detail':str(error)}, status=403)

    return JsonResponse(response, status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated,))
def GetDashboardMonitors(request):
    user_id = request.user.id

    try:
        from clusterCore.models import c123_dashboard_item
        monitors = c123_dashboardSerializer(c123_dashboard_item.objects.filter(created_by = request.user).order_by('order_id'), many = True).data

        return JsonResponse(monitors, safe = False , status=200)

    except Exception as e:
        error_message = str(e)
        print("GetUserMonitors Error: ", error_message)
        return JsonResponse({'detail': error_message}, status=500)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def ReorderDashboardMonitors(request):
    """
    Updates the order of a dashboard monitors
    """
    monitor_id = request.data.get("monitor_id")
    new_order_id = request.data.get("order_id")
    response = {}

    if not monitor_id or not new_order_id:
        return JsonResponse({'detail': 'Missing monitor_id or order_id'}, status=400)

    try:
        import psycopg2
        from django.conf import settings
        db = settings.DATABASES['default']
        with psycopg2.connect(
            database=db['NAME'],
            user=db['USER'],
            password=db['PASSWORD'],
            host=db['HOST'],
            port=db['PORT']
        ) as psy_conn:
            with psy_conn.cursor(cursor_factory=RealDictCursor) as cursor:
                cursor.execute("BEGIN;")
                cursor.execute(
                    """
                    SELECT order_id, created_by_id
                    FROM dashboard_item
                    WHERE monitor_id = %s;
                    """,
                    [monitor_id]
                )
                monitor = cursor.fetchone()

                if not monitor:
                    return JsonResponse({'detail': 'Monitor not found'}, status=404)

                current_order_id = monitor['order_id']
                created_by_id = monitor['created_by_id']

                if new_order_id > current_order_id:
                    # Shift order IDs down
                    cursor.execute(
                        """
                        UPDATE dashboard_item
                        SET order_id = order_id - 1
                        WHERE created_by_id = %s
                        AND order_id > %s AND order_id <= %s;
                        """,
                        [created_by_id, current_order_id, new_order_id]
                    )
                elif new_order_id < current_order_id:
                    # Shift order IDs up
                    cursor.execute(
                        """
                        UPDATE dashboard_item
                        SET order_id = order_id + 1
                        WHERE created_by_id = %s
                        AND order_id >= %s AND order_id < %s;
                        """,
                        [created_by_id, new_order_id, current_order_id]
                    )

                cursor.execute(
                    """
                    UPDATE dashboard_item
                    SET order_id = %s
                    WHERE monitor_id = %s;
                    """,
                    [new_order_id, monitor_id]
                )

                psy_conn.commit()

                response = {
                    'detail': 'Monitor order updated successfully',
                    'monitor_id': monitor_id,
                    'new_order_id': new_order_id
                }
                return JsonResponse(response, status=200)

    except Exception as e:
        error_message = str(e)
        print("ReorderDashboardMonitors Error: ", error_message)
        return JsonResponse({'detail': error_message}, status=500)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def AddDashboardMonitor(request):
    created_by_id = request.data.get("created_by")
    node_id = request.data.get("node_id")
    visualization_type = request.data.get("visualization_type")
    monitor_name = request.data.get("monitor_name")
    node_type = request.data.get("node_type")
    column_display_name = request.data.get("column_display_name")

    if not all([created_by_id, node_id, visualization_type, monitor_name]):
        return JsonResponse({'detail': 'All fields are required'}, status=400)

    try:
        from clusterCore.models import c123_dashboard_item
        new_monitor = c123_dashboard_item.objects.create(created_by = request.user,
                                                         node_id = c123_node.objects.filter(pk=node_id).first(),
                                                         visualization_type = visualization_type,
                                                         monitor_name = monitor_name,
                                                         node_type = node_type,
                                                         column_display_name = column_display_name)
        return JsonResponse({'detail': 'Monitor inserted successfully', 'monitor': c123_dashboardSerializer(new_monitor).data}, status=201)

    except Exception as e:
        error_message = str(e)
        print("InsertDashboardMonitor Error: ", error_message)
        return JsonResponse({'detail': error_message}, status=500)

    return JsonResponse({}, status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DeleteDashboardMonitor(request):
    monitor_id = request.data.get("monitor_id")

    if not monitor_id:
        return JsonResponse({'detail': 'Monitor ID is required'}, status=400)

    try:
        import psycopg2
        from django.conf import settings
        db = settings.DATABASES['default']
        with psycopg2.connect(
            database=db['NAME'],
            user=db['USER'],
            password=db['PASSWORD'],
            host=db['HOST'],
            port=db['PORT']
        ) as psy_conn:
            with psy_conn.cursor(cursor_factory=RealDictCursor) as cursor:
                cursor.execute(
                    """
                    SELECT monitor_id, order_id, created_by_id
                    FROM dashboard_item
                    WHERE monitor_id = %s;
                    """,
                    [monitor_id]
                )
                monitor = cursor.fetchone()

                if not monitor:
                    return JsonResponse({'detail': 'Monitor not found'}, status=404)

                order_id = monitor['order_id']
                created_by_id = monitor['created_by_id']

                cursor.execute(
                    """
                    DELETE FROM dashboard_item
                    WHERE monitor_id = %s;
                    """,
                    [monitor_id]
                )

                # Decrement order_id of items with a higher order_id
                cursor.execute(
                    """
                    UPDATE dashboard_item
                    SET order_id = order_id - 1
                    WHERE created_by_id = %s
                    AND order_id > %s;
                    """,
                    [created_by_id, order_id]
                )
                psy_conn.commit()

                return JsonResponse({'detail': 'Monitor deleted successfully', 'monitor_id': monitor_id}, status=200)

    except Exception as e:
        error_message = str(e)
        print("DeleteDashboardMonitor Error: ", error_message)
        return JsonResponse({'detail': error_message}, status=500)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def GetStatusCharts(request):
    host_id = request.data.get("host_id")
    time_period = request.data.get("time_period")
    col_name = request.data.get("col_name")
    org_id = 1
    monitor_name = "linux performance counters"
    node_type = OperatingSystem.LINUX.value
    response = []
    try:
        import psycopg2
        from django.conf import settings
        db = settings.DATABASES['default']
        with psycopg2.connect(database = db['NAME'], user = db['USER'], password = db['PASSWORD'], host = db['HOST'], port = db['PORT']) as psy_conn:
            with psy_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                response = list(fetch_data('SELECT app_123monitor.get_graph_details(%s,%s,%s,%s,%s,%s);FETCH ALL IN "get_graph_details_cursor";', cursor, [org_id, host_id, monitor_name, node_type, time_period, col_name]))
                ret_cursor = psy_conn.cursor("get_graph_details_cursor")
                ret_cursor.close()
    except Exception as e:
        error, = e.args
        print("GetStatusCharts: ", str(error))
        return JsonResponse({'detail':str(error)}, status=403)

    return JsonResponse(response, status = 200, safe=False)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def HostsCpuUsages(request):
    cpu_agg = request.data.get('cpu_agg')
    hosts = c123_host.objects.all()
    data = ''
    match cpu_agg:
        case 'minimum':
            data = 'min'
        case 'maximum':
            data = 'max'
        case 'average':
            data = 'avg'

    hosts_min_cpu_usages = {}

    try:
        with connection.cursor() as cursor:
            for host in hosts:
                hosts_min_cpu_usages[host.pk] = []
                for row in fetch_data(f"SELECT column2 FROM app_123monitor.rp_alert_detail_{data}_day CROSS JOIN (SELECT MAX(alert_detail_id) AS max_alert_detail_id FROM app_123monitor.rp_alert_detail_{data}_day) max_id WHERE alert_id IN (SELECT alert_id FROM app_123monitor.rp_alert_day d WHERE node_id={host.pk} AND node_type='{OperatingSystem.LINUX.value}' AND monitor_name='linux performance counters') AND column1 = 'cpuusage' ORDER BY alert_detail_id DESC;", cursor):
                    hosts_min_cpu_usages[host.pk].append(float(row[0]))
    except Exception as e:
        error, = e.args
        print("Hosts Cpu Usages: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

    return JsonResponse(hosts_min_cpu_usages, status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def GetNodeMonitorNames(request):
    node_id = request.data.get('node_id')
    node_type = request.data.get('node_type')
    node_monitor_names = []
    try:
        node_monitor_names = []
        with connection.cursor() as cursor:
            for row in fetch_data("SELECT monitor_name FROM app_123monitor.rp_monitors WHERE node_id = %s AND db_type = %s", cursor,  [node_id, node_type]):
                node_monitor_names.append(row[0])
    except Exception as e:
        error, = e.args
        print("Node Monitor Names: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

    return JsonResponse(node_monitor_names, status = 200, safe=False)
@api_view(['GET'])
@permission_classes((IsAuthenticated, ))
def GetNodeMonitorsTable(request, type):
    node_monitors_table = []
    if type == "line_chart":
        query = f''' SELECT
    *,
    ROW_NUMBER() OVER () AS index
FROM (
    SELECT
        CASE
            WHEN mc.node_type LIKE 'ORACLE%' THEN n."name" || '@' || (SELECT "name" FROM "clusterCore_c123_host" WHERE id = n.host_id) || ':' || n.port
            ELSE n.name
        END AS "name",
        mc.monitor_name,
        mc.display_name,
        mc.node_type,
        n.id
    FROM
        app_123monitor.rp_monitor_columns mc,
        "clusterCore_c123_node" n
    WHERE
        mc.org_id = 1
        AND mc.column_type = 'numeric'
        AND n.ind_deleted = 'false'
        AND mc.node_type = n."type"
        AND mc.node_type != '{OperatingSystem.LINUX.value}'
    UNION
    SELECT
        CASE
            WHEN mc.node_type = '{OperatingSystem.LINUX.value}' THEN h.name
        END AS "name",
        mc.monitor_name,
        mc.display_name,
        mc.node_type,
        h.id
    FROM
        app_123monitor.rp_monitor_columns mc,
        "clusterCore_c123_host" h
    WHERE
        mc.org_id = 1
        AND mc.column_type = 'numeric'
        AND h.ind_deleted = 'false'
        AND mc.node_type = '{OperatingSystem.LINUX.value}'
) AS subquery
ORDER BY
    4,
    2,
    1;'''
    else:
        query = f'''SELECT
    *,
    ROW_NUMBER() OVER () AS index
FROM (
    SELECT
        CASE
            WHEN m.db_type LIKE 'ORACLE%' THEN n."name" || '@' || (SELECT "name" FROM "clusterCore_c123_host" WHERE id = n.host_id) || ':' || n.port
            ELSE n.name
        END AS "name",
        m.monitor_name,
        m.db_type,
        n.id
    FROM
        app_123monitor.rp_monitors m,
        "clusterCore_c123_node" n
    WHERE
        m.org_id = 1
        AND n.ind_deleted = 'false'
        AND m.db_type = n."type"
        AND m.db_type != '{OperatingSystem.LINUX.value}'
    UNION
    SELECT
        CASE
            WHEN m.db_type = '{OperatingSystem.LINUX.value}' THEN h.name
        END AS "name",
        m.monitor_name,
        m.db_type,
        h.id
    FROM
        app_123monitor.rp_monitors m,
        "clusterCore_c123_host" h
    WHERE
        m.org_id = 1
        AND h.ind_deleted = 'false'
        AND m.db_type = '{OperatingSystem.LINUX.value}'
) AS subquery
ORDER BY
    4,
    2,
    1;'''

    try:
        with connection.connection as real_conn:
            with real_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                node_monitors_table = list (fetch_data(query, cursor))
    except Exception as e:
        error, = e.args
        print("Node Monitor Names: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

    return JsonResponse(node_monitors_table, status = 200, safe=False)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DeleteMonitor(request):
    org_id = request.data.get("org_id")
    node_id = request.data.get("node_id")
    monitor_name = request.data.get("monitor_name")
    monitor_type = request.data.get("type")

    try:
        with connection.cursor() as cursor:
            cursor.execute("SELECT app_123monitor.delete_monitor(%s, %s, %s, %s);", [org_id, node_id, monitor_name, monitor_type])
    except Exception as e:
        error, = e.args
        print("Delete Monitor: ", str(error))
        return JsonResponse({'detail':str(error)}, status=403)

    return JsonResponse({}, status = 204)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def GetAlertHistory(request):
    org_id = request.data.get("org_id")
    node_id = request.data.get("node_id")
    monitor_name = request.data.get("monitor_name")
    ind_interval = request.data.get("ind_interval")
    node_type = request.data.get('node_type')
    try:
         with connection.connection as real_conn:
            with real_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                    alert_history = list(fetch_data('SELECT app_123monitor.get_alerts_with_details(%s, %s, %s, %s, %s);FETCH ALL IN get_alerts_with_details_cursor', cursor, [node_id, monitor_name, node_type, ind_interval, org_id]))
    except Exception as e:
        error, = e.args
        print("Get Alert History: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

    return JsonResponse({"alert_history": alert_history}, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def GetLastNodeMonitorAlert(request):
    org_id = request.data.get("org_id")
    node_id = request.data.get("node_id")
    monitor_name = request.data.get("monitor_name")
    node_type = request.data.get('node_type')
    response = {}
    try:
        with connection.connection as real_conn:
            with real_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                cursor.execute("SELECT *, date_created AT TIME ZONE current_setting('TIMEZONE') AS date_created_with_tz FROM app_123monitor.rp_alert WHERE node_id = %s AND monitor_name = %s AND node_type = %s ORDER BY alert_id DESC LIMIT 1;", [node_id, monitor_name, node_type])
                node_monitor_last_alert = cursor.fetchone()

                if node_monitor_last_alert:
                    node_monitor_last_alert["details"] = list(fetch_data('SELECT app_123monitor.get_alert_details(%s, %s, %s,%s, %s);FETCH ALL IN get_alert_details_cursor', cursor, [org_id, None, monitor_name, node_monitor_last_alert.get('alert_id'), node_type]))
                    ret_cursor = real_conn.cursor('get_alert_details_cursor')
                    ret_cursor.close()
                    response = {"caller": request.user.pk, "alert": node_monitor_last_alert}
                else:
                    return JsonResponse({"caller": request.user.pk, "alert": 'no content'}, status = 204)
    except Exception as e:
        error, = e.args
        print("Get Last Node Monitor Alert: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def GetNodeAlertsSeverityCounts(request):
    node_id = request.data.get("node_id")
    response = {}

    if node_id is None:
        return JsonResponse({'detail': 'node_id is required.'}, status=400)
    try:
        node_id = int(node_id)
    except (ValueError, TypeError):
        return JsonResponse({'detail': 'node_id must be a valid integer.'}, status=400)

    try:
        with connection.connection as real_conn:
            with real_conn.cursor(cursor_factory=RealDictCursor) as cursor:
                cursor.execute("SELECT * FROM app_123monitor.get_node_alert_severity_counts(%s);", [node_id])

                severity_counts = cursor.fetchone()

                if severity_counts:
                    if severity_counts.get('severity_2_count') == -1:
                         print(f"Function get_node_alert_severity_counts reported an error for node_id: {node_id}")
                         return JsonResponse({'detail': 'Error retrieving severity counts from database function.'}, status=500)

                    response = {
                        "node_id": node_id,
                        "severity_counts": {
                            "severity_2": severity_counts.get('severity_2_count', 0),
                            "severity_3": severity_counts.get('severity_3_count', 0)
                        }
                    }
                    return JsonResponse(response, status=200)
                else:
                    print(f"Function get_node_alert_severity_counts did not return data for node_id: {node_id}")
                    return JsonResponse({'detail': 'Failed to retrieve severity counts.'}, status=500)

    except Exception as e:
        error_message = str(e)
        print(f"Error in GetNodeAlertsSeverityCounts API for node_id {node_id}: {error_message}")
        return JsonResponse({'detail': f'An error occurred: {error_message}'}, status=500)

    return JsonResponse(response, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def GetMonitorChartColumns(request):
    org_id = request.data.get("org_id")
    monitor_name = request.data.get("monitor_name")
    node_type = request.data.get('node_type')
    try:
        with connection.cursor() as cursor:
            cursor.execute("select display_name from app_123monitor.rp_monitor_columns where org_id=%s and monitor_name=%s and node_type=%s and column_type=%s;", [org_id, monitor_name, node_type, 'numeric'])
            columns_raw = cursor.fetchall()
            columns = [i[0] for i in columns_raw]
    except Exception as e:
        error, = e.args
        print("Get Monitor Chart Columns: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

    return JsonResponse(columns, safe=False, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def AddMonitor(request):
    parameters = request.data
    org_id = parameters.get("org_id")
    node_id = parameters.get("node_id")
    monitor_type = parameters.get("db_type")
    monitor_name = parameters.get("monitor_name")
    cmd_type = parameters.get("cmd_type")
    if OperatingSystem.LINUX.value in monitor_type and  MonitorCommand.SQL.value in cmd_type:
        cmd_type = MonitorCommand.PYTHON.value
    try:
        with connection.cursor() as cursor:
            cursor.execute("SELECT app_123monitor.check_monitor_name_exists(%s, %s, %s, %s);", [org_id, node_id, monitor_name, monitor_type])
            exists = cursor.fetchone()[0]
            if 'exists' in exists.lower():
                return JsonResponse({'detail': f'monitor {monitor_name} already exists'}, status=409)

            parameter_list = [
                    org_id,
                    node_id,
                    monitor_type,
                    monitor_name,
                    parameters.get("monitor_description"),
                    parameters.get("pre_script"),
                    parameters.get("monitor_script"),
                    parameters.get("ind_enabled"),
                    parameters.get("red_rule"),
                    parameters.get("yellow_rule"),
                    parameters.get("red_rule_message"),
                    parameters.get("yellow_rule_message"),
                    parameters.get("ind_send_email_notifications"),
                    parameters.get("monitor_interval"),
                    parameters.get("monitor_module"),
                    parameters.get("runtime_red_threshold"),
                    parameters.get("runtime_yellow_threshold"),
                    parameters.get("return_code_threshold"),
                    parameters.get("ind_status_graph"),
                    parameters.get("ind_create_timeline"),
                    parameters.get("ind_timeline_full_impact"),
                    cmd_type,
                    parameters.get("ind_first_column_unique"),
                    parameters.get("ind_alert_on_added_missing"),
                    parameters.get("alert_on_added_missing_severity"),
                    parameters.get("ind_alert_compare_to_previous_check"),
                    parameters.get("alert_compare_to_previous_check_severity"),
                    parameters.get("alert_compare_to_previous_check_higher_lower"),
                    parameters.get("alert_compare_to_previous_check_percent"),
                    parameters.get("ind_break_long_running"),
                    parameters.get("ind_compute_baseline"),
                    parameters.get("data_reduction_function"),
                    parameters.get("renew_alert_every"),
                    parameters.get("ind_red_by_row_count"),
                    parameters.get("red_by_row_count_option"),
                    parameters.get("red_by_row_count_items"),
                    parameters.get("ind_yellow_by_row_count"),
                    parameters.get("yellow_by_row_count_option"),
                    parameters.get("yellow_by_row_count_items"),
                    parameters.get("graph_aggregation"),
                    parameters.get("graph_aggregation_column"),
                    json.dumps(parameters.get("columns")),
                ]

            match monitor_type:
                case DBType.ORACLE_CDB.value:
                    target_objects = c123_node.objects.filter(type = DBType.ORACLE_CDB.value, ind_deleted = False)
                case DBType.ORACLE_PDB.value:
                    target_objects = c123_node.objects.filter(type = DBType.ORACLE_PDB.value, ind_deleted = False)
                case DBType.POSTGRES.value:
                    target_objects = c123_node.objects.filter(type = DBType.POSTGRES.value, ind_deleted = False)
                case DBType.MARIADB.value:
                    target_objects = c123_node.objects.filter(type = DBType.MARIADB.value, ind_deleted = False)
                case DBType.MONGO.value:
                    target_objects = c123_node.objects.filter(type = DBType.MONGO.value, ind_deleted = False).exclude(cluster_role = ClusterRoles.ARBITER.value)
                case DBType.MYSQL.value:
                    target_objects = c123_node.objects.filter(type = DBType.MYSQL.value, ind_deleted = False)
                case DBType.ELASTIC.value:
                    target_objects = c123_node.objects.filter(type = DBType.ELASTIC, ind_deleted = False)
                case DBType.REDIS.value:
                    target_objects = c123_node.objects.filter(type = DBType.REDIS.value, ind_deleted = False)
                case OperatingSystem.LINUX.value:
                    target_objects = c123_host.objects.filter(ind_deleted = False)
                case _:
                    target_objects = c123_node.objects.filter(type = monitor_type, ind_deleted = False)

            cmd_monitor = "SELECT app_123monitor.insert_monitor(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)"

            if monitor_type not in [OperatingSystem.LINUX.value, OperatingSystem.WINDOWS.value]:
                cmd_monitor_obj = "SELECT app_123monitor.insert_monitor_node(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)"
            else:
                cmd_monitor_obj = "SELECT app_123monitor.insert_monitor_host(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)"

            parameter_list[1] = None
            cursor.execute(cmd_monitor, parameter_list)
            print(cursor.query)

            for target in target_objects:
                try:
                    parameter_list[1] = target.pk
                    cursor.execute(cmd_monitor_obj, parameter_list)
                except Exception as e:
                    error, = e.args
                    print(f"Add Monitor: failed to add monitor for object {target.name} - {error}", )
                    continue

    except Exception as error:
        print("Add Monitor: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

    return JsonResponse({}, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def UpdateMonitor(request):

    try:

        parameters = request.data
        monitor_type = parameters.get("db_type")
        cmd_type = parameters.get("cmd_type")
        if OperatingSystem.LINUX.value in monitor_type and MonitorCommand.SQL.value in cmd_type:
            cmd_type = MonitorCommand.PYTHON.value

        parameter_list = [
            parameters.get("org_id"),
            parameters.get("node_id"),
            monitor_type,
            parameters.get("monitor_name"),
            parameters.get("monitor_description"),
            parameters.get("pre_script"),
            parameters.get("monitor_script"),
            parameters.get("ind_enabled"),
            parameters.get("red_rule"),
            parameters.get("yellow_rule"),
            parameters.get("red_rule_message"),
            parameters.get("yellow_rule_message"),
            parameters.get("ind_send_email_notifications"),
            parameters.get("monitor_interval"),
            parameters.get("monitor_module"),
            parameters.get("runtime_red_threshold"),
            parameters.get("runtime_yellow_threshold"),
            parameters.get("return_code_threshold"),
            parameters.get("ind_status_graph"),
            parameters.get("ind_create_timeline"),
            parameters.get("ind_timeline_full_impact"),
            cmd_type,
            parameters.get("ind_first_column_unique"),
            parameters.get("ind_alert_on_added_missing"),
            parameters.get("alert_on_added_missing_severity"),
            parameters.get("ind_alert_compare_to_previous_check"),
            parameters.get("alert_compare_to_previous_check_severity"),
            parameters.get("alert_compare_to_previous_check_higher_lower"),
            parameters.get("alert_compare_to_previous_check_percent"),
            parameters.get("ind_break_long_running"),
            parameters.get("ind_compute_baseline"),
            parameters.get("data_reduction_function"),
            parameters.get("renew_alert_every"),
            parameters.get("ind_red_by_row_count"),
            parameters.get("red_by_row_count_option"),
            parameters.get("red_by_row_count_items"),
            parameters.get("ind_yellow_by_row_count"),
            parameters.get("yellow_by_row_count_option"),
            parameters.get("yellow_by_row_count_items"),
            parameters.get("graph_aggregation"),
            parameters.get("graph_aggregation_column"),
            json.dumps(parameters.get("columns")),
        ]
        match monitor_type:
            case DBType.ORACLE_CDB.value:
                target_objects = c123_node.objects.filter(type = DBType.ORACLE_CDB.value, ind_deleted = False)
            case DBType.ORACLE_PDB.value:
                target_objects = c123_node.objects.filter(type = DBType.ORACLE_PDB.value, ind_deleted = False)
            case DBType.POSTGRES.value:
                target_objects = c123_node.objects.filter(type = DBType.POSTGRES.value, ind_deleted = False)
            case DBType.MARIADB:
                target_objects = c123_node.objects.filter(type = DBType.MARIADB.value, ind_deleted = False)
            case DBType.MONGO.value:
                target_objects = c123_node.objects.filter(type = DBType.MONGO.value, ind_deleted = False).exclude(cluster_role = ClusterRoles.ARBITER.value)
            case DBType.MYSQL.value:
                target_objects = c123_node.objects.filter(type = DBType.MYSQL.value, ind_deleted = False)
            case DBType.ELASTIC.value:
                target_objects = c123_node.objects.filter(type = DBType.ELASTIC.value, ind_deleted = False)
            case DBType.REDIS.value:
                target_objects = c123_node.objects.filter(type = DBType.REDIS.value, ind_deleted = False)
            case OperatingSystem.LINUX.value:
                target_objects = c123_host.objects.filter(ind_deleted = False)
            case _:
                target_objects = c123_node.objects.filter(type = monitor_type, ind_deleted = False)

        cmd_monitor = "SELECT app_123monitor.update_monitor(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)"
        if monitor_type not in [OperatingSystem.LINUX.value, OperatingSystem.WINDOWS.value]:
            cmd_monitor_obj = "SELECT app_123monitor.update_monitor_node(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)"
        else:
            cmd_monitor_obj = "SELECT app_123monitor.update_monitor_host(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)"

        parameter_list[1] = None


        with connection.cursor() as cursor:
            cursor.execute(cmd_monitor, parameter_list)
            for target in target_objects:
                parameter_list[1] = target.pk
                try:
                    cursor.execute(cmd_monitor_obj, parameter_list)
                except Exception as e:
                    error, = e.args
                    print(f"Update Monitor: failed to update monitor for object {target.name}", str(error))
                    continue


        return JsonResponse({}, status = 200)

    except Exception as e:
        print(f"Update Monitor: {e}")
        return JsonResponse({'detail':str(e)}, status=500)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def AddAgent(request):

    parameters = request.data

    parameter_list = [
        parameters.get("agent_ip"),
        parameters.get("agent_key"),
        parameters.get("ind_enabled"),
        parameters.get("role")
    ]

    cmd_agent = "SELECT app_123monitor.insert_agent(%s,%s,%s,%s)"

    try:
        with connection.connection as real_conn:
            with real_conn.cursor() as cursor:
                cursor.execute(cmd_agent, parameter_list)
    except Exception as e:
        error, = e.args
        print(f"Add Agent: {error}")
        return JsonResponse({'detail': str(error)}, status=500)

    return JsonResponse({}, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def EditAgent(request):

    parameters = request.data

    cmd_role = "SELECT app_123monitor.update_agent_role(%s,%s)"
    cmd_active = "SELECT app_123monitor.set_agent_active(%s,%s)"
    cmd_default = "SELECT app_123monitor.set_agent_default(%s)"
    agent_id = parameters.get("agent_id")

    try:
        with connection.connection as real_conn:
            with real_conn.cursor() as cursor:
                cursor.execute(cmd_role, [agent_id, parameters.get("role")])
                cursor.execute(cmd_active, [agent_id, parameters.get("ind_active")])
                cursor.execute(cmd_default, [agent_id])
    except Exception as e:
        error, = e.args
        print(f"Edit Agent: {error}")
        return JsonResponse({'detail': str(error)}, status=500)

    return JsonResponse({}, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def SnoozeMonitor(request):

    monitor_name = request.data.get("monitor_name")
    node_id = request.data.get("node_id")
    node_type = request.data.get("node_type")

    try:
        with connection.cursor() as cursor:
            cursor.callproc("app_123monitor.snooze_monitor", [node_id, monitor_name, node_type])
    except Exception as e:
        error, = e.args
        print("Snooze Monitor: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)
    return JsonResponse({}, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def GetAgentsChart(request):
    ind_interval = request.data.get("ind_interval")

    try:
       with connection.connection as real_conn:
           with real_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                agents_chart = list(fetch_data("SELECT app_123monitor.get_agents_graph(%s);FETCH ALL IN get_agents_graph_cursor", cursor, [ind_interval]))
    except Exception as e:
        error, = e.args
        print("Get Agents Chart: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

    return JsonResponse(agents_chart, status = 200, safe=False)
@api_view(['GET'])
@permission_classes((IsAuthenticated, ))
def GetAgentsSummary(request):

    try:
        with connection.connection as real_conn:
            with real_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                cursor.callproc("app_123monitor.get_agents_summary")
                agents_summary = cursor.fetchone()['get_agents_summary']
    except Exception as e:
        error, = e.args
        print("Get Agents Summary: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)

    return JsonResponse(agents_summary, status = 200, safe=False)
@api_view(['GET'])
@permission_classes((IsAuthenticated, ))
def alerts_list(request):
    alerts = []
    try:
        with connection.connection as real_conn:
            with real_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                for alert in fetch_data('SELECT app_123monitor.get_alerts();FETCH ALL IN "get_alerts_cursor"', cursor):
                    node, host = None, None
                    if 'linux' in alert["node_type"].lower():
                        host = c123_host.objects.filter(pk = alert["node_id"]).first()
                        if not host:
                            continue
                    else:
                        node = c123_node.objects.filter(pk = alert["node_id"]).first()
                        if not node:
                            continue
                    alert["type"] = node.type if not host else OperatingSystem.LINUX.value
                    alert["what"] = node.get_type_display() if not host else OperatingSystem.LINUX.value.capitalize()

                    if 'linux' not in alert["node_type"].lower():
                        match node.type:
                            case DBType.ORACLE_CDB.value | DBType.ORACLE_PDB.value:
                                alert["location"] = node.name + "@" + str(node.host)
                            case _:
                                alert["location"] = str(node.host) + ":" + str(node.port)
                    else:
                        alert["location"] = host.name

                    alerts.append(alert)
    except Exception as e:
        error, = e.args
        print("alerts_list: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)
    return JsonResponse(alerts, status = 200, safe = False)
@api_view(['GET'])
@permission_classes((IsAuthenticated, ))
def agents_list(request):
    try:
        with connection.connection as real_conn:
            with real_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                agents = list(fetch_data('SELECT app_123monitor.get_agents();FETCH ALL IN "get_agents_cursor"', cursor))
    except Exception as e:
        error, = e.args
        print("agents_list: ", str(error))
        return JsonResponse({'detail': str(error)}, status=403)
    return JsonResponse(agents, status = 200, safe = False)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def node_alerts_list(request, id):

    try:
        node_type = request.data.get("node_type", "")
        alerts = []
        node = c123_node.objects.filter(pk = id).first()
        if node:
             what = node.get_type_display()
             node_name = node.name
             node_host = node.host.name
             node_port = node.port
        else:
            what = node_type
            node_name = request.data.get('node_name')
            node_host = request.data.get('node_host')
            node_port = request.data.get('node_port')

        with connection.connection as real_conn:
            with real_conn.cursor(cursor_factory = RealDictCursor) as cursor:
                for alert in fetch_data('SELECT app_123monitor.get_node_alerts(%s,%s);FETCH ALL IN "get_node_alerts_cursor"', cursor, [id, node_type]):
                    alert["what"] = what
                    match node_type:
                        case DBType.ORACLE_CDB.value | DBType.ORACLE_PDB.value:
                            alert["location"] = node_name + "@" + str(node_host)
                        case _:
                            alert["location"] = str(node_host) + ":" + str(node_port)
                    alerts.append(alert)

        return JsonResponse(alerts, status = 200, safe = False)
    except Exception as e:
        error = str(e)
        print("node_alerts_list: ", error)
        return JsonResponse({'detail': error}, status=500)
