from rest_framework import serializers
import json,os,oracledb 
from clusterCore.tasks import drop_backup, format_rsa_key, license_after_create, check_os_version, get_fully_qualified_domain_name, format_password
from clusterCore.models import c123_cluster, c123_inventory, c123_host, c123_node, Task, c123_Oracle_Standalone, OracleClusterSizing, IntegratedService
from clusterCore.enums import DBType, ClusterRoles, TaskState, Schedule, Severity, Status, Inventory, DataBaseType, Node_Size_Allowed, OperatingSystem, IntegratedServiceTypes
from clusterCore.exceptions import *
from clusterCore.api.serializer import object_to_json, c123_CreateHostSerializer, c123_CreateTempClusterSerializer, FORBID_APP_HOST, RECREATION_CONFLICT, OS_UNSUPPORTED
from apscheduler.triggers.interval import IntervalTrigger
from django.utils import timezone
class c123_RegisterCDBSerializer(serializers.HyperlinkedModelSerializer):
    
    host = c123_CreateHostSerializer()
    cluster = c123_CreateTempClusterSerializer()
    class Meta:
        model = c123_node
        fields = ['name', 'type', 'status', 'sizing', 'port', 'dir_123Cluster', 'cluster_role', 'database_user', 'database_pwd', 'ex_params', 'cluster', 'host' ]
    
    @license_after_create 
    def create(self, validated_data):
        
        msg = ""
        create_run_linux = True
        is_error, hostname = get_fully_qualified_domain_name(validated_data["host"]["name"])
        if is_error:
            raise APIException401(hostname)
        on_app_host = c123_host.objects.filter(name = hostname, ex_params__myself = True).exists()
        if on_app_host:
            raise APIException406(FORBID_APP_HOST.format(hostname))
        
        host_exist = c123_host.objects.filter(name = hostname).exists()
        node_name = validated_data['name']
        host_data = validated_data.pop('host')
        host_data['name'] = hostname
        cdb_exists = c123_node.objects.filter(name = node_name)
        if cdb_exists.filter(ind_deleted = False, host__name = hostname).exists():
                raise APIException409(RECREATION_CONFLICT.format(DataBaseType.ORACLE_CDB.value, node_name))
        cdb_to_update = cdb_exists.filter(ind_deleted = True).first()
        
        current_user = self.context['request'].user
        host_data['created_by'] = current_user
        host_data['last_update_by'] = current_user
        host_data['password'] = format_password(host_data['password'])
        host_data['ex_params'] = {}
        
        (distributor, release) = check_os_version(host_data)
        if OperatingSystem.INVALID.value in distributor:
            raise APIException401(release['stderr'])
        
        if not OperatingSystem.CENTOS.value in distributor :
               raise APIException406(OS_UNSUPPORTED.format(DataBaseType.ORACLE_CDB.value, '', distributor, release))
        
        current_user = self.context['request'].user
        cluster_data = validated_data.pop('cluster')
        cluster_data['created_by'] = current_user
        cluster_data['last_update_by'] = current_user
        cluster_data['type'] = DBType.ORACLE_CDB.value
        validated_data['created_by'] = current_user
        validated_data['last_update_by'] = current_user
        validated_data['status'] = Status.REGISTERING.value
        validated_data['severity'] = Severity.NOT_OK.value 
        validated_data['type'] = DBType.ORACLE_CDB.value
        validated_data['sizing'] = validated_data['sizing'].upper()
        validated_data['database_pwd'] = format_password(validated_data['database_pwd'])
        validated_data['ex_params']['log_mode'] = 'NOARCHIVELOG'
        validated_data['ex_params']['flashback_on'] = 'NO'
        private_key = format_rsa_key(host_data['private_key'].strip())
        if host_exist:
            c123_host.objects.filter(name=hostname).update(**host_data)
            host = c123_host.objects.get(name=hostname)
            create_run_linux = False       
        else:
            host = c123_host.objects.create(**host_data)
           
        host.private_key = private_key
        host.save()
        validated_data['host'] = host
        dsn_tns = oracledb.makedsn(hostname, validated_data["port"], validated_data["name"])
        try:
            conn = oracledb.connect(user=validated_data["database_user"], password=validated_data["database_pwd"], dsn=dsn_tns, mode=oracledb.SYSDBA)
        except oracledb.DatabaseError as e:
            error, = e.args
            raise APIException401(str(error))
        if conn is not None: 
            try:
                cursor = conn.cursor()    
                cursor.execute("SELECT VERSION FROM V$INSTANCE")
                # query result
                if cdb_to_update:
                    validated_data["created_date"] = timezone.now()
                    c123_node.objects.filter(pk=cdb_to_update.pk).update(**validated_data)
                    cdb = c123_node.objects.get(pk=cdb_to_update.pk) 
                else:    
                    cdb = c123_node.objects.create(**validated_data)   
                cdb.version = cursor.fetchone()[0]
                cdb.ex_params["cdb_id"] = cdb.pk
                cdb.save()
            except oracledb.DatabaseError as e:
                error, = e.args
                raise APIException400(str(error))
    
            # step 2 verify oracle version
            ver = float(cdb.version.split('.')[0] + '.' + cdb.version.split('.')[1])
            
            if(ver < 12.1):
                # we only support versions 12 to 19
                msg = 'CDB version {} is not between 12.0 and 19.0'.format(cdb.version)
                raise APIException406(msg)    
            # step 3. if version over 12.1 check flashback support
            else:
                try:
                    cursor.execute("SELECT DATABASE_ROLE FROM v$database")
                    db_role = cursor.fetchone()
                except oracledb.DatabaseError as e:
                    error, = e.args
                    raise APIException410(str(error))
                
                if not db_role or ClusterRoles.PRIMARY.value not in db_role[0]:
                    msg = 'This is not a primary Oracle database. Cannot register this CDB'
                    raise APIException406(msg)
                try:
                    cursor.execute("SELECT LOG_MODE FROM v$database")
                    res = cursor.fetchone()
                except oracledb.DatabaseError as e:
                    error, = e.args
                    raise APIException410(str(error))
                else:
                    log_mode = res[0] if res else ""
                    cdb.ex_params["log_mode"] = log_mode
                    # 3.1 check archive mode
                    if 'NOARCHIVELOG' in log_mode or 'ARCHIVELOG' not in  log_mode:               
                        msg += "The database is not in archive log mode. Flashback/remote pdb copy is not supported.\n"
                    
                    try:
                        cursor.execute("SELECT FLASHBACK_ON FROM v$database")
                        res = cursor.fetchone()
                    except oracledb.DatabaseError as e:
                        error, = e.args
                        raise APIException410(str(error))
                 
                    cdb.ex_params["flashback_on"] = res[0] if res else ""
                    cdb.save()
                    if 'YES' not in cdb.ex_params["flashback_on"]:                       
                        msg += "The database is not in flashback mode.Flashback is not supported.\n"
                            
            # step 4 check oracle edition
            try:
                cursor.execute("SELECT banner FROM v$version WHERE UPPER(BANNER) LIKE '%DATABASE%'")
                res = cursor.fetchone()
            except oracledb.DatabaseError as e:
                error, = e.args
                raise APIException410(str(error))
        
            if res and 'Standard' in res[0]:
                #why could not?
                msg = 'This is a standard oracle database version and it supports only 1 pluggable database. Cannot register this CDB'
                raise APIException406(msg)
            
            # step 5 verify cdb support
            try:
                cursor.execute('SELECT CDB FROM v$database')
                res = cursor.fetchone()
            except oracledb.DatabaseError as e:
                error, = e.args
                raise APIException410(str(error))
         
            if (not res or 'YES' not in res[0]):       
                raise APIException406("This database is not a container database. Cannot register this CDB")    
        
            try:
                # cursor.execute("SELECT PDB_NAME FROM DBA_PDBS")
                cursor.execute("SELECT NAME, OPEN_MODE, TOTAL_SIZE/1024 FROM v$pdbs")
            except Exception as e:
                error, = e.args
                raise APIException410(str(error))
            
            pdbs_details = cursor.fetchall()
            found_pdbs_sizing = validated_data['ex_params'].get("found_pdbs_sizing")
            for pdb_item in pdbs_details:
                if 'PDB$SEED' not in pdb_item[0]:
                    pdb = c123_node.objects.filter(name=pdb_item[0], type=DBType.ORACLE_PDB.value, port=cdb.port, cluster_role=ClusterRoles.PDB.value, host=cdb.host, cluster = cdb.cluster, ex_params__cdb_id=cdb.id).first()
                    
                    ind_started = False if 'MOUNTED' not in pdb_item[1].upper()  else True
                    # size in GB
                    size = round(float(pdb_item[2]), 2)
                    if pdb:
                        pdb.size = size
                        pdb.ind_started = ind_started 
                        pdb.sizing = found_pdbs_sizing
                        pdb.ind_deleted = False
                        pdb.created_date = timezone.now()
                        pdb.created_by = current_user
                        pdb.last_update_by = current_user
                        pdb.save()
                    else:
                        c123_node.objects.create(name=pdb_item[0], type=DBType.ORACLE_PDB.value,
                                            created_by=current_user, last_update_by=current_user,
                                            port=cdb.port, size=size,
                                            cluster=cdb.cluster, cluster_role=ClusterRoles.PDB.value,
                                            host=cdb.host, ind_started=ind_started,
                                            severity='green', sizing=found_pdbs_sizing,
                                            dir_123Cluster=cdb.dir_123Cluster,
                                            ex_params={'cdb_id': cdb.id})
            tmp_data = json.dumps({'cdb_data': object_to_json(cdb), 'host_data': object_to_json(cdb.host), 'create_run_linux': create_run_linux, 'created_by': current_user.id})
            if self.context["request"].data.get("drop_inventories"):
                drop_backup(c123_inventory.objects.filter(node = cdb))
            else:
                c123_inventory.objects.filter(node = cdb).exclude(type = Inventory.EXPORT.value, ex_params__belong_to = cdb.type).exclude(type = Inventory.BACKUP.value, ex_params__belong_to = cdb.type, ex_params__node_version = cdb.version).exclude(type = Inventory.FLASHBACK.value, ex_params__belong_to = cdb.type, ex_params__node_version__lte = cdb.version).delete()
            
            task = Task.objects.create(task='register_cdb_task',
                                        displayment='register {}'.format(cdb.name),
                                        target_type=DBType.ORACLE_CDB.value,
                                        target_id=cdb.id,
                                        org_id = cdb.org_id,
                                        status=TaskState.PENDING.value,
                                        schedule=Schedule.NOW.value,
                                        parameters=tmp_data,
                                        created_by=current_user)
            
            from django.conf import settings
            from clusterCore.scheduler import UpdateOracleClusterRoles
            settings.scheduler.add_job(UpdateOracleClusterRoles,IntervalTrigger(minutes= 1, timezone=settings.TIME_ZONE), replace_existing=True, next_run_time=timezone.now() + timezone.timedelta(minutes = 5), id= f'update_oracle_roles_{cdb.cluster.id}',kwargs={"cluster_id": cdb.cluster.id})
        elif conn is None:       
            raise APIException401("Couldn't create a connection")
        if msg != '':
            #informational, means operation are partially successful...
            raise APIException206(msg)
        return cdb, task

class c123_OracleStandaloneSerializer(serializers.ModelSerializer):
    type = serializers.SerializerMethodField()
    
    def get_type(self, obj):
        return "STANDALONE"
    class Meta:
        model = c123_Oracle_Standalone
        fields = '__all__'
        
class c123_CreatePDBSerializer(serializers.ModelSerializer):
    cluster = c123_CreateTempClusterSerializer()
    class Meta:
        model = c123_node
        fields = ['name', 'type', 'status', 'sizing', 'port', 'cluster_role', 'database_user', 'database_pwd', 'ind_autostartup', 'ex_params', 'cluster', 'host' ]
    
    @license_after_create 
    def create(self, validated_data):
        
        current_user = self.context['request'].user
        node_name = validated_data['name'].upper()
        cdb = c123_node.objects.filter(type = DBType.ORACLE_CDB.value, pk = validated_data['ex_params']['cdb_id'], ind_deleted = False).first()
        if not cdb:
            raise APIException410(f"cdb id {validated_data['ex_params']['cdb_id']} not available")
        cluster_data = validated_data.pop('cluster')
        cluster_data['created_by'] = current_user
        cluster_data['last_update_by'] = current_user
        cluster_data['type'] = DBType.ORACLE_PDB.value
        pdb_exist = c123_node.objects.filter(name = node_name, ex_params__cdb_id = cdb.id)
        if pdb_exist.filter(ind_deleted = False).exists():
            raise APIException409(validated_data["name"]  + " already exists in " + cdb.name)
        pdb_to_update = pdb_exist.filter(ind_deleted = True).first()
        validated_data['cluster'] = cdb.cluster
        validated_data['created_by'] = current_user
        validated_data['last_update_by'] = current_user
        validated_data['host'] = cdb.host
        validated_data['type'] = DBType.ORACLE_PDB.value
        validated_data['severity'] = 'red'
        validated_data['dir_123Cluster'] = cdb.dir_123Cluster
        validated_data['database_pwd'] = format_password(validated_data['database_pwd'])
        validated_data['version'] = cdb.version
        validated_data['port'] = cdb.port
        validated_data['name'] = node_name
        validated_data['sizing'] = validated_data['sizing'].upper()
        validated_data['status'] = Status.REGISTERING.value
        validated_data['ind_deleted'] = False
        validated_data['ind_started'] = False
        # connect to standbys if selected
        # standby = request.POST.get('standby')
        # print(f"request.POST.get('standby') {standby}")
        # if standby == 'all':
        #     for stdby in OracleStandby.objects.filter(cdb=cdb):
        #         PluggableStandby.objects.create(
        #             created_by=request.user, pdb=pdb, standby=stdby)
        # elif standby != 'none':
        #     stdby = OracleStandby.objects.get(pk=standby)
        #     psb = PluggableStandby.objects.create(
        #         created_by=request.user, pdb=pdb, standby=stdby)
        #     print(f"{psb} created using {stdby} and {pdb}")
        if pdb_to_update:
            validated_data["created_date"] = timezone.now()
            c123_node.objects.filter(pk=pdb_to_update.pk).update(**validated_data)
            pdb = c123_node.objects.get(pk=pdb_to_update.pk) 
        else:    
            pdb = c123_node.objects.create(**validated_data)
        
        host = cdb.host 
        
        if self.context["request"].data.get("auto_delete"):
            pdb.auto_delete =  pdb.created_date + timezone.timedelta(days=float(self.context["request"].data.get("auto_delete_days")))
            pdb.save()
            
        # serialize the objects
        cdb_data = object_to_json(cdb)
        pdb_data = object_to_json(pdb)
        host_data = object_to_json(host)
        if self.context["request"].data.get("drop_inventories"):
            drop_backup(c123_inventory.objects.filter(node = pdb))
        else:
            c123_inventory.objects.filter(node = pdb).exclude(type = Inventory.EXPORT.value, ex_params__belong_to = pdb.type).exclude(type = Inventory.BACKUP.value, ex_params__belong_to = pdb.type, ex_params__node_version = pdb.version).exclude(type = Inventory.FLASHBACK.value, ex_params__belong_to = pdb.type, ex_params__node_version__lte = pdb.version).delete()
        tmp_data = json.dumps({'cdb_data': cdb_data, 'pdb_data': pdb_data, 'host_data': host_data, 'created_by': current_user.id})
        task = Task.objects.create(task='create_pdb_task',
                                    displayment='create {}'.format(pdb.name),
                                    target_type=DBType.ORACLE_PDB.value,
                                    target_id=pdb.id,
                                    org_id = pdb.org_id,
                                    status=TaskState.PENDING.value,
                                    schedule=Schedule.NOW.value,
                                    parameters=tmp_data,
                                    created_by=current_user)
        return pdb, task
    

class c123_InstallCDBSerializer(serializers.HyperlinkedModelSerializer):
    host = c123_CreateHostSerializer()
    cluster = c123_CreateTempClusterSerializer()
    class Meta:
        model = c123_node
        fields = ['name', 'cluster', 'host', 'port', 'version', 'database_user', 'database_pwd', 'dir_123Cluster', 'ex_params']

    @license_after_create 
    def create(self, validated_data):
        
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        
        create_run_linux = True
        host_data = validated_data.pop('host')
        if not host_data.get('ex_params'):
            host_data['ex_params'] = {}
        
        proxmox_host = validated_data['ex_params'].get('proxmox')
        if proxmox_host:
            service_name = validated_data['ex_params'].get('service_name')
            service = IntegratedService.objects.filter(name = service_name, type = IntegratedServiceTypes.PROXMOX.name).first()
            if not service:
                raise APIException404(f"{service_name} is not found")
        if not proxmox_host:
            is_error, hostname = get_fully_qualified_domain_name(host_data["name"])
            if is_error:
                raise APIException401(hostname)
        else:
            hostname = host_data['name'] 
        on_app_host = c123_host.objects.filter(name = hostname, ex_params__myself = True).exists()
        if on_app_host:
            raise APIException406(FORBID_APP_HOST.format(hostname))
        
        host_exist = c123_host.objects.filter(name = hostname).exists()
        host_data['name'] = hostname
        node_name = validated_data['name']
        cdb_exists = c123_node.objects.filter(name = node_name)
        if cdb_exists.filter(ind_deleted = False, host__name = hostname).exists():
                raise APIException409(RECREATION_CONFLICT.format(DataBaseType.ORACLE_CDB.value, node_name))
        cdb_to_update = cdb_exists.filter(ind_deleted = True).first()
        
        current_user = self.context['request'].user
        host_data['created_by'] = current_user
        host_data['last_update_by'] = current_user
        host_data['password'] = format_password(host_data['password'])
        host_data['ex_params'] = {}
        if not proxmox_host:
            (distributor, release) = check_os_version(host_data)
            if OperatingSystem.INVALID.value in distributor:
                raise APIException401(release['stderr'])
        else:
            distributor = validated_data['ex_params'].get('distributor')
            release = validated_data['ex_params'].get('release')
        
        if not OperatingSystem.CENTOS.value in distributor :
               raise APIException406(OS_UNSUPPORTED.format(DataBaseType.ORACLE_CDB.value, validated_data['version'], distributor, release))
        
        current_user = self.context['request'].user
        
        cluster_data = validated_data.pop('cluster')
        cluster_data['created_by'] = current_user
        cluster_data['last_update_by'] = current_user
        cluster_data['type'] = DBType.ORACLE_CDB.value
        cluster_data['name'] = node_name
        own_cluster = c123_cluster.objects.filter(name = cluster_data["name"]).first()
        #if recreated primary has not registered cluster for self in the past
        if not own_cluster:
            own_cluster, created = c123_cluster.objects.get_or_create(**cluster_data)
        validated_data['cluster'] = own_cluster
        validated_data['created_by'] = current_user
        validated_data['last_update_by'] = current_user
        validated_data['status'] = Status.CREATING.value
        validated_data['severity'] = Severity.NOT_OK.value 
        validated_data['type'] = DBType.ORACLE_CDB.value
        validated_data['cluster_role'] = ClusterRoles.CDB.value
        validated_data['sizing'] =  self.context['request'].data.get('sizing')
        validated_data['ind_started'] = False
        validated_data['database_pwd'] = format_password(validated_data['database_pwd'])
        validated_data['ind_deleted'] = False
        datadir = validated_data['dir_123Cluster'].strip()
        if os.path.basename(datadir).upper() == validated_data['name'].strip().upper():
            validated_data['dir_123Cluster'] = os.path.dirname(datadir)  
        private_key = format_rsa_key(host_data['private_key'].strip())
        if host_exist:
            c123_host.objects.filter(name=hostname).update(**host_data)
            host = c123_host.objects.get(name=hostname)
            create_run_linux = False  
        else:
            host = c123_host.objects.create(**host_data)
    
        host.private_key = private_key
        host.save()
        validated_data['host'] = host
        if cdb_to_update:
            validated_data["created_date"] = timezone.now()
            c123_node.objects.filter(pk=cdb_to_update.pk).update(**validated_data)
            cdb = c123_node.objects.get(pk=cdb_to_update.pk) 
        else:    
            cdb = c123_node.objects.create(**validated_data)

        if self.context["request"].data.get("auto_delete"):
            cdb.auto_delete =  cdb.created_date + timezone.timedelta(days=float(self.context["request"].data.get("auto_delete_days")))
            cdb.save()

        oracle_sys, created = OracleClusterSizing.objects.get_or_create()
        match cdb.sizing:
            case Node_Size_Allowed.SMALL.name:
                ram_gb = oracle_sys.small_memory
                cpu_cores = oracle_sys.small_cpu
                disk_gb = oracle_sys.small_disk
                flashback_keep_days_minutes = oracle_sys.small_restore_points_keep_days * 1440
            case Node_Size_Allowed.MEDIUM.name:
                cpu_cores = oracle_sys.medium_cpu
                disk_gb = oracle_sys.medium_disk
                ram_gb = oracle_sys.medium_memory
                flashback_keep_days_minutes = oracle_sys.medium_restore_points_keep_days * 1440           
            case Node_Size_Allowed.LARGE.name:
                cpu_cores = oracle_sys.large_cpu
                disk_gb = oracle_sys.large_disk
                ram_gb = oracle_sys.large_memory
                flashback_keep_days_minutes = oracle_sys.large_restore_points_keep_days * 1440 
            case Node_Size_Allowed.VERY_LARGE.name:
                cpu_cores = oracle_sys.very_large_cpu
                disk_gb = oracle_sys.very_large_disk
                ram_gb = oracle_sys.very_large_memory
                flashback_keep_days_minutes = oracle_sys.medium_restore_points_keep_days * 1440
            case _:
                ram_gb = 4
                cpu_cores = 4
                disk_gb = 70
                flashback_keep_days_minutes = 1440
        ram_mb = round((ram_gb * 1024) * 0.7)
        standalone = cdb.ex_params["standalone"]
        cdb.ex_params["ram_mb"] = ram_mb 
        cdb.ex_params["flashback_keep_days_minutes"] = flashback_keep_days_minutes
        cdb.ex_params["user_system_pwd"] = format_password(cdb.ex_params["user_system_pwd"])
        cdb.ex_params["user_sys_pwd"] = format_password(cdb.ex_params["user_sys_pwd"])
        cdb.ex_params["cdb_id"] = cdb.pk
        cdb.save()
        if standalone:
            #flexible, when we want for standalone to stay on layout but exclude from list of duplicate 
            # - can just remove that object by drop standalone dialog of layout, so cdb will be kept
            standalone_exists = c123_Oracle_Standalone.objects.filter(name = cdb.name).exists()
            if not standalone_exists:
                c123_Oracle_Standalone.objects.create(name = cdb.name, host_name = host.name, port = cdb.port, service_name = cdb.name, database_user = cdb.database_user, database_pwd = cdb.database_pwd, created_by = cdb.created_by)
            else:
                c123_Oracle_Standalone.objects.filter(name = cdb.name).update(name = cdb.name, host_name = host.name, port = cdb.port, service_name = cdb.name, database_user = cdb.database_user, database_pwd = cdb.database_pwd, created_by = cdb.created_by)

        # serialize the objects
        cdb_data = object_to_json(cdb)
        host_data = object_to_json(host)
        if self.context["request"].data.get("drop_inventories"):
            drop_backup(c123_inventory.objects.filter(node = cdb))
        else:
            c123_inventory.objects.filter(node = cdb).exclude(type = Inventory.EXPORT.value, ex_params__belong_to = cdb.type).exclude(type = Inventory.BACKUP.value, ex_params__belong_to = cdb.type, ex_params__node_version = cdb.version).exclude(type = Inventory.FLASHBACK.value, ex_params__belong_to = cdb.type, ex_params__node_version__lte = cdb.version).delete()
        
        app_host = object_to_json(app_host)
        repository = validated_data["ex_params"].get("repository")
        tmp_data = {'cdb_data': cdb_data,'host_data': host_data, 'app_host': app_host, 'ram_gb': ram_gb, 'disk_gb': disk_gb, 'cpu_cores': cpu_cores, 'create_run_linux': create_run_linux, 'repository': repository,  'created_by': current_user.id}
        if proxmox_host:
            cdb.ex_params['service_id'] = service.pk
            cdb.save()
            tmp_data['service_data'] = object_to_json(service)
        task = Task.objects.create(task='install_cdb_task',
                                    displayment='install {}'.format(cdb.name),
                                    target_type=DBType.ORACLE_CDB.value,
                                    target_id=cdb.id,
                                    org_id = cdb.org_id,
                                    status=TaskState.PENDING.value,
                                    schedule=Schedule.NOW.value,
                                    parameters=json.dumps(tmp_data),
                                    created_by=current_user)
        return cdb, task