from rest_framework import serializers
from clusterCore.models import c123_node, c123_host, c123_cluster, Task, IntegratedService, ClusterSizing
from clusterCore.enums import DBType, ClusterRoles, Inventory, Status, OperatingSystem, Severity, Schedule, TaskState, DataBaseType, IntegratedServiceTypes, Node_Size_Allowed
from clusterCore.api.serializer import c123_inventory, c123_CreateHostSerializer, c123_CreateTempClusterSerializer, object_to_json, FORBID_APP_HOST, RECREATION_CONFLICT
from clusterCore.exceptions import *
from clusterCore.tasks import check_os_version, drop_backup, format_rsa_key, license_after_create, get_fully_qualified_domain_name, format_password
import json
from django.utils import timezone
class c123_AddRedisNodeSerializer(serializers.HyperlinkedModelSerializer):
    
    host = c123_CreateHostSerializer()
    class Meta:
        model = c123_node
        fields = ['name', 'host', 'port', 'version', 'cluster_role', 'dir_123Cluster', 'ex_params']
    
    @license_after_create   
    def create(self, validated_data):
        create_run_linux = True
        host_data = validated_data.pop('host')
        if not host_data.get('ex_params'):
            host_data['ex_params'] = {}
        
        proxmox_host = validated_data['ex_params'].get('proxmox')
        if proxmox_host:
            service_name = validated_data['ex_params'].get('service_name')
            service = IntegratedService.objects.filter(name = service_name, type = IntegratedServiceTypes.PROXMOX.name).first()
            if not service:
                raise APIException404(f"{service_name} is not found")
        if not proxmox_host:
            is_error, hostname = get_fully_qualified_domain_name(host_data["name"])
            if is_error:
                raise APIException401(hostname)
        else:
            hostname = host_data['name']
        on_app_host = c123_host.objects.filter(name = hostname, ex_params__myself = True).exists()
        if on_app_host:
            raise APIException406(FORBID_APP_HOST.format(hostname))
        app_host = c123_host.objects.filter(ex_params__myself = True, ind_deleted = False).first()
        cluster_id  = self.context['request'].data.get('cluster_id')
        target_cluster = c123_cluster.objects.filter(pk = cluster_id).first()
        
        if not target_cluster:
            raise APIException410(f"cluster {cluster_id} is not available")
        cluster_nodes = c123_node.objects.filter(cluster = target_cluster, ind_deleted = False, type = DBType.REDIS.value)
        host_data['name'] = hostname
        node_name = f"{hostname}:{validated_data['port']}"
        validated_data['name'] = node_name
        node_exists = c123_node.objects.filter(name = node_name)
        if node_exists.filter(ind_deleted = False, host__name = hostname).exists():
                raise APIException409(RECREATION_CONFLICT.format(DataBaseType.REDIS.value, node_name))
        node_to_update = node_exists.filter(ind_deleted = True).first()
        current_user = self.context['request'].user
        host_data['created_by'] = current_user
        host_data['last_update_by'] = current_user
        host_data['password'] = format_password(host_data['password'])
        host_data['ex_params'] = {}
        if not proxmox_host:
            (distributor, release) = check_os_version(host_data)
            if OperatingSystem.INVALID.value in distributor:
                raise APIException401(release['stderr'])
        else:
            distributor = target_cluster.ex_params.get('distributor')
            release = target_cluster.ex_params.get('release')
        validated_data['ex_params']['distributor'] = distributor
        validated_data['ex_params']['release'] = release
        # this still take from db (because can be dozens of nodes to check)
        for node in cluster_nodes:
            if node.ex_params.get('distributor') != distributor or node.ex_params.get("release") != release:
               raise APIException406(f'node {node.name} ({node.ex_params.get("distributor")} {node.host.ex_params.get("release")}) and the new node({distributor} {release}) have not the same os version')
     
        validated_data['created_by'] = current_user
        validated_data['last_update_by'] = current_user
        validated_data['status'] = Status.CREATING.value
        validated_data['severity'] = Severity.NOT_OK.value 
        validated_data['type'] = DBType.REDIS.value
        validated_data['database_user'] = target_cluster.ex_params.get("database_user")
        validated_data['database_pwd'] = target_cluster.ex_params.get("database_pwd")
        validated_data['sizing'] =  self.context['request'].data.get('sizing')
        redis_sys, created = ClusterSizing.objects.get_or_create(db_type = DBType.REDIS.value)  
        match validated_data['sizing']:
            case Node_Size_Allowed.SMALL.name:
                ram_gb = redis_sys.small_memory
                cpu_cores = redis_sys.small_cpu
                disk_gb = redis_sys.small_disk
            case Node_Size_Allowed.MEDIUM.name:
                ram_gb = redis_sys.medium_memory
                cpu_cores = redis_sys.medium_cpu
                disk_gb = redis_sys.medium_disk
            case Node_Size_Allowed.LARGE.name:
                ram_gb = redis_sys.large_memory
                cpu_cores = redis_sys.large_cpu
                disk_gb = redis_sys.large_disk
            case Node_Size_Allowed.VERY_LARGE.name:
                ram_gb = redis_sys.very_large_memory
                cpu_cores = redis_sys.very_large_cpu
                disk_gb = redis_sys.very_large_disk
            case _:
                ram_gb = 1
                cpu_cores = 1
                disk_gb = 10 
        validated_data['cluster'] = target_cluster
        validated_data['ind_deleted'] = False
        private_key = format_rsa_key(host_data['private_key'].strip())
        host_exist = c123_host.objects.filter(name = hostname).exists()            
        if host_exist:
            c123_host.objects.filter(name=hostname).update(**host_data)
            host = c123_host.objects.get(name=hostname)
            create_run_linux = False
        else:
            host = c123_host.objects.create(**host_data)
           
        host.private_key = private_key
        host.save()
        validated_data['host'] = host
        if node_to_update:
            validated_data["created_date"] = timezone.now()
            c123_node.objects.filter(pk=node_to_update.pk).update(**validated_data)
            node = c123_node.objects.get(pk=node_to_update.pk)         
        else:    
            node = c123_node.objects.create(**validated_data)
        repository = target_cluster.ex_params['repository']
        node.ex_params["repository"] = repository
        node.ex_params["distributor"] = distributor
        node.ex_params["release"] = release
        if node.cluster_role == ClusterRoles.PRIMARY.value:
            node.ex_params["monitored_by"] = []
        node.save()
        if self.context["request"].data.get("drop_inventories"):
            drop_backup(c123_inventory.objects.filter(node = node))
        else:
            c123_inventory.objects.filter(node = node).exclude(type = Inventory.BACKUP.value, ex_params__belong_to = node.type, ex_params__node_version = node.version).delete()
     
        tmp_data = {'cluster_data': object_to_json(node.cluster), 'node_data': object_to_json(node), 'host_data': object_to_json(node.host), 'app_host': object_to_json(app_host), 'repository': repository, 'cpu_cores': cpu_cores, 'ram_gb': ram_gb, 'disk_gb': disk_gb, 'create_run_linux': create_run_linux, 'created_by': current_user.id}
        if proxmox_host:
            tmp_data['service_data'] = object_to_json(service)
        if node.cluster_role != ClusterRoles.SENTINEL.value:
             masters = c123_node.objects.filter(type = DBType.REDIS.value, cluster = node.cluster, ind_deleted = False, cluster_role = ClusterRoles.PRIMARY.value)
             master_list = [{"host": object_to_json(master.host), "node": object_to_json(master)} for master in masters]
             tmp_data['master_list'] = master_list
        node_label = 'sentinel' if node.cluster_role == ClusterRoles.SENTINEL.value else 'node'
        task = Task.objects.create(task=f"add_redis_{node_label}_task",
                                        displayment='adding redis {} {}'.format(node_label, node.name),
                                        target_type=DBType.REDIS.value,
                                        target_id=node.id,
                                        org_id=node.org_id,
                                        status=TaskState.PENDING.value,
                                        schedule=Schedule.NOW.value,
                                        parameters=json.dumps(tmp_data),
                                created_by=current_user)
    
        return node, task
    
class c123_AddRedisClusterSerializer(serializers.ModelSerializer):
    class Meta:
        model = c123_cluster
        exclude = ['created_by', 'last_update_by']
    def create(self, validated_data):
        data = self.context["request"].data
        if c123_cluster.objects.filter(name = data["name"]).exists():
            raise APIException409(f"Redis Cluster cluster {data['name']} already exists")

        current_user = self.context["request"].user
        cluster_data = {}
        cluster_data["ex_params"] = data.pop("ex_params")
        cluster_data["ex_params"]["database_user"] = data.get("database_user")
        cluster_data["ex_params"]["database_pwd"] = format_password(data.get('database_pwd'))
        
        cluster_data["ex_params"]["version"] = data.get("version")
        cluster_data["ex_params"]["sizing"] = data.get("sizing")
        cluster_data["ex_params"]["status"] = Status.PENDING.value
        cluster_data["ex_params"]["severity"] = Severity.NOT_OK.value
        cluster_data["ex_params"]["standalone"] = False
        cluster_data["type"] = DBType.REDIS.value
        cluster_data["created_by"] = current_user
        cluster_data["last_update_by"] = current_user
        cluster = c123_cluster.objects.create(**cluster_data)
        cluster.name = data["name"]
        cluster.save()
        
        return cluster
    
class c123_CreateRedisStandaloneSerializer(serializers.HyperlinkedModelSerializer):
    host = c123_CreateHostSerializer()
    cluster = c123_CreateTempClusterSerializer()
    class Meta:
        model = c123_node
        fields = ['name', 'cluster', 'host', 'port', 'version', 'database_user', 'database_pwd', 'dir_123Cluster', 'ex_params']
    
    @license_after_create
    def create(self, validated_data):
        host_data = validated_data.pop('host')
        if not host_data.get('ex_params'):
            host_data['ex_params'] = {}
        
        proxmox_host = validated_data['ex_params'].get('proxmox')
        if proxmox_host:
            service_name = validated_data['ex_params'].get('service_name')
            service = IntegratedService.objects.filter(name = service_name, type = IntegratedServiceTypes.PROXMOX.name).first()
            if not service:
                raise APIException404(f"{service_name} is not found")
        if not proxmox_host:
            is_error, hostname = get_fully_qualified_domain_name(host_data["name"])
            if is_error:
                raise APIException401(hostname)
        else:
            hostname = host_data['name'] 
        create_run_linux = True
        on_app_host = c123_host.objects.filter(name = hostname, ex_params__myself = True).exists()
        if on_app_host:
            raise APIException406(FORBID_APP_HOST.format(hostname))
        app_host = c123_host.objects.filter(ex_params__myself = True, ind_deleted = False).first()
        host_data['name'] = hostname
        node_name = f"{hostname}:{validated_data['port']}"
        validated_data['name'] = node_name
        node_exists = c123_node.objects.filter(name = node_name)
        if node_exists.filter(ind_deleted = False, host__name = hostname).exists():
                raise APIException409(RECREATION_CONFLICT.format(DataBaseType.REDIS.value, node_name))
        node_to_update = node_exists.filter(ind_deleted = True).first()
        current_user = self.context['request'].user
        host_data['created_by'] = current_user
        host_data['last_update_by'] = current_user
        host_data['password'] = format_password(host_data['password'])
        host_data['ex_params'] = {}
        if not proxmox_host:
            (distributor, release) = check_os_version(host_data)
            if OperatingSystem.INVALID.value in distributor:
                raise APIException401(release['stderr'])
        else:
            distributor = validated_data['ex_params'].get('distributor')
            release = validated_data['ex_params'].get('release')
        
        cluster_data = validated_data.pop('cluster')
        cluster_data['created_by'] = current_user
        cluster_data['last_update_by'] = current_user
        cluster_data['type'] = DBType.REDIS.value
        cluster_data['name'] = node_name
        own_cluster = c123_cluster.objects.filter(name = cluster_data["name"]).first()
        #if recreated primary has not registered cluster for self in the past
        if not own_cluster:
            own_cluster, created = c123_cluster.objects.get_or_create(**cluster_data)
        validated_data['created_by'] = current_user
        validated_data['last_update_by'] = current_user
        validated_data['status'] = Status.CREATING.value
        validated_data['severity'] = Severity.NOT_OK.value
        validated_data['type'] = DBType.REDIS.value
        validated_data['cluster_role'] = ClusterRoles.PRIMARY.value
        validated_data['sizing'] =  self.context['request'].data.get('sizing')
        redis_sys, created = ClusterSizing.objects.get_or_create(db_type = DBType.REDIS.value)  
        match validated_data['sizing']:
            case Node_Size_Allowed.SMALL.name:
                ram_gb = redis_sys.small_memory
                cpu_cores = redis_sys.small_cpu
                disk_gb = redis_sys.small_disk
            case Node_Size_Allowed.MEDIUM.name:
                ram_gb = redis_sys.medium_memory
                cpu_cores = redis_sys.medium_cpu
                disk_gb = redis_sys.medium_disk
            case Node_Size_Allowed.LARGE.name:
                ram_gb = redis_sys.large_memory
                cpu_cores = redis_sys.large_cpu
                disk_gb = redis_sys.large_disk
            case Node_Size_Allowed.VERY_LARGE.name:
                ram_gb = redis_sys.very_large_memory
                cpu_cores = redis_sys.very_large_cpu
                disk_gb = redis_sys.very_large_disk
            case _:
                ram_gb = 1
                cpu_cores = 1
                disk_gb = 10 
        validated_data["cluster"] = own_cluster
        validated_data["ex_params"]["standalone"] = True
        validated_data['database_pwd'] = format_password(validated_data['database_pwd'])
        validated_data['ind_deleted'] = False
         
        private_key = format_rsa_key(host_data['private_key'].strip())
        host_exist = c123_host.objects.filter(name = hostname).exists()     
        if not host_exist:
            host = c123_host.objects.create(**host_data)
        else:
            c123_host.objects.filter(name=hostname).update(**host_data)
            host = c123_host.objects.get(name=hostname)
        host.private_key = private_key
        host.save()
        validated_data['host'] = host
        if node_to_update:
            validated_data["created_date"] = timezone.now()
            c123_node.objects.filter(pk=node_to_update.pk).update(**validated_data)
            node = c123_node.objects.get(pk=node_to_update.pk)         
        else:    
            node = c123_node.objects.create(**validated_data)
        node.ex_params["slave_of"] = False
        node.save()
        if self.context["request"].data.get("auto_delete"):
            node.auto_delete =  node.created_date + timezone.timedelta(days=float(self.context["request"].data.get("auto_delete_days")))
            node.save()
        
        repository = self.context["request"].data.get("repository")
        if self.context["request"].data.get("drop_inventories"):
            drop_backup(c123_inventory.objects.filter(node = node))
        else:
            c123_inventory.objects.filter(node = node).exclude(type = Inventory.BACKUP.value, ex_params__belong_to = node.type, ex_params__node_version = node.version).delete()
        tmp_data = {'node_data': object_to_json(node), 'host_data': object_to_json(node.host), 'app_host': object_to_json(app_host), 'cpu_cores': cpu_cores, 'disk_gb': disk_gb, 'ram_gb': ram_gb, 'repository': repository, 'create_run_linux': create_run_linux, 'created_by': current_user.id}
        if proxmox_host:
            tmp_data['service_data'] = object_to_json(service)
        task = Task.objects.create(task="install_redis_standalone_task",
                                        displayment='creating redis standalone {}'.format(node.name),
                                        target_type=DBType.REDIS.value,
                                        target_id=node.id,
                                        org_id=node.org_id,
                                        status=TaskState.PENDING.value,
                                        schedule=Schedule.NOW.value,
                                        parameters=json.dumps(tmp_data),
                                        created_by=current_user)
        return node, task