
from unittest.util import _MAX_LENGTH
from django.contrib.auth.models import User
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import JSONField
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from clusterCore.enums import DataBaseType, IntegratedServiceTypes, User_Or_All_Choice, Node_Size_Allowed, ClusterRoles, ContainerOS, DBType
from dateutil.relativedelta import relativedelta

class CustomerKey(models.Model):
    """Licensing model"""
    uuid = models.CharField(primary_key=True, blank=False, null=False, max_length=160, default='0000000000000000000000000000000000000000000000000000000000000000400000000000000000008000000000000000000000000000000000000000000000000000000000000000000000000000', editable=False)
    key = models.CharField(max_length=500, blank=False, null=False)
    elasticsearch = models.IntegerField(default=0)
    postgresql = models.IntegerField(default=0)
    oracle = models.IntegerField(default=0)
    mongo = models.IntegerField(default=0)
    mariadb = models.IntegerField(default=0)
    mysql = models.IntegerField(default=0)
    redis = models.IntegerField(default=0)
    mssql = models.IntegerField(default=0)
    created_date = models.DateTimeField(verbose_name=_("Created Date"), auto_now_add=True, editable=False)

    def __str__(self):
        return self.key

class c123_cluster(models.Model):
    """Replication cluster model"""
    name = models.CharField(_('Name'), max_length=50, null=True, blank=True )
    type = models.CharField(_('Type'), max_length=50, null=True, blank=True, choices=[(db_type.name, db_type.value) for db_type in DataBaseType] )
    allow_children_from_all_users= models.BooleanField(_('Allow Children From All Users'), default=False, blank=False, null=False)
    created_date  = models.DateTimeField(verbose_name=_("Created Date"), auto_now_add=True, editable=False)
    created_by = models.ForeignKey(User, verbose_name=_("Created By"), related_name='cluster_created_by', on_delete=models.DO_NOTHING, null=False, blank=False)
    last_update_date  = models.DateTimeField(verbose_name=_("Last Update Date"), auto_now_add=True, editable=False)
    last_update_by = models.ForeignKey(User, verbose_name=_("Last Update By"), related_name='cluster_last_update_by', on_delete=models.DO_NOTHING, null=False, blank=False)
    ex_params = JSONField(_("Extras"),null=True, blank=True)

class c123_host(models.Model):
    """Remote machine model"""
    name = models.CharField(_('Name'), max_length=100, null=True, blank=True )
    ssh_port = models.CharField(_('SSH Port'), max_length=50, null=True, blank=True )
    username = models.CharField(_('Username'), max_length=50, null=True, blank=True )
    password = models.CharField(_('Password'), max_length=50, null=True, blank=True )
    private_key = models.TextField(verbose_name=_("Private Key"), max_length=1800, blank=True, null=True)
    created_date  = models.DateTimeField(verbose_name=_("Created Date"), auto_now_add=True, editable=False)
    created_by = models.ForeignKey(User, verbose_name=_("Created By"), related_name='host_created_by',  on_delete=models.DO_NOTHING, null=False, blank=False)
    last_update_date  = models.DateTimeField(verbose_name=_("Last Update Date"), auto_now_add=True, editable=False)
    last_update_by = models.ForeignKey(User, verbose_name=_("Last Update By"), related_name='host_last_update_by', on_delete=models.DO_NOTHING, null=False, blank=False)
    status = models.CharField(_('Status'), max_length=200, null=True, blank=True )
    ind_deleted = models.BooleanField(verbose_name=_("Is Deleted"), default=False)
    ex_params = JSONField(_("Extras"),null=True, blank=True)

    def __str__(self):
        return '%s' % self.name

class Organization(models.Model):
    "Users organization model"
    org_id = models.AutoField(primary_key=True)
    date_created = models.DateTimeField(null=True)
    ind_disable_all_nodes = models.CharField(max_length=1, null=True)
    organization_name = models.CharField(max_length=100, null=True)
    ind_enable_notifications = models.CharField(max_length=1, null=True)
    ind_housekeeper = models.CharField(max_length=1, null=True)
    housekeeper_short_retention = models.IntegerField(null=True)
    housekeeper_long_retention = models.IntegerField(null=True)
    max_notifications_per_hour = models.IntegerField(null=True)

    class Meta:
        db_table = 'organization'
        constraints = [
            models.UniqueConstraint(name = 'unique_organization', fields=['organization_name'])
        ]

    def __str__(self):
        return self.organization_name

class Profile(models.Model):
    """Extra user data model"""
    user = models.OneToOneField(User, on_delete=models.CASCADE)
    org_id = models.ForeignKey(Organization, verbose_name=_("Organization"), on_delete=models.DO_NOTHING, null=False, blank=False, default=1, db_column='org_id' )
    is_demo = models.BooleanField(verbose_name=_("Is Demo"), default=False)

class c123_host_monitor(models.Model):
    """metric_name in [cpu, mem, la]"""
    host = models.ForeignKey(c123_host, verbose_name=_("Host Name"), on_delete=models.DO_NOTHING, null=False, blank=False)
    metric_name = models.CharField(_('Metric Name'), max_length=50, blank=True, null=True)
    metric_value = models.CharField(_('Metric Value'), max_length=50, blank=True, null=True)
    metric_status = models.CharField(_('Metric Status'), max_length=50, blank=True, null=True)
    created_date  = models.DateTimeField(verbose_name=_("Created Date"), auto_now_add=True, editable=False)

class c123_node(models.Model):
    """Instance of database management system (for supported types - check DataBaseType enum)"""
    org_id = models.ForeignKey(Organization, verbose_name=_("Organization"), on_delete=models.DO_NOTHING, null=False, blank=False, default=1, db_column='org_id')
    name = models.CharField(_('Name'), max_length=100, null=True, blank=True )
    type = models.CharField(_('Type'), max_length=50, null=True, blank=True, choices=[(db_type.name, db_type.value) for db_type in DataBaseType] )
    created_date = models.DateTimeField(verbose_name=_("Created Date"), auto_now_add=True, editable=False)
    created_by = models.ForeignKey(User, verbose_name=_("Created By"),  related_name='node_created_by', on_delete=models.DO_NOTHING, null=False, blank=False)
    last_update_date = models.DateTimeField(verbose_name=_("Last Update Date"), auto_now_add=True, editable=False)
    last_update_by = models.ForeignKey(User, verbose_name=_("Last Update By"), related_name='node_last_update_by', on_delete=models.DO_NOTHING, null=False, blank=False)
    port= models.CharField(_('Port'),blank=True, null=True, max_length=50)
    cluster = models.ForeignKey(c123_cluster, verbose_name=_("Cluster"), on_delete=models.DO_NOTHING, null=True, blank=False)
    cluster_role = models.CharField(_('Cluster Role'), max_length= 50, choices=[(role.value, role.name) for role in ClusterRoles])
    status = models.CharField(_('Status'), max_length=200, null=True, blank=True )
    host = models.ForeignKey(c123_host, verbose_name=_("Host Name"), on_delete=models.DO_NOTHING, null=False, blank=False)
    ex_params = models.JSONField(_("Extras"),null=True, blank=True)
    version = models.CharField(verbose_name=_("Version"), max_length=50, blank=True, null=True)
    size = models.FloatField(verbose_name=_("Size"), blank=True, null=True)
    free = models.BigIntegerField(verbose_name=_("Free"), blank=True, null=True)
    temp = models.BigIntegerField(verbose_name=_("Temp"), blank=True, null=True)
    ind_deployed = models.BooleanField(verbose_name=_("Is Deployed"), default=False)
    ind_started = models.BooleanField(verbose_name=_("Is Started"), default=False)
    ind_autostartup = models.BooleanField(verbose_name=_("Is Autostart up"), default=False)
    ind_deleted = models.BooleanField(verbose_name=_("Is Deleted"), default=False)
    ind_license = models.BooleanField(verbose_name=_("Is Licensed"), default=False)
    agent_id = models.IntegerField(_('Agent ID'), blank=True, null=True)
    severity = models.CharField(verbose_name=_("Severity"), max_length=30, blank=True, null=True)
    sizing = models.CharField(verbose_name=_("Sizing"), max_length=11, blank=True, null=True)
    auto_delete = models.DateField(verbose_name=_("Auto Delete"),  blank=True, null=True)
    database_user = models.CharField(_("Database User"), max_length=100, blank=True, null=True)
    database_pwd = models.CharField(_("Database Password"), max_length=100, null=True, blank=True)
    dir_123Cluster = models.CharField(verbose_name=_("123 Cluster Directory"), max_length=1000, blank=True, null=True)


    @property
    def display_size(self):
        from clusterCore.api.serializer import format_size
        return '0 KB' if not self.size else format_size(self.size)

    @property
    def cdb_name(self):
        if self.type == DBType.ORACLE_PDB.value:
            cdb = c123_node.objects.filter(pk=self.ex_params.get("cdb_id")).first()
            return cdb.name if cdb else ""
        return ""

    @property
    def has_trial(self):
        return (timezone.now() - relativedelta(months=2)) > self.created_date

    def __str__(self):
        return self.name

class c123_database(models.Model):
    """Database model (for supported types - check DataBaseType enum)"""
    name = models.CharField(_('Name'), max_length=255, null=True, blank=True )
    type = models.CharField(_('Type'), max_length=51, null=True, blank=True, choices=[(db_type.name, db_type.value) for db_type in DataBaseType])
    node = models.ForeignKey(c123_node, verbose_name=_("Node Name"), on_delete=models.DO_NOTHING, null=False, blank=False)
    created_date  = models.DateTimeField(verbose_name=_("Created Date"), auto_now_add=True, editable=False)
    created_by = models.ForeignKey(User, verbose_name=_("Created By"),  related_name='database_created_by', on_delete=models.DO_NOTHING, null=False, blank=False)
    last_update_date  = models.DateTimeField(verbose_name=_("Last Update Date"), auto_now_add=True, editable=False)
    last_update_by = models.ForeignKey(User, verbose_name=_("Last Update By"), related_name='database_last_update_by', on_delete=models.DO_NOTHING, null=False, blank=False)
    status = models.CharField(_('Status'), max_length=200, null=True, blank=True )
    ex_params = models.JSONField(_("Extras"),null=True, blank=True)
    size = models.FloatField(verbose_name=_("Size"), blank=True, null=True)
    free = models.BigIntegerField(verbose_name=_("Free"), blank=True, null=True)
    temp = models.BigIntegerField(verbose_name=_("Temp"), blank=True, null=True)
    ind_deployed = models.BooleanField(verbose_name=_("Is Deployed"), default=False)
    ind_started = models.BooleanField(verbose_name=_("Is Started"), default=False)
    ind_autostartup = models.BooleanField(verbose_name=_("Is Autostart up"), default=False)
    ind_deleted = models.BooleanField(verbose_name=_("Is Deleted"), default=False)
    severity = models.CharField(verbose_name=_("Severity"), max_length=30, blank=True, null=True)
    auto_delete = models.DateField(verbose_name=_("Auto Delete"),  blank=True, null=True)
    database_user = models.CharField(_("Database User"), max_length=100, blank=True, null=True)
    database_pwd = models.CharField(_("Database Password"), max_length=100, null=True, blank=True)

    def __str__(self):
        return self.name

class c123_node_monitor(models.Model):
    """metric_name in [availability, lag]"""
    node = models.ForeignKey(c123_node, verbose_name=_("Node ID"), on_delete=models.DO_NOTHING, null=False, blank=False)
    metric_name = models.CharField(_('Metric Name'), max_length=50, blank=True, null=True)
    metric_value = models.CharField(_('Metric Value'), max_length=50, blank=True, null=True)
    metric_status = models.CharField(_('Metric Status'), max_length=50, blank=True, null=True)
    created_date  = models.DateTimeField(verbose_name=_("Created Date"), auto_now_add=True, editable=False)

class c123_dashboard_item(models.Model):
    """Dashboard items."""
    """visualization_type in [line_chart, table]"""
    """node_type in [PostgreSQL, ElasticSearch, LINUX, ...]"""
    monitor_id = models.AutoField(_('Monitor ID'), primary_key=True)
    created_by = models.ForeignKey(User, verbose_name=_("Created By"), on_delete=models.DO_NOTHING, blank=False, null=False)
    node_id = models.ForeignKey(c123_node, verbose_name=_("Node ID"), on_delete=models.DO_NOTHING, null=False, blank=False, db_column='node_id')
    order_id = models.IntegerField(_('Order ID'), blank=False, null=False, default=0)
    visualization_type = models.CharField(_('Visualization Type'), max_length=50, blank=True, null=True)
    node_type = models.CharField(_('Node Type'), max_length=50, blank=True, null=True)
    column_display_name = models.CharField(_('Column Display Name'), max_length=50, blank=True, null=True)
    monitor_name = models.CharField(_('Monitor Name'), max_length=50, blank=True, null=True)
    class Meta:
        db_table = 'dashboard_item'
        constraints = [
            models.UniqueConstraint(name = 'unique_item', fields=['created_by', 'node_id', 'visualization_type',
                                                                   'node_type', 'column_display_name', 'monitor_name'])
        ]

class Task(models.Model):
    "Database administration service task model"
    org_id = models.ForeignKey(Organization, verbose_name=_("Organization"), on_delete=models.DO_NOTHING, null=False, blank=False, default=1, db_column='org_id')
    task_id = models.AutoField(_('Task ID'), primary_key=True)
    created_date = models.DateTimeField(verbose_name=_("Created Date"), blank=True, null=True, auto_now_add=True,  editable=False)
    created_by = models.ForeignKey(User, verbose_name=_("Created By"), on_delete=models.DO_NOTHING, blank=True, null=True)
    status = models.CharField(verbose_name=_("Status"), max_length=50, blank=True, null=True)
    task = models.CharField(_('Task'), max_length=500, blank=True, null=True)
    monitor_name = models.CharField(_('Monitor Name'), max_length=50, blank=True, null=True)
    displayment = models.CharField(_('Displayment'), max_length=500, blank=True, null=True)
    target_type = models.CharField(_('Target Type'), max_length=50, blank=True, null=True)
    target_id = models.IntegerField(_('Target ID'), blank=True, null=True)
    agent_id = models.IntegerField(_('Agent ID'), blank=True, null=True)
    schedule = models.CharField(_('Schedule'), max_length=30, blank=True, null=False)
    parameters = JSONField(_("Parameters"), blank=True, null=True)
    next_due_date = models.DateTimeField(verbose_name=_("Next Due Date"), blank=True, null=True, default=timezone.now)

    class Meta:
        indexes = [
            models.Index(fields=['created_date'], name='task_idx'),
        ]

class IntegratedService(models.Model):

    def supported_databases_default():
        return {f'{t.name}': False for t in DataBaseType}

    def supported_os_default():
        return {f'{t.value}': False for t in ContainerOS}

    type = models.CharField(_('Type'), max_length=50, null=False, blank=False, choices=[(service_type.name, service_type.value) for service_type in IntegratedServiceTypes] )
    name = models.CharField(_("Name"), max_length=100, blank=True, null=True)
    host_name = models.CharField(_('Host Name'),blank=True, null=True, max_length=100)
    port = models.CharField(_('Port'),blank=True, null=True, max_length=50)
    username = models.CharField(_('Username'), max_length=50, null=True, blank=True)
    password = models.CharField(_('Password'), max_length=50, null=True, blank=True)
    ind_verified = models.BooleanField(verbose_name=_("Is Verified"), default=False)
    supported_databases = JSONField(_("Supported databases"), blank=True, null=True, default=supported_databases_default)
    supported_os = JSONField(_("Supported OS"), blank=True, null=True, default=supported_os_default)
    ex_params = JSONField(_("Extras"), null=True, blank=True)


    class Meta:
        db_table = 'integrated_services'
        indexes = [
            models.Index(fields=['name'], name='service_name_idx'),
            models.Index(fields=['host_name'], name='host_name_idx'),
        ]
        constraints = [
            models.UniqueConstraint(name = 'unique_service', fields=['name', 'host_name'])
        ]





class TaskLog(models.Model):
    "Database administration service task verbose log model"
    task_log_id = models.UUIDField(_('Task Log ID'), blank=True, null=True)
    task_id = models.IntegerField(_('Task ID'), blank=True, null=True)
    created_date = models.DateTimeField(verbose_name=_("Created Date"), blank=True, null=True)
    created_by = models.ForeignKey(User, verbose_name=_("Created By"), on_delete=models.DO_NOTHING, blank=True, null=True)
    status = models.CharField(verbose_name=_("Status"), max_length=50, blank=True, null=True)
    retcode = models.IntegerField(_("Return Code"), blank=True, null=True)
    retmsg = models.CharField(verbose_name=_("Return Message"), max_length=500, blank=True, null=True)
    retoutput = models.TextField(verbose_name=_("Return Output"), blank=True, null=True)
    duration = models.IntegerField(_("Duration"), blank=True, null=True)
    task = models.CharField(_('Task'), max_length=500, blank=True, null=True)
    target_type = models.CharField(_('Target Type'), max_length=50, blank=True, null=True)
    target_id = models.IntegerField(_('Target ID'), blank=True, null=True)
    agent_id = models.IntegerField(_('Agent ID'), blank=True, null=True)
    ext_field = JSONField(_("Extras"), null=True, blank=True)

    class Meta:
        indexes = [
            models.Index(fields=['created_date'], name='task_log_created_date_idx'),
            models.Index(fields=['created_by'], name='task_log_created_by_idx'),
            models.Index(fields=['task_id'], name='task_id_idx'),
        ]


class TaskStepLog(models.Model):
    "Database administration service task general log model"
    task_log_id = models.UUIDField(_('Task Log ID'), blank=True, null=True, db_index=True)
    created_date = models.DateTimeField(verbose_name=_("Created Date"), blank=True, null=True)
    status = models.CharField(verbose_name=_("Status"), max_length=50, blank=True, null=True)
    retcode = models.IntegerField(_("Return Code"), blank=True, null=True)
    retmsg = models.TextField(verbose_name=_("Return Message"), max_length=500, blank=True, null=True)
    retoutput = models.TextField(verbose_name=_("Return Output"), blank=True, null=True)
    duration = models.IntegerField(_("Duration"), blank=True, null=True)
    task = models.CharField(_('Task'), max_length=500, blank=True, null=True)
    step = models.CharField(_('Step'), max_length=500, blank=True, null=True)
    target_type = models.CharField(_('Target Type'), max_length=50, blank=True, null=True)
    target_id = models.IntegerField(_('Target ID'), blank=True, null=True)
    task_id = models.IntegerField(_('Task ID'), blank=True, null=True)

    class Meta:
        indexes = [
            models.Index(fields=['created_date'], name='task_step_log_created_date_idx'),
            models.Index(fields=['task_log_id'], name='task_lod_id_idx'),
        ]


class c123_Oracle_Standalone(models.Model):
    """Oracle Standalone Model"""
    org_id = models.ForeignKey(Organization, verbose_name=_("Organization"), on_delete=models.DO_NOTHING, null=False, blank=False, default=1, db_column='org_id')
    created_date  = models.DateTimeField(verbose_name=_("Created Date"), auto_now_add=True, editable=False)
    # relation attributes
    created_by = models.ForeignKey(User, verbose_name=_("Created By"),  related_name='standalone_created_by', on_delete=models.DO_NOTHING, null=False, blank=False)
    host_name = models.CharField(_('Host Name'),blank=True, null=True, max_length=100)
    port = models.CharField(_('Port'),blank=True, null=True, max_length=50)
    name = models.CharField(max_length=100, blank=True, null=True, unique=True)
    database_user = models.CharField(_("Database User"), max_length=100, blank=True, null=True)
    database_pwd = models.CharField(_("Database Password"), max_length=100, null=True, blank=True)
    service_name = models.CharField(_('Service Name'), max_length=50, null=True, blank=True )
    oracle_home = models.CharField(verbose_name=_("Oracle Home"), max_length=1000, blank=True, null=True)
    ind_started = models.BooleanField(verbose_name=_("Is Started"), default=False)
    severity = models.CharField(verbose_name=_("Severity"), max_length=30, default='green', blank=True, null=True)

    class Meta:
         constraints = [
            models.UniqueConstraint(name = 'unique_standalone', fields=['name', 'service_name'])
        ]

    def str(self):
        # host port service_name
        return self.name

class c123_inventory(models.Model):
    """Backups data storage model (for supported backup types - check Inventory enum)"""
    created_date = models.DateTimeField(verbose_name=_("Created Date"), auto_now_add=True)
    created_by = models.ForeignKey(User, verbose_name=_("Created By"), on_delete=models.DO_NOTHING, blank=True, null=True)
    node = models.ForeignKey(c123_node, verbose_name=_("Node Name"), on_delete=models.DO_NOTHING, blank=True, null=True)
    database = models.ForeignKey(c123_database, verbose_name=_("DB Name"), on_delete=models.DO_NOTHING, blank=True, null=True)
    type = models.CharField(_("Type"), max_length=100, blank=True, null=True)
    keep_days = models.PositiveSmallIntegerField(_("keep Days"), default=7, blank=True, null=True)
    name = models.CharField(_("Name"), max_length=100, blank=True, null=True)
    ex_params = JSONField(_("Extras"), null=True, blank=True)
    # status is was successful or not
    status = models.BooleanField(_("Status"), default=False)
    is_sched = models.BooleanField(_("Schedule"), default=False)
    run_log = models.TextField(_("Run Log"), blank=True, null=True)

class ClusterSizing(models.Model):
    """Node resources quota model"""
    db_type = models.CharField(max_length=50, blank=True, null=True, choices=[(dbt.name, dbt.value)
                                                                             for dbt in DataBaseType])
    small_cpu = models.PositiveSmallIntegerField(default=2, validators=[MaxValueValidator(16),
                                                                        MinValueValidator(1)])
    small_disk = models.PositiveSmallIntegerField(default=2, validators=[MaxValueValidator(1000),
                                                                         MinValueValidator(1)])
    small_memory = models.PositiveSmallIntegerField(default=2, validators=[MaxValueValidator(16),
                                                                           MinValueValidator(1)])
    small_backup_keep_days = models.PositiveSmallIntegerField(
        default=2, validators=[MinValueValidator(1), MaxValueValidator(100)])
    small_export_keep_days = models.PositiveSmallIntegerField(
        default=2, validators=[MinValueValidator(1), MaxValueValidator(100)])
    medium_cpu = models.PositiveSmallIntegerField(default=10, validators=[MaxValueValidator(16),
                                                                          MinValueValidator(1)])
    medium_disk = models.PositiveSmallIntegerField(default=4, validators=[MaxValueValidator(1000),
                                                                          MinValueValidator(1)])
    medium_memory = models.PositiveSmallIntegerField(default=14, validators=[MaxValueValidator(16),
                                                                             MinValueValidator(1)])
    medium_backup_keep_days = models.PositiveSmallIntegerField(
        default=7, validators=[MinValueValidator(1), MaxValueValidator(100)])
    medium_export_keep_days = models.PositiveSmallIntegerField(
        default=7, validators=[MinValueValidator(1), MaxValueValidator(100)])
    large_cpu = models.PositiveSmallIntegerField(default=10, validators=[MaxValueValidator(16),
                                                                         MinValueValidator(1)])
    large_disk = models.PositiveSmallIntegerField(default=10, validators=[MaxValueValidator(1000),
                                                                          MinValueValidator(1)])
    large_memory = models.PositiveSmallIntegerField(default=10, validators=[MaxValueValidator(16),
                                                                            MinValueValidator(1)])
    large_backup_keep_days = models.PositiveSmallIntegerField(
        default=30, validators=[MinValueValidator(1), MaxValueValidator(100)])
    large_export_keep_days = models.PositiveSmallIntegerField(
        default=30, validators=[MinValueValidator(1), MaxValueValidator(100)])
    very_large_cpu = models.PositiveSmallIntegerField(default=14, validators=[MaxValueValidator(16),
                                                                              MinValueValidator(1)])
    very_large_disk = models.PositiveSmallIntegerField(default=12, validators=[MaxValueValidator(1000),
                                                                               MinValueValidator(1)])
    very_large_memory = models.PositiveSmallIntegerField(default=14, validators=[MaxValueValidator(1000),
                                                                                 MinValueValidator(1)])
    very_large_backup_keep_days = models.PositiveSmallIntegerField(
        default=60, validators=[MinValueValidator(1), MaxValueValidator(100)])
    very_large_export_keep_days = models.PositiveSmallIntegerField(
        default=60, validators=[MinValueValidator(1), MaxValueValidator(100)])

class SystemConfig(models.Model):
    """Extra settings model"""
    db_type = models.CharField(max_length=50, blank=True, null=True, choices=[(dbt.name, dbt.value)
                                                                            for dbt in DataBaseType])
    backups_path = models.TextField(blank = True, null= True )
    exports_path = models.TextField(blank = True, null = True)

    def save(self, *args, **kwargs):
        if(not self.backups_path):
            match self.db_type:
                case DataBaseType.ORACLE_CDB.name | DataBaseType.ORACLE_PDB.name:
                    self.backups_path = "/backups/oracle/backups"
                case DataBaseType.POSTGRES.name:
                    self.backups_path = "/backups/postgres/backups"
                case DataBaseType.MONGO.name:
                     self.backups_path = "/backups/mongo/backups"
                case DataBaseType.MARIADB.name:
                     self.backups_path = "/backups/mariadb/backups"
                case DataBaseType.MYSQL.name:
                     self.backups_path = "/backups/mysql/backups"
                case DataBaseType.MSSQL.name:
                     self.backups_path = "/backups/mssql/backups"
                case DataBaseType.REDIS.name:
                     self.backups_path = "/backups/redis/backups"
                case DataBaseType.ELASTIC.name:
                     self.backups_path = "/backups/elasticsearch/snapshot/cluster"

        if(not self.exports_path):
            match self.db_type:
                case DataBaseType.ORACLE_CDB.name | DataBaseType.ORACLE_PDB.name:
                    self.exports_path = "/backups/oracle/exports"
                case DataBaseType.POSTGRES.name:
                    self.exports_path = "/backups/postgres/exports"
                case DataBaseType.MONGO.name:
                     self.exports_path = "/backups/mongo/exports"
                case DataBaseType.MARIADB.name:
                     self.exports_path = "/backups/mariadb/exports"
                case DataBaseType.MYSQL.name:
                     self.exports_path = "/backups/mysql/exports"
                case DataBaseType.MSSQL.name:
                     self.exports_path = "/backups/mssql/exports"
                case DataBaseType.REDIS.name:
                     self.exports_path = "/backups/redis/exports"
                case DataBaseType.ELASTIC.name:
                     self.exports_path = "/backups/elasticsearch/snapshot/indices"

        super().save(*args, **kwargs)

class SystemSettings(models.Model):
    """System settings model"""
    name = models.TextField(blank = True, null= True)
    value = models.CharField(blank = True, null = True)

class OracleSystemConfig(SystemConfig):
    """Oracle-specific extension of extra settings model"""
    pdb_path = models.TextField(blank = True, null = True, default = "/123cluster/oracle/pdbs")
    cdb_path =  models.TextField(blank = True, null = True, default = "/123cluster/oracle/cdbs")

    def save(self, *args, **kwargs):
        self.db_type = "ORACLE_CDB"
        super().save(*args, **kwargs)

class OracleClusterSizing(ClusterSizing):
    """Oracle-specific extension of node resources quota model"""
    small_shared_temp = models.PositiveSmallIntegerField(default=1, validators=[MaxValueValidator(100),
                                                                                MinValueValidator(1)])
    medium_shared_temp = models.PositiveSmallIntegerField(default=1, validators=[MaxValueValidator(100),
                                                                                 MinValueValidator(1)])
    large_shared_temp = models.PositiveSmallIntegerField(default=10, validators=[MaxValueValidator(100),
                                                                                 MinValueValidator(1)])
    very_large_shared_temp = models.PositiveSmallIntegerField(default=100, validators=[MaxValueValidator(100),
                                                                                       MinValueValidator(1)])
    small_restore_points_keep_days = models.PositiveSmallIntegerField(
        default=2, blank=True, null=True)
    medium_restore_points_keep_days = models.PositiveSmallIntegerField(
        default=7, blank=True, null=True)
    large_restore_points_keep_days = models.PositiveSmallIntegerField(
        default=30, blank=True, null=True)
    very_large_restore_points_keep_days = models.PositiveSmallIntegerField(
        default=60, blank=True, null=True)

class UserPrivileges(models.Model):
    "User resources quota model"
    # user type
    user = models.ForeignKey(User, on_delete=models.CASCADE, blank=True, null=True)
    # oracle priviliges
    oracle_pdb_quota = models.PositiveSmallIntegerField(default=3)
    oracle_pdb_view = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    oracle_pdb_create = models.CharField(
        max_length=10,
        default=Node_Size_Allowed.SMALL.name,
        choices=[(size.name, size.value) for size in Node_Size_Allowed])
    oracle_pdb_delete = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    # mongo priviliges
    mongo_quota = models.PositiveSmallIntegerField(default=3)
    mongo_view = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    mongo_create = models.CharField(
        max_length=10,
        default=Node_Size_Allowed.SMALL.name,
        choices=[(size.name, size.value) for size in Node_Size_Allowed])
    mongo_delete = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    # mysql priviliges
    mysql_quota = models.PositiveSmallIntegerField(default=3)
    mysql_view = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    mysql_create = models.CharField(
        max_length=10,
        default=Node_Size_Allowed.SMALL.name,
        choices=[(size.name, size.value) for size in Node_Size_Allowed])
    mysql_delete = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    # mssql priviliges
    mssql_quota = models.PositiveSmallIntegerField(default=3)
    mssql_view = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    mssql_create = models.CharField(
        max_length=10,
        default=Node_Size_Allowed.SMALL.name,
        choices=[(size.name, size.value) for size in Node_Size_Allowed])
    mssql_delete = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    # redis priviliges
    redis_quota = models.PositiveSmallIntegerField(default=3)
    redis_view = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    redis_create = models.CharField(
        max_length=10,
        default=Node_Size_Allowed.SMALL.name,
        choices=[(size.name, size.value) for size in Node_Size_Allowed])
    redis_delete = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    # mariadb priviliges
    mariadb_quota = models.PositiveSmallIntegerField(default=3)
    mariadb_view = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    mariadb_create = models.CharField(
        max_length=10,
        default=Node_Size_Allowed.SMALL.name,
        choices=[(size.name, size.value) for size in Node_Size_Allowed])
    mariadb_delete = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
     # postgres priviliges
    postgres_quota = models.PositiveSmallIntegerField(default=3)
    postgres_view = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    postgres_create = models.CharField(
        max_length=10,
        default=Node_Size_Allowed.SMALL.name,
        choices=[(size.name, size.value) for size in Node_Size_Allowed])
    postgres_delete = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    # postgres priviliges
    elasticsearch_quota = models.PositiveSmallIntegerField(default=3)
    elasticsearch_view = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    elasticsearch_create = models.CharField(
        max_length=10,
        default=Node_Size_Allowed.SMALL.name,
        choices=[(size.name, size.value) for size in Node_Size_Allowed])
    elasticsearch_delete = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])

class UserNotifications(models.Model):
    """Notification settings model"""
    # user type
    user = models.ForeignKey(User, on_delete=models.CASCADE, blank=True, null=True)
    email = models.EmailField(blank=True, unique=False)
    # email notification settings
    nodes_resolved_alerts = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
    completed_actions_alerts = models.CharField(
        max_length=3,
        default=User_Or_All_Choice.ONU.name,
        choices=[(allowed.name, allowed.value) for allowed in User_Or_All_Choice])
