import { useEffect, useState } from "react";
import moment from "moment";
import { Chart } from "primereact/chart";
import { timePeriods } from "../../config";
import 'chartjs-adapter-moment';

export default function AgentsChart({timePeriodType, dateKey, points, label, xAxisKey, yAxisKey, dateFormat }) {

    useEffect(() => {
        function getRandomColor(usedColors, backgroundColor) {
            let color = '';
          
            // Generate random hex color code
            do {
              color = '#' + Math.floor(Math.random() * 16777215).toString(16);
            } while (usedColors.includes(color) || isCloseToColor(color, backgroundColor));
          
            return color;
          }
          
          function isCloseToColor(color1, color2) {
            const threshold = 50; // Adjust this value to change the distance threshold
          
            const hex1 = color1.replace('#', '');
            const rgb1 = [parseInt(hex1.substring(0, 2), 16), parseInt(hex1.substring(2, 4), 16), parseInt(hex1.substring(4, 6), 16)];
          
            const hex2 = color2.replace('#', '');
            const rgb2 = [parseInt(hex2.substring(0, 2), 16), parseInt(hex2.substring(2, 4), 16), parseInt(hex2.substring(4, 6), 16)];
          
            const distance = Math.sqrt((rgb1[0] - rgb2[0]) ** 2 + (rgb1[1] - rgb2[1]) ** 2 + (rgb1[2] - rgb2[2]) ** 2);
          
            return distance <= threshold;
          }

        const timeFilter = (data, timePeriod) => {

            let timeUnit;
            let stepSize;

            switch (timePeriod) {

                case timePeriods.TEN_MIN:
                    timeUnit = "minute";
                    stepSize = 2;
                    break;
                case timePeriods.HOUR:
                    timeUnit = "minute";
                    stepSize = 10;
                    break;
                case timePeriods.SIX_HOURS:
                    timeUnit = "hour";
                    stepSize = 1.5;
                    break;
                case timePeriods.WEEK:
                    timeUnit = "day";
                    stepSize = 1;
                    break;
                case timePeriods.DAY:
                    timeUnit = "hour";
                    stepSize = 6;
                    break;
                default:
                    timeUnit = "minute";
                    stepSize = 2;
                    break;


            }

            const _filteredPoints = data?.filter((n) => {
                if (timePeriod === timePeriods.TEN_MIN) {
                    if (moment().diff(moment(n?.[dateKey]), "minutes") < 10)
                        return true;
                }
                else if (timePeriod === timePeriods.HOUR) {
                    if (moment().diff(moment(n?.[dateKey]), "hours") < 1)
                        return true;
                }
                else if (timePeriod === timePeriods.SIX_HOURS) {
                    if (moment().diff(moment(n?.[dateKey]), "hours") < 6)
                        return true;
                }
                else if (timePeriod === timePeriods.DAY) {
                    if (moment().diff(moment(n?.[dateKey]), "days") < 1)
                        return true;
                }
                else if (timePeriod === timePeriods.WEEK) {
                    if (moment().diff(moment(n?.[dateKey]), "days") < 7)
                        return true;
                }
                return false;
            });

            setChartOptions({
                responsive: true,
                maintainAspectRatio: false,
                plugins: {
                    legend: {
                        display: true,
                        labels: { boxWidth: 1, boxHeight : 1}
                    }
                },
                parsing: {
                    xAxisKey: xAxisKey,
                    yAxisKey: yAxisKey
                },
                scales: {
                    x: {
                        type: 'time',
                        bounds: 'ticks',
                        time: {
                            unit: timeUnit,
                            stepSize: stepSize,
                            displayFormats: {
                                day: dateFormat,
                                hour: dateFormat,
                                minute: dateFormat
                            }
                        },
                        ticks: {
                            color: 'white',
                            source: "labels",
                            precision: 0,
                        }
                    },
                    y: {
                        min: 0,
                        ticks: {
                            color: 'white',
                            precision: 0,
                        }
                    }
                }
            })
            return _filteredPoints;


        }

        let labels;
        let now = moment();
        switch (timePeriodType) {
            case timePeriods.TEN_MIN:
                labels = [now.clone().subtract(600, "seconds"), now.clone().subtract(400, "seconds"), now.clone().subtract(200, "seconds"), now];
                break;
            case timePeriods.HOUR:
                labels = [now.clone().subtract(60, "minutes"), now.clone().subtract(40, "minutes"), now.clone().subtract(20, "minutes"), now];
                break;
            case timePeriods.SIX_HOURS:
                labels = [now.clone().subtract(6, "hours"), now.clone().subtract(4, "hours"), now.clone().subtract(2, "hours"), now];
                break;
            case timePeriods.DAY:
                labels = [now.clone().subtract(24, "hours"), now.clone().subtract(16, "hours"), now.clone().subtract(8, "hours"), now];
                break;
            case timePeriods.WEEK:
                labels = [now.clone().subtract(168, "hours"), now.clone().subtract(112, "hours"), now.clone().subtract(56, "hours"), now];
                break;
            default:
                labels = [now.clone().subtract(600, "seconds"), now.clone().subtract(400, "seconds"), now.clone().subtract(200, "seconds"), now];
                break;

        }

        const chartData = timeFilter(points, timePeriodType);

        const datasets_data = chartData?.reduce((acc, obj) => {
            const dataset = obj.column1;
            if (!acc[dataset]) {
                acc[dataset] = [];
            }
            acc[dataset].push(obj);
            return acc;
        }, {});

        const usedColors = [];

        const datasets = [];

        // Set the first dataset's borderColor to lightgreen
        let isFirstDataset = true;
        const keys = datasets_data ? Object.keys(datasets_data) : [];
        if(keys.length < 2) setChartOptions(chartOptions => ({

            ...chartOptions, plugins: {...chartOptions.plugins, legend: { ...chartOptions.plugins.legend, labels: { boxWidth: 0}}}

        }))
        else setChartOptions(chartOptions => ({

            ...chartOptions, plugins: {...chartOptions.plugins, legend: { ...chartOptions.plugins.legend, position: "bottom"}}

        }))

        keys.forEach((k) => {
            const color = isFirstDataset ? 'lightgreen' : getRandomColor(usedColors, '#325b76');
            usedColors.push(color);
            const datasetLabel = keys?.length < 2 ? label: datasets_data[k][0].column1; 
            datasets.push({
                label: datasetLabel,
                data: datasets_data[k],
                fill: false,
                borderColor: color,
                backgroundColor: '#325b76',
                pointStyle: 'line',
                tension: 0.1
            });

            isFirstDataset = false;
        });
        setChartPoints({
            labels: labels,
            datasets: datasets
        });


    }, [timePeriodType, label, points, xAxisKey, yAxisKey, dateKey, dateFormat])

    const [chartPoints, setChartPoints] = useState(null);

    const [chartOptions, setChartOptions] = useState(null);

    return(<Chart type="line" data={chartPoints} options={chartOptions} />);


}