import { useState, useEffect } from "react";
import { Button } from "primereact/button";
import { Dropdown } from "primereact/dropdown";
import { InputText } from "primereact/inputtext";
import refreshTooltip from "../../refreshTooltip";
import { dbTypes, timePeriodTypes } from "../../../config";
import useInterval from "../../../util/useInterval";
import { useSelector, useDispatch } from "react-redux";
import { LogActions } from "../../../store/log/reducer";
import { Checkbox } from "primereact/checkbox";
import { ButtonGroup } from 'primereact/buttongroup';

const LogFilter = (props) => {

    const { db_type, time_period } = useSelector((state) => state.app);

    const [search, setSearch] = useState("");
    const [dbType, setDbType] = useState(db_type);
    const [timePeriodType, setTimePeriodType] = useState(time_period);
    const [logType, setLogType] = useState("task");
    const [status, setStatus] = useState("");
    const [autoRefresh, setAutoRefresh] = useState(false);
    const dispatch = useDispatch();

    useInterval(() => {
        if (timePeriodType) dispatch(LogActions.getLogs({ time_period: timePeriodType }));
    }, autoRefresh ? 30000 : null);

    const logTypes = [
        { label: 'Task Logs', value: 'task' },
        { label: 'System Log', value: 'system' }
    ];

    useEffect(() => {
        const storedSearch = localStorage.getItem("FilterSearch");
        const storedDbType = localStorage.getItem("FilterDbType");
        const storedTimePeriodType = localStorage.getItem("FilterTimePeriodType");
        const storedLogType = localStorage.getItem("FilterLogType");
        const storedStatus = localStorage.getItem("FilterStatus");
        const storedAutoRefresh = localStorage.getItem("FilterAutoRefresh");

        if (storedSearch) setSearch(storedSearch);
        if (storedDbType) setDbType(storedDbType);
        if (storedTimePeriodType) setTimePeriodType(storedTimePeriodType);
        if (storedLogType) setLogType(storedLogType);
        if (storedStatus) setStatus(storedStatus);
        if (storedAutoRefresh) setAutoRefresh(storedAutoRefresh === "true");
    }, []);

    useEffect(() => {
        localStorage.setItem("FilterSearch", search);
        localStorage.setItem("FilterDbType", dbType);
        localStorage.setItem("FilterTimePeriodType", timePeriodType);
        localStorage.setItem("FilterLogType", logType);
        localStorage.setItem("FilterStatus", status);
        localStorage.setItem("FilterAutoRefresh", autoRefresh);
    }, [search, dbType, timePeriodType, logType, status, autoRefresh]);

    const onStatusClick = (status) => () => {
        setStatus(status);
        if (props.onStatusChange)
            props.onStatusChange(status, search, timePeriodType, dbType, logType);

    }

    const onDbChange = (e) => {
        setDbType(e.value);
        if (props.onDbChange)
            props.onDbChange(e.value, search, timePeriodType, status, logType);
    }

    const onTimePeriodChange = (e) => {
        setTimePeriodType(e.target.value);
        if (props.onTimePeriodChange)
            props.onTimePeriodChange(e.value, search, status, dbType, logType);
    }

    const onLogTypeChange = (e) => {
        setLogType(e.value);
        if (props.onLogTypeChange)
            props.onLogTypeChange(e.value, search, status, dbType, logType);
    }

    const onSearchChange = (e) => {
        setSearch(e.target.value);
        if (props.onSearchChange)
            props.onSearchChange(e.target.value.toLowerCase(), timePeriodType, status, dbType, logType);

    }

    const onRefreshClick = () => {
        window.location.reload();
    }


    return (
        <div className="filter-component-wrapper">
            <div className="filter-component">
                <div className="filter-left-container flex flex-grow-0">
                    <div className="filter-title">Filter by:</div>
                    <Dropdown
                        className="p-inputtext-sm filter-field"
                        value={dbType}
                        options={dbTypes}
                        onChange={onDbChange}
                        placeholder="DB Types" />


                    <span className="p-input-icon-right">
                        <i className={search ? "pi pi-times" : "field-collapsed"} onClick={() => {
                            setSearch("");
                            if (props.onSearchChange)
                                props.onSearchChange(timePeriodType, status);

                        }} ></i>
                        <InputText className="p-inputtext-sm filter-field ml-2" placeholder="Search" tooltip="Search by task name, node type, log status, log created date" tooltipOptions={{ position: 'bottom' }} value={search} onChange={onSearchChange} />


                    </span>

                    <Dropdown
                        className="p-inputtext-sm filter-field ml-2 mr-auto"
                        value={logType}
                        options={logTypes}
                        onChange={onLogTypeChange}
                        placeholder="Log Type" />
                </div>

                <ButtonGroup>

                    <Button className={props.showSeverityGreen ? "severity-green toggled p-button-sm" : "severity-green p-button-sm"} onClick={onStatusClick("green")} label={`${props.statusGreenTotal}`} />
                    <Button className={props.showSeverityYellow ? "severity-yellow toggled p-button-sm" : "severity-yellow p-button-sm"} onClick={onStatusClick("yellow")} label={`${props.statusYellowTotal}`} />
                    <Button className={props.showSeverityRed ? "severity-red toggled p-button-sm" : "severity-red p-button-sm"} onClick={onStatusClick("red")} label={`${props.statusRedTotal}`} />

                </ButtonGroup>

                <div className="flex flex-row ml-auto filter-right-container">

                    <Dropdown
                        className={"filter-field time-period p-inputtext-sm"}
                        value={timePeriodType}
                        options={timePeriodTypes}
                        onChange={onTimePeriodChange}
                        placeholder="Time Period" />

                    <div className="align-self-center ml-2 mr-2">


                        <Checkbox inputId="autoupdate_cb" className="ml-1" checked={autoRefresh} value={autoRefresh} onChange={() => {
                            setAutoRefresh(!autoRefresh);
                        }} />
                        <label htmlFor="autoupdate_cb" className="p-checkbox-label"> Auto refresh</label>

                    </div>


                    <Button className={"filter-button refresh-button p-button-sm"} onClick={onRefreshClick} icon="pi pi-refresh" iconPos="right" />
                </div>

            </div >
            {refreshTooltip}
        </div>

    );
}


export default LogFilter;
