import React, { useEffect, useState } from 'react';
import { Chart } from 'primereact/chart';

const getRandomColor = () => '#' + Math.floor(Math.random() * 16777215).toString(16).padStart(6, '0');

const adjustColorBrightness = (hex, brightnessFactor) => {
  if (!hex || hex.length < 6) return '#cccccc';
  hex = hex.replace('#', '');
  if (hex.length === 3) {
      hex = hex[0] + hex[0] + hex[1] + hex[1] + hex[2] + hex[2];
  }
  let r = parseInt(hex.substring(0, 2), 16);
  let g = parseInt(hex.substring(2, 4), 16);
  let b = parseInt(hex.substring(4, 6), 16);

  r = Math.min(255, Math.max(0, Math.round(r * brightnessFactor)));
  g = Math.min(255, Math.max(0, Math.round(g * brightnessFactor)));
  b = Math.min(255, Math.max(0, Math.round(b * brightnessFactor)));

  return `#${r.toString(16).padStart(2, '0')}${g.toString(16).padStart(2, '0')}${b.toString(16).padStart(2, '0')}`;
};

export default function PieChart({ data, labelKey, valueKey }) {
  const [chartData, setChartData] = useState(null);
  const [chartOptions, setChartOptions] = useState(null);

  const OTHER_CATEGORY_COLOR = '#aaaaaa';

  useEffect(() => {
    if (!data || data.length === 0) {
        setChartData(null);
        return;
    }


    const sortedData = [...data].sort((a, b) => {
        const valA = typeof a[valueKey] === 'number' ? a[valueKey] : 0;
        const valB = typeof b[valueKey] === 'number' ? b[valueKey] : 0;
        return valB - valA;
    });

    let finalLabels = [];
    let finalValues = [];
    let backgroundColors = [];
    let hoverBackgroundColors = [];

    if (sortedData.length > 4) {
        const top4Data = sortedData.slice(0, 4);

        const otherSum = sortedData
            .slice(4)
            .reduce((sum, item) => sum + (typeof item[valueKey] === 'number' ? item[valueKey] : 0), 0);

        finalLabels = [...top4Data.map((item) => item[labelKey]), 'Other'];
        finalValues = [...top4Data.map((item) => item[valueKey]), otherSum];

        backgroundColors = top4Data.map(() => getRandomColor());
        backgroundColors.push(OTHER_CATEGORY_COLOR);

    } else {
        finalLabels = sortedData.map((item) => item[labelKey]);
        finalValues = sortedData.map((item) => item[valueKey]);
        backgroundColors = sortedData.map(() => getRandomColor());
    }


    hoverBackgroundColors = backgroundColors.map(color => adjustColorBrightness(color, 0.8));

    setChartData({
      labels: finalLabels,
      datasets: [
        {
          data: finalValues,
          backgroundColor: backgroundColors,
          hoverBackgroundColor: hoverBackgroundColors,
        },
      ],
    });


    setChartOptions({
      responsive: true,
      maintainAspectRatio: false,
      plugins: {
        tooltip: {
            callbacks: {
                label: function(context) {
                    let label = context.label || '';
                    if (label) {
                        label += ': ';
                    }
                    if (context.parsed !== null) {
                        label += context.parsed;
                    }
                    return label;
                }
            }
        },
        legend: {
          position: 'right',
          labels: {
            color: 'white',
            padding: 15,
            boxWidth: 12,
            usePointStyle: false
          }
        },
      },
    });

  }, [data, labelKey, valueKey]);

  if (!chartData) {
    return <div>No data available for pie chart.</div>;
  }

  return (
      <div style={{ position: 'relative', height: '300px', width: '100%' }}>
          <Chart type="pie" data={chartData} options={chartOptions} />
      </div>
  );
}
