import { useDispatch, useSelector } from "react-redux";
import { useEffect, Fragment, useState, useRef } from "react";
import { DataTable } from "primereact/datatable";
import { Column } from "primereact/column";
import { InputText } from "primereact/inputtext";
import { Button } from "primereact/button";
import { StatusActions } from "../../../store/status/reducer";
import ApplySettingsDialog from "../../../dialogs/applySettingsDialog";
import TooltipCell from "../../tooltipCell";


export default function CDBSettingsTable({ node }) {

    const dispatch = useDispatch();
    const { cdb_settings } = useSelector((state) => state.status);
    const [filteredSettings, setFilteredSettings] = useState([]);
    const [changedSettings, setChangedSettings] = useState([]);
    const [initialSettings, setInitialSettings] = useState([]);
    const [pendingSettings, setPendingSettings] = useState([]);
    const [settingSearch, setSettingSearch] = useState("");
    const [applySettings, setShowApplySettings] = useState(false);
    const [applyDisabled, setApplyDisabled] = useState(true);
    const [revertDisabled, setRevertDisabled] = useState(true);
    const dataTableRef = useRef(null);
    const [pageState, setPageState] = useState({ first: 0, rows: 10 });

    const handlePageChange = (event) => {
        setPageState({ first: event.first, rows: event.rows });
    };
    

    const searchFilter = (data, search) => {
        return data?.filter ? data?.filter((s) => {
            if (s?.['NAME']?.toLowerCase().includes(search) ||
                s?.['VALUE']?.toLowerCase().includes(search) ||
                s?.['ISDEFAULT']?.toLowerCase().includes(search) ||
                s?.['ISSYS_MODIFIABLE']?.toLowerCase().includes(search)
            )
                return true;
            return false;
        }) : [];
    }

    const settingBodyTemplate = (rowData, props) => <TooltipCell id={`${props.field}-${props.rowIndex}`} content={rowData['VALUE']} />
    const onApplySettings = (data) => dispatch(StatusActions.editCdbSettings(data, setShowApplySettings))

    useEffect(() => {
        if (node?.id) {

            if (!cdb_settings) dispatch(StatusActions.getCdbSettings(node?.id));

            if (cdb_settings?.length > 0) {
                const editableData = structuredClone(cdb_settings);
                setFilteredSettings(editableData);
                setChangedSettings(editableData);
            }

        }


    }, [dispatch, cdb_settings, node?.id]);

    useEffect(() => {
        setFilteredSettings(searchFilter(changedSettings, settingSearch));
    }, [changedSettings, settingSearch])


    const settingEditor = (options) => {
        return (<InputText value={options.value} onChange={(e) => options.editorCallback(e.target.value)} />)

    }

    const onCellEditComplete = (e) => {
        let { rowData, newValue, field } = e;
        const setting = cdb_settings?.filter((s) => s?.['NAME'] === rowData['NAME'])
        if (setting?.length > 0 && setting[0]['VALUE'] !== newValue)
            rowData['COLOR'] = 'red'
        else
            rowData['COLOR'] = 'default';
        const index = changedSettings?.findIndex((s) => s?.['NAME'] === rowData['NAME'])
        changedSettings[index] = rowData;
        const _pendingSettings = changedSettings?.filter((s) => s?.['COLOR'] === 'red');
        setChangedSettings(changedSettings);
        setPendingSettings(_pendingSettings);
        setApplyDisabled(_pendingSettings?.length < 1);
        if (rowData[field] !== newValue) rowData[field] = newValue
    }

    const actionBodyTemplate = (rowData, props) => {
        return (
            <div className="flex">
                <div className="mt-2">{rowData?.['ISSYS_MODIFIABLE'] === 'FALSE' ? "Yes" : "No"}</div>
                <Button icon="pi pi-pencil" severity="success" rounded size="small" onClick={() => {
                  const renderedRowIndex = (props.rowIndex - (pageState.first / pageState.rows) * pageState.rows) + 1;
                  dataTableRef.current?.getTable()?.rows?.[renderedRowIndex]?.cells[1]?.click()
                }} />
            </div>
        );
    }

    return (
        <Fragment>

            <div className="flex" >

                <span className="p-input-icon-right">
                    <i className={settingSearch ? "pi pi-times" : "field-collapsed"} onClick={() => setSettingSearch("")} />
                    
                    <InputText className="p-inputtext-sm filter-field mb-2"
                        placeholder="Search"
                        value={settingSearch}
                        onChange={(e) => {
                            setSettingSearch(e.target.value);
                        }} />
                </span>

                <div className="ml-auto">

                    <Button className="mr-2" label="Revert" disabled={revertDisabled} size="small" icon="pi pi-times" iconPos="right" onClick={() => {
                        setChangedSettings(cdb_settings);
                        setApplyDisabled(true);
                        setRevertDisabled(true);
                    }} />

                    <Button label="Apply" disabled={applyDisabled} size="small" icon="pi pi-check" iconPos="right" onClick={() => {
                        setInitialSettings(cdb_settings?.filter((s) => pendingSettings.filter((p) => p["NAME"] === s["NAME"])?.length > 0));
                        setShowApplySettings(true);
                    }} />

                </div>
            </div>



            <DataTable ref={dataTableRef} dataKey="NAME" editMode="cell" emptyMessage={<div className="text-sm pl-3">No parameters data</div>} className="monitors-table" value={filteredSettings}
                paginator={filteredSettings?.length > 10} rows={10} rowsPerPageOptions={[10, 20, 50, 100]} alwaysShowPaginator={false}
                paginatorTemplate="FirstPageLink PrevPageLink PageLinks NextPageLink LastPageLink RowsPerPageDropdown CurrentPageReport"
                currentPageReportTemplate="Showing {first} to {last} of {totalRecords} parameters"
                paginatorPosition="bottom"
                onPage={handlePageChange}
                first={pageState.first}
                rowClassName={(rowData) => {
                    return rowData['COLOR'] === 'red' ? 'text-red-500' : '';
                }}
            >

                <Column field="NAME" header="Name" style={{width: "40%"}} sortable ></Column>
                <Column field="VALUE" header="Setting" style={{width: "30%"}} body={settingBodyTemplate} editor={(options) => settingEditor(options)} onCellEditComplete={(options) => onCellEditComplete(options)} sortable></Column>
                <Column header="Restart required" body={actionBodyTemplate} ></Column>
            </DataTable>
            {applySettings ? <ApplySettingsDialog
                visible={applySettings}
                initial={initialSettings}
                settings={pendingSettings}
                node={node}
                onHide={() => setShowApplySettings(false)}
                onOkClick={onApplySettings}
            /> : null
            }
        </Fragment>
    );
}