import { useDispatch, useSelector } from "react-redux";
import { useEffect, Fragment, useState} from "react";
import { DataTable } from "primereact/datatable";
import { Column } from "primereact/column";
import { Button } from "primereact/button";
import { Dropdown } from "primereact/dropdown";
import { Checkbox } from "primereact/checkbox";
import { agentTypes } from "../../../config";
import moment from "moment";
import { ConfirmPopup } from "primereact/confirmpopup";
import RegisterAgentDialog from "../../../dialogs/monitoring/registerAgentDialog";
import { MonitoringActions } from "../../../store/monitoring/reducer";
import { timePeriodTypes } from "../../../config";
import { intervalByTimedelta } from "../nodeAlertsTable";
import AgentsChart from "../../agentsChart";


export default function NodeAgentsTable (props) {

    const dispatch = useDispatch();
    const { agents, agents_chart, agents_summary } = useSelector((state) => state.monitoring);
    const [popupElement, setPopupElement] = useState(false);
    const [agentToDelete, setAgentToDelete] = useState(false);
    const [showDeleteConfirm, setShowDeleteConfirm] = useState(false);
    const [showRegisterAgent, setShowRegisterAgent] = useState(false);
    const [filteredAgents, setFilteredAgents] = useState([]);
    const [agentInterval, setAgentInterval] = useState(0);
    const [timePeriodType, setTimePeriodType] = useState(timePeriodTypes[0].value);

    useEffect(() => {
        if(!agents) dispatch(MonitoringActions.getAgents());
        else if(agents?.length) setFilteredAgents(agents);

    }, [dispatch, agents]);

    useEffect(() => 
    {
     dispatch(MonitoringActions.getAgentsSummary());
     dispatch(MonitoringActions.getAgentsChart({"ind_interval": agentInterval}))
}, [agentInterval, dispatch])


    const onRegisterAgent = (data) => {
        dispatch(MonitoringActions.addAgent(data, setShowRegisterAgent));
    }

    const roleBodyTemplate = (rowData) => (<Dropdown className="p-inputtext-sm filter-field" value={rowData.role} options = {agentTypes} />);

    const lastHeartbeatBodyTemplate = (rowData) => {
        return rowData["Last heartbeat"] ? moment(rowData["Last heartbeat"]).format("MMM DD,YYYY, hh:mm A") : "no data";
    }

    const isDefaultBodyTemplate = (rowData) => {
        return <Checkbox disabled value={rowData.ind_default === "Y"} checked={rowData.ind_default === "Y"}/>
    }

    const activeBodyTemplate = (rowData) => {
        return <Checkbox disabled value={rowData.ind_active === "Y"} checked={rowData.ind_active === "Y"}/>;
    }


    const actionBodyTemplate = (rowData) => {
        return (
            <div className = "flex">
                <Button icon = "pi pi-trash" className="p-button-danger p-button-rounded ml-auto" id ={`delete_agent_${rowData.agent_id}`}  onClick = {() => {
                    setAgentToDelete(rowData);
                    setPopupElement(window.document.getElementById(`delete_agent_${rowData.agent_id}`));
                    setShowDeleteConfirm(true);
                    }}/>
            </div>
        );
    }

    const roleEditor = (options) =>
    {
        return (<Dropdown className="p-inputtext-sm filter-field" value={options.value} options = {agentTypes} onChange={(e) =>{options.editorCallback(e.value)}} />)

    }

    const checkboxEditor = (options) =>
    {
        return (<Checkbox value={options.value} checked = {options.value === "Y"} onChange={(e) => options.editorCallback(e.value === "Y" ? "N" : "Y")} />)

    }

    const onCellEditComplete = (e) => {
        let {rowData, newRowData, field} = e;
        if (rowData[field] !== newRowData[field]) dispatch(MonitoringActions.editAgent(newRowData, field));
        
    }



    return (
        <Fragment>
            <div className="flex flex-column mb-2">
                <div className="flex">
                <Dropdown className="p-inputtext-sm filter-field time-period mr-2 mb-2" 
             options={timePeriodTypes}  value={timePeriodType} onChange = {(e) => {
              setAgentInterval(intervalByTimedelta(e.value));
              setTimePeriodType(e.value)}} />
             
              <div className="ml-2 mr-1" style={{width: '48px', height:'48px'}}>
              <Button className="p-button-rounded p-button-info"  icon = "pi pi-plus" onClick={() => setShowRegisterAgent(true)}/>
              </div>
              <div className="ml-1 mr-2" style={{width: '48px', height:'48px'}}>
              <Button className="p-button-rounded p-button-secondary" icon = "pi pi-refresh" onClick={() => dispatch(MonitoringActions.setAgents(null))}/>
              </div>
             
              <div className="text-lg multiline">{agents_summary}</div>
                </div>
              <AgentsChart timePeriodType = {timePeriodType} dateFormat = "MM/DD/YYYY, HH:mm" label = {"Total monitors"} xAxisKey = {"Time"} yAxisKey = {"Total monitors"} dateKey = {"Time"} points = {agents_chart?.length ?  agents_chart : undefined}/>
            </div>
            
             
                <DataTable editMode="cell" emptyMessage = {<td className="text-sm pl-3">No agents data</td>} className="agents-table" value={filteredAgents}
                    paginator = {filteredAgents?.length > 10} rows={10}
                    paginatorTemplate="FirstPageLink PrevPageLink PageLinks NextPageLink LastPageLink CurrentPageReport"
                    currentPageReportTemplate="Showing {first} to {last} of {totalRecords} agents"
                    paginatorPosition="bottom">
                    <Column  field="status_text" header="Status" ></Column>
                    <Column  field="ip" header="IP address" sortable></Column>
                    <Column  field="Last heartbeat" header="Last heartbeat" body = {lastHeartbeatBodyTemplate} sortable></Column>
                    <Column  field="role" header="Role" cellEditValidator={(e) => { if (e.originalEvent.target.nodeName === 'LI') { return false; } return true; }} editor ={(options) => roleEditor(options)}  body={roleBodyTemplate} onCellEditComplete = {(options) => onCellEditComplete(options)} sortable></Column>
                    <Column  field="ind_default" header="Is default" editor = {(options) => checkboxEditor(options)} onCellEditComplete = {(options) => onCellEditComplete(options)}  body = {isDefaultBodyTemplate} sortable></Column>
                    <Column  field="ind_active" header="Active" editor = {(options) => checkboxEditor(options)} onCellEditComplete = {(options) => onCellEditComplete(options)} body = {activeBodyTemplate} sortable></Column>
                    <Column body={actionBodyTemplate}></Column>
                </DataTable>
                <ConfirmPopup onHide={() => setShowDeleteConfirm(false)} visible = {showDeleteConfirm} target={popupElement} acceptClassName = "p-button-danger" rejectIcon = "pi pi-times" message = "This will delete agent" icon = "pi pi-exclamation-triangle" acceptIcon = "pi pi-trash" acceptLabel="Confirm" rejectLabel="Cancel" reject={() => setShowDeleteConfirm(false)} accept = {() => dispatch(MonitoringActions.deleteAgent(agentToDelete.agent_id))} />
                {showRegisterAgent ? <RegisterAgentDialog visible= {showRegisterAgent} onHide = {() => setShowRegisterAgent(false)} onOkClick = {onRegisterAgent}/> : null}
                </Fragment>
    );
}