import { getColorByAlertSeverity } from "../../../util/colorBySeverity";
import MonitoringItems from '../../../dialogs/monitoring/monitorItems';
import { Checkbox } from 'primereact/checkbox';
import { useDispatch, useSelector } from "react-redux";
import { useState, useEffect, useMemo } from "react";
import { DataTable } from "primereact/datatable";
import { Column } from "primereact/column";
import timelineBodyTemplate from "../../../util/timelines";
import { MonitoringActions } from "../../../store/monitoring/reducer";
import { loadingMsg, timePeriods } from "../../../config";
import moment from "moment";
import React from 'react';

export const timeFilter = (data, timePeriod) => {
    return data?.filter ? data?.filter((n) => {
        if (timePeriod === timePeriods.TEN_MIN) {
            if (moment().diff(moment(n.date_created_with_tz), "minutes") < 10)
                return true;
        }
        else if (timePeriod === timePeriods.HOUR) {
            if (moment().diff(moment(n.date_created_with_tz), "hours") < 1)
                return true;
        }
        else if (timePeriod === timePeriods.SIX_HOURS) {
            if (moment().diff(moment(n.date_created_with_tz), "hours") < 6)
                return true;
        }
        else if (timePeriod === timePeriods.DAY) {
            if (moment().diff(moment(n.date_created_with_tz), "days") < 1)
                return true;
        }
        else if (timePeriod === timePeriods.WEEK) {
            if (moment().diff(moment(n.date_created_with_tz), "days") < 7)
                return true;
        }
        return false;
    }) : [];
}

export const intervalByTimedelta = (timedelta) => {
     switch (timedelta) {

        case timePeriods.TEN_MIN:
            return 0;
        case timePeriods.HOUR:
            return 1;
        case timePeriods.SIX_HOURS:
            return 2;
        case timePeriods.DAY:
            return 3;
        case timePeriods.WEEK:
            return 4;
        default:
            return 4;


    }
}


export function NodeAlertsTable({
    node,
    grouping,
    noHeader,
    noHeadings,
    monitorInterval,
    expanded,
    database,
    timelineVisible,
    timeVisible = true,
    filteredAlerts,
    groupSwitch,
    toggleTimeline,
    parentSortField = null,
    parentSortOrder = null
}) {

    const dispatch = useDispatch();
    const [selectedRecord, setSelectedRecord] = useState(null);
    const [showAlertHistory, setShowAlertHistory] = useState(false);
    const { isLoading, history_per_alerts } = useSelector((state) => state.monitoring);

    useEffect(() => {
        if (node) dispatch(MonitoringActions.getNodeAlerts({ "node_id": node.id, "node_type": node.type, "grouping": grouping }))
    }, [node, grouping, dispatch]);

    useEffect(() => {
        if (timelineVisible && filteredAlerts && monitorInterval != null)
            dispatch(MonitoringActions.getHistoryPerAlerts(filteredAlerts, monitorInterval));
    }, [filteredAlerts, timelineVisible, monitorInterval, dispatch]);


    const sortedFilteredAlerts = useMemo(() => {
        if (parentSortField === 'status' && parentSortOrder !== null && Array.isArray(filteredAlerts)) {

            const getAlertSeverityValue = (severity) => Math.abs(severity || 0);

            const alertsCopy = [...filteredAlerts];
            alertsCopy.sort((a, b) => {
                const severityA = getAlertSeverityValue(a.severity);
                const severityB = getAlertSeverityValue(b.severity);
                const comparison = severityA - severityB;
                return comparison * parentSortOrder;
            });
            return alertsCopy;
        }
        return filteredAlerts;
    }, [filteredAlerts, parentSortField, parentSortOrder]);


    const createdDateBodyTemplate = (item) => {
        return (<div className="nowrap-cell">{moment(item.date_created).format("MMM DD,YYYY, hh:mm A")}</div>);
    }

    const whatBodyTemplate = (item) => {
        return <div className={noHeader ? (expanded && database) ? 'parent-leaf' : expanded ? 'child-leaf' : '' : `severity-index-${getColorByAlertSeverity(item["severity"])}-row-outer`}>
            {item.monitor_name} (monitor)
        </div>
    }

    return (<div>
        <DataTable onSelectionChange={(e) => {
               const selectedAlert = e.value;
               setSelectedRecord(selectedAlert);
               if(selectedAlert) {
                dispatch(MonitoringActions.getAlertHistory({
                    "org_id": 1,
                    "node_id": selectedAlert.node_id,
                    "monitor_name": selectedAlert.monitor_name,
                    "node_type": selectedAlert.node_type,
                    "ind_interval": monitorInterval
                }));
                    dispatch(MonitoringActions.getMonitors({ 'org_id': 1, 'node_id': selectedAlert.node_id, "monitor_name": selectedAlert.monitor_name, "node_type": selectedAlert.node_type }));
                    setShowAlertHistory(true);
               }
        }} scrollable scrollHeight="flex"
            header={noHeader ? null : (
                <div className='flex align-items-start'>
                    {toggleTimeline && timelineVisible !== undefined && (
                        <>
                            <Checkbox id={`timeline-cb-${node?.id || 'default'}`} className='mr-1' checked={timelineVisible} onChange={toggleTimeline} />
                            <label className='p-checkbox-label' htmlFor={`timeline-cb-${node?.id || 'default'}`}>Timeline</label>
                        </>
                    )}
                </div>
            )}
            rowClassName={(rowData) => `severity-index-${getColorByAlertSeverity(rowData["severity"])}-row`}
            emptyMessage={isLoading ? <div className="text-sm empty-severity">{loadingMsg}</div> : <div className="text-sm empty-severity">No alerts data</div>}
            selectionMode="single" selection={selectedRecord} dataKey="alert_id"
            className={`severity-table ${noHeadings ? 'no-headings' : ''} alerts`}
            rowsPerPageOptions={[10, 20, 50, 100]} paginator alwaysShowPaginator={false} rows={100}
            value={sortedFilteredAlerts}
            >
            <Column
              field="monitor_name"
              style={{ width: "25%", maxWidth: "25%" }}
              header={groupSwitch ?
                React.cloneElement(groupSwitch, {},
                  React.cloneElement(groupSwitch.props.children[0], {
                    style: {
                      ...groupSwitch.props.children[0].props.style,
                       transform: 'translateX(0%)'
                    }
                  }),
                  groupSwitch.props.children[1]
                ) : "What"}
              body={whatBodyTemplate}
              sortable
            ></Column>
            <Column field="location" style={{ width: "22%", maxWidth: "22%" }} body={(item) => <div>{item.location}</div>} header="Location" sortable></Column>
            <Column field="severity" style={{ width: "25%", maxWidth: "25%" }} body={(item) => <div style={{ color: getColorByAlertSeverity(item["severity"]) }}>{getColorByAlertSeverity(item["severity"])?.toUpperCase()}</div>} header="Status" sortable></Column>
              {timeVisible ? <Column field="date_created_with_tz" style={{ width: "24%", maxWidth: "24%" }} body={createdDateBodyTemplate} header="Date" sortable></Column> : null}
              {timelineVisible && <Column field="timeline_data" style={{ width: "10%", maxWidth: "10%" }}
                header="Timeline"
                body={(item) => timelineBodyTemplate(item, history_per_alerts, monitorInterval)}
                sortable>

            </Column>}
        </DataTable>

        {showAlertHistory ? <MonitoringItems
            visible={showAlertHistory} monitorInterval={monitorInterval} onHide={() => setShowAlertHistory(false)}
            currentItem={selectedRecord}
        /> : null}
    </div>)
}
