import { useState, useMemo } from "react";
import TooltipCell from "../../tooltipCell";
import { useNavigate } from "react-router-dom";
import { types } from "../../../config";
import { loadingMsg } from "../../../config";
import { DataTable } from "primereact/datatable";
import { Column } from "primereact/column";
import { useSelector } from "react-redux";
import { NodeAlertsTable } from "../nodeAlertsTable";
import { Button } from "primereact/button";
import getColorBySeverity from "../../../util/colorBySeverity";
import moment from "moment";

export default function NodesTable({
    timelineVisible,
    expanded,
    database,
    data,
    groupSwitch,
    monitorInterval,
    noHeadings,
    parentSortField = null,
    parentSortOrder = null,
    nodeSeverityCounts = {}
}) {

    const { alerts_per_node } = useSelector((state) => state.monitoring);
    const { isLoading } = useSelector((state) => state.status);
    const [selectedRecord, setSelectedRecord] = useState(null);
    const [expandedRows, setExpandedRows] = useState(null);
    const navigate = useNavigate();

    const [internalSortField, setInternalSortField] = useState(null);
    const [internalSortOrder, setInternalSortOrder] = useState(null);


    const onSort = (event) => {
        setInternalSortField(event.sortField);
        setInternalSortOrder(event.sortOrder);
    };
    const effectiveSortField = parentSortField === 'status' ? parentSortField : internalSortField;
    const effectiveSortOrder = parentSortField === 'status' ? parentSortOrder : internalSortOrder;

    const sortedData = useMemo(() => {
      const sortField = parentSortField === 'status' ? parentSortField : internalSortField;
      const sortOrder = parentSortField === 'status' ? parentSortOrder : internalSortOrder;

      if (sortField && sortOrder !== null && Array.isArray(data)) {
          const dataCopy = [...data];

          dataCopy.sort((a, b) => {
              const itemA = a.data ?? a;
              const itemB = b.data ?? b;

              let valueA = itemA[sortField];
              let valueB = itemB[sortField];
              let comparison = 0;

              if (sortField === 'status') {
                  const statusOrder = (status) => (status === 'up' ? 1 : (status === 'down' ? 2 : 3));
                   comparison = statusOrder(valueA) - statusOrder(valueB);
              } else {
                  if (valueA == null && valueB != null) comparison = -1;
                  else if (valueA != null && valueB == null) comparison = 1;
                  else if (valueA == null && valueB == null) comparison = 0;
                  else if (typeof valueA === 'string' && typeof valueB === 'string')
                      comparison = valueA.localeCompare(valueB);
                  else comparison = valueA < valueB ? -1 : valueA > valueB ? 1 : 0;
              }

              return comparison * sortOrder;
          });
          return dataCopy;
      }
      return data;
    }, [data, parentSortField, parentSortOrder, internalSortField, internalSortOrder]);

    const nodeAlertsSortField = effectiveSortField === 'status' ? effectiveSortField : null;
    const nodeAlertsSortOrder = effectiveSortField === 'status' ? effectiveSortOrder : null;


    const getHref = (item) => {
         switch (item.type) {
            case types.ORACLE_PDB:
                return `oracle/cdb/${item.ex_params.cdb_id}/pdb/${item.id}`;
            case types.ORACLE_CDB:
                return `oracle/cdb/${item.id}`;
            case types.POSTGRES:
                return `postg/${item.id}`;
            case types.POSTGRES_K8S:
                return `postg/${item.id}`;
            case types.HAPROXY:
                return `haproxy/${item.id}`;
            case types.MONGO:
                return `mongo/${item.id}`;
            case types.MONGO_K8S:
                return `mongo_k8s/${item.id}`;
            case types.MARIADB:
                return `mariadb/${item.id}`;
            case types.ELASTIC:
                return `elastic/${item.id}`;
            case types.REDIS:
                return `redis/${item.id}`;
            case types.MYSQL:
                return `mysql/${item.id}`;
            case types.MSSQL:
                return `mssql/${item.id}`;
            default:
                return "";
        }
    }

     const createdDateBodyTemplate = (item) => {
          return (<div className="nowrap-cell">{moment(item.created_date).format("MMM DD,YYYY, hh:mm A")}</div>);
      };

     const whatBodyTemplate = (rowData) => {
          const name = rowData.data ? rowData.data.what : rowData.what;
          return <div className={expanded ? "parent-leaf" : ""}> {name} ({'database'})  <Button label="layout" link onClick={() =>  navigate('/layout', { state: { cluster_id: rowData.cluster } })} />  </div>;
     };
    const locationBodyTemplate = (rowData, props) => { return <TooltipCell id={`${props.field}-${rowData.data ? rowData.data.id : rowData.id}`} content={rowData.data ? rowData.data.location : rowData.location} /> }

    const statusBodyTemplate = (rowData) => {
      const item = rowData.data ?? rowData;
      const nodeId = item.id;
      const indLicense = item.ind_license;
      const severity = item.severity;
      const counts = nodeSeverityCounts[nodeId];
      const sev2 = counts?.severity_2 ?? 0;
      const sev3 = counts?.severity_3 ?? 0;

      let baseSymbol;
      if (severity === 'green' || severity === 'yellow') {
          baseSymbol = <span className="up-arrow">🡅</span>;
      } else {
          baseSymbol = <span className="down-arrow">🡇</span>;
      }

      const sev2Span = sev2 > 0 ? (
          <span style={{ color: 'orange', marginLeft: '5px' }}> {sev2}</span>
      ) : <span style={{ color: 'orange', marginLeft: '5px' }}>  0 </span>;

      const sev3Span = sev3 > 0 ? (
          <span style={{ color: 'red', marginLeft: '5px' }}> {sev3}</span>
      ) : <span style={{ color: 'red', marginLeft: '5px' }}> 0 </span>;

      const unlicensedText = indLicense === false ? ' (unlicensed)' : null;

      return (
          <>
              {baseSymbol}
              {sev2Span}
              {sev3Span}
              {unlicensedText}
          </>
      );
  };

  return (<DataTable tableStyle={{ tableLayout: "fixed", width: "100%" }} expandedRows={expandedRows} onRowToggle={(e) => {
      setExpandedRows(e.data)
  }} scrollable scrollHeight="flex"
      sortMode="single"
      onSort={onSort}
      sortField={effectiveSortField}
      sortOrder={effectiveSortOrder}
      emptyMessage={isLoading ? <div className="text-sm empty-severity">{loadingMsg}</div> : <div className="text-sm empty-severity">No nodes data</div>}
      selectionMode="single" selection={selectedRecord}
      onSelectionChange={(e) => setSelectedRecord(e.value)} onRowClick={(e) => { navigate(getHref(e.data)) }}
      rowExpansionTemplate={(rowData) => (
          <NodeAlertsTable
              expanded
              database={database}
              noHeadings
              node={rowData}
              grouping
              noHeader
              monitorInterval={monitorInterval}
              filteredAlerts={alerts_per_node?.[rowData.id]}
              timeVisible={true}
              timelineVisible={timelineVisible}
              parentSortField={nodeAlertsSortField}
              parentSortOrder={nodeAlertsSortOrder}
          />
      )}
      className={`severity-table nodes-table ${noHeadings ? 'no-headings' : ''}`} rowsPerPageOptions={[10, 20, 50, 100]} rowClassName={(rowData) => `severity-${getColorBySeverity(rowData.severity)}-row`} paginator alwaysShowPaginator={false} rows={100}
      value={sortedData}
      >
      <Column className="status-expander" expander/>
      <Column style={{width: "25%", maxWidth: "25%"}} field="what" body={whatBodyTemplate} header={groupSwitch} sortable></Column>
      <Column
          field="location"
          header='Location'
          body={locationBodyTemplate}
          style={{ width: "23%", maxWidth: "23%" }}
          sortable
      />
      <Column field="status" header="Status" body={statusBodyTemplate}  style={{width: "27%", maxWidth: "27%"}} sortable> </Column>
      <Column field="created_date" body={createdDateBodyTemplate}  style={{width: "25%", maxWidth: "25%"}} header="Last sample" sortable></Column>
      </DataTable>
  );
}
