import { useDispatch, useSelector } from "react-redux";
import { useEffect, Fragment, useState, useRef } from "react";
import { DataTable } from "primereact/datatable";
import { Column } from "primereact/column";
import { InputText } from "primereact/inputtext";
import { Button } from "primereact/button";
import { StatusActions } from "../../../store/status/reducer";
import ApplySettingsDialog from "../../../dialogs/applySettingsDialog";
import TooltipCell from "../../tooltipCell";


export default function PgSettingsTable({node}) {

    const dispatch = useDispatch();
    const { pg_settings } = useSelector((state) => state.status);
    const [filteredSettings, setFilteredSettings] = useState([]);
    const [changedSettings, setChangedSettings] = useState([]);
    const [initialSettings, setInitialSettings] = useState([]);
    const [pendingSettings, setPendingSettings] =  useState([]);
    const [settingSearch, setSettingSearch] = useState("");
    const [applySettings, setShowApplySettings] = useState(false);
    const [applyDisabled, setApplyDisabled] = useState(true);
    const [revertDisabled, setRevertDisabled] = useState(true);
    const dataTableRef = useRef(null);
    const [pageState, setPageState] = useState({ first: 0, rows: 10 });

    const handlePageChange = (event) => {
        setPageState({ first: event.first, rows: event.rows });
    };

    const searchFilter = (data, search) => {
        return data?.filter ? data?.filter((s) => {
            if (s?.name?.toLowerCase().includes(search) ||
                s?.setting?.toLowerCase().includes(search) ||
                s?.source?.toLowerCase().includes(search) ||
                s?.context?.toLowerCase().includes(search)
            )
                return true;
            return false;
        }) : [];
    }

    const onApplySettings = (data) => dispatch(StatusActions.editPgSettings(data, setShowApplySettings))

    const settingBodyTemplate = (rowData, props) => <TooltipCell id = {`${props.field}-${props.rowIndex}`} content = {rowData['setting']}/> 
    

    useEffect(() => {
        if (node?.id) {

            if (!pg_settings) dispatch(StatusActions.getPgSettings(node?.id));

            if (pg_settings?.length > 0) {
                const editableData = structuredClone(pg_settings);
                setFilteredSettings(editableData);
                setChangedSettings(editableData);
            }

        }


    }, [dispatch, pg_settings, node?.id]);

    useEffect(() => {

       if (node?.id) dispatch(StatusActions.setPgSettings(null));

    }, [node?.id, dispatch])


    useEffect(() => {
        setFilteredSettings(searchFilter(changedSettings, settingSearch));
    }, [changedSettings, settingSearch]);



    const settingEditor = (options) => {
        return (<InputText value={options.value} onChange={(e) => options.editorCallback(e.target.value)} />)

    }

    const actionBodyTemplate = (rowData, props) => {
        return (
            <div className="flex">
                <div className="mt-2">{rowData?.context === "postmaster" ? "Yes" : "No"}</div>
                <Button icon="pi pi-pencil" className="ml-auto" severity="success" rounded size="small" onClick={() => {
                     const renderedRowIndex = (props.rowIndex - (pageState.first / pageState.rows) * pageState.rows) + 1;
                     dataTableRef.current?.getTable()?.rows?.[renderedRowIndex]?.cells[1]?.click()
                }}/>
            </div>
        );
    }

    const onCellEditComplete = (e) => {
        let { rowData, newValue, field } = e;
        const setting = pg_settings?.filter((s) => s?.name === rowData['name'])
        if (setting?.length > 0 && setting[0]['setting'] !== newValue)
            rowData['color'] = 'red'
        else
            rowData['color'] = 'default';
        const index = changedSettings?.findIndex((s) => s?.name === rowData['name'])
        changedSettings[index] = rowData;
        const _pendingSettings = changedSettings?.filter((s) => s?.color === 'red');
        setPendingSettings(_pendingSettings);
        setApplyDisabled(_pendingSettings?.length < 1);
        setRevertDisabled(_pendingSettings?.length < 1);
        setChangedSettings(changedSettings);
        if (rowData[field] !== newValue) rowData[field] = newValue
    }



    return (
        <Fragment>

            <div className="flex" >
                <span className="p-input-icon-right">
                    <i className={settingSearch ? "pi pi-times" : "field-collapsed"} onClick={() => setSettingSearch("")} />
                    
                    <InputText className="p-inputtext-sm filter-field mb-2"
                        placeholder="Search"
                        value={settingSearch}
                        onChange={(e) => {
                            setSettingSearch(e.target.value);
                        }} />
                </span>
                <div className="ml-auto">

                <Button label="Revert" className="mr-2"  disabled = {revertDisabled} size="small" icon="pi pi-times" iconPos="right" onClick={() => {
                    setChangedSettings(pg_settings);
                    setApplyDisabled(true);
                    setRevertDisabled(true);
                    }}/>
                <Button label="Apply" disabled = {applyDisabled} size="small" icon="pi pi-check" iconPos="right" onClick={() => {
                    setInitialSettings(pg_settings?.filter((s) => pendingSettings.filter((p) => p.name === s.name)?.length > 0));
                    setShowApplySettings(true);
                    }}/>


                </div>
                
            </div>



            <DataTable ref={dataTableRef} editMode="cell" dataKey="name" emptyMessage={<div className="text-sm pl-3">No parameters data</div>} className="monitors-table" value={filteredSettings}
                paginator rows={10} rowsPerPageOptions={[10, 20, 50, 100]} alwaysShowPaginator={false}
                paginatorTemplate="FirstPageLink PrevPageLink PageLinks NextPageLink LastPageLink RowsPerPageDropdown CurrentPageReport"
                currentPageReportTemplate="Showing {first} to {last} of {totalRecords} parameters"
                paginatorPosition="bottom"
                onPage={handlePageChange}
                first={pageState.first}
                rowClassName={(rowData) => {
                    return rowData.color === 'red' ? 'text-red-500' : '';
                }}
            >

                <Column field="name" header="Name" style={{width: "40%"}} sortable ></Column>
                <Column field="setting" style={{width: "30%"}} body={settingBodyTemplate} header="Setting" editor={(options) => settingEditor(options)} onCellEditComplete={(options) => onCellEditComplete(options)} sortable></Column>
                <Column body={actionBodyTemplate} header="Restart required"></Column>
            </DataTable>
            {applySettings ? <ApplySettingsDialog
                visible={applySettings}
                initial = {initialSettings} 
                settings = {pendingSettings}
                node = {node}
                onHide={() => setShowApplySettings(false)}
                onOkClick={onApplySettings}
            /> : null
            }
        </Fragment>
    );
}