import { useState, useEffect, useCallback } from 'react';
import { useSelector } from 'react-redux';
import { useDispatch } from 'react-redux';
import { Button } from "primereact/button";
import { InputText } from "primereact/inputtext";
import { InputNumber } from "primereact/inputnumber";
import { Dialog } from "primereact/dialog";
import { Dropdown } from 'primereact/dropdown';
import validator from "../../util/validator";
import constraints from "./constraints";
import { LayoutActions } from '../../store/layout/reducer';
import ErrorTextBox from '../../components/errorTextBox';
import { dbTypes, serviceTypes, types, osTypes } from '../../config';
import { Checkbox } from 'primereact/checkbox';
import { data } from 'react-router-dom';

const AddIntegratedServiceDialog = ({visible, data, onHide, onOkClick}) => {

    const getSupportedDatabases = useCallback((type) => {

        const default_support = [types.PROXMOX, types.AWS].includes(type) ? true : false
        const supported_databases = dbTypes.filter((t) => t.value !== types.ALL).reduce((acc, obj) => {
            acc[obj.value] = default_support;
            return acc;
        }, {});
        switch (type) {
            case types.K8S:
                supported_databases[types.POSTGRES] = true;
                supported_databases[types.MONGO] = true;
                break;
            case types.PROXMOX:
            case types.AWS:
                supported_databases[types.MSSQL] = false;
                break;
            default:
                break
        }
        return supported_databases;

    }, [])


    const getSupportedOS = useCallback((type) => {

        const default_support = false
        const supported_os = osTypes.reduce((acc, obj) => {
            acc[obj.value] = default_support;
            return acc;
        }, {});
        switch (type) {
            case types.PROXMOX:
                supported_os[osTypes[0].value] = true;
                supported_os[osTypes[1].value] = true;
                break;
            default:
                break
        }
        return supported_os;

    }, [])

    const [pageIndex, setPageIndex] = useState(0);
    const dispatch = useDispatch();

    // First Page elements
    const [name, setName] = useState(data ? data.name : "");
    const [type, setType] = useState(data ? data.type : types.K8S);

    // Second page elements
    const [hostname, setHostname] = useState(data ? data.host_name : "");
    const [port, setPort] = useState(data ? data.port : 6443);
    const [bearer, setBearer] = useState(data ? data?.ex_params?.api_token : "");
    const [proxmoxTokenName, setProxmoxTokenName] = useState(data ? data?.ex_params?.proxmox_token_name : "");
    const [proxmoxTokenValue, setProxmoxTokenValue] = useState(data ? data?.ex_params?.proxmox_token_value : "");
    const [awsAccessToken, setAWSAccessToken] = useState(data ? data?.ex_params?.aws_access_token : "");
    const [awsSecretToken, setAWSSecretToken] = useState(data ? data?.ex_params?.aws_secret_token : "");
    const [verifySSL, setVerifySSL] = useState(data ? data?.ex_params?.verify_ssl : false);
    const [supportedDatabases, setSupportedDatabases] = useState(getSupportedDatabases(type));
    const [supportedOS, setSupportedOS] = useState(getSupportedOS(type));

    // First Page element errors
    const [nameError, setNameError] = useState("");


    // Second page element errors
    const [hostnameError, setHostnameError] = useState("");
    const [portError, setPortError] = useState("");
    const [bearerError, setBearerError] = useState("");
    const [proxmoxTokenNameError, setProxmoxTokenNameError] = useState("");
    const [proxmoxTokenValueError, setProxmoxTokenValueError] = useState("");
    const [awsAccessTokenError, setAWSAccessTokenError] = useState("");
    const [awsSecretTokenError, setAWSSecretTokenError] = useState("");

    const { errorMessage, isLoading, port_exists } = useSelector((state) => state.layout);


    const getConnectionCredentials = () => {

        switch(type){
            case types.AWS:
                return {
                    aws_access_token: awsAccessToken,
                    aws_secret_token: awsSecretToken,
                    type: type
                }
            case types.PROXMOX:
                return {
                    hostname: hostname,
                    proxmox_token_name: proxmoxTokenName,
                    proxmox_token_value: proxmoxTokenValue,
                    verify_ssl: verifySSL,
                    type: type
                }
            case types.K8S:
                return {
                    api_host: hostname,
                    api_port: port,
                    api_token: bearer,
                    type: type
                }
            default:
                return {}
        }

    }

    useEffect(() => {

        setPageIndex(0);
        const type = data ? data?.type :types.PROXMOX
        // First page elements and errors
        setName(data ? data?.name : "");
        setType(type);

        // Second page elements and errors
        setPort(data ? data?.port : 6443);
        setPortError("");
        setHostname(data ? data?.host_name : "");
        setHostnameError("");
        setProxmoxTokenName(data ? data?.ex_params?.proxmox_token_name : "");
        setProxmoxTokenValue(data ? data?.ex_params?.proxmox_token_value : "");
        setVerifySSL(data ? data?.ex_params?.verify_ssl : false);
        setProxmoxTokenNameError("");
        setProxmoxTokenValueError("");
        setAWSAccessToken(data ? data?.ex_params?.aws_access_token : "");
        setAWSSecretToken(data ? data?.ex_params?.aws_secret_token : "");
        setAWSAccessTokenError("");
        setAWSSecretTokenError("");
        setBearer(data ? data?.ex_params?.api_token : "");
        setBearerError("");
        setSupportedDatabases(getSupportedDatabases(type));
        setSupportedOS(getSupportedOS(type));


    }, [visible, data, getSupportedDatabases, getSupportedOS]);

    const validateFirst = () => {

        let errors = [];


        let error = validator(constraints)("name", name);
        (error == null || errors.push(error)) && setNameError(error);

        return errors;

    }


    useEffect(() => {
        if (port_exists) {
            if (port_exists?.detail) {

                setPortError("");

            }
            else {

                if (port_exists?.exists) {

                    setPortError("service available");
                }
                else {

                    setPortError("service unavailable");

                }

            }

        }
    }, [port_exists])



    const onThirdBackClick = () => {

        setPageIndex(1);

    }

    const onSaveClick = () => {
        const ex_params = {}

        if (type === types.PROXMOX) {

            ex_params["proxmox_token_name"] = proxmoxTokenName
            ex_params["proxmox_token_value"] = proxmoxTokenValue
            ex_params["verify_ssl"] = verifySSL

        }

        if (type === types.K8S) {
            ex_params["api_token"] = bearer
        }
        if(type === types.AWS) {
            ex_params["aws_access_token"] = awsAccessToken
            ex_params["aws_secret_token"] = awsSecretToken
        }

        const request_data = {
            id: data ? data?.id : null,
            name: name,
            host_name: hostname,
            type: type,
            username: "",
            password: "",
            port: [types.PROXMOX, types.AWS].includes(type) ? "" : port,
            supported_databases: supportedDatabases,
            supported_os: supportedOS,
            ex_params: ex_params
        }

        if (onOkClick) {
            onOkClick(request_data);
        }
    }

    const validateSecond = () => {

        let errors = [];
      

        if (type === types.K8S) {

            let error = validator(constraints)("hostname", hostname);
            (error == null || errors.push(error)) && setHostnameError(error);

            error = validator(constraints)("port", port);
            (error == null || errors.push(error)) && setPortError(error);

            error = validator(constraints)("bearer", bearer);
            (error == null || errors.push(error)) && setBearerError(error);

        }
        if (type === types.PROXMOX)
        {

            let error = validator(constraints)("hostname", hostname);
            (error == null || errors.push(error)) && setHostnameError(error);
            error = validator(constraints)("proxmoxTokenName", proxmoxTokenName);
            (error == null || errors.push(error)) && setProxmoxTokenNameError(error);
            error = validator(constraints)("proxmoxTokenValue", proxmoxTokenValue);
            (error == null || errors.push(error)) && setProxmoxTokenValueError(error);
            

        }
        if(type === types.AWS) {

            let error = validator(constraints)("awsAccessToken", awsAccessToken);
            (error == null || errors.push(error)) && setAWSAccessTokenError(error);
            error = validator(constraints)("awsSecretToken", awsSecretToken);
            (error == null || errors.push(error)) && setAWSSecretTokenError(error);

        }

        return errors;


    }

    const renderFooter = () => {
        if (pageIndex === 0) {
            return (
                <div style={{ display: "flex", justifyContent: "space-between" }}>
                    <Button label={"Close"} icon="pi pi-times" onClick={onHide} className="p-button-text" />
                    <Button
                        label={"Next"}
                        icon={`pi pi-arrow-right`}
                        iconPos="right"
                        onClick={onFirstNextClick} autoFocus />
                </div>
            );
        }
        else if (pageIndex === 1) {
            return (
                <div style={{ display: "flex", justifyContent: "space-between" }}>
                    <Button label={"Close"} icon="pi pi-times" onClick={onHide} className="p-button-text" />
                    <Button label={"Back"}
                        onClick={onSecondBackClick}
                        className="p-button-secondary"
                        icon="pi pi-arrow-left"
                        iconPos="left"
                    />
                    <Button
                        label={"Next"}
                        icon={`pi pi-arrow-right`}
                        iconPos="right"
                        onClick={onSecondNextClick} autoFocus />
                </div>
            );
        }
        else if (pageIndex === 2) {
            return (
                <div style={{ display: "flex", justifyContent: "space-between" }}>
                    <Button label={"Close"} icon="pi pi-times" onClick={onHide} className="p-button-text" />
                    <Button label={"Back"}
                        onClick={onThirdBackClick}
                        className="p-button-secondary"
                        icon="pi pi-arrow-left"
                        iconPos="left"
                    />

                    <Button
                        label={"Submit"}
                        icon={`pi pi-${isLoading ? 'spin pi-spinner' : 'check'}`}
                        id="submit-button"
                        className='submit-button'
                        iconPos="right"
                        onClick={onSaveClick} autoFocus />

                </div>
            );
        }
    }
    const renderPageSecond = () => {

        return (

            <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
                <div className="form-inner-part">
                    {[types.PROXMOX, types.K8S].includes(type) ? <div>
                        <div className="field grid">
                            <label className="col-fixed" style={{ width: "220px" }}>{"Hostname/IP:"}</label>
                            <div className="col">
                                <InputText className="p-inputtext-sm" placeholder="Hostname/IP" value={hostname} onChange={(e) => setHostname(e.target.value)} />
                            </div>
                        </div>
                        <div className="error-help" style={{ display: "flex", justifyContent: "flex-end", marginRight: "5px", marginTop: "-15px", marginBottom: "5px" }}>
                            <small id="name-help" className="p-error p-d-block">{hostnameError}</small>
                        </div>
                    </div> : null}




                    {type === types.K8S ? <div>



                        <div>
                            <div className="field grid">
                                <label className="col-fixed" style={{ width: "220px" }}>{"Bearer token:"}</label>
                                <div className="col">
                                    <InputText className="p-inputtext-sm" placeholder="Bearer token" value={bearer} onChange={(e) => setBearer(e.target.value)} />
                                </div>
                            </div>
                            <div className="error-help" style={{ display: "flex", justifyContent: "flex-end", marginRight: "5px", marginTop: "-15px", marginBottom: "5px" }}>
                                <small id="name-help" className="p-error p-d-block">{bearerError}</small>
                            </div>
                        </div>

                        <div>
                            <div className="field grid flex align-items-start">
                                <label className="col-fixed" style={{ width: "220px" }}>{"Port:"}</label>
                                <div className="col">
                                    <span className="p-input-icon-right flex">
                                        <InputNumber useGrouping={false} className="p-inputtext-sm" placeholder="Port" value={port} onValueChange={(e) => {
                                            setPort(e.value);
                                        }} />
                                        <i className={isLoading && !port_exists && port_exists !== undefined ? "pi pi-spin pi-spinner align-self-center" : ""} />

                                    </span>
                                </div>
                            </div>
                        </div>
                    </div> : null}

                    {type === types.AWS ? <div>


                        <div>
                            <div className="field grid">
                                <label className="col-fixed" style={{ width: "220px" }}>{"AWS Access Token:"}</label>
                                <div className="col">
                                    <InputText className="p-inputtext-sm" placeholder="AWS Access Token" value={awsAccessToken} onChange={(e) => setAWSAccessToken(e.target.value)} />
                                </div>
                            </div>
                            <div className="error-help" style={{ display: "flex", justifyContent: "flex-end", marginRight: "5px", marginTop: "-15px", marginBottom: "5px" }}>
                                <small id="name-help" className="p-error p-d-block">{awsAccessTokenError}</small>
                            </div>
                        </div>
                        <div>
                            <div className="field grid">
                                <label className="col-fixed" style={{ width: "220px" }}>{"AWS Secret Token:"}</label>
                                <div className="col">
                                    <InputText className="p-inputtext-sm" placeholder="AWS Secret Token" value={awsSecretToken} onChange={(e) => setAWSSecretToken(e.target.value)} />
                                </div>
                            </div>
                            <div className="error-help" style={{ display: "flex", justifyContent: "flex-end", marginRight: "5px", marginTop: "-15px", marginBottom: "5px" }}>
                                <small id="name-help" className="p-error p-d-block">{awsSecretTokenError}</small>
                            </div>
                        </div>

                    </div> : null}

                    {type === types.PROXMOX ? <div>


                        <div>
                            <div className="field grid">
                                <label className="col-fixed" style={{ width: "220px" }}>{"Proxmox Token Name:"}</label>
                                <div className="col">
                                    <InputText className="p-inputtext-sm" placeholder="Proxmox Token Name" value={proxmoxTokenName} onChange={(e) => setProxmoxTokenName(e.target.value)} />
                                </div>
                            </div>
                            <div className="error-help" style={{ display: "flex", justifyContent: "flex-end", marginRight: "5px", marginTop: "-15px", marginBottom: "5px" }}>
                                <small id="name-help" className="p-error p-d-block">{proxmoxTokenNameError}</small>
                            </div>
                        </div>
                        <div>
                            <div className="field grid">
                                <label className="col-fixed" style={{ width: "220px" }}>{"Proxmox Token Value:"}</label>
                                <div className="col">
                                    <InputText className="p-inputtext-sm" placeholder="Proxmox Token Value" value={proxmoxTokenValue} onChange={(e) => setProxmoxTokenValue(e.target.value)} />
                                </div>
                            </div>
                            <div className="error-help" style={{ display: "flex", justifyContent: "flex-end", marginRight: "5px", marginTop: "-15px", marginBottom: "5px" }}>
                                <small id="name-help" className="p-error p-d-block">{proxmoxTokenValueError}</small>
                            </div>
                        </div>
                        <div>
                            <div className="field grid">
                                <label className="col-fixed" style={{ width: "220px" }}>{"Verify SSL:"}</label>
                                <div className="col">
                                    <Checkbox onChange={() => setVerifySSL(!verifySSL)} checked={verifySSL} />
                                </div>
                            </div>
                        </div>

                    </div> : null}

                    <div>
                        <div className="field grid">
                            <label className="col-fixed" style={{ width: "220px" }}>{"Check connection:"}</label>
                            <div className="col">

                                <Button
                                    className="p-button-sm"
                                    label="Check"
                                    style={{ width: "80px" }}
                                    onClick={() => {

                                        const errors = validateSecond();
                                        if (errors.length === 0) dispatch(LayoutActions.checkServiceConnection(getConnectionCredentials()))

                                    }}
                                />
                                <div className={`${!['service available', 'service unavailable'].includes(portError) ? "p-error" : ""} p-d-block mt-2 text-sm`}>{portError}</div>
                            </div>
                        </div>
                    </div>

                </div>

                {errorMessage && <ErrorTextBox error={errorMessage?.message ? errorMessage.message : errorMessage} />}
            </div >



        );
    }



    const renderPageThird = () => {

        return (
            <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
                <div className="form-inner-part">

                    <div className='field'>List of database types which could be installed on this service:</div>
                    <ul>
                        {dbTypes.filter((t) => Object.keys(supportedDatabases).includes(t.value) && supportedDatabases[t.value] === true).map((t) => <li key={t.value}>{t.label}</li>)}
                    </ul>

                    {type === types.PROXMOX ? <div>

                        <div className='field'>List of operating systems which could be installed on this service:</div>
                        <ul>
                            {osTypes.map((t) => <li key={t.value}>{t.label}</li>)}
                        </ul>

                    </div> : null}

                </div>

                {errorMessage && <ErrorTextBox error={errorMessage?.message ? errorMessage.message : errorMessage} />}
            </div >



        );
    }

    const onFirstNextClick = () => {

        let errors = validateFirst();
        if (errors.length === 0) setPageIndex(1)
    }

    const onSecondBackClick = () => {
        setPortError("");
        setPageIndex(0);
    }

    const onSecondNextClick = () => {

        let errors = validateSecond();
        if (errors.length === 0) setPageIndex(2)
    }

    const renderPageFirst = () => {
        return (
            <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
                <div>
                    <div className="field grid">
                        <label className="col-fixed" style={{ width: "300px" }}>{"Name:"}</label>
                        <div className="col">
                            <InputText value={name} onChange={(e) => setName(e.target.value)} />
                        </div>
                    </div>
                    <div className="error-help" style={{ display: "flex", justifyContent: "flex-end", marginRight: "5px", marginTop: "-15px", marginBottom: "5px" }}>
                        <small id="name-help" className="p-error p-d-block">{nameError}</small>
                    </div>
                </div>

                <div>
                    <div className='field grid'>

                        <div className='col-fixed' style={{ width: "300px" }}>{"Service type:"}</div>
                        <div className='col'>

                            <Dropdown className="p-inputtext-sm"
                                value={type}
                                disabled = {Boolean(data) === true}
                                onChange={(e) => {
                                    const value = e.value
                                    setType(value);
                                    setSupportedDatabases(getSupportedDatabases(value));
                                    setSupportedOS(getSupportedOS(value));
                                }}
                                options={serviceTypes.filter((v) => v.value !== types.ALL)}
                                placeholder="Type options" />

                        </div>

                    </div>

                </div>






                {errorMessage && <ErrorTextBox error={errorMessage?.message ? errorMessage.message : errorMessage} />}
            </div>
        );
    }

    return (
        <Dialog
            className='add-integrated-service'
            header={`${data ? 'Edit' : 'Add a new' } Integrated Service`}
            visible={visible}
            style={{ width: '50vw' }}
            onHide={onHide}
            footer={renderFooter()}>
            <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
                {pageIndex === 0 && renderPageFirst()}
                {pageIndex === 1 && renderPageSecond()}
                {pageIndex === 2 && renderPageThird()}
            </div>

        </Dialog>
    );
}


export default AddIntegratedServiceDialog;