import { Dialog } from "primereact/dialog";
import { Button } from "primereact/button";
import { Dropdown } from "primereact/dropdown";
import { useState, useEffect } from "react";
import { InputText } from "primereact/inputtext";
import { RadioButton } from "primereact/radiobutton";
import { Calendar } from "primereact/calendar";
import moment from "moment";
import validator from "../../../util/validator";
import constraints from "./constraints";
import ApiBar from "../../../components/apiBar";
import copyToClipboard from "../../../util/copyToClipboard";
import createAPIcommand from "../../../util/createApiCommand";
import { useSelector } from "react-redux";
import ErrorTextBox from "../../../components/errorTextBox";
import config, {restoreOptions} from "../../../config";

export const RestoreElasticsearchCluster = (props) => {

  const [targetSnapshot, setTargetSnapshot] = useState(null);
  const [snapshotOptions, setSnapshotOptions] = useState([]);
  const [restoreConfirm, setRestoreConfirm] = useState("");
  const [targetSnapshotError, setTargetSnapshotError] = useState("");
  const [restoreTimeError, setRestoreTimeError] = useState("");
  const [apiTooltip, setApiTooltip] = useState(" ");
  const [restoreTime, setRestoreTime] = useState(null);
  const [restoreRadio, setRestoreRadio] = useState(restoreOptions.THIS.value);
  const { accessToken } = useSelector((state) => state.app);
  const { errorMessage, isLoading } = useSelector((state) => state.layout);
  const [restoreConfirmError, setRestoreConfirmError] = useState("");

  const validate = () => {

    const errors = [];


    let error = validator(constraints)("targetSnapshot", targetSnapshot);
    (error == null || errors.push(error)) && setTargetSnapshotError(error);

    error = validator(constraints)("restoreConfirm", restoreConfirm);
    (error == null || errors.push(error)) && setRestoreConfirmError(error);

    if (restoreRadio === restoreOptions.CUSTOM.value) {
      error = validator(constraints)("time", restoreTime);
      (error == null || errors.push(error)) && setRestoreTimeError(error);
    }
    else setRestoreTimeError("");

    return errors;

  }


  const handleOk = () => {

    let errors = validate();

    if (errors.length === 0) {

      props.onOkClick(
        {
          "sn_id": targetSnapshot?.id,
          "restore_option": restoreRadio,
          "restore_time": restoreTime ? moment(restoreTime).format("YYYY-MM-DD HH:mm:ss") : restoreTime
        });

    }
  }

  const onRestApiClick = async () => {

    let errors = validate()

    if (errors.length === 0) {

      setApiTooltip(await copyToClipboard(createAPIcommand({
        "sn_id": targetSnapshot?.name,
        "restore_option": restoreRadio,
        "restore_time": restoreTime ? moment(restoreTime).format("YYYY-MM-DD HH:mm:ss") : restoreTime

      }, accessToken, "restore_elasticsearch_cluster/")));
    }
  }

  useEffect(() => {

    if (props.data) {

      const temp = [];
      props.data.snapshots?.forEach((s) => temp.push({ label: `${moment(s.created_date).format("MMM DD,YYYY, hh:mm A")}  ${s.display_size}`, value: s}));
      setSnapshotOptions(temp)

    }

  }, [props.data]);

  const renderFooter = () => {
    return (
      <div style={{ display: "flex", justifyContent: "space-between" }}>
        <Button label={"Close"} icon="pi pi-times" onClick={props.onHide} className="p-button-text" />
        <Button
          label={"Submit"}
          icon= {`pi pi-${isLoading ? 'spin pi-spinner' : 'check'}`}
          disabled = {!targetSnapshot || targetSnapshot?.status === false}
          iconPos="right"
          onClick={handleOk} autoFocus />
      </div>
    );

  }



  return (
    <Dialog
      className='restore-elasticsearch-cluster'
      header={`Restore Cluster`}
      visible={props.visible}
      style={{ width: '40vw' }}
      onHide={props.onHide}
      footer={renderFooter()}>
      <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
        <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
          <div className="field grid">
            <div className="col">
              <div className="title">{"Choose snapshot to restore:"}</div>

              <div className="field">
                <Dropdown className="p-inputtext-sm"
                  value={targetSnapshot}
                  options={snapshotOptions}
                  onChange={(e) => setTargetSnapshot(e.value)}
                  placeholder="Select snapshot"
                  emptyMessage = "no snapshots found" 
                  />
                <div className="error-help">{<small className="p-error p-d-block">{targetSnapshotError}</small>}</div>
              </div>

              <div className="field grid">
                <div className="col">
                {
                    Object.keys(restoreOptions).map(key => (<div key={key} className="field-radiobutton">
                      <RadioButton inputId={key} name="restoreRadio" value={restoreOptions[key].value} checked={restoreRadio === restoreOptions[key].value} onChange={(e) => setRestoreRadio(e.value)} />
                      <label htmlFor={key}>{restoreOptions[key].label}</label>
                    </div>))
                  }
                </div>
              </div>

              <div className="field grid align-items-center">
                <div className="col-fixed">Select time:</div>
                <div className="col-6">
                  <Calendar
                    showSeconds
                    maxDate={new Date()}
                    showTime
                    hourFormat="24"
                    disabled={restoreRadio !== restoreOptions.CUSTOM.value}
                    value={restoreTime}
                    onChange={(e) => setRestoreTime(e.value)} />
                  <div className="error-help">{<small className="p-error p-d-block">{restoreTimeError}</small>}</div>
                </div>
              </div>
              <div className="field grid align-items-center">
                <div className="col">
                  <InputText
                    className="p-inputtext-sm"
                    placeholder={`Please type "${config.restoreConfirm}" here in order to continue`}
                    value={restoreConfirm}
                    onChange={(e) => setRestoreConfirm(e.target.value)} />
                  <div className="error-help">{<small className="p-error p-d-block">{restoreConfirmError}</small>}</div>
                </div>
              </div>

            </div>
          </div>
          {(errorMessage || targetSnapshot?.status === false) && <ErrorTextBox error={errorMessage ? errorMessage?.message ? errorMessage.message : errorMessage : "This snapshot did not complete successfully"} />}
          <ApiBar tooltip={apiTooltip} onClick={onRestApiClick} />
        </div>
      </div>
    </Dialog>)

};

export default RestoreElasticsearchCluster;