import { Dialog } from "primereact/dialog";
import { Button } from "primereact/button";
import { useEffect, useState, useRef } from "react";
import { InputText } from "primereact/inputtext";
import { Checkbox } from "primereact/checkbox";
import { Slider } from "primereact/slider";
import { useSelector, useDispatch } from "react-redux";
import { LayoutActions } from "../../../store/layout/reducer";
import validator from "../../../util/validator";
import constraints from "./constraints";
import ApiBar from "../../../components/apiBar";
import createAPIcommand from "../../../util/createApiCommand";
import copyToClipboard from "../../../util/copyToClipboard";
import ErrorTextBox from "../../../components/errorTextBox";
import getValidationError from '../../../util/errorPrompter';
import { Tooltip } from "primereact/tooltip";

export const AddMariaDBDatabaseDialog = (props) => {

  const [name, setName] = useState("");
  const { valid_name, isLoading, errorMessage } = useSelector((state) => state.layout);
  const dispatch = useDispatch();
  const [nameError, setNameError] = useState(null);

  const [automaticDelete, setAutomaticDelete] = useState(false);
  const [automaticDeleteDays, setAutomaticDeleteDays] = useState(1);
  const [apiTooltip, setApiTooltip] = useState(" ");
  const { accessToken } = useSelector((state) => state.app);
  const customEvent = new Event('beforeinput', { bubbles: true });
  const submitButtonRef = useRef(null);

  

  const validateDbName = () => {
    setNameError("");

    let errors = [];
    let error = validator(constraints)("name", name);
    (error == null || errors.push(error)) && setNameError(error);
    if (errors.length === 0) dispatch(LayoutActions.checkMariaDbDatabaseName({ node_id: props.node.id, name: name }));

  }

  useEffect(() => {
    if (valid_name) {

      if (valid_name?.error)
        setNameError(valid_name.error);
      else setNameError(valid_name?.response)
    }

  }, [valid_name]);


  const handleOk = () => {

    if (!nameError) setNameError("name not checked");

    if (nameError === "name available") {
      
      props.onOkClick(
        {
          "db_name": name,
          "node_id": props.node.id,
          "auto_delete": automaticDelete,
          "auto_delete_days": automaticDeleteDays,
        });

    }
    else submitButtonRef.current.dispatchEvent(customEvent);
  }

  const onRestApiClick = async () => {

    if (!nameError) setNameError("name not checked");

    if (nameError === "name available") {
      setApiTooltip(await copyToClipboard(createAPIcommand(
        {
          "db_name": name,
          "node_id": props.node.name,
          "auto_delete": automaticDelete,
          "auto_delete_days": automaticDeleteDays,
        }, accessToken, "add_mariadb_database/")));

    }
    else setApiTooltip(getValidationError());


  }

  const renderFooter = () => {
    return (
      <div style={{ display: "flex", justifyContent: "space-between" }}>
        <Button label={"Close"} icon="pi pi-times" onClick={props.onHide} className="p-button-text" />
        <Button
          label={"Submit"}
          className="submit-button"
          ref={submitButtonRef}
          icon= {`pi pi-${isLoading ? 'spin pi-spinner' : 'check'}`}
          iconPos="right"
          onClick={handleOk} autoFocus />
        <Tooltip   target='.submit-button' showEvent='beforeinput' position='top' >
          {getValidationError()}
        </Tooltip>
      </div>
    );

  }



  return (

    <Dialog
      className='add-mariadb-db'
      header={`Add new database`}
      visible={props.visible}
      style={{ width: '40vw' }}
      onHide={props.onHide}
      footer={renderFooter()}>
      <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
        <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
          <div className="field grid">
            <label style={{ marginTop: "7px" }} className="col-fixed align-self-start">{"Name:"}</label>
            <div className="col">
              <span className="p-input-icon-right flex">
                <i className={isLoading ? "pi pi-spin pi-spinner align-self-center" : ""} />
                <InputText className="p-inputtext-sm" placeholder="Name" value={name} onChange={(e) => {
                  setName(e.target.value);
                  setNameError("name not checked");
                  dispatch(LayoutActions.setValidName(undefined));
                }} />

              </span>
              <div className="error-help"><small className={nameError === "name available" ? "p-d-block" : "p-error p-d-block"}>{nameError}</small></div>
            </div>



            <div className="col-fixed">
              <Button className="p-button-sm" style={{ width: "80px" }} label="Check" onClick={validateDbName} />
            </div>
          </div>
          <div className="field grid flex align-items-center">
            <label htmlFor="cb1" className="p-checkbox-label col-fixed" style={{ width: "250px" }}>{!automaticDelete ? "Automatic delete: disabled" : `Automatic delete in: ${automaticDeleteDays} days`}</label>
            <div className="col-fixed">
              <Checkbox inputId="cb1" value="Automatic delete disabled:" onChange={() => setAutomaticDelete(!automaticDelete)} checked={automaticDelete}></Checkbox>
            </div>
            <div className="col">
              <Slider
                disabled={!automaticDelete}
                min={1}
                max={365}
                value={automaticDeleteDays}
                onChange={(e) => setAutomaticDeleteDays(e.value)} />
            </div>

          </div>
          {errorMessage && <ErrorTextBox error={errorMessage?.message ? errorMessage.message : errorMessage} />}
          <ApiBar tooltip={apiTooltip} onClick={onRestApiClick} />
        </div>
      </div>

    </Dialog>)

};

export default AddMariaDBDatabaseDialog;