import { Dialog } from "primereact/dialog";
import { InputNumber } from "primereact/inputnumber";
import { InputTextarea } from "primereact/inputtextarea";
import { InputText } from "primereact/inputtext";
import { Checkbox } from "primereact/checkbox";
import { Dropdown } from "primereact/dropdown";
import { Button } from "primereact/button";
import { RadioButton } from "primereact/radiobutton";
import { Tooltip } from "primereact/tooltip";
import CodeMirror from '@uiw/react-codemirror';
import { langs } from '@uiw/codemirror-extensions-langs';
import { useDispatch, useSelector } from "react-redux";
import { useState, useCallback, useEffect, useRef } from "react";
import { MonitoringActions } from "../../../store/monitoring/reducer";
import { debounce } from "lodash";
import validator from "../../../util/validator";
import copyToClipboard from "../../../util/copyToClipboard";
import constraints from "./constraints";
import {
    autocompletion,
    closeCompletion,
    startCompletion,
} from "@codemirror/autocomplete";

import { types } from "../../../config";

import { v4 as uuidv4 } from 'uuid';
import { vscodeDark } from '@uiw/codemirror-theme-vscode';

import { EditorView } from '@codemirror/view';

export default function MonitoringWizard({ user, update, visible, onHide, monitorType, data: monitorData, onOkClick }) {

    const dispatch = useDispatch();
    const mounted = useRef(null);
    const { isLoading, messages } = useSelector((state) => state.monitoring);

    const [pageIndex, setPageIndex] = useState(0);
    const [valid, setValid] = useState(update ? true : false);
    const [requestData, setRequestData] = useState({});
    //page 1 elements
    const [monitorName, setMonitorName] = useState(update ? monitorData?.monitor_name : "");
    const [monitorDescription, setMonitorDescription] = useState(update ? monitorData?.monitor_description : "");
    const [monitorModule, setMonitorModule] = useState(update ? monitorData?.module : "");
    const [cmdType, setCmdType] = useState(update ? monitorData?.cmd_type : monitorType === "LINUX" ? "ssh" : "sql");
    const [preScript, setPreScript] = useState(update ? monitorData?.pre_script : "");
    const [monitorScript, setMonitorScript] = useState(update ? monitorData?.monitor_script : "");
    const [nameError, setNameError] = useState("");
    const [cmdError, setCmdError] = useState("");
    //page 2 elements
    const columnTypeOptions = [{ label: "text", value: "text" }, { label: "numeric", value: "numeric" }];
    const [fieldObjects, setFieldObjects] = useState({ [uuidv4()]: { display_name: "column1", column_name: "column1", column_type: "text", placeholder: "Display name of column 1" } });
    const [fieldCount, setFieldCount] = useState(1);
    const [deletedFields, setDeletedFields] = useState([]);
    const [columnsError, setColumnsError] = useState("");
    //page 3 elements
    const monitorIntervalOptions = [

        { label: "15 Seconds", value: "15 seconds" },
        { label: "1 Minute", value: "1 minute" },
        { label: "5 Minutes", value: "5 minutes" },
        // { label: "10 Minutes", value: "10 minutes" },
        { label: "15 Minutes", value: "15 minutes" },
        { label: "1 Hour", value: "1 hour" },
        // { label: "6 Hours", value: "6 hours" },
        { label: "1 Day", value: "1 day" }
    ];
    const [monitorInterval, setMonitorInterval] = useState(update ? monitorData?.monitor_interval : '15 minutes');
    const [graphAggregationColumnOptions, setGraphAggregationColumnOptions] = useState([]);
    const [graphAggregationOptions, setGraphAggregationOptions] = useState([]);
    const [indEnabled, setIndEnabled] = useState(update ? monitorData?.ind_enabled === "Y" : true);
    const [indSendEmailNotifications, setIndSendEmailNotifications] = useState(update ? monitorData?.ind_send_email_notifications === "Y" : true);
    const [indStatusGraph, setIndStatusGraph] = useState(update ? monitorData?.ind_status_graph === "Y" : false);
    const [indCreateTimeline, setIndCreateTimeline] = useState(update ? monitorData?.ind_create_timeline === "Y" : true);
    const [graphAggregation, setGraphAggregation] = useState(update ? monitorData?.graph_aggregation : "average");
    const [graphAggregationColumn, setGraphAggregationColumn] = useState(update ? monitorData?.graph_aggregation_column : '');
    const [indTimelineFullImpact, setIndTimelineFullImpact] = useState(update ? monitorData?.ind_timeline_full_impact === "Y" : true);
    const [showGraphAggregationColumnOptions, setShowGraphAggregationColumnOptions] = useState(false);
    //page 4 elements
    const [yellowRule, setYellowRule] = useState(update ? monitorData?.yellow_rule === "1=1" ? "always" : monitorData?.yellow_rule === "1=2" ? "never" : monitorData?.yellow_rule : "always");
    const [redRule, setRedRule] = useState(update ? monitorData?.red_rule === "1=1" ? "always" : monitorData?.red_rule === "1=2" ? "never" : monitorData?.red_rule : "never");
    const [ruleOptions, setRuleOptions] = useState([]);
    const [yellowRuleMessage, setYellowRuleMessage] = useState(update ? monitorData?.yellow_rule_message : "This is a yellow alert");
    const [redRuleMessage, setRedRuleMessage] = useState(update ? monitorData?.red_rule_message : "This is a red alert");
    const [returnCodeThreshold, setReturnCodeThreshold] = useState(update ? monitorData?.return_code_threshold : "");
    const [redError, setRedError] = useState("");
    const [yellowError, setYellowError] = useState("");
    const [yellowMsgError, setYellowMsgError] = useState("");
    const [redMsgError, setRedMsgError] = useState("");
    //page 5 elements
    const [runtimeYellowThreshold, setRuntimeYellowThreshold] = useState(update ? monitorData?.runtime_yellow_threshold : 2000);
    const [runtimeRedThreshold, setRuntimeRedThreshold] = useState(update ? monitorData?.runtime_red_threshold : 10000);
    // const [timeout, setTimeout] = useState(update ? monitorData?.timeout : 10);
    //page 6 elements
    const rowCountOptions = [{ label: "above", value: "above" }, { label: "below", value: "below" }, { label: "exactly", value: "exactly" }];
    const [redByRowCountOption, setRedByRowCountOption] = useState(update ? monitorData?.red_by_row_count_option : "above");
    const [redByRowCountItems, setRedByRowCountItems] = useState(update ? monitorData?.red_by_row_count_items : 0);
    const [indRedByRowCount, setIndRedByRowCount] = useState(update ? monitorData?.ind_red_by_row_count === "Y" : false);
    const [yellowByRowCountOption, setYellowByRowCountOption] = useState(update ? monitorData?.yellow_by_row_count_option : "above");
    const [yellowByRowCountItems, setYellowByRowCountItems] = useState(update ? monitorData?.yellow_by_row_count_items : 0);
    const [indYellowByRowCount, setIndYellowByRowCount] = useState(update ? monitorData?.ind_yellow_by_row_count === "Y" : false);

    //page 7 elements
    const severityOptions = [
        { label: "Red", value: 3 },
        { label: "Yellow", value: 2 },
    ];

    const compareToPreviousCheckOptions = [
        { label: "higher", value: "higher" },
        { label: "lower", value: "lower" },
        { label: "different", value: "different" },
    ];

    const dataReductionFunctionOptions = [

        { label: "average", value: "average" },
        { label: "minimum", value: "minimum" },
        { label: "maximum", value: "maximum" },


    ];


    const renewAlertEveryOptions = [

        { label: "hour", value: "hour" },
        { label: "day", value: "day" },
        { label: "week", value: "week" },


    ];

    const [indFirstColumnUnique, setIndFirstColumnUnique] = useState(update ? monitorData?.ind_first_column_unique === "Y" : false);
    const [indAlertOnAddedMissing, setIndAlertOnAddedMissing] = useState(update ? monitorData?.ind_alert_on_added_missing === "Y" : false);
    const [alertOnAddedMissingSeverity, setAlertOnAddedMissingSeverity] = useState(update ? monitorData?.alert_on_added_missing_severity : 2);
    const [indAlertCompareToPreviousCheck, setIndAlertCompareToPreviousCheck] = useState(update ? monitorData?.ind_alert_compare_to_previous_check === "Y" : false);
    const [alertCompareToPreviousCheckSeverity, setAlertCompareToPreviousCheckSeverity] = useState(update ? monitorData?.alert_compare_to_previous_check_severity : 3);
    const [alertCompareToPreviousCheckHigherLower, setAlertCompareToPreviousCheckHigherLower] = useState(update ? monitorData?.alert_compare_to_previous_check_higher_lower : "higher");
    const [alertCompareToPreviousCheckPercent, setAlertCompareToPreviousCheckPercent] = useState(update ? monitorData?.alert_compare_to_previous_check_percent : 100);
    const [indComputeBaseline, setIndComputeBaseline] = useState(update ? monitorData?.ind_compute_baseline === "Y" : false);
    const [dataReductionFunction, setDataReductionFunction] = useState(update ? monitorData?.data_reduction_function : "average");
    const [renewAlertEvery, setRenewAlertEvery] = useState(update ? monitorData?.renew_alert_every : "day");

    const [isChanged, setIsChanged] = useState(update ? false : true);

    const commandLabel = useRef(monitorType === types.MONGO || monitorData?.db_type === types.MONGO ? "Mongo Shell" : monitorType === types.LINUX || (update && monitorData?.db_type === types.LINUX) ? "Python" : monitorType === types.REDIS || (update && monitorData?.db_type === types.REDIS) ? "Redis Command" : "SQL");
    const commandType = useRef(monitorType === types.LINUX || (update && monitorData?.db_type === types.LINUX) ? "python" : "sql");
    useEffect(() => {
        if (!mounted.current) {
            mounted.current = true;
        }
        else {
            let errors = [];
            let error = validator(constraints)("name", monitorName);
            (error == null || errors.push(error)) && setNameError(error);
            error = validator(constraints)("cmd", monitorScript);
            (error == null || errors.push(error)) && setCmdError(error);
            error = validator(constraints)("redRule", redRule);
            (error == null || errors.push(error)) && setRedError(error);
            error = validator(constraints)("yellowRule", yellowRule);
            (error == null || errors.push(error)) && setYellowError(error);
            error = validator(constraints)("redMsg", redRuleMessage);
            (error == null || errors.push(error)) && setRedMsgError(error);
            error = validator(constraints)("yellowMsg", yellowRuleMessage);
            (error == null || errors.push(error)) && setYellowMsgError(error);
            errors.length === 0 ? setValid(true) : setValid(false);

        }

    }, [monitorName, monitorScript, redRule, yellowRule, redRuleMessage, yellowRuleMessage])

    useEffect(() => {
        const hasChanged = (obj1, obj2) => {
            for (const [key, value] of Object.entries(obj2)) {
                if (key === 'columns') {
                    continue;
                }
                if (obj1?.hasOwnProperty(key) && obj1?.[key] !== value) {
                    return true;
                }
            }
            return false;
        }

        const hasColumnsChanged = (fields, columns) => {

            if (fields && columns) {


                for (const key in fields) {

                    const col = columns?.filter((c) => c.display_name === fields[key].display_name);
                    if (!col?.length || col?.length === 0 || col[0].column_type !== fields[key].column_type) {

                        return true;
                    }

                }

                for (const col of columns) {

                    let found = false;
                    for (const key in fields) {

                        if (fields[key].display_name === col.display_name) found = true;

                    }
                    if (!found) return true;
                }


            }




            return false;


        }

        let columns = {}
        let keys = Object.keys(fieldObjects);
        let _redRule = redRule;
        let _yellowRule = yellowRule;
        for (let i = 0; i < keys?.length; i++) {
            let column = fieldObjects[keys[i]]
            let column_name = column.column_name;
            let display_name = column.display_name;
            let column_type = column.column_type
            columns[column_name] =
            {
                "displayName": display_name,
                "type": column_type
            }

            if (column_type === "numeric") {
                _redRule = _redRule?.replaceAll(display_name, `${column_name}::numeric`);
                _yellowRule = _yellowRule?.replaceAll(display_name, `${column_name}::numeric`);
            }
            else {

                _redRule = _redRule?.replaceAll(display_name, column_name);
                _yellowRule = _yellowRule?.replaceAll(display_name, column_name);

            }



        }

        let data =
        {
            org_id: 1,
            node_id: null,
            db_type: update ? monitorData?.db_type : monitorType,
            monitor_name: monitorName,
            monitor_description: monitorDescription,
            monitor_module: monitorModule,
            pre_script: preScript,
            monitor_script: monitorScript,
            cmd_type: cmdType,
            monitor_interval: monitorInterval,
            ind_enabled: indEnabled ? "Y" : "N",
            ind_send_email_notifications: indSendEmailNotifications ? "Y" : "N",
            ind_status_graph: indStatusGraph ? "Y" : "N",
            ind_create_timeline: indCreateTimeline ? "Y" : "N",
            graph_aggregation: graphAggregation,
            graph_aggregation_column: graphAggregationColumn,
            ind_timeline_full_impact: indTimelineFullImpact ? "Y" : "N",
            yellow_rule: yellowRule === "always" ? "1=1" : yellowRule === "never" ? "1=2" : _yellowRule,
            red_rule: redRule === "always" ? "1=1" : redRule === "never" ? "1=2" : _redRule,
            yellow_rule_message: yellowRuleMessage,
            red_rule_message: redRuleMessage,
            return_code_threshold: returnCodeThreshold,
            runtime_yellow_threshold: runtimeYellowThreshold,
            runtime_red_threshold: runtimeRedThreshold,
            ind_red_by_row_count: indRedByRowCount ? "Y" : "N",
            ind_yellow_by_row_count: indYellowByRowCount ? "Y" : "N",
            yellow_by_row_count_option: yellowByRowCountOption,
            red_by_row_count_option: redByRowCountOption,
            yellow_by_row_count_items: yellowByRowCountItems,
            red_by_row_count_items: redByRowCountItems,
            ind_first_column_unique: indFirstColumnUnique ? "Y" : "N",
            ind_alert_on_added_missing: indAlertOnAddedMissing ? "Y" : "N",
            alert_on_added_missing_severity: alertOnAddedMissingSeverity,
            ind_alert_compare_to_previous_check: indAlertCompareToPreviousCheck ? "Y" : "N",
            alert_compare_to_previous_check_severity: alertCompareToPreviousCheckSeverity,
            alert_compare_to_previous_check_percent: alertCompareToPreviousCheckPercent,
            alert_compare_to_previous_check_higher_lower: alertCompareToPreviousCheckHigherLower,
            ind_compute_baseline: indComputeBaseline ? "Y" : "N",
            data_reduction_function: dataReductionFunction,
            ind_break_long_running: "Y",
            renew_alert_every: renewAlertEvery,
            // timeout: timeout,
        }

        if (update) {

            let changed = hasChanged(monitorData, data);
            if (!changed) {
                changed = hasColumnsChanged(fieldObjects, monitorData?.columns);
            }

            setIsChanged(changed);

        }

        data.columns = columns;



        setRequestData(data);
    }, [alertCompareToPreviousCheckHigherLower, alertCompareToPreviousCheckPercent, alertCompareToPreviousCheckSeverity, alertOnAddedMissingSeverity, cmdType, dataReductionFunction, fieldObjects, graphAggregation, graphAggregationColumn, indAlertCompareToPreviousCheck, indAlertOnAddedMissing, indComputeBaseline, indCreateTimeline, indEnabled, indFirstColumnUnique, indRedByRowCount, indSendEmailNotifications, indStatusGraph, indTimelineFullImpact, indYellowByRowCount, monitorData, monitorDescription, monitorModule, monitorInterval, monitorName, monitorScript, monitorType, preScript, redByRowCountItems, redByRowCountOption, redRule, redRuleMessage, renewAlertEvery, returnCodeThreshold, runtimeRedThreshold, runtimeYellowThreshold, update, yellowByRowCountItems, yellowByRowCountOption, yellowRule, yellowRuleMessage]);


    const onPreScriptEdit = (value) => {

        console.log(value);
        setPreScript(value);
    };


    const onMonitorScriptEdit = (value, viewUpdate) => {

        console.log(value);

        setMonitorScript(value)
    };


    useEffect(() => {

        let keys = Object.keys(fieldObjects);
        let _ruleOptions = [{ label: "always", type: "keyword" },
        { label: "never", type: "keyword" }
        ]

        keys.forEach((key) => _ruleOptions.push({ label: fieldObjects[key]?.display_name, type: 'keyword' }));

        setRuleOptions(_ruleOptions);

        setGraphAggregationColumnOptions(keys?.filter((key) => fieldObjects[key]?.column_type === "numeric").map((key) => ({ label: fieldObjects[key]?.display_name, value: fieldObjects[key]?.display_name })));
    },
        [fieldObjects]);

    useEffect(() => {
        const graphAggregationNumericOptions = [
            { label: "Average", value: "average" },
            { label: "Maximum", value: "maximum" },
            { label: "Minimum", value: "minimum" },
        ];

        const graphAggregationCommonOptions =
            [
                { label: "No aggregation", value: "no aggregation" },
                { label: "Number of fetched rows", value: "number_of_fetched_rows" },
                { label: "Monitor duration", value: "monitor_duration" },

            ]
        setGraphAggregationOptions(graphAggregationColumnOptions?.length > 0 ? [...graphAggregationCommonOptions, ...graphAggregationNumericOptions] : graphAggregationCommonOptions)
        setShowGraphAggregationColumnOptions(graphAggregationColumnOptions?.length > 0);
    }, [graphAggregationColumnOptions])

    useEffect(
        () => {

            switch (pageIndex) {

                case 2:
                    dispatch(MonitoringActions.getMessage([36]));
                    break;
                case 5:
                    dispatch(MonitoringActions.getMessage([32]));
                    break;
                case 6:
                    dispatch(MonitoringActions.getMessage([30, 29, 31]));
                    break;
                default:
                    break;


            }

        },
        [pageIndex, dispatch])


    const onAddField = () => {
        if (fieldCount > 9) {
            setColumnsError("could not add more than 10 columns");
            return;
        }
        else setColumnsError("");
        let uuid = uuidv4();
        let nextIndex = Object.keys(fieldObjects).length + 1;
        let column_name = `column${nextIndex}`
        let display_name = column_name;
        if (deletedFields.length > 0) {
            display_name = deletedFields.pop();
            setDeletedFields([...deletedFields]);

        }

        setFieldObjects({ ...fieldObjects, [uuid]: { display_name: display_name, column_name: column_name, column_type: 'text', placeholder: `Display name for ${display_name}` } });
        setFieldCount(fieldCount + 1);

    }


    useEffect(() => {

        if (update && monitorData) {


            let columns = {}
            let i = 0;
            monitorData?.columns?.forEach((c) => {
                i++;
                let uuid = uuidv4();
                columns[uuid] = { display_name: c.display_name, column_name: `column${i}`, column_type: c.column_type, placeholder: c.column_name };
            })

            setFieldObjects(columns);
            setFieldCount(i);

        }



    }, [update, monitorData]);

    const onDeleteField = (key) => {

        if (fieldCount < 2) {
            setColumnsError("monitor must have at least 1 column");
            return;
        }
        else setColumnsError("");

        if (fieldObjects[key]?.display_name.trim() !== "") {

            deletedFields.push(fieldObjects[key]?.display_name);
            setDeletedFields([...deletedFields]);

        }

        delete fieldObjects[key]
        setFieldObjects({ ...fieldObjects });
        setFieldCount(fieldCount - 1);


    }

    const onEditField = useCallback(
        (key, fieldName, value) => {

            setFieldObjects({
                ...fieldObjects, [key]:
                {
                    ...fieldObjects[key],
                    [fieldName]: value

                }
            });

        }, [fieldObjects]);

    const getHeader = (index) => {

        switch (index) {


            case 0:
                return `${update ? "Edit" : "Add"} Monitor ${monitorName} - Settings`;
            case 1:
                return `${update ? "Edit" : "Add"} Monitor ${monitorName} - Field names and types`;
            case 2:
                return `${update ? "Edit" : "Add"} Monitor ${monitorName} - Schedule`;
            case 3:
                return `${update ? "Edit" : "Add"} Monitor ${monitorName} - Thresholds on return values`;
            case 4:
                return `${update ? "Edit" : "Add"} Monitor ${monitorName} - Thresholds on runtime output`;
            case 5:
                return `${update ? "Edit" : "Add"} Monitor ${monitorName} - Rules on number of items`;
            case 6:
                return `${update ? "Edit" : "Add"} Monitor ${monitorName} - Rules based on first and columns retrieved`;
            default:
                return `${update ? "Edit" : "Add"} Monitor`;
        }

    }

    const onConfirm = () => {
        if (onOkClick) onOkClick(update, requestData);
    }

    function customCompletionDisplay() {
        return EditorView.updateListener.of(({ view, docChanged }) => {
            if (docChanged) {
                closeCompletion(view);

                debouncedStartCompletion(view);
            }
        });
    }

    const debouncedStartCompletion = debounce((view) => {
        startCompletion(view);
    }, 300);

    function completionSource(context) {
        const word = context.matchBefore(/\w*/);
        if (word.from === word.to || word.text.trim().length <= 0) return null;
        return {
            from: word.from,
            options: ruleOptions,
            filter: true,
        };
    }


    const renderFooter = () => {

        return (<div style={{ display: "flex", justifyContent: "space-between" }}>
            {pageIndex > 0 &&
                <Button
                    className="p-button-secondary"
                    label={"Back"}
                    icon="pi pi-arrow-left"
                    iconPos="left"
                    onClick={(() => setPageIndex(pageIndex - 1))} />
            }
            {pageIndex < 6 &&
                <Button
                    className="p-button-secondary"
                    label={"Next"}
                    icon="pi pi-arrow-right"
                    iconPos="right"
                    onClick={((e) => setPageIndex(pageIndex + 1))} />
            }

            <Button
                label={update ? "Update" : "Create"}
                icon={`pi pi-${isLoading ? 'spin pi-spinner' : 'check'}`}
                disabled={(update && !isChanged) || !valid || ! user?.is_superuser}
                iconPos="right"
                onClick={onConfirm} />

        </div>)

    }

    const renderPage1 = () => {

        return (<div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
            <div className="field">
                <label className="title">
                    Name
                </label>
                <InputText disabled={update || ! user?.is_superuser } placeholder="Enter the monitor`s name (required)" value={monitorName} className="p-inputtext-sm" onChange={(e) => {
                    setIsChanged(update ? monitorName !== monitorData?.monitor_name : monitorName !== e.target.value);
                    setMonitorName(e.target.value);
                }} />
                <div className="p-error text-xs">{nameError}</div>

            </div>
            <div className="field">
                <label className="title">
                    Description
                </label>
                <InputTextarea disabled = {! user?.is_superuser} className="monitor-description" value={monitorDescription} placeholder="Monitor description" onChange={(e) => {
                    setIsChanged(update ? monitorData?.monitor_description !== monitorDescription : monitorDescription !== e.target.value);
                    setMonitorDescription(e.target.value)
                }} />

            </div>
            <div className="field">
                <label className="title">
                    Module
                </label>
                <InputTextarea disabled = {! user?.is_superuser} className="monitor-module" value={monitorModule} placeholder="Monitor module" onChange={(e) => {
                    setIsChanged(update ? monitorData?.module !== monitorModule : monitorModule !== e.target.value);
                    setMonitorModule(e.target.value)
                }} />

            </div>
            <div className="field">
                <div className="flex">
                    <label className="title">
                        Pre script
                    </label>

                    <div className="field ml-auto mr-2">
                        <RadioButton disabled = {! user?.is_superuser} inputId="cmdType1" name="cmdType" value={commandType.current} checked={monitorType === types.LINUX || (update && monitorData?.db_type === types.LINUX) ? cmdType === "python" : cmdType === "sql"} onChange={(e) => {
                            setIsChanged(update ? monitorData?.cmd_type !== e.value : cmdType !== e.value);
                            setCmdType(e.value)
                        }} />
                        <label htmlFor="cmdType1" className="mb-auto mt-auto"> {commandLabel.current}</label>
                    </div>
                    <div className="field">
                        <RadioButton disabled = {! user?.is_superuser} inputId="cmdType2" name="cmdType" value="ssh" checked={cmdType === "ssh"} onChange={(e) => {
                            setIsChanged(update ? monitorData?.cmd_type !== e.value : cmdType !== e.value);
                            setCmdType(e.value)
                        }} />
                        <label htmlFor="cmdType2" className="mb-auto mt-auto">Linux/Unix shell</label>
                    </div>

                </div>

                <div className="field flex flex-column">
                    <CodeMirror
                        editable={user?.is_superuser}
                        key="preScriptEditor"
                        value={preScript}
                        theme={vscodeDark}
                        height="100px"
                        className="code-editor"
                        placeholder="Run any preparation logic before executing that command"
                        extensions={[monitorType === types.MONGO || monitorData?.db_type === types.MONGO ? langs.javascript() : langs.sql(), EditorView.lineWrapping]}
                        onChange={onPreScriptEdit}
                    />
                    <i className="pi pi-copy copy-script mr-2 mt-1 ml-auto" onClick={async () => await copyToClipboard(preScript)} />
                </div>




            </div>
            <div className="field">
                <div className="flex justify-content-between">
                    <label className="title">
                        Command
                    </label>
                </div>

                <div className="field flex flex-column">
                    <CodeMirror
                        editable={user?.is_superuser}
                        key="monitorScriptEditor"
                        theme={vscodeDark}
                        value={monitorScript}
                        className="code-editor"
                        extensions={[monitorType === types.MONGO || monitorData?.db_type === types.MONGO ? langs.javascript() : langs.sql(), EditorView.lineWrapping]}
                        height="100px"
                        placeholder="Command must return at least 1 column, multiple columns should be separated by spaces (required)"
                        onChange={onMonitorScriptEdit}
                    />
                    <i className="pi pi-copy copy-script mr-2 mt-1 ml-auto" onClick={async () => await copyToClipboard(monitorScript)} />
                    <div className="p-error text-xs">{cmdError}</div>
                </div>
            </div>
            <Tooltip target='.copy-script' content='copy'></Tooltip>
        </div>
        );

    }

    const renderPage2 = () => {

        return (<div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
            <div className="field">

                <label>Define the fields to be retrieved by monitoring script</label>

            </div>
            {Object.keys(fieldObjects).map((key) => (<div key={key} className="field grid">
                <div className="col">
                    <InputText disabled = {! user?.is_superuser} className={`p-inputtext-sm ${fieldObjects[key]?.display_name.trim() === "" ? "error-highlight" : ""}`} placeholder={fieldObjects[key]?.placeholder} value={fieldObjects[key]?.display_name} onChange={(e) => onEditField(key, "display_name", e.target.value)} />
                </div>
                <div className="col-fixed mt-2">
                    Type

                </div>
                <div className="col-fixed">
                    <Dropdown disabled = {! user?.is_superuser} style={{ width: "135px" }} options={columnTypeOptions} value={fieldObjects[key]?.column_type} onChange={(e) => onEditField(key, "column_type", e.value)} />
                </div>
                <div className="col-fixed">

                    <Button disabled = {! user?.is_superuser} className="p-button-danger" icon="pi pi-trash" onClick={(e) => onDeleteField(key)} />

                </div>

            </div>))}
            <div className="field grid">
                <div className="col">
                    <div><i className="pi pi-info-circle mr-2" />These names will be referenced later when setting up thresholds for raising alerts!</div>
                    <div className="p-error text-sm">{columnsError} </div>
                </div>
                <div className="col-fixed mt-2 ml-auto">
                    <Button disabled = {! user?.is_superuser}
                        label={"Add Field"}
                        icon="pi pi-plus"
                        iconPos="right"
                        onClick={onAddField} />
                </div>
            </div>
        </div>
        );

    }

    const renderPage3 = () => {

        return (<div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
            <div className="field grid">
                <div className="col-fixed">
                    <Checkbox disabled = {! user?.is_superuser} inputId="cb1" value={indEnabled} onChange={() => {

                        setIndEnabled(!indEnabled);
                    }
                    } checked={indEnabled}></Checkbox>
                </div>
                <label htmlFor="cb1" className="p-checkbox-label col-fixed">Enabled</label>
            </div>
            <div className="field grid">
                <div className="col-fixed">
                    <Checkbox disabled = {! user?.is_superuser} inputId="cb2" value={indSendEmailNotifications} onChange={() => {
                        setIndSendEmailNotifications(!indSendEmailNotifications);
                    }
                    } checked={indSendEmailNotifications}></Checkbox>
                </div>
                <label htmlFor="cb2" className="p-checkbox-label col-fixed">Send mail notifications</label>
            </div>
            <div className="field grid flex align-items-center">
                <div className="col-fixed">
                    <Checkbox className="mb-2" inputId="cb3" value={indStatusGraph} disabled={Object.keys(fieldObjects)?.filter((key) => fieldObjects[key]?.column_type === "numeric")?.length === 0 || ! user?.is_superuser} onChange={() => {
                        setIndStatusGraph(!indStatusGraph)
                    }} checked={indStatusGraph}></Checkbox>
                </div>
                <label htmlFor="cb3" className="p-checkbox-label col-fixed">Show graph in status panel</label>

                <div className="col-fixed">
                    <div className="flex align-items-center">
                        <Dropdown disabled={!indStatusGraph || ! user?.is_superuser} className="p-inputtext-sm mr-2" style={{ width: "135px" }} options={graphAggregationOptions} value={graphAggregation} onChange={(e) => {
                            setIsChanged(update ? monitorData?.graph_aggregation !== e.value : graphAggregation !== e.value);
                            setGraphAggregation(e.value)
                        }} />  {showGraphAggregationColumnOptions && <div className="flex align-items-center"> <div> of </div> <Dropdown disabled={!indStatusGraph} placeholder="Select column name" style={{ width: "220px" }} className="p-inputtext-sm ml-2 mr-2" options={graphAggregationColumnOptions} value={graphAggregationColumn} onChange={(e) => {
                            setIsChanged(update ? monitorData?.graph_aggregation_column !== e.value : graphAggregationColumn !== e.value);
                            setGraphAggregationColumn(e.value)
                        }} /></div>} <i className="pi pi-question-circle ml-2 graph-tip align-self-center" />
                        <Tooltip className="tooltip-base" target=".graph-tip" content={messages?.[36]} position="top" />

                    </div>
                </div>
            </div>
            <div className="field grid">
                <div className="col-fixed">
                    <Checkbox disabled = {! user?.is_superuser} inputId="cb4" value={indCreateTimeline} onChange={() => {
                        let value = !indCreateTimeline;
                        setIndCreateTimeline(value);
                        if (!value) setIndTimelineFullImpact(value);
                    }} checked={indCreateTimeline}></Checkbox>
                </div>
                <label htmlFor="cb4" className="p-checkbox-label col-fixed">Show in timelines</label>


            </div>
            <div className="field grid ml-3">
                <div className="col-fixed">
                    <Checkbox disabled = {! user?.is_superuser} inputId="cb5" value={indTimelineFullImpact} onChange={() => {
                        setIndTimelineFullImpact(!indTimelineFullImpact)
                    }
                    } checked={indTimelineFullImpact}></Checkbox>
                </div>
                <label htmlFor="cb5" className="p-checkbox-label col-fixed">This monitor has full impact on the overall timeline</label>

            </div>
            <div className="field">
                <div className="flex align-items-center">
                    <div>Run this monitor every</div> <Dropdown style={{ width: "200px" }} className="p-inputtext-sm ml-2" options={monitorIntervalOptions} value={monitorInterval} onChange={(e) => {
                        setIsChanged(update ? monitorData?.monitor_interval !== e.value : monitorInterval !== e.value)
                        setMonitorInterval(e.value);
                    }} />
                </div>
            </div>
            <div className="field">
                <div className="flex align-items-center">
                    <div>Renew alert every</div> <Dropdown disabled = {! user?.is_superuser} style={{ width: "220px" }} className="p-inputtext-sm ml-2" options={renewAlertEveryOptions} value={renewAlertEvery} onChange={(e) => {
                        setRenewAlertEvery(e.value)
                    }} />  <i className="pi pi-question-circle ml-2 renew-tip align-self-center" />
                    <Tooltip className="tooltip-base" target=".renew-tip" content={messages?.[31]} position="top" />
                </div>
            </div>
        </div>



        );

    }



    const renderPage4 = () => {

        return (<div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
            <div className="field grid">
                <div className="col-fixed title" style={{ width: "240px" }}>
                    Yellow threshold rule (required)
                </div>
                <div className="col">
                    <CodeMirror
                        editable={user?.is_superuser}
                        value={yellowRule}
                        theme={vscodeDark}
                        basicSetup={{ foldGutter: false, lineNumbers: false }}
                        className="rule-input p-inputtext p-component p-inputtext-sm"
                        placeholder="Yellow threshold rule"
                        extensions={[EditorView.lineWrapping, autocompletion({ activateOnTyping: true, filter: true, override: [completionSource] }), customCompletionDisplay()]}
                        onChange={(value) => {
                            let stripped = value?.replace(/(\r\n|\n|\r)/gm, "");
                            setYellowRule(stripped);
                        }} />

                </div>
                <div className="p-error text-xs">{yellowError}</div>


            </div>

            <div className="field grid">
                <div className="col-fixed title" style={{ width: "240px" }}>
                    Yellow threshold notification (required)
                </div>
                <div className="col">
                    <InputText disabled = {! user?.is_superuser} placeholder="Yellow threshold notification" className="p-inputtext-sm" value={yellowRuleMessage} onChange={(e) => {
                        setIsChanged(update ? monitorData?.yellow_rule_message !== e.target.value : yellowRuleMessage !== e.target.value)
                        setYellowRuleMessage(e.target.value)
                    }} />
                </div>
                <div className="p-error text-xs">{yellowMsgError}</div>

            </div>




            <div className="field grid">
                <div className="col-fixed title" style={{ width: "240px" }}>
                    Red threshold rule (required)
                </div>
                <div className="col">
                    <CodeMirror
                        editable = {user?.is_superuser}
                        value={redRule}
                        theme={vscodeDark}
                        basicSetup={{ foldGutter: false, lineNumbers: false }}
                        className="rule-input p-inputtext p-component p-inputtext-sm"
                        placeholder="Red threshold rule"
                        extensions={[EditorView.lineWrapping, autocompletion({ activateOnTyping: true, filter: true, override: [completionSource] }), customCompletionDisplay()]}
                        onChange={(value) => {
                            let stripped = value?.replace(/(\r\n|\n|\r)/gm, "");
                            setRedRule(stripped);
                        }} />
                    <div className="p-error text-xs">{redError}</div>
                </div>

            </div>

            <div className="field grid">
                <div className="col-fixed title" style={{ width: "240px" }}>
                    Red threshold notification (required)
                </div>
                <div className="col">
                    <InputText disabled = {! user?.is_superuser} className="p-inputtext-sm" value={redRuleMessage} placeholder="Red threshold notification" onChange={(e) => {
                        setRedRuleMessage(e.target.value)
                    }} />
                    <div className="p-error text-xs">{redMsgError}</div>
                </div>
            </div>
            <div className="field grid">
                <div className="col-fixed title" style={{ width: "240px" }}>
                    Return code
                </div>
                <div className="col">
                    <InputText disabled = {! user?.is_superuser} value={returnCodeThreshold} placeholder="Return code threshold" className="p-inputtext-sm" onChange={(e) => {
                        setReturnCodeThreshold(e.target.value)
                    }} />
                </div>

            </div>
        </div>
        );

    }

    const renderPage5 = () => {

        return (<div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
            <div className="field grid flex align-items-center">
                <div className="col-fixed title" style={{ width: "300px" }}>
                    Yellow if total run duration exceeds
                </div>
                <div className="col flex align-items-center">
                    <InputNumber disabled = {! user?.is_superuser} style={{ width: "135px" }} useGrouping={false} min={0} className="p-inputtext-sm mr-2" value={runtimeYellowThreshold} onValueChange={(e) => {
                        setRuntimeYellowThreshold(e.value)
                    }} /> <div> milliseconds</div>
                </div>
            </div>
            <div className="field grid flex align-items-center">
                <div className="col-fixed title" style={{ width: "300px" }}>
                    Red if total run duration exceeds
                </div>
                <div className="col flex align-items-center">
                    <InputNumber style={{ width: "135px" }} useGrouping={false} min={0} className="p-inputtext-sm mr-2" value={runtimeRedThreshold} onValueChange={(e) => {
                        setRuntimeRedThreshold(e.value)
                    }} /> <div> milliseconds</div>
                </div>
            </div>
            {/* <div className="field grid flex align-items-center">
            <div className="col-fixed title" style={{ width: "300px" }}>
                Timeout
            </div>
            <div className="col flex align-items-center">
                <InputNumber disabled = {! user?.is_superuser} style={{ width: "135px" }} useGrouping={false} min={0}  className="p-inputtext-sm mr-2" value={timeout} onValueChange={(e) => {
                    setTimeout(e.value)
                }} /> <div> seconds</div>
            </div>
            </div> */}
        </div>
        );

    }

    const renderPage6 = () => {

        return (<div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
            <div className="field grid flex align-items-center">
                <div className="col-fixed  mb-2">
                    <Checkbox disabled = {! user?.is_superuser} inputId="cb6" value={indRedByRowCount} onChange={() => {
                        setIndRedByRowCount(!indRedByRowCount)
                    }} checked={indRedByRowCount}></Checkbox>
                </div>
                <label htmlFor="cb6" className="p-checkbox-label col-fixed">Red alert if total number of fetched rows is</label>

                <div className="col-fixed">
                    <div className="flex">
                        <Dropdown  disabled={!indRedByRowCount || ! user?.is_superuser} style={{ width: "155px" }} className="p-inputtext-sm mr-2" options={rowCountOptions} value={redByRowCountOption} onChange={(e) => {
                            setRedByRowCountOption(e.value)
                        }} />   <InputNumber disabled={!indRedByRowCount} style={{ width: "140px" }} useGrouping={false} min={0} className="p-inputtext-sm" value={redByRowCountItems} onValueChange={(e) => {
                            setRedByRowCountItems(e.value)
                        }} />  <i className="pi pi-question-circle ml-2 items-tip align-self-center" />
                        <Tooltip className="tooltip-base" target=".items-tip" content={messages?.[32]} position="top" />
                    </div>
                </div>
            </div>
            <div className="field grid flex align-items-center">
                <div className="col-fixed mb-2">
                    <Checkbox disabled = {! user?.is_superuser} inputId="cb6" value={indYellowByRowCount} onChange={() => {
                        setIndYellowByRowCount(!indYellowByRowCount)
                    }} checked={indYellowByRowCount}></Checkbox>
                </div>
                <label htmlFor="cb6" className="p-checkbox-label col-fixed">Yellow alert if total number of fetched rows is</label>

                <div className="col-fixed">
                    <div className="flex">
                        <Dropdown disabled={!indYellowByRowCount || ! user?.is_superuser} style={{ width: "135px" }} className="p-inputtext-sm mr-2" options={rowCountOptions} value={yellowByRowCountOption} onChange={(e) => {
                            setYellowByRowCountOption(e.value)
                        }} />   <InputNumber disabled={!indYellowByRowCount} style={{ width: "140px" }} useGrouping={false} min={0} className="p-inputtext-sm" value={yellowByRowCountItems} onValueChange={(e) => {
                            setYellowByRowCountItems(e.value)
                        }} />
                    </div>
                </div>
            </div>
        </div>
        );
    }

    const renderPage7 = () => {

        return (<div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">

            <div className="field grid flex">
                <div className="col-fixed">
                    <Checkbox disabled = {! user?.is_superuser} inputId="cb7" value={indFirstColumnUnique} onChange={() => {
                        let value = !indFirstColumnUnique;
                        setIndFirstColumnUnique(value);
                        if (!value) {

                            setIndAlertOnAddedMissing(value);
                            setIndAlertCompareToPreviousCheck(value);

                        }
                    }} checked={indFirstColumnUnique}></Checkbox>
                </div>
                <label htmlFor="cb7" className="col-fixed p-checkbox-label"> {`${ruleOptions[2]?.label ? ruleOptions[2]?.label : "first column"} is unique - means there can be no two rows with the same name`}</label> <i className="pi pi-question-circle ml-2 unique-tip align-self-center" />
                <Tooltip className="tooltip-base" target=".unique-tip" content={messages?.[30]} position="top" />
            </div>
            <div className="field grid ml-3 flex align-items-center">

                <div className="col-fixed">
                    <Checkbox disabled={!indFirstColumnUnique || ! user?.is_superuser} inputId="cb8" value={indAlertOnAddedMissing} onChange={() => {
                        setIndAlertOnAddedMissing(!indAlertOnAddedMissing)
                    }} checked={indAlertOnAddedMissing}></Checkbox>
                </div>
                <div className="col-fixed flex align-items-center">
                    <label htmlFor="cb8" className="p-checkbox-label">Raise a </label>
                    <Dropdown disabled={!indAlertOnAddedMissing || ! user?.is_superuser} style={{ width: "120px" }} className="p-inputtext-sm ml-2 mr-2" options={severityOptions} value={alertOnAddedMissingSeverity} onChange={(e) => {
                        setAlertOnAddedMissingSeverity(e.value)
                    }} />
                    <div>{`alert if a  ${ruleOptions?.length > 2 ? ruleOptions[2].label : "first column"}  is added or removed compared to previous check`}</div>
                </div>

            </div>
            {ruleOptions?.length > 3 && (
              <div className="field grid ml-3 flex align-items-center">
                  <div className="col-fixed">
                      <Checkbox disabled={!indFirstColumnUnique || ! user?.is_superuser} inputId="cb9" value={indAlertCompareToPreviousCheck} onChange={() => {
                          setIndAlertCompareToPreviousCheck(!indAlertCompareToPreviousCheck)
                      }} checked={indAlertCompareToPreviousCheck}></Checkbox>
                  </div>
                  <div className="col-fixed flex align-items-center">
                      <label htmlFor="cb9" className="p-checkbox-label">
                        Raise a{" "}
                      </label>
                      <Dropdown
                        disabled={!indAlertCompareToPreviousCheck || !user?.is_superuser}
                        style={{ width: "120px" }}
                        className="p-inputtext-sm ml-2 mr-2"
                        options={severityOptions}
                        value={alertCompareToPreviousCheckSeverity}
                        onChange={(e) => {
                          setAlertCompareToPreviousCheckSeverity(e.value);
                        }}
                      />
                      <div>{`alert if current ${ruleOptions[3]?.label} value is`}</div>
                      <Dropdown
                        disabled={!indAlertCompareToPreviousCheck || !user?.is_superuser}
                        style={{ width: "150px" }}
                        className="p-inputtext-sm ml-2 mr-2"
                        options={compareToPreviousCheckOptions}
                        value={alertCompareToPreviousCheckHigherLower}
                        onChange={(e) => {
                          setAlertCompareToPreviousCheckHigherLower(e.value);
                        }}
                      />
                      {
                        alertCompareToPreviousCheckHigherLower !== "different" && (
                          <div className="flex align-items-center">
                            <div>by</div>
                            <InputNumber
                              disabled={!indAlertCompareToPreviousCheck || !user?.is_superuser}
                              style={{ width: "135px" }}
                              useGrouping={false}
                              min={0}
                              max={100}
                              className="p-inputtext-sm ml-2 mr-2"
                              value={alertCompareToPreviousCheckPercent}
                              onValueChange={(e) => {
                                setIsChanged(
                                  update
                                    ? monitorData?.alert_compare_to_previous_check_percent !== e.value
                                    : alertCompareToPreviousCheckPercent !== e.value
                                );
                                setAlertCompareToPreviousCheckPercent(e.value);
                              }}
                            />
                            <div>% </div>
                          </div>
                        )
                      }
                      <div className="ml-1">compared to previous check</div>
                  </div>
              </div>
            )}
            <div className="field grid flex align-items-center">
                <div className="col-fixed">
                    <Checkbox disabled = {! user?.is_superuser} className="mb-2" inputId="cb10" value={indComputeBaseline} onChange={() => {
                        setIndComputeBaseline(!indComputeBaseline)
                    }} checked={indComputeBaseline}></Checkbox>
                </div>
                <label htmlFor="cb10" className="p-checkbox-label col-fixed">Compute baseline</label>
                <div className="col-fixed">
                    <div className="flex align-items-center">
                        <Dropdown disabled={!indComputeBaseline || ! user?.is_superuser} className="p-inputtext-sm" style={{ width: "135px" }} options={dataReductionFunctionOptions} value={dataReductionFunction} onChange={(e) => {
                            setDataReductionFunction(e.value)
                        }} /><i className="pi pi-question-circle ml-2 baseline-tip align-self-center" />
                        <Tooltip className="tooltip-base" target=".baseline-tip" content={messages?.[29]} position="top" />
                    </div>
                </div>
            </div>

        </div>
        );

    }










    return (
        <Dialog
            className='monitoring-wizard'
            header={getHeader(pageIndex)}
            visible={visible}
            style={{ maxWidth: "90vw" }}
            onHide={onHide}
            footer={renderFooter()}>
            <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
                {pageIndex === 0 && renderPage1()}
                {pageIndex === 1 && renderPage2()}
                {pageIndex === 2 && renderPage3()}
                {pageIndex === 3 && renderPage4()}
                {pageIndex === 4 && renderPage5()}
                {pageIndex === 5 && renderPage6()}
                {pageIndex === 6 && renderPage7()}
            </div>
        </Dialog>
    );




}
