import { Dialog } from "primereact/dialog";
import { Button } from "primereact/button";
import { Tag } from 'primereact/tag';
import { Dropdown } from "primereact/dropdown";
import { useState, useEffect } from "react";
import { InputText } from "primereact/inputtext";
import { RadioButton } from "primereact/radiobutton";
import { Calendar } from "primereact/calendar";
import moment from "moment";
import validator from "../../../util/validator";
import constraints from "./constraints";
import ApiBar from "../../../components/apiBar";
import copyToClipboard from "../../../util/copyToClipboard";
import createAPIcommand from "../../../util/createApiCommand";
import { useSelector } from "react-redux";
import ErrorTextBox from "../../../components/errorTextBox";
import { Checkbox } from "primereact/checkbox";
import config, { clusterRoles, restoreOptions } from "../../../config";

export const RestoreMsSQLNode = (props) => {

  const [targetBackup, setTargetBackup] = useState(null);
  const [backupOptions, setBackupOptions] = useState([]);
  const [restoreConfirm, setRestoreConfirm] = useState("");

  const [targetBackupError, setTargetBackupError] = useState("");
  const [restoreTimeError, setRestoreTimeError] = useState("");
  const [apiTooltip, setApiTooltip] = useState(" ");
  const [restoreTime, setRestoreTime] = useState(null);
  const [restoreRadio, setRestoreRadio] = useState(restoreOptions.THIS.value);
  const [autostart, setAutoStart] = useState(true);
  const { accessToken } = useSelector((state) => state.app);
  const { errorMessage, isLoading } = useSelector((state) => state.layout);
  const [restoreConfirmError, setRestoreConfirmError] = useState("");

  const validate = () => {

    const errors = [];


    let error = validator(constraints)("targetBackup", targetBackup);
    (error == null || errors.push(error)) && setTargetBackupError(error);

    error = validator(constraints)("restoreConfirm", restoreConfirm);
    (error == null || errors.push(error)) && setRestoreConfirmError(error);

    if (restoreRadio === restoreOptions.CUSTOM.value) {
      error = validator(constraints)("time", restoreTime);
      (error == null || errors.push(error)) && setRestoreTimeError(error);
    }
    else setRestoreTimeError("");

    return errors;

  }


  const handleOk = () => {

    let errors = validate();

    if (errors.length === 0) {

      props.onOkClick(
        {
          "bu_id": targetBackup?.id,
          "autostart": autostart,
          "restore_option": restoreRadio,
          "restore_time": restoreTime ? moment(restoreTime).format("YYYY-MM-DD HH:mm:ss") : restoreTime
        });

    }
  }

  const onRestApiClick = async () => {

    let errors = validate()

    if (errors.length === 0) {

      setApiTooltip(await copyToClipboard(createAPIcommand({
        "bu_id": targetBackup?.name,
        "autostart": autostart,
        "restore_option": restoreRadio,
        "restore_time": restoreTime ? moment(restoreTime).format("YYYY-MM-DD HH:mm:ss") : restoreTime

      }, accessToken, "restore_mssql_node/")));
    }
  }

  useEffect(() => {

    if (props.data) {

      const temp = [];
      props.data.backups?.forEach((b) => temp.push({ label: `${moment(b.created_date).format("MMM DD,YYYY, hh:mm A")}  ${b.display_size}`, value: b }));
      props.data.nodes.forEach((s) => s.backups?.forEach((b) => temp.push({ label: `${moment(b.created_date).format("MMM DD,YYYY, hh:mm A")}  ${b.display_size}`, value: b })))
      setBackupOptions(temp)

    }

  }, [props.data]);

  const renderFooter = () => {
    return (
      <div style={{ display: "flex", justifyContent: "space-between" }}>
        <Button label={"Close"} icon="pi pi-times" onClick={props.onHide} className="p-button-text" />
        <Button
          label={"Submit"}
          icon= {`pi pi-${isLoading ? 'spin pi-spinner' : 'check'}`}
          disabled={!targetBackup || targetBackup?.status === false}
          iconPos="right"
          onClick={handleOk} autoFocus />
      </div>
    );

  }



  return (
    <Dialog
      className='restore-mssql_node'
      header={`Restore Primary`}
      visible={props.visible}
      style={{ width: '40vw' }}
      onHide={props.onHide}
      footer={renderFooter()}>
      <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
        <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
          <div className="field grid">
            <div className="col">
              <div className="title">{"Choose backup to restore:"}</div>

              <div className="field">
                <Dropdown className="p-inputtext-sm"
                  value={targetBackup}
                  options={backupOptions}
                  onChange={(e) => setTargetBackup(e.value)}
                  placeholder="Select backup"
                  emptyMessage = "no backups found" 
                  />
                <div className="error-help">{<small className="p-error p-d-block">{targetBackupError}</small>}</div>
              </div>

              {targetBackup?.is_sched && <div>
                <div className="field grid">
                  <div className="col">
                    {
                      Object.keys(restoreOptions).map(key => (<div key={key} className="field-radiobutton">
                        <RadioButton inputId={key} name="restoreRadio" value={restoreOptions[key].value} checked={restoreRadio === restoreOptions[key].value} onChange={(e) => setRestoreRadio(e.value)} />
                        <label htmlFor={key}>{restoreOptions[key].label}</label>
                      </div>))
                    }
                  </div>
                </div>

                <div className="field grid align-items-center">
                  <div className="col-fixed">Select time:</div>
                  <div className="col-6">
                    <Calendar
                      showSeconds
                      maxDate={new Date()}
                      showTime
                      hourFormat="24"
                      disabled={restoreRadio !== restoreOptions.CUSTOM.value || targetBackup?.is_sched === false}
                      value={restoreTime}
                      onChange={(e) => setRestoreTime(e.value)} />
                    <div className="error-help">{<small className="p-error p-d-block">{restoreTimeError}</small>}</div>
                  </div>
                </div>

              </div>}



              <div className="field grid align-items-center">
                <div className="col-fixed"> Start up instance after restore:</div>
                <div className="col">

                  <Checkbox checked={autostart} disabled={props.node?.cluster_role === clusterRoles.PRIMARY && props.node?.nodes?.filter((s) => s.status === 'green').length > 0} onChange={(e) => setAutoStart(e.checked)} />

                </div>
              </div>

              <div className="field">
                <Tag value="The current content of the node will be overridden, this operation is not reversible!" icon="pi pi-exclamation-triangle" severity="warning" />
              </div>

              <div className="field grid align-items-center">
                <div className="col">

                  <InputText
                    className="p-inputtext-sm"
                    placeholder={`Please type "${config.restoreConfirm}" here in order to continue`}
                    value={restoreConfirm}
                    onChange={(e) => setRestoreConfirm(e.target.value)} />

                  <div className="error-help">{<small className="p-error p-d-block">{restoreConfirmError}</small>}</div>
                </div>
              </div>

            </div>
          </div>
          {(errorMessage || targetBackup?.status === false) && <ErrorTextBox error={errorMessage ? errorMessage?.message ? errorMessage.message : errorMessage : "This backup did not complete successfully"} />}
          <ApiBar tooltip={apiTooltip} onClick={onRestApiClick} />
        </div>
      </div>

    </Dialog>)

};

export default RestoreMsSQLNode;