import { Dialog } from "primereact/dialog";
import { Button } from "primereact/button";
import { useState } from "react";
import { InputText } from "primereact/inputtext";
import { Password } from "primereact/password";
import { useDispatch, useSelector } from "react-redux";
import { StatusActions } from "../../../store/status/reducer";
import { LayoutActions } from "../../../store/layout/reducer";
import ErrorTextBox from "../../../components/errorTextBox";
import validator from "../../../util/validator";
import constraints from "./constraints";
import createAPIcommand from "../../../util/createApiCommand";
import copyToClipboard from "../../../util/copyToClipboard";
import ApiBar from "../../../components/apiBar";
import { base64Encode } from "../../../util/base64Encode";
import { footer } from "../../../util/passwordHelpers";


export const CredentialsDialog = (props) => {

    const [username, setUsername] = useState("");
    const [password, setPassword] = useState("");
    const [usernameError, setUsernameError] = useState(null);
    const [passwordError, setPasswordError] = useState(null);
    const [apiTooltip, setApiTooltip] = useState(" ");

    const { isLoading, valid_credentials } = useSelector(state => state.status);
    const {errorMessage} = useSelector((state) => state.layout);
    
    const { accessToken } = useSelector(state => state.app);

    const dispatch = useDispatch();

    const validate = () => {

        let errors = [];

        let error = validator(constraints)("username", username);
        (error == null || errors.push(error)) && setUsernameError(error);

        error = validator(constraints)("password", password);
        (error == null || errors.push(error)) && setPasswordError(error);

        return errors;


    }

    const handleOk = () => {

        let errors = validate();

        if (errors.length === 0)
        {

            if (valid_credentials?.result === "credentials valid") {

                props.onOkClick(
                    {

                        "username": username,
                        "password": base64Encode(password),
                        "pdb_id": props.node.id

                    });

            
            
                dispatch(StatusActions.setValidCredentials(null));
            }
    

        }
                
    

    }


    const onRestApiClick = async () => {

        let errors = validate();

        if (errors.length === 0)
        {

            if (valid_credentials?.result === "credentials valid") {


                setApiTooltip(await copyToClipboard(createAPIcommand(
                    {

                        "username": username,
                        "password": base64Encode(password),
                        "pdb_id": props.node.location

                    }
                    , accessToken, "update_pdb_credentials/")));

            
            
                dispatch(StatusActions.setValidCredentials(null));
            }
    

        }
                
    

    }

    const renderFooter = () => {
        return (
            <div style={{ display: "flex", justifyContent: "space-between" }}>
                <Button label={"Close"} icon="pi pi-times" onClick={props.onHide} className="p-button-text" />
                <Button
                    label={"Submit"}
                    icon= {`pi pi-${isLoading ? 'spin pi-spinner' : 'check'}`}
                    iconPos="right"
                    onClick={handleOk} autoFocus />
            </div>
        );

    }

    const validateCredentials = () => {

        let errors = validate();

        if (errors.length === 0)
        {

            dispatch(LayoutActions.checkPDBCredentials(
                {

                    "username": username,
                    "password": base64Encode(password),
                    "pdb_id": props.node_id

                }
            ));
        }
      

    }



    return (

        <Dialog
            className='credentials-dialog'
            header={`Database credentials`}
            visible={props.visible}
            style={{ width: '40hw'}}
            onHide={props.onHide}
            footer={renderFooter()}>
            <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
                <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
                    <div className="field grid">
                        <div className="col-fixed">Username:</div>
                        <div className="col">
                            <span className="p-input-icon-right flex">
                                <i className={isLoading ? "pi pi-spin pi-spinner align-self-center" : ""} />
                                <InputText className="p-inputtext-sm"
                                    placeholder="Username"
                                    value={username}
                                    onChange={(e) => 
                                    {

                                        setUsername(e.target.value);
                                        dispatch(StatusActions.setValidCredentials({"result": "credentials not checked"}))

                                    }}
                                    autoComplete="new-password" />
                            </span>
                            <div className="error-help"> <small className={"p-error p-d-block"}>{usernameError}</small></div>
                        </div>


                    </div>

                    <div className="field grid">
                        <div className="col-fixed" >Password:</div>
                        <div className="col">

                            <Password
                                className="p-inputtext-sm"
                                placeholder="Password"
                                value={password}
                                footer = {() => footer(password)}
                                autoComplete="new-password"
                                onChange={(e) => 
                                {

                                    dispatch(StatusActions.setValidCredentials({"result": "credentials not checked"}));
                                    setPassword(e.target.value);
                                }}
                                toggleMask
                                icon={<i className={isLoading ? "pi pi-spin pi-spinner" : ""} />}
                            />
                            <div className="error-help"><small className="p-error p-d-block">{passwordError}</small></div>
                            <div className="error-help"> <small style={{ display: "block" }} className={valid_credentials?.result === "credentials valid" ? "" : "p-error p-d-block"}>{valid_credentials?.result}</small></div>
                            <Button className="p-button-sm" style={{width: "80px"}} label="Check" onClick={validateCredentials} />
                        </div>
                    </div>
                    {errorMessage && <ErrorTextBox error = {errorMessage?.message ? errorMessage.message : errorMessage}/>}
                    <ApiBar tooltip = {apiTooltip} onClick = {onRestApiClick}/>
                </div>
            </div>

        </Dialog>)

};

export default CredentialsDialog;