
import { Dialog } from "primereact/dialog";
import { Button } from "primereact/button";
import { Dropdown } from "primereact/dropdown";
import { useState, useEffect } from "react";
import { useSelector, useDispatch } from "react-redux";
import { LayoutActions } from "../../../store/layout/reducer";
import moment from "moment";
import validator from "../../../util/validator";
import constraints from "./constraints";
import ApiBar from "../../../components/apiBar";
import createAPIcommand from "../../../util/createApiCommand";
import copyToClipboard from "../../../util/copyToClipboard";
import ErrorTextBox from "../../../components/errorTextBox";
import { clusterRoles } from "../../../config";


export const ImportCDBPDBDialog = (props) => {

  const [targetExport, setExport] = useState(null);
  const [exportOptions, setExportOptions] = useState([]);
  const [pageIndex, setPageIndex] = useState(0);
  const { user_exists, dump_exists, isLoading, errorMessage} = useSelector((state) => state.layout);
  const [dumpCheck, setDumpCheck] = useState("");
  const [userCheck, setUserCheck] = useState("");
  const [exportError, setExportError] = useState("");
  const [exportDetails, setExportDetails] = useState(null);
  const [apiTooltip, setApiTooltip] = useState(false);
  const {accessToken} = useSelector((state) => state.app); 
  const dispatch = useDispatch();

  useEffect(() => {

    if (props.exports) {
      const temp = [];
      props.exports.forEach((item) => temp.push({ label: `${item.node_name}_(${moment(item.created_date).format("MMM DD,YYYY, hh:mm A")})`, value: item}));
      setExportOptions(temp);
    }

  }, [props.exports]);

  useEffect(() => {

    if (dump_exists) {
      if (dump_exists?.error) { 
        setDumpCheck(`Dump file ${dump_exists.dump_file} does not exist.\n Cannot continue.`);

      }
      else setDumpCheck("ok");
    }

  }, [dump_exists]);

  useEffect(() => {

    if (user_exists) {
      if (!user_exists?.error) setUserCheck(`This will import ${user_exists.userName === "All users" ? "all users" : `user ${user_exists.userName}` } into the ${props.node.type.includes(clusterRoles.CDB) ? clusterRoles.CDB : clusterRoles.PDB} ${user_exists.pdbName}.`);
      else setUserCheck(`The schema ${user_exists.error} already exists.\nThis will DROP the user at the node and recreate it from the export file`);
    }

  }, [user_exists, props.node]);

  const onRestApiClick = async () => {

    if (!user_exists || user_exists?.detail || !dump_exists || dump_exists?.error) setApiTooltip(" ");
    else setApiTooltip(await copyToClipboard(createAPIcommand({
      "export_id": targetExport?.id
    }, accessToken,  props.node.type.includes(clusterRoles.PDB) ? 'import_pdb/' : 'import_cdb/')))


  }

  const renderImportPage = () => {

    return (<div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
      <div className="field grid flex">
        <label className="col-fixed align-self-start">{"Select export:"}</label>
        <div className="col">
          <Dropdown className="p-inputtext-sm"
            value={targetExport}
            options={exportOptions}
            onChange={(e) => {
              setExport(e.value);
              setExportDetails(props.exports.filter((n) => n.id === e.value)[0])
            }}
            placeholder={"Select export"}
            emptyMessage = "no exports found"  
            />
          <div className="error-help"><small className="p-error p-d-block">{exportError}</small> </div>
        </div>
      </div>

      <div className="field grid">
        <label className="col-fixed">{`Duration :`}</label>
        <div className="col">
          {exportDetails?.ex_params["duration"]}
        </div>
      </div>
      <div className="field grid">
        <label className="col-fixed">{`Exit code :`}</label>
        <div className="col">
          {exportDetails?.ex_params["exit_code"]}
        </div>
      </div>
      <div className="field grid">
        <label className="col-fixed">{`Export file name :`}</label>
        <div className="col">
          {exportDetails?.name}
        </div>
      </div>
      <div className="field grid">
        <label className="col-fixed">{`By :`}</label>
        <div className="col">
          {exportDetails?.created_by}
        </div>
      </div>
      <div className="field grid">
        <label className="col-fixed">{`Log :`}</label>
        {exportDetails?.run_log}
        <div className="col">
        </div>
      </div>
      {targetExport?.status === false && <ErrorTextBox error={"This export did not complete successfully"} />}
    </div>
    );
  }

  const renderCheckPage = () => {

    return (<div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">

      <div className="field grid">
        <label className="col-fixed">{`Checking dump file :`}</label>
        <div className="col multiline">
          {isLoading ? <i className="pi pi-spin pi-spinner" /> : dumpCheck}
        </div>
      </div>

      <div className="field grid">
        <label className="col-fixed">{`User exist check :`}</label>
        <div className="col multiline">
          {isLoading ? <i className="pi pi-spin pi-spinner" /> : userCheck}
        </div>
      </div>
      {errorMessage && <ErrorTextBox error = {errorMessage?.message ? errorMessage.message : errorMessage}/>}
      <ApiBar tooltip = {apiTooltip} onClick = {onRestApiClick} />
    </div>);

  }


  const handleOk = () => {

    props.onOkClick(
      {
        "export_id": targetExport?.name
      });

  }

  const renderFooter = () => {

    if (pageIndex === 1) {

      return (<div style={{ display: "flex", justifyContent: "space-between" }}>
        <Button label={"Close"} icon="pi pi-times" onClick={props.onHide} className="p-button-text" />
        <Button label={"Back"} 
        onClick={() => {
          dispatch(LayoutActions.setErrorMessage(""));
          setPageIndex(0);
        }} 
        className="p-button-secondary"
        icon="pi pi-arrow-left"
        iconPos="left"
         />
        <Button
          label={"Submit"}
          icon= {`pi pi-${isLoading ? 'spin pi-spinner' : 'check'}`}
          iconPos="right"
          onClick={handleOk} 
          autoFocus
          disabled = {!user_exists || user_exists?.detail || !dump_exists || dump_exists?.error}
          />
      </div>)

    }
    else if (pageIndex === 0) {

      return (<div style={{ display: "flex", justifyContent: "space-between" }}>
        <Button label={"Close"} icon="pi pi-times" onClick={props.onHide} className="p-button-text" />
        <Button
          label={"Next"}
          icon={`pi pi-${isLoading ? 'pi-spin pi-spinner' : 'arrow-right'}`}
          disabled = {!targetExport || targetExport?.status === false}
          iconPos="right"
          onClick={() => {
            const errors = [];

            let error = validator(constraints)("targetExport", targetExport);
            (error == null || errors.push(error)) && setExportError(error);

            if (errors.length === 0) {
              setDumpCheck("");
              setUserCheck("");
              dispatch(LayoutActions.checkDumpExists(
                {
                  "node_id": props.node.id,
                  "obj_type":  props.node.type.includes(clusterRoles.PDB) ? clusterRoles.PDB : clusterRoles.CDB,
                  "export_id": targetExport?.id,

                }));
              dispatch(LayoutActions.checkUserExists(
                {
                  "node_id": props.node.id,
                  "obj_type":  props.node.type.includes(clusterRoles.PDB) ? clusterRoles.PDB : clusterRoles.CDB,
                  "export_id": targetExport?.id,
                }));

              setPageIndex(1);

            }


          }} autoFocus />
      </div>)


    }

  }



  return (
    <Dialog
      className='import'
      header={`Import to ${props.node.type.includes(clusterRoles.PDB) ? props.node.name : props.node.location}`}
      visible={props.visible}
      style={{ width: '40vw'}}
      onHide={props.onHide}
      footer={renderFooter()}>
      <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
        {pageIndex === 0 && renderImportPage()}
        {pageIndex === 1 && renderCheckPage()}

      </div>

    </Dialog>)

};

export default ImportCDBPDBDialog;