import { Dialog } from "primereact/dialog";
import { Button } from "primereact/button";
import { Tag } from 'primereact/tag';
import { Dropdown } from "primereact/dropdown";
import { useState, useEffect } from "react";
import { InputText } from "primereact/inputtext";
import { RadioButton } from "primereact/radiobutton";
import { Checkbox } from "primereact/checkbox";
import { Calendar } from "primereact/calendar";
import moment from "moment";
import validator from "../../../util/validator";
import constraints from "./constraints";
import ApiBar from "../../../components/apiBar";
import copyToClipboard from "../../../util/copyToClipboard";
import createAPIcommand from "../../../util/createApiCommand";
import { useSelector } from "react-redux";
import ErrorTextBox from "../../../components/errorTextBox";
import { restoreOptions } from "../../../config";

export const RestorePDBDialog = (props) => {

  const [targetBackup, setTargetBackup] = useState(null);
  const [backupOptions, setBackupOptions] = useState([]);
  const [restoreConfirm, setRestoreConfirm] = useState("");
  const [targetBackupError, setTargetBackupError] = useState("");
  const [restoreConfirmError, setRestoreConfirmError] = useState("");
  const [apiTooltip, setApiTooltip] = useState(" ");
  const { accessToken } = useSelector((state) => state.app);
  const { errorMessage, isLoading } = useSelector((state) => state.layout);
  const [restoreTime, setRestoreTime] = useState(null);
  const [restoreRadio, setRestoreRadio] = useState(restoreOptions.THIS.value);
  const [autostart, setAutoStart] = useState(true);
  const [restoreTimeError, setRestoreTimeError] = useState("");

  const handleOk = () => {

    let errors = validate();

    if (errors.length === 0) {

      props.onOkClick(
        {
          "bu_id": targetBackup?.id,
          "autostart": autostart,
          "restore_option": restoreRadio,
          "restore_time": restoreTime ? moment(restoreTime).format("YYYY-MM-DD HH:mm:ss") : restoreTime
        });
    }
  }


  const validate = () => {

    let errors = [];

    let error = validator(constraints)("targetBackup", targetBackup);
    (error == null || errors.push(error)) && setTargetBackupError(error);

    error = validator(constraints)("restoreConfirm", restoreConfirm);
    (error == null || errors.push(error)) && setRestoreConfirmError(error);

    if (restoreRadio === restoreOptions.CUSTOM.value) {
      error = validator(constraints)("time", restoreTime);
      (error == null || errors.push(error)) && setRestoreTimeError(error);
    }
    else setRestoreTimeError("");

    return errors;

  }

  const onRestApiClick = async () => {


    let errors = validate();

    if (errors.length === 0) {

      setApiTooltip(await copyToClipboard(createAPIcommand({
        "bu_id": targetBackup?.name,
        "autostart": autostart,
        "restore_option": restoreRadio,
        "restore_time": restoreTime ? moment(restoreTime).format("YYYY-MM-DD HH:mm:ss") : restoreTime

      }, accessToken, "restore_backup_pdb/")));
    }
    else setApiTooltip(" ");


  }

  useEffect(() => {

    if (props.backups) {
      const temp = [];
      props.backups?.forEach((backup) => temp.push({ label: `${moment(backup.created_date).format("MMM DD,YYYY, hh:mm A")}   ${backup.display_size}`, value: backup }));
      setBackupOptions(temp)

    }

  }, [props.backups]);

  const renderFooter = () => {
    return (
      <div style={{ display: "flex", justifyContent: "space-between" }}>
        <Button label={"Close"} icon="pi pi-times" onClick={props.onHide} className="p-button-text" />
        <Button
          label={"Submit"}
          icon= {`pi pi-${isLoading ? 'spin pi-spinner' : 'check'}`}
          disabled={!targetBackup || targetBackup?.status === false}
          iconPos="right"
          onClick={handleOk} autoFocus />
      </div>
    );
  }

  return (
    <Dialog
      className='restore-pdb'
      header={`Restore PDB`}
      visible={props.visible}
      style={{ width: '50vw' }}
      onHide={props.onHide}
      footer={renderFooter()}>
      <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
        <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
          <div className="field grid">
            <div className="col">
              <div className="title">{"Choose backup to restore:"}</div>

              <div className="field">
                <Dropdown className="p-inputtext-sm"
                  value={targetBackup}
                  options={backupOptions}
                  onChange={(e) => setTargetBackup(e.value)}
                  placeholder="Select backup"
                  emptyMessage = "no backups found" 
                  />
                <div className="error-help">{<small className="p-error p-d-block">{targetBackupError}</small>}</div>
              </div>

              <div className="field grid align-items-center">
                <div className="field grid">
                  <div className="col">
                    {
                      Object.keys(restoreOptions).map(key => (<div key={key} className="field-radiobutton">
                        <RadioButton inputId={key} name="restoreRadio" value={restoreOptions[key].value} checked={restoreRadio === restoreOptions[key].value} onChange={(e) => setRestoreRadio(e.value)} />
                        <label htmlFor={key}>{restoreOptions[key].label}</label>
                      </div>))
                    }
                  </div>
                </div>
                <div className="col-fixed">Select time:</div>
                <div className="col-6">
                  <Calendar
                    showSeconds
                    maxDate={new Date()}
                    showTime
                    hourFormat="24"
                    disabled={restoreRadio !== restoreOptions.CUSTOM.value}
                    value={restoreTime}
                    onChange={(e) => setRestoreTime(e.value)} />
                  <div className="error-help">{<small className="p-error p-d-block">{restoreTimeError}</small>}</div>
                </div>
              </div>

              <div className="field grid align-items-center">
                <div className="col-fixed"> Start up pdb after restore:</div>
                <div className="col">

                  <Checkbox checked={autostart} onChange={(e) => setAutoStart(e.checked)} />

                </div>
              </div>

              <div className="field">
                <Tag icon="pi pi-exclamation-triangle" className="field" severity="warning" value="Once a pluggable database is restored previous data will be overwritten.This operation is not reversible." />
              </div>
              <div className="field grid flex align-items-center">
                <div className="col">
                  <InputText
                    className="p-inputtext-sm"
                    placeholder='Please type "restore me" here in order to continue'
                    value={restoreConfirm}
                    onChange={(e) => setRestoreConfirm(e.target.value)} />
                  <div className="error-help">{<small className="p-error p-d-block">{restoreConfirmError}</small>}</div>
                </div>
              </div>
            </div>

          </div>

          {(errorMessage || targetBackup?.status === false) && <ErrorTextBox error={errorMessage ? errorMessage?.message ? errorMessage.message : errorMessage : "This backup did not complete successfully"} />}
          <ApiBar tooltip={apiTooltip} onClick={onRestApiClick} />
        </div>
      </div>

    </Dialog>)

};

export default RestorePDBDialog;