import { Dialog } from "primereact/dialog";
import { Button } from "primereact/button";
import { Checkbox } from "primereact/checkbox";
import { Calendar } from "primereact/calendar";
import { RadioButton } from "primereact/radiobutton";
import { useState } from "react";
import { InputText } from "primereact/inputtext";
import { restoreOptions } from "../../../config";
import ApiBar from "../../../components/apiBar";
import createAPIcommand from "../../../util/createApiCommand";
import copyToClipboard from "../../../util/copyToClipboard";
import { useSelector } from "react-redux";
import moment from "moment";
import { Tag } from "primereact/tag";
import ErrorTextBox from "../../../components/errorTextBox";
import config from "../../../config";
import validator from "../../../util/validator";
import constraints from "./constraints";

const RestorePDBLayoutDialog = (props) => {

  const [restoreConfirm, setRestoreConfirm] = useState("");
  const [apiTooltip, setApiTooltip] = useState(" ");
  const [restoreConfirmError, setRestoreConfirmError] = useState("");
  const { accessToken } = useSelector((state) => state.app);
  const { errorMessage, isLoading } = useSelector((state) => state.layout);
  const [restoreTime, setRestoreTime] = useState(null);
  const [restoreRadio, setRestoreRadio] = useState(restoreOptions.THIS.value);
  const [autostart, setAutoStart] = useState(true);
  const [restoreTimeError, setRestoreTimeError] = useState("");
  const onRestApiClick = async () => {
    let errors = validate();

    if (errors.length === 0) {
      setApiTooltip(await copyToClipboard(createAPIcommand({
        "bu_id": props.backup.name,
        "autostart": autostart,
        "restore_option": restoreRadio,
        "restore_time": restoreTime ? moment(restoreTime).format("YYYY-MM-DD HH:mm:ss") : restoreTime
      }, accessToken, "restore_backup_pdb/")));
    }
    else setApiTooltip(" ");
  }

  const handleOk = () => {

    let errors = validate();
    if (errors.length === 0) {

      props.onOkClick(
        {
          "bu_id": props.backup.id,
          "autostart": autostart,
          "restore_option": restoreRadio,
          "restore_time": restoreTime ? moment(restoreTime).format("YYYY-MM-DD HH:mm:ss") : restoreTime
        }
      );

    }
  }

  const validate = () => {

    let errors = [];

    let error = validator(constraints)("restoreConfirm", restoreConfirm);
    (error == null || errors.push(error)) && setRestoreConfirmError(error);
    if (restoreRadio === restoreOptions.CUSTOM.value) {
      error = validator(constraints)("time", restoreTime);
      (error == null || errors.push(error)) && setRestoreTimeError(error);
    }
    else setRestoreTimeError("");

    return errors;

  }

  const renderFooter = () => {
    return (
      <div style={{ display: "flex", justifyContent: "space-between" }}>
        <Button label={"Close"} icon="pi pi-times" onClick={props.onHide} className="p-button-text" />
        <Button
          label={"Submit"}
          icon= {`pi pi-${isLoading ? 'spin pi-spinner' : 'check'}`}
          iconPos="right"
          onClick={handleOk} autoFocus />
      </div>
    );
  }

  return (
    <Dialog
      className='restore-pdb-backup'
      header={`Restore PDB backup`}
      visible={props.visible}
      style={{ width: '40vw' }}
      onHide={props.onHide}
      footer={renderFooter()}>
      <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
        <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
          <div className="field grid">
            <div className="col">
              <div className="field">
                This will restore the pluggable database {props.backup.node_name} to the backup, taken at {moment(props.backup.created_date).format("MMM DD,YYYY, hh:mm A")}
              </div>


              <div className="field grid">
                <div className="col">
                  {
                    Object.keys(restoreOptions).map(key => (<div key={key} className="field-radiobutton">
                      <RadioButton inputId={key} name="restoreRadio" value={restoreOptions[key].value} checked={restoreRadio === restoreOptions[key].value} onChange={(e) => setRestoreRadio(e.value)} />
                      <label htmlFor={key}>{restoreOptions[key].label}</label>
                    </div>))
                  }
                </div>
              </div>

              <div className="field grid align-items-center">
                <div className="col-fixed">Select time:</div>
                <div className="col-6">
                  <Calendar
                    showSeconds
                    maxDate={new Date()}
                    showTime
                    hourFormat="24"
                    disabled={restoreRadio !== restoreOptions.CUSTOM.value}
                    value={restoreTime}
                    onChange={(e) => setRestoreTime(e.value)} />
                  <div className="error-help">{<small className="p-error p-d-block">{restoreTimeError}</small>}</div>
                </div>
              </div>

              <div className="field grid align-items-center">
                <div className="col-fixed"> Start up pdb after restore:</div>
                <div className="col">

                  <Checkbox checked={autostart} onChange={(e) => setAutoStart(e.checked)} />

                </div>
              </div>

              <div className="field">
                <Tag value="The current content of the database will be overridden, this operation is not reversible!" icon="pi pi-exclamation-triangle" severity="warning" />
              </div>

              <div className="field grid flex align-items-center">
                <div className="col">
                  <InputText
                    className="p-inputtext-sm"
                    placeholder={`Please type "${config.restoreConfirm}" here in order to continue`}
                    value={restoreConfirm}
                    onChange={(e) => setRestoreConfirm(e.target.value)} />
                  <div className="error-help">{<small className="p-error p-d-block">{restoreConfirmError}</small>}</div>
                </div>
              </div>
            </div>
          </div>

          {errorMessage && <ErrorTextBox error={errorMessage?.message ? errorMessage.message : errorMessage} />}
          <ApiBar tooltip={apiTooltip} onClick={onRestApiClick} />
        </div>
      </div>
    </Dialog>)
};

export default RestorePDBLayoutDialog;