import { useState } from 'react';
import { InputText } from "primereact/inputtext";
import { Button } from "primereact/button";
import { Dialog } from "primereact/dialog";
import { useSelector, useDispatch } from 'react-redux';
import copyToClipboard from '../../../util/copyToClipboard';
import createAPIcommand from '../../../util/createApiCommand';
import ApiBar from '../../../components/apiBar';
import ErrorTextBox from '../../../components/errorTextBox';
import config, { clusterRoles, types } from '../../../config';
import validator from '../../../util/validator';
import constraints from './constraints';
import { Checkbox } from 'primereact/checkbox';
import { LayoutActions } from '../../../store/layout/reducer';
import { ConfirmPopup } from 'primereact/confirmpopup';

const PostgresNodeDeleteDialog = (props) => {

    const [dropConfirm, setDropConfirm] = useState("");
    const { accessToken } = useSelector((state) => state.app);
    const { errorMessage, isLoading } = useSelector((state => state.layout));
    const [apiTooltip, setApiTooltip] = useState(" ");
    const [dropConfirmError, setDropConfirmError] = useState("");
    const [finalBackup, setFinalBackup] = useState(true);
    const [showForceDeleteConfirm, setShowForceDeleteConfirm] = useState(false)
    const dispatch = useDispatch();

    const handleOk = () => props.onOkClick({
        [`${props.node.type === types.POSTGRES_K8S ? 'pod' : 'node'}_id`]: props.node.id,
        "type":  props.node.type,
        "final_backup": props.node.cluster_role === clusterRoles.PRIMARY ? finalBackup : false
    });

    const onDeletePgNodeClick = () => {
        let errors = validate();
        if (errors.length === 0) {
            if(props.node.type === types.POSTGRES_K8S) handleOk();
            else dispatch(LayoutActions.isNodeHostReachable({ 'node_id': props.node.id }, { allow: handleOk, confirm: setShowForceDeleteConfirm }));
        }
    }


    const validate = () => {

        let errors = [];

        let error = validator(constraints)("dropConfirm", dropConfirm);
        (error == null || errors.push(error)) && setDropConfirmError(error);
        return errors;

    }



    const onRestApiClick = async () => {

        let errors = validate();

        if (errors.length === 0) {

            setApiTooltip(await copyToClipboard(createAPIcommand(
                {

                    [`${props.node.type === types.POSTGRES_K8S ? 'pod' : 'node'}_id`]: props.node.id,
                    "final_backup": props.node.cluster_role === clusterRoles.PRIMARY ? finalBackup : false


                }
                , accessToken, props.node.type === types.POSTGRES_K8S ? 'drop_pg_pod/' : 'drop_pg_node/')));


        }
        else setApiTooltip(" ");

    }


    const renderFooter = () => {
        return (
            <div className="flex justify-content-between">
                <Button label={"Close"} icon="pi pi-times" onClick={props.onHide} className="p-button-text" />
                <Button
                    label={"Submit"}
                    id='submit-button'
                    icon= {`pi pi-${isLoading ? 'spin pi-spinner' : 'trash'}` } 
                    className="p-button-danger"
                    onClick={onDeletePgNodeClick} autoFocus />
                <ConfirmPopup
                    icon="pi pi-exclamation-triangle"
                    target={document.getElementById('submit-button')}
                    visible={showForceDeleteConfirm}
                    className="multiline"
                    onHide={() => setShowForceDeleteConfirm(false)}
                    accept={() => dispatch(LayoutActions.markNodeDeleted({ 'node_id': props.node.id }, props.onHide()))}
                    message={"Node host is currently offline, unable to perform drop.\nMark node as deleted anyway?"} />
            </div>
        );
    }


    return (
        <Dialog
            className='delete-pg-node'
            header={`Drop ${props.node.cluster_role.charAt(0)}${props.node.cluster_role.substring(1).toLowerCase()}`}
            visible={props.visible}
            style={{ width: '40vw' }}
            onHide={props.onHide}
            footer={renderFooter()}>
            <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
                <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">

                    <div className='form-inner-part border-none'>

                        <div style={{ marginBottom: "10px" }}>
                            This will drop {props.node?.name}. The {props.node.cluster_role.toLowerCase()} will no longer be manage by 123cluster.
                        </div>
                        <div>
                            Are you sure?
                        </div>
                        {props.node.cluster_role === clusterRoles.PRIMARY && (!props.node.nodes || props.node.nodes.length === 0) &&
                            <div>
                                Please consider to backup this node.
                            </div>}

                    </div>

                    <div className='form-inner-part border-none'>
                        <div className="field">
                            <div style={{ marginBottom: "10px" }} className="title">{`Please type "${config.dropConfirm}" here in order to continue`}</div>
                            <InputText className="p-inputtext-sm" placeholder={`Please type "${config.dropConfirm}" here in order to continue`} value={dropConfirm} onChange={(e) => setDropConfirm(e.target.value)} />
                            <div className="error-help">{<small className="p-error p-d-block">{dropConfirmError}</small>}</div>
                                 {props.node.cluster_role === clusterRoles.PRIMARY ? <div className='flex'>

                                <div className='ml-auto'>
                                    <Checkbox inputId='final-backup' className='mr-2' checked={finalBackup} value={finalBackup} onChange={() => setFinalBackup(!finalBackup)} />
                                    <label htmlFor='final-backup' className="p-checkbox-label" >run final backup on drop</label>
                                </div>

                            </div>  : null}

                        </div>
                    </div>

                    {errorMessage && <ErrorTextBox error={errorMessage?.message ? errorMessage.message : errorMessage} />}
                    <ApiBar tooltip={apiTooltip} onClick={onRestApiClick} />
                </div>
            </div>
        </Dialog>
    );
}

export default PostgresNodeDeleteDialog;