import { Dialog } from "primereact/dialog";
import { Button } from "primereact/button";
import { useState } from "react";
import { InputText } from "primereact/inputtext";
import ApiBar from "../../../components/apiBar";
import createAPIcommand from "../../../util/createApiCommand";
import copyToClipboard from "../../../util/copyToClipboard";
import { useSelector } from "react-redux";
import moment from "moment";
import { Tag } from "primereact/tag";
import { RadioButton } from "primereact/radiobutton";
import { Calendar } from "primereact/calendar";
import ErrorTextBox from "../../../components/errorTextBox";
import { Checkbox } from "primereact/checkbox";
import validator from "../../../util/validator";
import constraints from "./constraints";
import config, { clusterRoles, restoreOptions, types } from "../../../config";

const RestorePgNodeLayoutDialog = (props) => {

  const [restoreConfirm, setRestoreConfirm] = useState("");


  const [restoreRadio, setRestoreRadio] = useState(restoreOptions.THIS.value);

  const [restoreTime, setRestoreTime] = useState(null);
  const [restoreTimeError, setRestoreTimeError] = useState("");
  const [restoreConfirmError, setRestoreConfirmError] = useState("");
  const [autostart, setAutoStart] = useState(true);

  const [apiTooltip, setApiTooltip] = useState(" ");
  const { accessToken } = useSelector((state) => state.app);
  const { errorMessage, isLoading } = useSelector((state) => state.layout);


  const validate = () => {

    const errors = [];

    let error = validator(constraints)("restoreConfirm", restoreConfirm);
    (error == null || errors.push(error)) && setRestoreConfirmError(error);

    if (restoreRadio === restoreOptions.CUSTOM.value) {
      error = validator(constraints)("time", restoreTime);
      (error == null || errors.push(error)) && setRestoreTimeError(error);
    }
    else setRestoreTimeError("");

    return errors;

  }


  const onRestApiClick = async () => {

    let errors = validate();

    if (errors.length === 0) {

      setApiTooltip(await copyToClipboard(createAPIcommand({
        "bu_id": props.backup.name,
        "autostart": autostart,
        "restore_option": restoreRadio,
        "type": props.backup?.ex_params?.belong_to,
        "restore_time": restoreTime ? moment(restoreTime).format("YYYY-MM-DD HH:mm:ss") : restoreTime

      }, accessToken, props.backup?.ex_params?.belong_to === types.POSTGRES_K8S ? "restore_pg_pod/" : "restore_pg_node/")));

    }
    else setApiTooltip(" ");

  }

  const handleOk = () => {

    let errors = validate();

    if (errors.length === 0) {

      props.onOkClick(

        {
          "bu_id": props.backup.id,
          "autostart": autostart,
          "restore_option": restoreRadio,
          "restore_time": restoreTime ? moment(restoreTime).format("YYYY-MM-DD HH:mm:ss") : restoreTime,
          "type": props.backup?.ex_params?.belong_to
        }

      );

    }
  }

  const renderFooter = () => {
    return (
      <div style={{ display: "flex", justifyContent: "space-between" }}>
        <Button label={"Close"} icon="pi pi-times" onClick={props.onHide} className="p-button-text" />
        <Button
          label={"Submit"}
          icon= {`pi pi-${isLoading ? 'spin pi-spinner' : 'check'}`}
          iconPos="right"
          onClick={handleOk} autoFocus />
      </div>
    );

  }



  return (
    <Dialog
      className='restore-pg-node-backup'
      header={`Restore PostgreSQL node backup`}
      visible={props.visible}
      style={{ width: '40vw' }}
      onHide={props.onHide}
      footer={renderFooter()}>
      <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
        <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
          <div className="field grid">
            <div className="col">
              <div className="field">
                This will restore the node to the backup of {props.backup?.node_name}, taken at {moment(props.backup?.created_date).format("MMM DD,YYYY, hh:mm A")}
              </div>

              <div className="field grid">
                <div className="col">
                  {
                    Object.keys(restoreOptions).map(key => (<div key={key} className="field-radiobutton">
                      <RadioButton inputId={key} name="restoreRadio" value={restoreOptions[key].value} checked={restoreRadio === restoreOptions[key].value} onChange={(e) => setRestoreRadio(e.value)} />
                      <label htmlFor={key}>{restoreOptions[key].label}</label>
                    </div>))
                  }

                </div>
              </div>

              <div className="field grid align-items-center">
                <div className="col-fixed">Select time:</div>
                <div className="col-6">
                  <Calendar
                    showSeconds
                    maxDate={new Date()}
                    showTime
                    hourFormat="24"
                    disabled={restoreRadio !== restoreOptions.CUSTOM.value}
                    value={restoreTime}
                    onChange={(e) => setRestoreTime(e.value)} />
                  <div className="error-help">{<small className="p-error p-d-block">{restoreTimeError}</small>}</div>
                </div>
              </div>

              <div className="field grid align-items-center">
                <div className="col-fixed"> Start up instance after restore:</div>
                <div className="col">

                  <Checkbox checked={autostart} disabled={props.node?.cluster_role === clusterRoles.PRIMARY && props.node?.nodes?.filter((s) => s.status === 'green').length > 0} onChange={(e) => setAutoStart(e.checked)} />

                </div>
              </div>

              <div className="field">
                <Tag value="The current content of the node will be overridden, this operation is not reversible!" icon="pi pi-exclamation-triangle" severity="warning" />
              </div>

              <div className="field grid align-items-center">
                <div className="col">

                  <InputText
                    className="p-inputtext-sm"
                    placeholder={`Please type "${config.restoreConfirm}" here in order to continue`}
                    value={restoreConfirm}
                    onChange={(e) => setRestoreConfirm(e.target.value)} />

                  <div className="error-help">{<small className="p-error p-d-block">{restoreConfirmError}</small>}</div>
                </div>
              </div>



            </div>
          </div>
          {errorMessage && <ErrorTextBox error={errorMessage?.message ? errorMessage.message : errorMessage} />}
          <ApiBar tooltip={apiTooltip} onClick={onRestApiClick} />

        </div>
      </div>

    </Dialog>)

};

export default RestorePgNodeLayoutDialog;