import { useState } from 'react';
import { InputText } from "primereact/inputtext";
import { Button } from "primereact/button";
import { Dialog } from "primereact/dialog";
import { useDispatch, useSelector } from 'react-redux';
import copyToClipboard from '../../../util/copyToClipboard';
import createAPIcommand from '../../../util/createApiCommand';
import ApiBar from '../../../components/apiBar';
import ErrorTextBox from '../../../components/errorTextBox';
import config, { clusterRoles } from '../../../config';
import validator from '../../../util/validator';
import { Checkbox } from "primereact/checkbox";
import constraints from './constraints';
import { LayoutActions } from '../../../store/layout/reducer';
import { ConfirmPopup } from 'primereact/confirmpopup';

const DropRedisNodeDialog = (props) => {

    const [dropConfirm, setDropConfirm] = useState("");
    const { accessToken } = useSelector((state) => state.app);
    const { errorMessage, isLoading } = useSelector((state => state.layout));
    const [apiTooltip, setApiTooltip] = useState(" ");
    const [dataOnly, setDataOnly] = useState(false);
    const [dropConfirmError, setDropConfirmError] = useState("");
    const [finalBackup, setFinalBackup] = useState(true);
    const [showForceDeleteConfirm, setShowForceDeleteConfirm] = useState(false);
    const dispatch = useDispatch();

    const handleOk = () => props.onOkClick({
        "node_id": props.node.id,
        "data_only": dataOnly,
        "final_backup": props.node.cluster_role === clusterRoles.PRIMARY ? finalBackup : false
    });

    const onDeleteRedisNodeClick = () => {
        let errors = validate();
        if (errors.length === 0) dispatch(LayoutActions.isNodeHostReachable({ 'node_id': props.node.id }, { allow: handleOk, confirm: setShowForceDeleteConfirm }));
    }


    const validate = () => {

        let errors = [];

        let error = validator(constraints)("dropConfirm", dropConfirm);
        (error == null || errors.push(error)) && setDropConfirmError(error);
        return errors;

    }



    const onRestApiClick = async () => {

        let errors = validate();

        if (errors.length === 0) {

            setApiTooltip(await copyToClipboard(createAPIcommand(
                {

                    "node_id": props.node.name,
                    "data_only": dataOnly,
                    "final_backup": props.node.cluster_role === clusterRoles.PRIMARY ? finalBackup : false


                }
                , accessToken, 'drop_redis_node/')));


        }
        else setApiTooltip(" ");

    }


    const renderFooter = () => {
        return (
            <div className="flex justify-content-between">
                <Button label={"Close"} icon="pi pi-times" onClick={props.onHide} className="p-button-text" />
                <Button
                    label={"Submit"}
                    id='submit-button'
                    icon= {`pi pi-${isLoading ? 'spin pi-spinner' : 'trash'}`}
                    className="p-button-danger"
                    onClick={onDeleteRedisNodeClick} autoFocus />
                <ConfirmPopup
                    icon="pi pi-exclamation-triangle"
                    target={document.getElementById('submit-button')}
                    visible={showForceDeleteConfirm}
                    className="multiline"
                    onHide={() => setShowForceDeleteConfirm(false)}
                    accept={() => dispatch(LayoutActions.markNodeDeleted({ 'node_id': props.node.id }, props.onHide()))}
                    message={"Node host is currently offline, unable to perform drop.\nMark node as deleted anyway?"} />
            </div>
        );
    }


    return (
        <Dialog
            className='delete-redis-node'
            header={props.node.cluster_role ? `Drop ${props.node.cluster_role.charAt(0)}${props.node.cluster_role.substring(1).toLowerCase()}` : "Drop Cluster"}
            visible={props.visible}
            style={{ width: '40vw' }}
            onHide={props.onHide}
            footer={renderFooter()}>
            <div style={{ display: "flex", paddingBottom: "10px" }} className="form-container">
                <div style={{ flex: 1, padding: "10px" }} className="p-fluid form-container">
                    <div className='form-inner-part border-none'>


                        <div style={{ marginBottom: "10px" }}>
                            This will drop {props.node?.name}. The {props.node.cluster_role ? props.node.cluster_role.toLowerCase() : "cluster"} will no longer be manage by 123cluster.
                        </div>
                        <div className='field'>
                            Are you sure?
                        </div>
                        {props.node.cluster_role === clusterRoles.PRIMARY && !(props.slaves_count || props.slaves_count < 1) ?
                            <div className='field'>
                                Please consider to backup this node.
                            </div> : null}

                        {props.node?.cluster_role !== clusterRoles.SENTINEL ?
                            <div className="field">
                                <div style={{ marginBottom: "10px" }} className="col-fixed">{'Drop data only'}</div>
                                <div className='col'>
                                    <Checkbox checked={dataOnly} value={dataOnly} onChange={(e) => setDataOnly(!dataOnly)} />
                                </div>

                            </div> : null}

                    </div>

                    <div className='form-inner-part border-none'>

                        <div className="field">
                            <div style={{ marginBottom: "10px" }} className="title">{`Please type "${config.dropConfirm}" here in order to continue`}</div>
                            <InputText className="p-inputtext-sm" placeholder={`Please type "${config.dropConfirm}" here in order to continue`} value={dropConfirm} onChange={(e) => setDropConfirm(e.target.value)} />
                            <div className="error-help">{<small className="p-error p-d-block">{dropConfirmError}</small>}</div>
                                 {props.node.cluster_role === clusterRoles.PRIMARY ? <div className='flex'>

                                <div className='ml-auto'>
                                    <Checkbox inputId='final-backup' className='mr-2' checked={finalBackup} value={finalBackup} onChange={() => setFinalBackup(!finalBackup)} />
                                    <label htmlFor='final-backup' className="p-checkbox-label" >run final backup on drop</label>
                                </div>

                            </div>  : null}

                        </div>

                    </div>

                    {errorMessage && <ErrorTextBox error={errorMessage?.message ? errorMessage.message : errorMessage} />}
                    <ApiBar tooltip={apiTooltip} onClick={onRestApiClick} />

                </div>
            </div>
        </Dialog>
    );
}


export default DropRedisNodeDialog;