import { useState, useEffect } from "react";
import Menu from "../../components/menus/menu";
import { Button } from "primereact/button";
import { DataTable } from "primereact/datatable";
import { Column } from "primereact/column";
import { ConfirmPopup } from "primereact/confirmpopup";
import SelectDashboardTypeDialog from "../../dialogs/selectDashboardTypeDialog";
import { useSelector, useDispatch } from "react-redux";
import { MonitoringActions } from "../../store/monitoring/reducer";
import { Panel } from "primereact/panel";
import AgentsChart from "../../components/agentsChart"
import PieChart from "../../components/pieChart"
import useInterval from "../../util/useInterval";
import DashboardFilter from "../../components/filters/dashboardFilter";
import { AppActions } from "../../store/app/reducer";
import { types } from "../../config";
import notify from "../../util/notify";

const Dashboard = () => {

    const [showSelectDashboardTypeDialog, setShowSelectDashboardTypeDialog] = useState(false);
    const [dashboardToDelete, setDashboardToDelete] = useState(null);
    const [popupElement, setPopupElement] = useState(null);
    const [showDeleteConfirm, setShowDeleteConfirm] = useState(null);
    const dispatch = useDispatch();
    const { dashboard_monitors } = useSelector((state) => state.monitoring);
    const { db_type, time_period } = useSelector((state) => state.app);
    const [ timePeriod, setTimePeriod ] = useState(time_period);
    const [ filteredMonitors, setFilteredMonitors] = useState([]);
    const [search, setSearch] = useState("");
    const [dbType, setDbType] = useState(db_type);
    const [draggedMonitorIndex, setDraggedMonitorIndex] = useState(null);
    const isDragAndDropEnabled = !search && dbType === types.ALL;

    const searchFilter = (data, search) => {
      return data?.filter((n) =>
          ["column_display_name", "node_type", "monitor_name", "node_name"]
              .some((key) => n[key]?.toLowerCase()?.includes(search))
      );
  };

    const monitorFilter = (data, dbType) => {
        return dbType === types.ALL ? data : data?.filter ? data?.filter((n) => n.node_type?.includes(dbType)) : [];
    }

    useEffect(() => {
        if (dashboard_monitors?.length > 0) {
            let tempData = [...dashboard_monitors];
            if (search) tempData = searchFilter(tempData, search);
            if (dbType) tempData = monitorFilter(tempData, dbType);

            setFilteredMonitors(tempData);
        }
        else {
            dispatch(MonitoringActions.getDashboardMonitors(timePeriod));
        }
    }, [dashboard_monitors, dbType, timePeriod, search, dispatch]);

    useInterval(() => {
        dispatch(MonitoringActions.refreshDashboard());
    }, 60000);

    const onAddDashboardMonitor = (data) => {
        data["time_period"] = timePeriod;
        setShowSelectDashboardTypeDialog(false);
        dispatch(MonitoringActions.addDashboardMonitor(data));
  };

    const getDeleteIcon = (monitor_id) => {
        return (<i id={`delete_${monitor_id}`} className="pi pi-fw pi-times" onClick={() => {
            setDashboardToDelete({ monitor_id: monitor_id });
            setPopupElement(window.document.getElementById(`delete_${monitor_id}`));
            setShowDeleteConfirm(true);
        }} />);
    }

    const handleDragStart = (e, index) => {
      if (isDragAndDropEnabled) {
          setDraggedMonitorIndex(index);
      } else {
          e.preventDefault();
          notify.warning("Drag and Drop is disabled when filters are applied. Please remove filters to reorder monitors.")
      }
    };

    const handleDragOver = (e) => {
        e.preventDefault();
    };

    const handleDrop = (e, dropIndex) => {
        e.preventDefault();
        if (draggedMonitorIndex === dropIndex) return;

        const draggedMonitor = filteredMonitors[draggedMonitorIndex];
        const newOrderId = dropIndex + 1;

        const updatedOrder = [...filteredMonitors];
        const [draggedItem] = updatedOrder.splice(draggedMonitorIndex, 1);
        updatedOrder.splice(dropIndex, 0, draggedItem);

        setFilteredMonitors(updatedOrder);
        setDraggedMonitorIndex(null);

        dispatch(MonitoringActions.reorderDashboardMonitors({ monitor_id: draggedMonitor.monitor_id, order_id: newOrderId}));
    };

    const getTableHeader = (node, monitor_name, node_type) => {
        return (
            <div className="flex flex-column flex-shrink-0 justify-content-start">
                {`${node} ${monitor_name} ${node_type}`}
            </div>
        );
    };

    const onFilterMonitorChange = (dbType, time_period) => {
        setDbType(dbType);
        dispatch(AppActions.setDbType(dbType));
        dispatch(MonitoringActions.refreshDashboard({time_period: time_period, dbType: dbType}));
    };

    const onSearchChange = (search) => {
        setSearch(search);
    };

    const onTimePeriodChange = (timePeriod, dbType) => {
        dispatch(AppActions.setTimePeriod(timePeriod));
        dispatch(MonitoringActions.refreshDashboard({time_period: time_period, dbType: dbType}));
        setTimePeriod(timePeriod);
    };

    return (
      <div className="dashboard-page">
          <Menu />
          <div className="status-page">
              <DashboardFilter
                  onMonitorChange={onFilterMonitorChange}
                  onSearchChange={onSearchChange}
                  onTimePeriodChange={onTimePeriodChange}
              />
          </div>
          <div className="flex mb-4">
              <Button
                  className="p-button-rounded p-button-info ml-auto mr-auto"
                  icon="pi pi-plus"
                  onClick={() => setShowSelectDashboardTypeDialog(true)}
              />
          </div>
          <div>
        <div className="monitors-container">
          {filteredMonitors.map((item, index) => (
            <div
                key={item?.monitor_id}
                className="dashboard-chart-wrapper"
                draggable={isDragAndDropEnabled}
                onDragStart={(e) => handleDragStart(e, index)}
                onDragOver={handleDragOver}
                onDrop={(e) => handleDrop(e, index)}
            >
                {(() => {
                    switch (item.visualization_type) {
                        case "line_chart":
                            return (
                                <Panel
                                    className="dashboard-chart"
                                    icons={() => getDeleteIcon(item?.monitor_id)}
                                    header={
                                        <div className="multiline">
                                            {`${item?.node_name} — ${item?.monitor_name} — ${item?.node_type}`}
                                        </div>
                                    }
                                >
                                    <AgentsChart
                                        dateFormat="HH:mm"
                                        timePeriodType={timePeriod}
                                        points={item?.points}
                                        label={item.points?.[0]?.["column1"]}
                                        xAxisKey="dc"
                                        yAxisKey="col_name"
                                        dateKey="dc"
                                    />
                                </Panel>
                            );
                        case "pie_chart":
                            return (
                                <Panel
                                    className="dashboard-chart"
                                    icons={() => getDeleteIcon(item?.monitor_id)}
                                    header={
                                        <div className="multiline">
                                            {`${item?.node_name} — ${item?.monitor_name} — ${item?.node_type}`}
                                        </div>
                                    }
                                >
                                    <PieChart
                                        data={item?.points}
                                        labelKey="category"
                                        valueKey="value"
                                    />
                                </Panel>
                            );
                        case "monitor_status":
                            return (
                                <Panel
                                    className="dashboard-chart"
                                    icons={() => getDeleteIcon(item?.monitor_id)}
                                    header={
                                        <div className="multiline">
                                            {`${item?.node_name} — ${item?.monitor_name} — ${item?.node_type}`}
                                        </div>
                                    }
                                >
                                    {(() => {
                                        let severityText = '';
                                        let severityColor = '';

                                        switch (item?.severity) {
                                            case 1:
                                                severityText = 'GREEN';
                                                severityColor = 'green';
                                                break;
                                            case 2:
                                                severityText = 'ORANGE';
                                                severityColor = 'orange';
                                                break;
                                            case 3:
                                                severityText = 'RED';
                                                severityColor = 'red';
                                                break;
                                            default:
                                                severityText = 'UNKNOWN';
                                                severityColor = 'gray';
                                        }
                                        return <span style={{ color: severityColor }}>{severityText}</span>;
                                    })()}
                                          </Panel>
                                      );
                          default:
                              return (
                                  <Panel
                                      className="dashboard-chart"
                                      icons={() => getDeleteIcon(item?.monitor_id)}
                                      header={
                                          <div className="multiline">
                                              {`${item?.node_name} — ${item?.monitor_name} — ${item?.node_type}`}
                                          </div>
                                      }
                                  >
                                      <DataTable
                                          rows={10}
                                          paginator
                                          alwaysShowPaginator={false}
                                          rowClassName={(rowData) =>
                                              `severity-index-${rowData["color"]?.toLowerCase()}-row`
                                          }
                                          value={item?.alert?.details}
                                      >
                                          {item?.alert?.details?.[0] &&
                                              Object.keys(item?.alert?.details?.[0]).map((b, i) =>
                                                  b === "color" ? (
                                                      <Column key={i} />
                                                  ) : b === "message" ? (
                                                      <Column
                                                          key={i}
                                                          sortable
                                                          field={b}
                                                          header={b}
                                                          body={(d) => d.message?.replace(/\n\s*\n/g, "\n")}
                                                      />
                                                  ) : (
                                                      <Column key={i} field={b} header={b} sortable />
                                                  )
                                              )}
                                      </DataTable>
                                  </Panel>
                              );
                            }
                          })()}
                      </div>
                  ))}
              </div>
          </div>
          {showSelectDashboardTypeDialog && (
              <SelectDashboardTypeDialog
                  visible={showSelectDashboardTypeDialog}
                  onHide={() => setShowSelectDashboardTypeDialog(false)}
                  onOkClick={onAddDashboardMonitor}
              />
          )}
          <ConfirmPopup
              onHide={() => setShowDeleteConfirm(false)}
              visible={showDeleteConfirm}
              target={popupElement}
              acceptClassName="p-button-danger"
              rejectIcon="pi pi-times"
              message="This will delete dashboard"
              icon="pi pi-exclamation-triangle"
              acceptIcon="pi pi-trash"
              acceptLabel="Confirm"
              rejectLabel="Cancel"
              reject={() => setShowDeleteConfirm(false)}
              accept={() =>
                  dispatch(MonitoringActions.deleteDashboardMonitor(dashboardToDelete))
              }
        />
    </div>
  );
};

export default Dashboard;
