import { useCallback, useEffect, useState } from "react";
import { useDispatch, useSelector } from "react-redux";
import moment from "moment";
import { LogActions } from "../../store/log/reducer";
import { AppActions } from "../../store/app/reducer";
import Menu from "../../components/menus/menu";
import LogFilter from "../../components/filters/logFilter";
import { types } from "../../config";
import LogsTable from "../../components/tables/logsTable";

export const searchFilter = (data, search) => {
    return data?.filter ? data?.filter((n) => {
        if (n.what?.toLowerCase().includes(search) || n.target_type?.toLowerCase().includes(search) || n.status?.toLowerCase().includes(search) || (n.duration + "").includes(search) ||
            moment(n.when).format("MMM DD,YYYY, hh:mm A").toLowerCase().includes(search))
            return true;
        return false;
    }) : [];
}

const Log = () => {
    const dispatch = useDispatch();
    const { logs} = useSelector((state) => state.log);
    const [statusGreenTotal, setStatusGreenTotal] = useState(0);
    const [statusYellowTotal, setStatusYellowTotal] = useState(0);
    const [statusRedTotal, setStatusRedTotal] = useState(0);
    const [filteredLogs, setFilteredLogs] = useState(null);
    const { severity, db_type, time_period } = useSelector((state) => state.app);
    const [showSeverityGreen, setShowSeverityGreen] = useState(severity['green']);
    const [showSeverityYellow, setShowSeverityYellow] = useState(severity['yellow']);
    const [showSeverityRed, setShowSeverityRed] = useState(severity['red']);
    const [search, setSearch] = useState("");
    const [timePeriodType, setTimePeriodType] = useState(time_period);
    const [dbType, setDbType] = useState(db_type);

    const statusFilter = useCallback((data) => {
        let severityList = [];
        if (showSeverityGreen)
            severityList.push("green");
        if (showSeverityYellow)
            severityList.push("yellow");
        if (showSeverityRed)
            severityList.push("red");
        return data?.filter ? data?.filter((n) => severityList.indexOf(n.status) >= 0) : [];
    }, [showSeverityRed, showSeverityYellow, showSeverityGreen]);


    useEffect(() => {
        if (timePeriodType) dispatch(LogActions.getLogs({ time_period: timePeriodType }))
    }, [timePeriodType, dispatch])

    useEffect(() => {

        let tempStatusGreen = 0;
        let tempStatusYellow = 0;
        let tempStatusRed = 0;
        if (logs?.length > 0) {
            logs.forEach((n) => {
                switch (n.status) {
                    case "green":
                        ++tempStatusGreen;
                        break;
                    case "yellow":
                        ++tempStatusYellow;
                        break;
                    case "red":
                        ++tempStatusRed;
                        break;
                    default:
                        break;
                }
            });
            let tempData = logs;
            if (search)
                tempData = searchFilter(tempData, search);
            if (dbType) tempData = dbFilter(tempData, dbType);
            tempData = statusFilter(tempData);
            setFilteredLogs(tempData);
        }
        else setFilteredLogs(logs);
        setStatusGreenTotal(tempStatusGreen);
        setStatusYellowTotal(tempStatusYellow);
        setStatusRedTotal(tempStatusRed);
    }, [logs, dbType, search, statusFilter]);

    const dbFilter = (data, dbType) => {
        return dbType === types.ALL ? data : data?.filter ? data?.filter((n) => n.target_type?.includes(dbType)) : [];
    }
    const onFilterStatusChange = (status) => {
        let severityChoices = { green: showSeverityGreen, yellow: showSeverityYellow, red: showSeverityRed };
        if (status === "green") {

            severityChoices.green = !severityChoices.green;
            setShowSeverityGreen(severityChoices.green);

        }
        else if (status === "yellow") {


            severityChoices.yellow = !severityChoices.yellow;
            setShowSeverityYellow(severityChoices.yellow);

        }

        else if (status === "red") {
            severityChoices.red = !severityChoices.red;
            setShowSeverityRed(severityChoices.red);
        }
        dispatch(AppActions.setSeverity(severityChoices));
    }
    const onTimePeriodChange = (timePeriod, search, status, dbType) => {
        dispatch(AppActions.setTimePeriod(timePeriod));
        setTimePeriodType(timePeriod);
        let tempData = logs;
        if (search !== undefined && search)
            tempData = searchFilter(tempData, search);
        if (status !== undefined && status)
            tempData = statusFilter(tempData);
        if (dbType !== undefined && dbType)
            tempData = dbFilter(tempData, dbType);
        setFilteredLogs(tempData);
    }
    const onSearchChange = (search, status, dbType) => {
        setSearch(search);
        let tempData = logs;
        if (search)
            tempData = searchFilter(tempData, search);
        if (status !== undefined && status)
            tempData = statusFilter(tempData);
        if (dbType !== undefined && dbType)
            tempData = dbFilter(tempData, dbType);
        setFilteredLogs(tempData);
    }


    const onDbChange = (dbType, search, status) => {
        setDbType(dbType);
        dispatch(AppActions.setDbType(dbType));
        let tempData = logs;
        if (search)
            tempData = searchFilter(tempData, search);
        if (status !== undefined && status)
            tempData = statusFilter(tempData);
        if (dbType !== undefined && dbType)
            tempData = dbFilter(tempData, dbType);

        setFilteredLogs(tempData);
    }
    return (
        <div className="log-page">
            <Menu dbType={dbType} />
            <LogFilter
                statusGreenTotal={statusGreenTotal}
                statusYellowTotal={statusYellowTotal}
                statusRedTotal={statusRedTotal}
                onStatusChange={onFilterStatusChange}
                onTimePeriodChange={onTimePeriodChange}
                onSearchChange={onSearchChange}
                onDbChange={onDbChange}
                showSeverityGreen={showSeverityGreen}
                showSeverityYellow={showSeverityYellow}
                showSeverityRed={showSeverityRed}
            />
            <LogsTable filteredLogs={filteredLogs}
            timePeriodType={timePeriodType}
            />
        </div>
    );
}
export default Log;
