import { useCallback, useEffect, useState } from "react";
import { useDispatch, useSelector } from "react-redux";
import { Image } from "primereact/image";
import { InputText } from "primereact/inputtext";
import { Password } from "primereact/password";
import { Button } from "primereact/button";
import validator from "../../util/validator";
import notify from "../../util/notify";
import constraints from "./constraints";
import { AppActions } from "../../store/app/reducer";
import useKeyPress from "../../util/useKeyPress";
import logo from "../../assets/images/logo.gif";
import { SignupActions } from "../../store/signup/reducer";
import RequestDemoDialog from "../../dialogs/requestDemoDialog";
const Login = () => {
    const dispatch = useDispatch();
    const {username:cachedUsername, password:cachedPassword, org_name:cachedOrgName, errorMessage} = useSelector(state => state.app);
    const [username, setUsername] = useState(cachedUsername);
    const [password, setPassword] = useState(cachedPassword);
    const [orgName, setOrgName] = useState(cachedOrgName);
    const [usernameError, setUsernameError] = useState("");
    const [orgNameError, setOrgNameError] = useState("");
    const [passwordError, setPasswordError] = useState("");
    const [showDemo, setShowDemo] = useState(false);
    const enterPressed = useKeyPress("Enter");
    const onLoginButtonClick = useCallback(() => {
        const errors = [];
        let error = validator(constraints)("username", username);
        (error == null || errors.push(error)) && setUsernameError(error);
        error = validator(constraints)("org_name", orgName);
        (error == null || errors.push(error)) && setOrgNameError(error);
        error = validator(constraints)("password", password);
        (error == null || errors.push(error)) && setPasswordError(error);
        if (errors.length === 0)
            dispatch(AppActions.login(username, password, orgName));
    }, [dispatch, username, password, orgName])

    // const onDemoButtonClick = () => setShowDemo(true);
    
    useEffect(() => {
        if (errorMessage) {
            notify.error(errorMessage);
            dispatch(AppActions.setErrorMessage(""));
            dispatch(AppActions.logout());
        }
    }, [errorMessage, dispatch]);
    useEffect(() => 
    {
        if (enterPressed) onLoginButtonClick();
    }, [enterPressed, onLoginButtonClick]);
    return (
        <div className="login-page">
            <div className="flex justify-content-center align-items-center">
                <div className="form-container">
                    <div className="logo-container">
                        <div className="logo-wrapper">
                            <Image src={logo} alt="123 Cluster" width="180px" />
                        </div>
                    </div>
                    <div className="field">
                        <div className="p-inputgroup">
                            <span className="p-inputgroup-addon">
                                <i className="pi pi-building"></i>
                            </span>
                            <InputText type='text' name='org_name' placeholder="Organization Name" value={orgName} onChange={(e) => setOrgName(e.target.value)} />
                        </div>
                        {orgNameError && <small id="username2-help" className="p-error p-d-block">{orgNameError}</small>}
                    </div>
                    <div className="field">
                        <div className="p-inputgroup">
                            <span className="p-inputgroup-addon">
                                <i className="pi pi-user"></i>
                            </span>
                            <InputText type='text' name='username' placeholder="Username" value={username} onChange={(e) => setUsername(e.target.value)} />
                        </div>
                        {usernameError && <small id="username2-help" className="p-error p-d-block">{usernameError}</small>}
                    </div>
                    <div className="field">
                        <div className="p-inputgroup">
                            <span className="p-inputgroup-addon">
                                <i className="pi pi-key"></i>
                            </span>
                            <Password 
                            placeholder="Password" 
                            feedback={false} 
                            value={password}
                            autoComplete="new-password"
                            onChange={(e) => setPassword(e.target.value)}
                            toggleMask 
                            />
                        </div>
                        {passwordError && <small id="username2-help" className="p-error p-d-block">{passwordError}</small>}
                    </div>
                    <div className="field flex justify-content-center align-items-center">
                        <Button className="p-button mr-1" type="button" label="Sign in" onClick={onLoginButtonClick} />
                        {/* <Button className="p-button" type="button" label="Try it!" onClick={onDemoButtonClick} /> */}
                    </div>
                </div>
            </div>

            {showDemo ? <RequestDemoDialog visible = {showDemo} 
            onHide = {() => setShowDemo(false)}
            onOkClick = {(data) => dispatch(SignupActions.requestDemo(data, setShowDemo))}
            /> : null}
        </div>
    );
}
export default Login;
