import { useCallback, useEffect, useState } from "react";
import { useDispatch, useSelector } from "react-redux";
import moment from "moment";
import { MonitoringActions } from "../../store/monitoring/reducer";
import { AppActions } from "../../store/app/reducer";
import Menu from "../../components/menus/menu";
import MonitoringFilter from "../../components/filters/monitoringFilter";
import notify from "../../util/notify";
import { intervalByTimedelta, NodeAlertsTable } from '../../components/tables/nodeAlertsTable';
import { types } from "../../config";


export const searchFilter = (data, search) => {
    return data?.filter ? data?.filter((n) => {
        if (n.what?.toLowerCase().includes(search) || n.location?.toLowerCase().includes(search)
            || n.monitor_name?.toLowerCase().includes(search) ||
            moment(n.date_created_with_tz).format("MMM DD,YYYY, hh:mm A").includes(search))
            return true;
        if (n.severity && n.severity === parseInt(search))
            return true;
        return false;
    }) : [];
}

const Monitoring = () => {
    const dispatch = useDispatch();
    const { alerts, errorMessage } = useSelector((state) => state.monitoring);

    const [showTimelineColumn, setShowTimelineColumn] = useState(false);
    const [filteredAlerts, setFilteredAlerts] = useState(null);
    useEffect(() => {
        if (errorMessage) notify.error(errorMessage);
    }, [errorMessage]);
    const { severity, db_type, time_period } = useSelector((state) => state.app);
    const [showSeverityGreen, setShowSeverityGreen] = useState(severity['green']);
    const [showSeverityYellow, setShowSeverityYellow] = useState(severity['yellow']);
    const [showSeverityRed, setShowSeverityRed] = useState(severity['red']);
    const [severityGreenTotal, setSeverityGreenTotal] = useState(0);
    const [severityYellowTotal, setSeverityYellowTotal] = useState(0);
    const [severityRedTotal, setSeverityRedTotal] = useState(0);
    const [search, setSearch] = useState("");
    const [dbType, setDbType] = useState(db_type);
    const [timePeriodType, setTimePeriodType] = useState(time_period);
    const [monitorInterval, setMonitorInterval] = useState(intervalByTimedelta(timePeriodType));
    const severityFilter = useCallback((data) => {
        let severityList = [];
        if (showSeverityGreen) {
            severityList.push(1);
            severityList.push(-1);
        }
        if (showSeverityYellow) {
            severityList.push(2);
            severityList.push(-2);
        }
        if (showSeverityRed)
            severityList.push(3);
        return data?.filter ? data?.filter((n) => severityList.indexOf(n.severity) >= 0) : [];
    }, [showSeverityGreen, showSeverityYellow, showSeverityRed]);
    useEffect(() => {

        let tempSeverityGreen = 0;
        let tempSeverityYellow = 0;
        let tempSeverityRed = 0;
        if (!alerts)
            dispatch(MonitoringActions.getAlerts());

        if (alerts?.length > 0) {
            let tempData = alerts;
            if (search)
                tempData = searchFilter(tempData, search);
            if (dbType)
                tempData = dbFilter(tempData, dbType);
            tempData = severityFilter(tempData);
            setFilteredAlerts(tempData);
            alerts.forEach((n) => {
                switch (Math.abs(n.severity)) {
                    case 1:
                        ++tempSeverityGreen;
                        break;
                    case 2:
                        ++tempSeverityYellow;
                        break;
                    case 3:
                        ++tempSeverityRed;
                        break;
                    default:
                        break;
                }
            });
        }
        else setFilteredAlerts(alerts);
        setSeverityGreenTotal(tempSeverityGreen);
        setSeverityYellowTotal(tempSeverityYellow);
        setSeverityRedTotal(tempSeverityRed);
    }, [alerts, dbType, search, timePeriodType, dispatch, severityFilter]);
    const onFilterDbChange = (dbType, search, timePeriod, severity) => {
        setDbType(dbType);
        dispatch(AppActions.setDbType(dbType));
        let tempData = alerts;
        if (dbType)
            tempData = dbFilter(tempData, dbType);
        if (search !== undefined && search)
            tempData = searchFilter(tempData, search);
        if (severity !== undefined && severity)
            tempData = severityFilter(tempData);
        setFilteredAlerts(tempData);
    }
    const toggleTimeline = useCallback((flag) => {
        setShowTimelineColumn(flag);
    }, [])
    const onSearchChange = (search, dbType, timePeriod, severity) => {
        setSearch(search);
        let tempData = alerts;
        if (search)
            tempData = searchFilter(tempData, search);
        if (dbType !== undefined && dbType)
            tempData = dbFilter(tempData, dbType);
        if (severity !== undefined && severity)
            tempData = severityFilter(tempData);
        setFilteredAlerts(tempData);
    }

    const dbFilter = (data, dbType) => {
        return dbType === types.ALL ? data : data?.filter ? data?.filter((n) => n.type?.includes(dbType)) : [];
    }
    const onFilterSeverityChange = (severity, search, dbType, timePeriod) => {
        setSearch(search);
        setDbType(dbType);
        setTimePeriodType(timePeriod);
        let severityChoices = {green: showSeverityGreen, yellow: showSeverityYellow, red: showSeverityRed};
        if (severity === "green") {

            severityChoices.green = !severityChoices.green;
            setShowSeverityGreen(severityChoices.green);

        }
        else if (severity === "yellow") {


            severityChoices.yellow = !severityChoices.yellow;
            setShowSeverityYellow(severityChoices.yellow);

        }

        else if (severity === "red") {
            severityChoices.red = !severityChoices.red;
            setShowSeverityRed(severityChoices.red);
        }
        dispatch(AppActions.setSeverity(severityChoices));
    }
    const onTimePeriodChange = (timePeriod, dbType, search, severity) => {
        dispatch(AppActions.setTimePeriod(timePeriod));
        setTimePeriodType(timePeriod);
        let tempData = alerts;
        if (dbType !== undefined && dbType)
            tempData = dbFilter(tempData, dbType);
        if (search !== undefined && search)
            tempData = searchFilter(tempData, search);
        if (severity !== undefined && severity)
            tempData = severityFilter(tempData);
        setFilteredAlerts(tempData);
        setMonitorInterval(intervalByTimedelta(timePeriod));
    }

    return (
        <div className="monitoring-page">
            <Menu dbType={dbType} />
            <MonitoringFilter
                severityGreenTotal={severityGreenTotal}
                severityYellowTotal={severityYellowTotal}
                severityRedTotal={severityRedTotal}
                onDbChange={onFilterDbChange}
                onSeverityChange={onFilterSeverityChange}
                onSearchChange={onSearchChange}
                onTimePeriodChange={onTimePeriodChange}
                showSeverityGreen={showSeverityGreen}
                showSeverityYellow={showSeverityYellow}
                showSeverityRed={showSeverityRed}
                toggleTimeline={toggleTimeline}
            />
            <NodeAlertsTable noHeader monitorInterval={monitorInterval} filteredAlerts={filteredAlerts} timelineVisible={showTimelineColumn} toggleTimeline={() => setShowTimelineColumn(!showTimelineColumn)} />
        </div>
    );
}
export default Monitoring;