import { useCallback, useEffect, useState } from "react";
import { Column } from "primereact/column";
import { useDispatch, useSelector } from "react-redux";
import { StatusActions } from "../../store/status/reducer";
import Menu from "../../components/menus/menu";
import StatusFilter from "../../components/filters/statusFilter";
import getColorBySeverity from "../../util/colorBySeverity";
import { types, tableGroupingTypes, loadingMsg} from "../../config";
import { AppActions } from "../../store/app/reducer";
import { NodeAlertsTable, intervalByTimedelta } from "../../components/tables/nodeAlertsTable";
import { MonitoringActions } from "../../store/monitoring/reducer";
import moment from "moment";
import ChangeTableLayoutDialog from "../../dialogs/changeTableLayoutDialog";
import { DataTable } from "primereact/datatable";
import NodesTable from "../../components/tables/nodesTable";
import { useLocation, useNavigate } from "react-router-dom";
import { Button } from "primereact/button";

const Status = () => {
    const dispatch = useDispatch();
    const { nodes,  isLoading, nodeSeverityCounts } = useSelector((state) => state.status);
    const [selectedRecord, setSelectedRecord] = useState(null);
    const { severity, db_type, time_period, grouping, timeline } = useSelector((state) => state.app);
    const [showSeverityGreen, setShowSeverityGreen] = useState(severity['green']);
    const [showSeverityYellow, setShowSeverityYellow] = useState(severity['yellow']);
    const [showSeverityRed, setShowSeverityRed] = useState(severity['red']);
    const [severityGreenTotal, setSeverityGreenTotal] = useState(0);
    const [severityYellowTotal, setSeverityYellowTotal] = useState(0);
    const [severityRedTotal, setSeverityRedTotal] = useState(0);
    const [filteredNodes, setFilteredNodes] = useState(null);
    const [search, setSearch] = useState("");
    const [dbType, setDbType] = useState(db_type);
    const [timePeriodType, setTimePeriodType] = useState(time_period);
    const [showChangeTableLayout, setShowChangeTableLayout] = useState(false);
    const [tableLayout, setTableLayout] = useState(grouping);
    const [monitorInterval, setMonitorInterval] = useState(intervalByTimedelta(timePeriodType));
    const { alerts} = useSelector((state) => state.monitoring);
    const [filteredAlerts, setFilteredAlerts] = useState([]);
    const groupSwitch = (
      <div className="inline-flex">
        <i className="pi pi-bars mr-8" onClick={() => setShowChangeTableLayout(true)}
        style={{ transform: 'translateX(-200%)' }} />
        <div className="inline-flex">What</div>
      </div>
    );
    const [expandedRows, setExpandedRows] = useState(null);
    const [timelineVisible, setTimelineVisible] = useState(timeline);
    const { state } = useLocation();
    const navigate = useNavigate();

    const [mainSortField, setMainSortField] = useState(null);
    const [mainSortOrder, setMainSortOrder] = useState(null);

    const onMainSort = (event) => {
      setMainSortField(event.sortField);
      setMainSortOrder(event.sortOrder);
  };

    const onChangeTableLayout = useCallback((layoutType, timelineVisible) => {
        setShowChangeTableLayout(false);
        dispatch(AppActions.setGrouping(layoutType));
        dispatch(AppActions.setTimeline(timelineVisible));
        setTableLayout(layoutType);
        setTimelineVisible(timelineVisible);

    }, [dispatch])

    const dbFilterMonitors = (data, dbType) => {
        return dbType === types.ALL ? data : data?.filter ? data?.filter((n) => n.type?.includes(dbType)) : [];
    }

    const searchFilterMonitors = (data, search) => {
        return data?.filter ? data?.filter((n) => {
            if (n.what?.toLowerCase()?.includes(search) || n.location?.toLowerCase()?.includes(search)
                || n.monitor_name?.toLowerCase()?.includes(search) ||
                moment(n.date_created_with_tz)?.format("MMM DD,YYYY, hh:mm A")?.includes(search))
                return true;
            if (n.severity && n.severity === parseInt(search))
                return true;
            return false;
        }) : [];
    }

    const monitorSeverityFilter = useCallback((data) => {
        let severityList = [];
        if (showSeverityGreen) {
            severityList.push(1);
            severityList.push(-1);
        }
        if (showSeverityYellow) {
            severityList.push(2);
            severityList.push(-2);
        }
        if (showSeverityRed)
            severityList.push(3);
        return data?.filter ? data?.filter((n) => severityList.indexOf(n.severity) >= 0) : [];
    }, [showSeverityGreen, showSeverityYellow, showSeverityRed]);



    const severityFilter = useCallback((data) => {
        const filterRecursively = (node) => {
            const severityList = [];
            if (showSeverityGreen) severityList.push("green");
            if (showSeverityYellow) severityList.push("yellow");
            if (showSeverityRed) severityList.push("red");

            const matchesParent = severityList.indexOf(node.data ? node.data?.severity : node?.severity) >= 0;

            if (matchesParent) return true;

            // // Recursively check children
            // if (node.children) {
            //     node.children = node.children.filter(child => filterRecursively(child));
            //     return node.children.length > 0;
            // }

            return false;
        };

        return data?.filter ? data.filter(filterRecursively) : [];
    }, [showSeverityGreen, showSeverityYellow, showSeverityRed]);


    useEffect(() => {
        if (tableLayout !== tableGroupingTypes.NO_GROUPING){

            let tempSeverityGreen = 0;
            let tempSeverityYellow = 0;
            let tempSeverityRed = 0;

            const countSeverities = (node) => {
                switch (node.data ? node.data?.severity : node?.severity) {
                    case 'green':
                        ++tempSeverityGreen;
                        break;
                    case 'yellow':
                        ++tempSeverityYellow;
                        break;
                    case 'red':
                        ++tempSeverityRed;
                        break;
                    default:
                        break;
                }

                // // Recursively count severities in children
                // if (node.children) {
                //     node.children.forEach(child => countSeverities(child));
                // }
            };
            if (nodes?.length > 0) {
                nodes?.forEach(node => countSeverities(node));
                let tempData = structuredClone(nodes);
                if (search)
                    tempData = searchFilter(tempData, search);
                if (dbType)
                    tempData = dbFilter(tempData, dbType);
                tempData = severityFilter(tempData);
                if (state?.node_id && tableLayout === tableGroupingTypes.DATABASE)
                    tempData = nodes.filter((n) => n.id === state.node_id);
                if (state?.cluster_id && tableLayout === tableGroupingTypes.CLUSTER)
                    tempData = nodes.filter((c) => c.id === state.cluster_id);

                setFilteredNodes(tempData);

            }
            else setFilteredNodes(nodes);
            setSeverityGreenTotal(tempSeverityGreen);
            setSeverityYellowTotal(tempSeverityYellow);
            setSeverityRedTotal(tempSeverityRed);

        }

    }, [nodes, tableLayout, dbType, search, state, severityFilter, dispatch]);

    useEffect(() => {

        if (tableLayout !== tableGroupingTypes.NO_GROUPING) {
            setFilteredNodes([]);
            dispatch(StatusActions.getNodes(tableLayout === tableGroupingTypes.DATABASE));
        }

    }, [tableLayout, dispatch]);

    useEffect(() => {

        if (tableLayout === tableGroupingTypes.NO_GROUPING) {

            let tempSeverityGreen = 0;
            let tempSeverityYellow = 0;
            let tempSeverityRed = 0;
            if (!alerts)
                dispatch(MonitoringActions.getAlerts());

            if (alerts?.length > 0) {
                let tempData = alerts;
                if (search)
                    tempData = searchFilterMonitors(tempData, search);
                if (dbType)
                    tempData = dbFilterMonitors(tempData, dbType);
                tempData = monitorSeverityFilter(tempData);
                setFilteredAlerts(tempData);
                if (tableLayout === tableGroupingTypes.NO_GROUPING) {


                    alerts.forEach((n) => {
                        switch (Math.abs(n?.severity)) {
                            case 1:
                                ++tempSeverityGreen;
                                break;
                            case 2:
                                ++tempSeverityYellow;
                                break;
                            case 3:
                                ++tempSeverityRed;
                                break;
                            default:
                                break;
                        }
                    });

                }

            }
            else setFilteredAlerts(alerts);
            setSeverityGreenTotal(tempSeverityGreen);
            setSeverityYellowTotal(tempSeverityYellow);
            setSeverityRedTotal(tempSeverityRed);

        }


    }, [alerts, dbType, search, timePeriodType, tableLayout, dispatch, monitorSeverityFilter]);

    useEffect(() => {
      if (
          (tableLayout === tableGroupingTypes.CLUSTER || tableLayout === tableGroupingTypes.DATABASE) &&
          nodes && nodes.length > 0 && !isLoading
      ) {
          let nodeIds = [];
          if (tableLayout === tableGroupingTypes.CLUSTER) {
              nodes.forEach(cluster => {
                  if (cluster.nodes && Array.isArray(cluster.nodes)) {
                       cluster.nodes.forEach(node => {
                         const nodeId = node.data ? node.data.id : node.id;
                         if (nodeId != null) {
                             nodeIds.push(nodeId);
                         }
                       });
                  }
              });
          } else {
              nodeIds = nodes.map(node => node.data ? node.data.id : node.id).filter(id => id != null);
          }

          const uniqueNodeIds = [...new Set(nodeIds)];

          if (uniqueNodeIds.length > 0) {
              dispatch(StatusActions.getNodeSeverityCounts(uniqueNodeIds));
          }
      }
    }, [nodes, tableLayout, isLoading, dispatch]);

    const onFilterDbChange = useCallback((dbType, search, severity) => {
        setDbType(dbType);
        dispatch(AppActions.setDbType(dbType));
        let tempData = [];
        switch (tableLayout) {
            case tableGroupingTypes.NO_GROUPING:
                tempData = structuredClone(alerts);
                if (search)
                    tempData = searchFilterMonitors(tempData, search);
                if (dbType !== undefined && dbType)
                    tempData = dbFilterMonitors(tempData, dbType);
                if (severity !== undefined && severity)
                    tempData = monitorSeverityFilter(tempData);
                setFilteredAlerts(tempData);
                break;
            default:
                tempData = structuredClone(nodes);
                if (search)
                    tempData = searchFilter(tempData, search);
                if (dbType !== undefined && dbType)
                    tempData = dbFilter(tempData, dbType);
                if (severity !== undefined && severity)
                    tempData = severityFilter(tempData);
                setFilteredNodes(tempData);
                break;

        }
    }, [monitorSeverityFilter, severityFilter, alerts, dispatch, nodes, tableLayout])
    const onSearchChange = useCallback((search, dbType, severity) => {
        setSearch(search);
        let tempData = []
        switch (tableLayout) {
            case tableGroupingTypes.NO_GROUPING:
                tempData = structuredClone(alerts);
                if (search)
                    tempData = searchFilterMonitors(tempData, search);
                if (dbType !== undefined && dbType)
                    tempData = dbFilterMonitors(tempData, dbType);
                if (severity !== undefined && severity)
                    tempData = monitorSeverityFilter(tempData);
                setFilteredAlerts(tempData);
                break
            default:
                tempData = structuredClone(nodes);
                if (search)
                    tempData = searchFilter(tempData, search);
                if (dbType !== undefined && dbType)
                    tempData = dbFilter(tempData, dbType);
                if (severity !== undefined && severity)
                    tempData = severityFilter(tempData);
                setFilteredNodes(tempData);
                break;

        }

    }, [tableLayout, nodes, alerts, monitorSeverityFilter, severityFilter])

    const searchFilter = (data, search) => {
        const filterRecursively = (node) => {
            let matchesParent = false;
            if (node.data) {

                matchesParent = node.data.what?.toLowerCase()?.includes(search) ||
                    node.data?.location?.toLowerCase()?.includes(search) ||
                    node.data?.cluster_role?.toLowerCase()?.includes(search) ||
                    node.data?.status?.toLowerCase()?.includes(search) ||
                    node.data.alert?.toLowerCase()?.includes(search);

            }
            else {
                matchesParent = node.what?.toLowerCase()?.includes(search) ||
                    node.location?.toLowerCase()?.includes(search) ||
                    node.cluster_role?.toLowerCase()?.includes(search) ||
                    node.status?.includes(search) ||
                    node.alert?.includes(search);
            }

            if (matchesParent) return true;

            // Recursively check children
            if (node.children) {
                node.children = node.children.filter(child => filterRecursively(child));
                return node.children.length > 0;
            }

            return false;
        };

        return data?.filter ? data.filter(filterRecursively) : [];
    };

    const dbFilter = (data, dbType) => {
        const filterRecursively = (node) => {
            let matchesParent = false
            if (node.data && node.data.type) matchesParent = node.data.type.includes(dbType)
            else if (node.type) matchesParent = node.type.includes(dbType)

            if (matchesParent) return true;

            // Recursively check children
            if (node.children) {
                node.children = node.children.filter(child => filterRecursively(child));
                return node.children.length > 0;
            }

            return false;
        };

        return dbType === types.ALL ? data : data?.filter ? data.filter(filterRecursively) : [];
    };

    const statusBodyTemplate = (rowData) => {
      const severity = rowData.severity;

      if (severity === 'green' || severity === 'yellow' ) {
        return <span className="up-arrow">🡅</span>;
      } else if (severity === 'red') {
        return <span className="down-arrow">🡇</span>;
      } else {
        return severity;
      }
    };

    const onFilterSeverityChange = (severity, search, dbType) => {
        setSearch(search);
        setDbType(dbType);
        let severityChoices = { green: showSeverityGreen, yellow: showSeverityYellow, red: showSeverityRed }
        if (severity === "green") {

            severityChoices.green = !severityChoices.green;
            setShowSeverityGreen(severityChoices.green);

        }
        else if (severity === "yellow") {


            severityChoices.yellow = !severityChoices.yellow;
            setShowSeverityYellow(severityChoices.yellow);

        }

        else if (severity === "red") {
            severityChoices.red = !severityChoices.red;
            setShowSeverityRed(severityChoices.red);
        }
        dispatch(AppActions.setSeverity(severityChoices));
    }
    const onTimePeriodChange = (timePeriod) => {
        dispatch(AppActions.setTimePeriod(timePeriod));
        setTimePeriodType(timePeriod);
        setMonitorInterval(intervalByTimedelta(timePeriod));
    }

    const renderStatusTable = () => {

        switch (tableLayout) {

          case tableGroupingTypes.CLUSTER:
            return (
              <DataTable
              tableStyle={{ tableLayout: 'fixed', width: '100%' }}
              expandedRows={expandedRows}
              onRowToggle={(e) => {
                setExpandedRows(e.data);
              }}
              scrollable
              scrollHeight="flex"
              sortMode="single"
              onSort={onMainSort}
              sortField={mainSortField}
              sortOrder={mainSortOrder}
              emptyMessage={
                isLoading ? (
                  <div className="text-sm empty-severity">{loadingMsg}</div>
                ) : (
                  <div className="text-sm empty-severity">No nodes data</div>
                )
              }
              selectionMode="single"
              selection={selectedRecord}
              onSelectionChange={(e) => setSelectedRecord(e.value)}
              rowExpansionTemplate={(rowData) => {
                return (
                  <div>
                    {<NodesTable
                      expanded
                      groupSwitch={groupSwitch}
                      monitorInterval={monitorInterval}
                      timelineVisible={timelineVisible}
                      data={rowData.nodes}
                      parentSortField={mainSortField}
                      parentSortOrder={mainSortOrder}
                      nodeSeverityCounts={nodeSeverityCounts}
                      noHeadings
                      locationColumnWidth="25%"
                    /> }
                  </div>
                );
              }}
              className="severity-table"
              rowsPerPageOptions={[10, 20, 50, 100]}
              rowClassName={(rowData) => `severity-${getColorBySeverity(rowData.severity)}-row`}
              paginator
              alwaysShowPaginator={false}
              rows={100}
              value={filteredNodes}
            >
               <Column className="status-expander" expander />
               <Column
                field="what"
                header={groupSwitch}
                sortable
                body={(rowData) => (
                  <div>
                    {`${rowData.what} (cluster)`}{' '}
                    <Button
                      label="layout"
                      link
                      onClick={() =>
                        navigate('/layout', {
                          state: { cluster_id: rowData.nodes?.[0]?.cluster },
                        })
                      }
                    />{' '}
                  </div>
                )}
              ></Column>
               <Column
                field="name"
                header="Location"
                style={{ width: '23%', maxWidth: '23%' }}
                sortable
              ></Column>
              <Column
                field="status"
                header="Status"
                style={{ width: '26%', maxWidth: '26%' }}
                sortable
                body={statusBodyTemplate}
              > </Column>
              <Column
                header="Last sample"
                style={{ width: '25.7%', maxWidth: '25.7%' }}
                sortable
              > </Column>
             </DataTable>
          );

          case tableGroupingTypes.DATABASE:
            return (
                <NodesTable
                    database
                    groupSwitch={groupSwitch}
                    monitorInterval={monitorInterval}
                    timelineVisible={timelineVisible}
                    data={filteredNodes}
                    nodeSeverityCounts={nodeSeverityCounts}
                />
            );
            case tableGroupingTypes.NO_GROUPING:
                return <NodeAlertsTable noHeader groupSwitch={groupSwitch} monitorInterval={monitorInterval} filteredAlerts={filteredAlerts} timeVisible={true} timelineVisible={timelineVisible} />
            default:
                return null;


        }

    }

    return (
        <div className="status-page">
            <Menu dbType={dbType} />
            <StatusFilter
                severityGreenTotal={severityGreenTotal}
                severityYellowTotal={severityYellowTotal}
                severityRedTotal={severityRedTotal}
                onDbChange={onFilterDbChange}
                onSeverityChange={onFilterSeverityChange}
                onSearchChange={onSearchChange}
                onTimePeriodChange={onTimePeriodChange}
                showSeverityGreen={showSeverityGreen}
                showSeverityYellow={showSeverityYellow}
                showSeverityRed={showSeverityRed}
            />
            {renderStatusTable()}

            {showChangeTableLayout ? <ChangeTableLayoutDialog
                visible={showChangeTableLayout}
                selectedGrouping={tableLayout}
                timelineVisible = {timelineVisible}
                onHide={() => setShowChangeTableLayout(false)}
                onOkClick={onChangeTableLayout}

            /> : null}

        </div>
    );
}
export default Status;
