import { useMemo } from 'react';

const LineChart = ({
  data = [{ index: 0, col_name: 0 }],
  svgHeight = 200,
  svgWidth = 600,
  strokeWidth = 2,
  color = '#ffffff',
  fillColor,
}) => {

  // Use useMemo to calculate scales and processed data
  const { minX, minY, maxY, processedData, xScale, yScale } = useMemo(() => {
    const onlyX = data.map(obj => obj.index);
    const onlyY = data.map(obj => {
      const value = Number(obj.col_name);
      return isNaN(value) ? 0 : value; // Filter out non-numeric values
    });

    const processedData = data.map(obj => ({
      index: obj.index,
      col_name: Number(obj.col_name),
    }));

    const minX = Math.min(...onlyX);
    const maxX = Math.max(...onlyX);
    const minY = Math.min(...onlyY);
    const maxY = Math.max(...onlyY);

    // Pre-calculate scale factors with safeguards
    const xScale = maxX === minX ? 1 : svgWidth / (maxX - minX || 1);
    const yScale = maxY === minY ? 1 : svgHeight / (maxY - minY || 1);

    return { minX, maxX, minY, maxY, processedData, xScale, yScale };
  }, [data, svgWidth, svgHeight]);

  const pathD = useMemo(() => {
    if (!processedData.length) return '';

    return processedData.reduce(
      (path, point) => `${path} L ${(point.index - minX) * xScale} ${(svgHeight - (point.col_name - minY) * yScale)}`,
      `M ${(processedData[0].index - minX) * xScale} ${(svgHeight - (processedData[0].col_name - minY) * yScale)}`
    );
  }, [processedData, minX, minY, xScale, yScale, svgHeight]);

  const axisLines = useMemo(() => {
    const yAxis = {
      x1: 0,
      y1: svgHeight,
      x2: 0,
      y2: svgHeight - (svgHeight * (maxY - minY)) / (maxY || 1),
    };

    const xAxis = {
      x1: 0,
      y1: svgHeight - minY * yScale,
      x2: svgWidth,
      y2: svgHeight - minY * yScale,
    };

    return { xAxis, yAxis };
  }, [minY, maxY, yScale, svgWidth, svgHeight]);

    // Early return if there's not enough data
    if (!data || data.length < 2) {
      return <svg viewBox={`0 0 ${svgWidth} ${svgHeight}`} />;
    }
  return (
    <svg viewBox={`0 0 ${svgWidth} ${svgHeight}`} role="img" aria-label="Line Chart">
      <title>Line Chart</title>
      <desc>A line chart representing the provided data.</desc>
      <path
        className="linechart_path"
        d={pathD}
        stroke={color}
        fill={fillColor}
        strokeWidth={strokeWidth}
      />
      <g className="linechart_axis">
        <line {...axisLines.xAxis} stroke={color} strokeWidth={strokeWidth} />
        <line {...axisLines.yAxis} stroke={color} strokeWidth={strokeWidth} />
      </g>
    </svg>
  );
};
export default LineChart;
