--
-- TOC entry 532 (class 1255 OID 36427282)
-- Name: check_monitor_status(integer); Type: FUNCTION; Schema: public; Owner: -
--

CREATE OR REPLACE FUNCTION public.check_monitor_status(node_id integer) RETURNS void
    LANGUAGE plpgsql
    AS $$
DECLARE
    monitor_record RECORD;
    task_log_count INT;
    result_raw_count INT;
    alert_count INT;
    alert_detail_count INT;
    timeline_chunks_count INT;
BEGIN
    FOR monitor_record IN
        SELECT task_id, monitor_name 
        FROM public."clusterCore_task" cct 
        WHERE target_id = node_id AND monitor_name IS NOT NULL 
        GROUP BY task_id, monitor_name
    LOOP
        -- Check clusterCore_tasklog
        SELECT COUNT(*) INTO task_log_count
        FROM public."clusterCore_tasklog" cct2 
        WHERE cct2.task_id = monitor_record.task_id;

        -- Check rp_monitor_result_raw
        SELECT COUNT(*) INTO result_raw_count
        FROM app_123monitor.rp_monitor_result_raw rmrr 
        WHERE rmrr.node_id = node_id AND rmrr.task_id = monitor_record.task_id;

        -- Check rp_alert
        SELECT COUNT(*) INTO alert_count
        FROM app_123monitor.rp_alert ra 
        WHERE ra.node_id = node_id AND ra.monitor_name = monitor_record.monitor_name;

        -- Check rp_alert_detail
        IF alert_count > 0 THEN
            SELECT COUNT(*) INTO alert_detail_count
            FROM app_123monitor.rp_alert_detail rad 
            WHERE rad.alert_id IN (
                SELECT alert_id 
                FROM app_123monitor.rp_alert ra 
                WHERE ra.node_id = node_id AND ra.monitor_name = monitor_record.monitor_name
            );
        ELSE
            alert_detail_count := 0;
        END IF;

        -- Check rp_timeline_chunks
        SELECT COUNT(*) INTO timeline_chunks_count
        FROM app_123monitor.rp_timeline_chunks rtc 
        WHERE rtc.node_id = node_id AND rtc.timeline_name = monitor_record.monitor_name;

        IF task_log_count = 0 THEN
            RAISE info 'Monitor % (task_id: %) has no entries in clusterCore_tasklog.', monitor_record.monitor_name, monitor_record.task_id;
        END IF;

        IF result_raw_count = 0 THEN
            RAISE info 'Monitor % (task_id: %) has no entries in rp_monitor_result_raw.', monitor_record.monitor_name, monitor_record.task_id;
        END IF;

        IF alert_count = 0 THEN
            RAISE info 'Monitor % (task_id: %) has no entries in rp_alert.', monitor_record.monitor_name, monitor_record.task_id;
        END IF;

        IF alert_detail_count = 0 AND alert_count > 0 THEN
            RAISE info 'Monitor % (task_id: %) has no entries in rp_alert_detail.', monitor_record.monitor_name, monitor_record.task_id;
        END IF;

        IF timeline_chunks_count = 0 THEN
            RAISE info 'Monitor % (task_id: %) has no entries in rp_timeline_chunks.', monitor_record.monitor_name, monitor_record.task_id;
        END IF;
    END LOOP;
END;
$$;


--
-- TOC entry 560 (class 1255 OID 17219)
-- Name: insert_task(text, json, text, text, numeric, numeric, numeric, text); Type: FUNCTION; Schema: public; Owner: -
--

CREATE OR REPLACE FUNCTION public.insert_task(in_schedule text, in_params json, in_task text, in_target_type text, in_target_id numeric, in_org_id numeric, in_created_by numeric, in_monitor_name text DEFAULT NULL::text) RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE
module  text :='insert_task';
step    text :='start insert_task function';
errmsg text:=null;
ecode text:='';
status_task text:='ready';
t_id int:=null;
BEGIN
    if in_target_id is null 
   then status_task='done';
   end if;
   
    insert into "clusterCore_task" (org_id,schedule,created_date,parameters,status,task,target_type,target_id,created_by_id,monitor_name)
    VALUES (in_org_id,in_schedule,now(),in_params,status_task,in_task,in_target_type,in_target_id,in_created_by,in_monitor_name)
    returning task_id into t_id;
    raise info 'write log % % % ',module,step,now();
   return t_id;
exception
    when others then
     GET STACKED DIAGNOSTICS errmsg := MESSAGE_TEXT,
                              ecode := RETURNED_SQLSTATE;
        PERFORM write_log(-1,null,ecode,errmsg,module,step,3);
       return t_id;           
END;
$$;


--
-- TOC entry 591 (class 1255 OID 54555106)
-- Name: logger(numeric, uuid, text, text, numeric, numeric, text, text, text, numeric, json); Type: FUNCTION; Schema: public; Owner: -
--

CREATE OR REPLACE FUNCTION public.logger(task_id numeric, task_log_id uuid, task text, status text, duration numeric, created_by numeric, step text DEFAULT NULL::text, stdout text DEFAULT NULL::text, stderr text DEFAULT NULL::text, excode numeric DEFAULT NULL::numeric, ext_field json DEFAULT '{}'::json) RETURNS void
    LANGUAGE plpgsql
    AS $$
DECLARE
    errmsg TEXT := NULL;
    ecode TEXT := '1';
    v_target_type TEXT;
    v_target_id NUMERIC;
BEGIN
    -- Fetch target_type and target_id with explicit parameter reference using alias
    SELECT target_type, target_id
    INTO v_target_type, v_target_id
    FROM "clusterCore_task"
    WHERE "clusterCore_task".task_id = logger.task_id; -- Correct fix using function alias

    -- Decide where to insert the log entry
    IF step IS NULL OR step = 'null' THEN
        -- Insert into clusterCore_tasklog
        INSERT INTO "clusterCore_tasklog" (
            task_id, task_log_id, task, created_date, status, duration, 
            created_by_id, retoutput, retmsg, retcode, ext_field, target_type, target_id
        )
        VALUES (
            task_id, task_log_id, task, NOW(), status, duration, 
            created_by, stdout, stderr, excode, ext_field, v_target_type, v_target_id
        );
    ELSE
        -- Insert into clusterCore_tasksteplog
        INSERT INTO "clusterCore_tasksteplog" (
            task_id, task_log_id, task, created_date, status, duration, 
            step, retoutput, retmsg, retcode, target_type, target_id
        )
        VALUES (
            task_id, task_log_id, task, NOW(), status, duration, 
            step, stdout, stderr, excode, v_target_type, v_target_id
        );
    END IF;

    -- Log information
    RAISE INFO 'write log % % %', task, step, NOW();

EXCEPTION
    WHEN OTHERS THEN
        GET STACKED DIAGNOSTICS errmsg := MESSAGE_TEXT,
                                ecode := RETURNED_SQLSTATE;
        PERFORM write_log(-1, '', ecode, errmsg, task, step, 3);
END;
$$;


--
-- TOC entry 527 (class 1255 OID 17221)
-- Name: set_logfile_data(text, text, numeric); Type: FUNCTION; Schema: public; Owner: -
--

CREATE OR REPLACE FUNCTION public.set_logfile_data(subject text, in_logfile text, in_id numeric) RETURNS void
    LANGUAGE plpgsql
    AS $$
DECLARE
module  text :='set_logfile_data';
step    text :='start set_logfile_data function';
errmsg text:=null;
ecode text:='0';
begin
case subject
when 'BACKUP'
    then update "clusterCore_c123_inventory" set run_log=in_logfile where id=in_id;
    when 'EXPORT'
    then update "clusterCore_c123_inventory" set run_log=in_logfile where id=in_id;
    end case;
    raise info 'write log % % % ',module,step,now();
exception
    when others then
     GET STACKED DIAGNOSTICS errmsg := MESSAGE_TEXT,
                              ecode := RETURNED_SQLSTATE;
        PERFORM write_log(-1,ecode,errmsg,module,step,3);
           
END;
$$;


--
-- TOC entry 563 (class 1255 OID 17222)
-- Name: update_task(text, json, text, text, numeric, numeric, numeric, text); Type: FUNCTION; Schema: public; Owner: -
--

CREATE OR REPLACE FUNCTION public.update_task(in_schedule text, in_params json, in_task text, in_target_type text, in_target_id numeric, in_org_id numeric, in_created_by numeric, in_monitor_name text DEFAULT NULL::text) RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE
module  text :='update_task';
step    text :='start update_task function';
errmsg text:=null;
ecode text:='';
status_task text:='ready';
t_id int:=null;
BEGIN
    if in_target_id is null 
   then status_task='done';
   end if;
   
   if in_target_id is null 
   then
    update "clusterCore_task" set 
    org_id=in_org_id,
    schedule=in_schedule,
    created_date=now(),
    parameters=in_params,
    status=status_task,
    task=in_task,
    target_type=in_target_type,
    target_id=in_target_id,
    created_by_id=in_created_by,
    monitor_name=in_monitor_name
where 
org_id=in_org_id and 
task=in_task and 
target_type=in_target_type and 
target_id is null and 
monitor_name=in_monitor_name
    returning task_id into t_id;
   
   else
   update "clusterCore_task" set 
    org_id=in_org_id,
    schedule=in_schedule,
    created_date=now(),
    parameters=in_params,
    status=status_task,
    task=in_task,
    target_type=in_target_type,
    target_id=in_target_id,
    created_by_id=in_created_by,
    monitor_name=in_monitor_name
where 
org_id=in_org_id and 
task=in_task and 
target_type=in_target_type and 
target_id=in_target_id and 
monitor_name=in_monitor_name
    returning task_id into t_id;
   end if;
    raise info 'write log % % % ',module,step,now();
   return t_id;
exception
    when others then
     GET STACKED DIAGNOSTICS errmsg := MESSAGE_TEXT,
                              ecode := RETURNED_SQLSTATE;
        PERFORM write_log(-1,ecode,errmsg,module,step,3);
       return t_id;           
END;
$$;


--
-- TOC entry 570 (class 1255 OID 38062527)
-- Name: wipeout(); Type: FUNCTION; Schema: public; Owner: -
--

CREATE OR REPLACE FUNCTION public.wipeout() RETURNS void
    LANGUAGE plpgsql
    AS $$
BEGIN
--public
  TRUNCATE public."clusterCore_c123_cluster";
  DELETE FROM public."clusterCore_c123_host" WHERE ex_params->>'myself' <> 'true';
  TRUNCATE public."clusterCore_c123_node" cascade;
  TRUNCATE public."clusterCore_c123_database" cascade;
  TRUNCATE public."clusterCore_c123_inventory";

  DELETE FROM public."clusterCore_task" WHERE target_id IS NOT NULL;
  TRUNCATE public."clusterCore_tasklog";
  TRUNCATE public."clusterCore_tasksteplog";
  TRUNCATE public.db_log;

  TRUNCATE public."clusterCore_c123_host_monitor";
  TRUNCATE public."clusterCore_c123_node_monitor";

  TRUNCATE public.dashboard_item;
  
  TRUNCATE public.node_json;

  TRUNCATE public."clusterCore_c123_oracle_standalone";

  TRUNCATE public.django_admin_log;
  TRUNCATE public.django_apscheduler_djangojobexecution;

-- The following tables are explicitly NOT truncated:
-- "clusterCore_usernotifications";
-- "clusterCore_userprivileges";
-- "clusterCore_oracleclustersizing";
-- "clusterCore_oraclesystemconfig";
-- integrated_services;
-- c123_kubernetes_c123_cluster;
-- c123_kubernetes_c123_controlplane;
-- c123_kubernetes_c123_pod;
-- "clusterCore_clustersizing";
-- "clusterCore_systemconfig";
-- clusterCore_usernotifications";
-- clusterCore_userprivileges";
-- clusterCore_clustersizing";
-- clusterCore_systemconfig";
-- auth_group
-- clusterCore_profile;
-- auth_user
-- clusterCore_customerkey
-- django_apscheduler_djangojob
-- django_content_type
-- django_migrations
-- django_session
-- organization
-- auth_permission
-- auth_user_groups
-- auth_user_user_permissions
-- auth_group_permissions

--app_123monitor
  TRUNCATE app_123monitor.rp_nodes;
  TRUNCATE app_123monitor.r_params;
  TRUNCATE app_123monitor.rp_alert;
  TRUNCATE app_123monitor.rp_alert_day;
  TRUNCATE app_123monitor.rp_alert_detail;
  TRUNCATE app_123monitor.rp_alert_detail_avg_day;
  TRUNCATE app_123monitor.rp_alert_detail_hour;
  TRUNCATE app_123monitor.rp_alert_detail_max_day;
  TRUNCATE app_123monitor.rp_alert_detail_min_day;
  TRUNCATE app_123monitor.rp_alert_hour;
  TRUNCATE app_123monitor.rp_mail_queue;
  TRUNCATE app_123monitor.rp_monitor_result_raw;
  TRUNCATE app_123monitor.rp_timeline_chunks;
  TRUNCATE app_123monitor.tmp_added_missing;
  TRUNCATE app_123monitor.tmp_basic;
  TRUNCATE app_123monitor.tmp_basic_added_missing;
  TRUNCATE app_123monitor.tmp_compare_to_previous;
  TRUNCATE app_123monitor.tmp_current;
  TRUNCATE app_123monitor.tmp_final;
  TRUNCATE app_123monitor.tmp_previous;
  TRUNCATE app_123monitor.t_id;

  DELETE from app_123monitor.rp_monitors WHERE node_id IS NOT NULL;
  DELETE from app_123monitor.rp_timelines WHERE node_id IS NOT NULL;

-- The following tables are explicitly NOT truncated:
-- app_123monitor.monitor_text;
-- app_123monitor.rp_monitor_columns;
-- app_123monitor.rp_agent;
-- app_123monitor.rp_organization;


END;
$$;


--
-- TOC entry 596 (class 1255 OID 55743835)
-- Name: wipeout_logs(); Type: FUNCTION; Schema: public; Owner: -
--

CREATE OR REPLACE FUNCTION public.wipeout_logs() RETURNS void
    LANGUAGE plpgsql
    AS $$
BEGIN
--public
  TRUNCATE public."clusterCore_tasklog";
  TRUNCATE public."clusterCore_tasksteplog";
  TRUNCATE public.db_log;

  TRUNCATE public.django_admin_log;
  --TRUNCATE public.django_apscheduler_djangojobexecution;

--app_123monitor
  TRUNCATE app_123monitor.rp_alert;
  TRUNCATE app_123monitor.rp_alert_current;
  TRUNCATE app_123monitor.rp_alert_day;
  TRUNCATE app_123monitor.rp_alert_detail;
  TRUNCATE app_123monitor.rp_alert_detail_avg_day;
  TRUNCATE app_123monitor.rp_alert_detail_avg_hour;
  TRUNCATE app_123monitor.rp_alert_detail_avg_six_hours;
  TRUNCATE app_123monitor.rp_alert_detail_avg_week;
  TRUNCATE app_123monitor.rp_alert_detail_hour;
  TRUNCATE app_123monitor.rp_alert_detail_max_day;
  TRUNCATE app_123monitor.rp_alert_detail_max_hour;
  TRUNCATE app_123monitor.rp_alert_detail_max_six_hours;
  TRUNCATE app_123monitor.rp_alert_detail_max_week;
  TRUNCATE app_123monitor.rp_alert_detail_min_day;
  TRUNCATE app_123monitor.rp_alert_detail_min_hour;
  TRUNCATE app_123monitor.rp_alert_detail_min_six_hours;
  TRUNCATE app_123monitor.rp_alert_detail_min_week;
  TRUNCATE app_123monitor.rp_alert_hour;
  TRUNCATE app_123monitor.rp_alert_six_hours;
  TRUNCATE app_123monitor.rp_alert_week;

  TRUNCATE app_123monitor.rp_timeline CASCADE;
  TRUNCATE app_123monitor.rp_timeline_chunks;
  TRUNCATE app_123monitor.tmp_added_missing;
  TRUNCATE app_123monitor.tmp_basic;
  TRUNCATE app_123monitor.tmp_basic_added_missing;
  TRUNCATE app_123monitor.tmp_compare_to_previous;
  TRUNCATE app_123monitor.tmp_current;
  TRUNCATE app_123monitor.tmp_final;
  TRUNCATE app_123monitor.tmp_previous;
  TRUNCATE app_123monitor.t_id;



END;
$$;


--
-- TOC entry 512 (class 1255 OID 36193078)
-- Name: write_log(integer, text, text, text, text, text, integer); Type: FUNCTION; Schema: public; Owner: -
--

CREATE OR REPLACE FUNCTION public.write_log(in_node_id integer, in_monitor_name text, in_ercode text, in_msgerr text, in_module text, in_step text, in_severity integer) RETURNS void
    LANGUAGE plpgsql
    AS $$
DECLARE

BEGIN

    insert into db_log (node_id , monitor_name , ecode , errmsg , module , step , severity_id , date_created)
    VALUES (in_node_id,in_monitor_name,in_ercode,in_msgerr,in_module,in_step,in_severity, now());
    raise info 'write log % % % ',in_module,in_step,now();
exception when others then
    raise notice 'from write_log % %', SQLERRM, SQLSTATE;
end;
$$;
