import time, uuid
from abc import ABC

import urllib3
from typing import Any

from kubernetes import client

from kubernetes.client import CoreV1Api, ApiClient, CustomObjectsApi

from enums import Severity
from tasks_core import logger

# Disable only InsecureRequestWarning
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class StdErrException(Exception):
    def __init__(self, streams):
        super().__init__(streams)
        # remote_command in tasks_core.py does not return stdin
        # self.stdin = streams['stdin'].lower()
        self.stdout = streams["stdout"].lower()
        self.stderr = streams["stderr"].lower()
        self.excode = streams["excode"]


class ControlPlane:
    def __init__(self, api_data: dict[str, Any]):
        self._api_hostname = api_data["name"]
        self._api_port = api_data["port"]
        self._token = api_data["token"]

    def _get_api_client(self) -> ApiClient:
        aConfiguration = client.Configuration()
        aConfiguration.host = f"https://{self._api_hostname}:{self._api_port}"
        aConfiguration.verify_ssl = False
        aConfiguration.api_key = {"authorization": self._token}
        return client.ApiClient(aConfiguration)

    def get_core_v1_api(self) -> CoreV1Api:
        api_client = self._get_api_client()
        return client.CoreV1Api(api_client)

    def get_custom_objects_api(self) -> CustomObjectsApi:
        api_client = self._get_api_client()
        return client.CustomObjectsApi(api_client)


class Cluster(ABC):
    def __init__(self, cp: ControlPlane, cluster_data: dict[str, Any]):
        self.cp = cp
        self.name = cluster_data["name"]
        self.namespace = cluster_data["ex_params"]["namespace"]


class Host:
    def __init__(self, host_data: dict[str, Any]):
        self.name = host_data["name"]


class Pod(ABC):
    def __init__(self, cluster: Cluster, pod_data: dict[str, Any]):
        self.cluster = cluster
        self.id = pod_data["id"]

        if pod_data.get("ram_gb"):
            self.ram_gb = pod_data["ram_gb"]

        if pod_data.get("cpu_cores"):
            self.cpu_cores = pod_data["cpu_cores"]

        if pod_data.get("disk_gb"):
            self.disk_gb = pod_data["disk_gb"]


class TaskLogger:
    def __init__(self, task_id: int, task_name: str, created_by: int):
        self.id = task_id
        self.log_id = uuid.uuid4().hex
        self.name = task_name
        self.status = Severity.OK.value
        self.start_time = time.perf_counter()
        self.duration = 0
        self.created_by = created_by
        self.stdout = None
        self.stderr = None
        self.excode = None
        self._log_task_state()

    def __del__(self):
        self.duration = round(time.perf_counter() - self.start_time, 2)
        self._log_task_state()

    def step(self, step_name: str):
        return self._Step(step_name, self)

    def warning(self, warning_msg: str):
        self.status = Severity.WARNING.value
        self.stderr = warning_msg

    def error(self, error_msg: str, excode: int):
        self.status = Severity.NOT_OK.value
        self.stderr = error_msg
        self.excode = excode

    def _log_task_state(self):
        logger(
            self.id,
            self.log_id,
            self.name,
            "null",  # step name is null for task-level logs
            self.status,
            self.duration,
            self.created_by,
            self.stdout,
            self.stderr,
            self.excode,
        )

    class _Step:
        def __init__(self, step_name: str, logger):
            self.step_name = step_name
            self.step_start_time = time.perf_counter()
            self.step_duration = 0
            self.logger = logger

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            if exc_type is StdErrException:
                self._handle_stderr_exception(exc_val)
            elif exc_type is Exception:
                self._handle_generic_exception(exc_val)

            self.step_duration = round(time.perf_counter() - self.step_start_time, 2)
            self._log_step_state()
            return False

        def _handle_stderr_exception(self, exc_val):
            self.logger.status = Severity.NOT_OK.value
            self.logger.stdout = exc_val.stdout
            self.logger.stderr = exc_val.stderr
            self.logger.excode = exc_val.excode

        def _handle_generic_exception(self, exc_val):
            self.logger.status = Severity.NOT_OK.value
            self.logger.stdout = None
            self.logger.stderr = exc_val
            self.logger.excode = 9

        def _log_step_state(self):
            if self.step_name == "initializing task":
                logger(
                    self.logger.id,
                    self.logger.log_id,
                    self.logger.name,
                    self.step_name,
                    # blue color for initializing task
                    Severity.INFO.value,
                    self.step_duration,
                    self.logger.created_by,
                    self.logger.stdout,
                    self.logger.stderr,
                    self.logger.excode,
                )
                return

            logger(
                self.logger.id,
                self.logger.log_id,
                self.logger.name,
                self.step_name,
                self.logger.status,
                self.step_duration,
                self.logger.created_by,
                self.logger.stdout,
                self.logger.stderr,
                self.logger.excode,
            )
