# BELONG TO SERVICE TASK
import base64
import json, logging, yaml
import re
import socket
import time
from datetime import datetime
from pprint import pprint
from typing import Any, Optional

import psycopg2
from kubernetes import watch
from kubernetes.client import ApiException

from enums import Severity, Status, OperatingSystem
from k8s_tasks_core import TaskLogger, ControlPlane, Pod, Cluster, Host
from tasks_core import get_postgres_connection, connect_to_postgres_database, generate_complex_password, \
    format_for_shell, remote_command, find_tilda, check_os_version, get_date_modified_folder
from tasks_postgres import is_ip, find_postgres_bin


class PostgresCluster(Cluster):
    def __init__(self, cp: ControlPlane, cluster_data: dict[str, Any]):
        super().__init__(cp, cluster_data)
        self.superuser_secret_name = f"{self.name}-superuser"

    def wait_for_cluster_is_ready(self, timeout: int=300) -> None:
        custom_objects_api = self.cp.get_custom_objects_api()
        w = watch.Watch()

        for event in w.stream(
            custom_objects_api.list_namespaced_custom_object,
            group="postgresql.cnpg.io",
            version="v1",
            namespace=self.namespace,
            plural="clusters",
            timeout_seconds=timeout
        ):
            cluster = event["object"]
            if cluster.get("metadata", {}).get("name") == self.name:
                phase = cluster.get("status", {}).get("phase", "").lower()
                if phase == "cluster in healthy state":
                    w.stop()
                    return

        raise TimeoutError(f'Timeout reached. Cluster "{self.name}" is NOT Ready.')


class PostgresPod(Pod):
    def __init__(self, cluster: Cluster, pod_data: dict[str, Any]):
        super().__init__(cluster, pod_data)
        self._postgres = self.PostgresDatabase(pod_data)

    @property
    def postgres(self):
        return self._postgres
        
    class PostgresDatabase:
        def __init__(self, pod_data: dict[str, Any]):
            self.username = pod_data["database_user"]
            self.password = pod_data["database_pwd"]


def install_k8s_postgres_standalone_task(data, task_id):
    task_name = "install k8s postgres standalone task"

    try:
        data = json.loads(data)
        created_by = int(data["created_by"])

        cp = ControlPlane(json.loads(data["kubernetes_api_data"]))
        cluster = PostgresCluster(cp, json.loads(data["cluster_data"]))
        host = Host(json.loads(data["host_data"]))
        pod = PostgresPod(cluster, json.loads(data["pod_data"]))

        core_v1_api = cp.get_core_v1_api()
        custom_objects_api = cp.get_custom_objects_api()

        logger = TaskLogger(task_id, task_name, created_by)
    except Exception as error:
        logging.error(f"{task_name}: {str(error)}")
        return

    with logger.step("initializing task"):
        logger.status = Severity.OK.value
        pass

    with logger.step("updating configuration"):
        secret_manifest = {
            "apiVersion": "v1",
            "data": {
                "username": base64.b64encode(pod.postgres.username.encode('utf-8')).decode('utf-8'),
                "password": base64.b64encode(pod.postgres.password.encode('utf-8')).decode('utf-8')
            },
            "kind": "Secret",
            "metadata": {
                "name": cluster.superuser_secret_name,
                "namespace": cluster.namespace,
                "labels": {
                    "cnpg.io/reload": "true"
                }
            },
            "type": "kubernetes.io/basic-auth"
        }

        cluster_manifest = {
            "apiVersion": "postgresql.cnpg.io/v1",
            "kind": "Cluster",
            "metadata": {
                "name": cluster.name,
                "namespace": cluster.namespace
            },
            "spec": {
                "instances": 1,
                "managed": {
                    "roles": [
                        {
                            "name": pod.postgres.username,
                            "ensure": "present",
                            "login": True,
                            "superuser": True,
                            "passwordSecret": {
                                "name": cluster.superuser_secret_name
                            }
                        }
                    ]
                },
                "storage": {
                    "size": "1Gi"
                },
                # TODO: Add auth entry for backup host in the pg_hba.conf
                # "postgresql": {
                #     "pg_hba": [
                #         "hostssl app app 10.244.0.0/16 md5"
                #     ]
                # }
            }
        }

    with logger.step("creating cluster"):
        try:
            core_v1_api.read_namespace(cluster.namespace)
        except ApiException as e:
            if e.status == 404:  # Create namespace if not found
                namespace_body = {"metadata": {"name": cluster.namespace}}
                core_v1_api.create_namespace(namespace_body)
            else:
                raise e

        try:
            core_v1_api.create_namespaced_secret(
                namespace=cluster.namespace,
                body=secret_manifest
            )
        except ApiException as e:
            if e.status == 409:  # Conflict: Secret already exists
                existing_secret = core_v1_api.read_namespaced_secret(name=cluster.superuser_secret_name, namespace=cluster.namespace)
                existing_secret.data = secret_manifest["data"]
                core_v1_api.replace_namespaced_secret(name=cluster.superuser_secret_name, namespace=cluster.namespace, body=existing_secret)
            else:
                raise e

        custom_objects_api.create_namespaced_custom_object(
            group="postgresql.cnpg.io",
            version="v1",
            namespace=cluster.namespace,
            plural="clusters",
            body=cluster_manifest,
        )

        cluster.wait_for_cluster_is_ready()

    with logger.step("task finished"):
        pass

    with get_postgres_connection() as conn:
        with conn.cursor() as cur:
            cur.execute(
                """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""",
                (
                    logger.status,
                    Status.OK.value,
                    pod.id,
                ),
            )
            conn.commit()


def add_k8s_standby_repmgr_task(data, task_id):
    """
    Registers a standby repmgr pod to an existing primary repmgr pod in a Kubernetes cluster.
    The function registers every standby, no matter whether it is the first or the third.

    Args:
        data (str): Configuration data required for registering the standby instance.
        task_id (int): Unique identifier for the task being executed.
    """
    task_name = 'add k8s standby repmgr task'

    try:
        data = json.loads(data)
        created_by = int(data["created_by"])

        cp = ControlPlane(json.loads(data["cp_data"]))
        cluster = PostgresCluster(cp, json.loads(data["cluster_data"]))
        primary_pod = PostgresPod(cluster, json.loads(data["primary_data"]))
        standby_pod = PostgresPod(cluster, json.loads(data["standby_data"]))

        logger = TaskLogger(task_id, task_name, created_by)
    except Exception as error:
        logging.error(f"{task_name}: {str(error)}")
        raise
        # return

    with logger.step("initializing task"):
        pass

    with logger.step("checking installation possibility"):
        pass

    with logger.step("updating configuration"):
        values = {
            "namespace": standby_pod.namespace,
            "cluster": cluster.name,
            "dir123Cluster": standby_pod.dir_123Cluster,
            "pod": {
                "name": standby_pod.name,
                "id": standby_pod.id,
                "externalIp": standby_pod.externalIp,
                "externalPort": standby_pod.externalPort,
                "nodeAffinity": standby_pod.node_name,
                "image": standby_pod.image,
            },
            "resources": {
                # not enough resources on VM
                # "memory": standby_pod.ram_gb,
                # "cpu": standby_pod.cpu_cores,
                "memory": "256Mi",
                "cpu": "250m",
                "storage": standby_pod.disk_gb,
            },
            "primary": {
                "externalIp": primary_pod.externalIp,
                "externalPort": primary_pod.externalPort,
            },
            "database": {
                "username": standby_pod.postgres.username,
                "password": standby_pod.postgres.password,
                "repmgrPassword": standby_pod.postgres.repmgr_password
            },
        }

        values = yaml.dump(values, default_flow_style=False)
        cp.exec_command(f"mkdir -p {standby_pod.helm.cp_chart_dir}")
        cp.upload_file(values, standby_pod.helm.cp_values_filepath)

    with logger.step("updating primary's and other standbys pg_hba.conf and pgpass"):
        # update the pg_hba.conf and .pgpass for each pod, except for the current standby
        filtered_pods = tuple(p for p in cluster.pods if p.name != standby_pod.name)
        for p in filtered_pods:
            p.add_pg_hba_entries(source_pod=standby_pod)
            p.add_pgpass_entries(source_pod=standby_pod)

    with logger.step("installing postgresql"):
        if not cp.is_namespace_exists(standby_pod.namespace):
            logger.error(f'Namespace "{standby_pod.namespace}" does not exist!', excode=1)

        cp.transfer_file(local_path=standby_pod.helm.apphost_chart_filepath, remote_path=standby_pod.helm.cp_chart_filepath)
        cp.exec_command(f"helm install 123cluster-{standby_pod.name} {standby_pod.helm.cp_chart_filepath} -f {standby_pod.helm.cp_values_filepath} -n {standby_pod.namespace}")

        if not standby_pod.wait_db_container_ready():
            return logger.error(f"Error during the installation of the pod.", excode=1)

        standby_pod.exec_command(f"/usr/bin/repmgr -f {standby_pod.dir_123Cluster}/configs/repmgr.conf standby register")

    with get_postgres_connection() as conn:
        with conn.cursor() as cur:
            cur.execute(
                """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""",
                (
                    logger.status,
                    Status.OK.value,
                    standby_pod.id,
                ),
            )
            conn.commit()


def drop_k8s_pg_cluster_task(data, task_id):
    task_name = "drop k8s postgres cluster task"

    try:
        data = json.loads(data)
        created_by = int(data["created_by"])

        cp = ControlPlane(json.loads(data["kubernetes_api_data"]))
        cluster = PostgresCluster(cp, json.loads(data["cluster_data"]))
        cluster_hosts_data = [json.loads(p) for p in data["cluster_hosts_data"]]
        cluster_pods_data = [json.loads(p) for p in data["cluster_pods_data"]]
        
        core_v1_api = cp.get_core_v1_api()
        custom_objects_api = cp.get_custom_objects_api()

        logger = TaskLogger(task_id, task_name, created_by)
    except Exception as error:
        logging.error(f"{task_name}: {str(error)}")
        return

    with logger.step("initializing task"):
        pass

    with logger.step("dropping cluster configuration"):
        custom_objects_api.delete_namespaced_custom_object(
            group="postgresql.cnpg.io",
            version="v1",
            namespace=cluster.namespace,
            plural="clusters",
            name=cluster.name
        )

    with logger.step("dropping superuser secret"):
        core_v1_api.delete_namespaced_secret(
            name=cluster.superuser_secret_name,
            namespace=cluster.namespace
        )

    with logger.step("task finished"):
        pass

    with get_postgres_connection() as conn:
        with conn.cursor() as cur:
            for p in cluster_pods_data:
                cur.execute(
                    """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""",
                    (
                        logger.status,
                        Status.DELETED.value,
                        p["id"],
                    ),
                )

            for h in cluster_hosts_data:
                cur.execute(
                    """UPDATE "clusterCore_c123_host" SET status=%s, ind_deleted='true' WHERE id=%s""",
                    (
                        Status.DELETED.value,
                        h["id"],
                    ),
                )


def drop_k8s_pg_pod_task(data, task_id):
    """drop instance"""

    task_name = "drop k8s postgres pod task"

    try:
        data = json.loads(data)
        created_by = int(data["created_by"])

        cp = ControlPlane(json.loads(data["cp_data"]))
        cluster = PostgresCluster(cp, json.loads(data["cluster_data"]))
        pod = PostgresPod(cluster, json.loads(data["pod_data"]))

        logger = TaskLogger(task_id, task_name, created_by)
    except Exception as error:
        logging.error(f"{task_name}: {str(error)}")
        return

    with logger.step("initializing task"):
        pass

    with logger.step("unregister repmgr node"):
        for p in cluster.pods:
            if p.is_primary and p.is_running:
                if pod.is_primary:
                    p.exec_command(f"repmgr primary unregister -f {p.dir_123Cluster}/configs/repmgr.conf --node-id={pod.id} --force")
                else:
                    p.exec_command(f"repmgr standby unregister -f {p.dir_123Cluster}/configs/repmgr.conf --node-id={pod.id}")
                break

    with logger.step("deleting helm chart"):
        cp.exec_command(f"helm uninstall 123cluster-{pod.name} -n {pod.namespace}")
        cp.exec_command(f"rm -rf {pod.helm.cp_chart_dir}")

    with logger.step("updating primary's and other standbys pg_hba.conf and pgpass"):
        # update the pg_hba.conf and .pgpass for each pod, except for the current
        filtered_pods = tuple(p for p in cluster.pods if p.name != pod.name)
        for p in filtered_pods:
            p.remove_pg_hba_entries(source_pod=pod)
            p.remove_pgpass_entries(source_pod=pod)

    with logger.step("deleting data directory"):
        pod.node.exec_command(f"rm -rf {pod.dir_123Cluster}")

    with get_postgres_connection() as conn:
        with conn.cursor() as cur:
            cur.execute(
                """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""",
                (
                    logger.status,
                    Status.DELETED.value,
                    pod.id,
                ),
            )

    with logger.step("task finished"):
        pass


def pg_k8s_dump_backup_task(data, task_id):
    # NOTE:
    # Variable names and OOP structure remain unchanged to ensure
    # easy code copying from the same task for the VM.
    # Please refactor this code after completing the task,
    # adapting it to proper naming conventions and OOP structure.
    task_name = "dump k8s postgres standalone task"

    try:
        data = json.loads(data)
        created_by = int(data["created_by"])
        logger = TaskLogger(task_id, task_name, created_by)
    except Exception as error:
        logging.error(f"{task_name}: {str(error)}")
        return

    with logger.step("initializing task"):
        pass

    with logger.step('checking parameters'):
        cp = ControlPlane(data["cp_data"])
        cluster = PostgresCluster(cp, json.loads(data["cluster_data"]))
        pod = PostgresPod(cluster, json.loads(data["pod_data"]))

        host = {}
        host["name"] = pod.externalIp
        node = json.loads(data['pod_data'])
        backup = json.loads(data['backup_data'])

        backup_host = data['backup_host']
        if not is_ip(backup_host['name']):
            backup_host["name"] = socket.gethostbyname(backup_host['name'])

        # backup_type = dumpall or dump
        backup_type = data['backup_type']
        dumpall = 'dumpall' in backup_type
        if not dumpall:
            db_data = json.loads(data['db_data'])

        # TODO: Fetch the actual PostgreSQL version (currently returns None)
        # postgres_version = int(node['version'])
        postgres_version = 17

        path = backup['ex_params'].get('backup_path')
        if path.startswith('~'):
            jsn = find_tilda(backup_host)
            if jsn['stderr'] != '':
                logger.error(jsn['stderr'], jsn['excode'])
                return

            tilda_path = jsn['stdout']
            path = path.replace('~', tilda_path)

        compress = backup['ex_params'].get('compress')
        plain = backup['ex_params'].get('plain')
        include_data = data['include_data']
        include_publications = data['include_publications']
        include_subscriptions = data['include_subscriptions']
        use_inserts = data['use_inserts']

        exclude_filter = []
        if not include_data:
            exclude_filter.append("--schema-only")
        if not include_publications:
            exclude_filter.append("--no-publications")
        if not include_subscriptions:
            exclude_filter.append("--no-subscriptions")

        backup_host_distributor, backup_host_release = check_os_version(backup_host)
        if OperatingSystem.INVALID.value in backup_host_distributor:
            jsn = backup_host_release
            logger.error(jsn['stderr'], jsn['excode'])
            return

        postgres_bin_backup_host = find_postgres_bin(backup_host, postgres_version, backup_host_distributor)
        remote_command(backup_host, f"sudo mkdir -p '{path}'", False)
        remote_command(backup_host, f"sudo chown -R postgres:postgres '{path}'", False)

        date = datetime.now().strftime('%d%m%Y_%H%M%S')

    # for -Fc use .dump
    if compress:
       path = f'{path}/{backup["name"]}_{date}.gz'
    else:
        if dumpall:
            path = f'{path}/{backup["name"]}_{date}.out'
        else:
            path = f'{path}/{backup["name"]}_{date}.dump'

    with logger.step("adding pg_hba.conf entries"):
        entries = (
            f"host    all             {node['database_user']:<15} {backup_host['name']}/32      md5\n"
        )
        pod.add_pg_hba_entries(entries=entries)

    with logger.step("running dump command") as step:
        # -Fc = format custom
        if plain:
            dump_format = "-Fp"
        else:
            dump_format = "-Ft" if compress else "-Fc"

        if not dumpall:
            # Example command: "export PGPASSFILE='/var/lib/pgsql/.pgpass';pg_dump -h localhost -U user123cluster -d cluster123 > /root/123cluster/postgres/dumps/cluster123_dump_file.dump"
            command = (
                f"sudo -iu postgres "
                f"PGPASSWORD='{node['database_pwd'].strip()}' "
                f"{postgres_bin_backup_host}/pg_dump {dump_format} "
                f"-h '{host['name']}' "
                f"-U '{node['database_user']}' "
                f"-d '{db_data['name']}' "
                f"-p '{node['port']}' "
                f"--serializable-deferrable "
                f"--quote-all-identifiers "
                f"-c "
                f"--if-exists "
                f"-C "
                f"-f '{path}' "
                f"{'--inserts' if use_inserts else ''} "
                f"{' '.join(exclude_filter)}"
            )

            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                logger.error(jsn['stderr'], jsn['excode'])
                return
        else:
            # Example command: sudo -iu postgres /usr/lib/postgresql/13/bin/pg_dumpall > /root/123cluster/postgres/dumps/cluster123.out"
            parts = [
                "sudo -iu postgres",
                f"PGPASSWORD='{node['database_pwd'].strip()}'",
                f"{postgres_bin_backup_host}/pg_dumpall",
                f"-h '{host['name']}'",
                f"-U '{node['database_user']}'",
                f"-p '{node['port']}'",
                "-c --if-exists --quote-all-identifiers"
            ]

            if use_inserts:
                parts.append('--inserts')

            if postgres_version > 11:
                parts.append("--exclude-database='repmgr'")

            parts.append(f"| sudo gzip -c | sudo tee '{path}'" if compress else f"-f '{path}'")
            parts += exclude_filter

            command = " ".join(parts)
            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                logger.error(jsn['stderr'], jsn['excode'])
                return

    remote_command(backup_host, f"sudo chown -R postgres:postgres '{path}'", False)

    with logger.step("checking the backup size"):
        jsn = json.loads(remote_command(backup_host, f"sudo du -b '{path}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        if jsn['stderr'] != '':
            logger.error(jsn['stderr'], jsn['excode'])
            return

        size = jsn['stdout'].strip()
        backup_size = float(size)/1024
        if backup_size < 0.01:
            backup_size = 0.01

    with logger.step("getting dump modified date") as step:
        modified_date = get_date_modified_folder(hostdata=backup_host, folder_path=path)
        if isinstance(modified_date, tuple):
            # assume the logger will automatically catch the exception inside the context manager block
            pass

    with logger.step("task finished") as step:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                if dumpall:
                    cur.execute(
                        """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""",
                        (
                            Severity.OK.value,
                            Status.OK.value,
                            node["id"],
                        ),
                    )
                # else:
                #     cur.execute(
                #         """UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""",
                #         (Severity.OK.value, Status.OK.value, db_data['id'],))

                cur.execute(
                    """UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(
                        backup_size, backup["id"]
                    )
                )
                # cur.execute(
                #     """UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{backup_path}}', '"{0}"', true) WHERE id={1}""".format(
                #         path, backup['id']))
                cur.execute(
                    """UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{backup_path}}', '"{0}"', true) WHERE id={1}""".format(
                        path, backup['id']))
                cur.execute(
                    """UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}', status='true' WHERE id=%s"""
                    % (modified_date, backup["id"])
                )
                conn.commit()


def pg_k8s_dumpall_restore_task(data, task_id):
    task_name = "restore k8s postgres standalone task"
    try:
        data = json.loads(data)
        created_by = int(data["created_by"])

        cp = ControlPlane(data["cp_data"])
        cluster = PostgresCluster(cp, json.loads(data["cluster_data"]))
        pod = PostgresPod(cluster, json.loads(data["pod_data"]))

        logger = TaskLogger(task_id, task_name, created_by)

        host = {"name": pod.externalIp}
        node_data = json.loads(data["pod_data"])
        host_data = data["cp_data"]
        backup = json.loads(data["export"])
        compress = backup["ex_params"].get("compress")
        # host_ip = host_data['name']
        node_ip = node_data['name'].split(':')[0]

        backup_host = data['backup_host']
        if not is_ip(backup_host['name']):
            backup_host["name"] = socket.gethostbyname(backup_host['name'])

        # pg_user = node_data['database_user']
        # pg_password = node_data['database_pwd']
        pg_port = int(node_data['port'])
        # TODO: Fetch the actual PostgreSQL version (currently returns None)
        # postgres_version = int(node_data['version'])
        postgres_version = 17

        path = backup['ex_params'].get('backup_path')
        if path.startswith('~'):
            jsn = find_tilda(backup_host)
            if jsn['stderr'] != '':
                logger.error(jsn['stderr'], jsn['excode'])
                return

            tilda_path = jsn['stdout']
            path = path.replace('~', tilda_path)


        pod_name = re.sub(r"^[^a-z0-9]+|[^a-z0-9]+$", "", re.sub(r"[.:]", "-", f"postgres-{node_data['name']}"),
                          flags=re.IGNORECASE)
        with get_postgres_connection() as conn:
            with conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as cur:
                cur.execute("""SELECT * FROM "clusterCore_c123_cluster" WHERE id =%s""", (node_data["cluster_id"],))
                cluster = cur.fetchone()
                namespace = cluster["ex_params"].get("namespace")

        backup_host_distributor, backup_host_release = check_os_version(backup_host)
        if OperatingSystem.INVALID.value in backup_host_distributor:
            jsn = backup_host_release
            logger.error(jsn['stderr'], jsn['excode'])
            return

        postgres_bin_backup_host = find_postgres_bin(backup_host, postgres_version, backup_host_distributor)
        remote_command(backup_host, f"sudo mkdir -p '{path}'", False)
        remote_command(backup_host, f"sudo chown -R postgres:postgres '{path}'", False)



    except Exception as error:
        logging.error(f"{task_name}: {str(error)}")
        return

    with logger.step("initializing task"):
        pprint(data)

    with logger.step("adding pg_hba.conf entries"):
        entries = (
            f"host    all             {node_data['database_user']:<15} {backup_host['name']}/32      md5\n"
        )
        pod.add_pg_hba_entries(entries=entries)

    with logger.step("restore dump all") as step:
        # if compress:
        #     decompress_command = "gunzip -c /123cluster/postgres/illia-pgdata/backups/exports/backup.gz > /123cluster/postgres/illia-pgdata/backups/exports/"
        #     pod.exec_command(command=decompress_command)
        #
        #     restore_command = f"psql -U postgres -p {pg_port} -d postgres -f /123cluster/postgres/illia-pgdata/backups/exports/"
        #     pod.exec_command(command=restore_command)
        # else:
        #     restore_command = f"psql -U postgres -p {pg_port} -d postgres -f /123cluster/postgres/illia-pgdata/backups/exports/backup.out"
        #     pod.exec_command(command=restore_command)
        command = (f"sudo -iu postgres PGPASSWORD='{node_data['database_pwd'].strip()}' "
                   f"psql -h '{host['name']}' "
                   f"-U '{node_data['database_user']}' "
                   f"-d postgres "
                   f"-p '{node_data['port']}' < '{path}'") if not compress else (f"sudo gunzip < '{path}' | "
                                                                                 f"sudo -iu postgres PGPASSWORD='{node_data['database_pwd'].strip()}' "
                                                                                 f"psql -h '{host['name']}' "
                                                                                 f"-U '{node_data['database_user']}' "
                                                                                 f"-d postgres "
                                                                                 f"-p '{node_data['port']}'")


        jsn = json.loads(remote_command(backup_host, command, True), strict=False)

        if jsn['stdout'] != '':
            print(jsn['stdout'])

        ignorable_errors = [
            "cannot drop role postgres because it is required by the database system",
            "current user cannot be dropped",
            "role \"postgres\" already exists",
            "role \"root\" already exists",
            "database \"repmgr\" is being accessed by other users",
            "role \"repmgr\" cannot be dropped because some objects depend on it",
            "role \"repmgr\" already exists",
        ]

        if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
            errors = [line.strip() for line in jsn['stderr'].split('\n') if 'ERROR:' in line]

            if errors:
                # Identify errors that can be ignored
                ignored = [error for error in errors if any(ignorable in error for ignorable in ignorable_errors)]
                # Identify errors that should not be ignored
                non_ignorable_errors = [error for error in errors if
                                        not any(ignorable in error for ignorable in ignorable_errors)]

                # Log ignored errors for transparency
                if ignored:
                    pprint("Ignored expected errors:\n" + '\n'.join(ignored))

                # Fail the script if there are non-ignorable errors
                if non_ignorable_errors:
                    logger.error("Restore failed with critical errors:\n" + '\n'.join(non_ignorable_errors),
                                 jsn['excode'])
                    return

    with logger.step("check dump size") as step:
        size_command = f"du -sb '{node_data['dir_123Cluster']}' | awk '{{print $1}}'"
        jsn = pod.exec_command(command=size_command)
        size = jsn.strip()
        node_size = float(size)

    with logger.step("task finished") as step:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute(
                    """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""",
                    (Severity.OK.value, Status.OK.value, node_data['id'],))
                conn.commit()
                db_set_from_app = {}
                # for clarity - call dict_cur
                with conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as dict_cur:
                    dict_cur.execute(
                        """SELECT DISTINCT ON (name) id, name FROM "clusterCore_c123_database" WHERE node_id=%s""",
                        (node_data['id'],))
                    db_list_from_app = dict_cur.fetchall()
                    db_set_from_app = {db['name'].strip() for db in db_list_from_app}
                # TODO: pod execute command psql -c instead of conn_node or remote command if need exec command from backup host
                # conn_node = connect_to_postgres_database(node_ip, 'postgres', node_data['database_user'],
                #                                 node_data['database_pwd'].strip(), node_data['port'])


                cur.execute(
                    """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""",
                    (
                        Severity.OK.value,
                        Status.OK.value,
                        node_size,
                        node_data["id"],
                    ),
                )



def pg_k8s_basebackup_task(data, task_id):
    task_name = 'basebackup k8s postgres standalone task'
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])

        cp = ControlPlane(data["cp_data"])
        cluster = PostgresCluster(cp, json.loads(data["cluster_data"]))
        pod = PostgresPod(cluster, json.loads(data["pod_data"]))

        logger = TaskLogger(task_id, task_name, created_by)

        host = {"name": pod.externalIp}
        node_data = json.loads(data['pod_data'])
        host_data = data['cp_data']
        backup = json.loads(data['backup_data'])
        backup_host = data['backup_host']
        if not is_ip(backup_host['name']):
            backup_host["name"] = socket.gethostbyname(backup_host['name'])
        is_sched = backup['is_sched']
        compress = backup['ex_params'].get('compress')
        incremental = backup['ex_params'].get('incremental')
        incremental_physical = backup['ex_params'].get('incremental_physical')
        backup_path = backup['ex_params'].get('backup_path')
        if backup_path.startswith('~'):
            jsn = find_tilda(backup_host)
            if jsn['stderr'] != '':
                logger.error(jsn['stderr'], jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)
        node_ip = node_data['name'].split(':')[0]
        # host_ip = host_data['name']

        # pg_user = node_data['database_user']
        # pg_password = node_data['database_pwd']
        pg_port = int(node_data['port'])

        # TODO: Fetch the actual PostgreSQL version (currently returns None)
        # postgres_version = int(node_data['version'])
        postgres_version = 17
        if postgres_version < 17:
            incremental_physical = False

        log_archive_path = backup['ex_params'].get('log_archive_path')

        pod_name = re.sub(r"^[^a-z0-9]+|[^a-z0-9]+$", "", re.sub(r"[.:]", "-", f"postgres-{node_data['name']}"),
                          flags=re.IGNORECASE)
        with get_postgres_connection() as conn:
            with conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as cur:
                cur.execute("""SELECT * FROM "clusterCore_c123_cluster" WHERE id =%s""", (node_data["cluster_id"],))
                cluster = cur.fetchone()
                namespace = cluster["ex_params"].get("namespace")

        backup_host_distributor, backup_host_release = check_os_version(backup_host)
        if OperatingSystem.INVALID.value in backup_host_distributor:
            jsn = backup_host_release
            logger.error(jsn['stderr'], jsn['excode'])
            return

        postgres_bin_backup_host = find_postgres_bin(backup_host, postgres_version, backup_host_distributor)
        remote_command(backup_host, f"sudo mkdir -p '{backup_path}'", False)
        remote_command(backup_host, f"sudo chown -R postgres:postgres '{backup_path}'", False)

        backup_name = backup['name']

    except Exception as error:
        logging.error(f"{task_name}: {str(error)}")
        return

    with logger.step('initializing task'):
        pass

    with logger.step("adding pg_hba.conf entries"):
        entries = (
            f"host    all             {node_data['database_user']:<15} {backup_host['name']}/32      md5\n"
        )
        pod.add_pg_hba_entries(entries=entries)

    if incremental_physical:
        with logger.step('Getting system_identifier for storing backups') as step:
            conn = connect_to_postgres_database(node_ip, 'postgres', node_data['database_user'],
                                                node_data['database_pwd'].strip(), node_data['port'])
            with conn:
                with conn.cursor() as cur:
                    cur.execute("SELECT system_identifier FROM pg_control_system()")
                    system_identifier = cur.fetchone()[0]
                    backup_path = f"{backup_path}/{system_identifier}"

    with logger.step('Getting system_identifier for storing backups') as step:
        new_basebackup = True
        with get_postgres_connection() as conn:
            with conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as cur:
                cur.execute("""SELECT * FROM "clusterCore_c123_cluster" WHERE id =%s""", (node_data["cluster_id"],))
                cluster = cur.fetchone()
                new_basebackup = cluster["ex_params"].get("new_basebackup")

        if isinstance(incremental, str) and len(incremental) > 0 and not new_basebackup:
            backup_path = incremental

    with logger.step('check log_archive_path directory exist') as step:
        # pod.exec_command(command=f"ls {log_archive_path} || mkdir -p {log_archive_path}")
        remote_command(data=backup_host, cmd=f"sudo ls {log_archive_path} || sudo mkdir -p {log_archive_path}",
                       wait_for_output=False)

    with logger.step('check backup directory exist') as step:
        # pod.exec_command(command=f"ls {backup_path} || mkdir -p {backup_path}")
        remote_command(data=backup_host, cmd=f"sudo ls {backup_path} || sudo mkdir -p {backup_path}",
                       wait_for_output=False)

    with logger.step('Check if there is an old incremental backup') as step:
        subfolder = datetime.utcnow().strftime('%d%m%Y%H%M%S')
        command = f"find {backup_path} -mindepth 1 -maxdepth 1 -type d -exec stat -c \"%Y %n\" {{}} \; | sort -n | tail -1 | awk '{{print $2}}'"
        jsn = pod.exec_command(command=command)

        last_backup_folder = pod.exec_command(
            f"find {backup_path} -mindepth 1 -maxdepth 1 -type d "
            f"-exec stat -c \"%Y %n\" {{}} \; | sort -n | tail -1 | awk '{{print $2}}'"
        )

        if jsn:
            incremental_folder_empty = False
        else:
            incremental_folder_empty = True

        if not incremental_folder_empty and incremental_physical:
            backup_path = f"{backup_path}/{subfolder}"
            pod.exec_command(command=f"mkdir -p {backup_path}")


    with logger.step('Ensuring wal_level is correct') as step:
        try:
            with conn:
                with conn.cursor() as cur:
                    cur.execute("SHOW wal_level")
                    wal_level = cur.fetchone()[0]
                    if wal_level.strip().lower() == 'minimal':
                        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                        cur.execute("ALTER SYSTEM SET wal_level=replica")
                        cur.execute("SELECT pg_reload_conf()")
                        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
        except Exception as error:
            logger.error(f"Error wal_level: {error}", excode=1)

    with logger.step('Ensuring archive_mode is on') as step:
        try:
            with conn:
                with conn.cursor() as cur:
                    cur.execute("SHOW archive_mode")
                    archive_mode = cur.fetchone()[0]
                    if archive_mode.strip().lower() != 'on':
                        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                        cur.execute("ALTER SYSTEM SET archive_mode=on")
                        cur.execute("SELECT pg_reload_conf()")
                        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
        except Exception as error:
            logger.error(f"Error archive_mode: {error}", excode=1)

    with logger.step('Turn on summarize_wal') as step:
        try:
            with conn.cursor() as cur:
                cur.execute("SHOW summarize_wal")
                row = cur.fetchone()[0]
                if str(row).strip().lower() != 'on':
                    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                    cur.execute("ALTER SYSTEM SET summarize_wal=on")
                    cur.execute("SELECT pg_reload_conf()")
                    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
        except Exception as error:
            print(f"Error summarize_wal: {error}")

    with logger.step('Defining parameters for basebackup') as step:
        physical_incremental = f"-i '{incremental_physical}/backup_manifest'" if incremental_physical and not incremental_folder_empty else ''

    # TODO: Add check on empty backup folder
    if is_sched and incremental and not incremental_physical:
        with logger.step('Run backup') as step:
            try:
                with conn.cursor() as cur:
                    cur.execute("SELECT pg_current_wal_lsn()")
                    pg_current_wal_lsn = cur.fetchone()[0]
            except Exception as error:
                logger.error(f"Error retrieving WAL LSN: {error}", excode=1)
                raise

        wal_location = f"{backup_path}/pg_wal"
        backup_command = f"""
            pg_receivewal -h {node_ip} -U repmgr -p {pg_port} -D '{wal_location}' -n -E {pg_current_wal_lsn}
        """
        compress = False
    else:
        backup_format = '-Fp' if not compress else '-Ft -z'
        physical_incremental_param = (
            f"-i '{last_backup_folder}/backup_manifest'"
            if is_sched and incremental_physical and not incremental_folder_empty
            else ''
        )

        backup_command = f"""
            pg_basebackup -h {node_ip} -U repmgr -p {pg_port} -D '{backup_path}' {backup_format} -Xs -P -R {physical_incremental_param}
        """

    pod.exec_command(backup_command)

    # TODO: rsync don't work inside pod, need to find how copy pg_wal file in log_archive_path inside pod
    # with logger.step('Installing a command in PostgreSQL') as step:
    #     cur = conn.cursor()
    #     archive_command = f'''rsync -az --ignore-existing -e "ssh -p {host_data['ssh_port']} -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null" %p postgres@{host_data['name']}:{log_archive_path}'''
    #     conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    #     cur.execute(f"ALTER SYSTEM SET archive_command='{archive_command}'")
    #     conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)

    with logger.step("check dump size") as step:
        size_command = f"du -bs '{backup_path}' | awk '{{print $1}}'"
        jsn = pod.exec_command(command=size_command)
        size = jsn.strip()
        backup_size = float(size)/1024


    with logger.step('task finished') as step:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity='%s', status='%s', ind_started='true', ind_deleted='false' WHERE id=%s""" % (Severity.OK.value, Status.OK.value, node_data['id']))
                cur.execute(f"""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{incremental_physical}}', '"{backup_path}"', true) WHERE id={1}""")
                cur.execute(
                    """UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(
                        backup_size, backup['id']))
                # cur.execute(
                #     """UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{duration}}', '{0}', true) WHERE id={1}""".format(
                #         duration_task_end, backup['id']))
                conn.commit()




def pg_k8s_pod_restore_task(data, task_id):
    task_name = 'restore basebackup k8s postgres standalone task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])

        node_data = json.loads(data['node_data'])
        host_data = data['host_data']
        backup = json.loads(data['backup'])
        backup_path = f"/123cluster{backup['ex_params'].get('backup_path')}"

        pg_data_path = node_data['dir_123Cluster']
        node_ip = node_data['name'].split(':')[0]
        pg_port = int(node_data['port'])

        pod_name = re.sub(r"^[^a-z0-9]+|[^a-z0-9]+$", "", re.sub(r"[.:]", "-", f"postgres-{node_data['name']}"),
                          flags=re.IGNORECASE)
        with get_postgres_connection() as conn:
            with conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as cur:
                cur.execute("""SELECT * FROM "clusterCore_c123_cluster" WHERE id =%s""", (node_data["cluster_id"],))
                cluster = cur.fetchone()
                namespace = cluster["ex_params"].get("namespace")
    except Exception as error:
        logging.error(f"{task_name}: {str(error)}")
        return

    logger = TaskLogger(task_id, task_name, created_by)
    cp = ControlPlane(host_data)

    with logger.step('initializing task') as step:
        pass

    with logger.step('stop standbys') as step:
        pass

    with logger.step('stop primary') as step:
        command_stop_primary_pg_server = f"pg_ctl -D {pg_data_path} stop"
        # cp.pod_execute_command(pod_name, command_stop_primary_pg_server)

    with logger.step('delete old data') as step:
        command_delete_old_data = f"rm -rf {pg_data_path}/*"

    with logger.step('copy basebackup to primary') as step:
        command_delete_old_data = f"cp -r {backup_path}/* {pg_data_path}"
        pprint(data)

    with logger.step('check recovery.signal or standby.signal') as step:
        command_check = f"touch {pg_data_path}/recovery.signal"

    with logger.step('start primary') as step:
        command_start_primary_pg_server = f"pg_ctl -D {pg_data_path} start"

    with logger.step('Clear and recreate Standby') as step:
        # kubectl delete pvc <standby-pvc>
        # kubectl scale statefulset <standby-sts-name> --replicas=1
        pass


def create_pg_k8s_database_task(data, task_id):
    task_name = 'create pg database task'
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])

        cp = ControlPlane(json.loads(data["cp_data"]))
        cluster = PostgresCluster(cp, json.loads(data["cluster_data"]))
        pod = PostgresPod(cluster, json.loads(data["pod_data"]))

        logger = TaskLogger(task_id, task_name, created_by)

        pg_node = json.loads(data['pod_data'])
        host_data = json.loads(data['cp_data'])
        db = json.loads(data['db'])
        db_name = db['name']
        node_ip = pg_node['name'].split(':')[0]
        pg_port = pg_node['port']

        pod_name = re.sub(r"^[^a-z0-9]+|[^a-z0-9]+$", "", re.sub(r"[.:]", "-", f"postgres-{pg_node['name']}"),
                          flags=re.IGNORECASE)
        with get_postgres_connection() as conn:
            with conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as cur:
                cur.execute("""SELECT * FROM "clusterCore_c123_cluster" WHERE id =%s""", (pg_node["cluster_id"],))
                cluster = cur.fetchone()
                namespace = cluster["ex_params"].get("namespace")

    except Exception as error:
        logging.error(f"{task_name}: {str(error)}")
        return



    with logger.step('initializing task') as step:
        pass


    with logger.step('creating the database') as step:
        query = f"CREATE DATABASE {db_name};"
        command = f"psql -U postgres -p {pg_port} -d postgres -c \"{query}\""
        pod.exec_command(command)


    with logger.step('creating database users') as step:
        owner_user_name = f'{db_name}_owner_user'
        write_user_name = f'{db_name}_write_user'
        read_user_name = f'{db_name}_read_user'
        owner_user_pwd = generate_complex_password()
        read_user_pwd = generate_complex_password()
        write_user_pwd = generate_complex_password()
        logs_output = f'USERNAME = {owner_user_name}\nPASSWORD = {owner_user_pwd}\nUSERNAME = {read_user_name}\nPASSWORD = {read_user_pwd}\nUSERNAME = {write_user_name}\nPASSWORD = {write_user_pwd}'
        owner_user_pwd = format_for_shell(owner_user_pwd)
        write_user_pwd = format_for_shell(write_user_pwd)
        read_user_pwd = format_for_shell(read_user_pwd)
        alter_user_query = "ALTER USER %s WITH LOGIN NOCREATEDB NOSUPERUSER NOREPLICATION PASSWORD '%s'"
        create_user_query = "CREATE USER %s WITH LOGIN NOCREATEDB NOSUPERUSER NOREPLICATION PASSWORD '%s'"

        user_queries = [
            f"CREATE USER {owner_user_name} WITH LOGIN NOCREATEDB NOSUPERUSER NOREPLICATION PASSWORD '{owner_user_pwd}';",
            f"CREATE USER {write_user_name} WITH LOGIN NOCREATEDB NOSUPERUSER NOREPLICATION PASSWORD '{write_user_pwd}';",
            f"CREATE USER {read_user_name} WITH LOGIN NOCREATEDB NOSUPERUSER NOREPLICATION PASSWORD '{read_user_pwd}';",
        ]

        permission_queries =[
                f"ALTER DATABASE {db_name} OWNER TO {owner_user_name};",

                f"GRANT CONNECT ON DATABASE {db_name} TO {read_user_name};",
                f"GRANT USAGE ON SCHEMA public TO {read_user_name};",
                f"GRANT SELECT ON ALL TABLES IN SCHEMA public TO {read_user_name};",
                f"GRANT SELECT ON ALL SEQUENCES IN SCHEMA public TO {read_user_name};",
                f"ALTER DEFAULT PRIVILEGES IN SCHEMA public GRANT SELECT ON TABLES TO {read_user_name};",
                f"ALTER DEFAULT PRIVILEGES IN SCHEMA public GRANT SELECT ON SEQUENCES TO {read_user_name};",

                f"GRANT CONNECT ON DATABASE {db_name} TO {write_user_name};",
                f"GRANT USAGE, CREATE ON SCHEMA public TO {write_user_name};",
                f"GRANT SELECT, INSERT, UPDATE, DELETE ON ALL TABLES IN SCHEMA public TO {write_user_name};",
                f"GRANT USAGE ON ALL SEQUENCES IN SCHEMA public TO {write_user_name};",
                f"ALTER DEFAULT PRIVILEGES IN SCHEMA public GRANT SELECT, INSERT, UPDATE, DELETE ON TABLES TO {write_user_name};",
                f"ALTER DEFAULT PRIVILEGES IN SCHEMA public GRANT USAGE ON SEQUENCES TO {write_user_name};"
            ]


        for query in user_queries + permission_queries:
            command = f"psql -U postgres -p {pg_port} -d {db_name} -c \"{query}\""
            pod.exec_command(command)

        get_size_query = f"SELECT pg_database_size('{db_name}');"
        command = f"psql -U postgres -p {pg_port} -d {db_name} -t -c \"{get_size_query}\""

        res = pod.exec_command(command).strip()
        size = float(res) if res and res.isdigit() else 0.0

        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute(
                        """UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""",
                        (Severity.OK.value, Status.OK.value, size, db['id'],))
                    conn.commit()
        except Exception as error:
            return

    with logger.step('task finished') as step:
        pass

def pg_k8s_restore_DB_task(data, task_id):
    task_name = 'restore k8s postgres'
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])

        cp = ControlPlane(data["cp_data"])
        cluster = PostgresCluster(cp, json.loads(data["cluster_data"]))
        pod = PostgresPod(cluster, json.loads(data["pod_data"]))

        logger = TaskLogger(task_id, task_name, created_by)

        host = {"name": pod.externalIp}
        node_data = json.loads(data['pod_data'])
        db_data = json.loads(data['db_data'])
        db_name = db_data['name']
        host_data = data['cp_data']
        backup = json.loads(data['backup'])
        backup_host = data['backup_host']
        is_backup_host_ip = is_ip(backup_host['name'])
        if not is_backup_host_ip:
            backup_host_ip = socket.gethostbyname(backup_host['name'])

        compress = backup['ex_params'].get('compress')
        plain = backup['ex_params'].get('plain')
        incremental = backup['ex_params'].get('incremental')
        incremental_physical = backup['ex_params'].get('incremental_physical')
        backup_path = backup['ex_params'].get('backup_path')
        if backup_path.startswith('~'):
            jsn = find_tilda(backup_host)
            if jsn['stderr'] != '':
                logger.error(jsn['stderr'], jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)
        node_ip = node_data['name'].split(':')[0]
        # host_ip = host_data['name']

        # pg_user = node_data['database_user']
        # pg_password = node_data['database_pwd']
        pg_port = int(node_data['port'])

        # TODO: Fetch the actual PostgreSQL version (currently returns None)
        # postgres_version = int(node_data['version'])
        postgres_version = 17
        if postgres_version < 17:
            incremental_physical = False

        log_archive_path = backup['ex_params'].get('log_archive_path')

        pod_name = re.sub(r"^[^a-z0-9]+|[^a-z0-9]+$", "", re.sub(r"[.:]", "-", f"postgres-{node_data['name']}"),
                          flags=re.IGNORECASE)
        with get_postgres_connection() as conn:
            with conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as cur:
                cur.execute("""SELECT * FROM "clusterCore_c123_cluster" WHERE id =%s""", (node_data["cluster_id"],))
                cluster = cur.fetchone()
                namespace = cluster["ex_params"].get("namespace")

        backup_host_distributor, backup_host_release = check_os_version(backup_host)
        if OperatingSystem.INVALID.value in backup_host_distributor:
            jsn = backup_host_release
            logger.error(jsn['stderr'], jsn['excode'])
            return

        postgres_bin_backup_host = find_postgres_bin(backup_host, postgres_version, backup_host_distributor)
        remote_command(backup_host, f"sudo mkdir -p '{backup_path}'", False)
        remote_command(backup_host, f"sudo chown -R postgres:postgres '{backup_path}'", False)

        backup_name = backup['name']

        include_data = data['include_data']
        include_publications = data['include_publications']
        include_subscriptions = data['include_subscriptions']
        exclude_filter = []
        if not include_data:
            exclude_filter.append("--schema-only")
        if not include_publications:
            exclude_filter.append("--no-publications")
        if not include_subscriptions:
            exclude_filter.append("--no-subscriptions")

    except Exception as error:
        logging.error(f"{task_name}: {str(error)}")
        return

    with logger.step('initializing task'):
        pprint(data)

    with logger.step("adding pg_hba.conf entries"):
        entries = (
            f"host    all             {node_data['database_user']:<15} {backup_host['name']}/32      md5\n"
        )
        pod.add_pg_hba_entries(entries=entries)


    with logger.step("drop database") as step:
        terminate_query = f"SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname='{db_name}'"

        if postgres_version < 13:
            drop_query = f"DROP DATABASE IF EXISTS {db_name};"
        else:
            drop_query = f"DROP DATABASE IF EXISTS {db_name} WITH (FORCE);"

        terminate_command = f"psql -U postgres -p {pg_port} -d postgres -c \"{terminate_query}\""
        pod.exec_command(terminate_command)

        drop_command = f"psql -U postgres -p {pg_port} -d postgres -c \"{drop_query}\""
        pod.exec_command(drop_command)

    # TODO: ask about error
    with logger.step("preparing target for restore") as step:
        pass
        # pprint(f"DBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB NAMEEEEEEEEEEEEEE: {db_name}")
        # create_users_query = f"""
        # DO
        # $$
        # BEGIN
        #    IF NOT EXISTS (
        #       SELECT FROM pg_catalog.pg_user WHERE usename = '{db_name}_owner_user'
        #    ) THEN
        #       CREATE USER {db_name}_owner_user LOGIN PASSWORD '{db_name}_owner_user';
        #    END IF;
        #
        #    IF NOT EXISTS (
        #       SELECT FROM pg_catalog.pg_user WHERE usename = '{db_name}_read_user'
        #    ) THEN
        #       CREATE USER {db_name}_read_user LOGIN PASSWORD '{db_name}_read_user';
        #    END IF;
        #
        #    IF NOT EXISTS (
        #       SELECT FROM pg_catalog.pg_user WHERE usename = '{db_name}_write_user'
        #    ) THEN
        #       CREATE USER {db_name}_write_user LOGIN PASSWORD '{db_name}_write_user';
        #    END IF;
        # END
        # $$;
        # """
        #
        # queries = [
        #     create_users_query,
        #     "SHOW config_file"
        # ]
        #
        # for query in queries:
        #     command = f"psql -U postgres -p {pg_port} -d postgres -c \"{query}\""
        #     pod.exec_command(command)

            # commands = [
            #     f"sudo sed -i '/host\s*all\s*{node_data['database_user']}\s*{backup_host['name']}{chr(92) + '/32' if is_backup_host_ip else ''}\s*md5/d' '{config_dir}/pg_hba.conf'",
            #     f"echo -e 'host\tall\t{node_data['database_user']}\t{backup_host['name']}{'/32' if is_backup_host_ip else ''}\tmd5' | sudo tee -a '{config_dir}/pg_hba.conf'",
            # ]
            # if not is_backup_host_ip:
            #     commands += [
            #         f"sudo sed -i '/host\s*all\s*{node_data['database_user']}\s*{backup_host_ip}\/32\s*md5/d' '{config_dir}/pg_hba.conf'",
            #         f"echo -e 'host\tall\t{node_data['database_user']}\t{backup_host_ip}/32\tmd5' | sudo tee -a '{config_dir}/pg_hba.conf'",
            #     ]
            # for command in commands:
            #     jsn = json.loads(remote_command(host, command, True), strict=False)

    with logger.step("restoring the database") as step:
        if plain:
            command = f"sudo -iu postgres PGPASSWORD='{node_data['database_pwd'].strip()}' psql -h '{host['name']}' -U '{node_data['database_user']}' -d postgres -p '{node_data['port']}' < '{backup_path}'"
        else:
            dump_format = "-Ft" if compress else "-Fc"
            # Example command: "export PGPASSFILE='/var/lib/pgsql/.pgpass';psql -h localhost -U user123cluster -d cluster123 < /root/123cluster/postgres/dumps/cluster123_dump_file.txt"
            command = f"sudo -iu postgres PGPASSWORD='{node_data['database_pwd'].strip()}' {postgres_bin_backup_host}/pg_restore -c --if-exists -C {dump_format} -h '{host['name']}' -U '{node_data['database_user']}' -d postgres -p '{node_data['port']}' '{backup_path}' {' '.join(exclude_filter)}"

        jsn = json.loads(remote_command(backup_host, command, True), strict=False)

        if jsn['stdout'] != '':
            print(jsn['stdout'])

    with logger.step("check dump size") as step:
        size_command = f"du -sb '{node_data['dir_123Cluster']}' | awk '{{print $1}}'"
        jsn = pod.exec_command(command=size_command)
        size = jsn.strip()
        node_size = float(size)

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_data['id'],))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node_data['id'],))
                conn.commit()

    with logger.step("task finished") as step:
        pass


def drop_k8s_pg_database_task(data, task_id):
    task_name = 'drop k8s postgres DB'
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])

        cp = ControlPlane(data["cp_data"])
        cluster = PostgresCluster(cp, json.loads(data["cluster_data"]))
        pod = PostgresPod(cluster, json.loads(data["pod_data"]))

        logger = TaskLogger(task_id, task_name, created_by)

        host = {"name": pod.externalIp}
        node_data = json.loads(data['pod_data'])
        db_data = json.loads(data['db_data'])
        db = json.loads(data['db_data'])
        db_name = db_data['name']

        pg_port = int(node_data['port'])

        # TODO: Fetch the actual PostgreSQL version (currently returns None)
        # postgres_version = int(node_data['version'])
        postgres_version = 17






    except Exception as error:
        logging.error(f"{task_name}: {str(error)}")
        return

    with logger.step('initializing task'):
        pprint(data)


    with logger.step("dropping the database"):
        terminate_query = f"SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname='{db_name}'"

        if postgres_version < 13:
            drop_query = f"DROP DATABASE IF EXISTS {db_name};"
        else:
            drop_query = f"DROP DATABASE IF EXISTS {db_name} WITH (FORCE);"

        terminate_command = f"psql -U postgres -p {pg_port} -d postgres -c \"{terminate_query}\""
        pod.exec_command(terminate_command)

        drop_command = f"psql -U postgres -p {pg_port} -d postgres -c \"{drop_query}\""
        pod.exec_command(drop_command)

    with logger.step("check dump size") as step:
        size_command = f"du -sb '{node_data['dir_123Cluster']}' | awk '{{print $1}}'"
        jsn = pod.exec_command(command=size_command)
        size = jsn.strip()
        node_size = float(size)

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node_data['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DROPPED.value, db['id'],))
                conn.commit()

    with logger.step("task finished") as step:
        pass


