# BELONG TO SERVICE TASK
import json, time, uuid, pytz, os, re, logging
from socket import gethostbyname
from datetime import datetime
from psycopg2.extras import RealDictCursor
from elasticsearch import Elasticsearch
from elasticsearch.exceptions import NotFoundError, RequestError
from enums import OperatingSystem, SupportedOS, DBType, ClusterRoles, RestoreOptions, Status, Severity, Inventory, Schedule, LicenseType, Repository, Timeout
from tasks_core import insert_task, transfer_text_ram, remote_command, create_ssh_connection, get_postgres_connection, logger, init_container,\
     check_os_version, get_date_modified_folder, update_failed_object, install_offline_remote_to_remote, generate_complex_password, is_ram_enough, \
     TASK_TMP_FILES_DIR, transfer_files_rsync, get_log_rotate_config

#####################################################################
# ELASTICSEARCH TASKS 
#####################################################################

def check_elastic_version_exists(host, elastic_version, distributor):
    cmd = f"sudo rpm -qa | grep -E '^elasticsearch-{elastic_version}\.[0-9]+\.[0-9]'" if OperatingSystem.CENTOS.value in distributor else f"sudo dpkg -l | grep -E '^ii\s+elasticsearch' | grep '{elastic_version}.'"
    jsn = json.loads(remote_command(host, cmd, True), strict=False)
    return jsn['stdout'] != ''

def wait_for_being_writable(client, write_timeout = 120):
    "healthcheck function - wait until cluster health become yellow or green, means we can safely insert data to it"
    
    error = ''
    loaded_time = time.time()
    while time.time() - loaded_time < write_timeout:
        try:
            status = str(client.cluster.health()['status'])
            if status in [Severity.OK.value, Severity.WARNING.value]:
                return status
            
            time.sleep(1)
        except Exception as e:
            error = str(e)
    
    return error if error else status

def post_start_cleanup(hostname, port, username, password, roles):
    "revert the safety locks of prepare_to_stop_node function"
     
    jsn = {'stdout': '', 'stderr': '', 'excode': 0}
     
    try:
        es_client = connect_to_elasticsearch_insecure(hostname, username, password, port)
        if isinstance(es_client, tuple):
            jsn['stderr'] = str(es_client[1])
            jsn['excode'] = 2
            return jsn
        
        node_name = f"{hostname.strip()}_{port.strip()}"
        cluster_data = es_client.cat.nodes(h="name,id,pid", full_id=True, format="json")
        node_data =  next((i for i in cluster_data if node_name == i["name"].strip()), None)
        if not node_data:
            jsn['stderr'] = f"error: could not find {node_name} in cluster stats"
            return jsn

        if roles and "data" in roles:
            es_client.cluster.put_settings(persistent={
                "cluster.routing.allocation.enable": None
            })
        es_client.ml.set_upgrade_mode(enabled=False)
    except Exception as e:
        jsn['stderr'] = e.message if hasattr(e, 'message') else str(e)
        jsn['excode'] = 3
        return jsn
        
    try:
        es_client.shutdown.delete_node(node_id=node_data.get("id"))
    except NotFoundError:
        logging.warning(f"{node_name} is not in shutdown sequence")
    except Exception as e:
        jsn['stderr'] = e.message if hasattr(e, 'message') else str(e)
        jsn['excode'] = 4
        return jsn

    return jsn

def connect_to_elasticsearch_insecure(host, user, pwd, prt=9200, timeout=Timeout.SERVICE_ELASTIC.value):
    """connecting to elasticsearch cluster"""
    
    ex = 'unknown error'
    try:
        client = Elasticsearch(f"https://{host}:{prt}", basic_auth=(user, pwd), verify_certs=False)
        start_time = time.time()
           
        while time.time() - start_time < timeout:
            try:
                client.info()
                if client.ping():
                    return client
            except Exception as e:
                ex = e
                time.sleep(1)
        
        return (False, str(ex))
    except Exception as error:
        return (False, str(error))

def prepare_to_stop_node(host, node, type='restart') -> dict:
    "prepare node for graceful shutdown (helper function of rolling restart / full cluster restart / drop node)"
    
    try:
        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        es_client = connect_to_elasticsearch_insecure(host['name'], node['database_user'], node['database_pwd'], node['port'])
        if isinstance(es_client, tuple):
            jsn['stderr'] = str(es_client[1])
            jsn['excode'] = 1
            return jsn
        
        node_name = f"{host['name'].strip()}_{node['port'].strip()}"
        cluster_data = es_client.cat.nodes(h="name,id", full_id=True, format="json")
        node_data =  next((i for i in cluster_data if node_name == i["name"].strip()), None)
        if not node_data:
            jsn['stderr'] = f"error: could not find {node['name']} in cluster stats"
            return jsn
        
        es_client.shutdown.put_node(node_id=node_data.get("id"), reason=type, type=type)
        roles = node["ex_params"].get("roles")
        if roles and "data" in roles:
            es_client.cluster.put_settings(persistent={
                "cluster.routing.allocation.enable": "primaries"
            })
            es_client.cluster.put_settings(transient={
                "cluster.routing.allocation.enable": None
            })
            es_client.cluster.put_settings(transient={
                "cluster.routing.allocation.exclude._name": None
            })
        es_client.indices.flush()
        es_client.ml.set_upgrade_mode(enabled=True)
        return jsn
    except Exception as e:
        jsn['stderr'] = e.message if hasattr(e, 'message') else str(e)
        jsn['excode'] = 2
        return jsn

def stop_node(host, node, force = False) -> dict:
    jsn = {'stdout': '', 'stderr': '', 'excode': 0}

    try:
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node['port']}' | awk '{{print $7}}' | awk -F'/' '{{print $1}}'", True), strict=False)
        if jsn['stdout'] == '':
            # thats good, means node does not already running
            return {'stdout': '', 'stderr': '', 'excode': 0}
        
        port_pid = jsn['stdout']
        jsn = json.loads(remote_command(host, f"sudo kill {'-9' if force else ''} '{port_pid}'", True), strict=False)
        return jsn
    except Exception as e:
        jsn['stderr'] = e.message if hasattr(e, 'message') else str(e)
        jsn['excode'] = 2
        return jsn

def start_node(host, node, distributor, fresh=False, no_cleanup=False) -> dict:
    jsn = {'stdout': '', 'stderr': '', 'excode': 0}
    if fresh:
        username = 'elastic'
        password = node['ex_params'].get('elastic_pwd')
    else:
        username = node['database_user']
        password = node['database_pwd']
    
    es_client = connect_to_elasticsearch_insecure(host['name'], username, password, node['port'], timeout=Timeout.BACKEND_ELASTIC.value)
    if not isinstance(es_client, tuple):
        #no need to start when already started
        return jsn

    client = create_ssh_connection(host['name'], host['username'], host['password'], host['private_key'], host['ssh_port'])
    if isinstance(client, tuple):
        jsn['stderr'] = client[1].strerror if hasattr(client[1], "strerror") else str(client[1])
        jsn['excode'] = 1
        return jsn       
    
    node_conf = f"{node['dir_123Cluster'].strip()}/conf"
    elastic_bin = find_elastic_bin(host, distributor)

    commands = [
        f"sudo -u elasticsearch ES_PATH_CONF='{node_conf}' nohup {elastic_bin}/elasticsearch &> /dev/null &"
    ]
    if not fresh:
        commands.insert(0, f"sed -i '/cluster.initial_master_nodes:/d' '{node_conf}'")
    
    for command in commands:
        stdin, stdout, stderr = client.exec_command(command)
        exit_status = stdout.channel.recv_exit_status()# Blocking call
        stdin.channel.shutdown_write()
        try:
            stderr = stderr.read().decode('ascii').strip()
            stdout = stdout.read().decode('ascii').strip()
        except UnicodeDecodeError:
            stderr = stderr.read().decode('utf-8').strip()
            stdout = stdout.read().decode('utf-8').strip()

        if stderr != '' and 'error' in stderr:
            jsn['stderr'] = stderr
            jsn['stdout'] = stdout
            jsn['excode'] = exit_status
            client.close()
            return jsn
            
    client.close()
    
    if not no_cleanup:
        jsn = post_start_cleanup(host, node)
    
    return jsn

def restart_node(host, node, distributor, force=False) -> dict:
    jsn = stop_node(host, node, force=force)
    if jsn['stderr'] != '':
        return jsn
    
    jsn = start_node(host, node, distributor, no_cleanup=force)
    return jsn

def find_elastic_bin(host, distributor) -> str:
    """find location of elasticsearch bin folder"""

    tool = "dpkg-query -L" if OperatingSystem.UBUNTU.value in distributor else "rpm -ql"
    elastic_bin = ''

    try:
        jsn = json.loads(remote_command(host, f"sudo {tool} elasticsearch | grep '/bin$' | grep -vE '(jdk|modules)'", True), strict=False)
        if jsn['stdout'] != '':
            elastic_bin = jsn['stdout'].strip()

        if jsn['stderr'] != '' and 'permission denied' not in jsn['stderr'].lower():
            if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
                elastic_bin = '/usr/share/elasticsearch/bin'
            else:
                elastic_bin = '/usr/share/elasticsearch/bin'
    except Exception as error:
        logging.error(f"find_elastic_bin: {str(error)}")
    
    return elastic_bin

def install_elasticsearch_cluster_task(data, task_id):

    task = 'install elasticsearch cluster task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    
    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        app_host = json.loads(data['app_host'])
        # app_hostname = data['app_hostname']
        create_run_linux = data['create_run_linux']
        host_exists = True
        proxmox_container = node['ex_params'].get('proxmox')
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = node['ex_params'].get('distributor')
                release = node['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
        
        version_exists = check_elastic_version_exists(host, node['version'], distributor)
        # repository = network or local
        repository = data['repository']
        offline = True if Repository.LOCAL.value in repository else False
        es_version = node['version']
        node_name = f"{host['name'].strip()}_{node['port'].strip()}"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        update_failed_object(node['id'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)


    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    step = 'checking minimum requirements'
    duration_step_start = time.perf_counter()

    ram_mb = node['ex_params'].get('ram_mb')
    is_error, enough = is_ram_enough(host, ram_mb)
    if not enough or is_error:
        error = enough if is_error else f"Available RAM are less than required {ram_mb} MB"
        excode = 9 if is_error else 1
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,excode)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,excode)
        update_failed_object(node['id'])
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'running installation commands'
    duration_step_start = time.perf_counter()

    if not version_exists:
        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host, es_version, DBType.ELASTIC.value, app_host)
        else:
            jsn = install_elasticsearch(distributor, release, host, es_version)

        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'no such file', 'file not found', 'elasticsearch.service does not exist'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node['id'])
            return

        if offline:
            try:
                jsn = json.loads(remote_command(host, 'sudo systemctl disable elasticsearch', True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and any(err in error_lower for err in ['err', 'error', 'fail', 'errno']) and 'elasticsearch.service does not exist' not in error_lower:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(node['id'])
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                update_failed_object(node['id'])
                return
            
    jsn = create_elasticsearch_master(host, node, distributor)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()
    
    try:

        es_client = connect_to_elasticsearch_insecure(host['name'], node['database_user'], node['database_pwd'], node['port'])
        if isinstance(es_client, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(es_client[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(es_client[1]),1)
            return
        
        cluster_data = es_client.cat.nodes(h="name,du", full_id=True, format="json", bytes="b")
        node_data =  next((i for i in cluster_data if node_name == i["name"].strip()), None)
        if node_data:
            node_size = 0.0 if not node_data.get('du') else float(node_data['du']) / 1024.0
        else:
            node_size = 0.0
        
        remote_command(host, f"""echo "{get_log_rotate_config(node['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{node['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host), 'node_name': node['name']}, 'task': 'log_rotate_task',  'org_id': node['org_id'], 'monitor_name': None, 'org_id': node['org_id'],  'target_type': node['type'], 'target_id': node['id'], 'created_by': created_by}
        insert_task(params)

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s, version=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, es_version, node['id'],))
                conn.commit()
                # cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,node['id'],created_by])
                # conn.commit()
                # cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host['id'],created_by,OperatingSystem.LINUX.value])
                # conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
    
    # host['create_task'] = created_by
    # host['ssh_port'] = int(host['ssh_port'])
    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host, 'task': 'run_linux_performance_counters', 'target_type': 'host', 'target_id': host['id'], 'created_by': created_by}               
    #         insert_task(params)
    #     except Exception as error:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def install_elasticsearch(distributor, release, host, es_version) -> dict:
    
    if OperatingSystem.CENTOS.value in distributor:
        package_manager = 'dnf' if int(release) > 7 else 'yum'
        # Step 1: create the repository
        repo_file = f"""[elasticsearch]
name = Elasticsearch repository for {es_version}.x packages
baseurl = https://artifacts.elastic.co/packages/{es_version}.x/yum/
gpgcheck = 1
gpgkey = https://artifacts.elastic.co/GPG-KEY-elasticsearch
enabled = 0
autorefresh = 1
type = rpm-md
"""
        commands = [
            'sudo rpm --import https://artifacts.elastic.co/GPG-KEY-elasticsearch',
            "sudo mv '/etc/yum.repos.d/elasticsearch.repo' '/etc/yum.repos.d/elasticsearch.repo.bak' 2> /dev/null",
            f"echo '{repo_file}' | sudo tee '/etc/yum.repos.d/elasticsearch.repo'",
            f'sudo {package_manager} install -q -y --enablerepo=elasticsearch elasticsearch rsync net-tools'
        ]
    elif OperatingSystem.UBUNTU.value in distributor:
        commands = [
            "sudo debconf-set-selections <<< 'libssl1.1 libraries/restart-without-asking boolean true'",
            "sudo DEBIAN_FRONTEND=noninteractive apt-get -qq -y install  gpg apt-transport-https net-tools",
            "sudo wget -qO - https://artifacts.elastic.co/GPG-KEY-elasticsearch | sudo gpg -q --batch --yes --dearmor -o '/usr/share/keyrings/elasticsearch-keyring.gpg'",
            f"echo 'deb [signed-by=/usr/share/keyrings/elasticsearch-keyring.gpg] https://artifacts.elastic.co/packages/{es_version}.x/apt stable main' | sudo tee '/etc/apt/sources.list.d/elastic.list'",
            """sudo apt-get update && sudo DEBIAN_FRONTEND=noninteractive apt-get -o dpkg::Options::="--force-confdef" -qq -y install elasticsearch"""
        ]
    else:
        jsn = {}
        jsn['stderr'] = 'Distributor is not Ubuntu or CentOS'
        jsn['stdout'] = None
        jsn['excode'] = 99
        return jsn
    
    commands.append('sudo systemctl disable elasticsearch')
    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'file not found', 'no such file', 'elasticsearch.service does not exist'])
            if filter:
                return jsn
        except Exception as error:
            return {'stdout': None, 'stderr': str(error), 'excode': 9}
            
    return jsn

def create_elasticsearch_master(host, node, distributor) -> dict:
    """create one-node cluster with initial master"""
    
    try:
        jsn = {'stdout': '', 'stderr': '', 'ex_code': 0}

        if OperatingSystem.CENTOS.value in distributor:
            environment_variables = '/etc/sysconfig/elasticsearch'
        else:
            environment_variables = '/etc/default/elasticsearch'
        
        ram_mb = node["ex_params"].get("ram_mb")
        datadir = node['dir_123Cluster'].strip()
        hostname = host['name'].strip()
        ip = gethostbyname(hostname)
        node_name = f"{hostname}_{node['port'].strip()}"
        node_conf = f"{datadir}/conf"
        node_certs = f"{node_conf}/certs"
        elastic_bin = find_elastic_bin(host, distributor)
        cert_pass = node['ex_params'].get('elastic_pwd')
        unicast_entry =  f"""{hostname}:{node['ex_params'].get('transport_port')}\n"""
        conf_file = f"""reindex.ssl.verification_mode: none\n\
cluster.name: {node['ex_params'].get('cluster_name')}\n\
cluster.routing.allocation.enable: all\n\
cluster.routing.allocation.disk.threshold_enabled: false\n\
node.name: "{node_name}"\n\
node.roles: [master, data]\n\
network.host: 0.0.0.0
http.port: {node['port']}\n\
transport.port: {node['ex_params'].get('transport_port')}\n\
path.data: {datadir}\n\
path.logs: {datadir}/log\n\
discovery.seed_providers: file\n\
cluster.initial_master_nodes: ["{node_name}"]\n
xpack.security.enabled: true\n\
xpack.security.transport.ssl.enabled: true\n\
xpack.security.transport.ssl.verification_mode: certificate\n\
xpack.security.transport.ssl.client_authentication: required\n\
xpack.security.transport.ssl.keystore.path: certs/cert.p12\n\
xpack.security.transport.ssl.truststore.path: certs/cert.p12\n\
xpack.security.http.ssl.enabled: true\n\
xpack.security.http.ssl.keystore.path: certs/http.p12\n"""

        commands = [
            f"sudo mkdir -p '{node_conf}'",
            f"sudo rsync -a '/etc/elasticsearch/' '{node_conf}'",
            f"echo '-Xms{ram_mb}m\n-Xmx{ram_mb}m' | sudo tee '{node_conf}/jvm.options.d/jvm.options'",
            f"sudo rm -rf '{node_certs}'",
            f"sudo mkdir -p '{node_certs}'",
            f"sudo chown -R elasticsearch:elasticsearch '{datadir}'",
            f"echo '{unicast_entry}' | sudo tee '{node_conf}/unicast_hosts.txt'",
            f"echo '{conf_file}' | sudo tee '{node_conf}/elasticsearch.yml'",
            f"sudo sed -i '/ES_PATH_CONF/d' '{environment_variables}'",
            f"sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-certutil ca --out '{node_certs}/ca.p12' --pass '{cert_pass}'",
            f"sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-certutil cert --ca '{node_certs}/ca.p12' --ca-pass '{cert_pass}' --out '{node_certs}/cert.p12' --pass '{cert_pass}'",
            f"sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-certutil cert --ca '{node_certs}/ca.p12' --ca-pass '{cert_pass}' --dns '{hostname}' --ip '{ip}' --out '{node_certs}/http.p12' --pass '{cert_pass}'",
            f"sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-certutil cert --ca '{node_certs}/ca.p12' --ca-pass '{cert_pass}' --pem --dns '{hostname}' --ip '{ip}' --out '{node_certs}/http.pem' --pass '{cert_pass}'",
            f"echo 'y' | sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-keystore create",
            f"echo '{cert_pass}' | sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-keystore add -xf 'xpack.security.transport.ssl.keystore.secure_password'",
            f"echo '{cert_pass}' | sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-keystore add -xf 'xpack.security.transport.ssl.truststore.secure_password'",
            f"echo '{cert_pass}' | sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-keystore add -xf 'xpack.security.http.ssl.keystore.secure_password'",
            f"echo '{cert_pass}' | sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-keystore add -xf 'bootstrap.password'",
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            error_lower = jsn['stderr'].lower()
            if error_lower != '' and any(err in error_lower for err in ['err', 'fail']):
                return jsn
        
        jsn = start_node(host, node, distributor, fresh=True, no_cleanup=True)
        if jsn['stderr'] != '':
            return jsn

        es_client = connect_to_elasticsearch_insecure(host['name'], 'elastic', node['ex_params'].get('elastic_pwd'), node['port'])
        if isinstance(es_client, tuple):
            jsn['stdout'] = None
            jsn['excode'] = 1
            jsn['stderr'] = es_client[1].strerror if hasattr(es_client[1], 'strerror') else str(es_client[1])
        else:
            cluster_status = wait_for_being_writable(es_client)
            if cluster_status not in [Severity.OK.value, Severity.WARNING.value]:
                return {'stdout': None, 'stderr': cluster_status, 'excode':1}
        
            res = es_client.security.put_user(username=node['database_user'], password=node['database_pwd'], roles=["superuser"]).body
            if hasattr(res, 'error'):
                jsn['stdout'] = None
                jsn['excode'] = res.status_code
                jsn['stderr'] = res.error 
                return jsn

        return jsn
    except Exception as error:
           return {'stdout': None, 'stderr': str(error), 'excode': 9}

def create_elasticsearch_index_task(data, task_id):

    task = 'create elasticsearch index task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        db = json.loads(data['db'])
        settings = {
            "index.number_of_shards": db['ex_params'].get('num_shards'),
            "index.number_of_replicas": db['ex_params'].get('num_replicas')
        }
        owner = f"{db['name']}_owner_user"
        read = f"{db['name']}_read_user"
        write = f"{db['name']}_write_user"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating index'
    duration_step_start = time.perf_counter()

    es_client = connect_to_elasticsearch_insecure(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(es_client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(es_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(es_client[1]),1)
        return
    
    doc = {"doc_name": "cluster123", "host": f"{host['name']}"}
    try:
        es_client.indices.create(index=db['name'], settings=settings)
        es_client.index(index=db['name'], document=doc)
    except RequestError as err:
        if err.error == 'resource_already_exists_exception':
            err_msg = f"{db['name']} already exists"
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,err_msg,str(err),1)
            step = 'null'
            status = Severity.WARNING.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,err_msg,str(err),1)
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(err),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(err),1)
            return

    try:
        stats = es_client.indices.stats(index=db['name'], metric="store", forbid_closed_indices=False)
        index_size = round(float(stats["_all"]["total"]["store"]["total_data_set_size_in_bytes"]) / 1024, 2)
        if index_size < 0.01:
            index_size = 0.01        
    except RequestError as err:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(err),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(err),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'putting roles'
    duration_step_start = time.perf_counter()
    
    try:
        es_client.security.put_role(name=owner,indices= [
            {
              "names": [db['name']],
              "privileges": ["all"] 
            }
        ])
        
        es_client.security.put_role(name=read,indices= [
            {
              "names": [db['name']],
              "privileges": ["read"]
            }
        ])

        es_client.security.put_role(name=write,indices= [
            {
              "names": [db['name']],
              "privileges": ["read", "write"]
            }
        ])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating index users'
    duration_step_start = time.perf_counter()

    owner_user_pwd = generate_complex_password()
    read_user_pwd = generate_complex_password()
    write_user_pwd = generate_complex_password()
    logs_output = f'USERNAME = {owner}\nPASSWORD = {owner_user_pwd}\nUSERNAME = {read}\nPASSWORD = {read_user_pwd}\nUSERNAME = {write}\nPASSWORD = {write_user_pwd}'
    
    try:
        es_client.security.put_user(username=owner, password=owner_user_pwd, roles=[owner])
        es_client.security.put_user(username=read, password=read_user_pwd, roles=[read])
        es_client.security.put_user(username=write, password=write_user_pwd, roles=[write])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,logs_output,None,0)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, index_size,db['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def add_elasticsearch_node_task(data, task_id):
    """join/install node to existing cluster"""

    task = 'add elasticsearch node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_master = json.loads(data['host_master'])
        node = json.loads(data['node_data'])
        node_id = node['id']
        datadir = node['dir_123Cluster'].strip()
        host_node = json.loads(data['host_node'])
        app_host = json.loads(data['app_host'])
        host_exists = True
        proxmox_container = node['ex_params'].get('proxmox')
        distributor, release = check_os_version(host_node)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = node['ex_params'].get('distributor')
                release = node['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
        
        version_exists = check_elastic_version_exists(host_node, node['version'], distributor)

        hostname = host_node['name'].strip()
        ip = gethostbyname(hostname)
        node_name = f"{hostname.strip()}_{node['port'].strip()}"
        node_version = node['version']
        nodes = data['nodes']
        repository = data['repository']
        offline = True if Repository.LOCAL.value in repository else False
        datadir = node['dir_123Cluster'].strip()
        node_conf = f"{datadir}/conf"
        transport_port = node["ex_params"].get("transport_port")
        master = json.loads(data['master_data'])
        master_conf = f"{master['dir_123Cluster'].strip()}/conf"
        master_tmp_dir = f"{TASK_TMP_FILES_DIR}/{master['name'].strip().replace(':', '_')}"
        roles = node['ex_params'].get('roles')
        elastic_bin = find_elastic_bin(host_master, distributor)
        client = create_ssh_connection(host_master['name'], host_master['username'], host_master['password'], host_master['private_key'], host_master['ssh_port'])
        if isinstance(client, tuple):
            if hasattr(client[1], 'strerror'):
                error = client[1].strerror
            else:
                error = str(client[1])

            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            update_failed_object(node_id)
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host_node, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    step = 'checking minimum requirements'
    duration_step_start = time.perf_counter()

    ram_mb = node['ex_params'].get('ram_mb')
    is_error, enough = is_ram_enough(host_node, ram_mb)
    if not enough or is_error:
        error = enough if is_error else f"Available RAM are less than required {ram_mb} MB"
        excode = 9 if is_error else 1
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,excode)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,excode)
        update_failed_object(node_id)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if "master" in roles and "voting_only" not in roles:
        step = 'updating host providers'
        duration_step_start = time.perf_counter()

        commands = []
        
        for i in nodes:
            try:
                other_node = json.loads(i["node"])
                other_host = host_node if hostname in other_node['name'].strip() else json.loads(i["host"]) 
                other_unicast_file = f"{other_node['dir_123Cluster'].strip()}/conf/unicast_hosts.txt"
                jsn = json.loads(remote_command(other_host, f"sudo grep -q '{hostname}:{transport_port}' '{other_unicast_file}' || echo '{hostname}:{transport_port}' | sudo tee -a '{other_unicast_file}'", True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(node_id)
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                update_failed_object(node['id'])
                return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host_master['name'] == host_node['name']:
        step = 'configuring node'
        duration_step_start = time.perf_counter()

        data = f"""\nnode.name: "{node_name}"\n\
node.roles: {node['ex_params'].get('roles')}\n\
http.port: {node['port']}\n\
transport.port: {node['ex_params'].get('transport_port')}\n\
path.data: {datadir}\n\
path.logs: {datadir}/log\n"""

        commands = [
            f"sudo mkdir -p '{node_conf}'",
            f"sudo rsync -a '{master['dir_123Cluster']}/conf/' '{node_conf}'",
            f"echo '-Xms{ram_mb}m\n-Xmx{ram_mb}m' | sudo tee '{node_conf}/jvm.options.d/jvm.options'",
            f"sudo chown -R elasticsearch:elasticsearch '{datadir}'",
            f"sudo sed -i '/path.data:/d' '{node_conf}/elasticsearch.yml'",
            f"sudo sed -i '/path.logs:/d' '{node_conf}/elasticsearch.yml'",
            f"sudo sed -i '/node.name:/d' '{node_conf}/elasticsearch.yml'",
            f"sudo sed -i '/http.port:/d' '{node_conf}/elasticsearch.yml'",
            f"sudo sed -i '/transport.port:/d' '{node_conf}/elasticsearch.yml'",
            f"sudo sed -i '/node.roles:/d' '{node_conf}/elasticsearch.yml'",
            # must be last, to not break parsing (because that setting can be already removed in master)
            f"sudo sed -i '/cluster.initial_master_nodes:/d' '{node_conf}/elasticsearch.yml'",
            F"echo '{data}' | sudo tee -a '{node_conf}/elasticsearch.yml'",
            f"echo '{node['ex_params'].get('elastic_pwd')}' | sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-keystore add -xf 'bootstrap.password'",
        ]
        for command in commands:
            try:
                jsn = json.loads(remote_command(host_node, command, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(node_id)
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                update_failed_object(node['id'])
                return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    else:
        distributor, release = check_os_version(host_node)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node_id)
            return

        step = 'running installation commands'
        duration_step_start = time.perf_counter()

        if not version_exists:
            if offline:
                jsn = install_offline_remote_to_remote(distributor, release, host_node, node_version, DBType.ELASTIC.value, app_host)
            else:
                jsn = install_elasticsearch(distributor, release, host_node, node_version)
            
            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'no such file', 'file not found', 'elasticsearch.service does not exist'])
            if filter:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node_id)
                return

            if offline:
                try:
                    jsn = json.loads(remote_command(host_node, 'sudo systemctl disable elasticsearch', True), strict=False)
                    error_lower = jsn['stderr'].lower()
                    if error_lower != '' and any(err in error_lower for err in ['err', 'error', 'fail', 'errno']):
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        update_failed_object(node_id)
                        return
                except Exception as error:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                    return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        step = 'configuring node'
        duration_step_start = time.perf_counter()

        if OperatingSystem.CENTOS.value in distributor.lower():
            environment_file = '/etc/sysconfig/elasticsearch'
        else:
            environment_file = '/etc/default/elasticsearch'
        
        try:
            master_commands = [
                f"sudo mkdir -p '{master_tmp_dir}'",
                f"sudo rsync -a '{master_conf}/certs/ca.p12' '{master_tmp_dir}'",
                f"sudo rsync -a '{master_conf}/unicast_hosts.txt' '{master_tmp_dir}'"
            ]
            for command in master_commands:
                jsn = json.loads(remote_command(host_master, command, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(node_id)
                    return
            
            jsn = transfer_files_rsync(host_master, host_node, master_tmp_dir, [f"{master_tmp_dir}/"])
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
        
        node_conf = f"{node['dir_123Cluster']}/conf"
        node_certs = f"{node_conf}/certs"
        ca_pass = master['ex_params'].get('elastic_pwd')
        cert_pass = node['ex_params'].get('elastic_pwd')

        conf_file = f"""reindex.ssl.verification_mode: none\n\
cluster.name: {node['ex_params'].get('cluster_name')}\n\
cluster.routing.allocation.enable: all\n\
cluster.routing.allocation.disk.threshold_enabled: false\n\
node.name: "{node_name}"\n\
node.roles: {node["ex_params"].get("roles")}\n\
network.host: 0.0.0.0
http.port: {node['port']}\n\
transport.port: {node['ex_params'].get('transport_port')}\n\
path.data: {node['dir_123Cluster']}\n\
path.logs: {node['dir_123Cluster']}/log\n\
discovery.seed_providers: file\n
xpack.security.enabled: true\n\
xpack.security.transport.ssl.enabled: true\n\
xpack.security.transport.ssl.verification_mode: certificate\n\
xpack.security.transport.ssl.client_authentication: required\n\
xpack.security.transport.ssl.keystore.path: certs/cert.p12\n\
xpack.security.transport.ssl.truststore.path: certs/cert.p12\n\
xpack.security.http.ssl.enabled: true\n\
xpack.security.http.ssl.keystore.path: certs/http.p12\n"""
        
        commands = [
            f"sudo mkdir -p '{node_conf}'",
            f"sudo rsync -a '/etc/elasticsearch/' '{node_conf}'",
            f"echo '-Xms{ram_mb}m\n-Xmx{ram_mb}m' | sudo tee '{node_conf}/jvm.options.d/jvm.options'",
            f"sudo rm -rf '{node_certs}'",
            f"sudo mkdir -p '{node_certs}'",
            f"echo '{conf_file}' | sudo tee '{node_conf}/elasticsearch.yml'",
            f"sudo rsync -a '{master_tmp_dir}/ca.p12' '{node_certs}/ca.p12'",
            f"sudo rsync -a '{master_tmp_dir}/unicast_hosts.txt' '{node_conf}/unicast_hosts.txt'",
            f"sudo chown -R elasticsearch:elasticsearch '{datadir}'",
            f"sudo rm -rf  '{master_tmp_dir}'",
            f"sudo sed -i '/ES_PATH_CONF/d' '{environment_file}'",
            f"sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-certutil cert --ca '{node_certs}/ca.p12' --ca-pass '{ca_pass}' --out '{node_certs}/cert.p12' --pass '{cert_pass}'",
            f"sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-certutil cert --ca '{node_certs}/ca.p12' --ca-pass '{ca_pass}' --dns '{hostname}' --ip '{ip}' --out '{node_certs}/http.p12' --pass '{cert_pass}'",
            f"sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-certutil cert --ca '{node_certs}/ca.p12' --ca-pass '{ca_pass}' --pem --dns '{hostname}' --ip '{ip}' --out '{node_certs}/http.pem' --pass '{cert_pass}'",
            f"echo 'y' | sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-keystore create",
            f"echo '{cert_pass}' | sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-keystore add -xf 'xpack.security.transport.ssl.keystore.secure_password'",
            f"echo '{cert_pass}' | sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-keystore add -xf 'xpack.security.transport.ssl.truststore.secure_password'",
            f"echo '{cert_pass}' | sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-keystore add -xf 'xpack.security.http.ssl.keystore.secure_password'",
            f"echo '{cert_pass}' | sudo -u elasticsearch ES_PATH_CONF='{node_conf}' {elastic_bin}/elasticsearch-keystore add -xf 'bootstrap.password'"
        ]
        for command in commands:
            try:
                jsn = json.loads(remote_command(host_node, command, True), strict=False)
                if jsn['stderr'] != '' and ' elasticsearch keystore already exists' not in jsn['stderr'].lower():
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'starting node'
    duration_step_start = time.perf_counter()

    jsn = start_node(host_node, node, distributor, fresh=True, no_cleanup=True)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
       
    es_client = connect_to_elasticsearch_insecure(host_node['name'], 'elastic', node['ex_params'].get('elastic_pwd'), node['port'])
    if isinstance(es_client, tuple):
        if hasattr(es_client[1], 'strerror'):
            error = es_client[1].strerror
        else:
            error = str(es_client[1])

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    try:
        cluster_status = wait_for_being_writable(es_client)
        if cluster_status not in [Severity.OK.value, Severity.WARNING.value]:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,cluster_status,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,cluster_status,1)
            return

        es_client.security.put_user(username=node['database_user'], password=node['database_pwd'], roles=["superuser"]).body
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()

    try:
        
        cluster_data = es_client.cat.nodes(h="name,du", full_id=True, format="json", bytes="b")
        node_data = next((i for i in cluster_data if node_name == i["name"].strip()), None)
        if node_data:
            node_size = 0.0 if not node_data.get('du') else float(node_data['du']) / 1024.0
        else:
            node_size = 0.0
        
        remote_command(host_node, f"""echo "{get_log_rotate_config(node['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{node['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host_node), 'node_name': node['name']}, 'task': 'log_rotate_task',  'org_id': node['org_id'], 'monitor_name': None,  'target_type': node['type'], 'target_id': node['id'], 'created_by': created_by}
        insert_task(params)
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur = conn.cursor()
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s, version=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node_version, node['id'],))
                conn.commit()
                # cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,node['id'],created_by])
                # conn.commit()
                # cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host_node['id'],created_by,OperatingSystem.LINUX.value])
                # conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def delete_elasticsearch_index_task(data, task_id):
    """delete index"""

    task = 'delete elasticsearch index task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        db = json.loads(data['db'])
        users = [
            f"{db['name']}_owner_user",
            f"{db['name']}_read_user",
            f"{db['name']}_write_user",
        ]
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'deleting users'
    duration_step_start = time.perf_counter()

    es_client = connect_to_elasticsearch_insecure(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(es_client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(es_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(es_client[1]),1)
        return
    
    for user in users:
        try:
            es_client.security.delete_user(username=user)
        except NotFoundError:
            logging.warning(f'user {user} does not exist')
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        try:
            es_client.security.delete_role(name=user)
        except NotFoundError:
            logging.warning(f'role {user} does not exist')
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'deleting index'
    duration_step_start = time.perf_counter()

    try:
        es_client.indices.delete(index=db['name'], ignore_unavailable=True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = round(float(jsn['stdout']) / 1024, 2)
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DROPPED.value, db['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def close_elasticsearch_index_task(data, task_id):
    """close index"""
    
    task = 'close elasticsearch index task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        db = json.loads(data['db'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'closing index'
    duration_step_start = time.perf_counter()

    es_client = connect_to_elasticsearch_insecure(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(es_client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(es_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(es_client[1]),1)
        return

    not_found_index = False
    try:
        es_client.indices.close(index=db['name'])
    except NotFoundError:
        logging.warning(f"{db['name']} does not exist")
        not_found_index = True
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                if not_found_index:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Status.DELETED.value, db['id'],))
                else:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='false' WHERE id=%s""", (Status.OK.value, db['id'],))
                
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def elasticsearch_snapshot_task(data, task_id):
    """snapshot cluster/index"""
    # for local machine clusters only
    # dont have shared filesystem to work with this on cluster with remote nodes

    task = 'elasticsearch snapshot task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        snapshot = json.loads(data['backup_data'])
        snapshot_type = snapshot['ex_params'].get('backup_type')
        snapshot_path = snapshot['ex_params'].get('backup_path')
        compress = snapshot['ex_params'].get('compress')
        cluster_data = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in data["cluster_data"]]
        backup_host = data['backup_host']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
     
    step = 'updating cluster repository config'
    duration_step_start = time.perf_counter()

    commands = [
        f"sudo mkdir -p '{snapshot_path}'",
        f"sudo chmod -R ugo+rw '{snapshot_path}'"
    ]
    for command in commands:
        try:
            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    last_known_err = ''
    
    for i in cluster_data:
        host = i["host"]
        node = i["node"]
        jsn = prepare_to_stop_node(host, node)
        if jsn['stderr'] != '':
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

    for i in cluster_data:
        try:
            host = i["host"]
            node = i["node"]
            conf_path = f"{node['dir_123Cluster']}/conf/elasticsearch.yml"
            distributor, release = check_os_version(host)
            if OperatingSystem.INVALID.value in distributor:
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                continue

            jsn = json.loads(remote_command(host, f"sudo cat '{conf_path}'", True), strict=False)
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                continue

            updated_conf = None
            elasticsearch_conf = jsn['stdout']
            if 'path.repo' in elasticsearch_conf:
                match = re.search(r'path\.repo:\s*\[(.*?)\]', elasticsearch_conf)
                if match:
                    current_paths = match.group(1).split(', ')
                    if f'"{snapshot_path}"' not in current_paths:
                        current_paths.append(f'"{snapshot_path}"')
                        updated_paths = ', '.join(current_paths)
                        updated_conf = re.sub(r'path\.repo:\s*\[.*?\]', f'path.repo: [{updated_paths}]', elasticsearch_conf)
            else:
                path_repo = f'path.repo: ["{snapshot_path}"]'
                updated_conf = f'{elasticsearch_conf}\n{path_repo}'

            if not updated_conf:
                continue
            else:
                jsn = prepare_to_stop_node(host, node)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                jsn = stop_node(host, node)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                    continue


                jsn = json.loads(remote_command(host, f"echo '{updated_conf}' | sudo tee '{conf_path}'", True), strict=False)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                    continue

                jsn = start_node(host, node, distributor)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                    continue
        except Exception as error:
            last_known_err = {'stdout': '', 'stderr': str(error), 'ex_code': 9}
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
            continue

    status = Severity.OK.value if not last_known_err else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])
    
    step = 'creating snapshot'
    duration_step_start = time.perf_counter()

    repo_name = f"{node['ex_params'].get('cluster_name')}_indices" if snapshot_type == "index_snapshot" else f"{node['ex_params'].get('cluster_name')}_cluster"
    
    if snapshot_type == "index_snapshot":
        db_name = data['db_name']
    
    for i in cluster_data:
        es_client = connect_to_elasticsearch_insecure(i['host']['name'], i['node']['database_user'], i['node']['database_pwd'], i['node']['port'], Timeout.BACKEND_ELASTIC.value)
        if not isinstance(es_client, tuple):
            break

    if isinstance(es_client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(es_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(es_client[1]),1)
        return
       
    settings = {"compress": compress, "location": snapshot_path}
    
    try:
        #this create or update
        es_client.snapshot.create_repository(name=repo_name, settings=settings, type="fs")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    try:
        if snapshot_type == "index_snapshot":
            es_client.snapshot.create(repository=repo_name,include_global_state= False, snapshot=snapshot['name'], indices=db_name, partial=True, wait_for_completion = True)
        else:
            es_client.snapshot.create(repository=repo_name,include_global_state= False, indices = "*,-.*", snapshot=snapshot['name'], partial=True, wait_for_completion = True)
    except RequestError:
        print("already exists")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        stats = es_client.snapshot.status(repository=repo_name, snapshot=snapshot['name'])
        snapshot_size = round(float(stats['snapshots'][0]['stats']['total']['size_in_bytes'])/1024, 2)

        modified_date = get_date_modified_folder(hostdata=host, folder_path=snapshot_path)
        if isinstance(modified_date, tuple):
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"size":%s}', status='true' WHERE id=%s""", (snapshot_size, snapshot['id'],))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}', status='true' WHERE id=%s""" % (modified_date, snapshot['id']))
                if snapshot_type == 'index_snapshot':
                    query = """UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE id=%s"""
                    obj_id = data['db_id']
                else:
                    query = """UPDATE "clusterCore_c123_node" SET status=%s, ind_started='true' WHERE id=%s"""
                    obj_id = node['id']
                
                cur.execute(query, (Status.OK.value, obj_id,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
        
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def open_elasticsearch_index_task(data, task_id):
    """open index"""
    
    task = 'open elasticsearch index task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        db = json.loads(data['db'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'opening index'
    duration_step_start = time.perf_counter()

    es_client = connect_to_elasticsearch_insecure(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(es_client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(es_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(es_client[1]),1)
        return

    not_found_index = False
    try:
        es_client.indices.open(index=db['name'])
    except NotFoundError:
        logging.warning(f"{db['name']} does not exist")
        not_found_index = True
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                if not_found_index:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Status.DELETED.value, db['id'],))
                else:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE id=%s""", (Status.OK.value, db['id'],))
                
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restore_elasticsearch_snapshot_task(data, task_id):
    """restore cluster/index"""
    # for local machine clusters only
    # dont have shared filesystem to work with this on cluster with remote nodes

    task = 'restore elasticsearch snapshot task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        dump = json.loads(data['dump'])
        restore_date = data['date']
        restore_option = data["restore_option"].strip().lower()
        index_name = data.get('index_name')
        index_id = data.get('index_id')
        snapshot_path = dump['ex_params'].get('backup_path')
        compress = dump['ex_params'].get('compress')
        index_snapshot = dump['ex_params'].get('backup_type') == 'index_snapshot'
        cluster_data = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in data["cluster_data"]]
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
     
    step = 'updating cluster repository config'
    duration_step_start = time.perf_counter()

    last_known_err = ''
    
    for i in cluster_data:
        host = i["host"]
        node = i["node"]
        jsn = prepare_to_stop_node(host, node)
        if jsn['stderr'] != '':
            last_known_err = jsn
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
    
    for i in cluster_data:
        host = i["host"]
        node = i["node"]
        conf_path = f"{node['dir_123Cluster']}/conf/elasticsearch.yml"
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            last_known_err = release
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
            continue

        try:
            jsn = json.loads(remote_command(host, f"sudo cat '{conf_path}'", True), strict=False)
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                continue

            updated_conf = None
            elasticsearch_conf = jsn['stdout']
            if 'path.repo' in elasticsearch_conf:
                match = re.search(r'path\.repo:\s*\[(.*?)\]', elasticsearch_conf)
                if match:
                    current_paths = match.group(1).split(', ')
                    if f'"{snapshot_path}"' not in current_paths:
                        current_paths.append(f'"{snapshot_path}"')
                        updated_paths = ', '.join(current_paths)
                        updated_conf = re.sub(r'path\.repo:\s*\[.*?\]', f'path.repo: [{updated_paths}]', elasticsearch_conf)
            else:
                path_repo = f'path.repo: ["{snapshot_path}"]'
                updated_conf = f'{elasticsearch_conf}\n{path_repo}'

            if not updated_conf:
                continue
            else:
                jsn = prepare_to_stop_node(host, node, type='remove')
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                jsn = stop_node(host, node)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                    continue

                jsn = json.loads(remote_command(host, f"echo '{updated_conf}' | sudo tee '{conf_path}'", True), strict=False)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                    continue
                
                jsn = start_node(host, node, distributor)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                    continue
        
        except Exception as error:
            last_known_err = {'stdout': '', 'stderr': str(error), 'excode': 9}
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
            continue
        
    status = Severity.OK.value if not last_known_err else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'removing original indices'
    duration_step_start = time.perf_counter()

    for i in cluster_data:
        es_client = connect_to_elasticsearch_insecure(i['host']['name'], i['node']['database_user'], i['node']['database_pwd'], i['node']['port'], Timeout.BACKEND_ELASTIC.value)
        if not isinstance(es_client, tuple):
            break

    if isinstance(es_client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(es_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(es_client[1]),1)
        return

    try:
        if not index_snapshot:
            indices_to_delete = [i["index"] for i in es_client.cat.indices(h="index", format="json") if i.get("index") and not i.get("index").startswith(".")]
        else:
            indices_to_delete = [i["index"] for i in es_client.cat.indices(h="index", format="json") if i.get("index") == index_name]
        
        if len(indices_to_delete) > 0:
            for idx in indices_to_delete:
                es_client.indices.delete(index=idx, ignore_unavailable=True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    
    step = 'restoring snapshot'
    duration_step_start = time.perf_counter()

    settings = {"compress": compress, "location": snapshot_path}
    repo_name = f'{node["ex_params"].get("cluster_name")}_indices' if index_snapshot else f'{node["ex_params"].get("cluster_name")}_cluster'
    
    try:
        #this create or update
        es_client.snapshot.create_repository(name=repo_name, settings=settings, type="fs")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    snapshot_name = dump['name']
    try:
        if restore_option in [RestoreOptions.LATEST.value, RestoreOptions.CUSTOM.value]:
            if RestoreOptions.CUSTOM.value in restore_option:
                restore_date = datetime.strptime(restore_date, "%Y-%m-%d %H:%M:%S").astimezone(pytz.UTC).strftime('%Y-%m-%dT%H:%M:%S.%f') + 'Z'
            else:
                restore_date = datetime.now().astimezone(pytz.UTC).strftime('%Y-%m-%dT%H:%M:%S.%f') + 'Z'
            
            snapshots = es_client.snapshot.get(repository=repo_name,snapshot="_all", ignore_unavailable=True)
            if len(snapshots['snapshots']) > 0:
                snapshots['snapshots'].sort(key=lambda x: x["end_time"], reverse=True)
                latest_snapshot = None
                
                for snapshot in snapshots['snapshots']:
                    if snapshot["end_time"] <= restore_date:
                        if latest_snapshot is None or snapshot["end_time"] > latest_snapshot["end_time"]:
                            latest_snapshot = snapshot
                            
                if latest_snapshot:
                    snapshot_name = latest_snapshot['snapshot']

        es_client.snapshot.restore(repository=repo_name, indices="*,-.*", snapshot=snapshot_name, partial=True, include_global_state= False, wait_for_completion = True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

    step = 'updating app database'
    duration_step_start = time.perf_counter()

    try:
        cluster_stats = es_client.cat.nodes(h="name,du", full_id=True, format="json", bytes="b")
        db_set_from_restore = {i["index"] for i in es_client.cat.indices(h="index", format="json") if i.get("index") and not i.get("index").startswith(".")}
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    try:
        with get_postgres_connection() as conn:
            with conn.cursor(cursor_factory = RealDictCursor) as cur:
                cur.execute("""SELECT DISTINCT ON (name) id, name FROM "clusterCore_c123_database" WHERE node_id=%s""", (node['id'],))
                db_list_from_app = cur.fetchall()
                db_set_from_app = {db['name'].strip() for db in db_list_from_app}
                common_db = db_set_from_app.intersection(db_set_from_restore)
                missing_db = db_set_from_restore - db_set_from_app
                extra_db = db_set_from_app - db_set_from_restore
                for db in common_db:
                    try:
                        db_size = round(float(es_client.indices.stats(index=db)["_all"]["total"]["store"]["size_in_bytes"]) / 1024, 2)
                    except NotFoundError:
                        print(f"index {db} likely empty")
                        db_size = 0.0
                    
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, "size"=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, Status.OK.value, db_size, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                    conn.commit()
                
                for db in extra_db:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", [Severity.NOT_OK.value, Status.OK.value, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                    conn.commit()
                
                for db in missing_db:
                    try:
                        db_size = round(float(es_client.indices.stats(index=db)["_all"]["total"]["store"]["size_in_bytes"]) / 1024, 2)
                    except NotFoundError:
                        print(f"index {db} likely empty")
                        db_size = 0.0
                    
                    cur.execute("""INSERT INTO "clusterCore_c123_database" 
                    ("name", "type", created_date, last_update_date, size, ind_deployed, ind_started, ind_autostartup, ind_deleted, severity, sizing, database_user, database_pwd, created_by_id, last_update_by_id, node_id) 
                    VALUES (%s, %s, now(), now(), %s, 'false', 'true', 'false', 'false', %s, 'MEDIUM', %s, %s, %s, %s, %s)""", (db, DBType.MYSQL.value, db_size, Severity.OK.value, node['database_user'], node['database_pwd'], created_by, created_by, node['id']))
                    conn.commit()
                
                query_index = """UPDATE "clusterCore_c123_database" SET severity = %s, status=%s, ind_started='true', ind_deleted = 'false', size=%s WHERE id=%s"""
                query_node = """UPDATE "clusterCore_c123_node" SET severity = %s, status=%s, ind_started='true', size=%s WHERE id=%s"""
                if index_snapshot:
                    try:
                        db_size = round(float(es_client.indices.stats(index=index_name)["_all"]["total"]["store"]["size_in_bytes"]) / 1024, 2)
                    except NotFoundError:
                        db_size = 0.0
                    cur.execute(query_index, (Severity.OK.value, Status.OK.value, db_size, index_id))
                
                for i in cluster_data:
                    node = i['node']
                    node_name = f"{i['host']['name'].strip()}_{node['port'].strip()}"
                    node_data =  next((i for i in cluster_stats if node_name == i["name"].strip()), None)
                    if node_data:
                        node_size = 0.0 if not node_data.get('du') else float(node_data['du']) / 1024.0
                    else:
                        node_size = 0.0
                    cur.execute(query_node, (Severity.OK.value, Status.OK.value, node_size, node['id']))
                    conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
        
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def start_elasticsearch_node_task(data, task_id):
    """start node"""
    
    task = 'start elasticsearch node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'starting node'
    duration_step_start = time.perf_counter()

    jsn = start_node(host, node, distributor)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET status=%s, ind_started='true' WHERE id=%s""", (Status.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restart_elasticsearch_node_task(data, task_id):
    """restart node"""
    
    task = 'restart elasticsearch node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        node_unavailable = False
        
        es_client = connect_to_elasticsearch_insecure(host['name'], node['database_user'], node['database_pwd'], node['port'])
        if isinstance(es_client, tuple):
            node_unavailable = True
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'restarting node'
    duration_step_start = time.perf_counter()

    if not node_unavailable:
        jsn = prepare_to_stop_node(host, node)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

    jsn = restart_node(host, node, distributor, force=node_unavailable)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET status=%s, ind_started='true' WHERE id=%s""", (Status.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def upgrade_node(host, node, app_host, target_version) -> dict:
    "upgrade a single node to the target elasticsearch major version"

    try:
        jsn = {'stdout': '', 'stderr': '', 'excode': 0}

        offline = Repository.LOCAL.value in node['ex_params'].get('repository')
        
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            return release

        target_version_exists = check_elastic_version_exists(host, target_version, distributor)
        if not target_version_exists:
            if offline:
                jsn = install_offline_remote_to_remote(distributor, release, host, target_version, DBType.ELASTIC.value, app_host)
            else:
                jsn = install_elasticsearch(distributor, release, host, target_version)

            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'no such file', 'file not found', 'elasticsearch.service does not exist'])
            if filter:
                return jsn

            if offline:
                jsn = json.loads(remote_command(host,'sudo systemctl disable elasticsearch', True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and any(err in error_lower for err in ['err', 'error', 'fail', 'errno']) and 'elasticsearch.service does not exist' not in error_lower:
                    return jsn
            
        jsn = start_node(host, node, distributor, no_cleanup=True)
        if jsn['stderr'] != '':
            return jsn
        
        return jsn
    except Exception as e:
        return {'stdout': '', 'stderr': str(e), 'excode': 9}

def upgrade_elasticsearch_node_task(data, task_id):
    """upgrade node to the next major version"""
    
    task = 'upgrade elasticsearch node task'

    try:
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        app_host = json.loads(data['app_host'])
        cluster_nodes = [{'host': json.loads(n['host']), 'node': json.loads(n['node'])} for n in data['cluster_nodes']]
        # data-master-voting_only nodes also counts there, because able to store both system and data indices
        data_master_nodes = [n  for n in cluster_nodes if n['node']['ex_params'].get('roles') and len(n['node']['ex_params']['roles']) > 1 and 'data' in n['node']['ex_params']['roles'] and 'master' in n['node']['ex_params']['roles']]
        master_nodes = [n  for n in cluster_nodes if n['node']['ex_params'].get('roles') and len(n['node']['ex_params']['roles']) == 1 and 'master' in n['node']['ex_params']['roles']]
        data_nodes = [n  for n in cluster_nodes if n['node']['ex_params'].get('roles') and len(n['node']['ex_params']['roles']) == 1 and 'data' in n['node']['ex_params']['roles']]
        other_nodes = [n  for n in cluster_nodes if not n['node']['ex_params'].get('roles') or len(n['node']['ex_params']['roles']) == 0 or (len(n['node']['ex_params']['roles']) > 0 and 'data' not in n['node']['ex_params']['roles'] and 'master' not in n['node']['ex_params']['roles'])]
        target_version = data['target_version']
        cluster_id = cluster_nodes[0]['node']['cluster_id']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'shutdown cluster nodes'
    duration_step_start = time.perf_counter()

    last_known_err = ''
    
    try:
        # separate prepare to shutdown and shutdown itself
        for n in cluster_nodes:
            jsn = prepare_to_stop_node(n['host'], n['node'])
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
        
        for n in  cluster_nodes:
            jsn = stop_node(n['host'], n['node'])
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
    
    except Exception as error:
        last_known_err = str(error)
        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
    
    status = Severity.OK.value if not last_known_err else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'] if isinstance(last_known_err, dict) else None,last_known_err['stderr'] if isinstance(last_known_err, dict) else last_known_err,last_known_err['excode'] if isinstance(last_known_err, dict) else 1)
    
    step = 'upgrading cluster nodes'
    duration_step_start = time.perf_counter()

    last_known_err = ''
    
    try:
        for n in data_master_nodes:
            jsn = upgrade_node(n['host'], n['node'], app_host, target_version)
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
        
        for n in master_nodes:
           jsn = upgrade_node(n['host'], n['node'], app_host, target_version)
           if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
        
        for n in data_nodes:
           jsn = upgrade_node(n['host'], n['node'], app_host, target_version)
           if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        for n in other_nodes:
            jsn = upgrade_node(n['host'], n['node'], app_host, target_version)
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
    except Exception as error:
            last_known_err = str(error)
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
    
    status = Severity.OK.value if not last_known_err else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'] if isinstance(last_known_err, dict) else None,last_known_err['stderr'] if isinstance(last_known_err, dict) else last_known_err,last_known_err['excode'] if isinstance(last_known_err, dict) else 1)

    step = 'connecting to cluster'
    duration_step_start = time.perf_counter()
        
    config_client = None
    connect_error = ''
    for n in cluster_nodes:
        host = n["host"]
        node = n["node"]
        config_client = connect_to_elasticsearch_insecure(host['name'], node['database_user'], node['database_pwd'], node['port'], timeout=Timeout.BACKEND_ELASTIC.value)
        if isinstance(config_client, tuple):
            connect_error = str(config_client[1])
            continue
        
        break
        
    if not config_client or isinstance(config_client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,connect_error,9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,connect_error,9)
        return
    
    cluster_status = wait_for_being_writable(config_client)
    if cluster_status not in [Severity.OK.value, Severity.WARNING.value]:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,cluster_status,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,cluster_status,1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'removing safety locks from nodes'
    duration_step_start = time.perf_counter()

    try:
        last_known_err = ''

        for n in cluster_nodes:
            host = n['host']
            node = n['node']

            jsn = post_start_cleanup(host['name'], node['port'], node['database_user'], node['database_pwd'], node['ex_params'].get('roles'))
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
    except Exception as error:
        last_known_err = str(error)
        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

    status = Severity.OK.value if not last_known_err else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'] if isinstance(last_known_err, dict) else None,last_known_err['stderr'] if isinstance(last_known_err, dict) else last_known_err,last_known_err['excode'] if isinstance(last_known_err, dict) else 1)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET status=%s, ind_started='true', version=%s WHERE cluster_id=%s""", (Status.OK.value, target_version, cluster_id,))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def stop_elasticsearch_node_task(data, task_id):
    """stop node"""
    
    task = 'stop elasticsearch node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        node_unavailable = False
        
        es_client = connect_to_elasticsearch_insecure(host['name'], node['database_user'], node['database_pwd'], node['port'])
        if isinstance(es_client, tuple):
            node_unavailable = True
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stopping node'
    duration_step_start = time.perf_counter()
    
    warning = ''
    if not node_unavailable:
        jsn = prepare_to_stop_node(host, node)
        if jsn['stderr'] != '':
            warning = jsn
    
    jsn = stop_node(host, node, force=node_unavailable)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value if not warning else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if not warning else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,warning['stdout'],warning['stderr'],warning['excode'])
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='false' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def copy_elasticsearch_index_task(data, task_id):
    """copy index from some instance into another"""

    task = 'copy elasticsearch index task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_source = json.loads(data['host_source'])
        host_target = json.loads(data['host_target'])
        node_source = json.loads(data['node_source'])
        node_target = json.loads(data['node_target'])
        node_source_name = f"{host_source['name'].strip()}_{node_source['port'].strip()}"
        distributor, release = check_os_version(host_target)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        db_source = json.loads(data['db_source'])
        db_target = json.loads(data['db_target'])
        # db_option = 'override' or 'copy'
        db_option = data['db_option']
        db_exists = data['db_exists']
        keep_settings = data['keep_settings']
        owner = f"{db_target['name']}_owner_user"
        read = f"{db_target['name']}_read_user"
        write = f"{db_target['name']}_write_user"

        settings = None
        # copy_option = data['copy_option']
        other_cluster = host_source['name'] != host_target['name'] or node_source['port'] != node_target['port']
        if other_cluster: 
            src = {
                "remote": {
                "host": f"https://{host_source['name']}:{node_source['port']}",
                "username": node_source['database_user'],
                "password": node_source['database_pwd']
                },
                "index": db_source['name']
            }
        else:
            src = {"index": db_source['name']}

        dest = {"index": db_target['name']}

        es_client_target = connect_to_elasticsearch_insecure(host_target['name'], node_target['database_user'], node_target['database_pwd'], node_target['port'])
        if isinstance(es_client_target, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(es_client_target[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(es_client_target[1]),1)
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    if other_cluster:
        step = 'updating reindex remote whitelist'
        duration_step_start = time.perf_counter()

        try:
            conf_path = f"{node_target['dir_123Cluster']}/conf/elasticsearch.yml"
            jsn = json.loads(remote_command(host_target, f"sudo cat '{conf_path}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            updated_conf = None
            elasticsearch_conf = jsn['stdout']
            if 'reindex.remote.whitelist' in elasticsearch_conf:
                match = re.search(r'reindex\.remote\.whitelist:\s*\[(.*?)\]', elasticsearch_conf)
                if match:
                    current_names = match.group(1).split(', ')
                    if node_source_name not in current_names:
                        current_names.append(node_source_name)
                        updated_names = ', '.join(current_names)
                        updated_conf = re.sub(r'reindex\.remote\.whitelist:\s*\[.*?\]', f'reindex.remote.whitelist: [{updated_names}]', elasticsearch_conf)
            else:
                reindex_remote_whitelist = f'reindex.remote.whitelist: ["{node_source_name}"]'
                updated_conf = f'{elasticsearch_conf}\n{reindex_remote_whitelist}'

            if updated_conf:
                jsn = json.loads(remote_command(host_target, f"echo '{updated_conf}' | sudo tee '{conf_path}'", True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
                
                jsn = prepare_to_stop_node(host_target, node_target)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return

                jsn = stop_node(host_target, node_target)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return

                jsn = start_node(host_target, node_target, distributor)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if keep_settings:
        step = 'getting source index settings'
        duration_step_start = time.perf_counter()

        if host_source['name'] != host_target['name']:
            es_client_source = connect_to_elasticsearch_insecure(host_source['name'], node_source['database_user'], node_source['database_pwd'], node_source['port'])
            if isinstance(es_client_source, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(es_client_source[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(es_client_source[1]),1)
                return
            
        try:
           settings = es_client_target.indices.get_settings(index = db_source['name'], include_defaults=False)[db_source['name']]["settings"] if host_source['name'] == host_target['name'] else es_client_source.indices.get_settings(index = db_source['name'], include_defaults=False)[db_source['name']]["settings"]
           #remove unneeded settings
           settings["index"].pop("uuid")
           settings["index"].pop("version")
           settings["index"].pop("creation_date")
           settings["index"].pop("provided_name")
        except NotFoundError:
            settings = None
            logging.warning(f"{db_source['name']} does not exist")
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
         
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if db_exists and db_option == 'override':
        if other_cluster:
            step = 'dropping existing index'
            duration_step_start = time.perf_counter()

            try:
                es_client_target.indices.delete(index=db_source['name'], ignore_unavailable=True)
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
    
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    #elasticsearch could not copy to same index of same cluster
    #maybe remove override in case of same cluster?
    if not (db_option == 'override' and not other_cluster):
        step = 'copying the index'
        duration_step_start = time.perf_counter()
        
        try:
            if keep_settings and settings:
                es_client_target.indices.create(index=db_target['name'], settings=settings)
            else:
                es_client_target.indices.create(index=db_target['name'])
        except RequestError as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        try:
            res = es_client_target.reindex(source=src, dest=dest, wait_for_completion=True).body
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'putting roles'
    duration_step_start = time.perf_counter()
    
    try:
        es_client_target.security.put_role(name=owner,indices= [
            {
              "names": [db_target['name']],
              "privileges": ["all"] 
            }
        ])
        
        es_client_target.security.put_role(name=read,indices= [
            {
              "names": [db_target['name']],
              "privileges": ["read"]
            }
        ])

        es_client_target.security.put_role(name=write,indices= [
            {
              "names": [db_target['name']],
              "privileges": ["read", "write"]
            }
        ])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'putting users'
    duration_step_start = time.perf_counter()
    
    try:
        es_client_target.security.put_user(username=owner, password=owner, roles=[owner])
        es_client_target.security.put_user(username=read, password=read, roles=[read])
        es_client_target.security.put_user(username=write, password=write, roles=[write])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(host_target, f"sudo du -sb '{node_target['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = round(float(jsn['stdout']) / 1024, 2)
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node_target['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_target['id'],))
                if not db_exists:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_source['id'],))
                
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    
def drop_elasticsearch_node_task(data, task_id):
    """drop node"""
    
    task = 'drop elasticsearch node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        node_id = data['node_id']
        cluster_data = [{'host': json.loads(i['host']), 'node': json.loads(i['node'])} for i in data['cluster_data']]
        node_data = next(i for i in cluster_data if i['node']['id'] == node_id)
        host = node_data['host']
        node = node_data['node']
        other_nodes = [i for i in cluster_data if i['node']['id'] != node['id']]
        other_nodes_exists = len(other_nodes) > 0 
        node_name = f"{host['name'].strip()}_{node['port'].strip()}"
        final_backup = data.get('final_backup')
        backup_host = data['backup_host']
        roles = node["ex_params"].get("roles")
        should_update_config = other_nodes_exists and roles and "master" in roles and 'voting_only' not in roles
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node existence'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node['port']}' | awk '{{print $7}}' | awk -F'/' '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        port_pid = jsn['stdout']
        if port_pid == '':
            jsn = json.loads(remote_command(host, f"sudo rm -rf '{node['dir_123Cluster']}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'{node["name"]} is already deleted',1)

            try:
                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                        cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node["id"],))
                        conn.commit()
            except Exception as error:
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

            step = 'task finished'
            status = Severity.WARNING.value
            duration = 0
            logger(task_id,task_log_id,task,step,status,duration,created_by)
            step = 'null'
            status = Severity.WARNING.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    is_primary_error = None
    node_client = connect_to_elasticsearch_insecure(host['name'], node['database_user'], node['database_pwd'], node['port'], timeout=Timeout.BACKEND_ELASTIC.value)
    if isinstance(node_client, tuple):
        is_primary = False
        is_primary_error = str(node_client[1])
    else:
        try:
            res = node_client.cat.master(format='json')
            if len(res) > 0:
                is_primary = res[0].get('node') == node['name']
            else:
                is_primary = False
                is_primary_error = "no master found"
        except Exception as error:
            is_primary = False
            is_primary_error = str(error)
    
    if is_primary_error:
        status = Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,is_primary_error, 1)
    else:
        if is_primary and other_nodes_exists:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'{node["name"]} is primary with standbys.Switchover first.',1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,f'{node["name"]} is primary with nodes.Switchover first.',1)
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    if node['cluster_role'] != ClusterRoles.PRIMARY.value:
        final_backup = False
    if final_backup:
        step = 'snapshot the cluster'
        duration_step_start = time.perf_counter()
        
        snapshot_path = final_backup['ex_params'].get('backup_path')
        commands = [
            f"sudo mkdir -p '{snapshot_path}'",
            f"sudo chmod -R ugo+rw '{snapshot_path}'"
        ]
        for command in commands:
            try:
                jsn = json.loads(remote_command(backup_host, command, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return

        last_known_err = ''
        
        for i in cluster_data:
            host = i["host"]
            node = i["node"]
            jsn = prepare_to_stop_node(host, node)
            if jsn['stderr'] != '':
                last_known_err = jsn
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
        
        for i in cluster_data:
            host = i["host"]
            node = i["node"]
            conf_path = f"{node['dir_123Cluster']}/conf/elasticsearch.yml"
            distributor, release = check_os_version(host)
            if OperatingSystem.INVALID.value in distributor:
                last_known_err = release
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                continue

            try:
                jsn = json.loads(remote_command(host, f"sudo cat '{conf_path}'", True), strict=False)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                    continue

                updated_conf = None
                elasticsearch_conf = jsn['stdout']
                if 'path.repo' in elasticsearch_conf:
                    match = re.search(r'path\.repo:\s*\[(.*?)\]', elasticsearch_conf)
                    if match:
                        current_paths = match.group(1).split(', ')
                        if f'"{snapshot_path}"' not in current_paths:
                            current_paths.append(f'"{snapshot_path}"')
                            updated_paths = ', '.join(current_paths)
                            updated_conf = re.sub(r'path\.repo:\s*\[.*?\]', f'path.repo: [{updated_paths}]', elasticsearch_conf)
                else:
                    path_repo = f'path.repo: ["{snapshot_path}"]'
                    updated_conf = f'{elasticsearch_conf}\n{path_repo}'

                if not updated_conf:
                    continue
                else:
                    jsn = prepare_to_stop_node(host, node, type='remove')
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                   
                    jsn = stop_node(host, node)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                        continue

                    jsn = json.loads(remote_command(host, f"echo '{updated_conf}' | sudo tee '{conf_path}'", True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                        continue
                    
                    jsn = start_node(host, node, distributor)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                        continue
            
            except Exception as error:
                last_known_err = {'stdout': '', 'stderr': str(error), 'excode': 9}
                logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                continue

        repo_name = f"{node['ex_params'].get('cluster_name')}_cluster"

        for i in cluster_data:
            es_client = connect_to_elasticsearch_insecure(i['host']['name'], i['node']['database_user'], i['node']['database_pwd'], i['node']['port'], Timeout.BACKEND_ELASTIC.value)
            if not isinstance(es_client, tuple):
                break

        if isinstance(es_client, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(es_client[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(es_client[1]),1)
            return
        
        settings = {"compress": True, "location": snapshot_path}
        
        try:
            #this create or update
            es_client.snapshot.create_repository(name=repo_name, settings=settings, type="fs")
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        try:
             es_client.snapshot.create(repository=repo_name,include_global_state= False, indices = "*,-.*", snapshot=final_backup['name'], partial=True, wait_for_completion = True)
        except RequestError:
            print("already exists")
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])
    
    if should_update_config:
        step = 'updating host providers of other nodes'
        duration_step_start = time.perf_counter()

        for other in other_nodes:
            node_data = other["node"]
            host_data = other["host"]
            jsn = json.loads(remote_command(host_data, f"""sudo sed -i '/{host["name"]}:{node["port"]}/d' '{node_data["dir_123Cluster"]}/conf/unicast_hosts.txt'""", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        
        step = 'connecting to cluster'
        duration_step_start = time.perf_counter()
        
        config_client = None
        connect_error = ''
        if not is_primary_error:
            config_client = node_client
        else:
            for other in other_nodes:
                host_data = other["host"]
                node_data = other["node"]
                config_client = connect_to_elasticsearch_insecure(host_data['name'], node_data['database_user'], node_data['database_pwd'], node_data['port'], Timeout.BACKEND_ELASTIC.value)
                if isinstance(config_client, tuple):
                   connect_error = str(config_client)
                   continue
                
                break
           
        if not config_client or isinstance(config_client, tuple):
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,connect_error,1)
        else:
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

            if is_primary:
                master_candidates = [i for i in other_nodes if i['node']['ex_params'].get('roles') and 'master' in i['node']['ex_params']['roles'] and 'voting_only' not in i['node']['ex_params']['roles']]     
                if len(master_candidates) > 0:
                    step = 'switching to new master'
                    duration_step_start = time.perf_counter()
                        
                    target_client = None
                    target = None
                    for i in master_candidates:
                        host_data = i["host"]
                        node_data = i["node"]
                        #target should be available
                        target_client = connect_to_elasticsearch_insecure(host_data['name'], node_data['database_user'], node_data['database_pwd'], node_data['port'], Timeout.BACKEND_ELASTIC.value)
                        if not isinstance(target_client, tuple):
                            target = node_data
                            break

                    if isinstance(target_client, tuple):
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(target_client[1]),1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(target_client[1]),1)
                        return
                        
                    name_list = [i['node']['name'] for i in master_candidates if i['node']['name'] != target]
                    jsn = switchover_node(config_client, target['name'], name_list)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                    
                    try:
                        with get_postgres_connection() as conn:
                            with conn.cursor() as cur:
                                cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"elasticsearch_master_id":%s}', cluster_role=%s WHERE cluster_id=%s""", (target['id'], ClusterRoles.STANDBY.value, ['id'], target['id']))
                                cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"elasticsearch_master_id":%s}', cluster_role=%s WHERE id=%s""", (target['id'], ClusterRoles.PRIMARY.value, target['id']))
                                cur.execute("""UPDATE "clusterCore_c123_database" SET node_id = %s WHERE node_id=%s""", (target["id"], node["id"]))
                                conn.commit()
                    except Exception as error:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                        return
        
                    status = Severity.OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

            step = 'posting voting config exclusions'
            duration_step_start = time.perf_counter()
            
            voting_post_error = None
            try:
                config_client.cluster.post_voting_config_exclusions(node_names=node_name)
            except Exception as e:
                voting_post_error = e.message if hasattr(e, 'message') else str(e)
            
            if voting_post_error:
                status = Severity.WARNING.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,voting_post_error,1)
            else:
                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stopping node'
    duration_step_start = time.perf_counter()

    if not is_primary_error:
        jsn = prepare_to_stop_node(host, node, type='remove')
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
    jsn = stop_node(host, node, force=is_primary_error)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'removing node'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo rm -rf '{node['dir_123Cluster']}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
   
    if should_update_config:
        step = 'removing voting config exclusions'
        duration_step_start = time.perf_counter()
          
        voting_connect_error = None
        for other in other_nodes:
            host_data = other["host"]
            node_data = other["node"]
            config_client = connect_to_elasticsearch_insecure(host_data['name'], node_data['database_user'], node_data['database_pwd'], node_data['port'], timeout=Timeout.BACKEND_ELASTIC.value)
            if isinstance(config_client, tuple):
                voting_connect_error = str(config_client)
                continue
            
            break
          
        if voting_connect_error:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,connect_error,1)
        else:
            voting_remove_error = None
            try:
                config_client.cluster.delete_voting_config_exclusions(wait_for_removal=False)
            except Exception as e:
                voting_remove_error = e.message if hasattr(e, 'message') else str(e)
            
            if voting_remove_error:
                status = Severity.WARNING.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,voting_remove_error,1)
            else:
                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
                
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node["id"],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def switchover_node(es_client, target, name_list, timeout = 60):
    jsn = {'stdout': '', 'stderr': '', 'excode': 0}
    if len(name_list) == 0:
        #no need to change voting configuration - 
        # without competitors that node should be primary already
        return jsn
    #exclude other candidates
    try:
        es_client.cluster.post_voting_config_exclusions(node_names=','.join(name_list)) 
    except Exception as error:
        jsn['stderr'] = str(error)
        jsn['excode'] = 1
        return jsn
       
    #ensure that master is target node
    master_name = ''

    for _ in range(timeout):
        try:
            master = es_client.cat.master(format="json")
        except Exception as error:
           jsn['stderr'] = str(error)
           jsn['excode'] = 1
           return jsn
        
        if len(master) > 0:
            master_name = master[0].get("node")
            if target in master_name:
                break
        else:
           jsn['stderr'] = "could not obtain master info"
           jsn['excode'] = 1
           return jsn
        
        time.sleep(1)
    
    try:
        # no matter the result - try to remove voting exclusions, since we done
        es_client.cluster.delete_voting_config_exclusions(wait_for_removal=False)
        if target not in master_name:
            jsn['stderr'] = f'could not switchover master to {target}'
            return jsn
    except Exception as error:
        jsn['stderr'] = str(error)
        jsn['excode'] = 1
        return jsn
    
    return jsn

def switchover_elasticsearch_node_task(data, task_id):
    """switchover node"""
    
    task = 'switchover elasticsearch node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        master_data = json.loads(data['master_data'])
        host = data['host_data']
        node = json.loads(data['node_data'])
        name_list = data['name_list']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node availability'
    duration_step_start = time.perf_counter()

    es_client = connect_to_elasticsearch_insecure(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(es_client, tuple):            
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, node['id'],))
                    conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(es_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(es_client[1]),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'switching over node'
    duration_step_start = time.perf_counter()

    jsn = switchover_node(es_client, node['name'], name_list)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status = %s, ex_params = ex_params || '{"elasticsearch_master_id":%s}', cluster_role=%s WHERE cluster_id=%s""", (Severity.OK.value, Status.OK.value, node['id'], ClusterRoles.STANDBY.value, node['cluster_id']))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status = %s, ex_params = ex_params || '{"elasticsearch_master_id":%s}', cluster_role=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'], ClusterRoles.PRIMARY.value, node['id']))
                cur.execute("""UPDATE "clusterCore_c123_database" SET node_id = %s WHERE node_id=%s""", (node["id"], master_data["id"]))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)