# BELONG TO SERVICE TASK
import json, time, uuid, pytz, os, re, logging
from datetime import datetime
from math import inf as positive_infinity
from redis import Redis, Sentinel
from enums import OperatingSystem, SupportedOS, DBType, ClusterRoles, RestoreOptions, Status, Severity, Inventory, Schedule, LicenseType, Repository, Timeout
from tasks_core import insert_task, find_tilda, remote_command, get_postgres_connection, create_ssh_connection, logger, check_os_version, init_container,\
      get_date_modified_folder, transfer_text_ram, update_failed_object, install_offline_remote_to_remote, TASK_TMP_FILES_DIR, transfer_files_rsync, \
      get_log_rotate_config

#####################################################################
# REDIS TASKS 
#####################################################################

def find_redis_bin(host) -> str:
    """find location of redis bin folder for apt/remi distributions"""
    
    redis_bin = ''
    command = "sudo dirname $(which redis-server)"
    try:
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stdout'] != '':
            redis_bin = jsn['stdout'].strip()

        if jsn['stderr'] != '' and 'permission denied' not in jsn['stderr'].lower():
            if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
                redis_bin = '/usr/bin'
            else:
                redis_bin = '/usr/bin'
    except Exception as error:
        logging.error(f'find_postgres_bin: {str(error)}')
    
    return redis_bin


def check_redis_version_exists(host, redis_version, distributor):
    
    redis_bin = find_redis_bin(host)

    jsn = json.loads(remote_command(host, f"sudo {redis_bin}/redis-server --version | awk '{{print $3}}' | cut -d '=' -f2", True), strict = False)
    if jsn['stderr'] != '':
        cmd = f"sudo rpm -qa | grep 'redis-{redis_version}'" if OperatingSystem.CENTOS.value in distributor else f"sudo apt-cache policy redis | grep 'Installed:'  | xargs | cut -d'-' -f1 | cut -d':' -f3"
        jsn = json.loads(remote_command(host, cmd, True), strict = False)     
    return jsn['stdout'] != '' and str(redis_version) in jsn['stdout']

def connect_to_sentinel(host, user='default', pwd=None, prt=6379, timeout=Timeout.SERVICE_REDIS.value):
    """connecting to redis sentinel"""

    client = Sentinel([(host, prt)], sentinel_kwargs={'username': user, 'password': pwd}, socket_timeout=timeout, socket_connect_timeout=timeout, retry_on_timeout=True)
    try:
        ping = client.execute_command("ping")
    except Exception as e:
        return (False, e)

    if not ping:
        e = f'could not connect to {host}:{prt}'
        return (False, e)

    return client

def connect_to_redis(host, user='default', pwd=None, prt=6379, timeout=Timeout.SERVICE_REDIS.value):
    """connecting to redis cluster"""

    client = Redis(host=host, port=prt, username=user, password=pwd, socket_timeout=timeout, socket_connect_timeout=timeout, retry_on_timeout=True)
    try:
        ping = client.ping()
    except Exception as e:
        return (False, str(e))
        
    if not ping:
        e = f'could not connect to {host}:{prt}'
        return (False, e)

    return client


def install_redis(distributor, release, host, redis_version) -> dict:
    supported_versions = SupportedOS[distributor.upper()].value
    supported_versions_str = ", ".join(map(str, supported_versions))

    if OperatingSystem.CENTOS.value in distributor:
        if int(release) == 7:
            commands = [
                'sudo yum install -q -y tar net-tools epel-release',
                "sudo rpm --import https://rpms.remirepo.net/RPM-GPG-KEY-remi",
                f"sudo yum install -q -y http://rpms.remirepo.net/enterprise/remi-release-{release}.rpm",
                f"sudo yum --enablerepo=remi install -q -y redis-$(sudo yum --enablerepo=remi --showduplicates list redis | grep '{int(redis_version)}\.' | grep {redis_version} | sort -r | head -n1 | awk '{{print $2}}')"
            ]
        elif int(release) == 8:
            sources_dir = f'{TASK_TMP_FILES_DIR}/build/redis{redis_version}/{uuid.uuid4().hex}/'
            redis_link = "https://download.redis.io/releases/"
            commands = [
                'sudo dnf --enablerepo=appstream install -q -y gcc make python3 net-tools which tar rsync',
                f"sudo rm -rf '{sources_dir}'",
                f"sudo mkdir -p '{sources_dir}'",
                f"sudo curl -Lso '{sources_dir}redis.tar.gz' {redis_link}$(curl -s {redis_link} | grep -oE 'redis-{redis_version}\.[0-9]+\.tar\.gz' | sort -V | tail -1)",
                f"sudo tar zxfC '{sources_dir}redis.tar.gz' '{sources_dir}' --strip-components=1",
                f"sudo make -s -C '{sources_dir}'",
                f"sudo make install -s -C '{sources_dir}' 2>&1 | sudo tee '{sources_dir}make_install_log.txt'"
            ]
        elif int(release) == 9:
            commands = [
                'sudo dnf install -q -y epel-release net-tools',
                'sudo dnf install -q -y https://rpms.remirepo.net/enterprise/remi-release-9.rpm',
                'sudo dnf module reset -q -y redis',
                f'sudo dnf module enable -q -y redis:remi-{redis_version}',
                'sudo dnf install -q -y redis'
            ]
        else:
            jsn = {}
            jsn['stdout'] = 'Unknown release version'
            jsn['stderr'] = f'Release version of {distributor.capitalize()} is not {supported_versions_str}',
            jsn['excode'] = 1
            return jsn
    elif OperatingSystem.UBUNTU.value in distributor:
        release_codename_mapping = {
            18: 'bionic',
            20: 'focal',
            22: 'jammy',
        }
        version = release_codename_mapping.get(int(release), 'unknown')
        commands = [
            'sudo DEBIAN_FRONTEND=noninteractive apt-get -qq -y install gpg net-tools',
            "sudo wget -qO - https://packages.redis.io/gpg | sudo gpg -q --batch --yes --dearmor -o '/usr/share/keyrings/redis-archive-keyring.gpg'",
            f"echo 'deb [signed-by=/usr/share/keyrings/redis-archive-keyring.gpg] https://packages.redis.io/deb {version} main' | sudo tee '/etc/apt/sources.list.d/redis.list'",
            'sudo apt-get update',
            f"sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y redis=$(apt-cache policy redis | grep -E '6:{int(redis_version)}\.' | grep '{redis_version}' | head -1 | xargs |  sed 's/500$//' | grep -oE '6:{int(redis_version)}\..*') redis-server=$(apt-cache policy redis-server| grep -E '6:{int(redis_version)}\.' | grep '{redis_version}' | head -1 | xargs |  sed 's/500$//' | grep -oE '6:{int(redis_version)}\..*') redis-tools=$(apt-cache policy redis-tools | grep -E '6:{int(redis_version)}\.' | grep '{redis_version}' | head -1 | xargs |  sed 's/500$//' | grep -oE '6:{int(redis_version)}\..*')"
        ]
    else:
        jsn = {}
        jsn['stderr'] = 'Distributor is not Ubuntu or CentOS'
        jsn['stdout'] = None
        jsn['excode'] = 1
        return jsn
    
    if OperatingSystem.UBUNTU.value in distributor:
        commands += [
            'sudo systemctl stop redis-server',
            'sudo systemctl disable redis-server'
        ]

    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
            if filter:
                return jsn
        except Exception as error:
            return {'stdout': None, 'stderr': str(error), 'excode': 9}
            
    return jsn

def stop_redis_node_task(data, task_id):
    """stop node"""
    
    task = 'stop redis node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        sentinel = node['cluster_role'] == 'SENTINEL'
        redis_version = float(node['version'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node availability'
    duration_step_start = time.perf_counter()

    redis_client = connect_to_redis(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(redis_client, tuple):
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, node['id'],))
                    conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stopping node'
    duration_step_start = time.perf_counter()

    error = stop_redis_node(redis_client, sentinel=sentinel, version=redis_version)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false' WHERE id=%s""", (Severity.NOT_OK.value, Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def sentinel_monitor_master_task(data, task_id):
    """add monitor of certain master (+ its replicas) to certain sentinel node"""

    task = 'sentinel monitor master task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        node = data['node_data']
        sentinel = data['sentinel_data']
        quorum = data['quorum']
        down_ms = data['down_ms']
        failover_timeout = data['failover_timeout']
        parallel_syncs = data['parallel_syncs']
        affected_ids = data['affected_ids']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking sentinel availability'
    duration_step_start = time.perf_counter()

    redis_client = connect_to_redis(sentinel['host'], sentinel['database_user'], sentinel['database_pwd'], sentinel['port'])
    if isinstance(redis_client, tuple):
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status='down', ind_started='false' WHERE id=%s""", (Severity.NOT_OK.value, sentinel['id']))
                    conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'adding node to monitoring'
    duration_step_start = time.perf_counter()

    try:
        masters = redis_client.sentinel_masters()
        exists = next((i for i in masters.values() if i['ip'] == node['host'] and str(i['port']) == node['port']), None)
        if exists:
            redis_client.sentinel_remove(exists['name'])
        redis_client.sentinel_monitor(name=node['alias'], ip=node['host'], port=node['port'], quorum=quorum)
        redis_client.sentinel_set(node['alias'], 'auth-user', node['database_user'])
        redis_client.sentinel_set(node['alias'], 'auth-pass', node['database_pwd'])
        redis_client.sentinel_set(node['alias'], 'down-after-milliseconds', down_ms)
        redis_client.sentinel_set(node['alias'], 'failover-timeout', failover_timeout)
        redis_client.sentinel_set(node['alias'], 'parallel-syncs', parallel_syncs)
    except Exception as e:
        error = e.strerror if hasattr(e, 'strerror') else e.args[0]
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""SELECT ex_params->'monitored_nodes' FROM "clusterCore_c123_node" WHERE id=%s""", [sentinel['id']])
                sentinel_monitors = cur.fetchone()
                current_monitors = {} if not sentinel_monitors[0] else sentinel_monitors[0]
                if not current_monitors.get(node['alias']) or node['id'] not in current_monitors[node['alias']]:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity='{0}', status='{1}', ind_started='true', ind_deleted='false', ex_params = jsonb_set(ex_params,'{{monitored_by}}', COALESCE(ex_params->'monitored_by', '[]'::jsonb) || '["{2}"]'::jsonb, true) WHERE id={3}""".format(Severity.OK.value, Status.OK.value, sentinel['id'], node['id']))
                
                current_monitors[node['alias']] = affected_ids
                cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = jsonb_set(ex_params, '{{monitored_nodes}}', %s::jsonb) WHERE id = {0}""".format(sentinel['id']), [json.dumps(current_monitors)])
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def sentinel_remove_monitor_task(data, task_id):
    """remove monitor of certain master (+ its replicas) from certain sentinel node"""
    
    task = 'sentinel remove monitor task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        node = data['node_data']
        sentinel = data['sentinel_data']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking sentinel availability'
    duration_step_start = time.perf_counter()

    redis_client = connect_to_sentinel(sentinel['host'], sentinel['database_user'], sentinel['database_pwd'], sentinel['port'])
    if isinstance(redis_client, tuple):
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, sentinel['id'],))
                    conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'removing node from monitoring'
    duration_step_start = time.perf_counter()

    try:
        redis_client.sentinel_remove(node['alias'])
    except Exception as e:
        error = e.strerror if hasattr(e, 'strerror') else e.args[0]
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""SELECT ex_params->'monitored_nodes' FROM "clusterCore_c123_node" WHERE id=%s""", [sentinel['id']])
                sentinel_monitors = cur.fetchone()
                current_monitors = {} if not sentinel_monitors[0] else sentinel_monitors[0]
                ids = current_monitors.pop(node['alias'], None)
                if ids:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = jsonb_set(ex_params, '{{monitored_by}}', (ex_params->'monitored_by') - '{0}') WHERE id=ANY(%s)""".format(sentinel['id']), [ids])
                
                cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = jsonb_set(ex_params, '{{monitored_nodes}}', %s::jsonb) WHERE id = {0}""".format(sentinel['id']), [json.dumps(current_monitors)])
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

# def sentinel_reset_task(data, task_id):
#     """force sentinel node to rediscover certain master (+ its replicas)"""
    
#     task = 'sentinel reset task'
    
#     try:
#         data = json.loads(data)
#         created_by = int(data['created_by'])
#     except Exception as error:
#         logging.error(f"{task}: {str(error)}")
#         return

#     task_log_id = uuid.uuid4().hex
#     step = 'null'
#     status = Severity.OK.value
#     duration = 0
#     logger(task_id,task_log_id,task,step,status,duration,created_by)
#     duration_task_start = datetime.now()

#     step = 'initializing task'
#     status = Severity.OK.value
#     duration = 0
#     logger(task_id,task_log_id,task,step,status,duration,created_by)

#     step = 'checking parameters'
#     duration_step_start = time.perf_counter()

#     try:
#         node = json.loads(data['node_data'])
#         sentinel = json.loads(data['sentinel_data'])
#     except Exception as error:
#         status = Severity.NOT_OK.value
#         duration_step_end = time.perf_counter() - duration_step_start
#         logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
#         step = 'null'
#         status = Severity.NOT_OK.value
#         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
#         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
#         return

#     status = Severity.OK.value
#     duration_step_end = time.perf_counter() - duration_step_start
#     logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

#     step = 'checking sentinel availability'
#     duration_step_start = time.perf_counter()

#     redis_client = connect_to_sentinel(sentinel['host'], sentinel['database_user'], sentinel['database_pwd'], sentinel['port'])
#     if isinstance(redis_client, tuple):
#         try:
#             with get_postgres_connection() as conn:
#                 with conn.cursor() as cur:
#                     cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, sentinel['id'],))
#                     conn.commit()
#         except Exception as error:
#             status = Severity.NOT_OK.value
#             duration_step_end = time.perf_counter() - duration_step_start
#             logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
#             step = 'null'
#             status = Severity.NOT_OK.value
#             duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
#             logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
#             return

#         status = Severity.NOT_OK.value
#         duration_step_end = time.perf_counter() - duration_step_start
#         logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
#         step = 'null'
#         status = Severity.NOT_OK.value
#         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
#         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
#         return

#     status = Severity.OK.value
#     duration_step_end = time.perf_counter() - duration_step_start
#     logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

#     step = 'resetting sentinel'
#     duration_step_start = time.perf_counter()

#     try:
#         redis_client.sentinel_reset(node['alias'])
#     except Exception as e:
#         error = e.strerror if hasattr(e, 'strerror') else e.args[0]
#         status = Severity.NOT_OK.value
#         duration_step_end = time.perf_counter() - duration_step_start
#         logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
#         step = 'null'
#         status = Severity.NOT_OK.value
#         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
#         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
#         return

#     status = Severity.OK.value
#     duration_step_end = time.perf_counter() - duration_step_start
#     logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
#     try:
#         with get_postgres_connection() as conn:
#             with conn.cursor() as cur:
#                 cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"monitored_by":""}' WHERE id=%s""", (node['id'],))
#                 conn.commit()
#     except Exception as error:
#         step = 'null'
#         status = Severity.NOT_OK.value
#         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
#         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
#         return

#     step = 'task finished'
#     status = Severity.OK.value
#     duration = 0
#     logger(task_id,task_log_id,task,step,status,duration,created_by)
#     step = 'null'
#     status = Severity.OK.value
#     duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
#     logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def sentinel_failover_task(data, task_id):
    """force sentinel node to initiate failover of certain master (replace with its random replica)"""
    
    task = 'sentinel failover task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        node = data['node_data']
        sentinel = data['sentinel_data']
        other_nodes = data['other_nodes']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking sentinel availability'
    duration_step_start = time.perf_counter()

    redis_client = connect_to_sentinel(sentinel['host'], sentinel['database_user'], sentinel['database_pwd'], sentinel['port'])
    if isinstance(redis_client, tuple):
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, sentinel['id'],))
                    conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stepping down node'
    duration_step_start = time.perf_counter()

    try:
        redis_client.sentinel_failover(node['alias'])
        master = redis_client.discover_master(node['alias'])
    except Exception as e:
        error = e.strerror if hasattr(e, 'strerror') else e.args[0]
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                master_id = next((i['id'] for i in other_nodes if i['host'] == master[0] and i['port'] == int(master[1])), None)
                if not master_id:
                    master_id = node['id']
                
                slave_ids = [i['id'] for i in other_nodes if i['id'] != master_id]
                
                if master_id != node['id']:
                    slave_ids.append(node['id'])
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"slave_of":%s}', cluster_role=%s WHERE id=ANY(%s)""", (master_id, ClusterRoles.STANDBY.value, slave_ids))
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"slave_of":%s}', cluster_role=%s WHERE id=%s""", (False, ClusterRoles.PRIMARY.value, master_id))
                
                cur.execute("""SELECT ex_params->'monitored_by' FROM "clusterCore_c123_node" WHERE id=%s""", (node['id'],))
                sentinels = cur.fetchone()
                current_sentinels = [] if not sentinels[0] else sentinels[0]
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity='{0}', status='', ind_started='true', ind_deleted='false', ex_params = jsonb_set(ex_params, '{{monitored_by}}', %s::jsonb) WHERE id = ANY(%s)""".format(Severity.OK.value), (json.dumps(current_sentinels), slave_ids))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def install_redis_standalone_task(data, task_id):
    """Redis standalone installation"""

    task = 'install redis standalone task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        app_host = json.loads(data['app_host'])
        # app_hostname = data['app_hostname']
        create_run_linux = data.get('create_run_linux')
        # repository = network or local
        repository = data['repository']
        offline = True if Repository.LOCAL.value in repository else False
        redis_version = float(node['version'])
        has_aof = node["ex_params"].get("has_aof")
        host_exists = True
        proxmox_container = node['ex_params'].get('proxmox')
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = node['ex_params'].get('distributor')
                release = node['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
        
        version_exists = check_redis_version_exists(host, redis_version, distributor)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)


    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

   

    step = 'running installation commands'
    duration_step_start = time.perf_counter()

    if not version_exists:
        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host, redis_version, DBType.REDIS.value, app_host)
        else:
            jsn = install_redis(distributor, release, host, redis_version)

        err_lower = jsn['stderr'].lower()
        if 'E:' in jsn['stderr'] or any(err in err_lower for err in ['error', 'err', 'fail', 'errno'] if not any(err in err_lower for err in ['already installed', 'nothing to do', 'no such file', 'file not found'])):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node['id'])
            return

    jsn = create_redis_node(host, node)
    if jsn['stderr'] != '' and ('warning' not in jsn['stderr'].lower()):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return
    
    time.sleep(5)
    redis_client = connect_to_redis(host=host["name"], prt=node['port'])
    if isinstance(redis_client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
        return

    try:
        redis_client.acl_setuser(username=node["database_user"], passwords=[f"+{node['database_pwd']}"], keys=["*"], channels=["*"], commands=["+@all"], enabled=True)
        redis_client.acl_setuser(username="default", enabled=False)
        if has_aof:
            redis_client.config_set("appendonly", "yes")
        
        redis_client.config_rewrite()
        redis_client.client_kill_filter(user="default")
    except Exception as e:
        error = e.strerror if hasattr(e, 'strerror') else e.args[0]
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        
        node_size = float(jsn['stdout'].strip())/1024
        remote_command(host, f"""echo "{get_log_rotate_config(node['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{node['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host), 'node_name': node['name']}, 'task': 'log_rotate_task',  'org_id': node['org_id'], 'monitor_name': None,  'target_type': node['type'], 'target_id': node['id'], 'created_by': created_by}
        insert_task(params)
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size,  node['id'],))
                conn.commit()
                # cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,node['id'],created_by])
                # conn.commit()
                # cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host['id'],created_by,OperatingSystem.LINUX.value])
                # conn.commit()
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def add_redis_sentinel_task(data, task_id):
    """add redis sentinel node to cluster"""

    task = 'add redis sentinel task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        app_host = json.loads(data['app_host'])
        node = json.loads(data['node_data'])
        cluster = json.loads(data["cluster_data"])
        # app_hostname = data['app_hostname']
        create_run_linux = data.get('create_run_linux')
        # repository = network or local
        repository = data['repository']
        offline = True if Repository.LOCAL.value in repository else False
        redis_version = float(node['version'])
        host_exists = True
        proxmox_container = node['ex_params'].get('proxmox')
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = node['ex_params'].get('distributor')
                release = node['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
        
        version_exists = check_redis_version_exists(host, redis_version, distributor)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)


    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    
    step = 'running installation commands'
    duration_step_start = time.perf_counter()

    if not version_exists:
        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host, redis_version, DBType.REDIS.value, app_host)
        else:
            jsn = install_redis(distributor, release, host, redis_version)

        err_lower = jsn['stderr'].lower()
        if 'E:' in jsn['stderr'] or any(err in err_lower for err in ['error', 'err', 'fail', 'errno'] if not any(err in err_lower for err in ['no such file', 'file no found', 'already installed', 'nothing to do'])):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node['id'])
            return

    jsn = create_redis_node(host, node, False)
    if jsn['stderr'] !=  '' and ('warning' not in jsn['stderr'].lower()):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return
    
    time.sleep(5)
    redis_client = connect_to_redis(host = host["name"], prt=node['port'], pwd=cluster["ex_params"].get("database_user"))
    if isinstance(redis_client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
        return
    
    try:
        redis_client.acl_setuser(username=cluster['ex_params'].get('database_user'), passwords=[f"+{cluster['ex_params'].get('database_pwd')}"], channels=["*"], commands=["+@all"], enabled=True)
        redis_client.acl_setuser(username="default", enabled=False)
        redis_client.execute_command("sentinel config set", "sentinel-user", cluster['ex_params'].get('database_user'))
        redis_client.execute_command("sentinel config set", "sentinel-pass", cluster['ex_params'].get('database_pwd'))
        redis_client.client_kill_filter(user = "default")
    except Exception as e:
        error = e.strerror if hasattr(e, 'strerror') else e.args[0]
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()

    try:
        
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        remote_command(host, f"""echo "{get_log_rotate_config(node['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{node['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host), 'node_name': node['name']}, 'task': 'log_rotate_task',  'org_id': node['org_id'], 'monitor_name': None,  'target_type': node['type'], 'target_id': node['id'], 'created_by': created_by}
        insert_task(params)
        
        node_size = float(jsn['stdout'].strip())/1024
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                conn.commit()
                # cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,node['id'],created_by])
                # conn.commit()
                # cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host['id'],created_by,OperatingSystem.LINUX.value])
                # conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def create_redis_cluster_task(data, task_id):
    """Redis cluster installation"""

    task = 'create redis cluster task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        master_list = [{"host": json.loads(master["host"]), "node": json.loads(master["node"])} for master in data['master_list']]
        cluster_data = json.loads(data["cluster_data"])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'forming base cluster'
    duration_step_start = time.perf_counter()

    redis_bin = find_redis_bin(master_list[0]["host"])

    try:
        jsn = json.loads(remote_command(master_list[0]["host"], f"sudo {redis_bin}/redis-cli -u 'redis://{cluster_data['ex_params'].get('database_user')}:{cluster_data['ex_params'].get('database_pwd')}@{master_list[0]['host']['name']}:{master_list[0]['node']['port']}' --cluster create {master_list[0]['host']['name']}:{master_list[0]['node']['port']} {master_list[1]['host']['name']}:{master_list[1]['node']['port']} {master_list[2]['host']['name']}:{master_list[2]['node']['port']} --cluster-yes", True), strict=False)
        error_lower = jsn['stderr'].lower()
        if error_lower != '' and 'err' in error_lower:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    
    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"severity":\"%s\"}' WHERE id=%s""" % (Severity.OK.value,cluster_data['id'],))
                cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"status":\"\"}' WHERE id=%s""", (cluster_data['id'],))
                conn.commit()
    
        for master in master_list:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, master['node']['id'],))
                conn.commit()
                # cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,master['node']['id'],created_by])
                # conn.commit()
                # cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,master['host']['id'],created_by,OperatingSystem.LINUX.value])
                # conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
    
    # host['create_task'] = created_by
    # host['ssh_port'] = int(host['ssh_port'])
    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host, 'task': 'run_linux_performance_counters', 'target_type': 'host', 'target_id': host['id'], 'created_by': created_by}               
    #         insert_task(params)
    #     except Exception as error:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def add_redis_node_task(data, task_id):
    """configure redis node for cluster"""

    task = 'add redis node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        master_list = [{"host": json.loads(master["host"]), "node": json.loads(master["node"])} for master in data['master_list']]
        cluster_data = json.loads(data["cluster_data"])
        node = json.loads(data["node_data"])
        host = json.loads(data["host_data"])
        app_host = json.loads(data['app_host'])
        proxmox_container = node['ex_params'].get('proxmox')
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = node['ex_params'].get('distributor')
                release = node['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
        
        repository = data['repository']
        redis_version = float(cluster_data["ex_params"].get("version"))
        version_exists = check_redis_version_exists(host, redis_version, distributor)
        offline = True if Repository.LOCAL.value in repository else False
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)


    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    step = 'setting up node for cluster'
    duration_step_start = time.perf_counter()

    if not version_exists:
        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host, redis_version, DBType.REDIS.value, app_host)
        else:
            jsn = install_redis(distributor, release, host, redis_version)

        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'no such file', 'file not found'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node['id'])
            return

    jsn = create_redis_node(host, node, cluster_data)
    if jsn['stderr'] !=  '' and ('warning' not in jsn['stderr'].lower()):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return

    time.sleep(1)

    redis_client = connect_to_redis(host = host['name'], prt=node['port'])
    if isinstance(redis_client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
        return

    try:
        redis_client.acl_setuser(username=cluster_data["ex_params"].get("database_user"), passwords=[f"+{cluster_data['ex_params'].get('database_pwd')}"], keys=["*"], channels=["*"], commands=["+@all"], enabled=True)
        redis_client.acl_setuser(username="default", enabled=False)
        mode = redis_client.info().get("redis_mode")
        if mode:
            if 'sentinel' not in mode:
                redis_client.config_set("appendonly", "yes")
            if 'cluster' in mode:
                redis_client.config_set("masteruser", cluster_data["ex_params"].get('database_user'))
                redis_client.config_set("masterauth", cluster_data["ex_params"].get('database_pwd'))
            
            redis_client.sentinel_flushconfig() if 'sentinel' in mode else redis_client.config_rewrite()
        else:
            redis_client.config_rewrite()
        redis_client.client_kill_filter(user="default")
    except Exception as e:
        error = e.strerror if hasattr(e, 'strerror') else e.args[0]
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.WARNING.value, Status.PENDING.value, node['id'],))
                conn.commit()
                cur.execute("""SELECT COUNT(*) FROM "clusterCore_c123_node" WHERE cluster_id=%s AND cluster_role=%s AND ind_deleted = 'false'""", (cluster_data['id'], ClusterRoles.PRIMARY.value))
                create_cluster = int(cur.fetchone()[0]) > 2
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking cluster health'
    duration_step_start = time.perf_counter()
    
    other_masters = [i for i in master_list if i["node"]["name"] != node["name"]]
    master_up = other_masters[:-1] # take the last other master, as failback
    cluster_nodes = []
    slots = []
    redis_client = "no masters found to check cluster's health"
    cluster_state = None
    
    if len(other_masters) > 0:
        for master in master_list:
            redis_client = connect_to_redis(host=master['host']['name'], user=cluster_data["ex_params"].get("database_user"), prt=master['node']['port'], pwd=cluster_data["ex_params"].get("database_pwd"), timeout=10)
            if isinstance(redis_client, tuple):
                redis_client = str(redis_client[1])
                continue
            else:
                master_up = master
                break
    
    if redis_client and "no masters" not in str(redis_client).lower():
        try:
            cluster_state = redis_client.cluster('info').get('cluster_state')
            cluster_nodes = redis_client.cluster('nodes')
            slots = redis_client.cluster('slots')
            status = Severity.OK.value
        except Exception as e:
            error = e.strerror if hasattr(e, 'strerror') else e.args[0]
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    else:
        status = Severity.WARNING.value

    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if Severity.OK.value in status else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client),1)
    
    #if there are three masters and at least one of them are empty node - we can try create cluster
    if (not cluster_nodes or len(cluster_nodes) < 2) and (not slots or len(slots) == 0) and create_cluster:
        step = 'forming base cluster'
        duration_step_start = time.perf_counter()

        try:
            redis_bin = find_redis_bin(master_up["host"])
            master_addresses = [f"{i['host']['name']}:{i['node']['port']}" for i in master_list]
            jsn = json.loads(remote_command(master_up["host"], f"sudo {redis_bin}/redis-cli -u 'redis://{cluster_data['ex_params'].get('database_user')}:{cluster_data['ex_params'].get('database_pwd')}@{master_up['host']['name']}:{master_up['node']['port']}' --cluster create {' '.join(master_addresses)} --cluster-yes", True), strict=False)
            error_lower = jsn['stderr'].lower()
            if error_lower != '' and 'err' in error_lower:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
     
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        step = "updating 123cluster app metadata"
        duration_step_start = time.perf_counter()
         
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"severity":\"%s\"}' WHERE id=%s""" % (Severity.OK.value, cluster_data['id'],))
                    cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"status":\"\"}' WHERE id=%s""", (cluster_data['id'],))
                    conn.commit()
                    
                    for master in master_list:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, master['node']['id'],))
                        conn.commit()
                        # cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,master['node']['id'],created_by])
                        # conn.commit()
                        # cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,master['host']['id'],created_by,OperatingSystem.LINUX.value])
                        # conn.commit()
        
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
                      
        except Exception as error:
           status = Severity.WARNING.value
           duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
    
    # if cluster is initialized
    if ((slots and len(slots) > 0) or (cluster_nodes and len(cluster_nodes) > 1)):
        if cluster_state and 'fail' in cluster_state:
            step = 'repairing cluster'
            duration_step_start = time.perf_counter()
            
            try:
                command = f"echo 'yes' | sudo {redis_bin}/redis-cli -u 'redis://{master_up['node']['database_user']}:{master_up['node']['database_pwd']}@{master_up['host']['name']}:{master_up['node']['port']}' --cluster fix {master_up['host']['name']}:{master_up['node']['port']}"
                jsn = json.loads(remote_command(master_up['host'], command, True), strict=False)
                if error_lower != '' and "err" in error_lower:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        
        if mode and 'cluster' in mode:
            step = 'joining existing cluster'
            duration_step_start = time.perf_counter()

            try:
                if not node["ex_params"].get("slave_of"):
                    redis_bin = find_redis_bin(master_up["host"])
                    jsn = json.loads(remote_command(master_up["host"], f"sudo {redis_bin}/redis-cli -u 'redis://{cluster_data['ex_params'].get('database_user')}:{cluster_data['ex_params'].get('database_pwd')}@{master_up['host']['name']}:{master_up['node']['port']}' --cluster add-node {host['name']}:{node['port']} {master_up['host']['name']}:{master_up['node']['port']} --cluster-yes", True), strict=False)
                    error_lower = jsn['stderr'].lower()
                    if error_lower != '' and "err" in error_lower:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                else:
                    master_id = node['ex_params'].get('slave_of')
                    target_master = next((i for i in master_list if i['node']['id'] == node['ex_params'].get('slave_of')), None)
                    if target_master:
                        redis_client = connect_to_redis(host=target_master['host']['name'], prt=target_master['node']['port'], user=cluster_data["ex_params"].get("database_user"), pwd=cluster_data["ex_params"].get("database_pwd"))
                        if isinstance(redis_client, tuple):
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
                            return

                        try:
                            id = redis_client.cluster("myid").decode("ascii").strip()
                        except Exception as e:
                            error = e.strerror if hasattr(e, 'strerror') else e.args[0]
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                            return

                        redis_bin = find_redis_bin(target_master["host"])
                        jsn = json.loads(remote_command(target_master["host"], f"sudo {redis_bin}/redis-cli -u 'redis://{cluster_data['ex_params'].get('database_user')}:{cluster_data['ex_params'].get('database_pwd')}@{target_master['host']['name']}:{target_master['node']['port']}' --cluster add-node {host['name']}:{node['port']} {target_master['host']['name']}:{target_master['node']['port']} --cluster-slave --cluster-master-id {id} --cluster-yes", True), strict=False)
                        error_lower = jsn['stderr'].lower()
                        if error_lower != '' and "err" in error_lower:
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return
                    else:
                        error = f"could not found target master {master_id} for node {node['name']}"
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                        return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        remote_command(host, f"""echo "{get_log_rotate_config(node['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{node['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host), 'node_name': node['name']}, 'task': 'log_rotate_task',  'org_id': node['org_id'], 'monitor_name': None,  'target_type': node['type'], 'target_id': node['id'], 'created_by': created_by}
        insert_task(params)
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                if not node["ex_params"].get("slave_of"):
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                else:
                    cur.execute("""SELECT ex_params->'monitored_by' FROM "clusterCore_c123_node" WHERE id=%s""", [target_master['node']['id']])
                    sentinels = cur.fetchone()
                    current_sentinels = [] if not sentinels[0] else sentinels[0]
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity='{0}', status='{1}', ind_started='true', ind_deleted='false', size={2} ex_params = jsonb_set(ex_params, '{{monitored_by}}', %s::jsonb) WHERE id = {3}""".format(Severity.OK.value, Status.OK.value, node_size, node['id']), [json.dumps(current_sentinels)])
                    if len(current_sentinels) > 0:
                        cur.execute("""SELECT id, ex_params->'monitored_nodes' FROM "clusterCore_c123_node" WHERE id=ANY(%s)""", ([int(i) for i in current_sentinels],))
                        for i in cur.fetchall():
                            if i[1]:
                                for k, v in i[1].items():
                                    if target_master['node']['id'] in v and node['id'] not in v:
                                        i[1][k].append(node['id'])
                                
                                cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = jsonb_set(ex_params, '{{monitored_nodes}}', %s::jsonb) WHERE id ={0}""".format(i[0]), [json.dumps(i[1])])
                conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def start_redis_node_task(data, task_id):
    """start node"""
    
    task = 'start redis node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        node = json.loads(data['node_data'])
        sentinel = node['cluster_role'] == 'SENTINEL'
        conf_dir = f"{node['dir_123Cluster']}/conf"
        conf_file = f"{conf_dir}/{'sentinel' if sentinel else 'redis'}.conf"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node availability'
    duration_step_start = time.perf_counter()

    redis_client = connect_to_redis(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(redis_client, tuple):
        step = 'checking remote availability'
        duration_step_start = time.perf_counter()

        client = create_ssh_connection(host['name'], host['username'], host['password'], host['private_key'], host['ssh_port'])
        if isinstance(client, tuple):
            if hasattr(client[1], 'strerror'):
                error = client[1].strerror
            else:
                error = str(client[1])

            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        step = 'starting node'
        duration_step_start = time.perf_counter()

        redis_bin = find_redis_bin(host)
        command = f"sudo nohup {redis_bin}/redis-{'sentinel' if sentinel and OperatingSystem.CENTOS.value in distributor else 'server'} '{conf_file}' {'--sentinel' if sentinel and OperatingSystem.UBUNTU.value in distributor else ''} &> /dev/null &"
        
        try:
            stdin, stdout, stderr = client.exec_command(command)
            exit_status = stdout.channel.recv_exit_status() # Blocking call
            stdin.channel.shutdown_write()
            try:
                stderr = stderr.read().decode('ascii').strip()
                stdout = stdout.read().decode('ascii').strip("\n")
            except UnicodeDecodeError:
                stderr = stderr.read().decode('utf-8').strip()
                stdout = stdout.read().decode('utf-8').strip("\n")

            if stderr != '' and 'error' in stderr:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,str(stderr),exit_status)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,str(stderr),exit_status)
                return
            
            client.close()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        time.sleep(1)
        redis_client = connect_to_redis(host['name'], node['database_user'], node['database_pwd'], node['port'])
        if isinstance(redis_client, tuple):
            try:
                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, node['id'],))
                        conn.commit()
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

            error = redis_client[1]

            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def stop_redis_node(client, version, force=False, sentinel=False) -> str:
    error = ''
    if force:
        save = False
        lower_than_7 = not version or version < 7.0
        now = None if lower_than_7 else True
        force = None if lower_than_7 else True
    else:
        save = True
        now = None
        force = None

    try:
        client.shutdown() if sentinel else client.shutdown(save=save, now=now, force=force)
    except Exception as error:
        return str(error)
    
    return error

def drop_redis_node_task(data, task_id):
    """drop node"""
    
    task = 'drop redis node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        data_only = data.get('data_only')
        node = json.loads(data['node_data'])
        final_backup = data.get('final_backup')
        backup_host = data['backup_host']
        redis_version = float(node['version'])
        other_nodes = data.get('other_nodes')
        other_nodes_exists = isinstance(other_nodes, list) and len(other_nodes) > 0
        if other_nodes_exists:
            other_nodes = [{"host": json.loads(i['host']), "node": json.loads(i['node'])} for i in data['other_nodes']]
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'checking node existence'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node['port']}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    if jsn['stdout'] == '':
        try:
            jsn = json.loads(remote_command(host, f"sudo rm -rf '{node['dir_123Cluster']}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
        
        status = Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'{node["name"]} is already deleted', 1)

        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    if node['cluster_role'] == 'SENTINEL':
                        if len(other_nodes) > 0:
                            for i in other_nodes:
                                monitored_by = i['node']['ex_params'].get('monitored_by')
                                if monitored_by:
                                    if len(monitored_by) > 0:
                                        cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = jsonb_set(ex_params, '{{monitored_by}}',  (ex_params->'monitored_by') - '{0}') WHERE id=%s""".format(node['id']), [i['node']['id']])
                        
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity='{0}', status='{1}', ind_started='false', ind_deleted='true', ex_params = jsonb_set(ex_params, '{{monitored_nodes}}', %s::jsonb) WHERE id = {2}""".format(Severity.NOT_OK.value, Status.OK.value, node['id']), [json.dumps({})])
                    else:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.OK.value, node['id'],))
                    
                    conn.commit()
        except Exception as error:
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        step = 'task finished'
        status = Severity.WARNING.value
        duration = 0
        logger(task_id,task_log_id,task,step,status,duration,created_by)
        step = 'null'
        status = Severity.WARNING.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
        return
    
    step = 'checking node availability'
    duration_step_start = time.perf_counter()

    target_client = connect_to_redis(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(target_client, tuple):
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, node['id'],))
                    conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(target_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(target_client[1]),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    if node['cluster_role'] != ClusterRoles.PRIMARY.value:
        final_backup = False
    if final_backup:
        step = 'backing up node'
        duration_step_start = time.perf_counter()

        try:
            last_save = target_client.info("persistence")['rdb_last_save_time']
            target_client.bgsave()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        save_finished = False 

        while not save_finished:
            try:
                persistence_info = target_client.info("persistence")
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
            
            save_finished = persistence_info['rdb_bgsave_in_progress'] == 0 and persistence_info['rdb_last_bgsave_status'] == "ok" and persistence_info['rdb_last_save_time'] - last_save > 0
            save_failed = persistence_info['rdb_bgsave_in_progress'] == 0 and persistence_info['rdb_last_bgsave_status'] != "ok" and persistence_info['rdb_last_save_time'] - last_save == 0
            if save_failed:
                error = "save failed"
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
            
            time.sleep(2)
    
        try:
            backup_path = final_backup['ex_params'].get('backup_path')
            commands = [
                f"sudo rm -rf '{backup_path}'",
                f"sudo mkdir -p '{backup_path}'",
                f"sudo tar zcfC '{backup_path}/{final_backup['name']}.tar.gz' '{node['dir_123Cluster']}/data' dump.rdb"
            ]
            for cmd in commands:
                jsn = json.loads(remote_command(host, cmd, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            

            jsn = transfer_files_rsync(host, backup_host, backup_path, [f"{backup_path}/"])
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            jsn = json.loads(remote_command(backup_host,f"sudo du -b '{backup_path}/{final_backup['name']}.tar.gz' | awk '{{print $1}}'", True), strict=False) 
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            backup_size = round(float(jsn['stdout'])/1024, 2)
            modified_date = get_date_modified_folder(hostdata=backup_host, folder_path=backup_path)
            if isinstance(modified_date, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
                return
            
            remote_command(host, f"sudo rm -rf '{backup_path}'", True)

            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}', status='true' WHERE id=%s""" % (modified_date, final_backup['id']))
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size, final_backup['id']))
                    conn.commit()

            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        if not data_only:
            step = 'checking node type'
            duration_step_start = time.perf_counter()

            is_standalone = False
            redis_bin = find_redis_bin(host)

            try:
                id = target_client.cluster("myid").decode("ascii").strip()
            except Exception as e:
                error = e.strerror if hasattr(e, 'strerror') else e.args[0]
                if "cluster support disabled" not in error.lower():
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                    return
                else:
                    is_standalone = True

            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

            if not is_standalone:
                other_masters = None
                master_slaves = None
                is_master = False
                try:
                    role = target_client.info().get('role')
                    is_master = role and 'master' in role
                    if is_master:
                        cluster_info = target_client.cluster('nodes')
                        if isinstance(cluster_info, dict):
                            cluster_nodes = cluster_info.items()
                            cluster_other_masters = [k for k, v in cluster_nodes if 'master' in v["flags"] and 'fail' not in v["flags"] and node["name"] not in k]
                            cluster_master_slaves = [k for k, v in cluster_nodes if 'slave' in v["flags"] and 'fail' not in v["flags"] and id in v["master_id"]]
                            if other_nodes_exists:
                                other_masters = [i for i in other_nodes if i["node"]["name"] in cluster_other_masters]
                                master_slaves = [i for i in other_nodes if i["node"]["name"] in cluster_master_slaves]
                except Exception as e:
                    error = e.strerror if hasattr(e, 'strerror') else e.args[0]
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                    return
                
                if is_master:
                    other_masters_exists = isinstance(other_masters, list) and len(other_masters) > 0
                    master_slaves_exists = isinstance(master_slaves, list) and len(master_slaves) > 0
                    if not master_slaves_exists:
                        if other_masters_exists:
                            step = 'migrating data to other master'
                            duration_step_start = time.perf_counter()

                            # default - failback to last other master
                            master_up = other_masters[:-1]

                            for i in other_masters:
                                node_client = connect_to_redis(i['host']['name'], i['node']['database_user'], i['node']['database_pwd'], i['node']['port'])
                                if not isinstance(node_client, tuple):
                                    master_up = i
                                    break

                            try:
                                # check from receiver - its existence are more important for data migration
                                command = f"sudo {redis_bin}/redis-cli -u 'redis://{master_up['node']['database_user']}:{master_up['node']['database_pwd']}@{master_up['host']['name']}:{master_up['node']['port']}' --cluster check {master_up['host']['name']}:{master_up['node']['port']}"
                                jsn = json.loads(remote_command(master_up['host'], command, True), strict=False)
                                error_lower = jsn['stderr'].lower()
                                if error_lower != '' and "err" in error_lower:
                                    status = Severity.NOT_OK.value
                                    duration_step_end = time.perf_counter() - duration_step_start
                                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                                    step = 'null'
                                    status = Severity.NOT_OK.value
                                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                                    return

                                output = jsn['stdout']
                                # Define the sender and receiver target strings
                                sender_target = f"{host['name']}:{node['port']}"
                                receiver_target = f"{master_up['host']['name']}:{master_up['node']['port']}"

                                # Regular expressions to match the sender and receiver lines
                                sender_pattern = re.compile(f"M: .*? {re.escape(sender_target)}")
                                receiver_pattern = re.compile(f"M: .*? {re.escape(receiver_target)}")

                                # Extract the hexadecimal id for sender
                                sender_match = sender_pattern.search(output)
                                if sender_match:
                                    sender_hex_id = sender_match.group().split()[-2]

                                # Extract the hexadecimal id for receiver
                                receiver_match = receiver_pattern.search(output)
                                if receiver_match:
                                    receiver_hex_id = receiver_match.group().split()[-2]

                                # Extract the number of slots for sender
                                slots_pattern = re.compile(re.escape(sender_target) + r" \(.*?(\d+) slots")
                                slots_match = slots_pattern.search(output)
                                if slots_match:
                                    num_slots = int(slots_match.group(1))

                                # if collected migration info
                                if sender_match and receiver_match and slots_match:
                                    # if node has slots assigned
                                    if num_slots > 0:
                                        command = f"sudo {redis_bin}/redis-cli -u 'redis://{node['database_user']}:{node['database_pwd']}@{host['name']}:{node['port']}' --cluster reshard {host['name']}:{node['port']}  --cluster-from {sender_hex_id} --cluster-to {receiver_hex_id} --cluster-slots {num_slots} --cluster-yes"
                                        jsn = json.loads(remote_command(host, command, True), strict=False)
                                        error_lower = jsn['stderr'].lower()
                                        if error_lower != '' and "err" in error_lower:
                                            status = Severity.NOT_OK.value
                                            duration_step_end = time.perf_counter() - duration_step_start
                                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                                            step = 'null'
                                            status = Severity.NOT_OK.value
                                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                                            return
                                else:
                                    error = "Unable to obtain the parameters for data migration. Aborting."
                                    status = Severity.NOT_OK.value
                                    duration_step_end = time.perf_counter() - duration_step_start
                                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                                    step = 'null'
                                    status = Severity.NOT_OK.value
                                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                                    return
                            except Exception as error:
                                status = Severity.NOT_OK.value
                                duration_step_end = time.perf_counter() - duration_step_start
                                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                                step = 'null'
                                status = Severity.NOT_OK.value
                                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                                return
                                
                            status = Severity.OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
                    else:
                        try:
                            for i in master_slaves:
                                slave_host = i['host']
                                slave_node = i['node']
                                slave_client = connect_to_redis(slave_host['name'], slave_node['database_user'], slave_node['database_pwd'], int(slave_node['port']))
                                if isinstance(slave_client, tuple):
                                    continue
                                 
                                slave_client.cluster("failover force")
                                with get_postgres_connection() as conn:
                                    with conn.cursor() as cur:
                                        id_list = [s['node']['id'] for s in master_slaves if s['node']['name'] != slave_node['name']]
                                        id_list.append(node['id'])
                                        cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"slave_of":%s}', cluster_role=%s WHERE id=ANY(%s)""", (i['node']['id'], ClusterRoles.STANDBY.value, id_list))
                                        cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"slave_of":%s}', cluster_role=%s WHERE id=%s""", (False, ClusterRoles.PRIMARY.value, node['id']))
                                        cur.execute("""SELECT ex_params->'monitored_by' FROM "clusterCore_c123_node" WHERE id=%s""", (node['id'],))
                                        sentinels = cur.fetchone()
                                        current_sentinels = [] if not sentinels[0] else sentinels[0]
                                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity='{0}', status='{1}', ind_started='true', ind_deleted='false', ex_params = jsonb_set(ex_params, '{{monitored_by}}', %s::jsonb) WHERE id = ANY(%s)""".format(Severity.OK.value, Status.OK.value), (json.dumps(current_sentinels), id_list))
                                        conn.commit()
                                break
                        except Exception as e:
                            error = e.strerror if hasattr(e, 'strerror') else e.args[0]
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                            return
                        
                if other_nodes_exists:
                    step = 'removing node from config'
                    duration_step_start = time.perf_counter()

                    for i in other_nodes:
                        node_client = connect_to_redis(i['host']['name'], i['node']['database_user'], i['node']['database_pwd'], i['node']['port'])
                        if isinstance(node_client, tuple):
                            #TODO just skip unavailable nodes, or try to resurrect, or abort task?
                            pass
                        
                        try:
                            node_client.execute_command("cluster forget", id)
                        except Exception as e:
                            error = e.strerror if hasattr(e, 'strerror') else str(e.args[0])
                            if 'unknown node' not in error.lower():
                                status = Severity.NOT_OK.value
                                duration_step_end = time.perf_counter() - duration_step_start
                                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                                step = 'null'
                                status = Severity.NOT_OK.value
                                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
                                return

                    status = Severity.OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

                    step = 'removing node from cluster'
                    duration_step_start = time.perf_counter()
                    
                    try:
                        command = f"sudo {redis_bin}/redis-cli -u 'redis://{node['database_user']}:{node['database_pwd']}@{host['name']}:{node['port']}' --cluster del-node {host['name']}:{node['port']} {id} --cluster-yes"
                        jsn = json.loads(remote_command(host, command, True), strict=False)
                        error_lower = jsn['stderr'].lower()
                        if error_lower != '' and "err" in error_lower:
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return
                    except Exception as error:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                        return

                    status = Severity.OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

            step = 'dropping node'
            duration_step_start = time.perf_counter()

            error = stop_redis_node(target_client, force=True, version=redis_version, sentinel=node['cluster_role'] == 'SENTINEL')
            if error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

            try:
                jsn = json.loads(remote_command(host, f"sudo rm -rf '{node['dir_123Cluster']}'", True), strict = False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return

            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

            if not is_standalone and is_master and other_masters_exists:
                step = 'checking cluster health'
                duration_step_start = time.perf_counter()
                
                try:
                    command = f"echo 'yes' | sudo {redis_bin}/redis-cli -u 'redis://{master_up['node']['database_user']}:{master_up['node']['database_pwd']}@{master_up['host']['name']}:{master_up['node']['port']}' --cluster fix {master_up['host']['name']}:{master_up['node']['port']}"
                    jsn = json.loads(remote_command(master_up['host'], command, True), strict=False)
                    if error_lower != '' and "err" in error_lower:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                except Exception as error:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                    return

                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        else:
            step = 'flushing node'
            duration_step_start = time.perf_counter()

            try:
                target_client.flushall(asynchronous=False)
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
        
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    else: 
        step = 'dropping node'
        duration_step_start = time.perf_counter()

        error = stop_redis_node(target_client, force=True, version=redis_version, sentinel=node['cluster_role'] == 'SENTINEL')
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        try:
            jsn = json.loads(remote_command(host, f"sudo rm -rf '{node['dir_123Cluster']}'", True), strict = False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) 
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    if not data_only:
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    if node['cluster_role'] == 'SENTINEL':
                        if len(other_nodes) > 0:
                            for i in other_nodes:
                                monitored_by = i['node']['ex_params'].get('monitored_by')
                                if monitored_by:
                                    if len(monitored_by) > 0:
                                        cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = jsonb_set(ex_params, '{{monitored_by}}', (ex_params->'monitored_by') - '{0}') WHERE id=%s""".format(node['id']), [i['node']['id']])
                        
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity='{0}', status='{1}', ind_started='false', ind_deleted='true', ex_params = jsonb_set(ex_params, '{{monitored_nodes}}', %s::jsonb) WHERE id = {2}""".format(Severity.NOT_OK.value, Status.OK.value, node['id']), [json.dumps({})])
                    else:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.OK.value, node['id'],))
                    
                    conn.commit()
        except Exception as error:
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    else:
        try:
            cur.execute("""UPDATE "clusterCore_c123_node" SET status='flushed', ind_started='true', ind_deleted='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, node['id'],))
            conn.commit()
        except Exception as error:
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
      
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restart_redis_node_task(data, task_id):
    """restart node"""
    
    task = 'restart redis node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node['id'])
            return

        node = json.loads(data['node_data'])
        sentinel = node['cluster_role'] == 'SENTINEL'
        redis_version = float(node['version'])
        conf_dir = f"{node['dir_123Cluster']}/conf"
        conf_file = f"{conf_dir}/{'sentinel' if sentinel else 'redis'}.conf"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node availability'
    duration_step_start = time.perf_counter()

    redis_client = connect_to_redis(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(redis_client, tuple):
        error = redis_client[1]
        if hasattr(redis_client[1], 'strerror'):
            error = redis_client[1].strerror

        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, node['id'],))
                    conn.commit() 
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stopping node'
    duration_step_start = time.perf_counter()

    error = stop_redis_node(redis_client, sentinel=sentinel, version=redis_version)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'restarting node'
    duration_step_start = time.perf_counter()

    client = create_ssh_connection(host['name'], host['username'], host['password'], host['private_key'], host['ssh_port'])
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            error = client[1].strerror
        else:
            error = str(client[1])

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    redis_bin = find_redis_bin(host)
    command = f"sudo nohup {redis_bin}/redis-{'sentinel' if sentinel and OperatingSystem.CENTOS.value in distributor else 'server'} '{conf_file}' {'--sentinel' if sentinel and OperatingSystem.UBUNTU.value in distributor else ''} &> /dev/null &"
    
    try:
        stdin, stdout, stderr = client.exec_command(command)
        exit_status = stdout.channel.recv_exit_status() # Blocking call
        stdin.channel.shutdown_write()
        try:
            stderr = stderr.read().decode('ascii').strip()
            stdout = stdout.read().decode('ascii').strip("\n")
        except UnicodeDecodeError:
            stderr = stderr.read().decode('utf-8').strip()
            stdout = stdout.read().decode('utf-8').strip("\n")

        if stderr != '' and 'error' in stderr:
            error = stderr
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,str(error),exit_status)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,str(error),exit_status)
            return
        
        client.close()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    time.sleep(1)
    try: 
        redis_client.ping()
    except Exception as e:
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, node['id'],))
                    conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        error = e.strerror if hasattr(e, 'strerror') else e.args[0]
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET status=%s, ind_started='true' WHERE id=%s""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restore_redis_node_task(data, task_id):
    """restore node"""
    
    task = 'restore redis node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        backup_host = data['backup_host']
        node = json.loads(data['node_data'])
        backup = json.loads(data['backup'])
        schedule = backup['is_sched']
        backup_type = backup['ex_params'].get('backup_type')
        support_aof = node['ex_params'].get('has_aof')
        restore_option = data['restore_option'].strip().lower()
        compress = backup["ex_params"].get("compress")
        redis_version = float(node['version'])
        aof_object = "appendonly.aof" if redis_version < 7.0 else "appendonlydir"

        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        backup_path = backup['ex_params'].get('backup_path')
        if backup_path.startswith('~'):
            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)

        if schedule:
            if restore_option in [RestoreOptions.LATEST.value, RestoreOptions.CUSTOM.value]:
                if RestoreOptions.CUSTOM.value in restore_option:
                    timezone_cmd = 'cat /etc/timezone'
                    jsn = json.loads(remote_command(backup_host, timezone_cmd, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return

                    try:
                        timezone = pytz.timezone(jsn['stdout'])
                        target_timestamp = datetime.strptime(data['date'], "%Y-%m-%d %H:%M:%S").astimezone(timezone).timestamp()
                    except Exception as e:
                        error = f'Invalid server timezone: {jsn["stdout"]} error: {str(e)}'
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
                        return

                    command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n"
                else:
                    command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n | tail -1 | awk '{{print $2}}'"
            else:
                command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n | head -1 | awk '{{print $2}}'"

            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            if jsn['stdout'] == '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                return
            else:
                if restore_option in [RestoreOptions.THIS.value, RestoreOptions.LATEST.value]:
                    backup_path = jsn['stdout']
                else:
                    output_lines = [line.strip() for line in jsn['stdout'].split("\n") if line.strip()]
                    closest_subfolder = None
                    min_time_difference = positive_infinity

                    for line in output_lines:
                        subfolder_timestamp_str, subfolder = line.split(" ", 1)
                        subfolder_timestamp = int(subfolder_timestamp_str)
                        time_difference = abs(target_timestamp - subfolder_timestamp)
                        if time_difference < min_time_difference:
                            min_time_difference = time_difference
                            closest_subfolder = subfolder

                    if not closest_subfolder:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                        return

                    backup_path = closest_subfolder
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    redis_client = connect_to_redis(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(redis_client, tuple):
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, node['id'],))
                    conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
        return

    try:
        has_aof = "yes" in redis_client.config_get(pattern="appendonly")["appendonly"]
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    target_object = aof_object if backup_type == "aof" and has_aof else "dump.rdb"
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host['name'] != backup_host['name']:
        step = 'copying backup files of node to node host'
        duration_step_start = time.perf_counter()

        jsn = transfer_files_rsync(backup_host, host, backup_path, [f"{backup_path}/"])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stopping node'
    duration_step_start = time.perf_counter()

    if backup_type == "rdb" and has_aof:
        try:
            redis_client.config_set(name="appendonly", value= "no")
            redis_client.config_rewrite()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    error = stop_redis_node(redis_client, force=True, version=redis_version)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'restoring node'
    duration_step_start = time.perf_counter()

    client = create_ssh_connection(host['name'], host['username'], host['password'], host['private_key'], host['ssh_port'])
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            error = client[1].strerror
        else:
            error = str(client[1])
        
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    redis_bin = find_redis_bin(host)
    rm_command = f"sudo rm -rf '{node['dir_123Cluster']}/data/{target_object}'" if has_aof and backup_type == "aof" else  f"sudo rm -f '{node['dir_123Cluster']}/data/{target_object}'"
    move_command = f"sudo rsync -a --recursive '{backup_path}/{target_object}' '{node['dir_123Cluster']}/data'"
    if compress:
        filename_cmd = f"sudo ls -lt --file-type '{backup_path}' | grep '\\.tar.gz$' | head -n 1 | awk '{{print $9}}'"
        jsn = json.loads(remote_command(host, filename_cmd, True), strict = False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        archive_file = jsn['stdout']
        move_command = f"sudo tar xzfC '{backup_path}/{archive_file}' '{node['dir_123Cluster']}/data'"
    
    commands = [
        rm_command,
        move_command,
        f"sudo chown -R redis:redis '{node['dir_123Cluster']}/data/{target_object}'",
        f"sudo chown -R 660 '{node['dir_123Cluster']}/data/{target_object}'",
        f"sudo nohup {redis_bin}/redis-server '{node['dir_123Cluster']}/conf/redis.conf' &> /dev/null &",
        f"sudo rm -rf '{backup_path}'"
    ]

    try:
        for command in commands:
            stdin, stdout, stderr = client.exec_command(command)
            exit_status = stdout.channel.recv_exit_status() # Blocking call
            stdin.channel.shutdown_write()
            try:
                stderr = stderr.read().decode('ascii').strip()
                stdout = stdout.read().decode('ascii').strip("\n")
            except UnicodeDecodeError:
                stderr = stderr.read().decode('utf-8').strip()
                stdout = stdout.read().decode('utf-8').strip("\n")

            if stderr != '' and any (err in stderr.lower() for err in ['error', 'not recoverable']):
                error = stderr
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,str(error),exit_status)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,str(error),exit_status)
                return
        
        client.close()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    time.sleep(5)
    try:
        redis_client.ping()
    except Exception as e:
        error = e.strerror if hasattr(e, 'strerror') else e.args[0]
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    if support_aof and not has_aof:
        try:
            redis_client.config_set(name="appendonly", value="yes")
            redis_client.config_rewrite()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET status=%s, size=%s, ind_started='true' WHERE id=%s""", (Status.OK.value, node_size, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def switchover_redis_node_task(data, task_id):
    """switchover node"""
    
    task = 'switchover redis node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        id_list = data['id_list']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node availability'
    duration_step_start = time.perf_counter()

    redis_client = connect_to_redis(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(redis_client, tuple):
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, node['id'],))
                    conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'switchover node'
    duration_step_start = time.perf_counter() 
    
    try:
        redis_client.cluster("failover force")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"slave_of":%s}', cluster_role='STANDBY' WHERE id=ANY(%s)""", (node['id'], id_list))
                cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"slave_of":%s}', cluster_role='PRIMARY' WHERE id=%s""", (False, node['id']))
                cur.execute("""SELECT ex_params->'monitored_by' FROM "clusterCore_c123_node" WHERE id=%s""", (node['id'],))
                sentinels = cur.fetchone()
                current_sentinels = [] if not sentinels[0] else sentinels[0]
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity='{0}', status='{1}', ind_started='true', ind_deleted='false', ex_params = jsonb_set(ex_params, '{{monitored_by}}', %s::jsonb) WHERE id = ANY(%s)""".format(Severity.OK.value, Status.OK.value), (json.dumps(current_sentinels), id_list))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def create_redis_node(host, node, cluster = None) -> dict:

    sentinel = node['cluster_role'] == 'SENTINEL'

    if node['dir_123Cluster'].startswith('~'):
        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            return jsn
            
        tilda_path = jsn['stdout']
        node['dir_123Cluster'] = node['dir_123Cluster'].replace('~', tilda_path)

    redis_bin = find_redis_bin(host)
    conf_dir = f"{node['dir_123Cluster']}/conf"
    data_dir = f"{node['dir_123Cluster']}/data"
    log_dir = f"{node['dir_123Cluster']}/log"
    conf_file = f"{conf_dir}/{'sentinel' if sentinel else 'redis'}.conf"
    cluster_conf_file = f"{conf_dir}/nodes.conf"
    log_file = f"{log_dir}/redis.log"
    
    config_content = f"""bind {host["name"]} 127.0.0.1
port {node["port"]}
logfile {log_file}
dir {data_dir}
protected-mode no"""

    if cluster and not sentinel:
       config_content += f"""
cluster-enabled yes
cluster-config-file {cluster_conf_file}
"""


    commands = [
        f"sudo mkdir -p '{conf_dir}'",
        f"sudo mkdir -p '{data_dir}'",
        f"sudo mkdir -p '{log_dir}'",
        f"echo '{config_content}' | sudo tee '{conf_file}'"
    ]
    commands.append(f"sudo nohup {redis_bin}/redis-{'sentinel' if sentinel and os == OperatingSystem.CENTOS.value else 'server'} '{conf_file}' {'--sentinel' if sentinel and os == OperatingSystem.UBUNTU.value else ''} &> /dev/null &")

    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])
        except Exception as error:
            jsn = {'stdout': None, 'stderr': str(error), 'excode': 9}
            
    return jsn

def backup_redis_node_task(data, task_id):
    """backup node"""
    
    task = 'redis backup node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        backup_host = data['backup_host']
        node = json.loads(data['node_data'])
        backup = json.loads(data['backup_data'])
        backup_type = backup['ex_params'].get('backup_type')
        backup_path = backup['ex_params'].get('backup_path')
        compress = backup["ex_params"].get("compress")
        support_aof = node["ex_params"].get("has_aof")
        aof_object = "appendonly.aof" if float(node["version"]) < 7.0 else "appendonlydir"
        target_object = aof_object if backup_type == "aof" and support_aof else "dump.rdb"

        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        if backup_path.startswith('~'):
            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)

        if backup['is_sched']:
            start_date = datetime.now().astimezone(pytz.UTC).strftime('%d%m%Y%H%M%S')
            backup_path = f'{backup_path}/{start_date}'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node availability'
    duration_step_start = time.perf_counter()

    redis_client = connect_to_redis(host['name'], node['database_user'], node['database_pwd'], node['port'])
    if isinstance(redis_client, tuple):
        try:
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ind_started='false', severity=%s WHERE id=%s""", (Severity.NOT_OK.value, node['id'],))
                    conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(redis_client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(redis_client[1]),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if backup_type == "rdb":
        step = 'updating rdb'
        duration_step_start = time.perf_counter()

        try:
            last_save = redis_client.info("persistence")['rdb_last_save_time']
            redis_client.bgsave()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        save_finished = False 

        while not save_finished:
            try:
                persistence_info = redis_client.info("persistence")
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
            
            save_finished = persistence_info['rdb_bgsave_in_progress'] == 0 and persistence_info['rdb_last_bgsave_status'] == "ok" and persistence_info['rdb_last_save_time'] - last_save > 0
            save_failed = persistence_info['rdb_bgsave_in_progress'] == 0 and persistence_info['rdb_last_bgsave_status'] != "ok" and persistence_info['rdb_last_save_time'] - last_save == 0
            if save_failed:
                error = "save failed"
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
            
            time.sleep(2)
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'copying backup file'
    duration_step_start = time.perf_counter()

    remote_command(host, f"sudo mkdir -p '{backup_path}'", False)

    move_command = f"sudo rsync -a --recursive '{node['dir_123Cluster']}/data/{target_object}' '{backup_path}'" if not compress else f"sudo tar zcfC '{backup_path}/{backup['name']}.tar.gz' '{node['dir_123Cluster']}/data' {target_object}"

    try:
        jsn = json.loads(remote_command(host, move_command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
                
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host['name'] != backup_host['name']:
        step = 'copying backup files of node to backup host'
        duration_step_start = time.perf_counter()

        jsn = transfer_files_rsync(host, backup_host, backup_path, [f"{backup_path}/"])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'getting the backup size'
    duration_step_start= time.perf_counter()

    try:
        jsn = json.loads(remote_command(backup_host,f"sudo du -sb '{backup_path}' | awk '{{print $1}}'" if not compress else f"sudo du -b '{backup_path}/{backup['name']}.tar.gz' | awk '{{print $1}}'", True), strict=False) 
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        backup_size = round(float(jsn['stdout'])/1024, 2)
        modified_date = get_date_modified_folder(hostdata=backup_host, folder_path=backup_path)
        if isinstance(modified_date, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return
        
        remote_command(host, f"sudo rm -rf '{backup_path}'", True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}', status='true' WHERE id=%s""" % (modified_date, backup['id']))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size, backup['id']))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)