# BELONG TO SERVICE TASK
import os, json, time, uuid, pytz, logging,traceback
from datetime import datetime
from math import inf as positive_infinity
from psycopg2.extras import RealDictCursor
from enums import OperatingSystem, SupportedOS, DBType, ClusterRoles, RestoreOptions, Status, Severity, Inventory, Schedule, LicenseType, Repository
from tasks_core import insert_task, format_for_shell, create_ssh_connection, remote_command, check_remote_directory_exists, get_postgres_connection, logger,\
      check_os_version, find_tilda, get_date_modified_folder, transfer_text_ram, update_failed_object, install_offline_remote_to_remote, generate_complex_password, TASK_TMP_FILES_DIR,\
      transfer_files_rsync, get_mysql_connection, transfer_ssl_certificates, init_container, get_log_rotate_config

#####################################################################
# MARIADB TASKS
#####################################################################

def check_mariadb_version_exists(host, mariadb_version, distributor):
    cmd = f'rpm -qa | grep -i "MariaDB-server-{mariadb_version}"' if OperatingSystem.CENTOS.value in distributor else f'dpkg -l | grep -i "MariaDB-server" | grep "{mariadb_version}"'
    jsn = json.loads(remote_command(host, cmd, True), strict=False)
    return jsn['stdout'] != ''

def get_mariadb_config_file(id, port, is_standby = False, use_ssl = False):

    config = f"""[mysqld]
server_id='{id}'
bind-address=0.0.0.0
port='{port}'
default_storage_engine=InnoDB
datadir='$datadir'
log_bin='$datadir/binlog'
log_bin_index='$datadir/mariadb-bin.index'
relay-log=$(cat /etc/hostname)-relay-bin
binlog_format=ROW
expire_logs_days=2
innodb_buffer_pool_size=134217728
innodb_buffer_pool_instances=1
pid-file=/var/run/mariadb/mariadb_{port}.pid
character_set_server=utf8mb4
explicit_defaults_for_timestamp=ON
general_log = 1
general_log_file =$datadir/log/mariadb.log
replicate_ignore_db='mysql,sys'"""
    
    if use_ssl:
        config = f"""[client]
ssl-cert='$datadir/client-cert.pem'
ssl-key='$datadir/client-key.pem'
{config}"""
    
    if is_standby:
        if is_standby:
            config += """
read_only=ON"""
            if use_ssl:
                # mariadb does not support --ssl-mode option and bugs on detecting TLS version with --ssl option,
                # so for mariadb ssl connections mode are default PREFERRED unlike REQUIRED in mysql - only this way ssl works...
                config += """
ssl_ca=$datadir/ca.pem
ssl_cert=$datadir/server-cert.pem
ssl_key=$datadir/server-key.pem"""
    
    return config

def install_mariadb_standalone_task(data, task_id):

    task = 'install mariadb standalone task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        app_host = json.loads(data['app_host'])
        org_id = node['org_id']
        # app_hostname = data['app_hostname']
        create_run_linux = data['create_run_linux']
        # repository = network or local
        repository = data['repository']
        offline = True if Repository.LOCAL.value in repository else False
        mariadb_version = float(node['version'])
        
        host_exists = True
        proxmox_container = node['ex_params'].get('proxmox')
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = node['ex_params'].get('distributor')
                release = node['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
        
        version_exists = check_mariadb_version_exists(host, mariadb_version, distributor)

    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        update_failed_object(node['id'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

 
    step = 'running installation commands'
    duration_step_start = time.perf_counter()

    if not version_exists:
        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host, mariadb_version, DBType.MARIADB.value, app_host)
        else:
            jsn = install_mariadb(distributor, release, host, mariadb_version)

        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'mysql.service not loaded'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node['id'])
            return

    jsn = create_mariadb_node(host, node, distributor)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    
        node_size = round(float(jsn['stdout']) / 1024, 2)
        remote_command(host, f"""echo "{get_log_rotate_config(node['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{node['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host), 'node_name': node['name']}, 'task': 'log_rotate_task',  'org_id': node['org_id'], 'monitor_name': None,  'target_type': node['type'], 'target_id': node['id'], 'created_by': created_by}
        insert_task(params)
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,node['id'],created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host['id'],created_by,OperatingSystem.LINUX.value])
                conn.commit()
         
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
    
    # host['create_task'] = created_by
    # host['ssh_port'] = int(host['ssh_port'])
    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host, 'task': 'run_linux_performance_counters', 'target_type': 'host', 'target_id': host['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as error:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def install_mariadb(distributor, release, host, mariadb_version) -> dict:
    supported_versions = SupportedOS[distributor.upper()].value
    supported_versions_str = ", ".join(map(str, supported_versions))
    mariadb_version = float(mariadb_version)

    galera_3 = 'galera' if OperatingSystem.CENTOS.value in distributor else 'galera-3'
    galera = "galera-4" if mariadb_version >= 10.4 or mariadb_version in [10.10, 10.11] else galera_3
    
    if OperatingSystem.CENTOS.value in distributor:
        package_manager = 'dnf' if int(release) > 7 else 'yum'
        commands = [
            f"sudo {package_manager} --enablerepo=appstream install -q -y rsync net-tools",
            'sudo curl -LOs https://r.mariadb.com/downloads/mariadb_repo_setup',
            "sudo chmod +x mariadb_repo_setup",
            f"sudo ./mariadb_repo_setup --mariadb-server-version='{mariadb_version}'",
            f'sudo {package_manager} --enablerepo=appstream install -q -y epel-release MariaDB-server MariaDB-client MariaDB-backup MariaDB-common MariaDB-shared {galera}'
        ]
    elif OperatingSystem.UBUNTU.value in distributor:
        release_codename_mapping = {
            18: 'bionic',
            20: 'focal',
            22: 'jammy',
        }
        version = release_codename_mapping.get(int(release), 'unknown')

        if version == 'unknown':
            jsn = {
                'stdout': 'Unknown release version',
                'stderr': f'Release version of {distributor.capitalize()} is not {supported_versions_str}',
                'excode': 1
            }
            return jsn

        commands = [
            f"sudo debconf-set-selections <<< 'libssl1.1 libraries/restart-without-asking boolean true'",
            f'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y curl net-tools',
            f'sudo curl -LOs https://r.mariadb.com/downloads/mariadb_repo_setup',
            "sudo chmod +x mariadb_repo_setup",
            f"sudo ./mariadb_repo_setup --mariadb-server-version='{mariadb_version}'",
            f'sudo DEBIAN_FRONTEND=noninteractive apt-get -o dpkg::Options::="--force-confnew" install -qq -y mariadb-server mariadb-client mariadb-backup {galera}',
        ]

    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print('stdout:', jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
            if filter:
                return jsn
        except Exception as error:
            return {'stdout': None, 'stderr': str(error), 'excode': 9}
            
    return jsn

def find_mariadb_bin(host, distributor, version) -> str:
    """find location of mariadb bin folder"""

    tool = "dpkg-query -L" if OperatingSystem.UBUNTU.value in distributor else "rpm -ql"
    grep = "grep 'mysql_upgrade$'" if OperatingSystem.UBUNTU.value in distributor else "grep 'mariadbd-safe$'"
    package = "MariaDB-server" if OperatingSystem.CENTOS.value in distributor else f"mariadb-server-core-{version}"
    mariadb_bin = ''

    try:
        jsn = json.loads(remote_command(host, f'dirname $(sudo {tool} {package} | {grep})', True), strict=False)
        if jsn['stdout'] != '':
            mariadb_bin = jsn['stdout'].strip()

        if jsn['stderr'] != '' and 'permission denied' not in jsn['stderr'].lower():
            if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
                mariadb_bin = '/usr/share/mariadb/bin'
            else:
                mariadb_bin = '/usr/share/mariadb/bin'
    except Exception as error:
        logging.error(f"find_mariadb_bin: {str(error)}")
        
    return mariadb_bin

def create_mariadb_node(host, node, distributor) -> dict:
    try:
        mariadb_version = float(node['version'])
        use_ssl = node['ex_params'].get('ssl')
        cli = 'mysql' if mariadb_version < 10.4 else 'mariadb'
        data_dir = node['dir_123Cluster']
        if data_dir.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                return jsn

            tilda_path = jsn['stdout']
            data_dir = data_dir.replace('~', tilda_path)

        new_root_password = node['ex_params'].get('mariadb_root_password')
        change_password_root_command = f'''root_pwd='{new_root_password}';sudo {cli} -Nsfu root -e "SET PASSWORD FOR 'root'@'localhost' = PASSWORD('$root_pwd');"'''
        database_user = node['database_user']
        database_pwd = node['database_pwd']

        conf_path = "/etc/my.cnf.d/server.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mariadb.conf.d/50-server.cnf"
        config = get_mariadb_config_file(node['id'], node['port'], use_ssl = use_ssl)

        initialization_prefix = 'mysql' if mariadb_version < 10.4 and mariadb_version not in [10.10,10.11] else 'mariadb'
        
        # do not add key paths on initialization - only when have ssl certificates
        key_paths = """
ssl_ca=$datadir/ca.pem
ssl_cert=$datadir/server-cert.pem
ssl_key=$datadir/server-key.pem"""
        
        ssl_config_path = '/etc/pki/tls/mariadb.cnf' if OperatingSystem.CENTOS.value in distributor else '/etc/ssl/mariadb.cnf'
        #as default, as possible
        ssl_config = f"""[ req ]
prompt = no
distinguished_name = dn
[ dn ]
C = IL
ST = State
L = City
O = 123Cluster
OU = Organizational Unit
CN = MariaDB_{mariadb_version}_123Cluster_CERTIFICATE_Certificate"""
        
        # since mariadb-install-db dont have 'upgrade on install' option - we manually detect if use install or upgrade
        jsn = check_remote_directory_exists(host, data_dir, "d", is_dir_empty=True)
        if jsn['stderr'] != '' and 'no such' not in jsn['stderr'].lower():
           return jsn
        
        data_dir_not_empty = 'false' in jsn['stdout']
        tool = 'upgrade' if data_dir_not_empty else 'install-db'
        
        commands = [
            f"sudo mkdir -p '{data_dir}/log'",
            f"sudo chown -R mysql:mysql '{data_dir}'",
            # since mariadb-install-db dont have 'upgrade on install' option - we must manually start autogenerated instance (if exists) before upgrade
            "sudo systemctl start mariadb" if data_dir_not_empty else "sudo systemctl stop mariadb",
            f"[ -f '{conf_path}' ] && sudo cp '{conf_path}' '{conf_path}.bak'",
            f"""datadir='{data_dir}';echo "{config}" | sudo tee '{conf_path}'""",
            # mariadb-upgrade must use user root (because bugs with other users) with socket authentication (to bypass password requirement)
            f"""sudo {initialization_prefix}-{tool} --defaults-file='{conf_path}' {"--user=root --protocol=socket --force" if data_dir_not_empty else f"--user=mysql --datadir='{data_dir}'"} """,
        ]
        if not data_dir_not_empty:
            commands.insert(0, f"sudo mkdir -p '{data_dir}'")
        
        if use_ssl:
            commands += [
                # mariadb could not auto-generate certificates, so generate manually (for compatibility with mysql behaviour - store in node datadir)
                # mariadb docs are all incomplete, need this...
                # https://dev.mysql.com/doc/refman/8.4/en/creating-ssl-files-using-openssl.html + https://dev.mysql.com/doc/refman/8.4/en/creating-ssl-rsa-files-using-mysql.html
                f"echo '{ssl_config}' | sudo tee '{ssl_config_path}'".replace('CERTIFICATE', 'CA'),
                f"sudo openssl req -newkey rsa:2048 -x509 -nodes -keyout '{data_dir}/ca-key.pem' -out '{data_dir}/ca.pem' -days 3600 -config '{ssl_config_path}'",
                f"sudo openssl rsa -in '{data_dir}/ca-key.pem' -out '{data_dir}/ca-key.pem'",
                f"echo '{ssl_config}' | sudo tee '{ssl_config_path}'".replace('CERTIFICATE', 'Server'),
                f"sudo openssl req -newkey rsa:2048 -nodes -keyout '{data_dir}/server-key.pem' -out '{data_dir}/server-req.pem' -days 3600 -config '{ssl_config_path}'",
                f"sudo openssl rsa -in '{data_dir}/server-key.pem' -out '{data_dir}/server-key.pem'",
                f"sudo openssl x509 -req -in '{data_dir}/server-req.pem' -days 3600 -CA '{data_dir}/ca.pem' -CAkey '{data_dir}/ca-key.pem' -set_serial 01 -out '{data_dir}/server-cert.pem'",
                f"echo '{ssl_config}' | sudo tee '{ssl_config_path}'".replace('CERTIFICATE', 'Client'),
                f"sudo openssl req -newkey rsa:2048 -nodes -keyout '{data_dir}/client-key.pem' -out '{data_dir}/client-req.pem' -days 3600 -config '{ssl_config_path}'",
                f"sudo openssl rsa -in '{data_dir}/client-key.pem' -out  '{data_dir}/client-key.pem'",
                f"sudo openssl x509 -req -in '{data_dir}/client-req.pem' -days 3600 -CA '{data_dir}/ca.pem' -CAkey '{data_dir}/ca-key.pem' -set_serial 01 -out '{data_dir}/client-cert.pem'",
            ]

        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and 'error' in error_lower:
                return jsn
        
        if use_ssl:
            # before starting the node - must verify that certificate chain are correct
            jsn = json.loads(remote_command(host, f"sudo openssl verify -CAfile '{data_dir}/ca.pem' '{data_dir}/server-cert.pem' '{data_dir}/client-cert.pem'", True), strict=False)
            if 'OK' not in jsn['stdout']:
                return jsn
        
        err_list = ["error", "fail", "errno"]
        commands = [
            f"sudo mkdir -p '/var/run/mariadb/'",
            "sudo chown -R mysql:mysql '/var/run/mariadb/'",
            # certificate files should also belong to user 'mysql'
            f"sudo chown -R mysql:mysql '{data_dir}'",
            'sudo systemctl enable mariadb',
            'sudo systemctl start mariadb',
            change_password_root_command,
            # cursed mariadb shares the mysql older versions bug of empty user records after instance creation, which prevent network users to login
            f'''sudo {cli} -Nsfu root --password='{new_root_password}' -e "DELETE FROM mysql.user WHERE user = '';"''',
            f'''database_pwd='{database_pwd}';sudo {cli} -Nsfu root --password='{new_root_password}' -e "CREATE USER '{database_user}'@'%.%.%.%' IDENTIFIED VIA mysql_native_password USING PASSWORD('$database_pwd') {'REQUIRE SSL' if use_ssl else ''};"''',
            f'''sudo {cli} -Nsfu root --password='{new_root_password}' -e "GRANT ALL PRIVILEGES ON *.* TO '{database_user}'@'%.%.%.%' WITH GRANT OPTION;"''',
            f'''sudo {cli} -Nsfu root --password='{new_root_password}' -e "FLUSH PRIVILEGES;"'''
        ]
        
        if use_ssl:
            commands.insert(0, f"""datadir='{data_dir}';echo "{key_paths}" | sudo tee -a '{conf_path}'""")
        
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in err_list):
                return jsn

        queries = [
            f"CREATE USER '{database_user}'@'localhost' IDENTIFIED VIA mysql_native_password USING PASSWORD('{database_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
            f"CREATE USER '{database_user}'@'%' IDENTIFIED VIA mysql_native_password USING PASSWORD('{database_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
            f"GRANT ALL PRIVILEGES ON *.* TO '{database_user}'@'localhost' WITH GRANT OPTION",
            f"GRANT ALL PRIVILEGES ON *.* TO '{database_user}'@'%' WITH GRANT OPTION",
            "FLUSH PRIVILEGES"
        ]
        
        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            jsn['stderr'] = str(conn[1])
            return jsn
        
        with conn:
            with conn.cursor() as cur:
                for query in queries:
                    cur.execute(query)
                
        return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

def add_mariadb_standby_task(data, task_id):
    """add standby"""
    
    task = 'add mariadb standby task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_primary = json.loads(data['host_primary'])
        primary = json.loads(data['primary_data'])
        mariadb_version = float(primary['version'])
        use_ssl = primary['ex_params'].get('ssl')
        host_standby = json.loads(data['host_standby'])
        app_host = json.loads(data['app_host'])
        standby = json.loads(data['standby_data'])
        proxmox_container = standby['ex_params'].get('proxmox')
        host_exists = True
        distributor, release = check_os_version(host_standby)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = standby['ex_params'].get('distributor')
                release = standby['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(standby['id'])
                return
        
        version_exists = check_mariadb_version_exists(host_standby, mariadb_version, distributor)
        
        org_id = standby['org_id']
        # since standby are clone of primary - only primary root password are needed
        primary_root_password = primary['ex_params'].get('mariadb_root_password')
        # repository = network or local
        repository = data['repository']
        offline = True if Repository.LOCAL.value in repository else False
        rep_user = standby['ex_params'].get('rep_user')
        rep_pwd = standby['ex_params'].get('rep_pwd')
        has_read_only_admin = mariadb_version >= 10.5 or mariadb_version in [10.10, 10.11]
        conf_path = "/etc/my.cnf.d/server.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mariadb.conf.d/50-server.cnf"
        backup_cli = 'mariabackup' if mariadb_version < 11.0 else 'mariadb-backup'
        primary_clone = f"{TASK_TMP_FILES_DIR}/mariadb_clones/{primary['name'].replace(':', '_')}"
        standby_data_dir = standby['dir_123Cluster']
        if standby_data_dir.startswith('~'):
            jsn = find_tilda(host_standby)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            standby_data_dir = standby_data_dir.replace('~', tilda_path)
        
        primary_data_dir = primary['dir_123Cluster'].strip()
        if primary_data_dir.startswith('~'):
            jsn = find_tilda(host_primary)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            primary_data_dir = primary_data_dir.replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        update_failed_object(standby['id'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host_standby, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(standby['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        
        
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    if not version_exists:
        step = 'running installation commands'
        duration_step_start = time.perf_counter()

        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host_standby, mariadb_version, DBType.MARIADB.value, app_host)
        else:
            jsn = install_mariadb(distributor, release, host_standby, mariadb_version)
        
        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'mysql.service not loaded'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(standby['id'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'preparing primary commands'
    duration_step_start = time.perf_counter()

    try:
        now = datetime.utcnow().strftime('%d%m%Y%H%M%S')
        backup_name = f"{primary['name'].replace(':', '_')}_{now}"
        transfer_path = f"{primary_clone}/{backup_name}.gz"
        backup_cmd  = f"sudo {backup_cli} --backup --user=root --port='{primary['port']}' --password='{primary_root_password}' --target-dir='{primary_clone}' --stream=mbstream | sudo gzip -c | sudo tee '{transfer_path}'"
        
        queries = [
            f"CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED VIA mysql_native_password USING PASSWORD('{rep_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
            # each "create user if not exists" query should be confirmed by "alter user", against the cases of password misconfiguration
            f"ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED VIA mysql_native_password USING PASSWORD('{rep_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
            # all privileges should include replication slave, no?
            f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
            # but for insurance - better go explicit also
            f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
            "FLUSH PRIVILEGES",
            # now think that position before the start of backup are preferred over the position after backup restore
            "SHOW MASTER STATUS"
        ]
        
        conn = get_mysql_connection(host_primary, primary)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor(dictionary=True) as cur:
                for query in queries:
                    cur.execute(query)
                #its ok to do like this, with...as blocks variables scope are function-level
                master_replication = cur.fetchone()
                # revoke the right to modify from everyone, except the trusted superusers of 123Cluster app 
                # (because revoking this grants also breaks the connectivity by db drivers over the network,so trusted superusers need to have it...) 
                cur.execute(f'''SELECT CONCAT('REVOKE SUPER{',READ ONLY ADMIN' if has_read_only_admin else ''} ON *.* FROM ', QUOTE(user), '@', QUOTE(host), ';') AS REVOKE_QUERY FROM mysql.user WHERE user NOT IN ('root','PUBLIC', '{primary['database_user']}')''')
                revoke_queries = [i['REVOKE_QUERY'] for i in cur.fetchall()]
                for revoke_query in revoke_queries:
                    cur.execute(revoke_query)
                cur.execute("FLUSH PRIVILEGES")
        
        commands = [
            f"sudo mkdir -p '{primary_clone}'",
            backup_cmd
        ]
        for command in commands:
            jsn = json.loads(remote_command(host_primary, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(standby['id'])
                # if fails here - backup of primary is corrupted
                remote_command(host_primary, f"sudo rm -rf '{primary_clone}'", True)
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'running installation commands on standby'
    duration_step_start = time.perf_counter()
    
    try:
        jsn = transfer_files_rsync(host_primary, host_standby, primary_clone, [transfer_path])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(standby['id'])
            return

        config = get_mariadb_config_file(standby['id'], standby['port'], is_standby=True, use_ssl = use_ssl)
        
        commands = [
            f"sudo gunzip < '{transfer_path}' | sudo mbstream -C '{primary_clone}' -x",
            f"sudo chown -R mysql:mysql '{primary_clone}'",
            f"[ -f '{conf_path}' ] && sudo cp '{conf_path}' '{conf_path}.bak'",
            f"""datadir='{standby_data_dir}';echo "{config}" | sudo tee '{conf_path}'""",
            f"sudo {backup_cli} --prepare --target-dir='{primary_clone}'",
        ]
        
        for command in commands:
            jsn = json.loads(remote_command(host_standby, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(standby['id'])
                # if fails here - backup copy on standby host is corrupted
                remote_command(host_standby, f"sudo rm -rf '{primary_clone}'", True)
                return

        commands = [
            "sudo systemctl stop mariadb",
            # when cloning from backup - always wipe data directory, to avoid potential conflicts
            f"sudo rm -rf '{standby_data_dir}'",
            f"sudo mkdir -p '{standby_data_dir}'",
            f"sudo {backup_cli} --copy-back --target-dir='{primary_clone}'",
        ]
        for command in commands:
            jsn = json.loads(remote_command(host_standby, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        if use_ssl:
            jsn = transfer_ssl_certificates(primary, host_primary, host_standby, standby_data_dir)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        commands = [
            f"sudo chown -R mysql:mysql '{standby_data_dir}'",
            f"sudo mkdir -p '/var/run/mariadb/'",
            "sudo chown -R mysql:mysql '/var/run/mariadb/'",
            'sudo systemctl enable mariadb',
            'sudo systemctl start mariadb',
        ]
        for command in commands:
            jsn = json.loads(remote_command(host_standby, command, True), strict=False)
            if jsn['stderr'] != '' and 'synchronizing state' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        remote_command(host_primary, f"sudo rm -rf '{primary_clone}'", True)
        remote_command(host_standby, f"sudo rm -rf '{primary_clone}'", True)
        
        queries = [
            "STOP SLAVE",
            "RESET SLAVE ALL",
            f'''CHANGE MASTER TO MASTER_HOST='{host_primary["name"]}', MASTER_USER='{rep_user}', MASTER_PORT={primary["port"]}, {f"MASTER_SSL=1, MASTER_SSL_VERIFY_SERVER_CERT=0, MASTER_SSL_CAPATH='{standby_data_dir}', MASTER_SSL_CA='{standby_data_dir}/ca.pem', MASTER_SSL_CERT='{standby_data_dir}/client-cert.pem', MASTER_SSL_KEY='{standby_data_dir}/client-key.pem'," if use_ssl else ""} MASTER_PASSWORD='{rep_pwd}', MASTER_LOG_FILE='{master_replication["File"]}', MASTER_LOG_POS={master_replication["Position"]}''',
            "START SLAVE",
        ]

        conn = get_mysql_connection(host_standby, standby)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor() as cur:
                for query in queries:
                    cur.execute(query)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
        
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host_standby, f"sudo du -sb '{standby_data_dir}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = round(float(jsn['stdout']) / 1024, 2)
        remote_command(host_standby, f"""echo "{get_log_rotate_config(standby_data_dir)}" | sudo tee '/etc/logrotate.d/{standby['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host_standby), 'node_name': standby['name']}, 'task': 'log_rotate_task',  'org_id': standby['org_id'],'monitor_name': None,  'target_type': standby['type'], 'target_id': standby['id'], 'created_by': created_by}
        insert_task(params)
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, standby['id'],))
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,standby['id'],created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host_standby['id'],created_by,OperatingSystem.LINUX.value])
                conn.commit()
            
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
    
    # host_standby['create_task'] = created_by
    # host_standby['ssh_port'] = int(host_standby['ssh_port'])
    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host_standby, 'task': 'run_linux_performance_counters', 'target_type': 'host', 'target_id': host_standby['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as error:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def upgrade_mariadb_node_task(data, task_id):
    """upgrade mariadb node to the target major version"""

    task = 'upgrade mariadb node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        target_version = data['target_version']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        target_version_exists = check_mariadb_version_exists(host, target_version, distributor)

        conf_path = "/etc/my.cnf.d/server.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mariadb.conf.d/50-server.cnf"
        conf_file = os.path.basename(conf_path)
        standbys = data.get('standbys')
        has_standby = isinstance(standbys, list) and len(standbys) > 0
        if has_standby:
            standbys = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in standbys]
        
        repository = node['ex_params'].get('repository')
        offline = True if Repository.LOCAL.value in repository else False

        app_host = json.loads(data['app_host'])
        use_ssl = node['ex_params'].get('ssl')
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if has_standby:
        step = 'disconnecting standbys'
        duration_step_start = time.perf_counter()

        last_known_err = ''

        try:
            for item in standbys:
                standby_data = item['node']
                host_standby = item['host']
                rep_user = standby_data['ex_params'].get('rep_user')
                rep_pwd = standby_data['ex_params'].get('rep_pwd')
                conn = get_mysql_connection(host_standby, standby_data)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                "STOP SLAVE",
                                "RESET SLAVE ALL"
                            ]

                            for query in queries:
                                cur.execute(query)
        except Exception as error:
            last_known_err = str(error)
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
             
        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'stopping mariadb primary'
    duration_step_start = time.perf_counter()

    try:
        conn = get_mysql_connection(host, node)
        if not isinstance(conn, tuple):
            with conn:
                with conn.cursor(dictionary=True) as cur:
                    cur.execute("SET GLOBAL innodb_fast_shutdown = 0")

        tmp_conf_path = f"{TASK_TMP_FILES_DIR}/mariadb_conf/{node['name']}"
        conf_backup_path = f"{tmp_conf_path}/{conf_file}"
        commands = [
            f"sudo mkdir -p '{tmp_conf_path}'",
            f"sudo rsync -a '{conf_path}' '{conf_backup_path}'",
            'sudo systemctl stop mariadb'
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '' and 'not loaded' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if not target_version_exists:
        step = 'installing upgraded mariadb binaries'
        duration_step_start = time.perf_counter()

        try:
            if OperatingSystem.CENTOS.value in distributor:
                check_mariadb_exists = "sudo rpm -qa | grep 'MariaDB'"
            else:
                check_mariadb_exists = "sudo dpkg -l | grep 'mariadb'"
            
            jsn = json.loads(remote_command(host, check_mariadb_exists, True), strict=False)
            mariadb_exists = jsn['stdout'] != ''
            if mariadb_exists:
                if OperatingSystem.CENTOS.value in distributor:
                    package_manager = 'dnf' if int(release) > 7 else 'yum'
                    remove_mariadb_cmd = f"sudo {package_manager} -q -y remove MariaDB* galera**"
                else:
                    remove_mariadb_cmd = "sudo DEBIAN_FRONTEND=noninteractive apt-get -qq -y purge ^mariadb ^galera"
            
                jsn = json.loads(remote_command(host, remove_mariadb_cmd, True), strict=False)
                error_lower = jsn['stderr'].lower()
                filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['nothing to do'])
                if filter:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(node['id'])
                    return

            if OperatingSystem.CENTOS.value in distributor:
                check_mysql_exists = "sudo rpm -qa | grep 'mysql'"
            else:
                check_mysql_exists = "sudo dpkg -l | grep 'mysql'"
            
            jsn = json.loads(remote_command(host, check_mysql_exists, True), strict=False)
            mysql_exists = jsn['stdout'] != ''
            if mysql_exists:
                if OperatingSystem.CENTOS.value in distributor:
                    package_manager = 'dnf' if int(release) > 7 else 'yum'
                    remove_mysql_cmd = f"sudo {package_manager} -q -y remove mysql* percona* galera*"
                else:
                    remove_mysql_cmd = "sudo DEBIAN_FRONTEND=noninteractive apt-get -qq -y purge ^mysql ^percona ^galera"

                jsn = json.loads(remote_command(host, remove_mysql_cmd, True), strict=False)
                error_lower = jsn['stderr'].lower()
                filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['nothing to do'])
                if filter:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(node['id'])
                    return
            
            if offline:
                jsn = install_offline_remote_to_remote(distributor, release, host, target_version, DBType.MARIADB.value, app_host)
            else:
                jsn = install_mariadb(distributor, release, host, target_version)
            
            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
            if filter:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return

            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    step = 'starting mariadb primary'
    duration_step_start = time.perf_counter()

    try:
        commands = [
            f"sudo rsync -a '{conf_backup_path}' '{conf_path}'",
            "sudo systemctl stop mariadb",
            'sudo systemctl start mariadb'
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        if has_standby:
            conn = get_mysql_connection(host, node)
            if isinstance(conn, tuple):
                error = str(conn[1])
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                return
            
            with conn:
                with conn.cursor(dictionary=True) as cur:
                    cur.execute("SHOW MASTER STATUS")
                    master_replication = cur.fetchone()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    if has_standby:
        step = 'cloning standbys'
        duration_step_start = time.perf_counter()

        try:
            for item in standbys:
                standby_data = item['node']
                host_standby = item['host']
                rep_user = standby_data['ex_params'].get('rep_user')
                rep_pwd = standby_data['ex_params'].get('rep_pwd')
                
                standby_tmp_conf_dir = f"{TASK_TMP_FILES_DIR}/mariadb_conf/{standby_data['name']}"
                standby_data_dir = standby_data["dir_123Cluster"]
                if standby_data_dir.startswith('~'):
                    jsn = find_tilda(host_standby)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                    tilda_path = jsn['stdout']
                    standby_data_dir = standby_data_dir.replace('~', tilda_path)
                
                commands = [
                    f"sudo mkdir -p '{standby_tmp_conf_dir}'",
                    f"sudo rsync -a '{conf_path}' '{standby_tmp_conf_dir}/{conf_file}'",
                    "sudo systemctl stop mariadb"
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                conn = get_mysql_connection(host_standby, standby_data)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor(dictionary=True) as cur:
                            cur.execute("SET GLOBAL innodb_fast_shutdown = 0")
                
                commands = [
                    f"sudo mkdir -p '{standby_tmp_conf_dir}'",
                    f"sudo rsync -a '{conf_path}' '{standby_tmp_conf_dir}/{conf_file}'",
                    "sudo systemctl stop mariadb"
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                target_version_exists = check_mariadb_version_exists(host_standby, target_version, distributor)
                if not target_version_exists:
                    if OperatingSystem.CENTOS.value in distributor:
                        check_mariadb_exists = "sudo rpm -qa | grep 'MariaDB'"
                    else:
                        check_mariadb_exists = "sudo dpkg -l | grep 'mariadb'"
                    
                    jsn = json.loads(remote_command(host_standby, check_mariadb_exists, True), strict=False)
                    mariadb_exists = jsn['stdout'] != ''
                    if mariadb_exists:
                        if OperatingSystem.CENTOS.value in distributor:
                            package_manager = 'dnf' if int(release) > 7 else 'yum'
                            remove_mariadb_cmd = f"sudo {package_manager} -q -y remove MariaDB* galera**"
                        else:
                            remove_mariadb_cmd = "sudo DEBIAN_FRONTEND=noninteractive apt-get -qq -y purge ^mariadb ^galera"
                    
                        jsn = json.loads(remote_command(host_standby, remove_mariadb_cmd, True), strict=False)
                        error_lower = jsn['stderr'].lower()
                        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['nothing to do'])
                        if filter:
                            last_known_err = jsn
                            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                    if OperatingSystem.CENTOS.value in distributor:
                        check_mysql_exists = "sudo rpm -qa | grep 'mysql'"
                    else:
                        check_mysql_exists = "sudo dpkg -l | grep 'mysql'"
                    
                    jsn = json.loads(remote_command(host_standby, check_mysql_exists, True), strict=False)
                    mysql_exists = jsn['stdout'] != ''
                    if mysql_exists:
                        if OperatingSystem.CENTOS.value in distributor:
                            package_manager = 'dnf' if int(release) > 7 else 'yum'
                            remove_mysql_cmd = f"sudo {package_manager} -q -y remove mysql* percona* galera*"
                        else:
                            remove_mysql_cmd = "sudo DEBIAN_FRONTEND=noninteractive apt-get -qq -y purge ^mysql ^percona ^galera"

                        jsn = json.loads(remote_command(host_standby, remove_mysql_cmd, True), strict=False)
                        error_lower = jsn['stderr'].lower()
                        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['nothing to do'])
                        if filter:
                            last_known_err = jsn
                            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                    
                    if offline:
                        jsn = install_offline_remote_to_remote(distributor, release, host_standby, target_version, DBType.MARIADB.value, app_host)
                    else:
                        jsn = install_mariadb(distributor, release, host_standby, target_version)
                    
                    error_lower = jsn['stderr'].lower()
                    filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
                    if filter:
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                commands = [
                    f"sudo rsync -a '{standby_tmp_conf_dir}/{conf_file}' '{conf_path}' ",
                    "sudo systemctl stop mariadb",
                    "sudo systemctl start mariadb"
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                conn = get_mysql_connection(host, node)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                f"CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED VIA mysql_native_password USING PASSWORD('{rep_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
                                f"ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED VIA mysql_native_password USING PASSWORD('{rep_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
                                f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
                                f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
                                "FLUSH PRIVILEGES",
                            ]
                            for query in queries:
                                cur.execute(query)
                
                conn = get_mysql_connection(host_standby, standby_data)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                "STOP SLAVE",
                                "RESET SLAVE ALL",
                                f'''CHANGE MASTER TO MASTER_HOST='{host["name"]}', MASTER_PORT={node["port"]}, MASTER_USER='{rep_user}', MASTER_PASSWORD='{rep_pwd}', {f"MASTER_SSL=1, MASTER_SSL_VERIFY_SERVER_CERT=0, MASTER_SSL_CAPATH='{standby_data_dir}', MASTER_SSL_CA='{standby_data_dir}/ca.pem', MASTER_SSL_CERT='{standby_data_dir}/client-cert.pem', MASTER_SSL_KEY='{standby_data_dir}/client-key.pem'," if use_ssl else ""} MASTER_LOG_FILE='{master_replication['File']}', MASTER_LOG_POS={master_replication['Position']}''',
                                # does not fail if replication credentials are wrong or empty, always check replication status after restore
                                "START SLAVE"
                            ]
                            for query in queries:
                                cur.execute(query)
        except Exception as error:
            last_known_err = str(error)
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'] if isinstance(last_known_err, dict) else None,last_known_err['stderr'] if isinstance(last_known_err, dict) else last_known_err,last_known_err['excode'] if isinstance(last_known_err, dict) else 1)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='false' WHERE cluster_id=%s""", (Severity.NOT_OK.value, Status.OK.value, target_version, node['cluster_id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def create_mariadb_database_task(data, task_id):
    
    task = 'create mariadb database task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        use_ssl = node['ex_params'].get('ssl')
        db = json.loads(data['db'])
        db_name = db['name']
        db_id = db['id']
        mariadb_version = float(node['version'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating database'
    duration_step_start = time.perf_counter()
    
    try:
        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor() as cur:
                cur.execute(f"CREATE DATABASE IF NOT EXISTS {db_name} CHARACTER SET 'utf8' COLLATE 'utf8_general_ci'")
                # closing step = 'creating database'
                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

                step = 'creating database users'
                duration_step_start = time.perf_counter()

                owner_user_name = f'{db_name}_owner_user'
                write_user_name = f'{db_name}_write_user'
                read_user_name = f'{db_name}_read_user'
                owner_user_pwd = generate_complex_password()
                write_user_pwd = generate_complex_password()
                read_user_pwd = generate_complex_password()
                logs_output = f'USERNAME = {owner_user_name}\nPASSWORD = {owner_user_pwd}\nUSERNAME = {read_user_name}\nPASSWORD = {read_user_pwd}\nUSERNAME = {write_user_name}\nPASSWORD = {write_user_pwd}'
                owner_user_pwd = format_for_shell(owner_user_pwd)
                write_user_pwd = format_for_shell(write_user_pwd)
                read_user_pwd = format_for_shell(read_user_pwd)

                queries = [
                    f"CREATE USER IF NOT EXISTS '{owner_user_name}'@'localhost' IDENTIFIED VIA mysql_native_password USING PASSWORD('{owner_user_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
                    f"ALTER USER '{owner_user_name}'@'localhost' IDENTIFIED VIA mysql_native_password USING PASSWORD('{owner_user_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
                    f"GRANT ALL PRIVILEGES ON {db_name}.* TO '{owner_user_name}'@'localhost' WITH GRANT OPTION",
                    f"CREATE USER IF NOT EXISTS '{write_user_name}'@'localhost' IDENTIFIED VIA mysql_native_password USING PASSWORD('{write_user_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
                    f"ALTER USER '{write_user_name}'@'localhost' IDENTIFIED VIA mysql_native_password USING PASSWORD('{write_user_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
                    f"GRANT SELECT, INSERT, UPDATE ON {db_name}.* TO '{write_user_name}'@'localhost'",
                    f"CREATE USER IF NOT EXISTS '{read_user_name}'@'localhost' IDENTIFIED VIA mysql_native_password USING PASSWORD('{read_user_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
                    f"ALTER USER '{read_user_name}'@'localhost' IDENTIFIED VIA mysql_native_password USING PASSWORD('{read_user_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
                    f"GRANT SELECT ON {db_name}.* TO  '{read_user_name}'@'localhost'",
                    "FLUSH PRIVILEGES",
                    # if database are just created - size always be none, but still check, for the case of manual task rerun
                    f"SELECT SUM(data_length + index_length)/1024 FROM information_schema.TABLES WHERE table_schema='{db_name}' GROUP BY table_schema",
                ]

                for query in queries:
                    cur.execute(query)
                
                res = cur.fetchone()
                if res and res[0]:
                    db_size = res[0]

                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,logs_output,None,0)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        node_size = round(float(jsn['stdout']) / 1024, 2)
        

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_size,db_id))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id']))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
   
def drop_mariadb_database_task(data, task_id):

    task = 'drop mariadb database task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        db = json.loads(data['db_data'])
        db_name = db['name']
        db_id = db['id']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'dropping the database'
    duration_step_start = time.perf_counter()

    try:
        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor() as cur:
                # kill all the connections to db <db_name>
                cur.execute("SELECT id FROM information_schema.processlist WHERE db = %s", (db_name,))
                connection_ids = [i[0] for i in cur.fetchall()]
                #although some blazing fast client could reconnect during loop - thats the only way...
                for connection in connection_ids:
                    cur.execute(f"KILL {connection}")
                # drop db <db_name> with the associated dba users (order by importance ascending)
                queries = [
                    f"DROP USER IF EXISTS '{db_name}_read_user'@'localhost'",
                    f"DROP USER IF EXISTS '{db_name}_write_user'@'localhost'",
                    f"DROP USER IF EXISTS '{db_name}_owner_user'@'localhost'",
                    "FLUSH PRIVILEGES",
                    f"DROP DATABASE IF EXISTS {db_name}"
                ]
                for query in queries:
                    cur.execute(query)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = round(float(jsn['stdout']) / 1024, 2)
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DROPPED.value, db_id,))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def start_mariadb_node_task(data, task_id):
    """start mariadb node"""

    task = 'start mariadb node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node_id = data['node_id']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'starting mariadb node'
    duration_step_start = time.perf_counter()

    try:
        command = 'sudo systemctl start mariadb'
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_id,))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node_id,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def stop_mariadb_node_task(data, task_id):
    """stop mariadb node"""

    task = 'stop mariadb node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node_id = data['node_id']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stopping mariadb node'
    duration_step_start = time.perf_counter()

    try:
        command = 'sudo systemctl stop mariadb'
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false' WHERE id=%s""", (Severity.NOT_OK.value, Status.OK.value, node_id,))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='false' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node_id,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restart_mariadb_node_task(data, task_id):
    """restart mariadb node"""

    task = 'restart mariadb node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node_id = data['node_id']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'restarting mariadb node'
    duration_step_start = time.perf_counter()

    try:
        command = 'sudo systemctl restart mariadb'
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_id,))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node_id,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def backup_mariadb_node_task(data, task_id):
    """backup of mariadb node"""
    
    task = 'backup mariadb node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        backup_host = data['backup_host']
        node = json.loads(data['node_data'])
        backup = json.loads(data['backup_data'])
        node = json.loads(data['node_data'])
        compress = backup["ex_params"].get('compress')
        is_sched = backup['is_sched']
        incremental = backup['ex_params'].get("incremental")
        mariadb_version = float(node['version'])
        backup_path = backup['ex_params'].get('backup_path')
        if backup_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'backing up node'
    duration_step_start = time.perf_counter()

    command = f"sudo {'strace -e open' if mariadb_version == 10.3 else ''} maria{'' if mariadb_version < 11.0 else 'db-'}backup --backup --user=root --port='{node['port']}' --password='{node['ex_params'].get('mariadb_root_password')}'"

    if not isinstance(incremental, str):
        if not is_sched:
            command += f" --target-dir='{backup_path}'"
        elif incremental is False:
            start_date = datetime.now().astimezone(pytz.UTC).strftime('%d%m%Y%H%M%S')
            backup_path = f"{backup_path}/{start_date}"
            command += f" --target-dir='{backup_path}'"
        elif incremental is True:
            incremental = f"{backup_path}/base"
            command += f" --target-dir='{incremental}'"
    else:
        jsn = check_remote_directory_exists(backup_host, incremental, 'd')
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        incremental_exists = 'true' in jsn['stdout']
        if incremental_exists:
            jsn = json.loads(remote_command(backup_host, f"sudo grep 'to_lsn' '{incremental}/{'mariadb_' if mariadb_version >= 11.0 else 'xtra'}backup_checkpoints' | awk '{{print $3}}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            last_lsn = jsn['stdout']
            if last_lsn:
                incremental = f"{backup_path}/{last_lsn}"
                command += f" --incremental-lsn={last_lsn}"
            else:
                incremental = f"{backup_path}/base"
        else:
            base_dir = f"{backup_path}/base"
            jsn = check_remote_directory_exists(backup_host, base_dir, 'd')
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            base_exists = 'true' in jsn['stdout']
            if not base_exists:
                incremental = base_dir
            else:
                try:
                    jsn = json.loads(remote_command(backup_host, f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n | tail -1 | awk '{{print $2}}'", True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return

                    if jsn['stdout'] != '':
                        dir_list = jsn['stdout'].split('\n')
                        if len(dir_list) > 0:
                            incremental = dir_list[0]
                            jsn = json.loads(remote_command(backup_host, f"sudo grep 'to_lsn' '{incremental}/{'mariadb_' if mariadb_version >= 11.0 else 'xtra'}backup_checkpoints' | awk '{{print $3}}'", True), strict=False)
                            if jsn['stderr'] != '':
                                status = Severity.NOT_OK.value
                                duration_step_end = time.perf_counter() - duration_step_start
                                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                                step = 'null'
                                status = Severity.NOT_OK.value
                                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                                return

                            last_lsn = jsn['stdout']
                            if last_lsn:
                                incremental = f"{backup_path}/{last_lsn}"
                                command += f" --incremental-lsn={last_lsn}"
                except Exception as error:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                    return
    
        command += f" --target-dir='{incremental}'"

    try:
        jsn = json.loads(remote_command(host, f"sudo mkdir -p '{backup_path if not incremental else incremental}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        print(traceback.format_exc())
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    try:
        if compress and incremental:
            command += f" --extra-lsndir='{incremental}'"
        
        if compress:
            command += f" --stream=mbstream | sudo gzip -c | sudo tee '{incremental if incremental else backup_path}/{backup['name']}.gz'"

        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        error_lower = jsn['stderr'].lower()
        if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host['name'] != backup_host['name']:
        step = 'copying backup of node to backup host'
        duration_step_start = time.perf_counter()

        transfer_path = backup_path if not incremental else incremental
        jsn = transfer_files_rsync(host, backup_host, transfer_path, [f"{transfer_path}/"])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking the backup size'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"""sudo du -b{'' if compress else 's'} '{incremental if incremental else backup_path}{f'/{backup["name"]}.gz' if compress else ''}' | awk '{{print $1}}'""", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        size = jsn['stdout'].strip()
        backup_size = float(size)/1024
        
        modified_date = get_date_modified_folder(hostdata=backup_host, folder_path=backup_path)
        if isinstance(modified_date, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return
        
        remote_command(host, f"sudo rm -rf '{backup_path}'", True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
           with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,backup['id']))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}' WHERE id=%s""" % (modified_date, backup['id']))
                if isinstance(incremental, str):
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{incremental}}', '"{0}"', true) WHERE id={1}""".format(incremental,backup['id']))
               
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restore_mariadb_node_task(data, task_id):
    """restore node from backup"""

    task = 'restore mariadb node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        node = json.loads(data['node_data'])
        use_ssl = node['ex_params'].get('ssl')
        backup = json.loads(data['backup_data'])
        backup_host = data['backup_host']
        mariadb_version = float(node['version'])
        backup_cli = 'mariabackup' if mariadb_version < 11.0 else 'mariadb-backup'
        metadata_prefix = 'xtrabackup'if mariadb_version < 11.1 else 'mariadb_backup'
        compress = backup['ex_params'].get('compress')
        node_datadir = node['dir_123Cluster']
        restore_option = data['restore_option'].strip().lower()
        custom_pitr = RestoreOptions.CUSTOM.value in restore_option
        no_pitr = RestoreOptions.THIS.value in restore_option
        incremental = backup["ex_params"].get("incremental")
        node_root_pwd = node["ex_params"].get("mariadb_root_password")
        is_sched = backup['is_sched']
        conf_path = "/etc/my.cnf.d/server.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mariadb.conf.d/50-server.cnf"
        conf_file = os.path.basename(conf_path)
        primary_clone = f"{TASK_TMP_FILES_DIR}/mariadb_clones/{node['name'].replace(':', '_')}"
        tmp_conf_path = f"{TASK_TMP_FILES_DIR}/mariadb_conf/{node['name']}"
        standbys = data.get('standbys')
        has_standby = isinstance(standbys, list) and len(standbys) > 0
        if has_standby:
            standbys = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in standbys]

        backup_path = backup['ex_params'].get('backup_path')
        if backup_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)

        base_dir = f"{backup_path}/base"
        dir_list = None
        if is_sched and not (incremental and no_pitr):
            if restore_option in [RestoreOptions.LATEST.value, RestoreOptions.CUSTOM.value]:
                if custom_pitr:
                    timezone_cmd = 'cat /etc/timezone'
                    jsn = json.loads(remote_command(backup_host, timezone_cmd, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return

                    try:
                        timezone = pytz.timezone(jsn['stdout'])
                        target_datetime = datetime.strptime(data['date'], "%Y-%m-%d %H:%M:%S").astimezone(timezone)
                        target_datetime_str = target_datetime.strftime("%Y-%m-%d %H:%M:%S")
                        target_timestamp = target_datetime.timestamp()
                    except Exception as e:
                        error = f'Invalid server timezone: {jsn["stdout"]} error: {str(e)}'
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
                        return
    
                if incremental:
                    if custom_pitr:
                        command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d  ! -newermt '{target_datetime_str}' -exec stat -c '%Y %n' {{}} \; | grep -v '/base' | sort -n | awk '{{print $2}}'"
                    else:
                        command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; |  grep -v '/base' | sort -n | awk '{{print $2}}'"

                elif not incremental:
                    if custom_pitr:
                        command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n"
                    else:
                        command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n | tail -1 | awk '{{print $2}}'"
    
            elif no_pitr:
                command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n | head -1 | awk '{{print $2}}'"

            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            if jsn['stdout'] != '':
                if incremental and not no_pitr:
                    dir_list = jsn['stdout'].split('\n')

                if not incremental:
                    if jsn['stdout'] == '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                        return
                    else:
                        if restore_option in [RestoreOptions.THIS.value, RestoreOptions.LATEST.value]:
                            backup_path = jsn['stdout']
                        else:
                            output_lines = [line.strip() for line in jsn['stdout'].split("\n") if line.strip()]
                            closest_subfolder = None
                            min_time_difference = positive_infinity

                            for line in output_lines:
                                subfolder_timestamp_str, subfolder = line.split(" ", 1)
                                subfolder_timestamp = int(subfolder_timestamp_str)
                                time_difference = abs(target_timestamp - subfolder_timestamp)
                                if time_difference < min_time_difference:
                                    min_time_difference = time_difference
                                    closest_subfolder = subfolder

                            if not closest_subfolder:
                                status = Severity.NOT_OK.value
                                duration_step_end = time.perf_counter() - duration_step_start
                                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                                step = 'null'
                                status = Severity.NOT_OK.value
                                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                                return

                            backup_path = closest_subfolder
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host['name'] != backup_host['name']:
        step = 'copying backup files of node to new host'
        duration_step_start = time.perf_counter()

        jsn = transfer_files_rsync(backup_host, host, backup_path, [f"{backup_path}/"])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'preparing to restore'
    duration_step_start = time.perf_counter()

    commands = [
        f"sudo mkdir -p '{tmp_conf_path}'",
        f"sudo rsync -a '{conf_path}' '{tmp_conf_path}/{conf_file}'",
    ]
    
    if is_sched and incremental:
        if compress:
            filename_cmd = f"sudo ls -lt --file-type '{base_dir}' | grep '\.gz$' | head -n 1 | awk '{{print $9}}'"
            jsn = json.loads(remote_command(host, filename_cmd, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            archive_file = jsn['stdout']
            commands.append(f"sudo find '{base_dir}' -mindepth 1 -maxdepth 1 -type f -name '{metadata_prefix}*' -print0 | xargs -0 sudo rm -f && sudo gunzip < '{base_dir}/{archive_file}' | sudo mbstream -C '{base_dir}' -x && sudo chown -R mysql:mysql '{base_dir}'")
        
        commands.append(f"sudo {backup_cli} --prepare  --target-dir='{base_dir}'")
        
        if dir_list:
            for inc_dir in dir_list:
                inc_dir = inc_dir.strip()
                prepare_dir = [f"sudo {backup_cli} --prepare {'' if inc_dir != dir_list[:-1] else ''} --target-dir='{base_dir}' --incremental-dir='{inc_dir}'"]
                if compress:
                    filename_cmd = f"sudo ls -lt --file-type '{inc_dir}' | grep '\.gz$' | head -n 1 | awk '{{print $9}}'"
                    jsn = json.loads(remote_command(host, filename_cmd, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                    
                    archive_file = jsn['stdout']
                    prepare_dir.insert(0,f"sudo find '{inc_dir}' -mindepth 1 -maxdepth 1 -type f -name '{metadata_prefix}*' -print0 | xargs -0 sudo rm -f && sudo gunzip < '{inc_dir}/{archive_file}' | sudo mbstream -C '{inc_dir}' -x && sudo chown -R mysql:mysql '{inc_dir}'")

                commands += prepare_dir
    else:
        filename_cmd = f"sudo ls -lt --file-type '{backup_path}' | grep '\.gz$' | head -n 1 | awk '{{print $9}}'"
        jsn = json.loads(remote_command(host, filename_cmd, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        archive_file = jsn['stdout']
        commands.append(f"sudo chown -R mysql:mysql '{backup_path}'" if not compress else f"sudo gunzip < '{backup_path}/{archive_file}' | sudo mbstream -C '{backup_path}' -x && sudo chown -R mysql:mysql '{backup_path}'")
                
    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
                   
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if has_standby:
        step = 'disconnecting standbys'
        duration_step_start = time.perf_counter()

        last_known_err = None
        try:
            for item in standbys:
                standby_data = item['node']
                host_standby = item['host']
                conn = get_mysql_connection(host_standby, standby_data)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                "STOP SLAVE",
                                "RESET SLAVE ALL"
                            ]
                            for query in queries:
                                cur.execute(query)
        except Exception as error:
            logging.error(f"{task}@{step}: {task_id}: {str(error)}")

        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'restoring node from backup'
    duration_step_start = time.perf_counter()

    try:
        if use_ssl:
            jsn = transfer_ssl_certificates(node, host, host, mode='backup')
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        commands = [
            f"sudo {backup_cli} --prepare --target-dir='{backup_path if not (is_sched and incremental) else base_dir}'",
            f'sudo systemctl stop mariadb',
            f"sudo rm -rf '{node_datadir}'",
            f"sudo mkdir -p '{node_datadir}'",
            f"sudo {backup_cli} --copy-back --target-dir='{backup_path if not (is_sched and incremental) else base_dir}'",
            f"sudo rsync -a '{tmp_conf_path}/{conf_file}' '{conf_path}'",
            f"sudo rm -rf '{tmp_conf_path}'",
            f"sudo sed -i '/read_only=ON/d' '{conf_path}'",
            f"sudo chown -R mysql:mysql '{node_datadir}'",
            f"sudo mkdir -p '/var/run/mariadb'",
        ]

        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                if backup_host['name'] != host['name']:
                    remote_command(host, f"sudo rm -rf '{backup_path}'", True) 
                return
            
        if use_ssl:
            jsn = transfer_ssl_certificates(node, host, host, mode='restore')
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        commands = [
            f"sudo chown -R mysql:mysql '/var/run/mariadb'",
            f'sudo systemctl start mariadb'
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        remote_command(host, f"sudo rm -rf '{backup_path}'", True)
        
        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor(dictionary=True) as cur:
                cur.execute("SHOW MASTER STATUS")
                master_replication = cur.fetchone()
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role=%s, ex_params = ex_params || '{"mariadb_master_id":%s}', ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.PRIMARY.value, node['id'], node['id']))
                conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if has_standby:
        step = 'restoring replication'
        duration_step_start = time.perf_counter()
        
        now = datetime.utcnow().strftime('%d%m%Y%H%M%S')
        backup_name = f"{node['name'].replace(':', '_')}_{now}"
        transfer_path = f"{primary_clone}/{backup_name}.gz"
        backup_cmd = f"sudo {backup_cli} --backup --user=root --port='{node['port']}' --password='{node_root_pwd}' --target-dir='{primary_clone}' --stream=mbstream | sudo gzip -c | sudo tee '{transfer_path}'"
        last_known_err = None
        try:
            commands = [
                f"sudo mkdir -p '{primary_clone}'",
                backup_cmd
            ]
            for command in commands:
                jsn = json.loads(remote_command(host, command, True), strict=False)
                error_lower = jsn['stderr'].lower()
                if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
            for item in standbys:
                standby_data = item['node']
                host_standby = item['host']
                rep_user = standby_data["ex_params"].get("rep_user")
                rep_pwd = standby_data["ex_params"].get("rep_pwd")
                standby_tmp_conf_dir = f"{TASK_TMP_FILES_DIR}/mariadb_conf/{standby_data['name']}"
                standby_data_dir = standby_data["dir_123Cluster"]
                if standby_data_dir.startswith('~'):
                    jsn = find_tilda(host_standby)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                tilda_path = jsn['stdout']
                standby_data_dir = standby_data_dir.replace('~', tilda_path)
                
                commands = [
                    f"sudo mkdir -p '{standby_tmp_conf_dir}'",
                    f"sudo rsync -a '{conf_path}' '{standby_tmp_conf_dir}/{conf_file}'",
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                jsn = transfer_files_rsync(host, host_standby, primary_clone, [transfer_path])
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                conn = get_mysql_connection(host, node)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                f"CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED VIA mysql_native_password USING PASSWORD('{rep_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
                                f"ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED VIA mysql_native_password USING PASSWORD('{rep_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
                                f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
                                f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
                                "FLUSH PRIVILEGES",
                            ]
                            for query in queries:
                                cur.execute(query)
                
                commands = [
                    f"sudo gunzip < '{transfer_path}' | sudo mbstream -C '{primary_clone}' -x",
                    f"sudo chown -R mysql:mysql '{primary_clone}'",
                    f"sudo {backup_cli} --prepare --target-dir='{primary_clone}'",
                    f'sudo systemctl stop mariadb',
                    f"sudo rm -rf '{standby_data_dir}'",
                    f"sudo mkdir -p '{standby_data_dir}'",
                    f"sudo {backup_cli} --copy-back --target-dir='{primary_clone}'",
                    f"sudo rsync -a '{standby_tmp_conf_dir}/{conf_file}' '{conf_path}'",
                    f"sudo rm -rf '{standby_tmp_conf_dir}'",
                    "sudo mkdir -p '/var/run/mariadb'",
                    "sudo chown -R mysql:mysql '/var/run/mariadb'",
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    error_lower = jsn['stderr'].lower()
                    if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                if use_ssl:
                    jsn = transfer_ssl_certificates(node, host, host_standby, standby_data_dir)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                
                commands = [
                    f"sudo chown -R mysql:mysql '{standby_data_dir}'",
                    f"sudo systemctl start mariadb",
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                remote_command(host_standby, f"sudo rm -rf '{primary_clone}'", True)
                
                conn = get_mysql_connection(host_standby, standby_data)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                "STOP SLAVE",
                                "RESET SLAVE ALL",
                                #change master command does not fail if replication credentials are wrong or empty, always check replication status after restore
                                f'''CHANGE MASTER TO MASTER_HOST='{host["name"]}', MASTER_USER='{rep_user}', MASTER_PORT={node["port"]}, MASTER_PASSWORD='{rep_pwd}', {f"MASTER_SSL=1, MASTER_SSL_VERIFY_SERVER_CERT=0, MASTER_SSL_CAPATH='{standby_data_dir}', MASTER_SSL_CA='{standby_data_dir}/ca.pem', MASTER_SSL_CERT='{standby_data_dir}/client-cert.pem', MASTER_SSL_KEY='{standby_data_dir}/client-key.pem'," if use_ssl else ""} MASTER_LOG_FILE='{master_replication['File']}', MASTER_LOG_POS={master_replication['Position']}''',
                                "START SLAVE"
                            ]
                            for query in queries:
                                cur.execute(query)
                
                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role =%s, ex_params = ex_params || '{"mariadb_master_id":%s}', ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.STANDBY.value, node['id'], standby_data['id'],))
                        conn.commit()
                
            remote_command(host, f"sudo rm -rf '{primary_clone}'", True)
        except Exception as error:
            last_known_err = str(error)
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
    
        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'] if isinstance(last_known_err, dict) else None,last_known_err['stderr'] if isinstance(last_known_err, dict) else last_known_err,last_known_err['excode'] if isinstance(last_known_err, dict) else 1)
    
    last_known_err = None
    
    step = 'updating app database'
    duration_step_start = time.perf_counter()

    try:
        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor(dictionary=True) as cur:
                cur.execute("SELECT DISTINCT TABLE_SCHEMA FROM information_schema.TABLES WHERE TABLE_SCHEMA NOT IN ('information_schema', 'mysql', 'performance_schema', 'sys', 'test', 'tmp')")
                db_set_from_restore = {i['TABLE_SCHEMA'].strip() for i in cur.fetchall()}
                with get_postgres_connection() as app_conn:
                    with app_conn.cursor(cursor_factory=RealDictCursor) as app_cur:
                        app_cur.execute("""SELECT DISTINCT ON (name) id, name FROM "clusterCore_c123_database" WHERE node_id=%s""", (node['id'],))
                        db_list_from_app = app_cur.fetchall()
                        db_set_from_app = {db['name'].strip() for db in db_list_from_app}
                        common_db = db_set_from_app.intersection(db_set_from_restore)
                        missing_db = db_set_from_restore - db_set_from_app
                        extra_db = db_set_from_app - db_set_from_restore
                 
                        # in order of workload ascending
                        for db in extra_db:
                            app_cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", [Severity.NOT_OK.value, Status.OK.value, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                            app_conn.commit()

                        for db in common_db:
                            cur.execute("SELECT SUM(data_length + index_length)/1024 AS DB_SIZE_KB FROM information_schema.TABLES WHERE TABLE_SCHEMA = %s", (db,))
                            res = cur.fetchone()['DB_SIZE_KB']
                            db_size = 0.0 if not res else res
                            app_cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, "size"=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, Status.OK.value, db_size, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                            app_conn.commit()
                            
                        for db in missing_db:
                            cur.execute("SELECT SUM(data_length + index_length)/1024 AS DB_SIZE_KB FROM information_schema.TABLES WHERE TABLE_SCHEMA = %s", (db,))
                            res = cur.fetchone()['DB_SIZE_KB']
                            db_size = 0.0 if not res else res

                            app_cur.execute("""INSERT INTO "clusterCore_c123_database"
                            ("name", "type", created_date, last_update_date, size, ind_deployed, ind_started, ind_autostartup, ind_deleted, severity, sizing, database_user, database_pwd, created_by_id, last_update_by_id, node_id)
                            VALUES (%s, %s, now(), now(), %s, 'false', 'true', 'false', 'false', %s, 'MEDIUM', %s, %s, %s, %s, %s)""", (db, DBType.MARIADB.value, db_size, Severity.OK.value, node['database_user'], node['database_pwd'], created_by, created_by, node['id']))
                            app_conn.commit()
                        
                        jsn = json.loads(remote_command(host, f"sudo du -bs '{node_datadir}' | awk '{{print $1}}'", True), strict=False)
                        if jsn['stderr'] != '':
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return
                        
                        node_size = round(float(jsn['stdout']) / 1024, 2)
                        app_cur.execute("""UPDATE "clusterCore_c123_node" SET size=%s WHERE id=%s""", (node_size, node['id'],))
                        app_conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end, created_by)

def copy_mariadb_database_task(data, task_id):
    """copy database from some instance into another"""

    task = 'copy mariadb database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_source = json.loads(data['host_source'])
        host_target = json.loads(data['host_target'])
        node_source = json.loads(data['node_source'])
        node_target = json.loads(data['node_target'])
        use_ssl = node_target['ex_params'].get('ssl')
        db_source = json.loads(data['db_source'])
        db_target = json.loads(data['db_target'])
        # db_option = 'override' or 'copy'
        db_option = data['db_option']
        db_exists = data['db_exists']
        copy_users = data['copy_users']
        copy_name = data.get('copy_name')
        target_version = float(node_target['version'])
        source_version = float(node_source['version'])
        cli_target = 'mysql' if target_version < 10.4 else 'mariadb'
        source_prefix = 'mysql' if source_version < 11.0 else 'mariadb-'
        # copy_option = data['copy_option']
        new_db_name = copy_name if isinstance(copy_name, str) and len(copy_name.strip()) > 0 else db_target["name"]
        db_source_name = db_source['name'].strip() 
        node_source_name = node_source['name'].strip()
        host_source_name = host_source['name'].strip()
        host_target_name = host_target['name'].strip()
        tmp_folder = f"{TASK_TMP_FILES_DIR}/{node_source_name}"
        file = f"{tmp_folder}/{db_source_name}.dump"
        source_root_password = node_source["ex_params"].get("mariadb_root_password")
        target_root_password = node_target["ex_params"].get("mariadb_root_password")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'preparing to copy'
    duration_step_start = time.perf_counter()
    
    if host_source_name == host_target_name:
        # split same host dump preparation from copy - this runs now
        commands_target = [
            f"sudo mkdir -p '{tmp_folder}'",
            f"sudo {source_prefix}dump --default-character-set=utf8mb4 --port='{node_source['port']}' --password='{source_root_password}' --routines --events --quick {db_source_name} --result-file='{file}' ",
            f"sudo chown mysql:mysql '{file}'",
        ]
        for command in commands_target:
            jsn = json.loads(remote_command(host_target, command, True), strict=False)
            error_lower = jsn['stderr'].lower()
            if error_lower != '' and 'error' in error_lower:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        #this runs only after prepare dump + prepare target db
        commands_target = [
            f"sudo {cli_target} --default-character-set=utf8mb4 -Nsfu root --password='{target_root_password}' '{new_db_name}' < '{file}'"
        ]
    else:
        commands_source = [
            f"sudo mkdir -p '{tmp_folder}'",
            f"sudo chown -R mysql:mysql '{tmp_folder}'",
            f"sudo {source_prefix}dump --default-character-set=utf8mb4 --port='{node_source['port']}' --password='{source_root_password}' --routines --events --quick '{db_source_name}' --result-file='{file}'",
        ]
        for command in commands_source:
            try:
                jsn = json.loads(remote_command(host_source, command, True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and 'error' in error_lower:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
            
        jsn = transfer_files_rsync(host_source, host_target, tmp_folder, [file])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        commands_target = [
            f"sudo chown mysql:mysql '{file}'",
            f"sudo {cli_target} --default-character-set=utf8mb4 -Nsfu root --password='{target_root_password}' '{new_db_name}' < '{file}'",
        ]
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if db_exists and db_option == "override":
        step = 'removing target database'
        duration_step_start = time.perf_counter()

        try:
            conn = get_mysql_connection(host_target, node_target)
            if isinstance(conn, tuple):
                error = str(conn[1])
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                return
            
            with conn:
                with conn.cursor() as cur:
                    # kill all the connections to db <db_name>
                    cur.execute("SELECT id FROM information_schema.processlist WHERE db = %s", (new_db_name,))
                    connection_ids = [i[0] for i in cur.fetchall()]
                    #although some blazing fast client could reconnect during loop - thats the only way...
                    for connection in connection_ids:
                        cur.execute(f"KILL {connection}")
                    
                    cur.execute(f"DROP DATABASE IF EXISTS {new_db_name}")
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
         
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'copying the database'
    duration_step_start = time.perf_counter()

    try:
        conn = get_mysql_connection(host_target, node_target)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor() as cur:
                cur.execute(f"CREATE DATABASE IF NOT EXISTS {new_db_name} CHARACTER SET 'utf8' COLLATE 'utf8_general_ci'")

        for command in commands_target:
            jsn = json.loads(remote_command(host_target, command, True), strict=False)
            error_lower = jsn['stderr'].lower()
            if error_lower != '' and 'error' in error_lower:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    if copy_users and not (host_source_name == host_target_name and db_option == 'override'):
        step = 'copying database users'
        duration_step_start = time.perf_counter()

        create_users_builder = f'''SELECT CONCAT('CREATE USER IF NOT EXISTS ', QUOTE(mysql.user.user), '@', QUOTE(mysql.user.host), ' IDENTIFIED VIA ', mysql.user.plugin, ' USING ', 'PASSWORD(', '\\'initial_password\\'', ') {'REQUIRE SSL' if use_ssl else ''};') AS create_user_command FROM mysql.user JOIN mysql.db ON mysql.user.user = mysql.db.user WHERE Db = %s'''
        set_auth_builder = '''SELECT CONCAT('SET PASSWORD FOR ', QUOTE(user),'@',QUOTE(host),' =',QUOTE(password),';') AS set_password_query FROM mysql.user WHERE password IS NOT NULL AND EXISTS (SELECT 1 FROM mysql.db WHERE mysql.user.User = mysql.db.User AND mysql.user.Host = mysql.db.Host AND mysql.db.Db = %s);'''
        show_grants_builder = '''SELECT CONCAT('SHOW GRANTS FOR ', QUOTE(user), '@', QUOTE(host), ';') AS show_grants_command FROM mysql.db WHERE Db = %s'''
        
        try:
            conn = get_mysql_connection(host_source, node_source)
            if isinstance(conn, tuple):
               error = str(conn[1])
               status = Severity.NOT_OK.value
               duration_step_end = time.perf_counter() - duration_step_start
               logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
               step = 'null'
               status = Severity.NOT_OK.value
               duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
               logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
               return
        
            with conn:
                with conn.cursor() as cur:
                    cur.execute(create_users_builder, (db_source_name,))
                    migrate_user_queries = [i[0] for i in cur.fetchall()]
                    cur.execute(set_auth_builder, (db_source_name,))
                    migrate_user_queries += [i[0] for i in cur.fetchall()]
                    cur.execute(show_grants_builder, (db_source_name,))
                    show_grants_queries = [i[0] for i in cur.fetchall()]
                    for show_grants_query in show_grants_queries:
                        cur.execute(show_grants_query)
                        migrate_grants_queries = [i[0] for i in cur.fetchall()]
                        for migrate_grants_query in migrate_grants_queries:
                            migrate_user_queries.append(migrate_grants_query.replace("`", "'").replace("'.", ".").replace(f"{db_source_name}.", f"{new_db_name}.").replace("ON '", "ON "))
         
            migrate_user_queries.append("FLUSH PRIVILEGES")
            conn = get_mysql_connection(host_target, node_target)
            if isinstance(conn, tuple):
                error = str(conn[1])
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                return
        
            with conn:
                with conn.cursor() as cur:
                    for query in migrate_user_queries:
                        cur.execute(query)
        except Exception as error:
            print(traceback.format_exc())
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        jsn = json.loads(remote_command(host_target, f"sudo du -sb '{node_target['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = round(float(jsn['stdout']) / 1024, 2)
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node_target['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_target['id'],))
                if not db_exists:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_source['id'],))
                conn.commit()
        
        remote_command(host_source, f"sudo rm -rf '{tmp_folder}'", True)
        if host_source['name'] != host_target['name']:
            remote_command(host_source, f"sudo rm -rf '{tmp_folder}'", True)
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def drop_mariadb_node_task(data, task_id):
    """drop instance"""

    task = 'drop mariadb node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        host_primary = data.get('host_primary')
        primary = data.get('primary_data')
        node = json.loads(data['node_data'])
        final_backup = data.get('final_backup')
        backup_host = data['backup_host']
        mariadb_version = float(node['version'])
        backup_cli = 'mariabackup' if mariadb_version < 11.0 else 'mariadb-backup'
        rep_user = node["ex_params"].get("rep_user")
        standbys = data.get('standbys')
        has_standby = isinstance(standbys, list) and len(standbys) > 0
        if has_standby:
            standbys = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in data["standbys"]]

        data_directory = node['dir_123Cluster']
        if data_directory.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_directory = data_directory.replace('~', tilda_path)

        conf_path = "/etc/my.cnf.d/server.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mariadb.conf.d/50-server.cnf"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node existence'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node['port']}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        if jsn['stdout'] == '':
            jsn = json.loads(remote_command(host, f"sudo rm -rf '{data_directory}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'{node["name"]} is already deleted',1)

            try:
                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status='deleted', ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, node['id'],))
                        cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status='deleted', ind_started='false', ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, node['id'],))
                        conn.commit()
            except Exception as error:
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

            step = 'task finished'
            status = Severity.WARNING.value
            duration = 0
            logger(task_id,task_log_id,task,step,status,duration,created_by)
            step = 'null'
            status = Severity.WARNING.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    try:
        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,'',str(conn[1]),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,'',str(conn[1]),9)
            return
       
        is_primary = False
        with conn:
            #prefer dictionary, as more pythonic and readable
            with conn.cursor() as cur:
                cur.execute("SHOW SLAVE STATUS")
                res = cur.fetchone()
                if not res:
                    is_primary = True
    
        if is_primary and has_standby:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'{node["name"]} is primary with standbys.Switchover first.',1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,f'{node["name"]} is primary with standbys.Switchover first.',1)
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    if node['cluster_role'] != ClusterRoles.PRIMARY.value:
        final_backup = False
    if final_backup:
        try:
            step = 'backing up the node'
            duration_step_start = time.perf_counter()
            
            backup_path =  final_backup['ex_params'].get('backup_path')
            
            commands = [
                f"sudo rm -rf '{backup_path}'",
                f"sudo mkdir -p '{backup_path}'",
            ]
            for cmd in commands:
                jsn = json.loads(remote_command(host, cmd, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            
            cmd = f"sudo {'strace -e open' if mariadb_version == 10.3 else ''} {backup_cli} --backup --user=root --port='{node['port']}' --password='{node['ex_params'].get('mariadb_root_password')}' --target-dir='{backup_path}' --stream=mbstream | sudo gzip -c | sudo tee '{backup_path}/{final_backup['name']}.gz'"
            jsn = json.loads(remote_command(host, cmd, True), strict=False)

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            jsn = transfer_files_rsync(host, backup_host, backup_path, [f"{backup_path}/"])
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            jsn = json.loads(remote_command(backup_host, f"""sudo du -b '{backup_path}{f'/{final_backup["name"]}.gz'}' | awk '{{print $1}}'""", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            size = jsn['stdout'].strip()
            backup_size = float(size)/1024

            modified_date = get_date_modified_folder(hostdata=backup_host, folder_path=backup_path)
            if isinstance(modified_date, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
                return
            
            remote_command(host, f"sudo rm -rf '{backup_path}'", True)

            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size, final_backup['id']))
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}' WHERE id=%s""" % (modified_date, final_backup['id']))
                    conn.commit()
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    if host_primary and primary and rep_user and not is_primary:
        step = 'drop replication user (if exists)'
        duration_step_start = time.perf_counter()

        try:
            primary = json.loads(primary)
            conn = get_mysql_connection(host_primary, primary)
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,'',str(conn[1]),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,'',str(conn[1]),9)
                return
            
            queries = [
                f"DROP USER IF EXISTS '{rep_user}'@'%.%.%.%'",
                "FLUSH PRIVILEGES"
            ]
            with conn:
                with conn.cursor() as cur:
                    for query in queries:
                        cur.execute(query)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'removing node'
    duration_step_start = time.perf_counter()

    commands = [
        f'sudo systemctl stop mariadb',
        f"sudo rm -rf '{data_directory}'",
        f"sudo rm -f '{conf_path}'"
    ]
    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            err_list = ('does not exist')
            if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    should_reconfigure = has_standby and (is_primary or (node['cluster_role'] == ClusterRoles.PRIMARY.value and not is_primary) or not primary)
    if should_reconfigure:
        step = 'reconfiguring cluster'
        duration_step_start = time.perf_counter()

        new_primary = None
        error = None
        for i in standbys:
            try:
                standby_data = json.loads(i['node'])
                host_standby = json.loads(i['host'])
                #we search for standby, so doing the opposite of primary search
                is_standby = False
                conn = get_mysql_connection(host, node)
                if isinstance(conn, tuple):
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),9)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),9)
                    return
                
                with conn:
                    with conn.cursor():
                        cur.execute("SHOW SLAVE STATUS")
                        res = cur.fetchone()
                        if res:
                            is_standby = True
                
                if is_standby: 
                    jsn = switchover_node(host_target = host_standby, target = standby_data, distributor = distributor, standbys = standbys)
                    if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                        #return from here, unsuccessful switchover is not good in any case
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return

                    new_primary = standby_data
                    break
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return
        
        if not new_primary:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,error['stdout'],error['stderr'],error['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,error['stdout'],error['stderr'],error['excode'])
            return     

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                if should_reconfigure and new_primary:
                    standby_ids = [i['node']['id'] for i in standbys if i['node']['id'] != new_primary['id']]
                    if len(standby_ids) > 0:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"mariadb_master_id":%s}', cluster_role=%s WHERE id=ANY(%s)""", (new_primary['id'],ClusterRoles.STANDBY.value, standby_ids))
                    
                    cur.execute("""UPDATE "clusterCore_c123_node" SET status='%s', ex_params = ex_params || '{"mariadb_master_id":%s}' || '{"rep_user":""}' || '{"rep_pwd":""}', cluster_role='%s' WHERE id=%s""" % (Status.OK.value, new_primary['id'], ClusterRoles.PRIMARY.value, new_primary['id']))
                    cur.execute("""UPDATE "clusterCore_c123_database" SET node_id=%s WHERE node_id=%s""" % (new_primary['id'], node['id']))
                else:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                                
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def switchover_node(host_target, target, distributor, host_primary = None, primary = None, standbys = None):
    try:
        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        rep_user = target['ex_params'].get('rep_user')
        rep_pwd = target['ex_params'].get('rep_pwd')
        use_ssl = target['ex_params'].get('ssl')
        target_data_dir = target['dir_123Cluster'].strip()
        if target_data_dir.startswith('~'):
            jsn = find_tilda(host_target)
            if jsn['stderr'] != '':
                return jsn

            tilda_path = jsn['stdout']
            target_data_dir = target_data_dir.replace('~', tilda_path)
        
        conf_path = "/etc/my.cnf.d/server.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mariadb.conf.d/50-server.cnf"
        has_standby = isinstance(standbys, list) and len(standbys) > 0
        commands = [
            f"sudo sed -i '/read_only=ON/d' '{conf_path}'",
            f"sudo systemctl restart mariadb",
        ]
        for command in commands:
            jsn = json.loads(remote_command(host_target, command, True), strict=False)

            error_lower = jsn['stderr'].lower()
            if error_lower != '' and 'error' in error_lower:
                return jsn

        time.sleep(10)
        
        conn = get_mysql_connection(host_target, target)
        if isinstance(conn, tuple):
            jsn['stderr'] = str(conn[1])
            return jsn
        
        queries = [
            "STOP SLAVE",
            "RESET SLAVE ALL",
            f"CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED VIA mysql_native_password USING PASSWORD('{rep_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
            f"ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED VIA mysql_native_password USING PASSWORD('{rep_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
            f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
            f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
            "FLUSH PRIVILEGES",
            "SHOW MASTER STATUS"
        ]
        with conn:
            with conn.cursor(dictionary=True) as cur:
                for query in queries:
                    cur.execute(query)
                
                master_replication = cur.fetchone()

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role=%s, ex_params = ex_params || '{"mariadb_master_id":%s}', ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.PRIMARY.value, target['id'], target['id'],))
                conn.commit()
        
        change_master_command = f'''CHANGE MASTER TO MASTER_HOST='{host_target["name"]}', MASTER_USER='{rep_user}', MASTER_PORT={target["port"]}, MASTER_PASSWORD='{rep_pwd}', {f"MASTER_SSL=1, MASTER_SSL_VERIFY_SERVER_CERT=0, MASTER_SSL_CAPATH='{target_data_dir}', MASTER_SSL_CA='{target_data_dir}/ca.pem', MASTER_SSL_CERT='{target_data_dir}/client-cert.pem', MASTER_SSL_KEY='{target_data_dir}/client-key.pem'," if use_ssl else ""} MASTER_LOG_FILE='{master_replication['File']}', MASTER_LOG_POS={master_replication['Position']}'''
        
        if host_primary and primary:
            commands = [
                f"sudo sed -i '/read_only=ON/d' '{conf_path}'",
                "sudo systemctl restart mariadb",
            ]
            for command in commands:
                jsn = json.loads(remote_command(host_primary, command, True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and 'error' in error_lower:
                    return jsn
                
            time.sleep(10)
            
            queries = [
                f"STOP SLAVE",
                f"RESET SLAVE ALL",
                f"DROP USER IF EXISTS '{rep_user}'@'%.%.%.%'",
                "FLUSH PRIVILEGES",
                change_master_command,
                "START SLAVE"
            ]
            conn = get_mysql_connection(host_primary, primary)
            if isinstance(conn, tuple):
                jsn['stderr'] = str(conn[1])
                return jsn
            
            with conn:
                with conn.cursor() as cur:
                    for query in queries:
                        cur.execute(query)

            commands = [
                f"sudo grep -qxF '[mysqld]' '{conf_path}' && grep -qxF 'read_only=ON' '{conf_path}' || sudo sed -i '/\[mysqld\]/a read_only=ON' '{conf_path}'",
                "sudo systemctl restart mariadb",
            ]
            for command in commands:
                jsn = json.loads(remote_command(host_primary, command, True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and 'error' in error_lower:
                    return jsn
            
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity='%s', status='%s', ex_params = ex_params || '{"mariadb_master_id":%s}' || '{"rep_user":"%s"}' || '{"rep_pwd":"%s"}', cluster_role='%s' WHERE id=%s;""" % (Severity.OK.value, Status.OK.value, target['id'], rep_user, rep_pwd, ClusterRoles.STANDBY.value, primary['id']))
                    cur.execute("""UPDATE "clusterCore_c123_database" SET node_id=%s WHERE node_id=%s""" % (primary['id'], target['id']))
                    conn.commit()
        
        if has_standby:
            for i in standbys:
                standby_host = i['host']
                standby_data = i['node']
                rep_user = standby_data["ex_params"].get("rep_user")
                rep_pwd = standby_data["ex_params"].get("rep_pwd")
                change_master_command = f'''CHANGE MASTER TO MASTER_HOST='{host_target["name"]}', MASTER_USER='{rep_user}', MASTER_PORT={target["port"]}, MASTER_PASSWORD='{rep_pwd}', {f"MASTER_SSL=1, MASTER_SSL_VERIFY_SERVER_CERT=0, MASTER_SSL_CAPATH='{target_data_dir}', MASTER_SSL_CA='{target_data_dir}/ca.pem', MASTER_SSL_CERT='{target_data_dir}/client-cert.pem', MASTER_SSL_KEY='{target_data_dir}/client-key.pem'," if use_ssl else ""} MASTER_LOG_FILE='{master_replication['File']}', MASTER_LOG_POS={master_replication['Position']}'''
                #switchover function was wrong, in mariadb unlike the postgresql we made personal replication user for every node
                queries = [
                    f"CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED VIA mysql_native_password USING PASSWORD('{rep_pwd}') {'REQUIRE SSL' if use_ssl else ''}",
                    f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
                    f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
                    "FLUSH PRIVILEGES"
                ]

                conn = get_mysql_connection(host_target, target)
                if isinstance(conn, tuple):
                    jsn['stderr'] = str(conn[1])
                    return jsn
                
                with conn:
                    with conn.cursor() as cur:
                        for query in queries:
                            cur.execute(query)
                
                conn = get_mysql_connection(standby_host, standby_data)
                if isinstance(conn, tuple):
                    jsn['stderr'] = str(conn[1])
                    return jsn
                
                queries = [  
                    "STOP SLAVE",
                    "RESET SLAVE ALL",
                    change_master_command,
                    "START SLAVE"
                ]
                with conn:
                    with conn.cursor() as cur:
                        for query in queries:
                            cur.execute(query)
                
                commands = [
                    f"sudo grep -qxF '[mysqld]' '{conf_path}' && grep -qxF 'read_only=ON' '{conf_path}' || sudo sed -i '/\[mysqld\]/a read_only=ON' '{conf_path}'",
                    'sudo systemctl restart mariadb'
                ]
                for command in commands:
                    jsn = json.loads(remote_command(standby_host, command, True), strict=False)
                    error_lower = jsn['stderr'].lower()
                    if error_lower != '' and 'error' in error_lower:
                        return jsn
                
                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role=%s, ex_params = ex_params || '{"mariadb_master_id":%s}', ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.STANDBY.value, target['id'], standby_data['id'],))
                        conn.commit()
        
        return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

def switchover_mariadb_node_task(data, task_id):
    """switchover mariadb node"""

    task = 'switchover mariadb node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_primary = json.loads(data['host_primary'])
        primary = json.loads(data['primary_data'])
        host_standby = json.loads(data['host_standby'])
        standby = json.loads(data['standby_data'])
        distributor, release = check_os_version(host_standby)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        standbys = data.get('standbys')
        has_standby = isinstance(standbys, list) and len(standbys) > 0
        if has_standby:
            standbys = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in standbys]
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'switching over primary'
    duration_step_start = time.perf_counter()
    
    jsn = switchover_node(host_standby, standby, distributor, host_primary, primary, standbys)
    error_lower = jsn['stderr'].lower()
    if error_lower != '' and 'error' in error_lower:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)