# BELONG TO SERVICE TASK
import os, json, uuid, time, pytz, logging, traceback
from datetime import datetime
from math import inf as positive_infinity
from psycopg2.extras import RealDictCursor
from enums import OperatingSystem, SupportedOS, DBType, ClusterRoles, RestoreOptions, Status, Severity, Inventory, Schedule, LicenseType, Repository
from tasks_core import insert_task, remote_command, format_for_shell, create_ssh_connection, get_postgres_connection, logger, check_os_version, find_tilda,\
      check_remote_directory_exists, transfer_text_ram, transfer_files_rsync, get_date_modified_folder, update_failed_object, \
      install_offline_remote_to_remote, generate_complex_password, get_mysql_connection, TASK_TMP_FILES_DIR, transfer_ssl_certificates, \
      init_container, get_log_rotate_config

def check_mysql_version_exists(host, mysql_version, distributor):
    cmd = f'rpm -qa | grep -i "mysql-wsrep-{mysql_version}"' if OperatingSystem.CENTOS.value in distributor else f'dpkg -l | grep -i "mysql-wsrep-{mysql_version}"'
    jsn = json.loads(remote_command(host,cmd, True), strict=False)
    return jsn['stdout'] != ''

def install_mysql(distributor, release, host, mysql_version) -> dict:
    supported_versions = SupportedOS[distributor.upper()].value
    supported_versions_str = ", ".join(map(str, supported_versions))
    mysql_version = float(mysql_version)
    galera_version = 3 if mysql_version < 8.0 else 4
    xtrabackup_version_mapping = {
        5.7: '24',
        8.0: '80',
        8.4: '84'
    }
    xtrabackup_version = xtrabackup_version_mapping[mysql_version]
    xtrabackup_repo = 'tools' if mysql_version < 8.4 else 'pxb-84-lts'
    
    if OperatingSystem.CENTOS.value in distributor:
        package_manager = 'dnf' if int(release) > 7 else 'yum'
        if int(release) in [7, 9]:
            if int(release) > 7:
                distributor = OperatingSystem.REDHAT.name.lower()
            
            repo_file = f"""[galera]
name = Galera
baseurl = https://releases.galeracluster.com/galera-{galera_version}/{distributor}/{release}/x86_64
gpgkey = https://releases.galeracluster.com/GPG-KEY-galeracluster.com
gpgcheck = 1
[MySQL-wsrep]
name = MySQL-wsrep
baseurl = https://releases.galeracluster.com/mysql-wsrep-{mysql_version}/{distributor}/{release}/x86_64
gpgkey = https://releases.galeracluster.com/GPG-KEY-galeracluster.com
gpgcheck = 1
"""
            commands = [
                "sudo mv '/etc/yum.repos.d/galera.repo /etc/yum.repos.d/galera.repo.bak' 2>/dev/null",
                f"echo '{repo_file}' | sudo tee '/etc/yum.repos.d/galera.repo'",
            ]
        elif int(release) == 8:
            distributor = OperatingSystem.REDHAT.name.lower()
            galera_dir = f'{TASK_TMP_FILES_DIR}/galera{galera_version}'
            wsrep_dir = f'{TASK_TMP_FILES_DIR}/mysql-wsrep-{mysql_version}'
            wsrep_link = f"https://releases.galeracluster.com/mysql-wsrep-{mysql_version}/{distributor}/{release}/x86_64/"
            galera_link = f"https://releases.galeracluster.com/galera-{galera_version}/{distributor}/{release}/x86_64/"
            commands = [
                f"sudo {package_manager} --enablerepo=appstream install -q -y lsof socat stunnel numactl-libs rsync net-tools",
                f"sudo mkdir -p '{galera_dir}'",
                f"sudo mkdir -p '{wsrep_dir}'",
                f'''sudo curl -sL {wsrep_link} | sed -n 's/.*"\([^"]*\)".*/\\1/p' | grep -vE 'debug|devel|test|white' | xargs -I {{}} sudo curl -so {wsrep_dir}/{{}} {wsrep_link}{{}}''',
                f'''sudo curl -sL {galera_link} | sed -n 's/.*"\([^"]*\)".*/\\1/p' | xargs -I {{}} sudo curl -so {galera_dir}/{{}} {galera_link}{{}}''',
                f"sudo {package_manager} --enablerepo=appstream install -q -y {galera_dir}/*.rpm",
                f"sudo {package_manager} --enablerepo=appstream install -q -y {wsrep_dir}/*.rpm",
                f"sudo sed -i '/^KillSignal.*/a ExecStopPost=+systemctl daemon-reload' '/usr/lib/systemd/system/mysqld.service'",
                "sudo systemctl daemon-reload"
            ]
        else:
            jsn = {
                'stdout': 'Unknown release version',
                'stderr': f'Release version of {distributor.capitalize()} is not {supported_versions_str}',
                'excode': 1
            }
            return jsn
        
        commands += [
            f"sudo {package_manager} --enablerepo=appstream install -q -y net-tools {'hostname' if int(release) > 7 else ''} {f'galera-{galera_version} mysql-wsrep-{mysql_version}' if int(release) in [7, 9] else ''}",
            f"sudo {package_manager} install -q -y https://repo.percona.com/yum/percona-release-latest.noarch.rpm",
            f"sudo -iu root percona-release enable-only {xtrabackup_repo} release",
            f"sudo {package_manager} --enablerepo=appstream install -q -y percona-xtrabackup-{xtrabackup_version}"
        ]

    elif OperatingSystem.UBUNTU.value in distributor:
        mysql_install = []
        release_codename_mapping = {
            18: 'bionic',
            20: 'focal',
            22: 'jammy',
        }
        version = release_codename_mapping.get(int(release), 'unknown')

        if version == 'unknown':
            jsn = {
                'stdout': 'Unknown release version',
                'stderr': f'Release version of {distributor.capitalize()} is not {supported_versions_str}',
                'excode': 1
            }
            return jsn
        galera_repo = f"""deb [signed-by=/usr/share/keyrings/galera-keyring.gpg] https://releases.galeracluster.com/galera-4/{distributor} {version} main
deb [signed-by=/usr/share/keyrings/galera-keyring.gpg] https://releases.galeracluster.com/mysql-wsrep-{mysql_version}/{distributor} {version} main
"""
        galera_preferrence ="""Package: *
Pin: origin releases.galeracluster.com
Pin-Priority: 1001
"""
        mysql_install = [
            f"echo '{galera_repo}' | sudo tee /etc/apt/sources.list.d/galera.list",
            f"echo '{galera_preferrence}' | sudo tee -a '/etc/apt/preferences.d/galera.pref'",
            'sudo apt-get update',
            f'sudo DEBIAN_FRONTEND=noninteractive apt-get -o DPkg::Options::="--force-confnew" install -qq -y galera-{galera_version} galera-arbitrator-{galera_version} mysql-wsrep-{mysql_version}'
        ]

        pre_install = [
            "sudo debconf-set-selections <<< 'libssl1.1 libraries/restart-without-asking boolean true'",
            'sudo apt-get update',
            'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y ca-certificates software-properties-common curl gpg dirmngr net-tools',
            "sudo rm -f '/etc/apparmor.d/disable/usr.sbin.mysqld'",
            "sudo ln -s '/etc/apparmor.d/usr.sbin.mysqld' '/etc/apparmor.d/disable/'",
            "sudo mv '/etc/apt/sources.list.d/galera.list' '/etc/apt/sources.list.d/galera.list.bak' 2>/dev/null",
            'sudo dirmngr --list-crls &> /dev/null',
            "sudo touch '/root/.gnupg/dirmngr_ldapservers.conf'",
            "sudo gpg -q --batch --yes --no-default-keyring --keyring '/usr/share/keyrings/galera-keyring.gpg' --keyserver hkp://keyserver.ubuntu.com:80 --recv-keys 8DA84635"
        ]
        
        percona_preferrence ="""Package: *
Pin: release o=Percona Development Team
Pin-Priority: 1001
"""
        
        xtrabackup_install = [
            f"sudo wget -q https://repo.percona.com/apt/percona-release_latest.{version}_all.deb",
            f"sudo DEBIAN_FRONTEND=noninteractive dpkg -i percona-release_latest.{version}_all.deb",
            f"sudo -iu root percona-release enable-only {xtrabackup_repo} release",
            f"echo '{percona_preferrence}' | sudo tee '/etc/apt/preferences.d/00percona.pref'",
            'sudo apt-get update',
            f"sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y percona-xtrabackup-{xtrabackup_version}"
        ]
        commands = pre_install + mysql_install + xtrabackup_install   
        
    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
            if filter:
                return jsn
        except Exception as error:
            return {'stdout': None, 'stderr': str(error), 'excode': 9}
            
    return jsn
    
def find_mysql_bin(host, distributor,  component = "mysqld") -> str:
    """find location of mysql bin folder"""
    
    mysql_bin = ''
    command = f'sudo dirname $(which {component})' if OperatingSystem.UBUNTU.value in distributor else "sudo dirname $(rpm -ql $(rpm -qa | grep mysql | grep server) | grep 'mysqld$' | grep -v 'run')"
    try:
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stdout'] != '':
            mysql_bin = jsn['stdout'].strip()

        if jsn['stderr'] != '' and not ('warning' in jsn['stderr'].lower()) and 'permission denied' not in jsn['stderr'].lower():
            if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
                mysql_bin = '/usr/sbin'
            else:
                mysql_bin = '/usr/sbin'
        
        return mysql_bin
    except Exception as error:
        logging.error(f"find_mysql_bin: {str(error)}")

def get_mysql_config_file(id, port, version, is_standby=False, use_ssl = False):
    """mysql config file template"""
    
    version = float(version)
    #mysql_native_password does not support SSL and deprecated in mysql 8.x
    if use_ssl:
        authentication_plugin = 'sha256_password' if version < 8.0 else 'caching_sha2_password'
    else:
        authentication_plugin = 'mysql_native_password'
    
    #client section should go first, since mysqld section we edit for "read-only standbys"
    #ssl certificates are self-signed, since for trusted ones need buy domain names for machines
    config =f"""
[mysqld]
server_id='{id}'
bind-address=0.0.0.0
port='{port}'
default_storage_engine=InnoDB
innodb_buffer_pool_size=134217728
innodb_buffer_pool_instances=1
datadir='$datadir'
log_bin='$datadir/binlog'
relay-log=$(cat /etc/hostname)-relay-bin
pid-file=/var/run/mysql/mysql_{port}.pid
character_set_server=utf8mb4
collation_server=utf8mb4_general_ci
explicit_defaults_for_timestamp=ON
replicate_ignore_db='mysql,sys'
general_log = 1
general_log_file =$datadir/log/mysqld.log
performance_schema=ON"""
    if use_ssl:
        config = f"""[client]
ssl-mode=REQUIRED
ssl-cert='$datadir/client-cert.pem'
ssl-key='$datadir/client-key.pem'
{config}
auto_generate_certs=ON
require_secure_transport=ON"""
    
    #if that is standby - use ssl certificate files of primary
    if is_standby:
        config += """
super_read_only=ON"""
        if use_ssl:
            config += """
ssl_ca=$datadir/ca.pem
ssl_cert=$datadir/server-cert.pem
ssl_key=$datadir/server-key.pem"""

    #options deprecated/removed in mysql 8.x
    if version < 8.0:
        config += f"""
expire_logs_days=2
binlog_format=ROW
default_authentication_plugin={authentication_plugin}"""
    
    if version >= 8.0:
        config += f"""
binlog_expire_logs_seconds=172800
authentication_policy='{authentication_plugin}'"""
    
    # we want the ability to run mysql without ssl also, so enable_mysql_native_password plugin
    if version > 8.0:
        config+= f"""
mysql_native_password=ON
wsrep_sync_server_uuid=OFF
"""
        
    return config

def create_mysql_node(host, node, distributor, release) -> dict:
    try:
        appendix = "d" if OperatingSystem.CENTOS.value in distributor else ""

        data_dir = node['dir_123Cluster']
        mysql_version = float(node['version'])
        use_ssl = node['ex_params'].get('ssl')
        if use_ssl:
            authentication_plugin = 'sha256_password' if mysql_version < 8.0 else 'caching_sha2_password'
        else:
            authentication_plugin = 'mysql_native_password'

        if data_dir.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                return jsn

            tilda_path = jsn['stdout']
            data_dir = data_dir.replace('~', tilda_path)

        mysql_bin = find_mysql_bin(host, distributor)
        new_root_password = node['ex_params'].get('mysql_root_password')
        init_file = f"{data_dir}/mysql-init"
        change_password_root_command = f"""root_pwd='{new_root_password}';echo "ALTER USER 'root'@'localhost' IDENTIFIED BY '$root_pwd';" | sudo tee '{init_file}'"""

        database_user = node['database_user']
        database_pwd = node['database_pwd']

        conf_path = "/etc/my.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mysql.conf.d/mysqld.cnf"

        init_mysql_command = f"sudo {mysql_bin}/mysqld --defaults-file='{conf_path}' --initialize-insecure --user=mysql --datadir='{data_dir}' {'' if mysql_version < 8.0 else '--upgrade=FORCE'}"

        config = get_mysql_config_file(node['id'], node['port'], mysql_version, use_ssl=use_ssl)

        key_paths = """
ssl_ca=$datadir/ca.pem
ssl_cert=$datadir/server-cert.pem
ssl_key=$datadir/server-key.pem"""

        # should be added after mysql auto-generate certificates since replication docs mention it, but having it initially - prevents certificates to generate
        add_keys = f"""datadir='{data_dir}';echo "{key_paths}" | sudo tee -a '{conf_path}'"""
        pid_file = f"/var/run/mysql/mysql_{node['port']}.pid"
        commands = [
            f"sudo systemctl stop mysql{appendix}",
            f"sudo mkdir -p '{data_dir}'",
            f"sudo mkdir -p '{data_dir}/log'",
            f"sudo chown -R mysql:mysql '{data_dir}'",
            f"sudo mkdir -p '/var/run/mysql/'",
            f"[ -f '{conf_path}' ] && sudo cp '{conf_path}' '{conf_path}.bak'",
            f"""datadir='{data_dir}';echo "{config}"| sudo tee '{conf_path}'""",
            init_mysql_command,
            change_password_root_command,
            "sudo chown -R mysql:mysql '/var/run/mysql/'",
            f"sudo chown mysql:mysql '{init_file}'",
            # replacement for secure "mysqld --initialize" : provide init-file with password to mysqld daemon
            f"sudo {mysql_bin}/mysqld --defaults-file='{conf_path}' --init-file='{init_file}' --user=mysql --datadir='{data_dir}' --pid-file='{pid_file}' &> /dev/null &",
            # stop daemon in whatever safe way (should not break on that command in any case, and kill -9 should not be used!)
            f"sudo kill $(sudo cat '{pid_file}') &> /dev/null || sudo mysqladmin --password='{new_root_password}' shutdown  &> /dev/null || sudo kill $(sudo netstat -tunlp | grep '{node['port']}' | awk '{{print $7}}' | awk -F'/' '{{print $1}}') &> /dev/null",
            f"sudo rm -rf '{init_file}'",
            add_keys,
            f'sudo systemctl enable mysql{appendix}',
            # now mysql starts with our root password
            f'sudo systemctl start mysql{appendix}',
            f'''sudo mysql -Nsfu root --password='{new_root_password}' -e "SET GLOBAL connect_timeout=6000;"''',
             # backported against authentication issues in older versions
            f'''sudo mysql -Nsfu root --password='{new_root_password}' -e "DELETE FROM mysql.user WHERE user = '';"''',
            f'''database_pwd='{database_pwd}';sudo mysql -Nsfu root --password='{new_root_password}' -e "CREATE USER '{database_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '$database_pwd' {'REQUIRE SSL' if use_ssl else ''};"''',
            f'''sudo mysql -Nsfu root --password='{new_root_password}' -e "GRANT ALL PRIVILEGES ON *.* TO '{database_user}'@'%.%.%.%' WITH GRANT OPTION;"''',
            f'''sudo mysql -Nsfu root --password='{new_root_password}' -e "FLUSH PRIVILEGES;"''',
        ]
        if not use_ssl:
            commands.remove(add_keys)

        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            error_lower = jsn['stderr'].lower()
            if error_lower != '' and any(err in error_lower for err in ["error", "fail", "errno"]):
                return jsn
            
            if 'systemctl start' in command or ('mysqld' in command and '--pid-file' in command):
                #ssh may crash due to network issues - python better (TODO - reduce the intervals for production)
                time.sleep(5)

        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            jsn['stderr'] = str(conn[1])
            return jsn

        with conn:
            with conn.cursor() as cur:
                queries = [
                    f"""CREATE USER '{database_user}'@'localhost' IDENTIFIED WITH {authentication_plugin} BY '{database_pwd}' {'REQUIRE SSL' if use_ssl else ''}""",
                    f"""CREATE USER '{database_user}'@'%' IDENTIFIED WITH {authentication_plugin} BY '{database_pwd}' {'REQUIRE SSL' if use_ssl else ''}""",
                    f"GRANT ALL PRIVILEGES ON *.* TO '{database_user}'@'localhost' WITH GRANT OPTION",
                    f"GRANT ALL PRIVILEGES ON *.* TO '{database_user}'@'%' WITH GRANT OPTION",
                    "FLUSH PRIVILEGES",
                ]
                for query in queries:
                    cur.execute(query)

        return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': f"create_mysql_node error: {str(error)}", 'excode': 9}

def install_mysql_standalone_task(data, task_id):

    task = 'install mysql standalone task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        app_host = json.loads(data['app_host'])
        # app_hostname = data['app_hostname']
        create_run_linux = data['create_run_linux']
        # repository = network or local
        repository = data['repository']
        offline = True if Repository.LOCAL.value in repository else False
        mysql_version = float(node['version'])
        data_dir = node['dir_123Cluster'].strip()
        org_id = node['org_id']
        host_exists = True
        proxmox_container = node['ex_params'].get('proxmox')
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = node['ex_params'].get('distributor')
                release = node['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
        
        version_exists = check_mysql_version_exists(host, mysql_version, distributor)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    

    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return


    step = 'running installation commands'
    duration_step_start = time.perf_counter()

    if not version_exists:
        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host, mysql_version, DBType.MYSQL.value, app_host)
        else:
            jsn = install_mysql(distributor, release, host, mysql_version)
        
        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node['id'])
            return

    jsn = create_mysql_node(host, node, distributor, release)
    error_lower = jsn['stderr'].lower()
    if error_lower != '' and any(err in error_lower for err in ["error", "fail", "errno"]):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(node['id'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{data_dir}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        remote_command(host, f"""echo "{get_log_rotate_config(data_dir)}" | sudo tee '/etc/logrotate.d/{node['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host), 'node_name': node['name']}, 'task': 'log_rotate_task', 'monitor_name': None,  'org_id': node['org_id'],  'target_type': node['type'], 'target_id': node['id'], 'created_by': created_by}
        insert_task(params)
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id']))
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,node['id'],created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host['id'],created_by,OperatingSystem.LINUX.value])
                conn.commit()

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
    
    # host['create_task'] = created_by
    # host['ssh_port'] = int(host['ssh_port'])
    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host, 'task': 'run_linux_performance_counters', 'target_type': 'host', 'target_id': host['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as error:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def add_mysql_standby_task(data, task_id):

    task = 'add mysql standby task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_primary = json.loads(data['host_primary'])
        primary = json.loads(data['primary_data'])
        mysql_version = float(primary['version'])
        # No value threated as false, and should be always taken value of primary
        use_ssl = primary["ex_params"].get("ssl")
        host_standby = json.loads(data['host_standby'])
        standby = json.loads(data['standby_data'])
        app_host = json.loads(data['app_host'])
        proxmox_container = standby['ex_params'].get('proxmox')
        host_exists = True
        distributor, release = check_os_version(host_standby)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = standby['ex_params'].get('distributor')
                release = standby['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(standby['id'])
                return
        
        version_exists = check_mysql_version_exists(host_standby, mysql_version, distributor)

        # create_run_linux = data['create_run_linux']
        # repository = network or local
        repository = data['repository']
        offline = True if Repository.LOCAL.value in repository else False
        rep_user = standby['ex_params'].get('rep_user')
        rep_pwd = standby['ex_params'].get('rep_pwd')
        
        if use_ssl:
            authentication_plugin = 'sha256_password' if mysql_version < 8.0 else 'caching_sha2_password'
        else:
            authentication_plugin = 'mysql_native_password'
        
        appendix = "d" if OperatingSystem.CENTOS.value in distributor else ""
        replica = "REPLICA" if mysql_version >= 8.0 else "SLAVE"
        master = "SOURCE" if mysql_version >= 8.0 else "MASTER"
        conf_path = "/etc/my.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mysql.conf.d/mysqld.cnf"
        primary_clone = f"{TASK_TMP_FILES_DIR}/mysql_clones/{primary['name'].replace(':', '_')}"
        master_alias = "MASTER" if master == "MASTER" else "REPLICATION SOURCE"
        # since standby are clone of primary - only primary root password are needed
        primary_root_password = primary['ex_params'].get('mysql_root_password')
        standby_data_dir = standby['dir_123Cluster'].strip()
        if standby_data_dir.startswith('~'):
            jsn = find_tilda(host_standby)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            standby_data_dir = standby_data_dir.replace('~', tilda_path)
            
        primary_data_dir = primary['dir_123Cluster'].strip()
        if primary_data_dir.startswith('~'):
            jsn = find_tilda(host_primary)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            primary_data_dir = primary_data_dir.replace('~', tilda_path)
        
        org_id = standby["org_id"]
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host_standby, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(standby['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        
        
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    step = 'running installation commands'
    duration_step_start = time.perf_counter()

    if not version_exists:
        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host_standby, mysql_version, DBType.MYSQL.value, app_host)
        else:
            jsn = install_mysql(distributor, release, host_standby, mysql_version)

        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(standby['id'])
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'configuring the primary'
    duration_step_start = time.perf_counter()

    try:
        now = datetime.utcnow().strftime('%d%m%Y%H%M%S')
        backup_name = f"{primary['name'].replace(':', '_')}_{now}"
        transfer_path = f"{primary_clone}/{backup_name}.gz"
        backup_cmd = f"sudo xtrabackup --defaults-file='{conf_path}' --backup --user=root --port='{primary['port']}' --password='{primary_root_password}' --target-dir='{primary_clone}' --stream=xbstream | sudo gzip -c | sudo tee '{transfer_path}'"
        queries = [
            f"""CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' {'REQUIRE SSL' if use_ssl else ''}""",
            #each "create user if not exists" query should be confirmed by "alter user", against the cases of password misconfiguration
            f"ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' {'REQUIRE SSL' if use_ssl else ''}",
            #all privileges should include replication slave, no?
            f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
            #but for insurance - better go explicit also
            f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
            "FLUSH PRIVILEGES",
            #now think that position before the start of backup are preferred over the position after backup restore
            f"SHOW {'BINARY LOG' if mysql_version > 8.0 else 'MASTER'} STATUS"
        ]
        
        conn = get_mysql_connection(host_primary, primary)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor(dictionary=True) as cur:
                for query in queries:
                    cur.execute(query)
                #its ok to do like this, with...as blocks variables scope are function-level
                master_replication = cur.fetchone()
        
        commands = [
            f"sudo mkdir -p '{primary_clone}'",
            backup_cmd
        ]
        for command in commands:
            jsn = json.loads(remote_command(host_primary, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(standby['id'])
                # if fails here - backup of primary is corrupted
                remote_command(host_primary, f"sudo rm -rf '{primary_clone}'", True)
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'running installation commands on standby'
    duration_step_start = time.perf_counter()

    try:
        # backup preparation stage
        jsn = transfer_files_rsync(host_primary, host_standby, primary_clone, [transfer_path])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(standby['id'])
            return
        
        config = get_mysql_config_file(standby['id'], standby['port'], mysql_version, is_standby = True, use_ssl = use_ssl)
        
        commands = [
            f"sudo gunzip < '{transfer_path}' | sudo xbstream -C '{primary_clone}' -x",
            f"sudo chown -R mysql:mysql '{primary_clone}'",
            f"[ -f '{conf_path}' ] && sudo cp '{conf_path}' '{conf_path}.bak'",
            f"""datadir='{standby_data_dir}';echo "{config}" | sudo tee '{conf_path}'""",
            f"sudo xtrabackup --defaults-file='{conf_path}' --prepare --target-dir='{primary_clone}'",
        ]
        
        for command in commands:
            jsn = json.loads(remote_command(host_standby, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(standby['id'])
                # if fails here - backup copy on standby host is corrupted
                remote_command(host_standby, f"sudo rm -rf '{primary_clone}'", True)
                return

        # primary cloning stage
        commands = [
            f"sudo systemctl stop mysql{appendix}",
            # when cloning from backup - always wipe data directory, to avoid potential conflicts
            f"sudo rm -rf '{standby_data_dir}'",
            f"sudo mkdir -p '{standby_data_dir}'",
            f"sudo xtrabackup --defaults-file='{conf_path}' --copy-back --target-dir='{primary_clone}'"
        ]
        for command in commands:
            jsn = json.loads(remote_command(host_standby, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        if use_ssl:
            jsn = transfer_ssl_certificates(primary, host_primary, host_standby, standby_data_dir)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        # startup stage
        commands = [
            f"sudo chown -R mysql:mysql '{standby_data_dir}'",
            f"sudo mkdir -p '/var/run/mysql'",
            "sudo chown -R mysql:mysql '/var/run/mysql'",
            f'sudo systemctl enable mysql{appendix}',
            'sudo rm -rf /var/lib/mysql/auto.cnf',
            f'sudo rm -rf {standby_data_dir}/auto.cnf',
            f'sudo systemctl start mysql{appendix}',
        ]
        for command in commands:
            jsn = json.loads(remote_command(host_standby, command, True), strict=False)
            if jsn['stderr'] != '' and 'created symlink' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        # cleanup stage
        remote_command(host_primary, f"sudo rm -rf '{primary_clone}'", True)
        remote_command(host_standby, f"sudo rm -rf '{primary_clone}'", True)
          
        # replication stage
        queries = [
            f"STOP {replica}",
            f"RESET {replica} ALL",
            # since SSL certificates are autogenerated by mysql, filenames are constant - but paths must be absolute and we should use client certificate and key (replication are client connection)
            f'''CHANGE {master_alias} TO {master}_HOST='{host_primary['name']}', {master}_PORT={primary['port']}, {master}_USER='{rep_user}', {master}_PASSWORD='{rep_pwd}', {f"{master}_SSL=1, {master}_SSL_VERIFY_SERVER_CERT=0, {master}_SSL_CAPATH='{standby_data_dir}', {master}_SSL_CA='{standby_data_dir}/ca.pem', {master}_SSL_CERT='{standby_data_dir}/client-cert.pem', {master}_SSL_KEY='{standby_data_dir}/client-key.pem'," if use_ssl else ""} {master}_LOG_FILE='{master_replication["File"]}', {master}_LOG_POS={master_replication["Position"]}''',
            # credentials in both commands since docs still use it in CHANGE REPLICATION SOURCE for insurance, 
            # and START REPLICA with credentials complements ssl setup (explicit specify authentication plugin for replication connections)
            f"START {replica} USER = '{rep_user}' PASSWORD = '{rep_pwd}' DEFAULT_AUTH='{authentication_plugin}'"
        ]
    
        conn = get_mysql_connection(host_standby, standby)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor() as cur:
                for query in queries:
                    cur.execute(query)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)


    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host_standby, f"sudo du -sb '{standby_data_dir}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = round(float(jsn['stdout']) / 1024, 2)
        remote_command(host_standby, f"""echo "{get_log_rotate_config(standby_data_dir)}" | sudo tee '/etc/logrotate.d/{standby['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host_standby), 'node_name': standby['name']}, 'task': 'log_rotate_task', 'monitor_name': None,  'org_id': standby['org_id'], 'target_type': standby['type'], 'target_id': standby['id'], 'created_by': created_by}
        insert_task(params)
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, standby['id']))
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,standby['id'],created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host_standby['id'],created_by,OperatingSystem.LINUX.value])
                conn.commit()
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
    
    host_standby['create_task'] = created_by
    host_standby['ssh_port'] = int(host_standby['ssh_port'])

    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host_standby, 'task': 'run_linux_performance_counters', 'target_type': 'host', 'target_id': host_standby['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as error:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),None)
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def mysqlbinlog_dump_task(data, task_id):
    """a task to use backup with "mysqlbinlog" command to dump node binary logs as sql"""
    """just for node schedule backup"""

    task = 'mysqlbinlog dump task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        primary = json.loads(data['primary_data'])
        # host of backed up node
        host = data['host_data']
        dump = json.loads(data['backup_data'])
        binlog_time = dump['ex_params'].get('binlog_time')
        backup_dir = dump['ex_params'].get('backup_path')
        if backup_dir.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_dir = backup_dir.replace('~', tilda_path)

        # oplog_path = dump['ex_params'].get('oplog_path')
        timezone_cmd = 'cat /etc/timezone'
        jsn = json.loads(remote_command(host, timezone_cmd, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    timezone = jsn['stdout'].strip()
    try:
        timestamp = datetime.now().astimezone(pytz.timezone(timezone))
        path_date = timestamp.strftime("%Y_%m_%d_%H_%M_%S")
        binlog_time_now = timestamp.strftime("%Y-%m-%d %H:%M:%S")
        path = f'{backup_dir}/dumps_binlog/{path_date}'
    except:
        error = f'Invalid server timezone: {timezone}'
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return
    
    client = create_ssh_connection(host['name'], host['username'], host['password'], host['private_key'], int(host['ssh_port']))
    if isinstance(client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to host {host['name']}",str(client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to host {host['name']}",str(client[1]),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'running dump command'
    duration_step_start = time.perf_counter()

    start_datetime = f"--start-datetime='{binlog_time}'" if binlog_time else ''
    command = f"sudo mysqlbinlog $(find '{primary['dir_123Cluster']}' -type f -name 'binlog.[0-9]*' -printf '%p ') {start_datetime} --stop-datetime='{binlog_time_now}' > '{path}/{path_date}.sql'"
    
    try:
        binlog_time = timestamp.strftime("%Y-%m-%d %H:%M:%S")
        jsn = json.loads(remote_command(host, f"sudo mkdir -p '{path}'", True), strict = False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        stdin, stdout, stderr = client.exec_command(command)
        exit_status = stdout.channel.recv_exit_status() # Blocking call
        stdin.channel.shutdown_write()
        try:
            stderr = stderr.read().decode('ascii').strip()
            stdout = stdout.read().decode('ascii').strip()
        except UnicodeDecodeError:
            stderr = stderr.read().decode('utf-8').strip()
            stdout = stdout.read().decode('utf-8').strip()
   
        if stderr != '' and 'error' in stderr.lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,stderr,exit_status)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,stderr,exit_status)
            return
        
        client.close()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{binlog_path}}', '"{0}"', true) WHERE id={1}""".format(path,dump['id']))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{binlog_time}}', '"{0}"', true) WHERE id={1}""".format(binlog_time,dump['id']))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def drop_mysql_node_task(data, task_id):
    """drop mysql node"""

    task = 'drop mysql node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        appendix = "d" if OperatingSystem.CENTOS.value in distributor else ""
        host_primary = data.get('host_primary')
        primary = data.get('primary_data')
        node = json.loads(data['node_data'])
        mysql_version = float(node['version'])
        final_backup = data.get('final_backup')
        backup_host = data['backup_host']
        rep_user = node["ex_params"].get("rep_user")
        standbys = data.get("standbys")
        has_standby = isinstance(standbys, list) and len(standbys) > 0
        if has_standby:
            standbys = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in standbys]

        data_directory = node['dir_123Cluster']
        if data_directory.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_directory = data_directory.replace('~', tilda_path)

        conf_path = "/etc/my.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mysql.conf.d/mysqld.cnf"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'checking node existence'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node['port']}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        if jsn['stdout'] == '':
            jsn = json.loads(remote_command(host, f"sudo rm -rf '{data_directory}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'{node["name"]} is already deleted', 1)

            try:
                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                        cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                        conn.commit()
            except Exception as error:
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

            step = 'task finished'
            status = Severity.WARNING.value
            duration = 0
            logger(task_id,task_log_id,task,step,status,duration,created_by)
            step = 'null'
            status = Severity.WARNING.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
            return
        
        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,'',str(conn[1]),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,'',str(conn[1]),9)
            return
       
        is_primary = False
        with conn:
            #prefer dictionary, as more pythonic and readable
            with conn.cursor() as cur:
                cur.execute(f"SHOW {'SLAVE' if mysql_version < 8.0 else 'REPLICA'} STATUS")
                res = cur.fetchone()
                if not res:
                    is_primary = True

        if is_primary and has_standby:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'{node["name"]} is primary with standbys.Switchover first.',1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,f'{node["name"]} is primary with standbys.Switchover first.',1)
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    if node['cluster_role'] != ClusterRoles.PRIMARY.value:
        final_backup = False
    if final_backup:
        try:
            step = 'backing up the node'
            duration_step_start = time.perf_counter()
            
            backup_path =  final_backup['ex_params'].get('backup_path')
            commands = [
                f"sudo rm -rf '{backup_path}'",
                f"sudo mkdir -p '{backup_path}'",
            ]
            for cmd in commands:
                jsn = json.loads(remote_command(host, cmd, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            
            cmd = f"sudo xtrabackup --defaults-file='{conf_path}' --backup --user=root --port='{node['port']}' --password='{node['ex_params'].get('mysql_root_password')}' --target-dir='{backup_path}' --stream=xbstream | sudo gzip -c | sudo tee '{backup_path}/{final_backup['name']}.gz'"
            jsn = json.loads(remote_command(host, cmd, True), strict=False)

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            jsn = transfer_files_rsync(host, backup_host, backup_path, [f"{backup_path}/"])
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            jsn = json.loads(remote_command(backup_host, f"""sudo du -b '{backup_path}{f'/{final_backup["name"]}.gz'}' | awk '{{print $1}}'""", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            size = jsn['stdout'].strip()
            backup_size = float(size)/1024

            modified_date = get_date_modified_folder(hostdata=backup_host, folder_path=backup_path)
            if isinstance(modified_date, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
                return
            
            remote_command(host, f"sudo rm -rf '{backup_path}'", True)

            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size, final_backup['id']))
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}' WHERE id=%s""" % (modified_date, final_backup['id']))
                    conn.commit()
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
        
    if host_primary and primary and rep_user and not is_primary:
        step = 'drop replication user (if exists)'
        duration_step_start = time.perf_counter()

        try:
            primary = json.loads(primary)
            conn = get_mysql_connection(host_primary, primary)
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),9)
                return
            
            queries = [
                f"DROP USER IF EXISTS '{rep_user}'@'%.%.%.%'",
                "FLUSH PRIVILEGES"
            ]
            with conn:
                with conn.cursor() as cur:
                    for query in queries:
                        cur.execute(query)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'removing node'
    duration_step_start = time.perf_counter()

    commands = [
        f'sudo systemctl stop mysql{appendix}',
        f"sudo rm -rf '{data_directory}'",
        f"sudo rm -f '{conf_path}'"
    ]
    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            err_list = ('does not exist', 'not loaded')
            if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    should_reconfigure = has_standby and (is_primary or (node['cluster_role'] == ClusterRoles.PRIMARY.value and not is_primary) or not primary)
    if should_reconfigure:
        step = 'reconfiguring cluster'
        duration_step_start = time.perf_counter()

        new_primary = None
        error = None
        for i in standbys:
            try:
                standby_data = json.loads(i['node'])
                host_standby = json.loads(i['host'])
                #we search for standby, so doing the opposite of primary search
                is_standby = False
                conn = get_mysql_connection(host, node)
                if isinstance(conn, tuple):
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),9)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),9)
                    return
                
                with conn:
                    with conn.cursor():
                        cur.execute(f"SHOW {'SLAVE' if mysql_version < 8.0 else 'REPLICA'} STATUS")
                        res = cur.fetchone()
                        if res:
                            is_standby = True
                
                if is_standby: 
                    jsn = switchover_node(host_target = host_standby, target = standby_data, distributor = distributor, standbys = standbys)
                    if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                        #return from here, unsuccessful switchover is not good in any case
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return

                    new_primary = standby_data
                    break
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return
        
        if not new_primary:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,error['stdout'],error['stderr'],error['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,error['stdout'],error['stderr'],error['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                
                if should_reconfigure and new_primary:
                    standby_ids = [i['node']['id'] for i in standbys if i['node']['id'] != new_primary['id']]
                    if len(standby_ids) > 0:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"mysql_master_id":%s}', cluster_role=%s WHERE id=ANY(%s)""", (new_primary['id'], ClusterRoles.STANDBY.value, standby_ids))
                    
                    cur.execute("""UPDATE "clusterCore_c123_node" SET status = '%s', ex_params = ex_params || '{"mysql_master_id":%s}' || '{"rep_user":""}' || '{"rep_pwd":""}', cluster_role='%s' WHERE id=%s""" % (Status.OK.value, new_primary['id'], ClusterRoles.PRIMARY.value, new_primary['id']))
                    cur.execute("""UPDATE "clusterCore_c123_database" SET node_id = %s WHERE node_id=%s""" % (new_primary['id'], node['id']))
                else:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def create_mysql_database_task(data, task_id):
    
    task = 'create mysql database task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    
    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data']) # not standby
        use_ssl = node['ex_params'].get('')
        db = json.loads(data['db']) 
        db_name = db['name'].strip()
        db_id = db['id']
        if use_ssl:
            authentication_plugin = 'sha256_password' if float(node['version']) < 8.0 else 'caching_sha2_password'
        else:
            authentication_plugin = 'mysql_native_password'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating database'
    duration_step_start = time.perf_counter()

    try:
        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        db_size = 0.0
        with conn:
            with conn.cursor() as cur:
                cur.execute(f"CREATE DATABASE IF NOT EXISTS {db_name} CHARACTER SET 'utf8' COLLATE 'utf8_general_ci'")
                # closing step = 'creating database'
                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

                step = 'creating database users'
                duration_step_start = time.perf_counter()
                
                owner_user_name = f'{db_name}_owner_user'
                write_user_name = f'{db_name}_write_user'
                read_user_name = f'{db_name}_read_user'
                owner_user_pwd = generate_complex_password(20, DBType.MYSQL.value)
                write_user_pwd = generate_complex_password(20, DBType.MYSQL.value)
                read_user_pwd = generate_complex_password(20, DBType.MYSQL.value)
                logs_output = f'USERNAME = {owner_user_name}\nPASSWORD = {owner_user_pwd}\nUSERNAME = {read_user_name}\nPASSWORD = {read_user_pwd}\nUSERNAME = {write_user_name}\nPASSWORD = {write_user_pwd}'
                owner_user_pwd = format_for_shell(owner_user_pwd)
                write_user_pwd = format_for_shell(write_user_pwd)
                read_user_pwd  = format_for_shell(read_user_pwd)
                
                queries = [
                    f"""CREATE USER IF NOT EXISTS '{owner_user_name}'@'localhost' IDENTIFIED WITH {authentication_plugin} BY '{owner_user_pwd}' {'REQUIRE SSL' if use_ssl else ''}""",
                    f"""ALTER USER '{owner_user_name}'@'localhost' IDENTIFIED WITH {authentication_plugin} BY '{owner_user_pwd}' {'REQUIRE SSL' if use_ssl else ''}""",
                    f"GRANT ALL PRIVILEGES ON {db_name}.* TO '{owner_user_name}'@'localhost' WITH GRANT OPTION",
                    f"""CREATE USER IF NOT EXISTS '{write_user_name}'@'localhost' IDENTIFIED WITH {authentication_plugin} BY '{write_user_pwd}' {'REQUIRE SSL' if use_ssl else ''}""",
                    f"""ALTER USER '{write_user_name}'@'localhost' IDENTIFIED WITH {authentication_plugin} BY '{write_user_pwd}' {'REQUIRE SSL' if use_ssl else ''}""",
                    f"GRANT SELECT, INSERT, UPDATE ON {db_name}.* TO '{write_user_name}'@'localhost'",
                    f"""CREATE USER IF NOT EXISTS '{read_user_name}'@'localhost' IDENTIFIED WITH {authentication_plugin} BY '{read_user_pwd}' {'REQUIRE SSL' if use_ssl else ''}""",
                    f"""ALTER USER '{read_user_name}'@'localhost' IDENTIFIED WITH {authentication_plugin} BY '{read_user_pwd}' {'REQUIRE SSL' if use_ssl else ''}""",
                    f"GRANT SELECT ON {db_name}.* TO  '{read_user_name}'@'localhost'",
                    "FLUSH PRIVILEGES",
                    #if database are just created - size always be none, but still check, for the case of manual task rerun
                    f"SELECT SUM(data_length + index_length)/1024 FROM information_schema.TABLES WHERE table_schema='{db_name}' GROUP BY table_schema",
                ]

                for query in queries:
                    cur.execute(query)
                
                res = cur.fetchone()
                if res and res[0]:
                    db_size = res[0]

                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = round(float(jsn['stdout']) / 1024, 2)

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_size,db_id))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def drop_mysql_database_task(data, task_id):

    task = 'drop mysql database task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        db = json.loads(data['db_data'])
        db_name = db['name'].strip()
        db_id = db['id']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    #this can be done in single connection - so all the deletion steps can be merged to single one
    step = 'dropping the database'
    duration_step_start = time.perf_counter()

    try:
        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor() as cur:
                # kill all the connections to db <db_name>
                cur.execute("SELECT id FROM information_schema.processlist WHERE db = %s", (db_name,))
                connection_ids = [i[0] for i in cur.fetchall()]
                #although some blazing fast client could reconnect during loop - thats the only way...
                for connection in connection_ids:
                    cur.execute(f"KILL {connection}")
                # drop db <db_name> with the associated dba users (order by importance ascending)
                queries = [
                    f"DROP USER IF EXISTS '{db_name}_read_user'@'localhost'",
                    f"DROP USER IF EXISTS '{db_name}_write_user'@'localhost'",
                    f"DROP USER IF EXISTS '{db_name}_owner_user'@'localhost'",
                    "FLUSH PRIVILEGES",
                    f"DROP DATABASE IF EXISTS {db_name}"
                ]
                for query in queries:
                    cur.execute(query)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = round(float(jsn['stdout']) / 1024, 2)
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DROPPED.value, db_id,))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def mysqldump_task(data, task_id):
    """mysqldump of database/all databases"""

    task = 'mysqldump task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        backup_host = data['backup_host']
        node = json.loads(data['node_data'])
        type = node["type"]
        root_pwd = node["ex_params"].get("mysql_root_password") if DBType.MYSQL.name in type else node["ex_params"].get("mariadb_root_password")
        backup = json.loads(data['backup_data'])
        backup_type = data['backup_type']
        dumpall = 'dumpall' in backup_type
        prefix = 'mariadb-' if DBType.MARIADB.name in type and float(node['version']) >= 11.0 else 'mysql'
        if not dumpall:
            db_data = json.loads(data['db_data'])

        compress = data['compress']
        include_data = data['include_data']

        backup_path = backup['ex_params'].get('backup_path')
        if backup_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'updating modified date folder'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo mkdir -p '{backup_path}'", True), strict = False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'running mysqldump command'
    duration_step_start = time.perf_counter()

    data = '--routines --events ' if include_data else '--no-data'
    source = '--all-databases' if dumpall else f"'{db_data['name']}'"
    file = f'{backup_path}/{backup["name"]}.{"gz" if compress else "dump"}'

    try:
        command = f"sudo {prefix}dump --default-character-set=utf8mb4 --port='{node['port']}' --password='{root_pwd}' {data} --quick {source} --result-file='{file}' && sudo chown mysql:mysql '{file}'" if not compress else f"sudo {prefix}dump --default-character-set=utf8mb4 --port='{node['port']}' --password='{root_pwd}' {data} --quick {source} | sudo gzip -c | sudo tee '{file}'"
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if backup_host['name'] != host['name']:
        step = 'copying dump file to backup host'
        duration_step_start = time.perf_counter()
        
        jsn = transfer_files_rsync(host, backup_host, backup_path, [file])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        step = 'checking the backup size'
        duration_step_start = time.perf_counter()

        try:
            jsn = json.loads(remote_command(backup_host, f"sudo du -b '{file}' | awk '{{print $1}}'", True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            backup_size = float(jsn['stdout'])/1024
            
            modified_date = get_date_modified_folder(hostdata=backup_host, folder_path=backup_path)
            if isinstance(modified_date, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
                return
            
            remote_command(host, f"sudo rm -f '{file}'", True)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET status='true', ex_params = ex_params || '{"modified_date":"%s"}'  || '{"size":%s}'  WHERE id=%s""" % (modified_date, backup_size,  backup['id']))
                if not dumpall:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_data['id'],))
                else:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end, created_by)

def mysqldump_restore_task(data, task_id):
    """restore mysqldump of database/all databases"""

    task = 'restore mysqldump task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node = json.loads(data['node_data'])
        standbys = data.get('standbys')
        backup = json.loads(data['backup_data'])
        backup_path = backup['ex_params'].get('backup_path')
        backup_type = data['backup_type']
        backup_host = data['backup_host']
        mysql_version = float(node['version'])
        use_ssl = node['ex_params'].get('ssl')
        if use_ssl:
            authentication_plugin = 'sha256_password' if mysql_version < 8.0 else 'caching_sha2_password'
        else:
            authentication_plugin = 'mysql_native_password'

        node_datadir = node['dir_123Cluster'].strip()
        if node_datadir.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            node_datadir = node_datadir.replace('~', tilda_path)
        
        type = node['type']
         
        if DBType.MYSQL.name in type:
            conf_path = "/etc/my.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mysql.conf.d/mysqld.cnf"
        else:
            conf_path = "/etc/my.cnf.d/server.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mariadb.conf.d/50-server.cnf"
        
        appendix = 'd' if DBType.MYSQL.name in type and OperatingSystem.CENTOS.value in distributor else ''
         
        if DBType.MYSQL.name in type:
            open_read_write_commands = [
                f"sudo sed -i '/super_read_only=ON/d' '{conf_path}'",
                f"sudo systemctl restart mysql{appendix}",
            ]

            open_read_only_commands = [
                f"echo 'super_read_only=ON' | sudo tee -a '{conf_path}'",
                f'sudo systemctl restart mysql{appendix}'
            ]
        else:
            open_read_write_commands = [
                f"sudo sed -i '/read_only=ON/d' '{conf_path}'",
                "sudo systemctl restart mariadb",
            ]

            open_read_only_commands = [
                f"echo 'read_only=ON' | sudo tee -a '{conf_path}'",
                f'sudo systemctl restart mariadb'
            ]
        
        cli = 'mysql' if DBType.MYSQL.name in type or (DBType.MARIADB.name in type and mysql_version < 11.0) else 'mariadb'
        root_pwd = node["ex_params"].get("mysql_root_password") if DBType.MYSQL.name in type else node["ex_params"].get("mariadb_root_password")
        dumpall = 'dumpall' in backup_type
        prefix = 'mariadb-' if DBType.MARIADB.name in type and float(node['version']) >= 11.0 else 'mysql'
        if not dumpall:
            db_data = json.loads(data['db_data'])
            db_name = db_data['name'].strip()

        compress = backup['ex_params'].get('compress')

        if backup_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)
        
        file = f'''{backup_path}/{f"{backup['name']}.dump" if not compress else f"{backup['name']}.gz"}'''
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if backup_host['name'] != host['name']:
        step = 'copying dump file to primary'
        duration_step_start = time.perf_counter()

        jsn = transfer_files_rsync(backup_host, host, backup_path, [file])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'restoring dump'
    duration_step_start = time.perf_counter()

    try:
        target = '' if dumpall else f"'{db_name}'"
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node['port']}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        if jsn['stdout'] == '':
            service = f'mysql{appendix}' if DBType.MYSQL.value in type else 'mariadb' 
            jsn = json.loads(remote_command(host, f'sudo systemctl start {service}',  True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            time.sleep(5)
        
        if not dumpall:
            conn = get_mysql_connection(host, node)
            if isinstance(conn, tuple):
                error = str(conn[1])
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                return
            
            with conn:
                with conn.cursor() as cur:
                    cur.execute(f"CREATE DATABASE IF NOT EXISTS {db_name} CHARACTER SET 'utf8' COLLATE 'utf8_general_ci'")

        else:
            # dump database 'mysql' just for case, since it does not included in --all-databases
            mysql_backup_path = f"{TASK_TMP_FILES_DIR}/{node['name']}"
            mysql_dump = F"{mysql_backup_path}/{backup['name']}_mysql.dump"
            commands = [
                  f"sudo mkdir -p '{mysql_backup_path}'",
                  f"sudo {prefix}dump --default-character-set=utf8mb4 --port='{node['port']}' --password='{root_pwd}' --quick --databases mysql --result-file='{mysql_dump}' && sudo chown mysql:mysql '{mysql_dump}'"
            ]
            for command in commands:
                jsn = json.loads(remote_command(host,command, True), strict=False)
                if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
        
        jsn = json.loads(remote_command(host, f"sudo {cli} --default-character-set=utf8mb4 -Nsfu root --password='{root_pwd}' {target} < '{file}'" if not compress else f"sudo gunzip < '{file}' | sudo {cli} --default-character-set=utf8mb4 -Nsfu root --password='{root_pwd}' {target}", True), strict=False)
        if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        if dumpall:
            # restore database 'mysql' dump
            jsn = json.loads(remote_command(host, f"sudo {cli} --default-character-set=utf8mb4 -Nsfu root --password='{root_pwd}' mysql < '{mysql_dump}'", True), strict=False)
            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            remote_command(host, f"sudo rm -f '{mysql_dump}'", True)
            remote_command(host, f"sudo rm -f '{file}'", True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if standbys and len(standbys) > 0:
        step = 'restoring replication'
        duration_step_start = time.perf_counter()

        try:
            conn = get_mysql_connection(host, node)
            if isinstance(conn, tuple):
                error = str(conn[1])
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                return
        
            with conn:
                with conn.cursor(dictionary=True) as cur:
                    cur.execute(f"SHOW {'BINARY LOG' if DBType.MYSQL.name in type and mysql_version > 8.0 else 'MASTER'} STATUS")
                    master_replication = cur.fetchone()
        
            if DBType.MYSQL.name in type:
                master = "MASTER" if mysql_version < 8.0 else 'SOURCE'
                replica = "SLAVE" if mysql_version < 8.0 else 'REPLICA'
                master_alias = "MASTER" if master == "MASTER" else "REPLICATION SOURCE"
            else:
                replica = "REPLICA" if mysql_version >= 10.5 or mysql_version in [10.10, 10.11] else 'SLAVE'

            for item in standbys:
                standby_data = json.loads(item['node'])
                host_standby = json.loads(item['host'])
                rep_user = standby_data["ex_params"].get("rep_user")
                rep_pwd = standby_data["ex_params"].get("rep_pwd")
                if DBType.MYSQL.name in type:
                    change_master_command = f'''CHANGE {master_alias} TO {master}_HOST='{host["name"]}', {master}_PORT={node["port"]}, {master}_USER='{rep_user}', {master}_PASSWORD='{rep_pwd}', {f"{master}_SSL=1, {master}_SSL_VERIFY_SERVER_CERT=0, {master}_SSL_CAPATH='{node_datadir}', {master}_SSL_CA='{node_datadir}/ca.pem', {master}_SSL_CERT='{node_datadir}/client-cert.pem', {master}_SSL_KEY='{node_datadir}/client-key.pem'," if use_ssl else ""} {master}_LOG_FILE='{master_replication['File']}', {master}_LOG_POS={master_replication['Position']}'''
                else:
                    change_master_command = f'''CHANGE MASTER TO MASTER_HOST='{host["name"]}', MASTER_USER='{rep_user}', MASTER_PASSWORD='{rep_pwd}', MASTER_PORT={node["port"]}, {f"MASTER_SSL=1, MASTER_SSL_VERIFY_SERVER_CERT=0, MASTER_SSL_CAPATH='{node_datadir}', MASTER_SSL_CA='{node_datadir}/ca.pem', MASTER_SSL_CERT='{node_datadir}/client-cert.pem', MASTER_SSL_KEY='{node_datadir}/client-key.pem'," if use_ssl else ""} MASTER_LOG_FILE='{master_replication['File']}', MASTER_LOG_POS={master_replication['Position']}'''
                
                # restore on all standbys the same dump as on primary
                jsn = transfer_files_rsync(backup_host, host_standby, backup_path, [file])
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
                
                for command in open_read_write_commands:
                     jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                     if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                     
                jsn = json.loads(remote_command(host_standby, f"sudo {cli} --default-character-set=utf8mb4 -Nsfu root --password='{root_pwd}' {target} < '{file}'" if not compress else f"sudo gunzip < '{file}' | sudo {cli} --default-character-set=utf8mb4 -Nsfu root --password='{root_pwd}' {target}", True), strict=False)
                if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return

                remote_command(host_standby, f"sudo rm -f '{file}'", True)

                for command in open_read_only_commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return

                conn = get_mysql_connection(host, node)
                if isinstance(conn, tuple):
                    error = str(conn[1])
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                    return
            
                with conn:
                    with conn.cursor(dictionary=True) as cur:
                        queries = [
                            # we now care of preserve table 'mysql' on primary
                            # - but still add this, for insurance and conformance with restore backup task
                            f"""CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' REQUIRE SSL""" if DBType.MYSQL.value in type else f"""CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED VIA ed25519 USING PASSWORD('{rep_pwd}') {'REQUIRE SSL' if use_ssl else ''}""",
                            f"""ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' {'REQUIRE SSL' if use_ssl else ''}""" if DBType.MYSQL.value in type else f"""ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED VIA ed25519 USING PASSWORD('{rep_pwd}') {'REQUIRE SSL' if use_ssl else ''}""",
                            f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
                            f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
                            "FLUSH PRIVILEGES",
                        ]
                        for query in queries:
                            cur.execute(query)
                    
                conn = get_mysql_connection(host_standby, standby_data)
                if isinstance(conn, tuple):
                    error = str(conn[1])
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                    return
            
                with conn:
                    with conn.cursor(dictionary=True) as cur:
                        queries = [ 
                            f"STOP {replica}",
                            f"RESET {replica} ALL",
                            change_master_command,
                            f"START {replica} USER = '{rep_user}' PASSWORD = '{rep_pwd}' DEFAULT_AUTH='{authentication_plugin}'" if DBType.MYSQL.value in type else f"START {replica}"
                        ]
                        for query in queries:
                            cur.execute(query)
                
                if DBType.MYSQL.value in type:
                    query = """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role =%s, ex_params = ex_params || '{"mysql_master_id":%s}', ind_started='true' WHERE id=%s"""
                else:
                    query = """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role =%s, ex_params = ex_params || '{"mariadb_master_id":%s}', ind_started='true' WHERE id=%s"""

                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute(query, (Severity.OK.value, Status.OK.value, ClusterRoles.STANDBY.value, node['id'], standby_data['id'],))
                        conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    if dumpall:
        step = 'updating app database'
        duration_step_start = time.perf_counter()

        try:
            conn = get_mysql_connection(host, node)
            if isinstance(conn, tuple):
                error = str(conn[1])
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                return
            
            with conn:
                with conn.cursor(dictionary=True) as cur:
                    cur.execute("SELECT DISTINCT TABLE_SCHEMA FROM information_schema.TABLES WHERE TABLE_SCHEMA NOT IN ('information_schema', 'mysql', 'performance_schema', 'sys', 'test', 'tmp')")
                    db_set_from_restore = {i['TABLE_SCHEMA'].strip() for i in cur.fetchall()}
                    with get_postgres_connection() as app_conn:
                        with app_conn.cursor(cursor_factory=RealDictCursor) as app_cur:
                            app_cur.execute("""SELECT DISTINCT ON (name) id, name FROM "clusterCore_c123_database" WHERE node_id=%s""", (node['id'],))
                            db_list_from_app = app_cur.fetchall()
                            db_set_from_app = {db['name'].strip() for db in db_list_from_app}
                            common_db = db_set_from_app.intersection(db_set_from_restore)
                            missing_db = db_set_from_restore - db_set_from_app
                            extra_db = db_set_from_app - db_set_from_restore
                    
                            # in order of workload ascending
                            for db in extra_db:
                                app_cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", [Severity.NOT_OK.value, Status.OK.value, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                                app_conn.commit()

                            for db in common_db:
                                cur.execute("SELECT SUM(data_length + index_length)/1024 AS DB_SIZE_KB FROM information_schema.TABLES WHERE TABLE_SCHEMA = %s", (db,))    
                                res = cur.fetchone()['DB_SIZE_KB']
                                db_size = 0.0 if not res else res
                                app_cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, "size"=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, Status.OK.value, db_size, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                                app_conn.commit()
                                
                            for db in missing_db:
                                cur.execute("SELECT SUM(data_length + index_length)/1024 AS DB_SIZE_KB FROM information_schema.TABLES WHERE TABLE_SCHEMA = %s", (db,))    
                                res = cur.fetchone()['DB_SIZE_KB']
                                db_size = 0.0 if not res else res

                                app_cur.execute("""INSERT INTO "clusterCore_c123_database" 
                                ("name", "type", created_date, last_update_date, size, ind_deployed, ind_started, ind_autostartup, ind_deleted, severity, sizing, database_user, database_pwd, created_by_id, last_update_by_id, node_id) 
                                VALUES (%s, %s, now(), now(), %s, 'false', 'true', 'false', 'false', %s, 'MEDIUM', %s, %s, %s, %s, %s)""", (db, DBType.MYSQL.value if DBType.MYSQL.value in type else DBType.MARIADB.value, db_size, Severity.OK.value, node['database_user'], node['database_pwd'], created_by, created_by, node['id']))
                                app_conn.commit()
        except Exception as error:
            print(traceback.format_exc())
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        if not dumpall:
            jsn = json.loads(remote_command(host, f'''sudo {cli} -Nsfu root --password='{root_pwd}' -e "SELECT SUM(data_length + index_length)/1024 FROM information_schema.TABLES WHERE table_schema = '{db_name}';"''', True), strict=False)
            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            temp = jsn['stdout'].strip()
            db_size = 0 if temp == '' or 'null' in temp.lower() else float(temp) 

        jsn = json.loads(remote_command(host, f"sudo du -sb '{node_datadir}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = round(float(jsn['stdout']) / 1024, 2)
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id']))
                if not dumpall:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_size, db_data['id']))
                
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end, created_by)

def backup_mysql_node_task(data, task_id):
    """backup of mysql node"""

    task = 'backup mysql node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        backup_host = data['backup_host']
        node = json.loads(data['node_data'])
        backup = json.loads(data['backup_data'])
        node = json.loads(data['node_data'])
        compress = backup["ex_params"].get('compress')
        is_sched = backup['is_sched']
        incremental = backup['ex_params'].get("incremental")
        backup_path = backup['ex_params'].get('backup_path')
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        conf_path = "/etc/my.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mysql.conf.d/mysqld.cnf"
        if backup_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'backing up node'
    duration_step_start = time.perf_counter()

    command = f"sudo xtrabackup --defaults-file='{conf_path}' --backup --user=root --port='{node['port']}' --password='{node['ex_params'].get('mysql_root_password')}'"

    if not isinstance(incremental, str):
        if not is_sched:
            command += f" --target-dir='{backup_path}'"
        elif incremental is False:
            start_date = datetime.now().astimezone(pytz.UTC).strftime('%d%m%Y%H%M%S')
            backup_path = f"{backup_path}/{start_date}"
            command += f" --target-dir='{backup_path}'"
        elif incremental is True:
            incremental = f"{backup_path}/base"
            command += f" --target-dir='{incremental}'"
    else:
        jsn = check_remote_directory_exists(backup_host, incremental, 'd')
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        incremental_exists = 'true' in jsn['stdout']
        if incremental_exists:
            jsn = json.loads(remote_command(backup_host, f"sudo grep 'to_lsn' '{incremental}/xtrabackup_checkpoints' | awk '{{print $3}}'", True), strict = False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            last_lsn = jsn['stdout']
            if last_lsn:
                incremental = f"{backup_path}/{last_lsn}"
                command += f" --incremental-lsn={last_lsn}"
            else:
                incremental = f"{backup_path}/base"
        else:
            base_dir = f"{backup_path}/base"
            jsn = check_remote_directory_exists(backup_host, base_dir, 'd')
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            base_exists = 'true' in jsn['stdout']
            if not base_exists:
                incremental = base_dir
            else:
                try:
                    jsn = json.loads(remote_command(backup_host,f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n | tail -1 | awk '{{print $2}}'", True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return

                    if jsn['stdout'] != '':
                        dir_list = jsn['stdout'].split('\n')
                        if len(dir_list) > 0:
                            incremental = dir_list[0]
                            jsn = json.loads(remote_command(backup_host, f"sudo grep 'to_lsn' '{incremental}/xtrabackup_checkpoints' | awk '{{print $3}}'", True), strict = False)
                            if jsn['stderr'] != '':
                                status = Severity.NOT_OK.value
                                duration_step_end = time.perf_counter() - duration_step_start
                                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                                step = 'null'
                                status = Severity.NOT_OK.value
                                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                                return

                            last_lsn = jsn['stdout']
                            if last_lsn:
                                incremental = f"{backup_path}/{last_lsn}"
                                command += f" --incremental-lsn={last_lsn}"
                except Exception as error:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                    return
    
        command += f" --target-dir='{incremental}'"

    try:        
        jsn = json.loads(remote_command(host, f"sudo mkdir -p '{backup_path if not incremental else incremental}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    if compress and incremental:
        command += f" --extra-lsndir='{incremental}'"
    if compress:
        command += f" --stream=xbstream | sudo gzip -c | sudo tee '{incremental if incremental else backup_path}/{backup['name']}.gz'"
    
    try:
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        error_lower = jsn['stderr'].lower()
        if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
   
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host['name'] != backup_host['name']:
        step = 'copying backup of node to backup host'
        duration_step_start = time.perf_counter()

        transfer_path = backup_path if not incremental else incremental
        jsn = transfer_files_rsync(host, backup_host, transfer_path, [f"{transfer_path}/"])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'getting the backup size'
    duration_step_start= time.perf_counter()

    try:
        jsn = json.loads(remote_command(backup_host, f"""sudo du -b{'' if compress else 's'} '{incremental if incremental else backup_path}{f'/{backup["name"]}.gz' if compress else ''}' | awk '{{print $1}}'""", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        size = jsn['stdout'].strip()
        backup_size = float(size)/1024

        modified_date = get_date_modified_folder(hostdata=backup_host, folder_path=backup_path)
        if isinstance(modified_date, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return
        
        remote_command(host, f"sudo rm -rf '{backup_path}'", True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,backup['id']))           
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}' WHERE id=%s""" % (modified_date, backup['id']))
                if isinstance(incremental, str):
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{incremental}}', '"{0}"', true) WHERE id={1}""".format(incremental,backup['id']))
                
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end, created_by)

def copy_mysql_database_task(data, task_id):
    """copy database from some instance into another"""

    task = 'copy mysql database task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_source = json.loads(data['host_source'])
        host_target = json.loads(data['host_target'])
        node_source = json.loads(data['node_source'])
        node_target = json.loads(data['node_target'])
        db_source = json.loads(data['db_source'])
        db_target = json.loads(data['db_target'])
        # db_option = 'override' or 'copy'
        db_option = data['db_option']
        db_exists = data['db_exists']
        copy_name = data.get('copy_name')
        copy_users = data['copy_users']
        # copy_option = data['copy_option']
        host_source_name = host_source['name'].strip()
        host_target_name = host_target['name'].strip()
        db_source_name = db_source['name'].strip() 
        node_source_name = node_source['name'].strip()
        new_db_name = copy_name if isinstance(copy_name, str) and len(copy_name.strip()) > 0 else db_target["name"]
        tmp_folder = f"{TASK_TMP_FILES_DIR}/{node_source_name}"
        file = f"{tmp_folder}/{db_source_name}.dump"
        source_root_password = node_source["ex_params"].get("mysql_root_password")
        target_root_password = node_target["ex_params"].get("mysql_root_password")
        use_ssl = node_target["ex_params"].get("ssl")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
        
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'preparing to copy'
    duration_step_start = time.perf_counter()
    
    if host_source_name == host_target_name:
        # split same host dump preparation from copy - this runs now
        commands_target = [
            f"sudo mkdir -p '{tmp_folder}'",
            f"sudo mysqldump --default-character-set=utf8mb4 --port='{node_source['port']}' --password='{source_root_password}' --routines --events --quick '{db_source_name}' --result-file='{file}'",
            f"sudo chown mysql:mysql '{file}'",
        ]
        for command in commands_target:
            try:
                jsn = json.loads(remote_command(host_target, command, True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and 'error' in error_lower:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return
        
        commands_target = [
            f"sudo mysql --default-character-set=utf8mb4 -Nsfu root --password='{target_root_password}' '{new_db_name}' < '{file}'"
        ]
    else:
        commands_source = [
            f"sudo mkdir -p '{tmp_folder}'",
            f"sudo chown -R mysql:mysql '{tmp_folder}'",
            f"sudo mysqldump --default-character-set=utf8mb4 --port='{node_source['port']}' --password='{source_root_password}' --routines --events --quick --result-file='{file}' '{db_source_name}'"
        ]
        for command in commands_source:
            try:
                jsn = json.loads(remote_command(host_source, command, True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and 'error' in error_lower:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
            
        jsn = transfer_files_rsync(host_source, host_target, tmp_folder, [file])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        commands_target = [
            f"sudo chown -R mysql:mysql '{tmp_folder}'",
            f"sudo mysql --default-character-set=utf8mb4 -Nsfu root --password='{target_root_password}' '{new_db_name}' < '{file}'",
        ]
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if db_exists and db_option == "override":
        step = 'removing target database'
        duration_step_start = time.perf_counter()

        try:
            conn = get_mysql_connection(host_target, node_target)
            if isinstance(conn, tuple):
                error = str(conn[1])
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                return
            
            with conn:
                with conn.cursor() as cur:
                    # kill all the connections to db <db_name>
                    cur.execute("SELECT id FROM information_schema.processlist WHERE db = %s", (new_db_name,))
                    connection_ids = [i[0] for i in cur.fetchall()]
                    # although some blazing fast client could reconnect during loop - thats the only way...
                    for connection in connection_ids:
                        cur.execute(f"KILL {connection}")
                    
                    cur.execute(f"DROP DATABASE IF EXISTS {new_db_name}")
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
         
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'copying the database'
    duration_step_start = time.perf_counter()

    try:
        conn = get_mysql_connection(host_target, node_target)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor() as cur:
                cur.execute(f"CREATE DATABASE IF NOT EXISTS {new_db_name} CHARACTER SET 'utf8' COLLATE 'utf8_general_ci'")

        for command in commands_target:
            jsn = json.loads(remote_command(host_target, command, True), strict=False)
            error_lower = jsn['stderr'].lower()
            if error_lower != '' and 'error' in error_lower:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if copy_users and not (host_source_name == host_target_name and db_option == 'override'):
        step = 'copying database_users'
        duration_step_start = time.perf_counter()

        create_users_builder = f'''SELECT CONCAT('CREATE USER IF NOT EXISTS ', QUOTE(mysql.user.user), '@', QUOTE(mysql.user.host), ' IDENTIFIED WITH ', mysql.user.plugin, ' BY ', '\\'initial_password\\'', '{'REQUIRE SSL' if use_ssl else ''};') AS create_user_command FROM mysql.user JOIN mysql.db ON mysql.user.user = mysql.db.user WHERE Db = %s'''
        set_auth_builder = '''SELECT CONCAT('UPDATE mysql.user SET authentication_string = ', QUOTE(authentication_string), ' WHERE user = ', QUOTE(mysql.user.user), ' AND host = ', QUOTE(mysql.user.host), ';') AS update_user_command FROM mysql.user JOIN mysql.db ON mysql.user.user = mysql.db.user WHERE Db = %s'''
        show_grants_builder = '''SELECT CONCAT('SHOW GRANTS FOR ', QUOTE(user), '@', QUOTE(host), ';') AS show_grants_command FROM mysql.db WHERE Db = %s'''
        
        try:
            conn = get_mysql_connection(host_source, node_source)
            if isinstance(conn, tuple):
                error = str(conn[1])
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                return
        
            with conn:
                with conn.cursor() as cur:
                    cur.execute(create_users_builder, (db_source_name,))
                    migrate_user_queries = [i[0] for i in cur.fetchall()]
                    cur.execute(set_auth_builder, (db_source_name,))
                    migrate_user_queries += [i[0] for i in cur.fetchall()]
                    cur.execute(show_grants_builder, (db_source_name,))
                    show_grants_queries = [i[0] for i in cur.fetchall()]
                    for show_grants_query in show_grants_queries:
                        cur.execute(show_grants_query)
                        migrate_grants_queries = [i[0] for i in cur.fetchall()]
                        for migrate_grants_query in migrate_grants_queries:
                            migrate_user_queries.append(migrate_grants_query.replace("`", "'").replace("'.", ".").replace(f"{db_source_name}.", f"{new_db_name}.").replace("ON '", "ON "))
         
            migrate_user_queries.append("FLUSH PRIVILEGES")
            conn = get_mysql_connection(host_target, node_target)
            if isinstance(conn, tuple):
                error = str(conn[1])
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                return
        
            with conn:
                with conn.cursor() as cur:
                    for query in migrate_user_queries:
                        cur.execute(query)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        jsn = json.loads(remote_command(host_target, f"sudo du -sb '{node_target['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = round(float(jsn['stdout'].strip())/1024)
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, size=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node_target['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_target['id'],))
                if not db_exists:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_source['id'],))
                conn.commit()
        
        remote_command(host_source, f"sudo rm -rf '{tmp_folder}'", True)
        if host_source['name'] != host_target['name']:
            remote_command(host_source, f"sudo rm -rf '{tmp_folder}'", True)
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restore_mysql_node_task(data, task_id):
    """restore node from backup"""

    task = 'restore mysql node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node = json.loads(data['node_data'])
        use_ssl = node['ex_params'].get('ssl')
        backup = json.loads(data['backup_data'])
        backup_host = data['backup_host']
        mysql_version = float(node['version'])
        if use_ssl:
            authentication_plugin = 'sha256_password' if mysql_version < 8.0 else 'caching_sha2_password'
        else:
            authentication_plugin = 'mysql_native_password'
        replica = "REPLICA" if mysql_version >= 8.0 else "SLAVE"
        master = "SOURCE" if mysql_version >= 8.0 else "MASTER"
        compress = backup['ex_params'].get('compress')
        node_datadir = node['dir_123Cluster']
        restore_option = data['restore_option'].strip().lower()
        custom_pitr = RestoreOptions.CUSTOM.value in restore_option
        no_pitr = RestoreOptions.THIS.value in restore_option
        appendix = "d" if OperatingSystem.CENTOS.value in distributor else ""
        incremental = backup["ex_params"].get("incremental")
        node_root_pwd = node['ex_params'].get('mysql_root_password')
        is_sched = backup['is_sched']
        conf_path = "/etc/my.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mysql.conf.d/mysqld.cnf"
        conf_file = os.path.basename(conf_path)
        master_alias = "MASTER" if master == "MASTER" else "REPLICATION SOURCE"
        primary_clone = f"{TASK_TMP_FILES_DIR}/mysql_clones/{node['name'].replace(':', '_')}"
        tmp_conf_path = f"{TASK_TMP_FILES_DIR}/mysql_conf/{node['name']}"
        standbys = data.get('standbys')
        has_standby = isinstance(standbys, list) and len(standbys) > 0
        if has_standby:
            standbys = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in standbys]

        backup_path = backup['ex_params'].get('backup_path')
        if backup_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)

        base_dir = f"{backup_path}/base"
        dir_list = None
        if is_sched and not (incremental and no_pitr):
            if restore_option in [RestoreOptions.LATEST.value, RestoreOptions.CUSTOM.value]:
                if RestoreOptions.CUSTOM.value in restore_option:
                    timezone_cmd = 'cat /etc/timezone'
                    jsn = json.loads(remote_command(backup_host, timezone_cmd, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return

                    try:
                        timezone = pytz.timezone(jsn['stdout'])
                        target_datetime = datetime.strptime(data['date'], "%Y-%m-%d %H:%M:%S").astimezone(timezone)
                        target_datetime_str = target_datetime.strftime("%Y-%m-%d %H:%M:%S")
                        target_timestamp = target_datetime.timestamp()
                    except Exception as e:
                        error = f'Invalid server timezone: {jsn["stdout"]} error: {str(e)}'
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
                        return
    
                if incremental:
                    if custom_pitr:
                        command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d  ! -newermt '{target_datetime_str}' -exec stat -c '%Y %n' {{}} \; | grep -v '/base' | sort -n | awk '{{print $2}}'"
                    else:
                        command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; |  grep -v '/base' | sort -n | awk '{{print $2}}'"

                elif not incremental:
                    if custom_pitr:
                        command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d  -exec stat -c '%Y %n' {{}} \; | sort -n"
                    else:
                        command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n | tail -1 | awk '{{print $2}}'"
    
            elif no_pitr:
                command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n | head -1 | awk '{{print $2}}'"

            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            if jsn['stdout'] != '':
                if incremental and not no_pitr:
                    dir_list = jsn['stdout'].split('\n')

                if not incremental:
                    if jsn['stdout'] == '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                        return
                    else:
                        if restore_option in [RestoreOptions.THIS.value, RestoreOptions.LATEST.value]:
                            backup_path = jsn['stdout']
                        else:
                            output_lines = [line.strip() for line in jsn['stdout'].split("\n") if line.strip()]
                            closest_subfolder = None
                            min_time_difference = positive_infinity

                            for line in output_lines:
                                subfolder_timestamp_str, subfolder = line.split(" ", 1)
                                subfolder_timestamp = int(subfolder_timestamp_str)
                                time_difference = abs(target_timestamp - subfolder_timestamp)
                                if time_difference < min_time_difference:
                                    min_time_difference = time_difference
                                    closest_subfolder = subfolder

                            if not closest_subfolder:
                                status = Severity.NOT_OK.value
                                duration_step_end = time.perf_counter() - duration_step_start
                                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                                step = 'null'
                                status = Severity.NOT_OK.value
                                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                                return

                            backup_path = closest_subfolder
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host['name'] != backup_host['name']:
        step = 'copying backup files of node to new host'
        duration_step_start = time.perf_counter()

        jsn = transfer_files_rsync(backup_host, host, backup_path, [f"{backup_path}/"])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'preparing to restore'
    duration_step_start = time.perf_counter()

    commands = [
        f"sudo mkdir -p '{tmp_conf_path}'",
        f"sudo rsync -a '{conf_path}' '{tmp_conf_path}/{conf_file}'",
    ]
    
    if is_sched and incremental:
        if compress:
            filename_cmd = f"sudo ls -lt --file-type '{base_dir}' | grep '\.gz$' | head -n 1 | awk '{{print $9}}'"
            jsn = json.loads(remote_command(host, filename_cmd, True), strict = False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            archive_file = jsn['stdout']
            commands.append(f"sudo find '{base_dir}' -mindepth 1 -maxdepth 1 -type f -name 'xtrabackup*' -print0 | xargs -0 sudo rm -f && sudo gunzip < '{base_dir}/{archive_file}' | sudo xbstream -C '{base_dir}' -x && sudo chown -R mysql:mysql '{base_dir}'")
        
        commands.append(f"sudo xtrabackup --defaults-file='{conf_path}' --prepare --apply-log-only --target-dir='{base_dir}'")
        if dir_list:
            for inc_dir in dir_list:
                inc_dir = inc_dir.strip()
                prepare_dir = [f"sudo xtrabackup --defaults-file='{conf_path}' --prepare {'--apply-log-only' if inc_dir != dir_list[:-1] else ''} --target-dir='{base_dir}' --incremental-dir='{inc_dir}'"]
                if compress:
                    filename_cmd = f"sudo ls -lt --file-type '{inc_dir}' | grep '\.gz$' | head -n 1 | awk '{{print $9}}'"
                    jsn = json.loads(remote_command(host, filename_cmd, True), strict = False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                    
                    archive_file = jsn['stdout']
                    prepare_dir.insert(0,f"sudo find '{inc_dir}' -mindepth 1 -maxdepth 1 -type f -name 'xtrabackup*' -print0 | xargs -0 sudo rm -f && sudo gunzip < '{inc_dir}/{archive_file}' | sudo xbstream -C '{inc_dir}' -x && sudo chown -R mysql:mysql '{inc_dir}'""")
                
                commands += prepare_dir
    else:
        filename_cmd = f"sudo ls -lt --file-type '{backup_path}' | grep '\.gz$' | head -n 1 | awk '{{print $9}}'"
        jsn = json.loads(remote_command(host, filename_cmd, True), strict = False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        archive_file = jsn['stdout']
        commands.append(f"sudo chown -R mysql:mysql '{backup_path}'" if not compress else f"sudo gunzip < '{backup_path}/{archive_file}' | sudo xbstream -C '{backup_path}' -x && sudo chown -R mysql:mysql '{backup_path}'""")
    
    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if has_standby:
        step = 'disconnecting standbys'
        duration_step_start = time.perf_counter()

        last_known_err = ''

        try:
            for item in standbys:
                standby_data = item['node']
                host_standby = item['host']
                conn = get_mysql_connection(host_standby, standby_data)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                f"STOP {replica}",
                                f"RESET {replica} ALL"
                            ]

                            for query in queries:
                                cur.execute(query)
        except Exception as error:
            last_known_err = str(error)
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                       
        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'restoring node from backup'
    duration_step_start = time.perf_counter()
    
    try:
        if use_ssl:
            jsn = transfer_ssl_certificates(node, host, host, mode='backup')
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        commands = [
            f"sudo xtrabackup --defaults-file='{conf_path}' --prepare --target-dir='{backup_path if not (is_sched and incremental) else base_dir}'",
            f'sudo systemctl stop mysql{appendix}',
            f"sudo rm -rf '{node_datadir}'",
            f"sudo mkdir -p '{node_datadir}'",
            f"sudo xtrabackup --defaults-file='{conf_path}' --copy-back --target-dir='{backup_path if not (is_sched and incremental) else base_dir}'",
            f"sudo rsync -a '{tmp_conf_path}/{conf_file}' '{conf_path}'",
            f"sudo rm -rf '{tmp_conf_path}'",
            f"""sudo sed -i '/super_read_only=ON/d' '{conf_path}'""",
            f"sudo mkdir -p '/var/run/mysql'",
            "sudo chown -R mysql:mysql '/var/run/mysql'"
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                if backup_host['name'] != host['name']:
                    remote_command(host, f"sudo rm -rf '{backup_path}'", True)
                return
        
        if use_ssl:
            jsn = transfer_ssl_certificates(node, host, host, mode='restore')
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        commands = [
            f"sudo chown -R mysql:mysql '{node_datadir}'",
            'sudo rm -rf /var/lib/mysql/auto.cnf',
            f'sudo rm -rf {node_datadir}/auto.cnf',
            f'sudo systemctl start mysql{appendix}'
        ] 
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict = False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        remote_command(host, f"sudo rm -rf '{backup_path}'", True)
        
        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor(dictionary=True) as cur:
                cur.execute(f"SHOW {'BINARY LOG' if mysql_version > 8.0 else 'MASTER'} STATUS")
                master_replication = cur.fetchone()
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role =%s, ex_params = ex_params || '{"mysql_master_id":%s}', ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.PRIMARY.value, node['id'], node['id'],))
                conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if has_standby:
        step = 'restoring replication'
        duration_step_start = time.perf_counter()
        
        last_known_err = ''
        now = datetime.utcnow().strftime('%d%m%Y%H%M%S')
        backup_name = f"{node['name'].replace(':', '_')}_{now}"
        transfer_path = f"{primary_clone}/{backup_name}.gz"
        backup_cmd  = f"sudo xtrabackup --defaults-file='{conf_path}' --backup --user=root --port='{node['port']}' --password='{node_root_pwd}' --target-dir='{primary_clone}' --stream=xbstream | sudo gzip -c | sudo tee '{transfer_path}'"
        try:
            commands = [
                f"sudo mkdir -p '{primary_clone}'",
                backup_cmd
            ]
            for command in commands:
                jsn = json.loads(remote_command(host, command, True), strict=False)
                error_lower = jsn['stderr'].lower()
                if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
            for item in standbys:
                standby_data = item['node']
                host_standby = item['host']
                rep_user = standby_data["ex_params"].get("rep_user")
                rep_pwd = standby_data["ex_params"].get("rep_pwd")
                standby_tmp_conf_dir = f"{TASK_TMP_FILES_DIR}/mysql_conf/{standby_data['name']}"
                standby_data_dir = standby_data["dir_123Cluster"]
                if standby_data_dir.startswith('~'):
                    jsn = find_tilda(host_standby)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                    tilda_path = jsn['stdout']
                    standby_data_dir = standby_data_dir.replace('~', tilda_path)
                
                commands = [
                    f"sudo mkdir -p '{standby_tmp_conf_dir}'",
                    f"sudo rsync -a '{conf_path}' '{standby_tmp_conf_dir}/{conf_file}'"
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                        
                jsn = transfer_files_rsync(host, host_standby, primary_clone, [transfer_path])
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                conn = get_mysql_connection(host, node)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                f"CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' {'REQUIRE SSL' if use_ssl else ''}",
                                f"ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' {'REQUIRE SSL' if use_ssl else ''}",
                                f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
                                f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
                                "FLUSH PRIVILEGES",
                            ]
                            for query in queries:
                                cur.execute(query)
            
                commands = [
                    f"sudo gunzip < '{transfer_path}' | sudo xbstream -C '{primary_clone}' -x",
                    f"sudo chown -R mysql:mysql '{primary_clone}'",
                    f"sudo xtrabackup --defaults-file='{conf_path}' --prepare --target-dir='{primary_clone}'",
                    f"sudo systemctl stop mysql{appendix}",
                    f"sudo rm -rf '{standby_data_dir}'",
                    f"sudo mkdir -p '{standby_data_dir}'",
                    f"sudo xtrabackup --defaults-file='{conf_path}' --copy-back --target-dir='{primary_clone}'",
                    f"sudo rsync -a '{standby_tmp_conf_dir}/{conf_file}' '{conf_path}'",
                    f"sudo rm -rf '{standby_tmp_conf_dir}'",
                    f"sudo mkdir -p '/var/run/mysql'",
                    "sudo chown -R mysql:mysql '/var/run/mysql'",
                    "sudo chown -R 700 '/var/run/mysql'"
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    error_lower = jsn['stderr'].lower()
                    if jsn['stderr'] != '' and any(err in error_lower for err in ['error', 'fail', 'errno'] if 'completed ok' not in error_lower):
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                if use_ssl:
                    jsn = transfer_ssl_certificates(node, host, host_standby, standby_data_dir)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                
                commands = [
                    f"sudo chown -R mysql:mysql '{standby_data_dir}'",
                    'sudo rm -rf /var/lib/mysql/auto.cnf',
                    f'sudo rm -rf {standby_data_dir}/auto.cnf',
                    f"sudo systemctl start mysql{appendix}",
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                remote_command(host_standby, f"sudo rm -rf '{primary_clone}'", True)

                conn = get_mysql_connection(host_standby, standby_data)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                f"STOP {replica}",
                                f"RESET {replica} ALL",
                                f'''CHANGE {master_alias} TO {master}_HOST='{host["name"]}', {master}_PORT={node["port"]}, {master}_USER='{rep_user}', {master}_PASSWORD='{rep_pwd}', {f"{master}_SSL=1, {master}_SSL_VERIFY_SERVER_CERT=0, {master}_SSL_CAPATH='{standby_data_dir}', {master}_SSL_CA='{standby_data_dir}/ca.pem', {master}_SSL_CERT='{standby_data_dir}/client-cert.pem', {master}_SSL_KEY='{standby_data_dir}/client-key.pem'," if use_ssl else ""} {master}_LOG_FILE='{master_replication['File']}', {master}_LOG_POS={master_replication['Position']}''',
                                # does not fail if replication credentials are wrong or empty, always check replication status after restore
                                f"START {replica} USER = '{rep_user}' PASSWORD = '{rep_pwd}' DEFAULT_AUTH='{authentication_plugin}'"
                            ]
                            for query in queries:
                                cur.execute(query)
                
                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role =%s, ex_params = ex_params || '{"mysql_master_id":%s}', ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.STANDBY.value, node['id'], standby_data['id'],))
                        conn.commit()
            
            remote_command(host, f"sudo rm -rf '{primary_clone}'", True)
        except Exception as error:
            last_known_err = str(error)
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'] if isinstance(last_known_err, dict) else None,last_known_err['stderr'] if isinstance(last_known_err, dict) else last_known_err,last_known_err['excode'] if isinstance(last_known_err, dict) else 1)
    
    step = 'updating app database'
    duration_step_start = time.perf_counter()
    
    try:
        conn = get_mysql_connection(host, node)
        if isinstance(conn, tuple):
            error = str(conn[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
            return
        
        with conn:
            with conn.cursor(dictionary=True) as cur:
                cur.execute("SELECT DISTINCT TABLE_SCHEMA FROM information_schema.TABLES WHERE TABLE_SCHEMA NOT IN ('information_schema', 'mysql', 'performance_schema', 'sys', 'test', 'tmp')")
                db_set_from_restore = {i['TABLE_SCHEMA'].strip() for i in cur.fetchall()}
                with get_postgres_connection() as app_conn:
                    with app_conn.cursor(cursor_factory=RealDictCursor) as app_cur:
                        app_cur.execute("""SELECT DISTINCT ON (name) id, name FROM "clusterCore_c123_database" WHERE node_id=%s""", (node['id'],))
                        db_list_from_app = app_cur.fetchall()
                        db_set_from_app = {db['name'].strip() for db in db_list_from_app}
                        common_db = db_set_from_app.intersection(db_set_from_restore)
                        missing_db = db_set_from_restore - db_set_from_app
                        extra_db = db_set_from_app - db_set_from_restore
                 
                        # in order of workload ascending
                        for db in extra_db:
                            app_cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", [Severity.NOT_OK.value, Status.OK.value, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                            app_conn.commit()

                        for db in common_db:
                            cur.execute("SELECT SUM(data_length + index_length)/1024 AS DB_SIZE_KB FROM information_schema.TABLES WHERE TABLE_SCHEMA = %s", (db,))    
                            res = cur.fetchone()['DB_SIZE_KB']
                            db_size = 0.0 if not res else res
                            app_cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, "size"=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, Status.OK.value, db_size, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                            app_conn.commit()
                            
                        for db in missing_db:
                            cur.execute("SELECT SUM(data_length + index_length)/1024 AS DB_SIZE_KB FROM information_schema.TABLES WHERE TABLE_SCHEMA = %s", (db,))    
                            res = cur.fetchone()['DB_SIZE_KB']
                            db_size = 0.0 if not res else res

                            app_cur.execute("""INSERT INTO "clusterCore_c123_database" 
                            ("name", "type", created_date, last_update_date, size, ind_deployed, ind_started, ind_autostartup, ind_deleted, severity, sizing, database_user, database_pwd, created_by_id, last_update_by_id, node_id) 
                            VALUES (%s, %s, now(), now(), %s, 'false', 'true', 'false', 'false', %s, 'MEDIUM', %s, %s, %s, %s, %s)""", (db, DBType.MYSQL.value, db_size, Severity.OK.value, node['database_user'], node['database_pwd'], created_by, created_by, node['id']))
                            app_conn.commit()
                        
                        jsn = json.loads(remote_command(host,f"sudo du -bs '{node_datadir}' | awk '{{print $1}}'", True), strict=False)
                        if jsn['stderr'] != '':
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return
                        
                        node_size = round(float(jsn['stdout']) / 1024, 2)
                        app_cur.execute("""UPDATE "clusterCore_c123_node" SET size=%s WHERE id=%s""", (node_size, node['id'],))
                        app_conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end, created_by)

def start_mysql_node_task(data, task_id):
    """start mysql node"""

    task = 'start mysql node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_id = data['node_id']
        appendix = "d" if OperatingSystem.CENTOS.value in distributor else ""
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'starting mysql node'
    duration_step_start = time.perf_counter()

    try:
        command = f'sudo systemctl start mysql{appendix}'
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_id,))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='false' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node_id,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def stop_mysql_node_task(data, task_id):
    """stop mysql node"""

    task = 'stop mysql node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_id = data['node_id']
        appendix = "d" if OperatingSystem.CENTOS.value in distributor else ""
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stopping mysql node'
    duration_step_start = time.perf_counter()

    try:
        command = f'sudo systemctl stop mysql{appendix}'
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false' WHERE id=%s""", (Severity.NOT_OK.value, Status.OK.value, node_id,))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='false' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node_id,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)


def start_mysql_node_task(data, task_id):
    """start mysql node"""

    task = 'start mysql node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_id = data['node_id']
        appendix = "d" if OperatingSystem.CENTOS.value in distributor else ""
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'starting mysql node'
    duration_step_start = time.perf_counter()

    try:
        command = f'sudo systemctl start mysql{appendix}'
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_id,))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='false' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node_id,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def upgrade_mysql_node_task(data, task_id):
    """upgrade mysql node to the target major version"""

    task = 'upgrade mysql node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        data_dir = node['dir_123Cluster'].strip()
        if data_dir.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_dir = data_dir.replace('~', tilda_path)
        target_version = data['target_version']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        if float(target_version) > 8.0 and ( (OperatingSystem.UBUNTU.value in distributor and int(release) < 22) or (OperatingSystem.CENTOS.value in distributor and int(release) < 8)):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f"mysql {target_version} is not supported on {distributor} {release}",9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,f"mysql {target_version} is not supported on {distributor} {release}",9)
            return


        appendix = "d" if OperatingSystem.CENTOS.value in distributor else ""

        target_version_exists = check_mysql_version_exists(host, target_version, distributor)

        conf_path = "/etc/my.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mysql.conf.d/mysqld.cnf"
        standbys = data.get('standbys')
        has_standby = isinstance(standbys, list) and len(standbys) > 0
        if has_standby:
            standbys = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in standbys]
        
        repository = node['ex_params'].get('repository')
        offline = True if Repository.LOCAL.value in repository else False

        app_host = json.loads(data['app_host'])
        mysql_version = float(node['version'])
        use_ssl = node['ex_params'].get('ssl')
        node_root_pwd = node['ex_params'].get('mysql_root_password')
        if use_ssl:
            authentication_plugin = 'sha256_password' if mysql_version < 8.0 else 'caching_sha2_password'
        else:
            authentication_plugin = 'mysql_native_password'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if has_standby:
        step = 'disconnecting standbys'
        duration_step_start = time.perf_counter()

        last_known_err = ''

        try:
            replica = "REPLICA" if mysql_version >= 8.0 else "SLAVE"
            for item in standbys:
                standby_data = item['node']
                host_standby = item['host']
                rep_user = standby_data['ex_params'].get('rep_user')
                rep_pwd = standby_data['ex_params'].get('rep_pwd')
                conn = get_mysql_connection(host_standby, standby_data)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                f"STOP {replica}",
                                f"RESET {replica} ALL"
                            ]

                            for query in queries:
                                cur.execute(query)
        except Exception as error:
            last_known_err = str(error)
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
        
        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'stopping mysql primary'
    duration_step_start = time.perf_counter()

    try:
        conn = get_mysql_connection(host, node)
        if not isinstance(conn, tuple):
            with conn:
                with conn.cursor(dictionary=True) as cur:
                    cur.execute("SET GLOBAL innodb_fast_shutdown = 0")

 
        commands = [
            f'sudo systemctl stop mysql{appendix}'
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '' and 'not loaded' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if not target_version_exists:
        step = 'installing upgraded mysql binaries'
        duration_step_start = time.perf_counter()

        try:
            if OperatingSystem.CENTOS.value in distributor:
                check_mariadb_exists = "sudo rpm -qa | grep 'MariaDB'"
            else:
                check_mariadb_exists = "sudo dpkg -l | grep 'mariadb'"
            
            jsn = json.loads(remote_command(host, check_mariadb_exists, True), strict=False)
            mariadb_exists = jsn['stdout'] != ''
            if mariadb_exists:
                if OperatingSystem.CENTOS.value in distributor:
                    package_manager = 'dnf' if int(release) > 7 else 'yum'
                    remove_mariadb_cmd = f"sudo {package_manager} -q -y remove MariaDB* galera**"
                else:
                    remove_mariadb_cmd = "sudo DEBIAN_FRONTEND=noninteractive apt-get -qq -y purge ^mariadb ^galera"
            
                jsn = json.loads(remote_command(host, remove_mariadb_cmd, True), strict=False)
                error_lower = jsn['stderr'].lower()
                filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['nothing to do'])
                if filter:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(node['id'])
                    return
                
                if OperatingSystem.CENTOS.value in distributor:
                    remote_command(host, 'sudo rm -rf /etc/yum.repos.d/mariadb.repo', True)

            if OperatingSystem.CENTOS.value in distributor:
                check_mysql_exists = "sudo rpm -qa | grep 'mysql'"
            else:
                check_mysql_exists = "sudo dpkg -l | grep 'mysql'"
            
            jsn = json.loads(remote_command(host, check_mysql_exists, True), strict=False)
            mysql_exists = jsn['stdout'] != ''
            if mysql_exists:
                if OperatingSystem.CENTOS.value in distributor:
                    package_manager = 'dnf' if int(release) > 7 else 'yum'
                    remove_mysql_cmd = f"sudo {package_manager} -q -y remove mysql* percona* galera*"
                else:
                    remove_mysql_cmd = "sudo DEBIAN_FRONTEND=noninteractive apt-get -qq -y purge ^mysql ^percona ^galera"

                jsn = json.loads(remote_command(host, remove_mysql_cmd, True), strict=False)
                error_lower = jsn['stderr'].lower()
                filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['nothing to do'])
                if filter:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(node['id'])
                    return

            if offline:
                jsn = install_offline_remote_to_remote(distributor, release, host, target_version, DBType.MYSQL.value, app_host)
            else:
                jsn = install_mysql(distributor, release, host, target_version)
            
            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
            if filter:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(node['id'])
                return

            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    step = 'starting mysql primary'
    duration_step_start = time.perf_counter()
    
    try:

        conf_file = get_mysql_config_file(node['id'], node['port'], target_version, is_standby=False, use_ssl=node['ex_params'].get('ssl'))
        commands = [
            f"""datadir='{data_dir}';echo "{conf_file}"| sudo tee '{conf_path}'""",
            f'sudo systemctl stop mysql{appendix}',
            f'sudo systemctl start mysql{appendix}'
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        if has_standby:
            conn = get_mysql_connection(host, node)
            if isinstance(conn, tuple):
                error = str(conn[1])
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                return
            
            with conn:
                with conn.cursor(dictionary=True) as cur:
                    cur.execute(f"SHOW {'BINARY LOG' if float(target_version) > 8.0 else 'MASTER'} STATUS")
                    master_replication = cur.fetchone()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    if has_standby:
        step = 'cloning standbys'
        duration_step_start = time.perf_counter()

        try:
            for item in standbys:
                standby_data = item['node']
                host_standby = item['host']
                rep_user = standby_data['ex_params'].get('rep_user')
                rep_pwd = standby_data['ex_params'].get('rep_pwd')

                standby_data_dir = standby_data["dir_123Cluster"]
                if standby_data_dir.startswith('~'):
                    jsn = find_tilda(host_standby)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                    tilda_path = jsn['stdout']
                    standby_data_dir = standby_data_dir.replace('~', tilda_path)
                
                master = "SOURCE" if mysql_version >= 8.0 else "MASTER"
                master_alias = "MASTER" if master == "MASTER" else "REPLICATION SOURCE"
                
                commands = [
                    f"sudo systemctl stop mysql{appendix}"
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                conn = get_mysql_connection(host_standby, standby_data)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor(dictionary=True) as cur:
                            cur.execute("SET GLOBAL innodb_fast_shutdown = 0")
                
                commands = [
                    f"sudo systemctl stop mysql{appendix}"
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                target_version_exists = check_mysql_version_exists(host_standby, target_version, distributor)
                if not target_version_exists:
                    if OperatingSystem.CENTOS.value in distributor:
                        check_mariadb_exists = "sudo rpm -qa | grep 'MariaDB'"
                    else:
                        check_mariadb_exists = "sudo dpkg -l | grep 'mariadb'"
                    
                    jsn = json.loads(remote_command(host_standby, check_mariadb_exists, True), strict=False)
                    mariadb_exists = jsn['stdout'] != ''
                    if mariadb_exists:
                        if OperatingSystem.CENTOS.value in distributor:
                            package_manager = 'dnf' if int(release) > 7 else 'yum'
                            remove_mariadb_cmd = f"sudo {package_manager} -q -y remove MariaDB* galera**"
                        else:
                            remove_mariadb_cmd = "sudo DEBIAN_FRONTEND=noninteractive apt-get -qq -y purge ^mariadb ^galera"
                    
                        jsn = json.loads(remote_command(host_standby, remove_mariadb_cmd, True), strict=False)
                        error_lower = jsn['stderr'].lower()
                        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['nothing to do'])
                        if filter:
                            last_known_err = jsn
                            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                        
                        if OperatingSystem.CENTOS.value in distributor:
                            remote_command(host_standby, 'sudo rm -rf /etc/yum.repos.d/mariadb.repo', True)

                    if OperatingSystem.CENTOS.value in distributor:
                        check_mysql_exists = "sudo rpm -qa | grep 'mysql'"
                    else:
                        check_mysql_exists = "sudo dpkg -l | grep 'mysql'"
                    
                    jsn = json.loads(remote_command(host_standby, check_mysql_exists, True), strict=False)
                    mysql_exists = jsn['stdout'] != ''
                    if mysql_exists:
                        if OperatingSystem.CENTOS.value in distributor:
                            package_manager = 'dnf' if int(release) > 7 else 'yum'
                            remove_mysql_cmd = f"sudo {package_manager} -q -y remove mysql* percona* galera*"
                        else:
                            remove_mysql_cmd = "sudo DEBIAN_FRONTEND=noninteractive apt-get -qq -y purge ^mysql ^percona ^galera"

                        jsn = json.loads(remote_command(host_standby, remove_mysql_cmd, True), strict=False)
                        error_lower = jsn['stderr'].lower()
                        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['nothing to do'])
                        if filter:
                            last_known_err = jsn
                            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                    if offline:
                        jsn = install_offline_remote_to_remote(distributor, release, host_standby, target_version, DBType.MYSQL.value, app_host)
                    else:
                        jsn = install_mysql(distributor, release, host_standby, target_version)
                    
                    error_lower = jsn['stderr'].lower()
                    filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
                    if filter:
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                conf_file = get_mysql_config_file(standby_data['id'], standby_data['port'], target_version, is_standby=True, use_ssl=standby_data['ex_params'].get('ssl'))
                commands = [
                    f"""datadir='{standby_data_dir}';echo "{conf_file}"| sudo tee '{conf_path}'""",                 
                    f'sudo systemctl stop mysql{appendix}',
                    f"sudo systemctl start mysql{appendix}"
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                conn = get_mysql_connection(host, node)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                f"CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' {'REQUIRE SSL' if use_ssl else ''}",
                                f"ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' {'REQUIRE SSL' if use_ssl else ''}",
                                f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
                                f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
                                "FLUSH PRIVILEGES",
                            ]
                            for query in queries:
                                cur.execute(query)
                
                conn = get_mysql_connection(host_standby, standby_data)
                if not isinstance(conn, tuple):
                    with conn:
                        with conn.cursor() as cur:
                            queries = [
                                f"STOP {replica}",
                                f"RESET {replica} ALL",
                                f'''CHANGE {master_alias} TO {master}_HOST='{host["name"]}', {master}_PORT={node["port"]}, {master}_USER='{rep_user}', {master}_PASSWORD='{rep_pwd}', {f"{master}_SSL=1, {master}_SSL_VERIFY_SERVER_CERT=0, {master}_SSL_CAPATH='{standby_data_dir}', {master}_SSL_CA='{standby_data_dir}/ca.pem', {master}_SSL_CERT='{standby_data_dir}/client-cert.pem', {master}_SSL_KEY='{standby_data_dir}/client-key.pem'," if use_ssl else ""} {master}_LOG_FILE='{master_replication['File']}', {master}_LOG_POS={master_replication['Position']}''',
                                # does not fail if replication credentials are wrong or empty, always check replication status after restore
                                f"START {replica} USER = '{rep_user}' PASSWORD = '{rep_pwd}' DEFAULT_AUTH='{authentication_plugin}'"
                            ]
                            for query in queries:
                                cur.execute(query)
        except Exception as error:
            last_known_err = str(error)
            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'] if isinstance(last_known_err, dict) else None,last_known_err['stderr'] if isinstance(last_known_err, dict) else last_known_err,last_known_err['excode'] if isinstance(last_known_err, dict) else 1)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='false' WHERE cluster_id=%s""", (Severity.NOT_OK.value, Status.OK.value, target_version, node['cluster_id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restart_mysql_node_task(data, task_id):
    """restart mysql node"""

    task = 'restart mysql node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    
    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_id = data['node_id']
        appendix = "d" if OperatingSystem.CENTOS.value in distributor else ""
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'restarting mysql node'
    duration_step_start = time.perf_counter()

    try:
        command = f'sudo systemctl restart mysql{appendix}'
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_id,))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node_id,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    
def mysql_parameters_apply_task(data, task_id):
    """edit parameters of mysql node"""

    task = 'mysql parameters apply task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        parameters = data['parameters']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        appendix = "d" if OperatingSystem.CENTOS.value in distributor else ""
        conf_path = "/etc/my.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mysql.conf.d/mysqld.cnf"
        conf_file = os.path.basename(conf_path)
        tmp_conf_path = f"{TASK_TMP_FILES_DIR}/mysql_conf/{node['name']}"
        restore_config = f"sudo rsync -a '{tmp_conf_path}/{conf_file}' '{conf_path}'"
        restart_mysql = f"sudo systemctl restart mysql{appendix}"
        parameters_require_restart = []
        parameters_do_not_require_restart = []
        for p in parameters:
            (parameters_require_restart if p['source'].strip() and p['context'] != 'PERSISTED' else parameters_do_not_require_restart).append(p)
        
        port_parameter = next((p for p in parameters if p['name'].strip() == 'port'), None)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'applying parameters to node'
    duration_step_start = time.perf_counter()

    try:
        if len(parameters_do_not_require_restart) > 0:
            conn = get_mysql_connection(host, node)
            if isinstance(conn, tuple):
                error = str(conn[1])
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,9)
                return
            
            with conn:
                with conn.cursor() as cur:
                    for p in parameters_do_not_require_restart:
                        query = f"SET PERSIST {p['name']} = {p['setting']}"
                        cur.execute(query)
            
        if (len(parameters_require_restart) > 0):
            # since manual editing of config file are dangerous - backup the config file, just for case (interrupt task if could not backup config file)
            commands = [
                f"sudo mkdir -p '{tmp_conf_path}'",
                f"sudo rsync -a '{conf_path}' '{tmp_conf_path}/{conf_file}'",
            ]
            for command in commands:
                jsn = json.loads(remote_command(host, command, True), strict=False)
                if jsn['stderr'] != '' or jsn['stdout'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            
            for p in parameters_require_restart:
                name = p['name']
                value = p['setting']
                if name.strip() == 'datadir':
                    continue
                
                edit_config = f"""sudo grep -q "^{p['name']}=" '{conf_path}' && sudo sed -i "s|^{name}=.*|{name}={value}|" '{conf_path}' || sudo sed -i "/\[mysqld\]/a {name}={value}" '{conf_path}' """
                jsn = json.loads(remote_command(host, edit_config, True), strict=False)
                if jsn['stderr'] != '':
                    # attempt to restore previous config, interrupt task in any case
                    jsn_inner = json.loads(remote_command(host, restore_config, True), strict=False)
                    if jsn_inner['stderr'] != '' or jsn_inner['stdout'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn_inner['stdout'],jsn_inner['stderr'],jsn_inner['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn_inner['stdout'],jsn_inner['stderr'],jsn_inner['excode'])
                        return
                    
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            
            jsn = json.loads(remote_command(host, restart_mysql, True), strict=False)
            if jsn['stderr'] != '':
                # attempt to restore previous config and restart the node, interrupt task in any case
                commands = [
                    restore_config,
                    restart_mysql
                ]
                for command in commands:
                    jsn_inner = json.loads(remote_command(host, command, True), strict=False)
                    if jsn_inner['stderr'] != '' or jsn_inner['stdout'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn_inner['stdout'],jsn_inner['stderr'],jsn_inner['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn_inner['stdout'],jsn_inner['stderr'],jsn_inner['excode'])
                        return
                
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            if port_parameter:
                node['port'] = str(port_parameter['setting'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, port=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['port'], node['id'],))
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def switchover_node(host_target, target, distributor, host_primary = None, primary = None, standbys = None):
    try:
        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        mysql_version = float(target['version'])
        use_ssl = target['ex_params'].get('ssl')
        if use_ssl:
            authentication_plugin = 'sha256_password' if mysql_version < 8.0 else 'caching_sha2_password'
        else:
            authentication_plugin = 'mysql_native_password'
        
        target_data_dir = target['dir_123Cluster'].strip()
        if target_data_dir.startswith('~'):
            jsn = find_tilda(host_target)
            if jsn['stderr'] != '':
                return jsn

            tilda_path = jsn['stdout']
            target_data_dir = target_data_dir.replace('~', tilda_path)
        
        replica = "REPLICA" if mysql_version >= 8.0 else "SLAVE"
        master = "SOURCE" if mysql_version >= 8.0 else "MASTER"
        rep_user = target['ex_params'].get('rep_user')
        rep_pwd = target['ex_params'].get('rep_pwd')
        appendix = "d" if OperatingSystem.CENTOS.value in distributor else ""
        conf_path = "/etc/my.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mysql.conf.d/mysqld.cnf"
        master_alias = "MASTER" if master == "MASTER" else "REPLICATION SOURCE"
        has_standby = isinstance(standbys, list) and len(standbys) > 0
        commands = [
            f"sudo sed -i '/super_read_only=ON/d' '{conf_path}'",
            f"sudo systemctl restart mysql{appendix}"
        ]
        for command in commands:
            jsn = json.loads(remote_command(host_target, command, True), strict=False)
            error_lower = jsn['stderr'].lower()
            if error_lower != '' and 'error' in error_lower:
                return jsn
        
        time.sleep(5)
        
        conn = get_mysql_connection(host_target, target)
        if isinstance(conn, tuple):
            jsn['stderr'] = str(conn[1])
            return jsn
        
        queries = [
            f"STOP {replica}",
            f"RESET {replica} ALL",
            f"""CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' {'REQUIRE SSL' if use_ssl else ''}""",
            f"ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' {'REQUIRE SSL' if use_ssl else ''}",
            f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
            f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
            "FLUSH PRIVILEGES",
            f"SHOW {'BINARY LOG' if mysql_version > 8.0 else 'MASTER'} STATUS"
        ]
        with conn:
            with conn.cursor(dictionary=True) as cur:
                for query in queries:
                    cur.execute(query)
                master_replication = cur.fetchone()

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role=%s, ex_params = ex_params || '{"mysql_master_id":%s}', ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.PRIMARY.value, target['id'], target['id'],))
                conn.commit()
         
        change_master_command = f"""CHANGE {master_alias} TO {master}_HOST='{host_target['name']}', {master}_PORT={target['port']}, {master}_USER='{rep_user}', {master}_PASSWORD='{rep_pwd}', {f"{master}_SSL=1, {master}_SSL_VERIFY_SERVER_CERT=0, {master}_SSL_CAPATH='{target_data_dir}', {master}_SSL_CA='{target_data_dir}/ca.pem', {master}_SSL_CERT='{target_data_dir}/client-cert.pem', {master}_SSL_KEY='{target_data_dir}/client-key.pem'," if use_ssl else ""} {master}_LOG_FILE='{master_replication['File']}', {master}_LOG_POS={master_replication['Position']}"""
        
        if host_primary and primary:
            commands = [
                f"sudo sed -i '/super_read_only=ON/d' '{conf_path}'",
                f"sudo systemctl restart mysql{appendix}",
            ]
            for command in commands:
                jsn = json.loads(remote_command(host_primary, command, True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and 'error' in error_lower:
                    return jsn
                
            time.sleep(5)
            
            queries = [
                f"STOP {replica}",
                f"RESET {replica} ALL",
                f"DROP USER IF EXISTS '{rep_user}'@'%.%.%.%'",
                "FLUSH PRIVILEGES",
                change_master_command,
                f"START {replica} USER='{rep_user}' PASSWORD='{rep_pwd}' DEFAULT_AUTH='{authentication_plugin}'"
            ]
            conn = get_mysql_connection(host_primary, primary)
            if isinstance(conn, tuple):
                jsn['stderr'] = str(conn[1])
                return jsn
            
            with conn:
                with conn.cursor() as cur:
                    for query in queries:
                        cur.execute(query)

            commands = [
                f"sudo grep -qxF '[mysqld]' '{conf_path}' && grep -qxF 'super_read_only=ON' '{conf_path}' || sudo sed -i '/\[mysqld\]/a super_read_only=ON' '{conf_path}'",
                f"sudo systemctl restart mysql{appendix}",
            ]
            for command in commands:
                jsn = json.loads(remote_command(host_primary, command, True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and 'error' in error_lower:
                    return jsn
               
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity='%s', status='%s', ex_params = ex_params || '{"mysql_master_id":%s}' || '{"rep_user":"%s"}' || '{"rep_pwd":"%s"}', cluster_role='%s' WHERE id=%s;""" % (Severity.OK.value, Status.OK.value, target['id'], rep_user, rep_pwd, ClusterRoles.STANDBY.value, primary['id']))
                    cur.execute("""UPDATE "clusterCore_c123_database" SET node_id=%s WHERE node_id=%s""" % (primary['id'], target['id']))
                    conn.commit()
        
        if has_standby:
            for i in standbys:
                standby_host = i['host']
                standby_data = i['node']
                rep_user = standby_data["ex_params"].get("rep_user")
                rep_pwd = standby_data["ex_params"].get("rep_pwd")
                
                change_master_command = f"""CHANGE {master_alias} TO {master}_HOST='{host_target['name']}', {master}_PORT={target['port']}, {master}_USER='{rep_user}', {master}_PASSWORD='{rep_pwd}', {f"{master}_SSL=1, {master}_SSL_VERIFY_SERVER_CERT=0, {master}_SSL_CAPATH='{target_data_dir}', {master}_SSL_CA='{target_data_dir}/ca.pem', {master}_SSL_CERT='{target_data_dir}/client-cert.pem', {master}_SSL_KEY='{target_data_dir}/client-key.pem'," if use_ssl else ""} {master}_LOG_FILE='{master_replication['File']}', {master}_LOG_POS={master_replication['Position']}"""

                # switchover function was wrong, in mysql unlike the postgresql we made personal replication user for every node
                queries = [
                    f"""CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' {'REQUIRE SSL' if use_ssl else ''}""",
                    f"ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED WITH {authentication_plugin} BY '{rep_pwd}' {'REQUIRE SSL' if use_ssl else ''}",
                    f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
                    f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
                    "FLUSH PRIVILEGES"
                ]

                conn = get_mysql_connection(host_target, target)
                if isinstance(conn, tuple):
                    jsn['stderr'] = str(conn[1])
                    return jsn
                
                with conn:
                    with conn.cursor() as cur:
                        for query in queries:
                            cur.execute(query)
                
                conn = get_mysql_connection(standby_host, standby_data)
                if isinstance(conn, tuple):
                    jsn['stderr'] = str(conn[1])
                    return jsn
                
                queries = [
                    f"STOP {replica}",
                    f"RESET {replica} ALL",
                    change_master_command,
                    f"START {replica} USER = '{rep_user}' PASSWORD = '{rep_pwd}' DEFAULT_AUTH='{authentication_plugin}'"
               ]
                with conn:
                    with conn.cursor() as cur:
                        for query in queries:
                            cur.execute(query)
                
                commands = [
                    f"sudo grep -qxF '[mysqld]' '{conf_path}' && grep -qxF 'super_read_only=ON' '{conf_path}' || sudo sed -i '/\[mysqld\]/a super_read_only=ON' '{conf_path}'",
                    f'sudo systemctl restart mysql{appendix}'
                ]
                for command in commands:
                    jsn = json.loads(remote_command(standby_host, command, True), strict=False)
                    error_lower = jsn['stderr'].lower()
                    if error_lower != '' and 'error' in error_lower:
                        return jsn
                
                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role =%s, ex_params = ex_params || '{"mysql_master_id":%s}', ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.STANDBY.value, target['id'], standby_data['id'],))
                        conn.commit()
        
        return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

def switchover_mysql_node_task(data, task_id):
    """switchover mysql node"""

    task = 'switchover mysql node task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_primary = json.loads(data['host_primary'])
        primary = json.loads(data['primary_data'])
        host_standby = json.loads(data['host_standby'])
        standby = json.loads(data['standby_data'])
        standbys = data.get('standbys')
        has_standby = isinstance(standbys, list) and len(standbys) > 0
        if has_standby:
            standbys = [{"host": json.loads(i["host"]), "node": json.loads(i["node"])} for i in standbys]
        
        distributor, release = check_os_version(host_standby)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'switching over primary'
    duration_step_start = time.perf_counter()
    
    jsn = switchover_node(host_standby, standby, distributor, host_primary, primary, standbys)
    error_lower = jsn['stderr'].lower()
    if error_lower != '' and 'error' in error_lower:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)