# BELONG TO SERVICE TASK
import re, json, time, uuid, logging, os, pytz, traceback
from enum import Enum
from math import inf as positive_infinity
import oracledb
from datetime import datetime, timedelta
from sql import spec_statement, body_statement
from enums import  OperatingSystem, DBType, ClusterRoles, RestoreOptions, Status, Severity, Inventory, Schedule, LicenseType, Repository, Timeout
from tasks_core import insert_task, check_os_version, remote_command, create_ssh_connection, get_postgres_connection, install_offline, logger, find_tilda, init_container, \
      get_date_modified_folder, update_failed_object, transfer_files_rsync, TASK_TMP_FILES_DIR, is_ram_enough, install_offline_remote_to_remote, check_remote_directory_exists

class IncrementalLevel(Enum):
    NO_INCREMENTAL = ''
    CUMULATIVE = 'INCREMENTAL LEVEL 1 CUMULATIVE'
    DIFFERENTIAL = 'INCREMENTAL LEVEL 1'
    BASE = 'INCREMENTAL LEVEL 0'

#####################################################################
# CDB TASKS
#####################################################################

def check_oracle_version_exists(host, oracle_version):
    jsn = json.loads(remote_command(host, f"sudo -iu oracle lsnrctl status | grep -oP 'Version\s+\K\d+' | head -1", True), strict = False)
    return jsn['stdout'] != ''

def register_cdb_task(data, task_id):
    """a task responsible for registering an oracle container database"""
    
    task = 'register CDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()
    
    try:
        cdb = json.loads(data['cdb_data'])
        host = json.loads(data['host_data'])
        # create_run_linux = data['create_run_linux']
        org_id = cdb.get('org_id', 1)
        
        if cdb['dir_123Cluster'].startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(cdb['id'])
                return

            tilda_path = jsn['stdout']
            cdb['dir_123Cluster'] = cdb['dir_123Cluster'].replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # step 0 create the 123cluster directory
    step = 'creating directory for 123cluster(oracle)'
    duration_step_start= time.perf_counter()

    try:
        commands = [
            f"sudo mkdir -p '{cdb['dir_123Cluster']}'",
            f"sudo chown oracle:dba '{cdb['dir_123Cluster']}'"
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(cdb['id'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    # step 5 check instance - can support creating pluggable databases and create the package
    step = 'checking instance - can support creating pdbs'
    duration_step_start= time.perf_counter()

    # establish a connection
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        update_failed_object(cdb['id'])
        return

    try:
        cursor = conn.cursor()
        cursor.execute(spec_statement)
        cursor.execute(body_statement)
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    # step 6 check instance - is type CDB
    # try:
    #     cursor.execute("SELECT status FROM dba_objects WHERE OWNER='SYS' AND OBJECT_NAME='PCK_PDB_MAINT'")
    # except Exception as e:
    #     error, = e.args
    #     step = 'null'
    #     status = Severity.NOT_OK.value
    #     duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #     logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    #     raise Exception('task_id:', task_id, str(error))

    # for idx, val in enumerate(cursor):
    #     if 'INVALID' in val:
    #         msg = 'PDB support is not available with machine {}'.format(cdb)
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    #         raise Exception('task_id:', task_id, msg)

    # create the 123Cluster folder
    step = 'creating directory dir_123cluster'
    duration_step_start= time.perf_counter()
    # try:
    #     cursor.execute("DROP DIRECTORY dir_123cluster")
    # except:
    #     pass

    try:
        cursor.execute("CREATE OR REPLACE DIRECTORY dir_123cluster AS '{}'".format(cdb['dir_123Cluster']))
    except Exception as e:
        error = str(e.args[0])
        if 'ORA-00955' in error:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    else:
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'analyzing CDB size'
    duration_step_start= time.perf_counter()

    try:
        cursor.execute("SELECT SUM(bytes) FROM dba_data_files")
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    else:
        try:
            fetched_cdb_size = cursor.fetchone()[0]
            if fetched_cdb_size is not None:
                cdb_size = float(fetched_cdb_size)/1024
            else:
                cdb_size = 0
        except:
            cdb_size = 0
    finally:
        cursor.close()
        conn.close()

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    '''client = create_ssh_connection(cdb.host, cdb.user, cdb.pwd, cdb.ssh_port)
    if isinstance(client, tuple):
        TaskLogger.error(str(client), task=tpl_id)
        raise Exception(tpl.id, str(client))

    Logger()._info("established connection")
    # run tnsping utility using paramiko to get the standby body
    TaskLogger.info(f"container has {len(aliases)} standby aliases")

    for alias in aliases:
        tns_entry = fetchTNSEntry(client, cdb.oracle_home, alias[0])
        host, port, sid = parseTNSEntry(tns_entry)

        # create the found standby
        osb = OracleStandby.objects.create(
            alias=alias[0],
            host=host,
            port=port,
            service_name=sid,
            database_user=cdb.database_user,
            database_pwd=cdb.database_pwd,
            version=cdb.version,
            cdb=cdb
        )
        cdb.standby = osb
        cdb.save()'''
    
    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, cdb_size, cdb['id']))
                conn.commit()
                # cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,cdb['id'],created_by])
                # conn.commit()
                # cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host['id'],created_by,OperatingSystem.LINUX.value])
                # conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)

    # host['create_task'] = created_by
    # host['ssh_port'] = int(host['ssh_port'])
    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host, 'task': 'run_linux_performance_counters', 'target_type': 'host', 'target_id': host['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as error:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

# def insert_cdb(cdb):
#     try:
#         cdb = json.dumps(cdb) #convert to json string
#         conn = get_postgres_connection()
#         cur = conn.cursor()
#         cur.callproc('insert_cdb', (cdb,))
#         conn.commit()
#         cur.close()
#         conn.close()
#     except Exception as error:
#         raise Exception(err)

# def create_linux_performance_counters(host_data):
#     dir_name = '123cluster'
#     file_content = '''#!/usr/bin/bash 
# #sar 1 1 | grep Average | awk '{print "{\\"cpuusage\\":"u$3+$5",\\"iowait\\":"0+$6",\\"memusage\\":56,\\"memtotal\\":12,\\"cpucount\\":8}"}' 
# mem=`free |grep Mem|awk '{print "\\"memusage\\":"$3 ",\\"memtotal\\":"$2}'` 
# swap=`free |grep Swap|awk '{print "\\"swapusage\\":"$3 ",\\"swaptotal\\":"$2}'` 
# cu=`sar 1 1 | grep Average | awk '{print "\\"cpuusage\\":"u$3+$5",\\"iowait\\":"0+$6""}'` 
# cpuc=\\"cpucount\\":`cat /proc/cpuinfo |grep processor|wc -l` 
# info=`echo {$cu"," $mem"," $cpuc"," $swap}`
# echo $info'''

#     try:
#         jsn = []
#         jsn.append(json.loads(remote_command(host_data, f'mkdir -p /root/{dir_name}', True))) 
#         client = create_ssh_connection(host_data['name'], host_data['username'], host_data['password'], host_data['private_key'], int(host_data['ssh_port']))
#         sftp = client.open_sftp()
#         f = sftp.open(f'/root/{dir_name}/linux_performance_counters.sh', mode='w')
#         f.write(file_content) 
#         f.close()
#         sftp.close()
#         jsn.append(json.loads(remote_command(host_data, f'chmod +x /root/{dir_name}/linux_performance_counters.sh', True)))
#     except Exception as error:
#         print('create_linux_performance_counters Error: ', str(err))

def unregister_cdb_task(data, task_id):
    """unregister an oracle container database"""
    
    task = 'unregister CDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()
    
    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        host = data['host_data']
        cdb_data = json.loads(data['cdb_data'])
        if cdb_data['dir_123Cluster'].startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            cdb_data['dir_123Cluster'] = cdb_data['dir_123Cluster'].replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # Drop the 123Cluster folder
    step = 'dropping directory'
    duration_step_start= time.perf_counter()
    
    # establish a oracle connection
    error, conn = get_cxoracle_connection(cdb=cdb_data, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return

    try:
        cursor = conn.cursor()
        cursor.execute("DROP DIRECTORY '{}'".format(cdb_data['dir_123Cluster']))
    except Exception as e:
        error = str(e.args[0])
        if 'ORA-22929' in error:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    else:
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    finally:
        cursor.close()
        conn.close()

    step = 'unregistering CDB from app'
    duration_step_start= time.perf_counter()

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, cdb_data['id']))
                conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) 
                  
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

####################################################################################################################

'''@shared_task
def fetchTNSEntry(connection, oracle_home, alias):
    """retrieve the entry using tns ping"""
    command_string = f"export ORACLE_HOME={oracle_home};{oracle_home}/bin/tnsping {alias}"
    jsn = json.loads(remote_command(
        connection, command_string, True), strict=False)
    return jsn['stdout']

@shared_task
def parseTNSEntry(entry):
    """parse the TNS entry from a string"""
    host_s = entry.find("HOST = ") + 7
    host_e = entry[host_s:].find(")")
    host = entry[host_s:host_s+host_e]

    port_s = entry.find("PORT = ") + 7
    port_e = entry[port_s:].find(")")
    port = entry[port_s:port_s+port_e]

    sid_s = entry.find("SID = ") + 6
    sid_e = entry[sid_s:].find(")")
    sid = entry[sid_s:sid_s+sid_e]
    return (host, port, sid)
'''

def duplicate_oracle_standalone_task(data, task_id):
    """duplicate_oracle_standalone_task is a task to duplicate a
    standalone database into a pluggable database"""

    task = 'duplicate oracle standalone task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        cdb = json.loads(data['cdb_data']) 
        host = data['host_data']
        target_PDB = data['target_PDB']
        file_dest = data['file_dest']
        start_after_copy = data['start_after_copy']
        duplicated_sa_id = data['duplicated_sa_id']
        # standalone user, should be sysdba, but not user sys -
        # only that way could register task by current duplicate standalone view...
        target_db_user = data['target_db_user']
        target_db_pwd = data['target_db_pwd']
        link_name = data['link_name']
        duplicate_sql_script_path = '/tmp/duplicate_sql_script.sh'
        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        if file_dest.startswith('~'):
            tilda_path = jsn['stdout']
            file_dest = file_dest.replace('~', tilda_path)

        if cdb['dir_123Cluster'].startswith('~'):
            tilda_path = jsn['stdout']
            cdb['dir_123Cluster'] = cdb['dir_123Cluster'].replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    # Step 9 - copy the database
    step = 'creating PDB from link'
    duration_step_start= time.perf_counter()
    
    # connect to target container
    error, conn2 = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn2,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn2,1)
        return

    try:
        cursor = conn2.cursor()
        cursor.execute("CREATE PLUGGABLE DATABASE {0} FROM NON$CDB@{1} USER_TABLESPACES=ALL CREATE_FILE_DEST='{2}' standbys=none".format(target_PDB, link_name, file_dest))
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # Step 10 - drop the link
    step = 'dropping database link'
    duration_step_start= time.perf_counter()
    
    try:
        cursor.execute(f"DROP DATABASE LINK {link_name}")
    except Exception as e:
        error, = e.args
        if 'ORA-04054' not in str(error):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

   
    step = 'running sql script'
    duration_step_start= time.perf_counter()

    try:
        cursor.execute("SELECT SYS_CONTEXT ('USERENV','ORACLE_HOME') FROM DUAL")
        res = cursor.fetchone()
        # script may run from several tens of minutes to several hours - depending on how much non-cdb are populated, so close connection
        cursor.close()
        conn2.close()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    if not res:
        error = f"could not obtain the oracle home of target cdb {cdb['name']}"
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return
    
    oracle_home = res[0]
    
    # sqlplus scripts must end with explicit exit, otherwise - may stuck
    file_content = f'''export ORACLE_HOME={oracle_home}\nexport ORACLE_SID={cdb['name']}\n{oracle_home}/bin/sqlplus / as sysdba<<EOF
ALTER SESSION SET CONTAINER={target_PDB};
@{oracle_home}/rdbms/admin/noncdb_to_pdb.sql
ALTER PLUGGABLE DATABASE {target_PDB} OPEN;
CREATE OR REPLACE DIRECTORY dir_123cluster AS '{cdb['dir_123Cluster']}';
GRANT DBA, CONNECT, RESOURCE, CREATE PLUGGABLE DATABASE TO {target_db_user};
exit;
EOF'''

    try:
        jsn = json.loads(remote_command(host, f"""echo "{file_content}" | sudo tee '{duplicate_sql_script_path}'""", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    try:
        jsn = json.loads(remote_command(host, f"sudo chmod +x '{duplicate_sql_script_path}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        command = f"sudo -iu oracle '{duplicate_sql_script_path}'"
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        elif int(jsn['excode']) != 0:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])

        jsn = json.loads(remote_command(host, f"sudo rm -f '{duplicate_sql_script_path}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    if status == Severity.WARNING.value:
        pass
    else:
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)


    # Step 11 - reopen with specified mode
    if start_after_copy != "nothing":
        step = 'reopenning PDB'
        duration_step_start= time.perf_counter()

        error, conn2 = get_cxoracle_connection(cdb=cdb, host=host)
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn2,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn2,1)
            return

        if start_after_copy == "openre":
            try:
                cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE".format(target_PDB))
                target_open_query = "ALTER PLUGGABLE DATABASE {} OPEN READ ONLY".format(target_PDB) 
                cursor.execute(target_open_query)
            except Exception as e:
                error, = e.args
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
        
        cursor.close()
        conn2.close()

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn_postgres:
            with conn_postgres.cursor() as cur_post:
                if start_after_copy != "nothing":
                    cur_post.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_deleted='false', ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, duplicated_sa_id))
                else:
                    cur_post.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, duplicated_sa_id))
                conn_postgres.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def backup_cdb_task(data, task_id):
    """backup an Oracle Container database task"""

    task = 'backup CDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        cdb = json.loads(data['cdb_data'])
        standalone = cdb['ex_params'].get('standalone')
        host = data['host_data']
        backup_host = data['backup_host']
        backup = json.loads(data['backup_data'])
        backup_option = data['options']
        now = datetime.utcnow().strftime('%d%m%Y%H%M%S')
        #must not contain special characters
        backup_path = backup['ex_params'].get('backup_path')
        schedule = backup['is_sched']
        compress = backup['ex_params'].get('compress')
        # options dropdown
        incremental_level = backup['ex_params'].get('incremental_level').strip().upper()
        if backup_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)
        
        if schedule:
            if incremental_level == IncrementalLevel.NO_INCREMENTAL.value:
                backup_path = f'{backup_path}/{now}'
            # if incremental are used - all the subsequent backups must go to the same backup path
            else:
                backup_path = f'{backup_path}/incremental'

        datadir = cdb['dir_123Cluster']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating backup directory'
    duration_step_start= time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo mkdir -p '{backup_path}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        if schedule and incremental_level != IncrementalLevel.NO_INCREMENTAL.value:
            #check if incremental backup ever completed successfully
            jsn = check_remote_directory_exists(backup_host, backup_path, 'd', is_dir_empty=True)
            if jsn['stderr'] != '' and 'no such' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
            backup_folder_empty = True if 'true' in jsn['stdout'] else False
            #if incremental folder does not contain files
            if backup_folder_empty:
                #default to level 0 backup
                incremental_level = IncrementalLevel.BASE.value
            else:
                #download all the existing incremental backup files back to node host
                jsn = transfer_files_rsync(backup_host, host, backup_path, [f"{backup_path}/"])
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return

    try:
        cursor = conn.cursor()
        cursor.execute("SELECT version FROM v$instance")
        major_version = int(cursor.fetchone()[0].split('.')[0])
        if major_version in [18, 19]:
            #for all pdbs (including PDB$SEED) - set empty db_create_file_dest
            cursor.execute("SELECT name FROM v$pdbs")
            pdb_names = [i[0] for i in cursor.fetchall()]
            pdb_file_dests = {}
            for i in pdb_names:
                cursor.execute('ALTER SESSION SET CONTAINER={}'.format(i))
                cursor.execute("SELECT value FROM v$parameter WHERE name='db_create_file_dest'")
                pdb_file_dests[i] = cursor.fetchone()[0]
                cursor.execute("ALTER SYSTEM SET db_create_file_dest='' scope=both")
                conn.commit()
        
        cursor.close()
        conn.close()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'creating RMAN script'
    duration_step_start= time.perf_counter()

    # 3. create the rman script
    rman_script_file = backup_path + "/rman_" + \
    cdb['name'] + "_" + now + ".rman"
    logfile_name = cdb['name'] + "_" + now + ".log"
    log_file = backup_path + "/" + logfile_name
    backup_full_path = f"{backup_path}/%U"

    line1 = "ALLOCATE CHANNEL c1 TYPE DISK FORMAT '{}';".format(backup_full_path)
    # backup by option
    # line2 = "BACKUP DATABASE PLUS ARCHIVELOG;"
    line2 = f"BACKUP {'AS COMPRESSED BACKUPSET' if compress else ''} {incremental_level} DATABASE ROOT;" if not standalone and backup_option != 'all' else f"BACKUP {'AS COMPRESSED BACKUPSET' if compress else ''} {incremental_level} DATABASE;"

    try:
        commands = [
            f"""echo -e "run {{\n{line1}\n{line2}\nRELEASE CHANNEL c1;}}" | sudo tee '{rman_script_file}'""",
            f"sudo chown -R oracle:dba '{backup_path}'",
            f"sudo chmod -R 755 '{backup_path}'",
            f"sudo chown -R oracle:dba '{datadir}'",
            f"sudo chmod -R 755 '{datadir}'",
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    # 4. run the script
    step = 'running the RMAN script'
    duration_step_start= time.perf_counter()
    
    try:
        # command = "export ORACLE_HOME={0};{0}/bin/rman target=sys/{1}@{2}:{3}/{4} @{5} log={6}".format(cdb['ex_params'].get('oracle_home'), cdb['database_pwd'], host['name'], cdb['port'], cdb['name'], rman_script_file, log_file)
        # command = "sudo -u oracle ORACLE_HOME={0} {0}/bin/rman target='sys/{1}@{2}:{3}/{4}' @{5} log={6}".format(cdb['ex_params'].get('oracle_home'), cdb['database_pwd'], host['name'], cdb['port'], cdb['name'], rman_script_file, log_file)
        # borrowed this from backup pdb task - also works and does not expose the password
        command = "sudo -iu oracle ORACLE_HOME={0} ORACLE_SID={1} {0}/bin/rman target=/ @{2} log={3}".format(cdb['ex_params'].get('oracle_home'), cdb['name'], rman_script_file, log_file)
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    if major_version in [18, 19]:
        error, conn = get_cxoracle_connection(cdb=cdb, host=host)
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
            return

        try:
            cursor = conn.cursor()
            for i in pdb_names:
                cursor.execute("ALTER SESSION SET CONTAINER={}".format(i))
                cursor.execute("ALTER SYSTEM SET db_create_file_dest='{}'".format(pdb_file_dests[i]))
                conn.commit()
            cursor.close()
            conn.close()
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # 5. read the log file to see if there where any errors
    step = 'reading the log file'
    duration_step_start= time.perf_counter()
    
    try:
        stdout = "The backup failed. Please check the log file {} for details.".format(log_file)
        jsn = json.loads(remote_command(host, f"sudo cat '{log_file}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        backup['run_log'] = jsn['stdout']
        with get_postgres_connection() as conn_postgres:
            with conn_postgres.cursor() as cur_post:
                cur_post.callproc('set_logfile_data', (Inventory.BACKUP.value,backup['run_log'],backup['id']))
                conn_postgres.commit()
                # search for errors in the backup log
                for line in backup['run_log'].splitlines():
                    if 'RMAN-' in line:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,line,None)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,line,None)
                        return
                
                # backup is successful
                backup['status'] = True
                cur_post.execute("""UPDATE "clusterCore_c123_inventory" SET status={} WHERE id={}""".format(backup['status'], backup['id']))
                conn_postgres.commit()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    if host['name'] != backup_host['name']:
        step = 'copying backup of CDB to backup host'
        duration_step_start = time.perf_counter()
        
        jsn = transfer_files_rsync(host, backup_host, backup_path, [f"{backup_path}/"])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'getting the backup size'
    duration_step_start= time.perf_counter()

    try:
        # 6. get the backup size
        jsn = json.loads(remote_command(backup_host, f"sudo du -s '{backup_path}' | cut -f1", True))
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        modified_date = get_date_modified_folder(hostdata=backup_host, folder_path=backup_path) 
        if isinstance(modified_date, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return
        
        # backup['ex_params']['size'] = round(int(jsn['stdout'])/1024/1024, 2)
        backup['ex_params']['size'] = float(jsn['stdout'].strip())/1024
        backup['ex_params']['size'] = str(backup['ex_params'].get('size'))
        # if we need update both before and after transfer - for after transfer need new connection
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true) WHERE id={1}""".format(backup['ex_params'].get('size'), backup['id']))
                #relate to size - if have unexpected format in modified_date = something wrong with backup file
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}', status='true' WHERE id=%s""" % (modified_date, backup['id']))
                #if backup data are ok - assume that node ok
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, cdb['id']))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{exit_code}}', '{0}', true) WHERE id={1}""".format(0, backup['id']))
                conn.commit()
                remote_command(host, f"sudo rm -rf '{backup_path}'", True)

                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

                step = 'task finished'
                status = Severity.OK.value
                duration = 0
                logger(task_id,task_log_id,task,step,status,duration,created_by)
                step = 'null'
                status = Severity.OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

                try:
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{duration}}', '{0}', true) WHERE id={1}""".format(duration_task_end, backup['id']))
                    conn.commit()
                except Exception as duration_exc: #notice-level
                    error, = duration_exc.args
                    logging.error(f"{task}@{step}: {task_id}: {str(error)}")    
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

'''
@shared_task(bind=True)
def restore_cdb_task_notify(self, cdb_data, tpl_id):
    """restore a CDB task"""
    # link the task to the logger
    # tpl = task_from_json(tpl_data)
    tpl = TaskProgressLog.objects.get(pk=tpl_id)
    tpl.celery_task_id = self.request.id
    tpl.state = AsyncResult(self.request.id).state
    tpl.save()
    tpl_data = object_to_json(tpl)

    TaskLogger = Logger(task_ident=tpl.id)

    # extract the args
    task_args = eval(tpl.args)
    buid = task_args['bu_id']
    autostart = task_args['autostart']

    # deserialize the objects
    cdb = node_from_json(cdb_data, 'CDBInstance')
    backup = NodeBackup.objects.get(pk=buid)

    TaskLogger.info("starting restore {}".format(backup))

    # create a unique string with the date
    now = timezone.now().strftime('%d%m%Y%H%M%S')

    # 1. connect using paramiko
    client = create_ssh_connection(cdb.host, cdb.user, cdb.pwd, cdb.ssh_port)
    if isinstance(client, tuple):
        TaskLogger.error(str(client))
        raise Exception(tpl.id, str(client))

    TaskLogger.info("established connection")
    # 2. create the rman restore script
    restore_path = backup.backup_path + "/restore_" + cdb.service_name + "_" + now
    rman_restore_script_path = restore_path + ".rman"
    logfile_name = cdb.service_name + "_restore_" + now + ".log"
    logfile_path = backup.backup_path + "/" + logfile_name
    backup_full_path = "{}/%U".format(backup.backup_path)

    # create the rman script line by line
    line1 = "ALLOCATE CHANNEL c1 TYPE DISK FORMAT '{}/%U';".format(
        restore_path)
    line2 = "RESTORE PLUGGABLE DATABASE {};".format(cdb.service_name)
    line3 = "RECOVER PLUGGABLE DATABASE {};".format(cdb.service_name)
    line4 = "ALTER PLUGGABLE DATABASE {} OPEN;".format(cdb.service_name)

    # 3. create the rman script
    jsn = json.loads(remote_command(
        client, "echo -e \"run {\n" + line1 + "\n" + line2 + "\n" + line3 + "\n" + line4 + "\n}\" | sudo tee " + rman_restore_script_path, True), strict=False)

    if jsn['stderr'] != '':
        print(jsn['stderr'])
        raise Exception(tpl.id, jsn['stderr'])
    else:
        TaskLogger.info("created the rman script")

    # 4. run the script
    jsn = json.loads(remote_command(
        client, "export ORACLE_HOME={0};{1}/bin/rman target=sys/{2}@{3}:{4}/{5} @{6} log={7}".format(cdb.oracle_home, cdb.oracle_home, cdb.pwd, cdb.host, cdb.port, cdb.service_name, rman_restore_script_path, logfile_path), True), strict=False)

    print("export oracle home {}".format(jsn))
    if jsn['stderr'] != '':
        print(jsn['stderr'])
        raise Exception(tpl.id, jsn['stderr'])
    else:
        TaskLogger.info("backup script executed")

    # 4. read the script log
    sftp = client.open_sftp()
    remote_file = sftp.open(logfile_path)

    TaskLogger.info("reading log..")

    # search for errors in the backup log
    try:
        for line in remote_file:
            if 'RMAN' in line:
                TaskLogger.error(line)
                raise Exception()
    except:
        raise Exception(tpl.id, "unfortunately restore didn't succeed, please see {}".format(
            logfile_path))
    finally:
        remote_file.close()
        sftp.close()
        client.close()

    return True

'''
def create_startup_triggers_task(host, cdb_data, pdb_data, task, task_id, task_log_id, created_by):
    """a task to recreate the container startup trigger once a
    pluggable database has the property ind_autostartup set to True"""
    
    step = 'starting create_startup_triggers_task'
    duration_step_start= time.perf_counter()
    
    error = ''
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""SELECT name FROM "clusterCore_c123_node" WHERE ind_autostartup='true' AND type=%s""", (DBType.ORACLE_PDB.value,))
                startup_pdbs = cur.fetchall()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        return str(error)

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # deserialize the objects
    cdb = cdb_data
    pdb = pdb_data
    
    # beggining part of the script
    trigger_sql = "CREATE OR REPLACE TRIGGER SYS.OPEN_PDBS \
                    AFTER STARTUP ON DATABASE \
                    BEGIN \
                    "
    # append the middle part of the script with the pdb names
    for pdb in startup_pdbs:
        trigger_sql = trigger_sql + \
            "EXECUTE IMMEDIATE 'ALTER PLUGGABLE DATABASE {} OPEN'".format(pdb)

    # append the final part of the script
    trigger_sql = trigger_sql + " END open_pdbs"

    step = 'connecting to cdb and execute the trigger query'
    duration_step_start= time.perf_counter()

    # connect to the cdb
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        return str(conn)
    else:
        error = ''

    # execute the trigger query
    try:
        cursor = conn.cursor()
        cursor.execute(trigger_sql)
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        return str(error)

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # check if the trigger was created
    step = 'checking if the trigger was created'
    duration_step_start= time.perf_counter()

    try:
        cursor.execute("SELECT STATUS FROM dba_triggers WHERE trigger_name='OPEN_PDBS'")
        result = cursor.fetchone()
        if not result or 'ENABLED' not in result[0]:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,None)
            return "{} trigger execution failed".format(cdb)
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        return str(error)
    
    cursor.close()
    conn.close()

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    return error

####################################################################
# PDB TASKS
####################################################################

def get_cluster_sizing_per_size(size):
    """return a dict with oracle cluster sizing dict"""
    # ocs = OracleClusterSizing()
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute('SELECT * FROM "clusterCore_oracleclustersizing"')
                ocs = cur.fetchall()
                cur.execute("""SELECT * FROM "clusterCore_clustersizing" WHERE db_type=%s""", (DBType.ORACLE_CDB.value,))
                cs = cur.fetchall() 
    except Exception as e:
        error, = e.args
        return (False, str(error))

    pdb_size_scheme = {
        'SMALL': {
            'cpu': cs[0][2],
            'disk': cs[0][3],
            'memory': cs[0][4],
            'shared_temp': ocs[0][1],
            'backup': cs[0][5],
            'restore': ocs[0][5],
            'export': cs[0][6],
        },
        'MEDIUM': {
            'cpu': cs[0][7],
            'disk': cs[0][8],
            'memory': cs[0][9],
            'shared_temp': ocs[0][2],
            'backup': cs[0][10],
            'restore': ocs[0][6],
            'export': cs[0][11],
        },
        'LARGE': {
            'cpu': cs[0][12],
            'disk': cs[0][13],
            'memory': cs[0][14],
            'shared_temp': ocs[0][3],
            'backup': cs[0][15],
            'restore': ocs[0][7],
            'export': cs[0][16],
        },
        'VERY_LARGE': {
            'cpu': cs[0][17],
            'disk': cs[0][18],
            'memory': cs[0][19],
            'shared_temp': ocs[0][4],
            'backup': cs[0][20],
            'restore': ocs[0][8],
            'export': cs[0][21],
        },
    }
    
    return pdb_size_scheme[size]

def create_pdb_task(data, task_id):
    """a task that creates a pluggable database in a container"""
    
    task = 'create PDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        cdb = json.loads(data['cdb_data'])
        pdb = json.loads(data['pdb_data'])
        host = json.loads(data['host_data'])
        cluster_size = get_cluster_sizing_per_size(pdb['sizing'])
        if isinstance(cluster_size, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(cluster_size[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(cluster_size[1]),1)
            return

        pdb_ind_started = True

        if pdb['ex_params'].get('file_dest').startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            pdb['ex_params']['file_dest'] = pdb['ex_params'].get('file_dest').replace('~', tilda_path)

        # create the destination folder specified by the user
        path = pdb['ex_params'].get('file_dest') + "/" + pdb['name']
        re.sub("//", "/", path) # in case an extra / was added by the user
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating file destination'
    duration_step_start= time.perf_counter()

    try:
        commands = [
            "sudo mkdir -p '{}'".format(path),
            f"sudo chown -R oracle:dba '{pdb['ex_params'].get('file_dest')}'"
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating the PDB'
    duration_step_start= time.perf_counter()

    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return 

    # prepare the query for pdb creation
    query = f"CREATE PLUGGABLE DATABASE {pdb['name']} ADMIN USER {pdb['database_user']} IDENTIFIED BY {pdb['database_pwd']} CREATE_FILE_DEST='{path}'"
    # append standby if selected
    # if args['standby'] == 'all':
    #     query = f"{query} STANDBYS=ALL"
    #     TaskLogger.info("adding standby's all")
    # elif args['standby'] == 'none':
    query = f"{query} STANDBYS=NONE"
    # else:
    #     standby = OracleStandby.objects.get(pk=args['standby'])
    #     query = query + \
    #         f" STANDBYS=('{standby.alias}')"
    #     TaskLogger.info(f"adding standby{standby}")

    # create the database
    try:
        cursor = conn.cursor()
        cursor.execute(query)
    except Exception as e:
        # if there was an exception print it
        logging.debug("CREATE PLUGGABLE DATABASE {} ADMIN USER {} IDENTIFIED BY {} CREATE_FILE_DEST='{}'".format(pdb['name'], pdb['database_user'], pdb['database_pwd'], path))
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'openning and setting the PDB'
    duration_step_start= time.perf_counter()
    
    try:
        # open the pdb
        cursor.execute("ALTER PLUGGABLE DATABASE {} OPEN".format(pdb['name']))
        # set as the session
        cursor.execute("ALTER SESSION SET CONTAINER={}".format(pdb['name']))
        # grant dba to database user
        cursor.execute("GRANT DBA, CONNECT, RESOURCE, CREATE PLUGGABLE DATABASE TO {}".format(pdb['database_user']))
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    # try:
    #     oracle_cluster_size = OracleClusterSizing()
    # except Exception as e:
    #     error, = e.args

    # set disk limit according to size scheme
    # cluster_size_commands = [
    #     # "ALTER PLUGGABLE DATABASE {} OPEN".format(pdb['name']),
    #     f"ALTER SESSION SET CONTAINER={pdb['name']}",
    #     f"ALTER PLUGGABLE DATABASE {pdb['name']} STORAGE (MAXSIZE {cluster_size['disk']}G)",
    #     f"ALTER SYSTEM SET CPU_COUNT={cluster_size['cpu']} SCOPE=BOTH",
    #     f"ALTER PLUGGABLE DATABASE STORAGE (MAX_SHARED_TEMP_SIZE {cluster_size['shared_temp']}G)"
    # ]
    # step = 'execute size commands'
    # duration_step_start= time.perf_counter()
    # status = Severity.OK.value
    # # execute size commands
    # try:
    #     cursor.execute(cluster_size_commands[0])
    # except Exception as e:
    #     error, = e.args
    #     status = Severity.NOT_OK.value
    # try:
    #     cursor.execute(cluster_size_commands[1])
    # except Exception as e:
    #     error, = e.args
    #     status = Severity.NOT_OK.value
    # try:
    #     cursor.execute(cluster_size_commands[2])
    # except Exception as e:
    #     error, = e.args
    #     status = Severity.NOT_OK.value
    # try:
    #     cursor.execute(cluster_size_commands[3])
    # except Exception as e:
    #     error, = e.args
    #     status = Severity.NOT_OK.value
    # duration_step_end = time.perf_counter() - duration_step_start
    # logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    error = config_pdb(host, cdb, pdb, cluster_size, task, task_id, task_log_id, created_by)
    if error != '':
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'creating dir_123cluster folder'
    duration_step_start= time.perf_counter()

    if cdb['dir_123Cluster'].startswith('~'):
        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        tilda_path = jsn['stdout']
        cdb['dir_123Cluster'] = cdb['dir_123Cluster'].replace('~', tilda_path)
    
    # create the 123Cluster folder
    try:
        cursor.execute("CREATE DIRECTORY dir_123cluster AS '{}'".format(cdb['dir_123Cluster']))
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    else:
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'measuring PDB size'
    duration_step_start= time.perf_counter()

    try:
        cursor.execute(
            f"SELECT p.PDB_NAME \"Pluggable database name\", \
            DECODE(SUM(d.BYTES), NULL, 'Not Available since the PDB is down', (SUM(d.BYTES))) \"Size(B)\" \
            FROM DBA_PDBS p, CDB_DATA_FILES d \
            WHERE p.PDB_ID=d.CON_ID(+) AND p.PDB_NAME='{pdb['name']}' \
            GROUP BY p.PDB_ID, p.PDB_NAME \
            ORDER BY SUM(d.BYTES) DESC")
    except Exception as e:
        error,= e.args
        if "ORA-01219" in str(error):
            pdb_ind_started = False
        else:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    else:
        try:
            fetched_pdb_size = cursor.fetchone()[1]
            if fetched_pdb_size is not None:
                pdb_size = float(fetched_pdb_size)/1024
            else:
                pdb_size = 0
        except Exception as error:
            logging.error(f"{task}@{step}: {task_id}: {str(error)}")
            pdb_size = 0

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # call the auto startup subtask
    # if pdb['is_autostartup']:
    #     # TaskLogger.info("setting auto-start up")
    #     SetAutoStartPDB.delay(cdb['id'])

    # # call the auto delete subtask
    # if pdb['auto_delete'] is not None:
    #     # TaskLogger.info("setting auto-delete on {}".format(pdb.auto_delete))
    #     SetAutoDelete.delay(cdb_data, pdb_data)

    if pdb['ind_autostartup']:
        error = create_startup_triggers_task(host, cdb, pdb, task, task_id, task_log_id, created_by)
        if error != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    cursor.close()
    conn.close()

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started=%s, ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, pdb_ind_started, pdb_size, pdb['id']))
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def open_pdb_task(data, task_id):
    """open a pluggable database task"""

    task = 'open PDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        cdb = json.loads(data['cdb_data'])
        pdb_id = data['pdb_id']
        host = data['host_data']
        open_query = data['open_query']
        # pdb_name = data['pdb_name']
        # option = data['option']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'openning PDB'
    duration_step_start= time.perf_counter()

    # step 1 establish a connection
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return

    # duration_step_start= time.perf_counter()
    # step = 'check the database current open mode'
    
    # status = Severity.OK.value
    # # Check the database current open mode:
    # try:
    #     cursor.execute("Select OPEN_MODE from v$PDBS where name = '{}'".format(pdb_name))
    # except Exception as e:
    #     error, = e.args
    #     status = Severity.NOT_OK.value
    #     duration_step_end = time.perf_counter() - duration_step_start
    #     logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    #     step = 'null'
    #     status = Severity.NOT_OK.value
    #     duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #     logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    #     raise Exception('task_id:', task_id, str(error))
    # else:
    #     open_mode = cursor.fetchone()[0]
    # duration_step_end = time.perf_counter() - duration_step_start
    # logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # open_query = "alter pluggable database {} open ".format(pdb_name)
    # # the database is already open in the requested mode
    # if open_mode == 'READ WRITE':
    #     if option == 'RW':
    #         already_open_msg = "pluggable database {} already open please close it first".format(pdb_name)
    #         print(already_open_msg)
    #         # TaskLogger.warning(already_open_msg)
    #         step = 'null'
    #         status = 'orange'
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    #         return 
    #     elif option == 'RO':
    #         open_query += "READ ONLY"
    # elif open_mode == 'READ ONLY':
    #     if option == 'RO':
    #         already_open_msg = "pluggable database {} already open please close it first".format(pdb_name)
    #         print(already_open_msg)
    #         # TaskLogger.warning(already_open_msg)
    #         step = 'null'
    #         status = 'orange'
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    #         return 
    # elif open_mode == 'MOUNTED':
    #     if option == 'RO':
    #         open_query += "READ ONLY"

    try:
        cursor = conn.cursor()
        cursor.execute(open_query)
    except Exception as e:
        error, = e.args
        error = str(error)
        if error.find('ORA-65019') != -1:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    cursor.close()
    conn.close()

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, pdb_id))
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

    # open in replication section
    # try:
    #     pdb = PDB_Instance.objects.get(pk=args['obj_id'])
    # except:
    #     TaskLogger.warning("unable to get PDB object for replication autostarts")
    # else:
    #     if pdb.also_open_standbys:
    #         TaskLogger.info("calling open database standbys task")
    #         open_pdb_standbys_task.delay(args['obj_id'])


def cdb_parameters_apply_task(data, task_id):
    """edit CDB parameters task"""

    task = 'CDB parameters apply task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        cdb = json.loads(data['cdb_data'])
        host = json.loads(data['host_data'])
        parameters = data['parameters']
        require_restart = any(p for p in parameters if p['ISSYS_MODIFIABLE'] == 'FALSE')
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'applying CDB settings'
    duration_step_start= time.perf_counter()

    try:
        error, conn = get_cxoracle_connection(cdb=cdb, host=host)
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
            return
        with conn:
            with conn.cursor() as cur:
                for p in parameters:
                    query = f"""ALTER SYSTEM SET {p['NAME']} = {p['VALUE']} {'SCOPE=SPFILE' if p['ISSYS_MODIFIABLE'] == 'FALSE' else ''}"""
                    print(query)
                    cur.execute(query)
    except Exception as e:
        print(traceback.format_exc())
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    if require_restart:
        cdb_restart_script = """SHUTDOWN IMMEDIATE;
STARTUP MOUNT;
ALTER DATABASE OPEN;
exit;        
"""
        jsn = json.loads(remote_command(host, f"echo '{cdb_restart_script}' | sudo -iu oracle sqlplus / as sysdba",True), strict = False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, cdb['id']))
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)


def install_cdb_task(data, task_id):
    """install a container database task"""

    task = 'install CDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        cdb = json.loads(data['cdb_data'])
        host = json.loads(data['host_data'])
        app_host = json.loads(data['app_host'])
        proxmox_container = cdb['ex_params'].get('proxmox')
        user_sys_pwd = cdb["ex_params"].get("user_sys_pwd")
        user_system_pwd = cdb["ex_params"].get("user_system_pwd")
        database_user = cdb["database_user"]
        database_pwd = cdb["database_pwd"]
        cdb_name = cdb['name'].strip()
        cdb_port = cdb['port'].strip()
        listener = f"{cdb_name}:{cdb_port}"
        cdb_id = cdb['id']
        # datadir must not end with cdb sid! - throws ORA-19800 (undocumented issue of dbca?) and must not contain special characters
        datadir = cdb["dir_123Cluster"].strip()
        if os.path.basename(datadir).upper() == cdb_name.upper():
            datadir = os.path.dirname(datadir)
        oracle_version = cdb['version']
        repository = data['repository']
        ram_mb = cdb["ex_params"].get("ram_mb")
        redo_log_mb = cdb["ex_params"].get("redo_log_mb")
        flashback_keep_days_minutes = cdb["ex_params"].get("flashback_keep_days_minutes")
        standalone = cdb["ex_params"].get("standalone")
        create_run_linux = data['create_run_linux']
        response_file_dir = f'{TASK_TMP_FILES_DIR}/oracle/{cdb_name}'
        response_file = f'{response_file_dir}/createdb.rsp'
        version_exists = check_oracle_version_exists(host, oracle_version)
        #less change to mess up when having our custom response files than copy bundled template-add/edit parameters-remove conficting parameters
        response_files = {'19': f"""
##############################################################################
##                                                                          ##
##                            DBCA response file                            ##
##                            ------------------                            ##
## Copyright(c) Oracle Corporation 1998,2019. All rights reserved.         ##
##                                                                          ##
## Specify values for the variables listed below to customize 			    ##
## your installation.                                         			    ##
##                                                            			    ##                                                               			##
## IMPORTANT NOTE: This file contains plain text passwords and   			##
## should be secured to have read permission only by oracle user 			##
## or db administrator who owns this installation.               			##
##############################################################################

#-------------------------------------------------------------------------------
# Do not change the following system generated value. 
#-------------------------------------------------------------------------------
responseFileVersion=/oracle/assistants/rspfmt_dbca_response_schema_v19.0.0

#-------------------------------------------------------------------------------
# Database configuration parameters
#-------------------------------------------------------------------------------
gdbName={cdb_name}
sid={cdb_name}
templateName=General_Purpose.dbc
createAsContainerDatabase={"false" if standalone else "true"}
databaseType=MULTIPURPOSE

#-------------------------------------------------------------------------------
# Memory and storage parameters
#-------------------------------------------------------------------------------
memoryMgmtType=AUTO_SGA
totalMemory={ram_mb}
storageType=FS
datafileDestination={datadir}
recoveryAreaDestination={datadir}
recoveryAreaSize=10240

#-------------------------------------------------------------------------------
# Character set parameters
#-------------------------------------------------------------------------------
characterSet=AL32UTF8

#-------------------------------------------------------------------------------
# Administrative passwords
#-------------------------------------------------------------------------------
sysPassword={user_sys_pwd}
systemPassword={user_system_pwd}

#-------------------------------------------------------------------------------
# Network parameters
#-------------------------------------------------------------------------------
createListener={listener}

#-------------------------------------------------------------------------------
# Log and archive parameters
#-------------------------------------------------------------------------------
enableArchive=true
archiveLogDest={datadir}
redoLogFileSize={redo_log_mb}

#-------------------------------------------------------------------------------
# Enterprise Manager (EM) configuration
#-------------------------------------------------------------------------------
emConfiguration=NONE

#-------------------------------------------------------------------------------
# Miscellaneous parameters
#-------------------------------------------------------------------------------
ignorePreReqs=true
"""}
        
        if proxmox_container:
            jsn = json.loads(remote_command(host, "echo 1", True), strict=False)
            if jsn['stderr'] != '':
                host_exists = False
            else:
                host_exists = True



    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            distributor = cdb["ex_params"].get("distributor")
            release = cdb["ex_params"].get("release")
            jsn = init_container(service_data, host, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(cdb['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    step = 'checking minimum requirements'
    duration_step_start = time.perf_counter()

    is_error, enough = is_ram_enough(host, ram_mb)
    if not enough or is_error:
        error = enough if is_error else f"Available RAM are less than required {ram_mb} MB"
        excode = 9 if is_error else 1
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,excode)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,excode)
        update_failed_object(cdb_id)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)



    if not version_exists:
        step = 'installing oracle software'
        duration_step_start= time.perf_counter()
    
        jsn = install_oracle(app_host, host, oracle_version, repository)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(cdb_id)
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'initializing CDB'
    duration_step_start= time.perf_counter()
    
    #rpm install of redhat has strict path standards,enough to find oracle home when no other means
    jsn = json.loads(remote_command(host, "sudo find '/opt' -type d -name 'dbhome_*'", True), strict=False)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        update_failed_object(cdb_id)
        return
    
    oracle_home = jsn['stdout']
    user_oracle_profile = "/home/oracle/.bash_profile"
    environment = f"""export ORACLE_HOME={oracle_home}
export PATH=$PATH:$ORACLE_HOME/bin
export LD_LIBRARY_PATH=$ORACLE_HOME/lib:/lib:/usr/lib
export CLASSPATH=$ORACLE_HOME/jlib:$ORACLE_HOME/rdbms/jlib
export ORACLE_UNQNAME={cdb_name}
export ORACLE_SID={cdb_name}"""
    
    commands = [
        f"sudo mkdir -p '{datadir}'",
        # for initialization - datadir must belong to oracle:oinstall
        f"sudo chown -R oracle:oinstall '{datadir}'",
        f"sudo chmod -R 775 '{datadir}'",
        f"""sudo grep -q '{environment}' '{user_oracle_profile}' || echo '{environment}' | sudo tee -a '{user_oracle_profile}'""",
        f"sudo mkdir -p '{response_file_dir}'",
        f"echo '{response_files[oracle_version]}' | sudo tee '{response_file}'",
        f"sudo chown oracle:oinstall '{response_file}'", 
        f"sudo chmod 0600 '{response_file}'", 
        f"sudo -iu oracle dbca -silent -createDatabase -responseFile '{response_file}'",
        # since we finished to work on datadir
        # - change access to oracle:dba
        f"sudo chown -R oracle:dba '{datadir}'",
        f"sudo sed -i 's|{cdb_name}:{oracle_home}:N|{cdb_name}:{oracle_home}:Y|' '/etc/oratab'",
    ]
    for command in commands:
        jsn = json.loads(remote_command(host, command, True), strict = False)
        if jsn['stderr'] != '' or ('DBT-' in jsn['stdout'] and '[FATAL]' in jsn['stdout'] and 'DBT-10317' not in jsn['stdout']):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
    if database_user not in ["sys", "system"]:
        step = 'creating superuser'
        duration_step_start= time.perf_counter()

        cdb["database_user"] = 'sys'
        cdb["database_pwd"] = user_sys_pwd
        error, conn = get_cxoracle_connection(cdb=cdb, host=host)
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
            return

        try:
            with conn:
                with conn.cursor() as cur:
                    #One note - custom CDB database username must start with C## prefix for oracle safety reasons, 
                    #should we enforce that as rule, or it is up to user?
                    #That name restriction does not apply to standalones
                    cur.execute(f"SELECT COUNT(*) FROM dba_users WHERE username='{database_user.upper()}'")
                    users_count = cur.fetchone()[0]
                    if users_count == 0:
                        cur.execute(f"CREATE USER {database_user} IDENTIFIED BY {database_pwd}")
                    cur.execute(f"GRANT SYSDBA,DBA,CREATE SESSION,RESOURCE,UNLIMITED TABLESPACE TO {database_user} {'container=all' if not standalone else ''}")
        except Exception as e:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(e),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(e),1)
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'registering CDB'
    duration_step_start= time.perf_counter()

    cdb["database_user"] = database_user
    cdb["database_pwd"] = database_pwd
    cdb_size = 0.0
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return
    
    try:
        with conn:
            with conn.cursor() as cur:
                cur.execute(spec_statement)
                cur.execute(body_statement)
                cur.execute("CREATE OR REPLACE DIRECTORY dir_123cluster AS '{}'".format(datadir))
                cur.execute("ALTER SYSTEM SET DB_FLASHBACK_RETENTION_TARGET={} SCOPE = BOTH".format(flashback_keep_days_minutes))
                cur.execute("SELECT FLASHBACK_ON FROM V$DATABASE")
                flashback_state = cur.fetchone()[0]
                if 'NO' in flashback_state:
                    cur.execute("ALTER DATABASE FLASHBACK ON")
                
                cur.execute("ALTER SYSTEM SET DB_CREATE_FILE_DEST='{}' SCOPE = BOTH".format(datadir))
                cur.execute("SELECT STATUS FROM V$BLOCK_CHANGE_TRACKING")
                change_tracking_state = cur.fetchone()[0]
                if 'DISABLED' in change_tracking_state:
                    cur.execute("ALTER DATABASE ENABLE BLOCK CHANGE TRACKING")
                
                cur.execute("SELECT SUM(bytes) FROM dba_data_files")
                try:
                    fetched_cdb_size = cur.fetchone()[0]
                    if fetched_cdb_size is None:
                        cdb_size = 0
                    else:
                        cdb_size = float(fetched_cdb_size)/1024
                except:
                    cdb_size = 0
                
                cur.execute("SELECT VERSION FROM V$INSTANCE")
                try:
                    oracle_version = cur.fetchone()[0]
                except:
                    pass
    except Exception as e:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
        
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s, version=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, cdb_size, oracle_version, cdb_id))
                cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = jsonb_set(ex_params, '{{oracle_home}}', '"{0}"', true) WHERE id={1}""".format(oracle_home,cdb_id))
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
        
    # host['create_task'] = created_by
    # host['ssh_port'] = int(host['ssh_port'])
    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host, 'task': 'run_linux_performance_counters', 'target_type': OperatingSystem.LINUX.value, 'target_id': host['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as error:
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def install_oracle(app_host, host, oracle_version, repository):
    "install oracle database software"
    try:
        binaries_package = f'{TASK_TMP_FILES_DIR}/oracle/{oracle_version}/oracle_software.rpm'
        
        jsn = {'stderr': '', 'stdout': '' ,'excode': 0}
        #need this,oracle could be installed on EL 7/8/9
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            return release
        
        #any EL distributive may go,but we have what we have
        if OperatingSystem.CENTOS.value not in distributor: 
            jsn['stderr'] = f"Only {OperatingSystem.CENTOS.value} are currently supported for Oracle installation.Detected OS are {distributor}."
            return jsn
         
        #example path
        if Repository.LOCAL.value in repository:
            jsn = install_offline_remote_to_remote(distributor, release, host, oracle_version, 'oracle', app_host)
            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
            if filter:
                return jsn

            # when install are completely offline, 
            # oracle preinstall improperly configures user...
            post_install = [
                # delete user "oracle" with all user-related files, non-force
                "sudo userdel -r oracle",
                # add user "oracle" with proper uid, groups, home directory and shell
                "sudo useradd -m -u 54321 -g oinstall -G dba,oper oracle -d '/home/oracle' -s '/bin/bash'"
            ]
            for command in post_install:
                jsn = json.loads(remote_command(host, command, True), strict=False)
                if jsn['stderr'] != '':
                    return jsn
        else:
            # oracle pre-install usually miss some packages, so manual install goes first
            # for now - just for centos 8, for 7 - will be different
            requirements = """bc
binutils
elfutils-libelf
elfutils-libelf-devel
fontconfig-devel
glibc
glibc-devel
ksh
libaio
libaio-devel
libXrender
libXrender-devel
libX11
libXau
libXi
libXtst
libgcc
librdmacm-devel
libstdc++
libstdc++-devel
libxcb
make
net-tools
nfs-utils
targetcli
smartmontools
sysstat
gcc
unixODBC
libnsl
libnsl2
unzip
file
file-libs
rsync
"""
            requirements_file = "oracle_requirements.txt"
            commands = [
                f"echo '{requirements}' | sudo tee '{requirements_file}'",
                f"sudo yum --enablerepo=appstream install -q -y $(sudo cat '{requirements_file}')",
                'sudo yum --enablerepo=appstream install -q -y https://yum.oracle.com/repo/OracleLinux/OL7/0/base/x86_64/getPackage/compat-libstdc++-33-3.2.3-71.el7.x86_64.rpm',
                'sudo yum --enablerepo=appstream install -q -y https://yum.oracle.com/repo/OracleLinux/OL7/0/base/x86_64/getPackage/compat-libcap1-1.10-7.el7.x86_64.rpm',
                'sudo yum --enablerepo=appstream install -q -y https://yum.oracle.com/repo/OracleLinux/OL8/appstream/x86_64/getPackage/oracle-database-preinstall-19c-1.0-2.el8.x86_64.rpm',
            ]
            for command in commands:
                jsn = json.loads(remote_command(host, command, True), strict=False)
                if jsn['stderr'] != '':
                    return jsn
                
            jsn = json.loads(remote_command(app_host, "sudo find / -type d -name 'rpms' 2>/dev/null | tail -n 1", True), strict=False)
            if jsn['stderr'] != '':
                return jsn
            
            if jsn['stdout'] == '':
                jsn['stderr'] == 'Could not determine local repository location'
                return jsn
            
            local_rpms_dir = jsn['stdout']
            # assume that local repository has the structure from install_offine function
            oracle_path = f"{local_rpms_dir}/{distributor}{release}/oracle/{oracle_version}/oracle_software.rpm"
            
            #any oracle binaries are several GB for sure, rsync transfer are mandatory
            jsn = transfer_files_rsync(app_host, host, os.path.dirname(binaries_package), [oracle_path])
            if jsn['stderr'] != '':
                return jsn
            
            jsn = json.loads(remote_command(host, f"sudo yum localinstall -q -y '{binaries_package}'", True), strict = False)
            return jsn
        
        return jsn
    except Exception as e:
        jsn['stderr'] = str(e)
        return jsn

'''
@shared_task
def open_pdb_standbys_task(pdb_id):
    """open a pluggable database task"""

    try:
        pdb = PDB_Instance.objects.get(pk=pdb_id)
    except Exception as e:
        error = str(e.args[0])
        Logger()._error(error, user='system')
        raise Exception(error)
    else:
        standbys = PluggableStandby.objects.filter(pdb=pdb)

    Logger()._info(f"open database in {len(standbys)} standbys", user='system')
    # step 1 establish a connection
    for standby in standbys:
        try:
            conn = get_cxoracle_connection(0, cdb=standby.standby)
        except Exception as e:
            error = str(e.args[0])
            Logger()._error(error, user='system')
            raise Exception(error)
        else:
            Logger()._info(f"connected to {standby}", user='system')
        
        try:
            cursor = conn.cursor()
        except Exception as error:
            raise Exception('task_id:', task_id, str(error))

        # Check the database current open mode:
        try:
            cursor.execute(f"alter pluggable database {pdb.name} open READ ONLY")
        except Exception as e:
            error = str(e.args[0])
            if error.find('ORA-65019') != -1:
                Logger()._warning(error, user='system')
                return True
            else:
                Logger()._error(error, user='system')
                raise Exception(error)
        finally:
            cursor.close()
            conn.close()

    return True

'''
def close_pdb_task(data, task_id):
    """close a pluggable database task"""

    task = 'close PDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        cdb = json.loads(data['cdb_data'])
        pdb_id = data['pdb_id']
        host = data['host_data']
        close_query = data['close_query']
        # pdb_name = data['pdb_name']
        # option = data['option']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'closing PDB'
    duration_step_start= time.perf_counter()

    # step 1 establish a connection
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return

    # duration_step_start= time.perf_counter()
    # step = 'check the database current open mode'
    
    # status = Severity.OK.value
    # check open mode
    # try:
    #     cursor.execute("Select open_mode from v$pdbs where name ='{}'".format(pdb_name))
    # except Exception as e:
    #     error, = e.args
    #     status = Severity.NOT_OK.value
    #     duration_step_end = time.perf_counter() - duration_step_start
    #     logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    #     step = 'null'
    #     status = Severity.NOT_OK.value
    #     duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #     logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    #     raise Exception('task_id:', task_id, str(error))
    # else:
    #     open_state = cursor.fetchone()[0]
    # duration_step_end = time.perf_counter() - duration_step_start
    # logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    # # PDB already closed
    # if 'mounted' in open_state.lower():
    #     err_msg = "pluggable database {} already closed".format(pdb_name)
    #     print('task_id:', task_id, err_msg)
    #     step = 'null'
    #     status = Severity.WARNING.value
    #     duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #     logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    #     return
    
    # close the PDB
    # close_query = "alter pluggable database {} close ".format(pdb_name)
    # if option == 'CI':
    #     close_query += "immediate"
    # elif option == 'CA':
    #     close_query += "abort"
    
    error = ''
    try:
        cursor = conn.cursor()
        cursor.execute(close_query)
    except Exception as e:
        error, = e.args
        error = str(error)
        if error.find('ORA-65020:') == -1:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return

    cursor.close()
    conn.close()

    status = Severity.OK.value if not error else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, pdb_id))
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def chk_pdb_exist(cursor, pdb_name):
    """return True if the pdb exists else False"""

    try:
        cursor.execute("SELECT PDB_NAME FROM DBA_PDBS")
        pdb_list = cursor.fetchall()
    except Exception as e:
        error, = e.args
        return (False, str(error))

    for item in pdb_list:
        if item[0] == pdb_name:
            return True

    return False

def close_pdb(cursor, pdb_name) -> tuple:
    """close PDB function"""

    status, error = Severity.OK.value, ''
    try:
        pdb_exist = chk_pdb_exist(cursor, pdb_name)
        if isinstance(pdb_exist, tuple):
            return (Severity.NOT_OK.value, str(pdb_exist[1]))

        if pdb_exist:
            cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE IMMEDIATE".format(pdb_name))
        else:
            logging.warning(f"{pdb_name} doesn't exist")
    except Exception as e:
        error, = e.args
        error = str(error)
        if error.find('ORA-65020') == -1:
            return (Severity.NOT_OK.value, error)
        else:
            return (Severity.WARNING.value, error)

    return (status, error)
    
def drop_pdb_task(data, task_id):
    """
    a task to drop a pluggable database from a container
    parameters:
        cdb_data: the local model containter
        pdb_data: the local model pluggable
        task_id: id of the task
    """

    task = 'drop PDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        cdb = json.loads(data['cdb_data'])
        pdb = json.loads(data['pdb_data'])
        host = data['host_data']
        option = data['option']
        is_autostartup = False
        # auto_delete_task_id = pdb['auto_delete_task_id']
        error, conn = get_cxoracle_connection(cdb=cdb, host=host)
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
            return
        
        cursor = conn.cursor()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if option == "justDrop":
        step = 'closing the PDB'
        duration_step_start= time.perf_counter()

        step_status, error = close_pdb(cursor, pdb['name'])
        if step_status == Severity.NOT_OK.value and str(error).find('ORA-01219') == -1:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        elif step_status == Severity.WARNING.value:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        else:
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
      
        step = 'dropping the PDB including datafiles'
        duration_step_start= time.perf_counter()

        try:
            cursor.execute("DROP PLUGGABLE DATABASE {} INCLUDING DATAFILES".format(pdb['name']))
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        is_autostartup = pdb['ind_autostartup']
    # dropping the database but keeping the metadata 
    else:
        step = 'closing the PDB'
        duration_step_start= time.perf_counter()

        step_status, error = close_pdb(cursor, pdb['name'])
        if step_status == Severity.NOT_OK.value:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        elif step_status == Severity.WARNING.value:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        else:
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        step = 'dropping the PDB but keep PDB metadata'
        duration_step_start= time.perf_counter()

        try:
            cursor.execute("ALTER PLUGGABLE DATABASE {} OPEN".format(pdb['name']))

            step_status, error = close_pdb(cursor, pdb['name'])
            if step_status == Severity.NOT_OK.value:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
        
            cursor.execute("ALTER PLUGGABLE DATABASE {} OPEN READ ONLY".format(pdb['name']))
        except Exception as e:
            error, = e.args
            error = str(error)
            if error.find('ORA-65019') != -1:
                status = Severity.WARNING.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            else:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
        
        try:
            cursor.execute("SELECT d.FILE_NAME FROM DBA_PDBS p, CDB_DATA_FILES d WHERE p.PDB_ID=d.CON_ID AND pdb_name='{}'".format(pdb['name']))
            file_names = cursor.fetchall()
            file_list = ''
            for file_n in file_names:
                file_list += file_n[0] + ' '
            
            remote_command(host, "sudo rm -f '{}'".format(file_list), False)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        is_autostartup = pdb['ind_autostartup']
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    # PDB_Instance.objects.filter(id=pdb.id).delete()
    # try:
    #     conn_postgres = get_postgres_connection()
    #     cur = conn_postgres.cursor()
    #     cur.execute('DELETE FROM oracle_pdb_instance WHERE id={}'.format(pdb['id']))
    #     conn_postgres.commit()
    # except Exception as e:
    #     print('ERROR:drop_pdb_task:connection_postgres:',e)
    #     step = 'null'
    #     status = Severity.NOT_OK.value
    #     duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #     logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    #     raise Exception(str(error)[:99])
    # cur.close()
    # conn_postgres.close()

    if is_autostartup:
        error = create_startup_triggers_task(host, cdb, pdb, task, task_id, task_log_id, created_by)
        if error != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    cursor.close()
    conn.close()

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_deleted='true', ind_started='false' WHERE id=%s""", (Severity.NOT_OK.value, Status.DROPPED.value, pdb['id']))
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def rename_pdb_task(data, task_id):
    """rename PDB task"""

    task = 'rename PDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        cdb = json.loads(data['cdb_data'])
        pdb = json.loads(data['pdb_data'])
        host = data['host_data']
        new_name = data['new_name'] 
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'closing the PDB'
    duration_step_start= time.perf_counter()

    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return
   
    try:
        cursor = conn.cursor()
        cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE IMMEDIATE".format(pdb['name']))
    except Exception as e:
        error, = e.args
        if str(error).find('ORA-65020') == -1:
            # the error isn't already closed
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        else:
            pass
            # already closed error
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'openning and renaming PDB'
    duration_step_start= time.perf_counter()

    try:
        cursor.execute("ALTER PLUGGABLE DATABASE {} OPEN RESTRICTED".format(pdb['name']))
        cursor.execute("ALTER SESSION SET CONTAINER={}".format(pdb['name']))
        cursor.execute("ALTER PLUGGABLE DATABASE {0} RENAME global_name TO {1}".format(pdb['name'],new_name))
        cursor.execute("ALTER SESSION SET CONTAINER = CDB$ROOT")
        cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE IMMEDIATE".format(new_name))
        cursor.execute("ALTER PLUGGABLE DATABASE {} OPEN".format(new_name))
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if pdb['ind_autostartup']:
        error = create_startup_triggers_task(host, cdb, pdb, task, task_id, task_log_id, created_by)
        if error != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    cursor.close()
    conn.close()

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET name=%s, severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [new_name.upper(), Severity.OK.value, Status.OK.value, pdb['id']])
                conn.commit()
                # Query: Select task_id for updating node name in monitor_task
                cur.execute("""SELECT task_id FROM "clusterCore_task" WHERE target_id=%s AND target_type=%s AND task='monitor_task'""", [pdb['id'], pdb['type']])
                task_ids = [task_id[0] for task_id in cur.fetchall()]
                # Query: Update parameters -> node_data -> name
                cur.execute("""UPDATE "clusterCore_task" SET parameters=jsonb_set(parameters, '{node_data,name}', %s::jsonb) WHERE task_id=ANY(%s)""", [json.dumps({"name": new_name}), task_ids])
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def copy_pdb_task(data, task_id):
    """copy PDB task"""

    task = 'copy PDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        cdb = json.loads(data['cdb_data'])
        pdb = json.loads(data['pdb_data'])
        host = data['host_data']
        copy_name = data['copy_name']
        autostart = data['autostart']
        source_restart = data['source_restart']
        target_restart = data['target_restart']
        pdb_target_id = data['pdb_target_id']
        file_dest = data['file_dest']
        if file_dest.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            file_dest = file_dest.replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating directory'
    duration_step_start= time.perf_counter()

    commands = [
        f"sudo mkdir -p '{file_dest}'",
        f"sudo chown oracle:dba '{file_dest}'"
    ]
    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'copying the PDB'
    duration_step_start= time.perf_counter()

    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return

    # close PDB
    try:
        cursor = conn.cursor()
        cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE IMMEDIATE".format(pdb['name']))
        # open readonly
        cursor.execute("ALTER PLUGGABLE DATABASE {} OPEN READ ONLY".format(pdb['name']))
        # create a PDB copy
        # try:
        #     cursor.callproc("SYS.PCK_PDB_MAINT.pc_copy_pdb", [pdb['name'], copy_name])
        # except Exception as e:
        #     error, = e.args
        #     raise Exception(str(error))
        # else:
        #     print("copy {} as {} in progress".format(pdb, copy_name))
        cursor.execute("CREATE PLUGGABLE DATABASE {0} FROM {1}  USER_TABLESPACES=ALL CREATE_FILE_DEST='{2}' standbys=none".format(copy_name, pdb['name'], file_dest))
    except Exception as e:
        error, = e.args
        if 'ORA-65020' not in str(error):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        else:
            pass

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # open the new copy according to option
    step = 'openning the new copy and close/re-open the original PDB'
    duration_step_start= time.perf_counter()

    target_restart_query = f"ALTER PLUGGABLE DATABASE {copy_name} OPEN"
    if target_restart == 'RO':
        try:
            cursor.execute(target_restart_query)
            cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE IMMEDIATE".format(copy_name))
        except Exception as e:
            error, = e.args
            if 'ORA-65020' not in str(error):
                error, = e.args
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

        target_restart_query = target_restart_query + " READ ONLY"
    
    if target_restart != 'NA':
        try:
            cursor.execute(target_restart_query)
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    # close the original PDB again
    try:
        cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE IMMEDIATE".format(pdb['name']))
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    # reopen the original PDB again according to option
    source_restart_query = f"ALTER PLUGGABLE DATABASE {pdb['name']} OPEN"
    if source_restart == 'RO':
        source_restart_query = source_restart_query + " READ ONLY"

    if source_restart != 'NA':
        try:
            cursor.execute(source_restart_query)
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    try:
        cursor.close()
        conn.close()
    except Exception:
        pass
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if autostart == 'true':
        error = create_startup_triggers_task(host, cdb, pdb, task, task_id, task_log_id, created_by)
        if error != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                if source_restart != 'NA':
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, pdb['id']))
                
                if target_restart != 'NA':
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, pdb_target_id))
                
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def config_pdb(host, cdb, pdb, new_size, task, task_id, task_log_id, created_by) -> str:
    """a function to configure a pluggable oracle database"""
    
    step = 'configuring PDB'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    error = ''
    max_size = new_size['disk']
    shared_temp = new_size['shared_temp']
    cpu_count = new_size['cpu']
    # autostart = task_args['autostart']
    max_size = int(max_size)
    shared_temp = int(shared_temp)
    cpu_count = int(cpu_count)
    # autostart = bool(autostart)
    cluster_size = get_cluster_sizing_per_size(pdb['sizing'])
    if isinstance(cluster_size, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(cluster_size[1]),1)
        return str(cluster_size[1])

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'setting and limiting PDB size'
    duration_step_start= time.perf_counter()

    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn),1)
        return conn
    else:
        error = ''

    try:
        cursor = conn.cursor()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        return str(error)
    
    try:
        cursor.execute("ALTER PLUGGABLE DATABASE {} OPEN".format(pdb['name']))
        cursor.execute("ALTER SESSION SET CONTAINER={}".format(pdb['name']))
    except Exception as e:
        error, = e.args
        if 'ORA-65019' in str(error):
            error = ''
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            return str(error)
    
    # set max disk size limit
    # set to unlimited size
    if max_size == 10001:
        try:
            cursor.execute("ALTER PLUGGABLE DATABASE {} STORAGE UNLIMITED".format(pdb.name))
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            return str(error)
    # limited to size max_size
    elif max_size != cluster_size['memory']:
        try:
            cursor.execute("ALTER PLUGGABLE DATABASE {} STORAGE (MAXSIZE {}G)".format(pdb['name'], max_size))
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            return str(error)

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # set the shared temp size
    step = 'setting the shared temp size'
    duration_step_start= time.perf_counter()
    
    if shared_temp == 10001:
        try:
            cursor.execute("ALTER PLUGGABLE DATABASE {} MAX_SHARED_TEMP_SIZE UNLIMITED".format(pdb['name']))
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            return str(error)
    elif shared_temp != cluster_size['shared_temp']:
        try:
            cursor.execute("ALTER PLUGGABLE DATABASE {} STORAGE (MAX_SHARED_TEMP_SIZE {}G)".format(pdb['name'], shared_temp))
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            return str(error)

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # set the cpu count
    step = 'setting the cpu count'
    duration_step_start= time.perf_counter()

    if cpu_count != cluster_size['cpu']:
        try:
            cursor.execute("ALTER SYSTEM SET cpu_count={} scope=both".format(cpu_count))
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            return str(error)

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # 6. autostartup
    # if autostart != pdb.is_autostartup:
    #     try:
    #         SetAutoStartPDB.delay(cdb.id, tpl.id)
    #     except Exception as e:
    #         error, = e.args
    #         raise Exception(task_id, error)
    #     else:
    #         pdb.is_autostartup = autostart
    #         pdb.save()
    
    try:
        cursor.close()
        conn.close()
    except Exception:
        pass

    step = 'configuring PDB finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    return error

def change_pdb_cluster_size_task(data, task_id):
    """a task to change a pluggable oracle database max size, cpu and memory"""

    task = 'change PDB cluster size task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()
    
    try:
        cdb = json.loads(data['cdb_data'])
        pdb = json.loads(data['pdb_data'])
        host = data['host_data']
        size_scheme = data['new_size']
        new_size = get_cluster_sizing_per_size(size_scheme)
        if isinstance(new_size, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(new_size[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(new_size[1]),1)
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    error = config_pdb(host, cdb, pdb, new_size, task, task_id, task_log_id, created_by)
    if error != '':
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET sizing=%s WHERE id=%s""", (size_scheme, pdb['id']))
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def copy_to_another_cdb_task(data, task_id):
    """copy PDB to another CDB task"""

    task = 'copy PDB to another CDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        host_source = data['host_source_data']
        host_target = data['host_target_data']
        source = json.loads(data['source_data'])
        target = json.loads(data['target_data'])
        file_dest = data['file_dest']    
        pdb_source_id = data['pdb_source_id']
        pdb_target_id = data['pdb_target_id']
        pdb_name = data['pdb_name']
        pdb_user = data['pdb_user']
        pdb_pwd = data["pdb_pwd"]
        copy_name = data['copy_name']
        source_pdb_restart = data['source_restart']
        target_pdb_restart = data['target_restart']
        jsn = find_tilda(host_target)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        tilda_path = jsn['stdout']

        if file_dest and file_dest.startswith('~'):
            file_dest = file_dest.replace('~', tilda_path)

        if target['dir_123Cluster'] and target['dir_123Cluster'].startswith('~'):
            target['dir_123Cluster'] = target['dir_123Cluster'].replace('~', tilda_path)
        
        clone_link = "clone_link_123cluster"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'preparing destination folder'
    duration_step_start= time.perf_counter()

    dest_dir = f"{file_dest}/{copy_name}"
    commands = [
        f"sudo mkdir -p '{dest_dir}'",
        f"sudo chown -R oracle:dba '{dest_dir}'",
        f"sudo chmod -R 755 '{dest_dir}'"
    ]
    for command in commands:
        jsn = json.loads(remote_command(host_target, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'starting source PDB as READ ONLY'
    duration_step_start= time.perf_counter()

    # 1. Connect to the PDB
    error, conn = get_cxoracle_connection(cdb=source, host=host_source)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return

    try:
        cursor = conn.cursor()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    # 2. Make sure the source pluggable database is started RO
    try:
        cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE IMMEDIATE".format(pdb_name))
    except Exception as e:
        error, = e.args
        if str(error).find('ORA-65020') != -1:
            pass
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    try:
        cursor.execute("ALTER PLUGGABLE DATABASE {} OPEN READ ONLY".format(pdb_name))
        cursor.close()
        conn.close()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'creating PDB in target'
    duration_step_start= time.perf_counter()
    
    # 7. Create the PDB in the target
    try:
        create_link = f"""CREATE DATABASE LINK {clone_link} CONNECT TO {pdb_user} IDENTIFIED BY {pdb_pwd} USING '(DESCRIPTION =
	(ADDRESS_LIST =
  	(ADDRESS = (PROTOCOL = TCP)(HOST = {host_source['name']})(PORT = {source['port']}))
	)
	(CONNECT_DATA =
  	(SERVICE_NAME = {pdb_name})
	)
  )'"""
        drop_link = f"DROP DATABASE LINK {clone_link}"
        create_target_pdb = '''CREATE PLUGGABLE DATABASE {} FROM {}@{} USER_TABLESPACES=ALL CREATE_FILE_DEST='{}' standbys=none'''.format(copy_name, pdb_name, clone_link, dest_dir)
        
        error, conn = get_cxoracle_connection(cdb=target, host=host_target)
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
            return
        
        try:
            cursor = conn.cursor()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        try:
            cursor.execute(drop_link)
        except Exception as link_drop_exc:
            error, = link_drop_exc.args
            error = str(error)
            if error.find('ORA-02024') == -1:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
                return
        
        cursor.execute(create_link)
        cursor.execute(f"SELECT COUNT(*) FROM all_directories WHERE directory_name = '{copy_name}'")
        count = cursor.fetchone()[0]

        # Create the directory object if it does not exist
        if count == 0:
            cursor.execute(f"CREATE DIRECTORY {copy_name} AS '{dest_dir}'")
        cursor.execute(create_target_pdb)
        # 8. check the new pdb state
        cursor.execute("SELECT name, open_mode FROM v$pdbs WHERE name='{}'".format(copy_name.upper()))
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    try:
        cursor.execute("ALTER PLUGGABLE DATABASE {} OPEN".format(copy_name))
    except Exception as e:
        error, = e.args
        if 'ORA-65019' not in str(error):
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    try:
        cursor.execute("ALTER SESSION SET CONTAINER={}".format(copy_name))
        cursor.execute("CREATE OR REPLACE DIRECTORY dir_123cluster AS '{}'".format(target['dir_123Cluster']))
    except Exception as e:
        error, = e.args
        if 'ORA-04043' not in str(error):
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    try:
        cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE IMMEDIATE".format(copy_name))
        cursor.close()
        conn.close()
    except Exception as e:
        error, = e.args
        if 'ORA-65020' not in str(error):
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'opening source and target according to checkboxes'
    duration_step_start= time.perf_counter()
    # 9. According to checkboxes for source and also target
        # Do nothing (NA)
        # Open RW
        # Open RO
    
    if target_pdb_restart != "NA":
        target_open_query = "ALTER PLUGGABLE DATABASE {} OPEN READ ONLY".format(
            copy_name) if target_pdb_restart == "RO" else "ALTER PLUGGABLE DATABASE {} OPEN".format(copy_name)

        error, conn = get_cxoracle_connection(cdb=target, host=host_target)
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
            return
        
        try:
            cursor = conn.cursor()
            cursor.execute(target_open_query)
            cursor.close()
            conn.close()
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
    if source_pdb_restart == "RW":
        # reconnect to the source container to make the changes
        error, conn = get_cxoracle_connection(cdb=source, host=host_source)
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
            return

        try:
            cursor = conn.cursor()
            cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE".format(pdb_name))
            cursor.execute("ALTER PLUGGABLE DATABASE {} OPEN".format(pdb_name))
            cursor.close()
            conn.close()
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                if source_pdb_restart == "NA":
                    query = """UPDATE "clusterCore_c123_node" SET severity=%d, status=%s WHERE id=%s"""
                else:
                    query = """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s"""

                cur.execute(query, (Severity.OK.value, Status.OK.value, pdb_source_id))
                conn.commit()

                if target_pdb_restart == "NA":
                    query = """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s WHERE id=%s"""
                else:
                    query = """UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s"""

                cur.execute(query, (Severity.OK.value, Status.OK.value, pdb_target_id))
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
        
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

####################################################################
# PDB BACKUPS
####################################################################

def backup_pdb_task(data, task_id):
    """backup PDB task"""

    task = 'backup PDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        pdb_name = data['pdb_name']
        pdb_id = data['pdb_id']
        cdb = json.loads(data['cdb_data'])
        host = data['host_data']
        backup_host = data['backup_host']
        backup = json.loads(data['backup_data'])
        #must not contain special characters
        backup_path = backup['ex_params'].get('backup_path')
        schedule = backup['is_sched']
        compress = backup['ex_params'].get('compress')
        # options dropdown
        incremental_level = backup['ex_params'].get('incremental_level').strip().upper()
        if backup_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    
            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)
        
        # create a unique string with the date
        now = datetime.utcnow().strftime('%d%m%Y%H%M%S')
        if schedule:
            if incremental_level == IncrementalLevel.NO_INCREMENTAL.value:
                backup_path = f'{backup_path}/{now}'
            # if incremental are used - all the subsequent backups must go to the same backup path
            else:
                backup_path = f'{backup_path}/incremental'
       
        datadir = cdb['dir_123Cluster']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
        
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # 2. create the backup directory
    step = 'creating backup directory'
    duration_step_start= time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo mkdir -p '{backup_path}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        if schedule and incremental_level != IncrementalLevel.NO_INCREMENTAL.value:
            # check if incremental backup ever completed successfully
            jsn = check_remote_directory_exists(backup_host, backup_path, 'd', is_dir_empty=True)
            if jsn['stderr'] != '' and 'no such' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
            backup_folder_empty = True if 'true' in jsn['stdout'] else False
            # if incremental folder does not contain files
            if backup_folder_empty:
                # default to level 0 backup
                incremental_level = IncrementalLevel.BASE.value
            else:
                # download all the existing incremental backup files back to node host
                jsn = transfer_files_rsync(backup_host, host, backup_path, [f"{backup_path}/"])
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
        
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'creating RMAN script'
    duration_step_start= time.perf_counter()
    
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return

    try:
        cursor = conn.cursor()
        cursor.execute("SELECT version FROM v$instance")
        major_version = int(cursor.fetchone()[0].split('.')[0])
        if major_version in [18, 19]:
            cursor.execute("ALTER SESSION SET CONTAINER={}".format(pdb_name))
            cursor.execute("SELECT value FROM v$parameter WHERE name='db_create_file_dest'")
            db_create_file_dest = cursor.fetchone()[0]
            cursor.execute("ALTER SYSTEM SET db_create_file_dest='' scope=both")
            conn.commit()
        cursor.close()
        conn.close()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    # 3. create the rman script
    rman_script_file = backup_path + \
        "/rman_" + pdb_name + "_" + now + ".rman"

    logfile_name = pdb_name + "_" + now + ".log"
    log_file = backup_path + "/" + logfile_name
    backup_full_path = "{}/%U".format(backup_path)
    
    line1 = "ALLOCATE CHANNEL c1 TYPE DISK FORMAT '{}';".format(backup_full_path)
    line2 = "BACKUP {} {} PLUGGABLE DATABASE '{}';".format('AS COMPRESSED BACKUPSET' if compress else '', incremental_level , pdb_name)

    try:
        commands = [
            f"""echo -e "run {{\n{line1}\n{line2}\nRELEASE CHANNEL c1;}}" | sudo tee '{rman_script_file}'""",
            f"sudo chown -R oracle:dba '{backup_path}'",
            f"sudo chmod -R 755 '{backup_path}'",
            f"sudo chown -R oracle:dba '{datadir}'",
            f"sudo chmod -R 755 '{datadir}'",
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # 4. run the script
    # jsn = json.loads(remote_command(
    #     client, "export ORACLE_HOME={0};{1}/bin/rman target=sys/{2}@{3}:{4}/{5} @{6} log={7}".format(cdb['oracle_home'], cdb['oracle_home'], cdb['pwd'], cdb['host'], cdb['port'], cdb['service_name'], rman_script_file, log_file), True), strict=False)
    
    step = 'running the RMAN script'
    duration_step_start= time.perf_counter()

    try:
        # command = "export ORACLE_HOME={0};export ORACLE_SID={1};{0}/bin/rman target=/ @{2} log={3}".format(cdb['ex_params'].get('oracle_home'), cdb['name'], rman_script_file, log_file)
        command = "sudo -iu oracle ORACLE_HOME={0} ORACLE_SID={1} {0}/bin/rman target=/ @{2} log={3}".format(cdb['ex_params'].get('oracle_home'), cdb['name'], rman_script_file, log_file)
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    if major_version in [18, 19]:
        error, conn = get_cxoracle_connection(cdb=cdb, host=host)
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
            return

        try:
            cursor = conn.cursor()
            cursor.execute("ALTER SESSION SET CONTAINER={}".format(pdb_name))
            cursor.execute(f"ALTER SYSTEM SET db_create_file_dest='{db_create_file_dest}'")
            conn.commit()
            cursor.close()
            conn.close()
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # 5. read the log file to see if there where any errors
    step = 'reading the log file'
    duration_step_start= time.perf_counter()

    # save the file in the local backup object
    try:
        stdout = "The backup failed. Please check the log file {} for details.".format(log_file)
        jsn = json.loads(remote_command(host, f"sudo cat '{log_file}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        backup['run_log'] = jsn['stdout']
        with get_postgres_connection() as conn_postgres:
            with conn_postgres.cursor() as cur_post:
                cur_post.callproc('set_logfile_data', (Inventory.BACKUP.value,backup['run_log'],backup['id']))
                conn_postgres.commit()
                for line in backup['run_log'].splitlines():
                    if 'RMAN-' in line:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,line,1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,line,1)
                        return
                
                # backup is successful
                backup['status'] = True
                cur_post.execute("""UPDATE "clusterCore_c123_inventory" SET status=%s WHERE id=%s""", (backup['status'], backup['id']))
                conn_postgres.commit()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # 6.1 upload to backup host
    if host['name'] != backup_host['name']:
        step = 'copying backup of PDB to backup host'
        duration_step_start = time.perf_counter()

        jsn = transfer_files_rsync(host, backup_host, backup_path, [f"{backup_path}/"])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    # 6.2. get the backup size
    step = 'getting the backup size'
    duration_step_start= time.perf_counter()

    try:
        # 6. get the backup size
        jsn = json.loads(remote_command(backup_host, f"sudo du -sb '{backup_path}' | awk '{{print $1}}'", True))
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        modified_date = get_date_modified_folder(hostdata=backup_host, folder_path=backup_path) 
        if isinstance(modified_date, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return
        
        # backup['ex_params']['size'] = round(int(jsn['stdout'])/1024/1024, 2)
        backup['ex_params']['size'] = float(jsn['stdout'])/1024
        backup['ex_params']['size'] = str(backup['ex_params'].get('size'))
        #if we need update both before and after transfer - for after transfer need new connection
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true) WHERE id={1}""".format(backup['ex_params'].get('size'), backup['id']))
                #relate to size - if have unexpected format in modified_date = something wrong with backup file
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}', status='true' WHERE id=%s""" % (modified_date, backup['id']))
                #if backup data are ok - assume that node ok
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, pdb_id))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{exit_code}}', '{0}', true) WHERE id={1}""".format(0, backup['id']))
                conn.commit()

                remote_command(host, f"sudo rm -rf '{backup_path}'", True)

                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

                step = 'task finished'
                status = Severity.OK.value
                duration = 0
                logger(task_id,task_log_id,task,step,status,duration,created_by)
                step = 'null'
                status = Severity.OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

                try:
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{duration}}', '{0}', true) WHERE id={1}""".format(duration_task_end, backup['id']))
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"start_date":\"%s\"}', status='true' WHERE id=%s""" % (duration_task_start.strftime("%Y-%m-%d %H:%M:%S"), backup['id']))
                    conn.commit()
                except Exception as duration_exc: #notice-level
                    error, = duration_exc.args
                    logging.error(f"{task}@{step}: {task_id}: {str(error)}")
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
def restore_pdb_from_backup_task(data, task_id):
    """restore an Oracle pluggable database from a backup task"""

    task = 'restore PDB from backup task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        pdb_name = data['pdb_name']
        pdb_id = data['pdb_id']
        # bu_id = task_args['bu_id']
        autostart = data['autostart']
        cdb = json.loads(data['cdb_data'])
        host = data['host_data']
        backup_host = data['backup_host']
        backup = json.loads(data['backup_data'])
        backup_path = backup['ex_params'].get('backup_path')
        schedule = backup["is_sched"]
        restore_option = data['restore_option'].lower()
        incremental_level = backup['ex_params'].get('incremental_level').strip().upper()
        date = data['date']
        if backup_path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            backup_path = backup_path.replace('~', tilda_path)

        now = datetime.utcnow().strftime('%d%m%Y%H%M%S')
        timezone_cmd = 'cat /etc/timezone'
        jsn = json.loads(remote_command(backup_host, timezone_cmd, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        timezone = pytz.timezone(jsn['stdout'])
        if schedule and incremental_level != IncrementalLevel.NO_INCREMENTAL.value:
            backup_path = f'{backup_path}/incremental'
         
        elif schedule:
            if restore_option in [RestoreOptions.LATEST.value, RestoreOptions.CUSTOM.value]:
                if RestoreOptions.CUSTOM.value in restore_option:
                    target_timestamp = datetime.strptime(date, "%Y-%m-%d %H:%M:%S").astimezone(timezone).timestamp()
                    command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | grep -v '/incremental' | sort -n"
                else:
                    command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | grep -v '/incremental' | sort -n | tail -1 | awk '{{print $2}}'"
            else:
                command = f"sudo find '{backup_path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \;  | grep -v '/incremental' | sort -n | head -1 | awk '{{print $2}}'"

            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            if jsn['stdout'] == '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                return
            else:
                if restore_option in [RestoreOptions.THIS.value, RestoreOptions.LATEST.value]:
                    backup_path = jsn['stdout']
                else:
                    output_lines = [line.strip() for line in jsn['stdout'].split("\n") if line.strip()]
                    closest_subfolder = None
                    min_time_difference = positive_infinity

                    for line in output_lines:
                        subfolder_timestamp_str, subfolder = line.split(" ", 1)
                        subfolder_timestamp = int(subfolder_timestamp_str)
                        time_difference = abs(target_timestamp - subfolder_timestamp)
                        if time_difference < min_time_difference:
                            min_time_difference = time_difference
                            closest_subfolder = subfolder

                    if not closest_subfolder:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                        return

                    backup_path = closest_subfolder
        
        # "this" are datetime without timezone in timezone of app (UTC)
        if RestoreOptions.THIS.value in restore_option:
            start_date = backup['ex_params'].get('start_date')
            restore_date = (datetime.strptime(start_date, "%Y-%m-%d %H:%M:%S").astimezone(timezone) + timedelta(seconds=int(backup['ex_params'].get('duration')))).strftime("%d.%m.%Y %H:%M:%S")
            # restore_date = datetime.strptime(date, "%Y-%m-%d %H:%M:%S").replace(tzinfo=pytz.UTC).strftime("%d.%m.%Y %H:%M:%S")
        elif RestoreOptions.LATEST.value in restore_option:
            restore_date = ''
        else:
            restore_date = datetime.strptime(date, "%Y-%m-%d %H:%M:%S").astimezone(timezone).strftime("%d.%m.%Y %H:%M:%S")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'copying backup of PDB to PDB host'
    duration_step_start = time.perf_counter()

    jsn = transfer_files_rsync(backup_host, host, backup_path, [f"{backup_path}/"])
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating RMAN restore script'
    duration_step_start= time.perf_counter()

    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return

    try:
        cursor = conn.cursor()
        cursor.execute("SELECT open_mode FROM v$pdbs WHERE name ='{}'".format(pdb_name))
        open_mode_row = cursor.fetchone()
        cursor.close()
        conn.close()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
        return

    open_state = open_mode_row[0] if open_mode_row and open_mode_row[0] else 'unknown'
    is_pdb_closed = 'mounted' in open_state.lower()
        
     # 2. create the script file
    rman_script_file = backup_path + \
        "/restore_" + pdb_name + "_" + now + ".rman"
    restore_logfile_name = "restore_" + pdb_name + "_" + now + ".log"
    log_file = backup_path + "/" + restore_logfile_name

    # 3. create the rman restore script
    open_mode = "OPEN RESETLOGS" if restore_date else "OPEN"
    logs_destination = f"AUXILIARY DESTINATION='{backup_path}'" if restore_date else ''
    lines = [
        "run {",
        'CROSSCHECK BACKUP;',
        'ALTER PLUGGABLE DATABASE \\"{}\\" CLOSE IMMEDIATE;'.format(pdb_name) if not is_pdb_closed else '',
        "ALLOCATE CHANNEL c1 TYPE DISK FORMAT '{}/%U';".format(backup_path),
         """SET UNTIL TIME \\"TO_DATE('{}','DD.MM.YYYY HH24:MI:SS')\\";""".format(restore_date) if restore_date else '',
        'RESTORE VALIDATE PLUGGABLE DATABASE \\"{}\\";'.format(pdb_name),
        'RESTORE PREVIEW PLUGGABLE DATABASE \\"{}\\";'.format(pdb_name),
        'RESTORE PLUGGABLE DATABASE \\"{}\\";'.format(pdb_name),
        '''RECOVER PLUGGABLE DATABASE \\"{}\\" {};'''.format(pdb_name, logs_destination),
        'RELEASE CHANNEL c1;',
        'ALTER PLUGGABLE DATABASE \\"{}\\" {};'.format(pdb_name, open_mode),
        "}"
    ]
    script_lines = "\n".join(lines)
    # for backup/restore -
    # both backup and datadir must belong to oracle:dba (default rman conect user)
    commands = [
        f"""echo -e "{script_lines}" | sudo tee '{rman_script_file}'""",
        "sudo chown -R oracle:dba '{}'".format(backup_path),
        "sudo chown -R oracle:dba '{}'".format(cdb['dir_123Cluster'])
    ]
    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # 4. run the restore script
    step = 'running the RMAN restore script'
    duration_step_start= time.perf_counter()

    try:
        # command = "export ORACLE_HOME={0};export ORACLE_SID={1};{0}/bin/rman target=/ @{2} log={3}".format(cdb['ex_params'].get('oracle_home'), cdb['name'], rman_script_file, log_file)
        command = "sudo -iu oracle ORACLE_HOME={0} {1}/bin/rman target='sys/{2}@{3}:{4}/{5}' @{6} log={7}".format(cdb['ex_params'].get('oracle_home'), cdb['ex_params'].get('oracle_home'), cdb['database_pwd'], host['name'], cdb['port'], cdb['name'], rman_script_file, log_file)
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'reading the log file'
    duration_step_start= time.perf_counter()

    # save the file in the local backup object
    try:
        jsn = json.loads(remote_command(host, f"sudo cat '{log_file}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        backup['run_log'] = jsn['stdout']
        with get_postgres_connection() as conn_postgres:
            with conn_postgres.cursor() as cur_post:
                cur_post.callproc('set_logfile_data', (Inventory.BACKUP.value,backup['run_log'],backup['id']))
                conn_postgres.commit()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    stdout = "The restore failed. Please check the log file {} for details.".format(log_file)
    # search for errors in the restore log
    try:
        for line in backup['run_log'].splitlines():
            if 'RMAN-' in line:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,line,None)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,line,None)
                return
        
        remote_command(host, f"sudo rm -rf '{backup_path}'", True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if not autostart:
        step = 'closing the PDB'
        duration_step_start= time.perf_counter()

        error, conn = get_cxoracle_connection(cdb=cdb, host=host)
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
            return

        try:
            cursor = conn.cursor()
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        try:
            cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE".format(pdb_name))
        except Exception as e:
            error, = e.args
            logging.error(f"{task}@{step}: {task_id}: {str(error)}")
        finally:
            cursor.close()
            conn.close()

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, pdb_id))
                conn.commit()
        
        jsn = json.loads(remote_command(host, f"sudo rm -rf '{backup_path}'", True), strict=False)
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

####################################################################
# PDB FLASHBACKS
####################################################################

def flashback_cdb_pdb_task(data, task_id):
    """flashback_cdb_pdb_task"""

    task = 'flashback CDB PDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        obj_id = data['obj_id']
        obj_name = data['obj_name']
        obj_type = data['obj_type']
        cdb = json.loads(data['cdb_data'])
        host = data['host_data']
        flashback = json.loads(data['flashback'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'creating restore point'
    duration_step_start= time.perf_counter()
    
    # 1. connect to the database
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return

    # 2. create the restore point
    # determine the query according to object type
    if obj_type == ClusterRoles.PDB.value:
        if flashback['ex_params'].get('guaranteed'):
            query = f"CREATE RESTORE POINT {flashback['name']} FOR PLUGGABLE DATABASE {obj_name} GUARANTEE FLASHBACK DATABASE"
        else:
            query = f"CREATE RESTORE POINT {flashback['name']} FOR PLUGGABLE DATABASE {obj_name}"
    elif obj_type == ClusterRoles.CDB.value:
        if flashback['ex_params'].get('guaranteed'):
            query = f"CREATE RESTORE POINT {flashback['name']} GUARANTEE FLASHBACK DATABASE"
        else:
            query = f"CREATE RESTORE POINT {flashback['name']}"

    try:
        cursor = conn.cursor()
        cursor.execute(query)
    except Exception as e:
        error, = e.args
        if str(error).find('ORA-39893') != -1:
            try:
                cursor.execute('ALTER PLUGGABLE DATABASE {} OPEN'.format(obj_name))
            except Exception as e:
                error, = e.args
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return 
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1) 
            return
    else:
        flashback['run_log'] = "Restore point was created."
        cursor.close()
        conn.close()

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # was successful
    flashback['status'] = True
    flashback['run_log'] = "Restore point was created."

    step = 'updating flashback object'
    duration_step_start= time.perf_counter()

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{exit_code}}', '{0}', true), status='{1}', run_log='{2}' WHERE id={3}""".format(0, flashback['status'], flashback['run_log'], flashback['id']))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, obj_id))
                conn.commit()
                
                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

                step = 'task finished'
                status = Severity.OK.value
                duration = 0
                logger(task_id,task_log_id,task,step,status,duration,created_by)
                step = 'null'
                status = Severity.OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

                try:
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{duration}}', '{0}', true) WHERE id={1}""".format(duration_task_end, flashback['id']))
                    conn.commit()
                except Exception as duration_exc:
                    error, = duration_exc.args
                    logging.error(f"{task}@{step}: {task_id}: {str(error)}")        
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
def restore_pdb_from_flashback_task(data, task_id):
    """flashback to a restore point"""
    
    task = 'restore PDB from flashback task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        pdb_name = data['pdb_name']
        pdb_id = data['pdb_id']
        cdb = json.loads(data['cdb_data'])
        host = data['host_data']
        flashback = json.loads(data['flashback_data'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'flashbacking to restore point'
    duration_step_start= time.perf_counter()
    
    # 1. connect to the database
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return
    
    try:
        cursor = conn.cursor()
        # 2. close the pluggable database
        cursor.execute("ALTER PLUGGABLE DATABASE {} CLOSE IMMEDIATE".format(pdb_name))
    except Exception as e:
        error, = e.args
        if str(error).find('ORA-65020') == -1:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            error, = e.args
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    try:
        # 3. flashback to restore point
        cursor.execute("FLASHBACK PLUGGABLE DATABASE {} TO RESTORE POINT {}".format(pdb_name, flashback['name']))
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    # 4. re-open with resetlogs
    step = 're-openning with resetlogs'
    duration_step_start= time.perf_counter()

    try:
        cursor.execute("ALTER PLUGGABLE DATABASE {} OPEN RESETLOGS".format(pdb_name))
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    finally:
        cursor.close()
        conn.close()

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, pdb_id))
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def drop_restore_point_task(data, task_id):
    """a task to remove a flashback restore point of a pluggable database"""
    
    task = 'drop restore point task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        pdb_id = data['pdb_id']
        pdb_name = data['pdb_name']
        cdb = json.loads(data['cdb_data'])
        host = data['host_data']
        flashback = json.loads(data['flashback_data'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'dropping restore point'
    duration_step_start= time.perf_counter()
    
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return

    try:
        cursor = conn.cursor()
        cursor.execute("DROP RESTORE POINT {} FOR PLUGGABLE DATABASE {}".format(flashback['name'], pdb_name))
    except Exception as e:
        error, = e.args
        if 'ORA-38780' in str(error):
            pass
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    finally:
        cursor.close()
        conn.close()

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""DELETE FROM "clusterCore_c123_inventory" WHERE id=%s""", [flashback['id']])
                conn.commit()
    except Exception as e:
        error, = e.args
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

####################################################################
# PDB AUTOMATIC FUNCTIONALITY
####################################################################
# @shared_task
# def SetAutoDelete(cdb_data, pdb_data):
#     """reschedule delete tasks for automatic delete and reschedule email sending"""

#     # cdb = node_from_json(cdb_data, 'CDBInstance')
#     # pdb = node_from_json(pdb_data, 'PDB_Instance')
#     cdb = cdb_data
#     pdb = pdb_data
#     if pdb['auto_delete_task_id'] is not None:
#         celery.task.control.revoke(pdb.auto_delete_task_id)

#     subject = "{} will be delete in one {}"
#     body = "Hi {},\n{} will be delete in one {}. \
#         \nIf you wish to save your data please backup \
#             or reschedule the automatic delete date.\
#                 \n\n123Cluster"

#     args = {
#         'obj_type': 'PDB_Instance',
#         'obj_id': pdb['id'],
#         'cdb_id': cdb['id'],
#         'callback_operation': SCHEDULED_DELETE,
#         'start': str(datetime.now().replace(tzinfo=None))
#     }

#     # delete_tpl = TaskProgressLog.objects.create(user=pdb.created_by,
#     #                                             task='Delete {} scheduled on {}'.format(
#     #                                                 pdb, pdb.auto_delete),
#     #                                             args=repr(args)
#     #                                             )

#     delete_task = drop_pdb_task.si(cdb_data, pdb_data, delete_tpl.id)
#     delete_task.apply_async(eta=pdb.auto_delete)

#     week_message_task = send_email_from_celery.si((subject.format(
#         pdb, 'one week'), body.format(pdb.created_by.username, pdb, 'week'), [pdb.created_by.email]))
#     week_message_task.apply_async(eta=pdb.auto_delete - timedelta(weeks=1))
    
#     hour_message_task = send_email_from_celery.si((subject.format(
#         pdb, 'one week'), body.format(pdb.created_by.username, pdb, 'hour'), [pdb.created_by.email]))
#     hour_message_task.apply_async(eta=pdb.auto_delete - timedelta(hours=1))
    
#     pdb.auto_delete_task_id = delete_task.id
#     pdb.save()

#     return True

# @shared_task
# def SetAutoStartPDB(cdb_id):
#     cdb = CDBInstance.objects.get(pk=cdb_id)

#     # TaskLogger = Logger(task_ident=tpl_id)
#     # TaskLogger.info("setting up auto startup")

#     # create the connection
#     try:
#         conn = get_cxoracle_connection(cdb=cdb)
#     except Exception as e:
#         error, = e.args
#         # TaskLogger.error(error)
#         raise Exception(error)

#     try:
#         cursor = conn.cursor()
#     except Exception as error:
#         raise Exception('task_id:', task_id, str(error))

#     # TaskLogger.info("connection established")
#     auto_pdbs = PDB_Instance.objects.filter(is_autostartup=True)
#     query = "CREATE OR REPLACE TRIGGER SYS.OPEN_PDBS \
#             AFTER STARTUP ON DATABASE \
#             BEGIN "
#     for pas in auto_pdbs:
#         query += " EXECUTE IMMEDIATE 'ALTER PLUGGABLE DATABASE " + pas.name + " OPEN'; "

#     query += "END open_pdbs"

#     try:
#         cursor.execute(query)
#     except Exception as e:
#         print('autostart error')
#         error, = e.args
#         # TaskLogger.error(error)
#         raise Exception(str(error))

#     finally:
#         cursor.close()
#         TaskLogger.info("auto start up set up successfully")

####################################################################
# PDB EXPORT/IMPORT
####################################################################

def export_cdb_pdb_task(data, task_id):
    """a task to perform async export of CDB/PDB using paramiko"""
    
    task = 'export CDB PDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        cdb = json.loads(data['cdb_data'])
        host = data['host_data']
        backup_host = data['backup_host']
        export = json.loads(data['export'])
        include_data = data['include_data']
        #must not contain special characters
        backup_path = cdb['dir_123Cluster']
        dir_name = export['ex_params'].get('db_path_name')
        pdb_data = data.get('pdb_data')
        dump_file = export['ex_params'].get("dump_file")[0]
        parfile_path = f"{backup_path}/{dump_file.replace('.dmp', '.par')}"
        oracle_home = cdb['ex_params'].get('oracle_home')
        log_name = export['ex_params'].get('log_file')[0]
        schema = export['ex_params'].get('schema')
        full = schema == 'all_schemas'
        export_log_path = export['ex_params'].get('remote_path')
        export_file_path = f"{backup_path}/{dump_file}"
        export_log_file = f"{export_log_path}/{log_name}"
        
        #follow rather available task data - for export of cdb we dont need the pdb
        if pdb_data:
            pdb = json.loads(pdb_data)
            name_to_use = pdb['name']
            user_to_use = pdb['database_user'].strip() 
            pwd_to_use = pdb['database_pwd']
            obj_type = ClusterRoles.PDB.value
        else:
            name_to_use = cdb['name']
            user_to_use = cdb['database_user'].strip()
            pwd_to_use = cdb['database_pwd']
            obj_type = ClusterRoles.CDB.value
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating directory (if not exists)'
    duration_step_start= time.perf_counter()
    
    # oracle recommend using parameter files when configuring expdp/impdp - more clean input 
    parfile_content = f"""
DIRECTORY={dir_name}
DUMPFILE="{dir_name}:{dump_file}"
LOGFILE="{dir_name}:{log_name}"
{"FULL=Y" if full else f'SCHEMAS="{schema}"' }
{"" if include_data else "CONTENT=metadata_only"}
"""

    commands = [
        f"sudo mkdir -p '{backup_path}'",
        f"echo '{parfile_content}' | sudo tee '{parfile_path}'",
        f"sudo chown -R oracle:dba '{backup_path}'"
    ]
    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return
    
    try:
        cursor = conn.cursor()
    except Exception as e:
        error = str(e.args[0])
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    if obj_type == ClusterRoles.PDB.value:
        try:
            cursor.execute("ALTER SESSION SET CONTAINER={}".format(pdb['name']))
        except Exception as e:
            error, = e.args
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
      
    try:
        # if backup_path are changed over time - this replace logical object of directory in db, not the physical backup_path 
        cursor.execute(f"CREATE OR REPLACE DIRECTORY {dir_name} AS '{backup_path}'")
        # grant read-write permissions on directory to user_to_use
        cursor.execute(f"GRANT READ, WRITE ON DIRECTORY {dir_name} TO  {user_to_use}")
    except Exception as e:
        error = str(e.args[0])
        if 'ORA-00955' in error:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    else:
        status = Severity.WARNING.value if status == Severity.WARNING.value else Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    finally:
        cursor.close()
        conn.close()
    
    # ensure ORACLE_HOME is active
    # prepare the export query
    if user_to_use == 'sys':
        admin = f'''\\"{user_to_use}/$database_pwd@{host['name']}:{cdb['port']}/{name_to_use} as sysdba\\"'''
    else:
        admin = f'''\\"{user_to_use}/$database_pwd@{host['name']}:{cdb['port']}/{name_to_use}\\"'''

    export_cmd = f"database_pwd='{pwd_to_use}'; sudo -iu oracle echo x/x | ORACLE_HOME={oracle_home} {cdb['ex_params'].get('oracle_home')}/bin/expdp {admin} PARFILE='{parfile_path}'"
    
    step = 'running export command'
    duration_step_start= time.perf_counter()
    
    try:
        jsn = json.loads(remote_command(host, export_cmd, True), strict=False)
        if int(jsn['excode']) != 0:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode']) 
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # inspect export operation log using paramiko
    # read the log file
    step = 'reading the log file'
    duration_step_start= time.perf_counter()
    
    if export_log_path.startswith('~'):
        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        tilda_path = jsn['stdout']
        export_log_path = export_log_path.replace('~', tilda_path)
    
    # save the file in the local backup object
    try:
        remote_file = f"{export_log_path}/{log_name}"
        jsn = json.loads(remote_command(host, f"sudo cat '{remote_file}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        run_log = jsn['stdout']
        export['run_log'] = run_log
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    stdout = "The export failed. Please check the log file {} for details.".format(remote_file)
    try:
        for line in run_log.splitlines():
            if any(substring in line for substring in ("ORA-", "EXP-")):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,line,None)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,line,None)
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host['name'] != backup_host['name']:
        step = f'copying export of {obj_type.upper()} to backup host'
        duration_step_start = time.perf_counter()

        jsn = transfer_files_rsync(host, backup_host, backup_path, [export_file_path])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking the backup size'
    duration_step_start= time.perf_counter()

    try:
        jsn = json.loads(remote_command(backup_host, f"sudo du -b '{export_file_path}' | cut -f1", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        size = jsn['stdout'].strip()
        backup_size = float(size)/1024

        if float(backup_size) < 0.01:
            backup_size == 0.01
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'updating export and PDB or CDB objects'
    duration_step_start= time.perf_counter()

    # was successful
    export['status'] = True

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status={1} WHERE id={2}""".format(backup_size, export['status'], export['id']))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{exit_code}}', '{0}', true) WHERE id={1}""".format(0, export['id']))
                conn.commit()
                cur.callproc('set_logfile_data', (Inventory.EXPORT.value,export['run_log'],export['id']))
                conn.commit()

                if obj_type == ClusterRoles.PDB.value:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, pdb['id']))
                elif obj_type == ClusterRoles.CDB.value:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, cdb['id']))
                
                conn.commit()
                #Never try to remove directory - since it oracle directory, it may be used not only for export.Remove the export-related files only.
                # remove export file
                remote_command(host, f"sudo rm -f '{export_file_path}'", True)
                # remove parameters file, since all the parameters read are included into log
                remote_command(host, f"sudo rm -f '{parfile_path}'", True)
                # by default - do not remove log, it might be needed for troubleshooting
                # remote_command(host, f"sudo rm -f '{export_log_file}'", True)

                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

                step = 'task finished'
                status = Severity.OK.value
                duration = 0
                logger(task_id,task_log_id,task,step,status,duration,created_by)
                step = 'null'
                status = Severity.OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

                try:
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{duration}}', '{0}', true) WHERE id={1}""".format(duration_task_end, export['id']))
                    conn.commit()
                except Exception as duration_exc:
                    error, = duration_exc.args
                    logging.error(f"{task}@{step}: {task_id}: {str(error)}")
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
  
def import_pdb_task(data, task_id):
    """a task to perform import of an exported PDB"""
    
    task = 'import PDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        export = json.loads(data['export'])
        cdb = json.loads(data['cdb_data'])
        pdb = json.loads(data['pdb_data'])
        host = data['host_data']
        backup_host = data['backup_host']
        pdb_name = pdb['name']
        user_to_use = pdb['database_user'].strip()
        dir_name = export['ex_params'].get('db_path_name')
        #must not contain special characters
        backup_path = cdb['dir_123Cluster']
        dump_file = export['ex_params'].get('dump_file')[0]
        parfile_path = f"{backup_path}/{dump_file.replace('.dmp', '.par')}"
        schema = export['ex_params'].get('schema')
        oracle_home = cdb['ex_params'].get('oracle_home')
        import_log_path = export['ex_params'].get('remote_path')
        export_file_path = f"{backup_path}/{dump_file}"
        full = schema == 'all_schemas'
        # create a log file for the import
        log_name = f"imp_{export['ex_params'].get('log_file')[0].strip('/')}"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host['name'] != backup_host['name']:
        step = 'copying export of PDB to PDB host'
        duration_step_start = time.perf_counter()

        jsn = transfer_files_rsync(backup_host, host, backup_path, [f"{backup_path}/{dump_file}"])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'setting the container'
    duration_step_start= time.perf_counter()
    
    # 1. connect to the host
    error, conn = get_cxoracle_connection(cdb=cdb, host=host)
    if error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
        return

    try:
        cursor = conn.cursor()
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    # 2. set the current container
    try:
        cursor.execute("ALTER SESSION SET CONTAINER={}".format(pdb_name))
    except Exception as e:
        error, = e.args
        status = Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
    else:
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'creating directory (if not exists)'
    duration_step_start= time.perf_counter()
    
    try:
        # oracle recommend using parameter files when configuring expdp/impdp - more clean input 
        parfile_content = f"""
DIRECTORY={dir_name}
DUMPFILE="{dir_name}:{dump_file}"
LOGFILE="{dir_name}:{log_name}"
{"FULL=Y" if full else f'SCHEMAS="{schema}"' }
"""
        
        commands = [
            f"echo '{parfile_content}' | sudo tee '{parfile_path}'",
            f"sudo chown -R oracle:dba '{backup_path}'",
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict = False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        # if backup_path are changed over time - this replace logical object of directory in db, not the physical backup_path 
        cursor.execute(f"CREATE OR REPLACE DIRECTORY {dir_name} AS '{backup_path}'")
        # grant read-write permissions on directory to user_to_use
        cursor.execute(f"GRANT READ, WRITE ON DIRECTORY {dir_name} TO  {user_to_use}")
    except Exception as e:
        error = str(e.args[0])
        if 'ORA-00955' in error:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    else:
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if export['ex_params'].get('schema') != 'all_schemas':
        step = 'dropping user'
        duration_step_start= time.perf_counter()

        # 3. drop the previous user
        try:
            cursor.execute("DROP USER {} CASCADE".format(schema))
        except Exception as e:
            error, = e.args
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        else:
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
  
    # get the command output
    step = 'running import command'
    duration_step_start= time.perf_counter()
    
    try:
        cursor.close()
        conn.close()
        
        # 6. prepare the import query
        # ensure ORACLE_HOME is active 
        admin = f'''\\"{user_to_use}/$database_pwd@{host['name']}:{cdb['port']}/{pdb_name} {'as sysdba' if user_to_use == 'sys' else ''}\\"'''
        import_cmd = f"database_pwd='{pdb['database_pwd']}'; sudo -iu oracle ORACLE_HOME={oracle_home} {oracle_home}/bin/impdp {admin} PARFILE='{parfile_path}'"
        
        jsn = json.loads(remote_command(host, import_cmd, True), strict=False)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    # 7. check for errors
    if int(jsn['excode']) != 0:
        status = Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        # step = 'null'
        # status = Severity.NOT_OK.value
        # duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        # logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode']) 
        # return
    else:
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # inspect import operation log using paramiko
    # read the log file
    step = 'reading the log file'
    duration_step_start= time.perf_counter()
    
    if import_log_path.startswith('~'):
        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        tilda_path = jsn['stdout']
        import_log_path = import_log_path.replace('~', tilda_path)
     
    import_log_file = f"{import_log_path.rstrip('/')}/{log_name}"
    
    stdout = "The import completed with errors. Please check the log file {} for details.".format(import_log_file)
    
    # save the file in the local backup object
    try:
        jsn = json.loads(remote_command(host, f"sudo cat '{import_log_file}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        run_log = jsn['stdout']
        export['run_log'] = run_log
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,'',str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,'',str(error),1)
        return
    
    errors = ''
    try:
        for line in run_log.splitlines():
            if any(substring in line for substring in ("ORA-", "EXP-")) and 'exists' not in line:
                errors = f"{errors}{line}\n"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,str(error),1)
        return
    
    status = Severity.WARNING.value if errors else Severity.OK.value 
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,errors,1) if errors else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'updating export and PDB objects'
    duration_step_start= time.perf_counter()

    # was successful
    export['status'] = True

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET status=%s WHERE id=%s""", [export['status'], export['id']])
                conn.commit()
                cur.callproc('set_logfile_data', (Inventory.EXPORT.value,export['run_log'],export['id']))
                conn.commit()
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, pdb['id']))
                conn.commit()
        
        
        #Never try to remove directory - since it oracle directory, it may be used not only for export.Remove the export files only.
        # remove export file
        remote_command(host, f"sudo rm -f '{export_file_path}'", True)
        # remove parameters file, since all the parameters read are included into log
        remote_command(host, f"sudo rm -f '{parfile_path}'", True)
        # by default - do not remove log, it might be needed for troubleshooting
        # remote_command(host, f"sudo rm -f '{import_log_file}'", True)
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def import_cdb_task(data, task_id):
    """a task to perform import of an exported CDB"""
    
    task = 'import CDB task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0 
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start= time.perf_counter()

    try:
        export = json.loads(data['export'])
        cdb = json.loads(data['cdb_data'])
        user_to_use = cdb['database_user'].strip()
        host = data['host_data']
        backup_host = data['backup_host']
        #must not contain special characters
        backup_path = cdb['dir_123Cluster']
        dir_name = export['ex_params'].get('db_path_name')
        logname =  f"imp_{export['ex_params'].get('log_file')[0].strip('/')}"
        dump_file = export['ex_params'].get('dump_file')[0]
        parfile_path = f"{backup_path}/{dump_file.replace('.dmp', '.par')}"
        schema = export['ex_params'].get('schema')
        oracle_home = cdb['ex_params'].get('oracle_home')
        import_log_path = export['ex_params'].get('remote_path')
        export_file_path = f"{backup_path}/{dump_file}"
        full = schema == 'all_schemas'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host['name'] != backup_host['name']:
        step = 'copying export of CDB to CDB host'
        duration_step_start = time.perf_counter()

        jsn = transfer_files_rsync(backup_host, host, backup_path, [f"{backup_path}/{dump_file}"])
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        
    step = 'creating directory (if not exists)'
    duration_step_start= time.perf_counter()
    
    try:
        # oracle recommend using parameter files when configuring expdp/impdp - more clean input 
        parfile_content = f"""
DIRECTORY={dir_name}
DUMPFILE="{dir_name}:{dump_file}"
LOGFILE="{dir_name}:{logname}"
{"FULL=Y" if full else f'SCHEMAS="{schema}"' }
"""
        commands = [
            f"echo '{parfile_content}' | sudo tee '{parfile_path}'",
            f"sudo chown -R oracle:dba '{backup_path}'",
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict = False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
        # 1. connect to the host
        error, conn = get_cxoracle_connection(cdb=cdb, host=host)
        if error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,conn,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,conn,1)
            return

        try:
            cursor = conn.cursor()
        except Exception as e:
            error, = e.args
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        # if backup_path are changed over time - this replace logical object of directory in db, not the physical backup_path   
        cursor.execute(f"CREATE OR REPLACE DIRECTORY {dir_name} AS '{backup_path}'")
        # grant read-write permissions on directory to user_to_use
        cursor.execute(f"GRANT READ, WRITE ON DIRECTORY {dir_name} TO  {user_to_use}")
    except Exception as e:
        error = str(e.args[0])
        if 'ORA-00955' in error:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if not full:
        step = 'preparing export user'
        duration_step_start= time.perf_counter()

        # 3. create user if not exist
        try:
            cursor.execute("CREATE USER {}".format(schema))
        except Exception as e:
            error, = e.args
            error = str(error)
            #ORA-01920 is ok ,"exists/not exists" syntax supported only since oracle 23c
            if error.find('ORA-01920') == -1:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
                return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # get the command output
    step = 'running import command'
    duration_step_start= time.perf_counter()
    
    try:
        # 5. close the db connection
        cursor.close()
        conn.close()

        # 6. prepare the import query
        # create a log file for the import
        # ensure ORACLE_HOME is active 
      
        admin = f'''\\"{user_to_use}/$database_pwd@{host['name']}:{ cdb['port']}/{cdb['name']} {'as sysdba' if user_to_use == 'sys' else ''}\\"''' 
            
        import_cmd = f"database_pwd='{cdb['database_pwd']}'; sudo -iu oracle ORACLE_HOME={oracle_home} {oracle_home}/bin/impdp {admin} PARFILE='{parfile_path}'"
    
        jsn = json.loads(remote_command(host, import_cmd, True), strict=False)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
   
    if import_log_path.startswith('~'):
        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        tilda_path = jsn['stdout']
        import_log_path = import_log_path.replace('~', tilda_path)

    client = create_ssh_connection(host['name'], host['username'], host['password'], host['private_key'], int(host['ssh_port']))
    if isinstance(client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(client[1]),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    # inspect import operation log using paramiko
    # read the log file
    step = 'reading the log file'
    duration_step_start= time.perf_counter()
    
    import_log_file = f"{import_log_path.rstrip('/')}/{logname}"

    stdout = "The import completed with errors. Please check the log file {} for details.".format(import_log_file)
    
    # save the file in the local backup object
    try:
        jsn = json.loads(remote_command(host, f"sudo cat '{import_log_file}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        run_log = jsn['stdout']
        export['run_log'] = run_log
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,str(error),1)
        return
    
    errors = ''
    try:
        for line in run_log.splitlines():
            if any(substring in line for substring in ("ORA-", "EXP-")) and 'exists' not in line:
                errors = f"{errors}{line}\n"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,'',str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,'',str(error),1)
        return

    status = Severity.WARNING.value if errors else Severity.OK.value 
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,errors,1) if errors else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'updating export and CDB objects'
    duration_step_start= time.perf_counter()

    # was successful
    export['status'] = True

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET status=%s WHERE id=%s""", [export['status'], export['id']])
                conn.commit()
                cur.callproc('set_logfile_data', (Inventory.EXPORT.value,export['run_log'],export['id']))
                conn.commit()
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, cdb['id']))
                conn.commit()
        
        #Never try to remove directory - since it oracle directory, it may be used not only for export.Remove the export-related files only.
        # remove export file
        remote_command(host, f"sudo rm -f '{export_file_path}'", True)
        # remove parameters file, since all the parameters read are included into log
        remote_command(host, f"sudo rm -f '{parfile_path}'", True)
        # by default - do not remove log, it might be needed for troubleshooting
        #remote_command(host, f"sudo rm -f '{import_log_file}'", True)
    except Exception as e:
        error, = e.args
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

'''
@shared_task(bind=True)
def delete_pdb_export_task(self, cdb_data, export_id, tpl_id):
    """a task that will remove the export after export.keep_days expires"""
    print("delete_pdb_export_task")
    tpl = TaskProgressLog.objects.get(pk=tpl_id)
    tpl.celery_task_id = self.request.id
    tpl.save()

    # create a logger object set with user and task
    TaskLogger = Logger(task_ident=tpl.id)
    TaskLogger.info("starting remove expired export task")

    # extract the args
    args = eval(tpl.args)
    export = PDBExport.objects.get(pk=export_id)
    pdb_name = args['pdb_name']
    # deserialize the container
    cdb = node_from_json(cdb_data, 'CDBInstance')

    TaskLogger.info("extracted task arguments")

    # 1. connect to the host
    client = create_ssh_connection(cdb.host, cdb.user, cdb.pwd, cdb.ssh_port)
    if isinstance(client, tuple):
        TaskLogger.error(str(client))
        raise Exception(tpl.id, str(client))

    TaskLogger.info("connected to server")

    dump_file = export.remote_path + "/" + export.dump_file
    dump_file = re.sub("//", "/", dump_file)
    log_file = export.remote_path + "/" + export.log_file
    log_file = re.sub("//", "/", log_file)

    jsn = json.loads(remote_command(
        client, ("rm {} {}; ".format(dump_file, log_file)), True), strict=False)

    TaskLogger.info("executed remove dump and log files")

    # check for errors
    if int(jsn['excode']) != 0:
        print("the command returned " + str(jsn['excode']))
        print("the error:")
        print(jsn['stderr'])
        TaskLogger.warning("the script returned " + str(jsn['excode']))
    else:
        TaskLogger.info("export files deleted")

    client.close()
    TaskLogger.info("server connection closed")

    return True

####################################################################
# performance tasks
####################################################################

@shared_task
def pdb_heartbeat_task(host, port, pdb_data, standby_id=None):
    """a task to check remote pluggable database connectivity, connection speed and open mode
    the task will be called by celery every 15 seconds and the result will be saved in django model"""
    try:
        system = User.objects.get(username='system')
    except:
        system = User.objects.create(username='system', is_staff=True, is_superuser=True, email='support@123cluster.com')
    
    try:
        pdb = node_from_json(pdb_data, 'PDB_Instance')
    except Exception as e:
        error = str(e.args[0])
        Logger()._error(
            f"{error} unable to deserialize objects for heartbeat", user=system.pk)
        print(f"{error} unable to deserialize objects for heartbeat")
        raise Exception(0, error)
    else:
        start = datetime.now().replace(tzinfo=None)

        if not standby_id:
            node_type = 'PDB_Instance'
            node_id = pdb.id
        else:
            node_type = 'OracleStandby'
            node_id = standby_id
        try:
            conn = get_cx_pdb_connection(host=host, port=port, pdb=pdb)
        except Exception as e:
            # failure
            error, = e.args
            Logger()._error(str(error)[:90], user=system.pk)
            pdb.is_started = False
            pdb.save()
            HeartBeat.objects.create(
                node_type=node_type, node_id=node_id)
            return False
        else:
            # success
            end = datetime.now().replace(tzinfo=None)
            cur = conn.cursor()
            try:
                cur.execute(
                    "select 1 from dual")
                cur.execute(
                    "SELECT open_mode FROM v$pdbs WHERE name='{}'".format(pdb.name))
            except Exception as e:
                # failure
                error, = e.args
                Logger()._error(str(error)[:99], user=system.pk)
                pdb.is_started = False
                pdb.save()

                HeartBeat.objects.create(
                    node_type=node_type, node_id=node_id)
                return False
            else:
                speed = end - start
                open_mode_result = cur.fetchone()
                open_mode = "" if open_mode_result is not None else open_mode_result[0]
                HeartBeat.objects.create(node_type=node_type, node_id=node_id,
                                         connectivity=True, connection_speed=speed, open_mode=open_mode)
                pdb.is_started = True
                pdb.save()

        finally:
            conn.close()

    return True

@shared_task
def pdb_standby_lag_task(cdb_data, pdb_data):
    """measure lag between PDB and it's standbys"""

    sys_user = User.objects.get(username='system')

    cdb = node_from_json(cdb_data, 'CDBInstance')
    pdb = node_from_json(pdb_data, 'PDB_Instance')

    standbys = PluggableStandby.objects.filter(pdb=pdb)
    # 1. connect to the container
    try:
        conn = get_cxoracle_connection(0, cdb=cdb)
    except Exception as e:
        error = str(e.args[0])[:99]
        # failure establishing connection
        Logger()._error(error, user=sys_user.pk)
        print(f"unable to connect to {cdb}")
        raise Exception(0, error)
    else:
        try:
            cursor = conn.cursor()
        except Exception as error:
            raise Exception('task_id:', task_id, str(error))

    # 2. set the PDB as the container
    try:
        cursor.execute(f"alter session set container={pdb.name}")
    except Exception as e:
        error = str(e.args[0])[:99]
        Logger()._error(error, user=sys_user.pk)
        raise Exception(0, error)

    # 3. create a table
    try:
        cursor.execute("create table t_123_cluster_master_time (master_time date)")
    except Exception as e:
        pass

    # 4. delete the previous cursor
    try:
        cursor.execute("delete t_123_cluster_master_time")
    except Exception as e:
        pass

    # 5. insert a value to the table in the PDB
    try:
        cursor.execute("insert into t_123_cluster_master_time values(sysdate)")
    except Exception as e:
        error = str(e.args[0])[:99]
        Logger()._error(error, user=sys_user.pk)
        raise Exception(0, error)

    # 6. get the time at the primary and close the connection
    try:
        cursor.execute("select master_time from t_123_cluster_master_time")
    except Exception as e:
        error = str(e.args[0])[:99]
        Logger()._error(error, user=sys_user.pk)
        raise Exception(0, error)

    else:
        server_time = cursor.fetchone()[0]
        conn.close()

    for standby in standbys:
        try:
            standby_conn = get_cxoracle_connection(0, cdb=standby.standby)
        except Exception as e:
            error = str(e.args[0])[:99]
            Logger()._error(error, user=sys_user.pk)
            print(f"unable to connect to {standby}")
            StandbyLagTime.objects.create(
                standby=standby, server_time=server_time, lag_state="red")
            break
        else:
            standby_cursor = standby_conn.cursor()
            standby_cursor.execute(f"alter session set container={pdb.name}")
            try:
                standby_cursor.execute("select master_time from t_123_cluster_master_time")
            except Exception as e:
                error = str(e.args[0])[:99]
                Logger()._error(error, user=sys_user.pk)
                StandbyLagTime.objects.create(
                    standby=standby, server_time=server_time, lag_state="red")
                break
            else:
                standby_time = standby_cursor.fetchone()[0]
                lag = server_time - standby_time
                if hasattr(lag, 'minutes'):
                    if lag.minutes > 5:
                        lag_state = 'red'
                elif hasattr(lag, 'seconds'):
                    if lag.seconds > 300:
                        lag_state = 'red'
                    elif lag.seconds > 60:
                        lag_state = 'orange'
                    else:
                        lag_state = 'green'

                StandbyLagTime.objects.create(
                    standby=standby, server_time=server_time, lag=lag, lag_state=lag_state)

    return True

# insert into t_123_cluster_master_time values(sysdate)
# Commit
# get the time at the primary
# select master_time from t_123_cluster_master_time

# get the time at the standby
# select master_time from t_123_cluster_master_time

# The lag is master_time(from primary) - master_time(from standby)
# Standby icon + line colors
# orange if lag > 1 minute
# red if lag > 5 minutes'''

####################################################################
# Connection tasks
####################################################################

def get_cxoracle_connection(**kwargs):
    """create a cx-oracle api connection and return it"""

    cdb = kwargs.pop('cdb')
    host = kwargs.pop('host')
    mode = kwargs.pop('mode', None)
    if not mode:
        mode = oracledb.SYSDBA

    conn = None
    try:
        dsn_tns = oracledb.makedsn(host['name'], cdb['port'], cdb['name'])
        conn = oracledb.connect(user=cdb['database_user'], password=cdb['database_pwd'], dsn=dsn_tns, mode=mode, tcp_connect_timeout=Timeout.SERVICE_ORACLE.value)
    except Exception as e:
        error, = e.args
        return (True, str(error))
    
    return (False, conn)

'''@shared_task
def get_cx_pdb_connection(**kwargs):
    """create a cx-oracle api connection for PDBs and return it"""

    print("get_cx_pdb_connection called")
    host = kwargs.pop('host')
    port = kwargs.pop('port')
    pdb = kwargs.pop('pdb')

    dsn_tns = oracledb.makedsn(host, port, service_name=pdb.name)

    try:
        conn = oracledb.connect(user=pdb.database_user, password=pdb.database_pwd, dsn=dsn_tns)
    except Exception as e:
        error, = e.args
        Logger()._error("connection parameters: db_user {} db_pwd {} dsn_tns {}".format(pdb.database_user, pdb.database_pwd, dsn_tns))
        Logger()._error(error)
        raise Exception(error)
    else:
        return conn

def get_task_connection(**kwargs):
    print("function get_task_connection called")
    cdb = kwargs.pop('cdb')
    print("cdb parameter {} ".format(cdb.id))
    dsn_tns = oracledb.makedsn(cdb.host, cdb.port, cdb.service_name)
    print("trying to create a connection")
    try:
        conn = oracledb.connect(user=cdb.database_user, password=cdb.database_pwd, dsn=dsn_tns, mode=oracledb.SYSDBA)
    except Exception as e:
        print("connection exception")
        error, = e.args
        print(error)
        return error
    return conn

####################################################################
"""Callback tasks"""
####################################################################

####################################################################
# Test tasks
####################################################################

@shared_task(bind=True)
def fail_task(self, tpl_id):
    raise Exception(tpl_id, 'this happened')

@shared_task(bind=True)
def win_task(self, pdb_str, tpl_str):
    pass
    # print(pdb_str, tpl_str)
    # pdb = deserializer_utility(pdb_str)
    # tpl = task_deserializer_utility(tpl_str)
    # print(tpl)
    # tpl = TaskProgressLog.objects.get(pk=tpl_id)
    # tpl.celery_task_id = self.request.id
    # tpl.state = AsyncResult(tpl.celery_task_id).state
    # # tpl.save()
    # time.sleep(5)'''