# BELONG TO SERVICE TASK
import os, json, time, uuid, pytz, re, logging, socket
from math import inf as positive_infinity
import paramiko
import psycopg2, psycopg2.extras
from datetime import datetime
from enums import OperatingSystem, SupportedOS, DBType, ClusterRoles, RestoreOptions, Status, Severity, Inventory, Schedule, LicenseType, Packages, Repository, AESMode, RSAKeySize
from enum import Enum
from tasks_core import insert_task, local_command, init_container, format_for_shell, remote_command, create_ssh_connection, get_postgres_connection, logger, find_tilda, check_port_existence, check_os_version, get_date_modified_folder,\
     connect_to_postgres_database, update_failed_object, install_offline_remote_to_remote, check_remote_directory_exists, check_space, generate_complex_password, TASK_TMP_FILES_DIR, \
     transfer_files_rsync, transfer_ssl_certificates, write_to_local_file, create_ec2_machine, get_log_rotate_config, TASK_CERTIFICATE_FILES_DIR, find_local_tilda, check_local_os_version,  check_local_directory_exists, \
     sql_file_executor

def bind_pg_stat_statements(hostname, node):
    try:
        conn = connect_to_postgres_database(hostname, 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
        if isinstance(conn, tuple):
            return str(conn[1])
        
        error = sql_file_executor("create func for pgss.txt", conn)
        if error != '':
            return error
        
        error = sql_file_executor("create tables for pgss.txt", conn)
        if error != '':
            return error
        
        conn.close()

    
        return ''
    
    except Exception as e:
        return str(e)

class RepmgrVersions(Enum):
    REDHAT_PG_10_DEFAULT = '5.3.2'
    REDHAT7 = '5.4.0'
    REDHAT_PG_LEGACY = '5.4.1'
    LATEST = '5.5.0'

#####################################################################
# POSTGRES TASKS 
#####################################################################

def check_pg_version_exists(host, pg_version, distributor):
    cmd = f"sudo rpm -qa | grep 'postgresql{pg_version}'" if OperatingSystem.CENTOS.value in distributor else f"sudo dpkg -l | grep -E '^ii\s+postgresql-{pg_version}'"
    jsn = json.loads(remote_command(host, cmd, True), strict=False)
    return jsn['stdout'] != ''

PHYSICAL_REPLICATION_SLOT = "slot_123cluster"

def is_ip(hostname):
    return bool(re.search(r'\b(?:[0-9]{1,3}\.){3}[0-9]{1,3}\b', hostname))

def generate_ssl_certificates(host, node, AESMode=AESMode.AES_128.value, RSAKeySize=RSAKeySize.RSA_2048.value, days=3650, certificate_authentication=False, ssl_backup_host=None):
    """generate certificates for ssl connections support\n
       cerificate_authentication - whether to generate client ssl certificates\n
       ssl_backup_host - host to backup node ssl files into
    """
    # TODO: -subj

    try:
        password = node['database_pwd'].strip()
        # generate in data directory, like are stated in postgresql docs
        cert_root = node['dir_123Cluster'].strip()
        # for compatibility (because may be ip), if use dns - name from /etc/hostname are preferred
        hostname = host['name']
        node_name = node['name'].strip()
        database_user = node['database_user'].strip()
        # sometimes (on centos 9, for example) - we could be prohibited from write into /tmp, better use path under ours control
        # (and include host_port to know which target are failed if something go wrong)
        passphrase_file = f'{TASK_TMP_FILES_DIR}/{node_name}/123cluster_passphrase.txt'
        server_key_path = f"{cert_root}/server.key"
        client_key_path = f"{cert_root}/client.key"
        server_cert_path = f"{cert_root}/server.crt"
        client_sign_request = f"{cert_root}/client.csr"
        client_cert_path = f"{cert_root}/client.crt"
        ca_path = f"{cert_root}/root.crt"

        commands = [
            f"sudo mkdir -p '{os.path.dirname(passphrase_file)}'",
            f"echo '{password}' | sudo tee '{passphrase_file}'",
            f"sudo openssl genrsa -{AESMode} -passout file:{passphrase_file} {RSAKeySize} | sudo tee '{server_key_path}'",
            f"sudo openssl rsa -in '{server_key_path}' --passin file:{passphrase_file} -out '{server_key_path}'",
            f"sudo chmod 0400 '{server_key_path}'",
            f"sudo chown postgres:postgres '{server_key_path}'",
            # here we both creating signing request and self-sign the certificate in same command (enough for ca certificate)
            f'''sudo openssl req -new -key '{server_key_path}' -days {days} -out '{server_cert_path}' -x509 -subj "/C=IL/ST=Test/L=Tel-Aviv/O=Test/OU=Test/CN={hostname}/emailAddress=Test@test.com"''',
            # for simplicity - server certificate = ca certificate (and always preserve permissions)
            f"sudo rsync -a '{server_cert_path}' '{ca_path}'",
        ]
        if certificate_authentication:
            commands += [
                f"sudo openssl genrsa -{AESMode} -passout file:{passphrase_file} {RSAKeySize} | sudo tee '{client_key_path}'",
                f"sudo openssl rsa -in '{client_key_path}' --passin file:{passphrase_file} -out '{client_key_path}'",
                # client certificate should be signed by ours CA certificate and key, so now writing separate signing request
                f'''sudo openssl req -new -key '{client_key_path}' -out '{client_sign_request}' -subj "/C=IL/ST=Test/L=Tel-Aviv/O=Test/OU=Test/CN={database_user}/emailAddress=Test@test.com"''',
                # then sign the client certificate (keep sha256?) by ours CA certificate and key
                f"sudo openssl x509 -req -in '{client_sign_request}' -CA '{ca_path}' -CAkey '{server_key_path}' -CAcreateserial -out '{client_cert_path}' -days 365 -sha256",
                f"sudo chmod 0400 '{client_key_path}'",
                f"sudo chown postgres:postgres '{client_key_path}'",
            ]

        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                return jsn
        
        if ssl_backup_host:
            jsn = transfer_ssl_certificates(node, host, ssl_backup_host)
            if jsn['stderr'] != '':
                return jsn
            
            ssl_hostname = ssl_backup_host['name'].strip()
            task_service_hostname = socket.getfqdn()
            task_service_ip = socket.gethostbyname(task_service_hostname)
            if ssl_hostname not in task_service_hostname and ssl_hostname not in task_service_ip:
                ssl_backup_path = f"{TASK_CERTIFICATE_FILES_DIR}/{hostname}_{node['port']}"
                # just directly fetch by python ssh to RAM, 
                # since node host might not have sftp server configured and task service os might not have built-in scp/rsync 
                jsn = json.loads(remote_command(host, f"sudo cat '{ca_path}'", True), strict=False)
                if jsn['stderr'] != '':
                    return jsn
                
                jsn = write_to_local_file(f"{ssl_backup_path}/{os.path.basename(ca_path)}", jsn['stdout'])
                if jsn['stderr'] != '':
                    return jsn
                
                # write client certificate and key only if verify-full ssl mode are requested
                if certificate_authentication:
                    jsn = json.loads(remote_command(host, f"sudo cat '{client_cert_path}'", True), strict=False)
                    if jsn['stderr'] != '':
                        return jsn
                    
                    jsn = write_to_local_file(f"{ssl_backup_path}/{os.path.basename(client_cert_path)}", jsn['stdout'])
                    if jsn['stderr'] != '':
                        return jsn
                    
                    jsn = json.loads(remote_command(host, f"sudo cat '{client_key_path}'", True), strict=False)
                    if jsn['stderr'] != '':
                        return jsn
                    
                    jsn = write_to_local_file(f"{ssl_backup_path}/{os.path.basename(client_key_path)}", jsn['stdout'])
                    if jsn['stderr'] != '':
                        return jsn
                    
        # cleanup - wait for finish, dont care if not successful
        remote_command(host, f"sudo rm -rf '{passphrase_file}'", True)

        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        return jsn
    except Exception as e:
        return {'stdout': '', 'stderr': str(e), 'excode': 9}

def setup_postgresql_ssl(distributor, host, node, AESMode=AESMode.AES_128.value, RSAKeySize=RSAKeySize.RSA_2048.value, days=3650, certificate_authentication=False, ssl_backup_host=None) -> dict:
    """configure instance to support ssl connections\n
       cerificate_authentication - whether to enforce authentication by client certificate\n
       ssl_backup_host - host to backup node ssl files into
    """
    
    try:
        # now exact certificate generation are in helper function, because will be used more than once
        jsn = generate_ssl_certificates(host, node, AESMode, RSAKeySize, days, certificate_authentication, ssl_backup_host)
        if jsn['stderr'] != '':
            # if fail to generate certificates - interrupt immediately
            return jsn
        
        data_directory = node['dir_123Cluster'].strip()
        database_user = node['database_user'].strip()
        postgres_version = node['version']
        postgres_bin = find_postgres_bin(host, postgres_version, distributor)
        # stick to more clarity when we are not forced to use legacy pg_hba.conf syntax
        strict_ssl = 'verify-full'  if int(postgres_version) > 11 else '1'
        pg_hba_lines = f"""hostssl	 {database_user}         all          0.0.0.0/0    		{f'cert     		clientcert={strict_ssl}' if certificate_authentication else 'md5'}
hostssl   repmgr         repmgr          0.0.0.0/0        md5
hostssl	 all         all          0.0.0.0/0    		{f'cert     		clientcert={strict_ssl}' if certificate_authentication else 'md5'}"""
        ssl_config_parts = [
            f"echo \"ssl = on\n"
            f"ssl_ca_file = '{data_directory}/server.crt'\n"
            f"ssl_cert_file = '{data_directory}/root.crt'\n"
            # isnt it the empty string by default (anyway - is good for clarity)?
            f"ssl_crl_file = ''\n"
            f"ssl_key_file = '{data_directory}/server.key'\n"
            f"ssl_ciphers = 'HIGH:MEDIUM:+3DES:!aNULL' # allowed SSL ciphers\n"
            f"ssl_prefer_server_ciphers = on\" | sudo tee -a '{data_directory}/postgresql.conf'",
            # allow ssl connections from any origin (sslmode = verify-full if certificate authentication are enforced)
            f"echo '{pg_hba_lines}' | sudo tee -a '{data_directory}/pg_hba.conf'"
        ]
        
        if OperatingSystem.CENTOS.value in distributor.lower():
            ssl_config_parts.append(f"sudo systemctl restart postgresql-{postgres_version}_{node['port']}.service")
        elif OperatingSystem.UBUNTU.value in distributor.lower():
            ssl_config_parts.append(f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{data_directory}' restart -l /dev/null")

        for command in ssl_config_parts:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                return jsn

        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        return jsn
    except Exception as e:
        return {'stdout': '', 'stderr': str(e), 'excode': 9}

def pg_wal_replay_resume(host, node, distributor, max_attempts=60):
    attempt = 0
    postgres_bin = find_postgres_bin(host, node['version'], distributor)

    while attempt < max_attempts:
        try:
            jsn = json.loads(remote_command(host, f"sudo -iu postgres {postgres_bin}/pg_isready -h '{host['name']}' -p '{node['port']}' -U '{node['database_user']}'", True), strict=False)
            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                return jsn['stderr']

            if 'accepting connections' in jsn['stdout'].lower():
                break
            elif 'no response' in jsn['stdout'].lower():
                return f"No response from the PostgreSQL server {host['name']}:{node['port']}"

            attempt += 1
            time.sleep(1)
        except Exception as error:
            return str(error)

    if attempt == max_attempts:
        return "Timeout: Unable to resume WAL replay within the specified time."

    try:
        conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
        if isinstance(conn, tuple):
           return str(conn[1])

        with conn:
            with conn.cursor() as cur:
                cur.execute('SELECT pg_wal_replay_resume()')
                return ''
    except Exception as error:
        strerror = str(error)
        if 'recovery is not in progress' not in strerror.lower():
            return strerror
        
        return ''

def install_postgres_standalone_task(data, task_id): 
    """
    installing postgres and creating superuser - before installation should be installed:
    # yum install -y redhat-lsb-core which sudo
    """
    
    task = 'install postgres standalone task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
        global PHYSICAL_REPLICATION_SLOT
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        pg_node = json.loads(data['node_data'])
        host = json.loads(data['host_data'])
        # machine where app
        app_host = json.loads(data['app_host'])
        app_hostname = data['app_hostname']
        create_run_linux = data['create_run_linux']
        postgres_version = int(pg_node['version'])
        # repository = network or local
        repository = data['repository']
        max_connections = pg_node["ex_params"].get('max_connections', 100)
        ssl = pg_node['ex_params'].get('ssl', False)
        certificate_authentication = pg_node['ex_params'].get('certificate_authentication', False)
        org_id = pg_node['org_id']
        offline = True if Repository.LOCAL.value in repository else False
        app_host_pg_password = data['app_host_pg_password']
        log_archive_path = data['log_archive_path']
        proxmox_container = pg_node['ex_params'].get('proxmox')
        aws_host = host['ex_params'].get('aws_username')
        host_exists = True
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment, AWS EC2 VM WOULD not exist at the moment
            if proxmox_container or aws_host :
                host_exists = False
                distributor = pg_node['ex_params'].get('distributor')
                release = pg_node['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(pg_node['id'])
                return
        
        version_exists = check_pg_version_exists(host, postgres_version, distributor)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if not host_exists:

        try:

            if proxmox_container:

                step = 'creating proxmox container'
                duration_step_start = time.perf_counter()

                service_data = json.loads(data['service_data'])
                ram_gb = data.get('ram_gb')
                cpu_cores = data.get('cpu_cores')
                disk_gb = data.get('disk_gb')
                jsn = init_container(service_data, host, distributor, release, ram_gb, cpu_cores, disk_gb)
                if jsn['stderr'] != '' or jsn['excode'] != 0:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(pg_node['id'])
                    return
            

                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

            elif aws_host:
                step = 'creating ec2 machine'
                duration_step_start = time.perf_counter()

                disk_gb = data.get('disk_gb')

                jsn = create_ec2_machine(host, pg_node['ex_params'].get('distributor'), pg_node['ex_params'].get('release'), disk_gb, pg_node['sizing'])
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(pg_node['id'])
                    return
            
                
                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    if not version_exists:
        step = 'installing postgresql {}'.format(str(postgres_version))
        duration_step_start = time.perf_counter()

        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host, postgres_version, DBType.POSTGRES.value, app_host)
        else:
            jsn = install_postgres_online(distributor, release, host, postgres_version)
        
        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(pg_node['id'])
            return
        elif 'warning' in error_lower:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        else:
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking psql existence'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, "sudo grep -qxF 'postgres    ALL=(ALL) NOPASSWD: ALL' '/etc/sudoers' || echo 'postgres    ALL=(ALL) NOPASSWD: ALL' | sudo tee -a '/etc/sudoers'", True), strict=False)
        if 'error' in jsn['stderr'].lower() or 'errno' in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        jsn = json.loads(remote_command(host, 'sudo which psql', True), strict=False)
        if 'error' in jsn['stderr'].lower() or 'not found' in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(pg_node['id'])
            return
        elif '/bin/psql' not in jsn['stdout'].lower() and '/sbin/psql' not in jsn['stdout'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(pg_node['id'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
         
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    data = {
        'host': host,
        'repository': repository,
        'app_hostname': app_hostname,
        'pg_node': pg_node,
        'postgres_version': postgres_version,
        'pgdata': pg_node['dir_123Cluster'],
        'distributor': distributor,
        'release': release,
        'max_connections': max_connections,
        'ssl': ssl,
        'certificate_authentication': certificate_authentication,
        'app_host': app_host
    }

    jsn = create_node(data, task, task_id, task_log_id, duration_task_start, created_by)
    if jsn['stderr'] != '':
        logging.error(f"create_node: {jsn['stderr']}")
        return

    step = 'configuring instance'
    duration_step_start = time.perf_counter()
    
    # test the superuser with psycopg2
    if certificate_authentication:
        conn = connect_to_postgres_database(host['name'], 'postgres', pg_node['database_user'], pg_node['database_pwd'].strip(), pg_node['port'], strict_ssl=True)
    else:
        conn = connect_to_postgres_database(host['name'], 'postgres', pg_node['database_user'], pg_node['database_pwd'].strip(), pg_node['port'])

    if isinstance(conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
        return
        
    try:
        cur = conn.cursor()
        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        cur.execute("CREATE EXTENSION IF NOT EXISTS pg_stat_statements")
        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
        cur.execute("SHOW wal_level")
        row = cur.fetchone()[0]

        if 'minimal' in str(row).strip().lower():
            #for replication - wal level should be >= 'replica'
            conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
            cur.execute("ALTER SYSTEM SET wal_level=replica")
            conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)

        cur.execute("SELECT slot_name FROM pg_replication_slots WHERE slot_name = %s", (PHYSICAL_REPLICATION_SLOT,))
        slot_exists = cur.fetchone()
        if not slot_exists:
            cur.execute("SELECT pg_create_physical_replication_slot(%s)", (PHYSICAL_REPLICATION_SLOT,))
        conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = "configuring user 'postgres' passwordless authentication with app host"
    duration_step_start = time.perf_counter()

    distributor_app_host, release_app_host = check_local_os_version()
    if OperatingSystem.INVALID.value in distributor:
        jsn = release
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    centos_dist_app_host = OperatingSystem.CENTOS.value in distributor_app_host
    if centos_dist_app_host:
        package_manager = 'dnf' if int(release) > 7 else 'yum'

    postgres_home_dir_app_host = '/var/lib/pgsql' if centos_dist_app_host else '/var/lib/postgresql'
    private_filepath_app_host = f'{postgres_home_dir_app_host}/.ssh/id_rsa'
    pg_password_path_app_host = f"""{TASK_TMP_FILES_DIR}/{app_host['name']}"""
    commands_app_host = [
        f'sudo {package_manager} install -q -y openssl' if centos_dist_app_host else 'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y openssl',
    ]
    if not centos_dist_app_host:
        commands_app_host.append('sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y sshpass')
    else:
        if int(release_app_host) == 7:
            commands_app_host.append(f'sudo {package_manager} install -q -y https://vault.centos.org/7.6.1810/extras/x86_64/Packages/sshpass-1.06-2.el7.x86_64.rpm')
        elif int(release_app_host) == 8:
            commands_app_host.append(f"sudo {package_manager} install -q -y https://dl.rockylinux.org/pub/rocky/8/AppStream/x86_64/os/Packages/s/sshpass-1.09-4.el8.x86_64.rpm")
        elif int(release_app_host) == 9:
            commands_app_host.append(f'sudo {package_manager} install -q -y https://mirror.stream.centos.org/9-stream/AppStream/x86_64/os/Packages/sshpass-1.09-4.el9.x86_64.rpm')
        
    commands_app_host += [
        f"sudo usermod --password $(echo '{app_host_pg_password}' | openssl passwd -1 -stdin) postgres",
        f"sudo -u postgres ssh-keygen -t rsa -P '' -f '{private_filepath_app_host}' <<< y",
    ]
    for command in commands_app_host:
        jsn = json.loads(local_command(command, True, step), strict=False)
        if jsn['stderr'] != '' and 'created directory' not in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    
    pg_password = pg_node['ex_params'].get('pg_password')
    centos_dist_node = OperatingSystem.CENTOS.value in distributor
    postgres_home_dir_node = '/var/lib/pgsql' if centos_dist_node else '/var/lib/postgresql'
    private_filepath_node = f'{postgres_home_dir_node}/.ssh/id_rsa'
    ssh_file_part_node = f"-i '{private_filepath_node}.pub'"

    if centos_dist_node:
        package_manager = 'dnf' if int(release) > 7 else 'yum'
    
    commands_node = [
        f'sudo {package_manager} install -q -y openssl' if centos_dist_node else 'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y openssl',
    ]
    if not centos_dist_node:
        commands_node.append('sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y sshpass')
    else:
        if int(release) == 7:
            commands_node.append(f'sudo {package_manager} install -q -y https://mirror.stream.centos.org/9-stream/AppStream/x86_64/os/Packages/sshpass-1.09-4.el9.x86_64.rpm')
        elif int(release) == 8:
            commands_node.append(f"sudo {package_manager} install -q -y https://dl.rockylinux.org/pub/rocky/8/AppStream/x86_64/os/Packages/s/sshpass-1.09-4.el8.x86_64.rpm")
        elif int(release) == 9:
            commands_node.append(f'sudo {package_manager} install -q -y https://vault.centos.org/7.6.1810/extras/x86_64/Packages/sshpass-1.06-2.el7.x86_64.rpm')
    
    commands_node += [
        f"sudo usermod --password $(echo '{pg_password}' | openssl passwd -1 -stdin) postgres",
        f"sudo -u postgres ssh-keygen -t rsa -P '' -f '{private_filepath_node}' <<< y",
    ]
    
    for command in commands_node:
        jsn = json.loads(remote_command(host, command, True, step), strict=False)
        if jsn['stderr'] != '' and 'created directory' not in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    
    # 1. Write password of the REMOTE HOST user to the file.
    # 2. Read password from file to stdin of ssh-copy-id.
    # 3. Provide identity file of the LOCAL HOST user to ssh-copy-id, to authenticate the user of LOCAL HOST on REMOTE HOST.
    # 4. Copy identity file of the REMOTE HOST user to the LOCAL HOST by connecting to the REMOTE HOST from the LOCAL HOST.
    jsn = json.loads(remote_command(host, f"sudo mkdir -p '{pg_password_path_app_host}' && echo '{app_host_pg_password}' | sudo tee '{pg_password_path_app_host}/pg_password' && sudo chown postgres:postgres '{pg_password_path_app_host}' && sudo -iu postgres sshpass -f '{pg_password_path_app_host}/pg_password' ssh-copy-id -p '{app_host['ssh_port']}' -o StrictHostKeyChecking=no {ssh_file_part_node} 'postgres@{app_host['name']}'", True), strict=False)
    if jsn['stderr'] != '' and 'permission denied' in jsn['stderr']:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    
    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()
    
    
    try:

        # postgresql-specific - remove the previous log archive of cluster, since it is not usable after postgresql cluster recreation
        local_command(f"sudo rm -rf '{log_archive_path}' | awk '{{print $1}}'", True)
        
        jsn = json.loads(remote_command(host, f"sudo du -sb '{pg_node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        remote_command(host, f"""echo "{get_log_rotate_config(pg_node['dir_123Cluster'])}" | sudo tee '/etc/logrotate.d/{pg_node['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host), 'node_name': pg_node['name']}, 'task': 'log_rotate_task',  'org_id': pg_node['org_id'], 'monitor_name': None,  'target_type': pg_node['type'], 'target_id': pg_node['id'], 'created_by': created_by}
        insert_task(params)

        error = bind_pg_stat_statements(host['name'], pg_node)
        if error != '':
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)


        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, postgres_version, node_size, pg_node['id'],))
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,pg_node['id'],created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host['id'],created_by,OperatingSystem.LINUX.value])
                conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)

    # host['create_task'] = created_by
    # host['ssh_port'] = int(host['ssh_port'])
    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host, 'task': 'run_linux_performance_counters', 'target_type': OperatingSystem.LINUX.value, 'target_id': host['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as error:
    #         return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    
def install_postgres_online(distributor, release, host, postgres_version) -> dict:
    supported_versions = SupportedOS[distributor.upper()].value
    supported_versions_str = ", ".join(map(str, supported_versions))

    if OperatingSystem.UBUNTU.value in distributor.lower():
        release_codename_mapping = {
            18: 'bionic',
            20: 'focal',
            22: 'jammy',
        }
        version = release_codename_mapping.get(int(release), 'unknown')
        commands = [
            'sudo apt-get update',
            'sudo apt-get install -qq -y gpg net-tools',
            "sudo wget -q -O - https://www.postgresql.org/media/keys/ACCC4CF8.asc | sudo gpg -q --batch --yes --dearmor -o '/usr/share/keyrings/pgdg-keyring.gpg'",
            f"echo 'deb [signed-by=/usr/share/keyrings/pgdg-keyring.gpg] http://apt{'-archive' if int(release) == 18 else ''}.postgresql.org/pub/repos/apt {version}-pgdg main' | sudo tee '/etc/apt/sources.list.d/pgdg.list'",
            'sudo apt-get update'
        ]
                    
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            error_lower = jsn['stderr'].lower()
            if error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do']):
                return jsn
            elif 'not found' in jsn['stderr'].lower():
                return jsn

        if postgres_version == '':
            jsn = json.loads(remote_command(host, 'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y postgresql && sudo systemctl stop postgresql && sudo systemctl disable postgresql', True), strict=False)
            error_lower = jsn['stderr'].lower()
            if error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do']):
                return jsn
        else:
            jsn = json.loads(remote_command(host, 'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y postgresql-{} && sudo systemctl stop postgresql && sudo systemctl disable postgresql'.format(str(postgres_version)), True), strict=False)
            error_lower = jsn['stderr'].lower()
            if error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['error', 'fail', 'err', 'errno'] if 'error: nothing to do' not in error_lower)):
                return jsn

    elif OperatingSystem.CENTOS.value in distributor.lower() or 'red hat' in distributor.lower():
        archive_repo = f"""[pgdg{postgres_version}-archive]
name=PostgreSQL {postgres_version} RPMs for RHEL/Rocky Linux/AlmaLinux {release}
baseurl=https://yum-archive.postgresql.org/{postgres_version}/{OperatingSystem.REDHAT.name.lower()}/{OperatingSystem.REDHAT.value}-{release}-x86_64
enabled=1
gpgcheck=1
gpgkey=https://yum.postgresql.org/keys/{f"PGDG-RPM-GPG-KEY-{OperatingSystem.REDHAT.value.upper()}" if int(postgres_version) > 10 else "RPM-GPG-KEY-PGDG"}
"""
        archive_setup = f"echo '{archive_repo}' | sudo tee '/etc/yum.repos.d/pgdg.repo'"
        package_manager = 'dnf' if int(release) > 7 else 'yum'
        if int(release) in [7, 8, 9]:
            if int(postgres_version) < 12:
                command = archive_setup
            else:
                command = f'sudo {package_manager} install -q -y https://download.postgresql.org/pub/repos/yum/reporpms/EL-{release}-x86_64/pgdg-{OperatingSystem.REDHAT.name.lower()}-repo-latest.noarch.rpm'
            jsn = json.loads(remote_command(host, command, True), strict=False)
            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                return jsn 
            elif 'command not found' in jsn['stderr']:
                return jsn
            
            if int(release) == 8:
                jsn = json.loads(remote_command(host, f'sudo {package_manager} -q -y module disable postgresql', True), strict=False)
                error_lower = jsn['stderr'].lower()
                if jsn['stderr'] != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                    return jsn

            if int(postgres_version) >= 15:
                jsn = json.loads(remote_command(host, f'sudo {package_manager} install -q -y epel-release', True), strict=False)
                error_lower = jsn['stderr'].lower()
                if jsn['stderr'] != '' and  any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                    return jsn
            
            #if postgres_version = '' - this transforms to just postgresql-server
            jsn = json.loads(remote_command(host, f'sudo {package_manager} install -q -y postgresql{postgres_version}-server rsync net-tools tar which', True), strict=False)
            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                return jsn
            
            if int(release) == 7:
                jsn = json.loads(remote_command(host, f'sudo {package_manager} install -q -y postgresql{postgres_version}-client', True), strict=False)
                error_lower = jsn['stderr'].lower()
                if jsn['stderr'] != '' and  any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                    return jsn
                
            jsn = json.loads(remote_command(host, f'sudo {package_manager} --enablerepo=appstream install -q -y postgresql{postgres_version}-contrib', True), strict=False)
            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and  any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                return jsn
        else:
            jsn = {
                'stdout': 'Unsupported release version',
                'stderr': f'Release version of {distributor.capitalize()} is not in {supported_versions_str}',
                'excode': 1
            }
            return jsn
        
        jsn = json.loads(remote_command(host, f'sudo {package_manager} list installed|grep postgres', True), strict=False)
        if 'error' in jsn['stderr'].lower() or 'errno' in jsn['stderr'].lower():
            return jsn
        elif jsn['stdout'] == '':
            return jsn
    else:
        jsn = {
            'stderr': 'Distributor is not Ubuntu or CentOS',
            'stdout': None,
            'excode': 99
        }
        return jsn

    return jsn

def find_postgres_bin(host, postgres_version, distributor) -> str:
    """find location of POSTGRES BIN"""

    postgres_bin = ''
    try:
        if OperatingSystem.CENTOS.value in distributor:
            jsn = json.loads(remote_command(host, f'sudo dirname $(rpm -ql postgresql{postgres_version}-server.x86_64|grep initdb|grep -v share)', True), strict=False)
        else:
            jsn =  json.loads(remote_command(host, f"sudo dpkg-query -L postgresql-{postgres_version} | grep '/bin$'", True), strict=False)

        if jsn['stdout'] != '':
            postgres_bin = jsn['stdout'].strip()

        if jsn['stderr'] != '' and 'permission denied' not in jsn['stderr'].lower():
            if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
                postgres_bin = f'/usr/lib/postgresql/{postgres_version}/bin'
            else:
                postgres_bin = f'/usr/lib/postgresql/{postgres_version}/bin'
    except Exception as error:
        logging.error(f'find_postgres_bin: {str(error)}')

    return postgres_bin

def find_local_postgres_bin(postgres_version, distributor) -> str:
    """find location of POSTGRES BIN"""

    postgres_bin = ''
    try:
        if OperatingSystem.CENTOS.value in distributor:
            jsn = json.loads(local_command(f'sudo dirname $(rpm -ql postgresql{postgres_version}-server.x86_64|grep initdb|grep -v share)', True), strict=False)
        else:
            jsn =  json.loads(local_command(f"sudo dpkg-query -L postgresql-{postgres_version} | grep '/bin$'", True), strict=False)

        if jsn['stdout'] != '':
            postgres_bin = jsn['stdout'].strip()

        if jsn['stderr'] != '' and 'permission denied' not in jsn['stderr'].lower():
            if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
                postgres_bin = f'/usr/lib/postgresql/{postgres_version}/bin'
            else:
                postgres_bin = f'/usr/lib/postgresql/{postgres_version}/bin'
    except Exception as error:
        logging.error(f'find_postgres_bin: {str(error)}')

    return postgres_bin


def install_haproxy_task(data, task_id): 
    """installing haproxy on standalone or cluster"""
    
    task = 'install haproxy task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_haproxy = json.loads(data['host_haproxy_data'])
        is_host_haproxy_ip = is_ip(host_haproxy['name'])
        if not is_host_haproxy_ip:
            host_haproxy_ip = socket.gethostbyname(host_haproxy['name'])
        
        node_haproxy = json.loads(data['node_haproxy_data'])
        app_host = json.loads(data['app_host'])
        # hosts_and_nodes = [{'host':..., 'node':...}, {...}, ...]
        hosts_and_nodes =[{'host': json.loads(i['host']), 'node': json.loads(i['node'])} for i in data['hosts_and_nodes']]
        node_list = [{'node_id': entry['node']['id'], 'node_port': entry['node']['port'], 'host_ip': entry['host']['name']} for entry in hosts_and_nodes]
        # repository = network or local
        repository = data['repository']
        # if host has other haproxy nodes
        has_other_haproxies = data['has_other_haproxies']
        #if haproxy package installed on host
        haproxy_exists = data['haproxy_exists']
        offline = True if Repository.LOCAL.value in repository else False
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    distributor, release = check_os_version(host_haproxy)
    if OperatingSystem.INVALID.value in distributor:
        jsn = release
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
 
    step = 'installing haproxy'
    duration_step_start = time.perf_counter()

    if not haproxy_exists:
        jsn = install_haproxy(distributor, release, host_haproxy, app_host, offline)
        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'] if 'error: nothing to do' not in error_lower and 'already installed' not in error_lower))
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

    conf_content = append_haproxy_conf(node_list, node_haproxy['port']) if has_other_haproxies else generate_haproxy_conf(node_list, node_haproxy['port'])
    commands = [
        "sudo test ! -f '/etc/haproxy/haproxy.cfg.bak' && sudo rsync -a '/etc/haproxy/haproxy.cfg' '/etc/haproxy/haproxy.cfg.bak'",
        f"echo '{conf_content}' | sudo tee {'-a' if has_other_haproxies else ''} '/etc/haproxy/haproxy.cfg'",
        "[[ $(sudo tail -n 1 '/etc/haproxy/haproxy.cfg') != '' ]] && printf '\\n' | sudo tee -a '/etc/haproxy/haproxy.cfg'",
        "sudo systemctl restart haproxy",
        "sudo systemctl status haproxy"
    ]
    for command in commands:
        try:
            jsn = json.loads(remote_command(host_haproxy, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
        
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'installing healthcheck'
    duration_step_start = time.perf_counter()

    try:
        for host_and_node in hosts_and_nodes:
            host = host_and_node['host']
            node = host_and_node['node']
            
            jsn = install_healthcheck(distributor, release, host, node, app_host, offline)
            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'] if 'error: nothing to do' not in error_lower and 'already installed' not in error_lower))
            if filter:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return  
            
            with conn:
                with conn.cursor() as cur:
                    cur.execute("SHOW config_file")
                    config_file = cur.fetchone()[0]
                    if config_file.endswith('/'):
                        config_file = config_file[:-1]
                    
                    config_dir = os.path.dirname(config_file) 
                        
                    commands = [
                        f"sudo sed -i \'/host\s*all\s*{node['database_user']}\s*{host_haproxy['name']}{chr(92) + '/32' if is_host_haproxy_ip else ''}\s*md5/d' '{config_dir}/pg_hba.conf'",
                        f"echo -e 'host\tall\t{node['database_user']}\t{host_haproxy['name']}{'/32' if is_host_haproxy_ip else ''}\tmd5' | sudo tee -a '{config_dir}/pg_hba.conf'",
                    ]
                    if not is_host_haproxy_ip:
                        commands += [
                            f"sudo sed -i \'/host\s*all\s*{node['database_user']}\s*{host_haproxy_ip}\/32\s*md5/d' '{config_dir}/pg_hba.conf'",
                            f"echo -e 'host\tall\t{node['database_user']}\t{host_haproxy_ip}/32'\tmd5' | sudo tee -a '{config_dir}/pg_hba.conf'",
                        ]
                    
                    for command in commands:
                        jsn = json.loads(remote_command(host, command, True), strict=False)
                        if jsn['stderr'] != '':
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return
                        
                        cur.execute('SELECT pg_reload_conf()')
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_host" SET status=%s, ind_deleted='false' WHERE id=%s""", (Status.OK.value, host_haproxy['id'],))
                cur.execute("""UPDATE "clusterCore_c123_node" SET status='HAProxy', severity=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, node_haproxy['id'],))
                cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"ind_haproxy":true}' WHERE cluster_id = %s AND ind_deleted = 'false' AND type != '%s';""" % (node_haproxy['cluster_id'], DBType.HAPROXY.value))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def install_haproxy(distributor, release, host, app_host, offline) -> dict:
    # offline = True or False
    try:
        if not offline:
            if OperatingSystem.UBUNTU.value in distributor.lower():
                jsn = json.loads(remote_command(host, 'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y haproxy', True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do']):
                    return jsn
            elif OperatingSystem.CENTOS.value in distributor.lower() or 'red hat' in distributor.lower():
                jsn = json.loads(remote_command(host, f'sudo {"dnf" if int(release) > 7 else "yum"} --enablerepo=appstream install -q -y haproxy', True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                    return jsn
            else:
                jsn = {
                    'stderr': 'Distributor is not Ubuntu or CentOS',
                    'stdout': None,
                    'excode': 99
                }
        else:
            jsn = install_offline_remote_to_remote(distributor, release, host, None, DBType.POSTGRES.value, app_host, Packages.HAPROXY.value)
    except Exception as error:
        jsn = {
            'stderr': str(error),
            'stdout': None,
            'excode': 9
        }
    
    return jsn

def start_haproxy_task(data, task_id): 
    """starting haproxy"""

    task = 'start haproxy task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node_ids = data['node_ids']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'starting haproxy'
    duration_step_start = time.perf_counter()

    try:
        command = "sudo systemctl start haproxy"
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
        
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET status='HAProxy', severity=%s, ind_started='true', ind_deleted='false' WHERE id=ANY(%s)""", (Severity.OK.value, node_ids,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def stop_haproxy_task(data, task_id):
    """stopping haproxy"""

    task = 'stop haproxy task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node_ids = data['node_ids']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stopping haproxy'
    duration_step_start = time.perf_counter()

    try:
        command = 'sudo systemctl stop haproxy'
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false' WHERE id=ANY(%s)""", (Severity.NOT_OK.value, Status.OK.value, node_ids,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restart_haproxy_task(data, task_id):
    """restarting haproxy"""

    task = 'restart haproxy task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node_ids = data['node_ids']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'restarting haproxy'
    duration_step_start = time.perf_counter()

    try:
        command = 'sudo systemctl restart haproxy'
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, ind_started='true' WHERE id=ANY(%s)""", (Severity.OK.value, node_ids,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def drop_haproxy_task(data, task_id): 
    """uninstall haproxy"""
    
    task = 'drop haproxy task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        haproxy_data = [{"port": i['port'], "has_other_on_host": i['has_other_on_host'], "host": json.loads(i['host'])} for i in data['haproxy_data']]
        # has other than we going to delete haproxies in cluster
        has_other_haproxies = data['has_other_haproxies']
        haproxy_ids = data['haproxy_ids']
        cluster_id = data['cluster_id']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'removing haproxy'
    duration_step_start = time.perf_counter()

    known_hosts = []

    for i in haproxy_data:
        host_haproxy = i['host']
        has_other_on_host = i['has_other_on_host']
        commands = ["sudo systemctl restart haproxy" if has_other_on_host else "sudo systemctl stop haproxy"]
        # config edit branch - 
        # can run several times per host,
        # to unregister target haproxy nodes on same host
        if has_other_on_host:
            haproxy_port = i['port']
            commands = [
                "sudo test ! -f '/etc/haproxy/haproxy.cfg.bak' && sudo rsync -a '/etc/haproxy/haproxy.cfg' '/etc/haproxy/haproxy.cfg.bak'",
                f"sudo sed -i '/^listen cluster_{haproxy_port}/,/^\\s*$/d' '/etc/haproxy/haproxy.cfg'",
            ] + commands
        # haproxy purge branch -
        # can run only once for every host,
        # only if host doesnt have haproxies other than we deleting
        else:
            if host_haproxy['name'] in known_hosts:
                # if already processed -
                # haproxy already removed from this host, skipping
                continue
            else:
                distributor, release = check_os_version(host_haproxy)
                if OperatingSystem.INVALID.value in distributor:
                    jsn = release
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return

                package_manager = f'{"dnf" if int(release) > 7 else "yum"} remove' if OperatingSystem.CENTOS.value in distributor else 'apt-get purge'
                commands +=  [
                    f'sudo {package_manager} -y haproxy',
                    "sudo rm -f '/etc/haproxy/haproxy.cfg'"
                ]

        for command in commands:
            try:
                jsn = json.loads(remote_command(host_haproxy, command, True), strict=False)
                if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in ['no such', 'not found', 'unable to locate package', 'nothing to do', 'not loaded']):
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return

        known_hosts.append(host_haproxy['name'])

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=ANY(%s)""", (Severity.NOT_OK.value, Status.OK.value, haproxy_ids,))
                if not has_other_haproxies:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"ind_haproxy":false}' WHERE cluster_id=%s""" % (cluster_id,))
                
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def create_node(data, task, task_id, task_log_id, duration_task_start, created_by) -> dict:
    """create node function"""
    # after creation, change port in postgresql.conf (if needed)

    step = 'checking parameters for node creation'
    duration_step_start = time.perf_counter()

    host = data['host']
    app_host = data['app_host']
    repository = data['repository']
    distributor = data['distributor']
    release = data['release']
    app_hostname = data['app_hostname']
    is_app_host_ip = is_ip(app_hostname)
    if not is_app_host_ip:
        app_host_ip = socket.gethostbyname(app_hostname)
    
    is_host_ip = is_ip(host['name'])
    if not is_host_ip:
        host_ip = socket.gethostbyname(host['name'])
    
    pg_node = data['pg_node']
    postgres_version = int(data['postgres_version'])
    # wal_size = data['wal_size']
    # encoding = data['encoding']
    max_connections = pg_node['ex_params'].get('max_connections', 100)
    ssl = data['ssl']
    certificate_authentication = data['certificate_authentication']
        
    pgdata = data['pgdata'].strip()
    if pgdata.startswith('~'):
        jsn = find_tilda(host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return jsn

        tilda_path = jsn['stdout']
        pgdata = pgdata.replace('~', tilda_path)

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    jsn = check_port_existence(host, pg_node['port'])
    port_existence = True if jsn['stdout'] != '' else False

    if jsn['stderr'] != '':
        return jsn

    if port_existence is False:
        step = 'creating directories for the cluster'
        duration_step_start = time.perf_counter()
    
        jsn = json.loads(remote_command(host, f"sudo mkdir -p '{pgdata}'", True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(pg_node['id'])
            return jsn

        remote_command(host, f"sudo chown -R postgres:postgres '{pgdata}'", False)
        remote_command(host, f"sudo chmod 0700 '{pgdata}'", False)

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        # step = 'comparing RAM with shared buffers'
        # duration_step_start = time.perf_counter()

        # command = "sudo free -m | grep 'Mem:' | awk '{print $2}' | xargs"
        # jsn = json.loads(remote_command(host, command, True), strict=False)
        # if jsn['stderr'] != '' or jsn['stdout'] == '':
        #     status = Severity.NOT_OK.value
        #     duration_step_end = time.perf_counter() - duration_step_start
        #     logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        #     step = 'null'
        #     status = Severity.NOT_OK.value
        #     duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        #     logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        #     update_failed_object(pg_node['id'])
        #     remote_command(host, f'sudo rm -rf {pgdata}', False)
        #     return jsn

        # RAM = int(jsn['stdout'].strip())
        # shared_buffers = RAM - 1024 - (4 * max_connections)

        # if shared_buffers <= 0:
        #     jsn = {
        #         'stdout': 'HINT: Consider reducing the maximum number of connections or increasing the available RAM',
        #         'stderr': f'shared buffers {shared_buffers}MB are higher than available RAM {RAM}',
        #         'excode': 1
        #     }
        #     status = Severity.NOT_OK.value
        #     duration_step_end = time.perf_counter() - duration_step_start
        #     logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        #     step = 'null'
        #     status = Severity.NOT_OK.value
        #     duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        #     logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        #     update_failed_object(pg_node['id'])
        #     remote_command(host, f"sudo rm -rf '{pgdata}'", False)
        #     return jsn

        # status = Severity.OK.value
        # duration_step_end = time.perf_counter() - duration_step_start
        # logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        # example: postgres_bin = '/usr/pgsql-10/bin'
        postgres_bin = find_postgres_bin(host, postgres_version, distributor)
        
        step = 'initdb the cluster'
        duration_step_start = time.perf_counter()
        # wal-segsize option exists in postgres 11 and above
        # if postgres_version == 10:
        jsn = json.loads(remote_command(host, f"sudo -iu postgres {postgres_bin}/initdb -D '{pgdata}'", True), strict=False)

        # elif postgres_version > 10:
        #     jsn = json.loads(remote_command(host, f'sudo -i -u postgres {postgres_bin}/initdb -D {pgdata} -E {encoding} --wal-segsize={wal_size}', True), strict=False)
        if 'error' in jsn['stderr'].lower() and 'error: nothing to do' not in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(pg_node['id'])
            if 'exists but is not empty' not in jsn['stderr'].lower():
                remote_command(host, f"sudo rm -rf '{pgdata}'", False)
            return jsn
            
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        if OperatingSystem.UBUNTU.value in distributor.lower():
            remote_command(host, f"sudo mkdir -p '{pgdata}/log'", False)
            remote_command(host, f"sudo chown -R postgres:postgres '{pgdata}'", False)

        step = 'updating postgresql.conf'
        duration_step_start = time.perf_counter()

        commands = [
            f"echo 'port = {pg_node['port']}' | sudo tee -a '{pgdata}/postgresql.conf'",
            f"""echo "listen_addresses = '*'" | sudo tee -a '{pgdata}/postgresql.conf'""",
            # f"sudo sed -i 's/shared_buffers = 128MB/shared_buffers = {shared_buffers}MB/g' '{pgdata}/postgresql.conf'",
            f"sudo sed -i 's/max_connections = 100/max_connections = {max_connections}/g' '{pgdata}/postgresql.conf'",
            f"""echo "archive_command = '/bin/true'" | sudo tee -a '{pgdata}/postgresql.conf'""",
            f"""echo "shared_preload_libraries = 'pg_stat_statements'" | sudo tee -a '{pgdata}/postgresql.conf'""",
            f"echo 'logging_collector = on' | sudo tee -a '{pgdata}/postgresql.conf'",
            f"""echo "log_filename = 'postgresql.log'" | sudo tee -a '{pgdata}/postgresql.conf'""",
            f"echo 'log_rotation_age = 0'| sudo tee -a '{pgdata}/postgresql.conf'",
            f"echo 'log_rotation_size = 0'| sudo tee -a '{pgdata}/postgresql.conf'"
        ]
        # store up to 64 wal segments by default (1024 MB of data)
        if postgres_version < 13:
            commands.append(f"""echo 'wal_keep_segments = 64' | sudo tee -a '{pgdata}/postgresql.conf'""")
        else:
            commands.append(f"""echo 'wal_keep_size = 1024' | sudo tee -a '{pgdata}/postgresql.conf'""")

        if postgres_version >= 17:
            commands.append(f"""echo "summarize_wal = 'on'" | sudo tee -a '{pgdata}/postgresql.conf'""")

        if OperatingSystem.CENTOS.value in distributor.lower():
            centos_commands = [
                f"sudo sed -i '/^PGDATA[^ ]*/c PGDATA={pgdata}' '/var/lib/pgsql/.bash_profile'",
                f"sudo rsync -a '/usr/lib/systemd/system/postgresql-{postgres_version}.service' '/usr/lib/systemd/system/postgresql-{postgres_version}_{pg_node['port']}.service'",
                f"sudo sed -i '/Environment=PGDATA/c Environment=PGDATA={pgdata}' '/usr/lib/systemd/system/postgresql-{postgres_version}_{pg_node['port']}.service'",
                "sudo systemctl daemon-reload",
                f"sudo systemctl enable postgresql-{postgres_version}_{pg_node['port']}.service",
                f"sudo systemctl start postgresql-{postgres_version}_{pg_node['port']}.service"
            ]
            if int(release) == 8:
                centos_commands.insert(2, f"sudo sed -i '/^KillSignal.*/a ExecStopPost=+systemctl daemon-reload' '/usr/lib/systemd/system/postgresql-{postgres_version}_{pg_node['port']}.service'")
            
            commands += centos_commands
        else:
            ubuntu_commands = [
                f"echo 'export PGDATA={pgdata}' | sudo tee -a '/var/lib/postgresql/.bashrc'",
                "sudo -s source '/var/lib/postgresql/.bashrc'",
            ]
            commands += ubuntu_commands
        
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])
                
            if jsn['stderr'] != '' and 'created symlink' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                # remote_command(host, f"sudo rm -rf '{pgdata}'", False)
                return jsn
      
        if OperatingSystem.UBUNTU.value in distributor.lower():
            err = start_node(host, postgres_version, pgdata)
            if err != '':
                return {'stdout': None, 'stderr': err,  'excode': 1}
        
        if ssl:
            jsn = setup_postgresql_ssl(distributor, host, pg_node, certificate_authentication=certificate_authentication, ssl_backup_host=app_host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return jsn
       
       # all the md5 authentication pg_hba.conf directives must be not defined earlier than ssl setup
        commands = [
            f"echo -e 'host\tall\t{pg_node['database_user']}\t{app_hostname}{'/32' if is_app_host_ip else ''}\tmd5' | sudo tee -a '{pgdata}/pg_hba.conf'",
            f"echo -e 'host\treplication\t{pg_node['database_user']}\t{app_hostname}{'/32' if is_app_host_ip else ''}\tmd5' | sudo tee -a '{pgdata}/pg_hba.conf'",
            f"echo -e 'host\tall\tpostgres\t{host['name']}{'/32' if is_host_ip else ''}\tmd5' | sudo tee -a '{pgdata}/pg_hba.conf'",
            f"echo -e 'host\tall\t{pg_node['database_user']}\t{host['name']}{'/32' if is_host_ip else ''}\tmd5' | sudo tee -a '{pgdata}/pg_hba.conf'",
        ]
        if not is_app_host_ip:
            commands += [
                f"echo -e 'host\tall\t{pg_node['database_user']}\t{app_host_ip}/32\tmd5' | sudo tee -a '{pgdata}/pg_hba.conf'",
                f"echo -e 'host\treplication\t{pg_node['database_user']}\t{app_host_ip}/32\tmd5' | sudo tee -a '{pgdata}/pg_hba.conf'",
            ]
        
        if not is_host_ip:
            commands += [
            f"echo -e 'host\tall\tpostgres\t{host_ip}/32\tmd5' | sudo tee -a '{pgdata}/pg_hba.conf'",
            f"echo -e 'host\tall\t{pg_node['database_user']}\t{host_ip}/32\tmd5' | sudo tee -a '{pgdata}/pg_hba.conf'"
        ]
        
        if OperatingSystem.CENTOS.value in distributor.lower():
            commands.append(f"sudo systemctl restart postgresql-{postgres_version}_{pg_node['port']}.service")
        elif OperatingSystem.UBUNTU.value in distributor.lower():
            commands.append(f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{pgdata}' restart -l /dev/null")
        commands.append(f"""database_pwd='{pg_node['database_pwd'].strip()}'; echo "CREATE USER {pg_node['database_user']} WITH LOGIN SUPERUSER REPLICATION PASSWORD '$database_pwd';" | sudo -iu postgres psql -p '{pg_node['port']}'""")
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return jsn
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    else:
        jsn = {}
        jsn['stdout'] = None
        jsn['stderr'] = f'This port {pg_node["port"]} is already in use'
        jsn['excode'] = 1
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return jsn

    jsn = {}
    jsn['stdout'] = None
    jsn['stderr'] = ''
    jsn['excode'] = 0
    return jsn

    # host = json.dumps(host)
    # pg_node = json.dumps(pg_node)
    # host_data = json.dumps({'host': host, 'db_data': pg_node})
    # try:
    #     params = {'schedule': '15 seconds', 'data': host_data, 'task': 'create_primary_lag_table_task', 'target_type': 'postgresql', 'target_id': pg_node['id'], 'created_by': created_by}
    #     insert_task(params)
    # except Exception as error:
    #     step = 'null'
    #     status = Severity.NOT_OK.value
    #     duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #     logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
  
def create_repmgr_task(data, task_id):
    """add standby to primary with repmgr"""

    task = 'create repmgr task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    
    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        primary_data = json.loads(data['primary_data'])
        standby_data = json.loads(data['standby_data'])
        host_primary = json.loads(data['host_primary'])
        host_standby = json.loads(data['host_standby'])
        org_id = primary_data['org_id']
        app_host = json.loads(data['app_host'])
        use_ssl = primary_data['ex_params'].get('ssl')
        certificate_authentication = primary_data['ex_params'].get('certificate_authentication')
        create_run_linux = data['create_run_linux']
        postgres_version = int(primary_data['version'])
        pg_password = data['pg_password'].strip()
        repmgr_pwd = primary_data['ex_params'].get('repmgr_pwd').strip()
        app_host_pg_password = data['app_host_pg_password']
        proxmox_container = standby_data['ex_params'].get('proxmox')
        host_exists = True
        distributor, release = check_os_version(host_standby)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = standby_data['ex_params'].get('distributor')
                release = standby_data['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(standby_data['id'])
                return
        
        version_exists = check_pg_version_exists(host_standby, postgres_version, distributor)
        package_manager =  "dnf" if int(release) > 7 else "yum"
        repository = data['repository']
        offline = True if Repository.LOCAL.value in repository else False
        haproxy_exists = data.get('haproxy_exists')
        repmgr_path = f'/etc/repmgr/{postgres_version}'
        # repmgr_config_file = f"repmgr{primary_data['id']}.conf"
        repmgr_config_file = "repmgr.conf"
        repmgr_pwd = primary_data["ex_params"].get("repmgr_pwd").strip()

        same_host = False
        if host_primary['name'] == host_standby['name']:
            same_host = True

        centos_dist = OperatingSystem.CENTOS.value in distributor
        postgres_home_dir = '/var/lib/pgsql' if centos_dist else '/var/lib/postgresql'
        if centos_dist:
            json_c = "http://vault.centos.org/centos/8-stream/BaseOS/x86_64/os/Packages/json-c-0.13.1-3.el8.x86_64.rpm" if int(release) == 8 else "json-c"
        
        if int(release) < 8:
             repmgr_version = RepmgrVersions.REDHAT7.value
        else:
            repmgr_version = RepmgrVersions.REDHAT_PG_LEGACY.value if postgres_version < 13 else RepmgrVersions.LATEST.value
        
        if postgres_version < 11:
            repmgr_version = RepmgrVersions.REDHAT_PG_10_DEFAULT.value
        
        pgpass_path = f'{postgres_home_dir}/.pgpass'
        private_filepath = f'{postgres_home_dir}/.ssh/id_rsa'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)


    if proxmox_container and not host_exists:

        step = 'creating proxmox container'
        duration_step_start = time.perf_counter()

        try:


            service_data = json.loads(data['service_data'])

            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host_standby, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(standby_data['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    if not version_exists:
        step = 'installing postgresql {}'.format(str(postgres_version))
        duration_step_start = time.perf_counter()

        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host_standby, postgres_version, DBType.POSTGRES.value, app_host)
        else:
            jsn = install_postgres_online(distributor, release, host_standby, postgres_version)
        
        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'exception', 'errno'] if not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration'])))
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(standby_data['id'])
            return
        elif 'warning' in error_lower:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        else:
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(host_standby, "sudo grep -qxF 'postgres    ALL=(ALL) NOPASSWD: ALL' '/etc/sudoers' || echo 'postgres    ALL=(ALL) NOPASSWD: ALL' | sudo tee -a '/etc/sudoers'", True), strict=False)
        if 'error' in jsn['stderr'].lower() or 'errno' in jsn['stderr'].lower():
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    isStandby = False
    # instaling and configuration of repmgr
    for _ in range(2):
        step = f'configuring {"primary" if not isStandby else "standby"} for repmgr'
        duration_step_start = time.perf_counter()
        
        last_warning = None
        node_id = primary_data['id']
        postgres_bin = find_postgres_bin(host_primary, postgres_version, distributor)
        
        if isStandby:
            host_primary, host_standby = host_standby, host_primary
            primary_data, standby_data = standby_data, primary_data
            node_id = primary_data['id']

            if offline:
                jsn = install_offline_remote_to_remote(distributor, release, host_primary, postgres_version, DBType.POSTGRES.value, app_host, Packages.REPMGR.value)
                error_lower = jsn['stderr'].lower()
                filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
                if filter:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    if isStandby:
                        update_failed_object(primary_data['id'])
                    return
                elif 'warning' in error_lower:
                    last_warning = jsn
            else:
                if OperatingSystem.CENTOS.value in distributor:
                    repmgr_install = f"sudo {package_manager} -y install {json_c if int(release) > 7 else ''} {f'repmgr_{postgres_version}-{repmgr_version}' if not (postgres_version == 10 and int(release) < 9) else f'https://dl.2ndquadrant.com/default/release/browse/rpm/packages/rhel/{release}/x86_64/10/repmgr10-{RepmgrVersions.REDHAT7.value}-1.el{release}.x86_64.rpm'} "
                elif OperatingSystem.UBUNTU.value in distributor:
                    repmgr_install = f"sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y postgresql-{postgres_version}-repmgr"
                
                jsn = json.loads(remote_command(host_primary, repmgr_install, True), strict=False)
                error_lower = jsn['stderr'].lower()
                if jsn['stderr'] != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(primary_data['id'])
                    return
            
            # variable to partially remove the hardcode
            standby_datadir = primary_data['dir_123Cluster'].strip()
            if standby_datadir.startswith('~'):
                jsn = find_tilda(host_primary)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(primary_data['id'])
                    return

                tilda_path = jsn['stdout']
                standby_datadir = standby_datadir.replace('~', tilda_path)
            
            # repmgr docs recommends to call standby clone only on empty directory, so for insurance
            prepare_datadir_commands = [
                f"sudo rm -rf '{standby_datadir}'",
                f"sudo mkdir -p '{standby_datadir}'",
                f"sudo chown -R postgres:postgres '{standby_datadir}'",
                f"sudo chmod 0700 '{standby_datadir}'",
            ]
            for cmd in prepare_datadir_commands:
                jsn = json.loads(remote_command(host_primary, cmd, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    update_failed_object(primary_data['id'])
                    return
        
        # on primary
        else:
            try:
                if offline:
                    jsn = install_offline_remote_to_remote(distributor, release, host_primary, postgres_version, DBType.POSTGRES.value, app_host, Packages.REPMGR.value)
                    error_lower = jsn['stderr'].lower()
                    filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
                    if filter:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        if isStandby:
                            update_failed_object(primary_data['id'])
                        return
                    elif 'warning' in error_lower:
                        last_warning = jsn
                else:
                    if OperatingSystem.CENTOS.value in distributor:
                        repmgr_install = f"sudo {package_manager} -y install {json_c if int(release) > 7 else ''} {f'repmgr_{postgres_version}-{repmgr_version}' if not (postgres_version == 10 and int(release) < 9) else f'https://dl.2ndquadrant.com/default/release/browse/rpm/packages/rhel/{release}/x86_64/10/repmgr10-5.4.0-1.el{release}.x86_64.rpm'} "
                    elif OperatingSystem.UBUNTU.value in distributor:
                        repmgr_install = f"sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y postgresql-{postgres_version}-repmgr"
                    
                    jsn = json.loads(remote_command(host_primary, repmgr_install, True), strict=False)
                    error_lower = jsn['stderr'].lower()
                    if jsn['stderr'] != '' and  any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        update_failed_object(standby_data['id'])
                        return
                
                is_host_primary_ip = is_ip(host_primary['name'])
                if not is_host_primary_ip:
                    host_primary_ip = socket.gethostbyname(host_primary['name'])
                
                is_host_standby_ip = is_ip(host_standby['name'])
                if not is_host_standby_ip:
                    host_standby_ip = socket.gethostbyname(host_primary['name'])
                
                remove_lines = "/hot_standby/d;/wal_level/d;/max_wal_senders/d;/max_replication_slots/d;/wal_log_hints/d;/archive_mode/d;/archive_command/d;/hot_standby_feedback/d;/shared_preload_libraries/d;/listen_addresses/d"
                append_lines = "hot_standby = on\nwal_level = replica\nmax_wal_senders = 10\nmax_replication_slots = 10\nwal_log_hints = on\narchive_mode = on\narchive_command = '/bin/true'\nshared_preload_libraries = 'repmgr,pg_stat_statements'\nhot_standby_feedback = on\nlisten_addresses = '*'\n"
                hba_lines = f"host\trepmgr\trepmgr\t{host_primary['name']}{'/32' if is_host_primary_ip else ''}\tmd5\n\
host\treplication\trepmgr\t{host_primary['name']}{'/32' if is_host_primary_ip else ''}\tmd5\n"
                if not is_host_primary_ip:
                    hba_lines += f"host\trepmgr\trepmgr\t{host_primary_ip}/32\tmd5\n\
host\treplication\trepmgr\t{host_primary_ip}/32\tmd5\n"

                if not same_host:
                    hba_lines += f"host\trepmgr\trepmgr\t{host_standby['name']}{'/32' if is_host_standby_ip else ''}\tmd5\n\
host\treplication\trepmgr\t{host_standby['name']}{'/32' if is_host_standby_ip else ''}\tmd5\n\
host\tall\tpostgres\t{host_standby['name']}{'/32' if is_host_standby_ip else ''}\tmd5\n\
host\tall\t{primary_data['database_user']}\t{host_standby['name']}{'/32' if is_host_standby_ip else ''}\tmd5\n"
                    
                    if not is_host_standby_ip:
                          hba_lines += f"host\trepmgr\trepmgr\t{host_standby_ip}/32\tmd5\n\
host\treplication\trepmgr\t{host_standby_ip}/32\tmd5\n\
host\tall\tpostgres\t{host_standby_ip}/32\tmd5\n\
host\tall\t{primary_data['database_user']}\t{host_standby_ip}/32\tmd5\n"

                commands = [
                    f"sudo sed -i '{remove_lines}' '{primary_data['dir_123Cluster']}/postgresql.conf'",
                    f"""echo "{append_lines}" | sudo tee -a '{primary_data['dir_123Cluster']}/postgresql.conf'""",
                    f"echo -e '{hba_lines}' | sudo tee -a '{primary_data['dir_123Cluster']}/pg_hba.conf'",
                    f'sudo systemctl stop repmgr{"d" if OperatingSystem.UBUNTU.value in distributor else "-{}".format(postgres_version)} | true',
                    "sudo pkill -9 -f repmgr",
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_primary, command, True), strict=False)
                    if jsn['stdout'] != '':
                        print(jsn['stdout'])

                    error_lower = jsn['stderr'].lower()
                    filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'] if 'error: nothing to do' not in error_lower and 'already installed' not in error_lower)) and not any(err in error_lower for err in ['role "repmgr" does not exist', 'not loaded', 'notice', 'warning'])
                    if filter:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                
                #after killing repmgr, primary goes into crash recovery mode for a little while, need to wait before connecting with db driver
                time.sleep(1)
                
                # we should have access to primary, due to create_node() function
                conn = connect_to_postgres_database(host_primary['name'], 'postgres',  primary_data['database_user'], primary_data['database_pwd'].strip(), int(primary_data['port']))
                if isinstance(conn, tuple):
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                    return
                
               #ensure that create database and drop database getting personal transaction by setting autocommit before any other command 
               #and by not using with ... as
                conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT) 
                cur = conn.cursor()
                cur.execute("SELECT pg_reload_conf()") 
                cur.execute("SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname='repmgr'")
                cur.execute(f'''DROP DATABASE IF EXISTS repmgr {'WITH (FORCE)' if postgres_version > 12 else ''}''')
                cur.execute("DROP USER IF EXISTS repmgr")
                cur.execute("CREATE USER repmgr WITH LOGIN SUPERUSER PASSWORD %s", (repmgr_pwd,))
                cur.execute("CREATE DATABASE repmgr WITH OWNER repmgr")
                # for value of search_path see https://www.repmgr.org/docs/current/quickstart-repmgr-user-database.html
                cur.execute("""ALTER USER repmgr SET search_path TO repmgr, "$user", public""")
                # do not create extension, 'repmgr primary register' will do it for us
                # https://www.repmgr.org/docs/current/quickstart-primary-register.html
                cur.close()
                conn.close()
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return

            client = create_ssh_connection(host_primary['name'], host_primary['username'], host_primary['password'], host_primary['private_key'], int(host_primary['ssh_port']))
            if isinstance(client, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,f"cannot connect to host {host_primary['name']}",str(client[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,f"cannot connect to host {host_primary['name']}",str(client[1]),1)
                return
            
            command = f"sudo systemctl restart postgresql-{postgres_version}_{primary_data['port']}.service" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{primary_data['dir_123Cluster']}' restart -l /dev/null"
            try:
                stdin, stdout, stderr = client.exec_command(command)
                exit_status = stdout.channel.recv_exit_status()# Blocking call
                stdin.channel.shutdown_write()
                try:
                    stderr = stderr.read().decode('ascii').strip()
                    stdout = stdout.read().decode('ascii').strip()
                except UnicodeDecodeError:
                    stderr = stderr.read().decode('utf-8').strip()
                    stdout = stdout.read().decode('utf-8').strip()

                if stderr != '' and 'error' in stderr:
                    if OperatingSystem.CENTOS.value in distributor:
                        lock_file_postmaster = status_pg_service(client, primary_data['port'], postgres_version)
                        if lock_file_postmaster:
                            commands = [
                                f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{primary_data['dir_123Cluster']}' stop -l /dev/null",
                                f"sudo systemctl start postgresql-{postgres_version}_{primary_data['port']}.service"
                            ]
                            for command in commands:
                                stdin, stdout, stderr = client.exec_command(command)
                                exit_status = stdout.channel.recv_exit_status()# Blocking call
                                stdin.channel.shutdown_write()
                                try:
                                    stderr = stderr.read().decode('utf-8').strip()
                                    stdout = stdout.read().decode('utf-8').strip()
                                except UnicodeDecodeError:
                                    stderr = stderr.read().decode('ascii').strip()
                                    stdout = stdout.read().decode('ascii').strip()
                                
                                if stderr != '':
                                    break
                                
                                time.sleep(5)
                    if stderr != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,stderr,exit_status)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,stderr,exit_status)
                        return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

        # passwordless SSH connection for the Linux user "postgres"
        ssh_file_part = f"-i '{private_filepath}.pub'" if OperatingSystem.UBUNTU.value in distributor else ''
        pg_password_path = f"""{TASK_TMP_FILES_DIR}/{host_primary['name']}"""
        # some passwords could not be read from stdin by sshpass, pass through file required 
        sshpassfile = f"sudo mkdir -p '{pg_password_path}' && echo '{pg_password}' | sudo tee '{pg_password_path}/pg_password' && sudo chown postgres:postgres '{pg_password_path}' && sudo -iu postgres sshpass -f '{pg_password_path}/pg_password' ssh-copy-id -p '{host_standby['ssh_port']}' -o StrictHostKeyChecking=no {ssh_file_part} 'postgres@{host_standby['name']}'"

        commands = [
            f'sudo {package_manager} install -q -y openssl' if OperatingSystem.CENTOS.value in distributor else 'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y openssl',
            f'sudo {package_manager} install -q -y sshpass' if OperatingSystem.CENTOS.value in distributor else 'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y sshpass',
            f"sudo usermod --password $(echo '{pg_password}' | openssl passwd -1 -stdin) postgres",
            f"sudo -u postgres ssh-keygen -t rsa -P '' -f '{private_filepath}' <<< y",
            f"echo 'export PATH=$PATH:{postgres_bin}' | sudo tee -a '{postgres_home_dir}/.bash_profile'",
            f"sudo -s source '{postgres_home_dir}/.bash_profile'"
        ]

        if not isStandby:
            for host in [host_primary, host_standby]:
                host_commands = commands.copy()
                if OperatingSystem.CENTOS.value in distributor:
                    if  int(release) == 9:
                        host_commands = list(map(lambda command: command.replace(f'sudo {package_manager} install -q -y sshpass', f'sudo {package_manager} install -q -y https://mirror.stream.centos.org/9-stream/AppStream/x86_64/os/Packages/sshpass-1.09-4.el9.x86_64.rpm'), host_commands))
                    elif int(release) == 8:
                        host_commands = list(map(lambda command: command.replace(f'sudo {package_manager} install -q -y sshpass', f'sudo {package_manager} install -q -y https://dl.rockylinux.org/pub/rocky/8/AppStream/x86_64/os/Packages/s/sshpass-1.09-4.el8.x86_64.rpm'), host_commands))
                    elif int(release) == 7:
                        host_commands = list(map(lambda command: command.replace(f'sudo {package_manager} install -q -y sshpass', f'sudo {package_manager} install -q -y https://vault.centos.org/7.6.1810/extras/x86_64/Packages/sshpass-1.06-2.el7.x86_64.rpm'), host_commands))
                    
                    host_commands = host_commands[2:] if offline else host_commands
                else:
                    host_commands = host_commands[2:] if offline else host_commands

                if host == host_standby and offline:
                    jsn = install_offline_remote_to_remote(distributor, release, host, postgres_version, DBType.POSTGRES.value, app_host, Packages.REPMGR.value)
                    error_lower = jsn['stderr'].lower()
                    filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration'])
                    if filter:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                    elif 'warning' in error_lower:
                        last_warning = jsn
                            
                for command in host_commands:
                    try:
                        jsn = json.loads(remote_command(host, command, True), strict=False)
                        if jsn['stdout'] != '':
                            print(jsn['stdout'])

                        error_lower = jsn['stderr'].lower()
                        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'] if 'error: nothing to do' not in error_lower and 'already installed' not in error_lower)) 
                        if filter:
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return
                    except Exception as error:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                        return

        try:    
            jsn = json.loads(remote_command(host_primary, sshpassfile, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if jsn['stderr'] != '' and 'info' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
        # public_filepath = sshpassfile_dir
        # generateSSHKey(host_primary, host_standby, private_filepath, public_filepath)

        if primary_data['dir_123Cluster'].startswith('~'):
            jsn = find_tilda(host_primary)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return 

            tilda_path = jsn['stdout']
            primary_data['dir_123Cluster'] = primary_data['dir_123Cluster'].replace('~', tilda_path)
        # repmgr docs are strongly against using password in connection string, and it limits passwords to alphanumeric-only, better use pgpass_path
        append_lines = f"node_id={str(node_id)}\nnode_name='pgnode{str(node_id)}'\n\
conninfo='host={host_primary['name']} port={primary_data['port']} dbname=repmgr user=repmgr passfile={pgpass_path}'\n\
data_directory='{primary_data['dir_123Cluster']}'\n\
pg_bindir='{postgres_bin}'\n\
use_replication_slots=yes\n\
replication_type='physical'\n\
failover='automatic'\n\
passfile='{pgpass_path}'\n\
use_primary_conninfo_password=false\n\
promote_command='{postgres_bin}/repmgr standby promote -f {repmgr_path}/{repmgr_config_file} --log-to-file'\n\
follow_command='{postgres_bin}/repmgr standby follow -f {repmgr_path}/{repmgr_config_file} --log-to-file --upstream-node-id=%n'\n"

        if OperatingSystem.CENTOS.value in distributor:
            append_lines += f"service_start_command='sudo systemctl start postgresql-{postgres_version}_{primary_data['port']}'\n\
service_stop_command='sudo systemctl stop postgresql-{postgres_version}_{primary_data['port']}'\n\
service_restart_command='sudo systemctl restart postgresql-{postgres_version}_{primary_data['port']}'\n\
service_reload_command='sudo systemctl reload postgresql-{postgres_version}_{primary_data['port']}'\n"

        elif OperatingSystem.UBUNTU.value in distributor:
            append_lines += f"service_start_command='{postgres_bin}/pg_ctl -D {primary_data['dir_123Cluster']} start -l /dev/null '\n\
service_stop_command='{postgres_bin}/pg_ctl -D {primary_data['dir_123Cluster']} stop -m fast -l /dev/null '\n\
service_restart_command='{postgres_bin}/pg_ctl -D {primary_data['dir_123Cluster']} restart -m fast -l /dev/null '\n\
service_reload_command='{postgres_bin}/pg_ctl -D {primary_data['dir_123Cluster']} reload -l /dev/null'\n\
repmgrd_service_start_command='sudo systemctl start repmgrd'\n\
repmgrd_service_stop_command='sudo systemctl stop repmgrd'\n"

        remove_append_lines = "/node_id/d;/node_name/d;/conninfo/d;/data_directory/d;/pg_bindir/d;/use_replication_slots/d;/replication_type/d;/failover/d;/promote_command/d;/repmgrd_service_start_command/d;/repmgrd_service_stop_command/d"

        if int(host_primary['ssh_port']) != 22:
            remove_append_lines = remove_append_lines + ";/ssh_options/d"
            append_lines = append_lines + f"ssh_options='-p {str(host_primary['ssh_port'])} -q -o ConnectTimeout=10'\n"

        # ubuntu: '/etc/init.d/repmgrd' or '/etc/default/repmgrd'
        service_location = '/usr/lib/systemd/system/' if OperatingSystem.CENTOS.value in distributor else '/etc/default/repmgrd'
        commands = [
            # f"sudo sed -i '/Environment=REPMGRDCONF/c Environment=REPMGRDCONF={repmgr_path}/{repmgr_config_file}' '{service_location}/repmgr-{postgres_version}.service'",
            # f"sudo rsync -a '{repmgr_path}/repmgr.conf' '{repmgr_path}/{repmgr_config_file}'",
            f"sudo rsync -a '{repmgr_path}/repmgr.conf' '{repmgr_path}/repmgr.conf.bak'",
            f"sudo sed -i '{remove_append_lines}' '{repmgr_path}/{repmgr_config_file}'",
            f"""echo "{append_lines}" | sudo tee -a '{repmgr_path}/{repmgr_config_file}'"""
        ]

        if OperatingSystem.UBUNTU.value in distributor:
            ubuntu_commands = [
                f"sudo mkdir -p '{repmgr_path}'",
                f"sudo touch '{repmgr_path}/repmgr.conf'",
                f"sudo sed -i '/REPMGRD_ENABLED=no/c REPMGRD_ENABLED=yes' '{service_location}'",
                f"""sudo sed -i '/REPMGRD_CONF=/c REPMGRD_CONF="{repmgr_path}/{repmgr_config_file}"' '{service_location}'""",
                "sudo systemctl daemon-reload",
            ]
            commands = ubuntu_commands + commands

        for command in commands:
            jsn = json.loads(remote_command(host_primary, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])
            
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        if not isStandby:
            #should be run before starting to work with repmgr, not after
            pgpass_content = f"\n{host_primary['name']}:{primary_data['port']}:repmgr:repmgr:$repmgr_pwd\n{host_primary['name']}:{primary_data['port']}:replication:repmgr:$repmgr_pwd\n{host_standby['name']}:{standby_data['port']}:repmgr:repmgr:$repmgr_pwd\n{host_standby['name']}:{standby_data['port']}:replication:repmgr:$repmgr_pwd"
            
            if not is_host_primary_ip:
                pgpass_content = f"{pgpass_content}\n{host_primary_ip}:{primary_data['port']}:repmgr:repmgr:$repmgr_pwd\n{host_primary_ip}:{primary_data['port']}:replication:repmgr:$repmgr_pwd"
            
            if not is_host_standby_ip:
                 pgpass_content = f"""{pgpass_content}\n{host_standby_ip}:{standby_data['port']}:repmgr:repmgr:$repmgr_pwd\n{host_standby_ip}:{standby_data['port']}:replication:repmgr:$repmgr_pwd"""
            
            commands = [
                f"""repmgr_pwd='{repmgr_pwd}'; echo "{pgpass_content}" | sudo tee '{pgpass_path}'""",
                f"sudo chown postgres:postgres '{pgpass_path}'",
                f"sudo chmod 0600 '{pgpass_path}'",
            ]

            for host in [host_standby, host_primary]:
                for command in commands:
                    jsn = json.loads(remote_command(host, command, True), strict=False)
                    if jsn['stdout'] != '':
                        print(jsn['stdout'])

                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
            
            commands = [
                f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' primary register",
                "sudo mkdir -p '/run/repmgr'",
                f"sudo touch '/run/repmgr/repmgrd-{postgres_version}.pid'",
                "sudo chown -R postgres:postgres '/run/repmgr'",
                f"sudo systemctl enable repmgr-{postgres_version}" if OperatingSystem.CENTOS.value in distributor else "echo 'systemctl enable repmgrd is not a native service'"
            ]

            # repmgr 10 5.4.0 available only for redhat < 9, on redhat 9 are repmgr 10 5.3.2
            if OperatingSystem.CENTOS.value in distributor and int(release) < 9 and postgres_version == 10:
                # maintainer of repmgr 10 5.4.0 has different name of service - rename to expected
                commands = [
                    "sudo rsync -a /lib/systemd/system/repmgr10.service /etc/systemd/system/repmgr-10.service",
                    "sudo systemctl daemon-reload"
                ] + commands

            for command in commands:
                jsn = json.loads(remote_command(host_primary, command, True), strict=False)
                if jsn['stdout'] != '':
                    print(jsn['stdout'])

                err_list = ('notice: primary node record', 'this node is already registered', 'created symlink')
                if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list): 
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
        # standby commands
        else:
            # give standby string connection of primary
            commands = [
                f"echo 'export PATH=$PATH:{postgres_bin}' | sudo tee -a '{postgres_home_dir}/.bash_profile'",
                f"sudo -s source '{postgres_home_dir}/.bash_profile'",
                # deprecate libpq connection strings whenever possible - when using such strings with url-unsafe passwords, cluster will break
                f"""sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby clone -h '{host_standby['name']}' -p '{standby_data['port']}' -U repmgr -d repmgr --force --fast-checkpoint""",
                "sudo mkdir -p '/run/repmgr'",
                f"sudo touch '/run/repmgr/repmgrd-{postgres_version}.pid'",
                "sudo chown -R postgres:postgres /run/repmgr"
            ]

            if OperatingSystem.CENTOS.value in distributor:
                centos_commands = [
                    f"sudo systemctl enable repmgr-{postgres_version}",
                    f"sudo rsync -a '/usr/lib/systemd/system/postgresql-{postgres_version}.service' '/usr/lib/systemd/system/postgresql-{postgres_version}_{primary_data['port']}.service'",
                    f"sudo sed -i '/Environment=PGDATA/c Environment=PGDATA={standby_datadir}' '/usr/lib/systemd/system/postgresql-{postgres_version}_{primary_data['port']}.service'",
                    "sudo systemctl daemon-reload",
                    f"sudo systemctl enable postgresql-{postgres_version}_{primary_data['port']}.service",
                    f"sudo systemctl start postgresql-{postgres_version}_{primary_data['port']}.service"
                ]
                if int(release) == 8:
                    centos_commands.insert(2, f"sudo sed -i '/^KillSignal.*/a ExecStopPost=+systemctl daemon-reload' '/usr/lib/systemd/system/postgresql-{postgres_version}_{primary_data['port']}.service'")
                
                # repmgr 10 5.4.0 available only for redhat < 9, on redhat 9 are repmgr 10 5.3.2
                if int(release) < 9 and postgres_version == 10:
                    # maintainer of repmgr 10 5.4.0 has different name of service - rename to expected
                    centos_commands = [
                        "sudo rsync -a /lib/systemd/system/repmgr10.service /etc/systemd/system/repmgr-10.service",
                        "sudo systemctl daemon-reload"
                    ] + centos_commands
                
                commands += centos_commands

            for command in commands:
                if '~' in command:
                    jsn = find_tilda(host_primary)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return 

                    tilda_path = jsn['stdout']
                    command = command.replace('~', tilda_path)

                jsn = json.loads(remote_command(host_primary, command, True), strict=False)
                if jsn['stdout'] != '':
                    print(jsn['stdout'])

                err_list = ('info', 'notice', 'file exists', 'created symlink')
                if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            
            if OperatingSystem.UBUNTU.value in distributor:
                stderr = start_node(host_primary, postgres_version, standby_datadir)
                if stderr != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(stderr),1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(stderr),1)
                    return
            
            command = f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby register --force --wait-sync"
            jsn = json.loads(remote_command(host_primary, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            err_list = ('info', 'notice')
            if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        isStandby = True
        
        status = Severity.WARNING.value if last_warning else Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_warning['stdout'],last_warning['stderr'],last_warning['excode']) if last_warning else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # host_primary and host_standby was switched, also primary_data and standby_data in loop above
    
    step = 'starting repmgr service'
    duration_step_start = time.perf_counter()

    # do exactly what are meant in step
    for host in [host_standby, host_primary]:
        jsn = json.loads(remote_command(host, f"sudo systemctl start repmgr-{postgres_version}" if OperatingSystem.CENTOS.value in distributor else "sudo systemctl restart repmgrd", True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    # means sslmode = verify-full are requested
    if use_ssl and certificate_authentication:
        # so need to ignore 'repmgr standby clone' and generate unique certificates for host of standby...
        step = 'generating standby ssl certificates'
        duration_step_start = time.perf_counter()
        
        jsn = generate_ssl_certificates(host_primary, primary_data, certificate_authentication=certificate_authentication, ssl_backup_host=app_host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        cmd = f"sudo systemctl restart postgresql-{postgres_version}_{primary_data['port']}.service" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{primary_data['dir_123Cluster']}' restart -l /dev/null"
        jsn = json.loads(remote_command(host_primary, cmd, True, step), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = "configuring user 'postgres' passwordless authentication with app host"
    duration_step_start = time.perf_counter()
    
    pg_password_path_app_host = f"{TASK_TMP_FILES_DIR}/{app_host['name']}"
    
    jsn = json.loads(remote_command(host_primary,  f"sudo mkdir -p '{pg_password_path_app_host}' && echo '{app_host_pg_password}' | sudo tee '{pg_password_path_app_host}/pg_password' && sudo chown postgres:postgres '{pg_password_path_app_host}' && sudo -iu postgres sshpass -f '{pg_password_path_app_host}/pg_password' ssh-copy-id -p '{app_host['ssh_port']}' -o StrictHostKeyChecking=no {ssh_file_part} 'postgres@{app_host['name']}'", True), strict=False)
    if jsn['stderr'] != '' and 'permission denied' in jsn['stderr'].lower():
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
      
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    # primary and standby were switched
    # host_primary['create_task'] = created_by
    # host_primary['ssh_port'] = int(host_primary['ssh_port'])
    
    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host_primary, 'task': 'run_linux_performance_counters', 'target_type': OperatingSystem.LINUX.value, 'target_id': host_primary['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as error:
    #         step = 'null'
    #         status = Severity.NOT_OK.value
    #         duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    #         logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
    #         return
    
        
    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()
    
    try:

    
        jsn = json.loads(remote_command(host_primary, f"sudo du -sb '{standby_datadir}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        
        remote_command(host_primary, f"""echo "{get_log_rotate_config(standby_datadir)}" | sudo tee '/etc/logrotate.d/{primary_data['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host_primary), 'node_name': primary_data['name']}, 'task': 'log_rotate_task', 'monitor_name': None,  'org_id': primary_data['org_id'],  'target_type': primary_data['type'], 'target_id': primary_data['id'], 'created_by': created_by}
        insert_task(params)


        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, postgres_version, node_size, primary_data['id'],))
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,primary_data['id'],created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host_primary['id'],created_by,OperatingSystem.LINUX.value])
                conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    except Exception as error:
        status = Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)


    if haproxy_exists:
        step = 'configuring HAProxy'
        duration_step_start = time.perf_counter()

        try:
            conn = connect_to_postgres_database(host_primary['name'], 'postgres', primary_data['database_user'], primary_data['database_pwd'].strip(), primary_data['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return  
            
            with conn:
                with conn.cursor() as cur:
                    cur.execute("SHOW config_file")
                    config_file = cur.fetchone()[0]
                    if config_file.endswith('/'):
                        config_file = config_file[:-1]
                    
                    config_dir = os.path.dirname(config_file)
                        
            haproxy_data = [{"host": json.loads(i['host']), "port": i['port']} for i in data['haproxy_data']]
            pg_creds = f"pg{primary_data['id']} {host_primary['name']}:{primary_data['port']}"
            for i in haproxy_data:
                host_haproxy = i['host']
                haproxy_port = i['port']
                is_host_haproxy_ip = is_ip(host_haproxy['name'])
                if not is_host_haproxy_ip:
                    host_haproxy_ip = socket.gethostbyname(host_haproxy['name'])

                jsn = json.loads(remote_command(host_haproxy, f"sudo grep -qF 'server {pg_creds}' '/etc/haproxy/haproxy.cfg' && echo 'already exists'", True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
                
                if 'already exists' not in jsn['stdout']:
                    commands = [
                        "sudo test ! -f '/etc/haproxy/haproxy.cfg.bak' && sudo rsync -a '/etc/haproxy/haproxy.cfg' '/etc/haproxy/haproxy.cfg.bak'",
                        f'''last_listen_line=$(sudo grep -n '^listen cluster' '/etc/haproxy/haproxy.cfg' | sudo awk -F: '{{print $1}}' | sudo tail -n 1) &&
        if [[ $(sudo grep -n '^listen cluster_{haproxy_port}' '/etc/haproxy/haproxy.cfg' | sudo awk -F: '{{print $1}}') -eq $last_listen_line ]]; then
        echo '    server {pg_creds} check port 80' | sudo tee -a '/etc/haproxy/haproxy.cfg'
        else
        empty_line_index=$(sudo awk '/^listen cluster_{haproxy_port}/ {{start=NR}} /^$/{{if (start) {{print NR; exit}}}}' '/etc/haproxy/haproxy.cfg')
        
        if [ -n "$empty_line_index" ]; then
            sudo sed -i "${{empty_line_index}}i\ \ \ \ server {pg_creds} check port 80" '/etc/haproxy/haproxy.cfg'
        fi
        fi
        ''',
                        """sudo sed -i '/^[[:space:]]*$/d' '/etc/haproxy/haproxy.cfg' && sudo awk 'BEGIN { is_blank = 1 } /^[[:alpha:]]/ { if (!is_blank) print "" } NF { is_blank = 0 } 1' '/etc/haproxy/haproxy.cfg' | sudo tee 'temp_haproxy.cfg' && sudo mv 'temp_haproxy.cfg' '/etc/haproxy/haproxy.cfg'""",
                        "sudo systemctl restart haproxy",
                        "sudo systemctl status haproxy"
                    ]
                    for command in commands:
                        jsn = json.loads(remote_command(host_haproxy, command, True), strict=False)
                        if jsn['stderr'] != '':
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return

                commands = [
                    f"sudo sed -i '/host\s*all\s*{primary_data['database_user']}\s*{host_haproxy['name']}{chr(92) + '/32' if is_host_haproxy_ip else ''}\s*md5/d' '{config_dir}/pg_hba.conf'",
                    f"echo -e 'host\tall\t{primary_data['database_user']}\t{host_haproxy['name']}{'/32' if is_host_haproxy_ip else ''}\tmd5' | sudo tee -a '{config_dir}/pg_hba.conf'",
                ]
                if not is_host_haproxy_ip:
                    commands += [
                        f"sudo sed -i '/host\s*all\s*{primary_data['database_user']}\s*{host_haproxy_ip}\/32\s*md5/d' '{config_dir}/pg_hba.conf'",
                        f"echo -e 'host\tall\t{primary_data['database_user']}\t{host_haproxy_ip}/32\tmd5' | sudo tee -a '{config_dir}/pg_hba.conf'",
                    ]
                    
                for command in commands:
                    jsn = json.loads(remote_command(host_primary, command, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
            
            jsn = install_healthcheck(distributor, release, host_primary, primary_data, app_host, offline)
            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'] if 'error: nothing to do' not in error_lower and 'already installed' not in error_lower))
            if filter:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            conn = connect_to_postgres_database(host_primary['name'], 'postgres', primary_data['database_user'], primary_data['database_pwd'].strip(), primary_data['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return  
            
            with conn:
                with conn.cursor() as cur:
                    cur.execute('SELECT pg_reload_conf()')
            
            # ensure that connection does not closed, 
            # because that block could be needed only if have haproxy
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"ind_haproxy":true}' WHERE id = %s;""" % (primary_data['id'],))
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def add_standby_repmgr_task(data, task_id):
    """add second or more standby"""

    task = 'add standby task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        primary_data = json.loads(data['primary_data'])
        new_standby_data = json.loads(data['standby_data'])
        org_id = new_standby_data['org_id']
        host_primary = json.loads(data['host_primary'])
        host_new_standby = json.loads(data['host_standby'])
        use_ssl = primary_data['ex_params'].get('ssl')
        certificate_authentication = primary_data['ex_params'].get('certificate_authentication')
        is_host_new_standby_ip = is_ip(host_new_standby['name'])
        app_host_pg_password = data['app_host_pg_password']
        if not is_host_new_standby_ip:
            host_new_standby_ip = socket.gethostbyname(host_new_standby['name'])
        
        is_host_primary_ip = is_ip(host_primary['name'])
        if not is_host_primary_ip:
            host_primary_ip = socket.gethostbyname(host_primary['name'])
        
        app_host = json.loads(data['app_host'])
        repmgr_pwd = primary_data['ex_params'].get("repmgr_pwd").strip()
        proxmox_container = new_standby_data['ex_params'].get('proxmox')
        host_exists = True
        distributor, release = check_os_version(host_new_standby)
        if OperatingSystem.INVALID.value in distributor:
            # proxmox container may not exist at the moment
            if proxmox_container:
                host_exists = False
                distributor = new_standby_data['ex_params'].get('distributor')
                release = new_standby_data['ex_params'].get('release')
            else:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(new_standby_data['id'])
                return
        
        package_manager = "dnf" if int(release) > 7 else "yum"
        json_c = "http://vault.centos.org/centos/8-stream/BaseOS/x86_64/os/Packages/json-c-0.13.1-3.el8.x86_64.rpm" if int(release) == 8 else "json-c"
        create_run_linux = data['create_run_linux']
        postgres_version = int(primary_data['version'])
        version_exists = check_pg_version_exists(host_new_standby, postgres_version, distributor)
        pg_password = primary_data['ex_params'].get('pg_password').strip()
        # standbys_host = [{'host':..., 'node_id':..., 'node_port':...}, {...}, ...]
        standbys_host = data['standbys']
        repmgr_pwd = primary_data['ex_params'].get('repmgr_pwd').strip()
        repository = data['repository']
        offline = True if Repository.LOCAL.value in repository else False
        haproxy_exists = data.get('haproxy_exists')
        standby_datadir = new_standby_data['dir_123Cluster'].strip()
        repmgr_path = f'/etc/repmgr/{postgres_version}'
        # repmgr_config_file = f"repmgr{primary_data['id']}.conf"
        repmgr_config_file = "repmgr.conf"
        postgres_home_dir = '/var/lib/pgsql' if OperatingSystem.CENTOS.value in distributor else '/var/lib/postgresql'
        private_filepath = f'{postgres_home_dir}/.ssh/id_rsa'
        pgpass_path = f'{postgres_home_dir}/.pgpass'
        pgpass_content = f"{host_primary['name']}:{primary_data['port']}:repmgr:repmgr:$repmgr_pwd\n{host_primary['name']}:{primary_data['port']}:replication:repmgr:$repmgr_pwd\n{host_new_standby['name']}:{new_standby_data['port']}:repmgr:repmgr:$repmgr_pwd\n{host_new_standby['name']}:{new_standby_data['port']}:replication:repmgr:$repmgr_pwd"""
        
        if not is_host_primary_ip:
            pgpass_content = f"{pgpass_content}\n{host_primary_ip}:{primary_data['port']}:repmgr:repmgr:$repmgr_pwd\n{host_primary_ip}:{primary_data['port']}:replication:repmgr:$repmgr_pwd"
        
        if not is_host_new_standby_ip:
            pgpass_content = f"{pgpass_content}\n{host_new_standby_ip}:{new_standby_data['port']}:repmgr:repmgr:$repmgr_pwd\n{host_new_standby_ip}:{new_standby_data['port']}:replication:repmgr:$repmgr_pwd"""
        
        standbys_host = [{"host": json.loads(i["host"]), "node_id": i["node_id"], "node_port": i["node_port"], "db_usr": i["db_usr"], "db_pwd": i["db_pwd"], "data_directory": i['data_directory']} for i in standbys_host]
        for standby in standbys_host:
            standby_host = standby['host']
            standby_port = standby['node_port']
            is_standby_host_ip = is_ip(standby_host['name'])
            if not is_standby_host_ip:
                standby_host_ip = socket.gethostbyname(standby_host['name'])
            
            pgpass_content += f"\n{standby_host['name']}:{standby_port}:repmgr:repmgr:$repmgr_pwd\n{standby_host['name']}:{standby_port}:replication:repmgr:$repmgr_pwd"
            
            if not is_standby_host_ip:
                pgpass_content += f"\n{standby_host_ip}:{standby_port}:repmgr:repmgr:$repmgr_pwd\n{standby_host_ip}:{standby_port}:replication:repmgr:$repmgr_pwd"""
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if proxmox_container and not host_exists:

        try:

            step = 'creating proxmox container'
            duration_step_start = time.perf_counter()

            service_data = json.loads(data['service_data'])
            ram_gb = data.get('ram_gb')
            cpu_cores = data.get('cpu_cores')
            disk_gb = data.get('disk_gb')
            jsn = init_container(service_data, host_new_standby, distributor, release, ram_gb, cpu_cores, disk_gb)
            if jsn['stderr'] != '' or jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                update_failed_object(new_standby_data['id'])
                return
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    if not version_exists:
        step = 'installing postgresql {}'.format(str(postgres_version))
        duration_step_start = time.perf_counter()

        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host_new_standby, postgres_version, DBType.POSTGRES.value, app_host)
        else:
            jsn = install_postgres_online(distributor, release, host_new_standby, postgres_version)
        
        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(new_standby_data['id'])
            return
        elif 'warning' in error_lower:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        else:
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'installing repmgr on standby'
    duration_step_start = time.perf_counter()
    
    try:
        jsn = json.loads(remote_command(host_new_standby, "sudo grep -qxF 'postgres    ALL=(ALL) NOPASSWD: ALL' '/etc/sudoers' || echo 'postgres    ALL=(ALL) NOPASSWD: ALL' | sudo tee -a '/etc/sudoers'", True), strict=False)
        if 'error' in jsn['stderr'].lower() or 'errno' in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    if standby_datadir.startswith('~'):
        jsn = find_tilda(host_new_standby)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(new_standby_data['id'])
            return

        tilda_path = jsn['stdout']
        standby_datadir = standby_datadir.replace('~', tilda_path)
    
    node_id = new_standby_data['id']
    postgres_bin = find_postgres_bin(host_new_standby, postgres_version, distributor)

    if offline:
        jsn = install_offline_remote_to_remote(distributor, release, host_new_standby, postgres_version, DBType.POSTGRES.value, app_host, Packages.REPMGR.value)
        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        elif 'warning' in error_lower:
            status = Severity.WARNING.value
    else:
        if OperatingSystem.CENTOS.value in distributor:
            if int(release) < 8:
                repmgr_version = RepmgrVersions.REDHAT7.value
            else:
                repmgr_version = RepmgrVersions.REDHAT_PG_LEGACY.value if postgres_version < 13 else RepmgrVersions.LATEST.value
        
            if postgres_version < 11:
                repmgr_version = RepmgrVersions.REDHAT_PG_10_DEFAULT.value
            
            command = f"sudo {package_manager} install -q -y {json_c if int(release) > 7 else ''}  { f'repmgr_{postgres_version}-{repmgr_version}'  if not (postgres_version == 10 and int(release) < 9) else f'https://dl.2ndquadrant.com/default/release/browse/rpm/packages/rhel/{release}/x86_64/10/repmgr10-{RepmgrVersions.REDHAT7.value}-1.el{release}.x86_64.rpm'}"
        elif OperatingSystem.UBUNTU.value in distributor:
            command = f"sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y postgresql-{postgres_version}-repmgr"

        try:
            jsn = json.loads(remote_command(host_new_standby, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    status = Severity.WARNING.value if status == Severity.WARNING.value else Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'configuring standby for repmgr'
    duration_step_start = time.perf_counter()

    set_pg_password = f"sudo usermod --password $(echo '{pg_password}' | openssl passwd -1 -stdin) postgres"
    
    commands = [
        f'sudo {package_manager} install -q -y openssl' if OperatingSystem.CENTOS.value in distributor else 'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y openssl',
        f'sudo {package_manager} install -q -y sshpass' if OperatingSystem.CENTOS.value in distributor else 'sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y sshpass',
        set_pg_password,
        f"sudo -u postgres ssh-keygen -t rsa -P '' -f '{private_filepath}' <<< 'y'",
        f"echo 'export PATH=$PATH:{postgres_bin}' | sudo tee -a '{postgres_home_dir}/.bash_profile'",
        f"sudo -s source '{postgres_home_dir}/.bash_profile'"
    ]

    if OperatingSystem.CENTOS.value in distributor:
        if  release == '9':
            commands = list(map(lambda command: command.replace(f'sudo {package_manager} install -q -y sshpass', f'sudo {package_manager} install -q -y https://mirror.stream.centos.org/9-stream/AppStream/x86_64/os/Packages/sshpass-1.09-4.el9.x86_64.rpm'), commands))
        elif release == '8':
            commands = list(map(lambda command: command.replace(f'sudo {package_manager} install -q -y sshpass', f'sudo {package_manager} install -q -y https://dl.rockylinux.org/pub/rocky/8/AppStream/x86_64/os/Packages/s/sshpass-1.09-4.el8.x86_64.rpm'), commands))
        elif release == '7':
            commands = list(map(lambda command: command.replace(f'sudo {package_manager} install -q -y sshpass', f'sudo {package_manager} install -q -y https://vault.centos.org/7.6.1810/extras/x86_64/Packages/sshpass-1.06-2.el7.x86_64.rpm'), commands))

    commands = commands[3:] if offline else commands
    for command in commands:
        try:
            jsn = json.loads(remote_command(host_new_standby, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration'])
            if filter:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    ssh_file_part = f"-i '{private_filepath}.pub'" if OperatingSystem.UBUNTU.value in distributor else ''
    pgpass_add_lines = f"\n{host_new_standby['name']}:{new_standby_data['port']}:repmgr:repmgr:$repmgr_pwd\n{host_new_standby['name']}:{new_standby_data['port']}:replication:repmgr:$repmgr_pwd"

    hba_lines = f"host\trepmgr\trepmgr\t{host_new_standby['name']}{'/32' if is_host_new_standby_ip else ''}\tmd5\n\
host\treplication\trepmgr\t{host_new_standby['name']}{'/32' if is_host_new_standby_ip else ''}\tmd5\n"
    
    if not is_host_new_standby_ip:
        hba_lines += f"host\trepmgr\trepmgr\t{host_new_standby_ip}/32\tmd5\n\
host\treplication\trepmgr\t{host_new_standby_ip}/32\tmd5\n"

    pg_password_path = f"""{TASK_TMP_FILES_DIR}/{host_new_standby['name']}"""
    # send keys of every standby host (except the new standby host, because of it already have its own key) to new standby host
    for standby in standbys_host:
        try:
            standby_host = standby['host']
            sshpassfile = f"sudo mkdir -p '{pg_password_path}' && echo '{pg_password}' | sudo tee '{pg_password_path}/pg_password' && sudo chown postgres:postgres '{pg_password_path}' && sudo -iu postgres sshpass -f '{pg_password_path}/pg_password' ssh-copy-id -p '{standby_host['ssh_port']}' -o StrictHostKeyChecking=no {ssh_file_part} 'postgres@{standby_host['name']}'"
            jsn = json.loads(remote_command(host_new_standby, sshpassfile, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno', 'permission denied'])
            
            if filter:
                if 'ssh-copy-id' in command and 'permission denied' in error_lower:
                    jsn = json.loads(remote_command(standby_host, set_pg_password, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                    
                    jsn = json.loads(remote_command(host_new_standby, sshpassfile, True), strict=False)
                    filter = error_lower != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno', 'permission denied'])
                    if filter:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            pg_password_path = f"""{TASK_TMP_FILES_DIR}/{standby_host['name']}""" 
            sshpassfile = f"sudo mkdir -p '{pg_password_path}' && echo '{pg_password}' | sudo tee '{pg_password_path}/pg_password' && sudo chown postgres:postgres '{pg_password_path}' && sudo -iu postgres sshpass -f '{pg_password_path}/pg_password' ssh-copy-id -p '{host_new_standby['ssh_port']}' -o StrictHostKeyChecking=no {ssh_file_part} 'postgres@{host_new_standby['name']}'"

            commands = [
                f"""repmgr_pwd='{repmgr_pwd}'; echo "{pgpass_add_lines}" | sudo tee -a '{pgpass_path}'""",
                f"sudo sed -i '/host\s*repmgr\s*repmgr\s*{host_new_standby['name']}{chr(92) + '/32' if is_host_new_standby_ip else ''}\s*md5/d' '{standby['data_directory']}/pg_hba.conf'",
                f"sudo sed -i '/host\s*replication\s*repmgr\s*{host_new_standby['name']}{chr(92) + '/32' if is_host_new_standby_ip else ''}\s*md5/d' '{standby['data_directory']}/pg_hba.conf'",
                f"echo '{hba_lines}' | sudo tee -a '{standby['data_directory']}/pg_hba.conf'",
                f'''sudo -iu postgres PGPASSWORD='{standby['db_pwd'].strip()}' psql -p '{standby['node_port']}' -U '{standby['db_usr']}' -d postgres -c "SELECT pg_reload_conf()"''',
                sshpassfile,
            ]

            for command in commands:
                jsn = json.loads(remote_command(standby_host, command, True), strict=False)
                if jsn['stdout'] != '':
                    print(jsn['stdout'])
                
                error_lower = jsn['stderr'].lower()
                filter = error_lower != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno', 'permission denied'])
                if filter:
                    if 'ssh-copy-id' in command and 'permission denied' in error_lower:
                        jsn = json.loads(remote_command(host_new_standby, set_pg_password, True), strict=False)
                        if jsn['stderr'] != '':
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return
                        
                        jsn = json.loads(remote_command(standby_host, sshpassfile, True), strict=False)
                        filter = error_lower != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno', 'permission denied'])
                        if filter:
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return
                    else:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    pg_password_path = f"""{TASK_TMP_FILES_DIR}/{host_new_standby['name']}"""
    # send key of primary host to new standby host
    sshpassfile = f"sudo mkdir -p '{pg_password_path}' && echo '{pg_password}' | sudo tee '{pg_password_path}/pg_password' && sudo chown postgres:postgres '{pg_password_path}' && sudo -iu postgres sshpass -f '{pg_password_path}/pg_password' ssh-copy-id -p '{host_primary['ssh_port']}' -o StrictHostKeyChecking=no {ssh_file_part} 'postgres@{host_primary['name']}'"

    try:
        jsn = json.loads(remote_command(host_new_standby, sshpassfile, True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno', 'permission denied'])
        if filter:
            if 'permission denied' in error_lower:
                jsn = json.loads(remote_command(host_primary, set_pg_password, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
                
                jsn = json.loads(remote_command(host_new_standby, sshpassfile, True), strict=False)
                filter = error_lower != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno', 'permission denied'])
                if filter:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            
            else:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    pg_password_path = f"""{TASK_TMP_FILES_DIR}/{host_primary['name']}"""
    # at last - send key of new standby host to primary host
    sshpassfile = f"sudo mkdir -p '{pg_password_path}' && echo '{pg_password}' | sudo tee '{pg_password_path}/pg_password' && sudo chown postgres:postgres '{pg_password_path}' && sudo -iu postgres sshpass -f '{pg_password_path}/pg_password' ssh-copy-id -p '{host_new_standby['ssh_port']}' -o StrictHostKeyChecking=no {ssh_file_part} 'postgres@{host_new_standby['name']}' "
    
    commands = [
        f"sudo sed -i '/host\s*repmgr\s*repmgr\s*{host_new_standby['name']}{chr(92) + '/32' if is_host_new_standby_ip else ''}\s*md5/d' '{primary_data['dir_123Cluster']}/pg_hba.conf'",
        f"sudo sed -i '/host\s*replication\s*repmgr\s*{host_new_standby['name']}{chr(92) + '/32' if is_host_new_standby_ip else ''}\s*md5/d' '{primary_data['dir_123Cluster']}/pg_hba.conf'",
        f"sudo sed -i '/host\s*all\s*{new_standby_data['database_user']}\s*{host_new_standby['name']}{chr(92) + '/32' if is_host_new_standby_ip else ''}\s*md5/d' '{primary_data['dir_123Cluster']}/pg_hba.conf'",
        f"echo -e '{hba_lines}' | sudo tee -a '{primary_data['dir_123Cluster']}/pg_hba.conf'",
        f"echo -e 'host\tall\t{new_standby_data['database_user']}\t{host_new_standby['name']}{'/32' if is_host_new_standby_ip else ''}\tmd5' | sudo tee -a  '{primary_data['dir_123Cluster']}/pg_hba.conf'",
        f'''sudo -iu postgres PGPASSWORD='{primary_data['database_pwd'].strip()}' psql -p '{primary_data['port']}' -U '{primary_data['database_user']}' -d postgres -c "SELECT pg_reload_conf()"''',
        sshpassfile,
        f"""repmgr_pwd='{repmgr_pwd}'; echo "{pgpass_add_lines}" | sudo tee -a '{pgpass_path}'"""
    ]
    for command in commands:
        try:
            jsn = json.loads(remote_command(host_primary, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])
            
            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno', 'permission denied'])
            if filter:
                if 'ssh-copy-id' in command and 'permission denied' in error_lower:
                    jsn = json.loads(remote_command(host_new_standby, set_pg_password, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                    
                    jsn = json.loads(remote_command(host_primary, sshpassfile, True), strict=False)
                    filter = error_lower != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno', 'permission denied'])
                    if filter:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                else:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    append_lines = f"node_id={str(node_id)}\nnode_name='pgnode{str(node_id)}'\n\
conninfo='host={host_new_standby['name']} port={new_standby_data['port']} dbname=repmgr user=repmgr passfile={pgpass_path}'\n\
data_directory='{standby_datadir}'\n\
pg_bindir='{postgres_bin}'\n\
use_replication_slots=yes\n\
replication_type='physical'\n\
failover='automatic'\n\
passfile='{pgpass_path}'\n\
use_primary_conninfo_password=false\n\
promote_command='{postgres_bin}/repmgr standby promote -f {repmgr_path}/{repmgr_config_file} --log-to-file'\n\
follow_command='{postgres_bin}/repmgr standby follow -f {repmgr_path}/{repmgr_config_file} --log-to-file --upstream-node-id=%n'\n"

    if OperatingSystem.CENTOS.value in distributor:
        append_lines += f"service_start_command='sudo systemctl start postgresql-{postgres_version}_{new_standby_data['port']}'\n\
service_stop_command='sudo systemctl stop postgresql-{postgres_version}_{new_standby_data['port']}'\n\
service_restart_command='sudo systemctl restart postgresql-{postgres_version}_{new_standby_data['port']}'\n\
service_reload_command='sudo systemctl reload postgresql-{postgres_version}_{new_standby_data['port']}'\n"

    elif OperatingSystem.UBUNTU.value in distributor:
        append_lines += f"service_start_command='{postgres_bin}/pg_ctl -D {standby_datadir} start -l /dev/null '\n\
service_stop_command='{postgres_bin}/pg_ctl -D {standby_datadir} stop -m fast -l /dev/null '\n\
service_restart_command='{postgres_bin}/pg_ctl -D {standby_datadir} restart -m fast -l /dev/null '\n\
service_reload_command='{postgres_bin}/pg_ctl -D {standby_datadir} reload -l /dev/null'\n\
repmgrd_service_start_command='sudo systemctl start repmgrd'\n\
repmgrd_service_stop_command='sudo systemctl stop repmgrd'\n"

    remove_append_lines = "/node_id/d;/node_name/d;/conninfo/d;/data_directory/d;/pg_bindir/d;/use_replication_slots/d;/replication_type/d;/failover/d;/promote_command/d;/repmgrd_service_start_command/d;/repmgrd_service_stop_command/d"
    if int(host_new_standby['ssh_port']) != 22:
        remove_append_lines = remove_append_lines + ";/ssh_options/d"
        append_lines = append_lines + f"ssh_options='-p {str(host_new_standby['ssh_port'])} -q -o ConnectTimeout=10'\n"

    # ubuntu: '/etc/init.d/repmgrd' or '/etc/default/repmgrd'
    service_location = '/usr/lib/systemd/system/' if OperatingSystem.CENTOS.value in distributor else '/etc/default/repmgrd'

    commands = [
        f"sudo rsync -a '{repmgr_path}/repmgr.conf' '{repmgr_path}/repmgr.conf.bak'",
        f"sudo sed -i '{remove_append_lines}' '{repmgr_path}/{repmgr_config_file}'",
        f"""echo "{append_lines}" | sudo tee -a '{repmgr_path}/{repmgr_config_file}'"""
    ]
    
    if OperatingSystem.UBUNTU.value in distributor:
        ubuntu_commands = [
            f"sudo mkdir -p '{repmgr_path}'",
            f"sudo touch '{repmgr_path}/repmgr.conf'",
            f"sudo sed -i '/REPMGRD_ENABLED=no/c REPMGRD_ENABLED=yes' '{service_location}'",
            f"""sudo sed -i '/REPMGRD_CONF=/c REPMGRD_CONF="{repmgr_path}/{repmgr_config_file}"' '{service_location}'""",
            "sudo systemctl daemon-reload",
        ]
        commands = ubuntu_commands + commands

    for command in commands:
        try:
            jsn = json.loads(remote_command(host_new_standby, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if jsn['stderr'] != '' and 'extension "repmgr" already exists' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    commands = [
        f"""repmgr_pwd='{repmgr_pwd}'; echo "{pgpass_content}" | sudo tee '{pgpass_path}'""",
        f"sudo chown postgres:postgres '{pgpass_path}'",
        f"sudo chmod 0600 '{pgpass_path}'",
        f"sudo rm -rf '{standby_datadir}'",
        f"sudo mkdir -p '{standby_datadir}'",
        f"sudo chown -R postgres:postgres '{standby_datadir}'",
        f"sudo chmod 0700 '{standby_datadir}'",
        f"""sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby clone -h '{host_primary['name']}' -p '{primary_data['port']}' -U repmgr -d repmgr --fast-checkpoint""",
        f"sudo rm -f '{standby_datadir}/postmaster.pid'",
        "sudo mkdir -p '/run/repmgr'",
        f"sudo touch '/run/repmgr/repmgrd-{postgres_version}.pid'",
        f"sudo chown -R postgres:postgres '/run/repmgr'",
    ]

    if OperatingSystem.CENTOS.value in distributor:
        centos_commands = [
            f'sudo systemctl enable repmgr-{postgres_version}',
            f"sudo rsync -a '/usr/lib/systemd/system/postgresql-{postgres_version}.service' '/usr/lib/systemd/system/postgresql-{postgres_version}_{new_standby_data['port']}.service'",
            f"sudo sed -i '/Environment=PGDATA/c Environment=PGDATA={standby_datadir}' '/usr/lib/systemd/system/postgresql-{postgres_version}_{new_standby_data['port']}.service'",
            "sudo systemctl daemon-reload",
            f"sudo systemctl enable postgresql-{postgres_version}_{new_standby_data['port']}.service",
            f"sudo systemctl start postgresql-{postgres_version}_{new_standby_data['port']}.service"
        ]

        if int(release) == 8:
            centos_commands.insert(2, f"sudo sed -i '/^KillSignal.*/a ExecStopPost=+systemctl daemon-reload' '/usr/lib/systemd/system/postgresql-{postgres_version}_{new_standby_data['port']}.service'")
        
        # repmgr 10 5.4.0 available only for redhat < 9, on redhat 9 are repmgr 10 5.3.2 
        if int(release) < 9 and postgres_version == 10:
            # maintainer of repmgr 10 5.4.0 has different name of service - rename to expected
            centos_commands = [
                "sudo rsync -a /lib/systemd/system/repmgr10.service /etc/systemd/system/repmgr-10.service",
                "sudo systemctl daemon-reload"
            ] + centos_commands
        
        commands += centos_commands

    for command in commands:
        if '~' in command:
            jsn = find_tilda(host_new_standby)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            command = command.replace('~', tilda_path)

        try:
            jsn = json.loads(remote_command(host_new_standby, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            err_list = ('info', 'notice', 'file exists', 'created symlink')
            if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    if OperatingSystem.UBUNTU.value in distributor:
        stderr = start_node(host_new_standby, postgres_version, standby_datadir)
        if stderr != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(stderr),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(stderr),1)
            return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'starting repmgr service'
    duration_step_start = time.perf_counter()
    
    try:
        commands = [
            f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby register --force --wait-sync",
            f"sudo systemctl start repmgr-{postgres_version}" if OperatingSystem.CENTOS.value in distributor else "sudo systemctl restart repmgrd"
        ]
        for command in commands:
            jsn = json.loads(remote_command(host_new_standby, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            err_list = ('info', 'notice')
            if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    # means sslmode = verify-full are requested
    if use_ssl and certificate_authentication:
        # so need to ignore 'repmgr standby clone' and generate unique certificates for host of standby...
        step = 'generating standby ssl certificates'
        duration_step_start = time.perf_counter()
        
        jsn = generate_ssl_certificates(host_new_standby, new_standby_data, certificate_authentication=certificate_authentication, ssl_backup_host=app_host)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        cmd = f"sudo systemctl restart postgresql-{postgres_version}_{new_standby_data['port']}.service" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{standby_datadir}' restart -l /dev/null"
        jsn = json.loads(remote_command(host_new_standby, cmd, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if haproxy_exists:
        step = 'configuring HAProxy'
        duration_step_start = time.perf_counter()

        try:
            conn = connect_to_postgres_database(host_new_standby['name'], 'postgres', new_standby_data['database_user'], new_standby_data['database_pwd'].strip(), new_standby_data['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return
            
            with conn:
                with conn.cursor() as cur:
                    cur.execute("SHOW config_file")
                    config_file = cur.fetchone()[0]
                    if config_file.endswith('/'):
                        config_file = config_file[:-1]
                    
                    config_dir = os.path.dirname(config_file)
            
            haproxy_data = [{"host": json.loads(i['host']), "port": i['port']} for i in data['haproxy_data']]
            pg_creds = f"pg{new_standby_data['id']} {host_new_standby['name']}:{new_standby_data['port']}"
            for i in haproxy_data:
                host_haproxy = i['host']
                haproxy_port = i['port']
                is_host_haproxy_ip = is_ip(host_haproxy['name'])
                if not is_host_haproxy_ip:
                    host_haproxy_ip = socket.gethostbyname(host_haproxy['name'])

                jsn = json.loads(remote_command(host_haproxy, f"sudo grep -qF 'server {pg_creds}' '/etc/haproxy/haproxy.cfg' && echo 'already exists'", True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
                
                if 'already exists' not in jsn['stdout']:
                    commands = [
                        "sudo test ! -f '/etc/haproxy/haproxy.cfg.bak' && sudo rsync -a '/etc/haproxy/haproxy.cfg' '/etc/haproxy/haproxy.cfg.bak'",
                        f'''last_listen_line=$(sudo grep -n '^listen cluster' '/etc/haproxy/haproxy.cfg' | sudo awk -F: '{{print $1}}' | sudo tail -n 1) &&
        if [[ $(sudo grep -n '^listen cluster_{haproxy_port}' '/etc/haproxy/haproxy.cfg' | sudo awk -F: '{{print $1}}') -eq $last_listen_line ]]; then
        echo '    server {pg_creds} check port 80' | sudo tee -a '/etc/haproxy/haproxy.cfg'
        else
        empty_line_index=$(sudo awk '/^listen cluster_{haproxy_port}/ {{start=NR}} /^$/{{if (start) {{print NR; exit}}}}' '/etc/haproxy/haproxy.cfg')
        
        if [ -n "$empty_line_index" ]; then
            sudo sed -i "${{empty_line_index}}i\ \ \ \ server {pg_creds} check port 80" '/etc/haproxy/haproxy.cfg'
        fi
        fi
        ''',
                        """sudo sed -i '/^[[:space:]]*$/d' '/etc/haproxy/haproxy.cfg' && sudo awk 'BEGIN { is_blank = 1 } /^[[:alpha:]]/ { if (!is_blank) print "" } NF { is_blank = 0 } 1' '/etc/haproxy/haproxy.cfg' | sudo tee 'temp_haproxy.cfg' && sudo mv 'temp_haproxy.cfg' '/etc/haproxy/haproxy.cfg'""",
                        "sudo systemctl restart haproxy",
                        "sudo systemctl status haproxy"
                    ]
                    for command in commands:
                        jsn = json.loads(remote_command(host_haproxy, command, True), strict=False)
                        if jsn['stderr'] != '':
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return

                commands = [
                    f"sudo sed -i '/host\s*all\s*{new_standby_data['database_user']}\s*{host_haproxy['name']}{chr(92) + '/32' if is_host_haproxy_ip else ''}\s*md5/d' '{config_dir}/pg_hba.conf'",
                    f"echo -e 'host\tall\t{new_standby_data['database_user']}\t{host_haproxy['name']}{'/32' if is_host_haproxy_ip else ''}\tmd5' | sudo tee -a '{config_dir}/pg_hba.conf'",
                ]
                if not is_host_haproxy_ip:
                    commands += [
                        f"sudo sed -i '/host\s*all\s*{new_standby_data['database_user']}\s*{host_haproxy_ip}\/32\s*md5/d' '{config_dir}/pg_hba.conf'",
                        f"echo -e 'host\tall\t{new_standby_data['database_user']}\t{host_haproxy_ip}/32\tmd5' | sudo tee -a '{config_dir}/pg_hba.conf'",
                    ]
                
                for command in commands:
                    jsn = json.loads(remote_command(host_new_standby, command, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
            
            jsn = install_healthcheck(distributor, release, host_new_standby, new_standby_data, app_host, offline)
            error_lower = jsn['stderr'].lower()
            filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'fail', 'errno'] if 'error: nothing to do' not in error_lower and 'already installed' not in error_lower))
            if filter:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            conn = connect_to_postgres_database(host_new_standby['name'], 'postgres', new_standby_data['database_user'], new_standby_data['database_pwd'].strip(), new_standby_data['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return  
            
            with conn:
                with conn.cursor() as cur:
                    cur.execute('SELECT pg_reload_conf()')
            
            # ensure that connection does not closed, 
            # because that block could be needed only if have haproxy
            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"ind_haproxy":true}' WHERE id = %s;""" % (new_standby_data['id'],))
                    conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = "configuring user 'postgres' passwordless authentication with app host"
    duration_step_start = time.perf_counter()
    
    pg_password_path_app_host = f"{TASK_TMP_FILES_DIR}/{app_host['name']}"
    
    jsn = json.loads(remote_command(host_new_standby,  f"sudo mkdir -p '{pg_password_path_app_host}' && echo '{app_host_pg_password}' | sudo tee '{pg_password_path_app_host}/pg_password' && sudo chown postgres:postgres '{pg_password_path_app_host}' && sudo -iu postgres sshpass -f '{pg_password_path_app_host}/pg_password' ssh-copy-id -p '{app_host['ssh_port']}' -o StrictHostKeyChecking=no {ssh_file_part} 'postgres@{app_host['name']}'", True, step), strict=False)
    if jsn['stderr'] != '' and 'permission denied' in jsn['stderr'].lower():
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
      
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    # if create_run_linux:
    #     try:
    #         params = {'schedule': '60 seconds', 'data': host_new_standby, 'task': 'run_linux_performance_counters', 'target_type': OperatingSystem.LINUX.value, 'target_id': host_new_standby['id'], 'created_by': created_by}
    #         insert_task(params)
    #     except Exception as error:
    #         return

    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()

    try:

        jsn = json.loads(remote_command(host_new_standby, f"sudo du -sb '{standby_datadir}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        remote_command(host_new_standby, f"""echo "{get_log_rotate_config(standby_datadir)}" | sudo tee '/etc/logrotate.d/{new_standby_data['name']}'""", True)
        params =  {'schedule': '1 hour', 'data': {'created_by': created_by, 'host_data': json.dumps(host_new_standby), 'node_name': new_standby_data['name']}, 'task': 'log_rotate_task', 'monitor_name': None,  'org_id': new_standby_data['org_id'],  'target_type': new_standby_data['type'], 'target_id': new_standby_data['id'], 'created_by': created_by}
        insert_task(params)

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, postgres_version, node_size, new_standby_data['id'],))
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,new_standby_data['id'],created_by])
                conn.commit()
                cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host_new_standby['id'],created_by,OperatingSystem.LINUX.value])
                conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    except Exception as error:
        status = Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)

    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def generateSSHKey(host_primary, host_standby, private_filepath, public_filepath):
    """
    Generate public and private keys. The filepath parameters 
    are the paths top the respective public and private key files.
    """
    from io import StringIO

    key = paramiko.RSAKey.generate(1024)
    prkey = StringIO()
    key.write_private_key(prkey)

    client = create_ssh_connection(host_primary['name'], host_primary['username'], host_primary['password'], host_primary['private_key'], int(host_primary['ssh_port']))
    if isinstance(client, tuple):
        return

    sftp = client.open_sftp()
    f = sftp.open(private_filepath, mode='w')
    f.write(prkey.getvalue().decode())
    f.close()
    sftp.close()
    
    client.close()

    client = create_ssh_connection(host_standby['name'], host_standby['username'], host_standby['password'], host_standby['private_key'], int(host_standby['ssh_port']))
    if isinstance(client, tuple):
        return

    sftp = client.open_sftp()
    f = sftp.open(public_filepath, mode='w')
    f.write(f"{key.get_name()} {key.get_base64()}")
    f.close()
    sftp.close()
    
    client.close()

def create_pg_database_task(data, task_id):
    """
        Example to create database
        con = psycopg2.connect(host = '176.9.226.3', database="cluster123", user ='user123cluster', password = '1234', port=5432)
        con.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        cursor = con.cursor()
        cursor.execute('CREATE DATABASE db2')
    """

    task = 'create pg database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        pg_node = json.loads(data['node_data'])
        host = json.loads(data['host_data'])
        db = json.loads(data['db'])
        db_name = db['name']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        step = 'creating the database'
        duration_step_start = time.perf_counter()
      
        jsn = create_database(host, pg_node, db_name)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    try:
        step = 'creating database users'
        duration_step_start = time.perf_counter()

        owner_user_name = f'{db_name}_owner_user'
        write_user_name = f'{db_name}_write_user'
        read_user_name = f'{db_name}_read_user'
        owner_user_pwd = generate_complex_password()
        read_user_pwd = generate_complex_password()
        write_user_pwd = generate_complex_password()
        logs_output = f'USERNAME = {owner_user_name}\nPASSWORD = {owner_user_pwd}\nUSERNAME = {read_user_name}\nPASSWORD = {read_user_pwd}\nUSERNAME = {write_user_name}\nPASSWORD = {write_user_pwd}'
        owner_user_pwd = format_for_shell(owner_user_pwd)
        write_user_pwd = format_for_shell(write_user_pwd)
        read_user_pwd  = format_for_shell(read_user_pwd)
        alter_user_query = "ALTER USER %s WITH LOGIN NOCREATEDB NOSUPERUSER NOREPLICATION PASSWORD '%s'"
        create_user_query = "CREATE USER %s WITH LOGIN NOCREATEDB NOSUPERUSER NOREPLICATION PASSWORD '%s'"
        
        conn = connect_to_postgres_database(host['name'], db_name, pg_node['database_user'], pg_node['database_pwd'].strip(), pg_node['port'])
        if isinstance(conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
            return
        
        with conn:
            with conn.cursor() as cur:
                cur.execute("SELECT 1 FROM pg_catalog.pg_roles WHERE rolname = %s", (owner_user_name,))
                owner_exists = cur.fetchone()
                cur.execute(alter_user_query % (owner_user_name, owner_user_pwd,) if owner_exists else create_user_query % (owner_user_name, owner_user_pwd,))
                cur.execute("SELECT 1 FROM pg_catalog.pg_roles WHERE rolname = %s", (write_user_name,))
                write_exists = cur.fetchone()
                cur.execute(alter_user_query % (write_user_name, write_user_pwd,) if write_exists else create_user_query % (write_user_name, write_user_pwd,))
                cur.execute("SELECT 1 FROM pg_catalog.pg_roles WHERE rolname = %s", (read_user_name,))
                read_exists = cur.fetchone()
                cur.execute(alter_user_query % (read_user_name, read_user_pwd,) if read_exists else create_user_query % (read_user_name, read_user_pwd,))
                cur.execute("ALTER DATABASE %s OWNER TO %s" % (db_name, owner_user_name,))
                cur.execute("GRANT CONNECT ON DATABASE %s TO %s" % (db_name, read_user_name,))
                cur.execute("GRANT USAGE ON SCHEMA public TO %s" % (read_user_name,))
                cur.execute("GRANT SELECT ON ALL TABLES IN SCHEMA public TO %s" % (read_user_name,))
                cur.execute("GRANT SELECT ON ALL SEQUENCES IN SCHEMA public TO %s" % (read_user_name,))
                cur.execute("ALTER DEFAULT PRIVILEGES IN SCHEMA public GRANT SELECT ON TABLES TO %s" % (read_user_name,))
                cur.execute("ALTER DEFAULT PRIVILEGES IN SCHEMA public GRANT SELECT ON SEQUENCES TO %s" % (read_user_name,))
                cur.execute("GRANT CONNECT ON DATABASE %s TO %s" % (db_name, write_user_name,))
                cur.execute("GRANT USAGE, CREATE ON SCHEMA public TO %s" % (write_user_name,))
                cur.execute("GRANT SELECT, INSERT, UPDATE, DELETE ON ALL TABLES IN SCHEMA public TO %s" % (write_user_name,))
                cur.execute("GRANT USAGE ON ALL SEQUENCES IN SCHEMA public TO %s" % (write_user_name,))
                cur.execute("ALTER DEFAULT PRIVILEGES IN SCHEMA public GRANT SELECT, INSERT, UPDATE, DELETE ON TABLES TO %s" % (write_user_name,))
                cur.execute("GRANT USAGE ON ALL SEQUENCES IN SCHEMA public TO %s" % (write_user_name,))
                cur.execute("ALTER DEFAULT PRIVILEGES IN SCHEMA public GRANT USAGE ON SEQUENCES TO %s" % (write_user_name,))
                cur.execute("SELECT pg_database_size(%s)", (db_name,))
                res = cur.fetchone()
                size = res if res and res[0] else 0.0

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,logs_output,None,0)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, size,db['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def drop_pg_database_task(data, task_id):
    """drop database"""
    
    task = 'drop pg database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        db = json.loads(data['db_data'])
        db_name = db['name'].strip()
        postgres_version = int(node['version'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'dropping the database'
    duration_step_start = time.perf_counter()
    
    jsn = drop_database(host, node, db_name, postgres_version)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DROPPED.value, db['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def drop_pg_node_task(data, task_id):
    """drop instance"""

    task = 'drop pg node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
        global PHYSICAL_REPLICATION_SLOT
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        postgres_version = int(node['version'])
        final_backup = data.get('final_backup')
        backup_host = data['backup_host']
        
        PG_PASSWORD = node['database_pwd'].strip()
        global PHYSICAL_REPLICATION_SLOT

        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        backup_host_distributor, backup_host_release = check_local_os_version()
        if OperatingSystem.INVALID.value in backup_host_distributor:
            jsn = backup_host_release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        postgres_bin_backup_host = find_local_postgres_bin(postgres_version, backup_host_distributor)
        
        standby_data = data.get('standby_data')
        has_standby = isinstance(standby_data, list) and len(standby_data) > 0
        if has_standby:
            standbys = [{'host': json.loads(i['host']), 'node': json.loads(i['node'])} for i in standby_data]

        haproxy_exists = data.get('haproxy_exists')
        if haproxy_exists:
            haproxy_hosts = [json.loads(i) for i in data ['haproxy_hosts']]

        data_directory = node['dir_123Cluster']
        if data_directory.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_directory = data_directory.replace('~', tilda_path)

        repmgr_file_name = "repmgr.conf"
        repmgr_config_file = f"/etc/repmgr/{postgres_version}/{repmgr_file_name}"
        stop_repmgr_command = f'sudo systemctl stop repmgrd' if OperatingSystem.UBUNTU.value in distributor else f"sudo systemctl stop repmgr-{postgres_version}"
        postgres_bin = find_postgres_bin(host, postgres_version, distributor)
        repmgr_executable = f"{postgres_bin}/repmgr"
        jsn = check_remote_directory_exists(host, repmgr_executable, 'f', is_dir_empty=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    
        repmgr_exists = 'true' in jsn['stdout']
        if repmgr_exists:
            repmgr_pwd = node['ex_params'].get('repmgr_pwd')
            if repmgr_pwd:
                repmgr_pwd = repmgr_pwd.strip()
            
            check_node_command = f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} node check -f '{repmgr_config_file}'"
            primary_command = f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} primary unregister -f '{repmgr_config_file}' --force"
            standby_command = f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} standby unregister -f '{repmgr_config_file}' --force"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node existence'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node['port']}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        if jsn['stdout'] == '':
            jsn = json.loads(remote_command(host, f"sudo rm -rf '{data_directory}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            if repmgr_exists:
                repmgr_clear = [stop_repmgr_command]
                jsn = json.loads(remote_command(host, check_node_command, True), strict=False)

                if "is primary" in jsn['stdout'] or (node["cluster_role"] == ClusterRoles.PRIMARY.value and 'is the server running' in jsn['stderr'].lower()):
                    repmgr_clear.insert(0, primary_command)
                elif "is standby" in jsn['stdout'] or (node["cluster_role"] == ClusterRoles.STANDBY.value and 'is the server running' in jsn['stderr'].lower()):
                    repmgr_clear.insert(0, standby_command)

                for command in repmgr_clear:
                    jsn = json.loads(remote_command(host, command, True), strict=False)
                    err_list = ('is the server running', 'info', f'directory "{data_directory}" does not exist', 'service not loaded', 'no such file or directory')
                    if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return

            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'{node["name"]} is already deleted',1)

            try:
                with get_postgres_connection() as conn:
                    with conn.cursor() as cur:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                        cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node["id"],))
                        conn.commit()
            except Exception as error:
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

            step = 'task finished'
            status = Severity.WARNING.value
            duration = 0
            logger(task_id,task_log_id,task,step,status,duration,created_by)
            step = 'null'
            status = Severity.WARNING.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
      
    is_primary_error = None
    conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
    if isinstance(conn, tuple):
        is_primary = False
        is_primary_error = str(conn[1])
    
    if is_primary_error:
        status = Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,is_primary_error,1)
    else:
        try:
            cur = conn.cursor()
            cur.execute('SELECT pg_is_in_recovery()')
            res = cur.fetchone()
            cur.close()
            conn.close()
        except Exception as error:
            is_primary = False
            is_primary_error = str(error)
    
    if is_primary_error:
        status = Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,is_primary_error,1)
    else:
        is_primary = res and not res[0]
        if is_primary and has_standby:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,f'{node["name"]} is primary with standbys. Switchover first.',1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,f'{node["name"]} is primary with standbys. Switchover first.',1)
            return
    
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    if node['cluster_role'] != ClusterRoles.PRIMARY.value:
        final_backup = False
    if final_backup:
        try:
            step = 'backing up the node'
            duration_step_start = time.perf_counter()

            conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return
            
            cur = conn.cursor()
            
            if postgres_version < 15:
                cur.execute("SELECT pg_is_in_backup()")
                is_in_backup = cur.fetchone()[0]
                if is_in_backup:
                    stdout = f"The database instance {host['name']}:{node['port']} is already in backup mode"
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,None,2)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,None,2)
                    return
            
            cur.execute("SELECT pg_is_in_recovery()")
            is_in_recovery = cur.fetchone()[0]
        
            if not is_in_recovery:
                try:
                    cur.execute("SHOW wal_level")
                    row = cur.fetchone()[0]
                except Exception as error:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                    return

                if 'minimal' in str(row).strip().lower():
                    # when 'minimal' - basebackup does not store wal, so we change
                    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                    cur.execute("ALTER SYSTEM SET wal_level=replica")
                    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
                
                cur.execute("SHOW archive_mode")
                row = cur.fetchone()[0]

                if str(row).strip().lower() == 'off':
                    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                    cur.execute("ALTER SYSTEM SET archive_mode=on")
                    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
                
                cur.execute("SELECT slot_name FROM pg_replication_slots WHERE slot_name=%s", (PHYSICAL_REPLICATION_SLOT,))
                slot_exists = cur.fetchone()
                if not slot_exists:
                    cur.execute("SELECT pg_create_physical_replication_slot(%s)", (PHYSICAL_REPLICATION_SLOT,))
                    conn.commit()

            backup_path =  final_backup['ex_params'].get('backup_path')

            use_method_stream = '-Xs' if is_primary else ''
            use_custom_replication_slot = f'-S {PHYSICAL_REPLICATION_SLOT}' if is_primary else ''
            commands = [
                f"sudo rm -rf '{backup_path}'",
                f"sudo mkdir -p '{backup_path}'",
                f"sudo chown -R postgres:postgres '{backup_path}'",
                f"sudo chmod 700 '{backup_path}'",
                f"sudo -iu postgres PGPASSWORD='{PG_PASSWORD}' {postgres_bin_backup_host}/pg_basebackup -h '{host['name']}' -U '{node['database_user']}' -p '{node['port']}' -D '{backup_path}' -Ft -z --checkpoint=fast {use_method_stream} {use_custom_replication_slot}"
            ]
            for cmd in commands:
                jsn = json.loads(local_command(cmd, True), strict=False)
                if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            
            jsn = json.loads(local_command(f"""sudo du -bs '{backup_path}' | awk '{{print $1}}'""", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            size = jsn['stdout'].strip()
            backup_size = float(size)/1024

            modified_date = get_date_modified_folder(folder_path=backup_path)
            if isinstance(modified_date, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
                return
            
            remote_command(host, f"sudo rm -rf '{backup_path}'", True)

            with get_postgres_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,final_backup['id']))           
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}' WHERE id=%s""" % (modified_date, final_backup['id']))
                    conn.commit()
            
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    if has_standby and (is_primary or (node['cluster_role'] == ClusterRoles.PRIMARY.value and not is_primary)):
        step = 'reconfiguring cluster'
        duration_step_start = time.perf_counter()
                
        for i in standbys:
            standby_host = i['host']
            standby = i['node']
            conn = connect_to_postgres_database(standby_host['name'], 'postgres', standby['database_user'], standby['database_pwd'].strip(), standby['port'])
            if not isinstance(conn, tuple):
                break
        
        if isinstance(conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
            return
        
        siblings = [{'host': s['host'], 'node':s['node']} for s in standbys if s['node']['id'] != standby['id']]
        jsn = switchover_repmgr(standby_host, postgres_version, standby["ex_params"].get("repmgr_pwd"), distributor, siblings)
        if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower() or 'error' in jsn['stdout'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        try:
            cur.execute("""UPDATE "clusterCore_c123_node" SET severity = %s, status = %s, ex_params = ex_params || '{"pg_primary_id":%s}', cluster_role=%s WHERE cluster_id=%s""", (Severity.OK.value, Status.OK.value, standby['id'], ClusterRoles.STANDBY.value, standby['cluster_id']))
            cur.execute("""UPDATE "clusterCore_c123_node" SET cluster_role=%s WHERE id=%s""", (standby['id'], ClusterRoles.PRIMARY.value, standby['id']))
            cur.execute("""UPDATE "clusterCore_c123_database" SET node_id = %s WHERE node_id=%s""", (standby['id'], node['id']))
            cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"new_basebackup": true}' WHERE cluster_id=%s""", (node['cluster_id'],))
            conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null' 
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        conn = connect_to_postgres_database(standby_host['name'], 'postgres', standby['database_user'], standby['database_pwd'].strip(), standby['port'])
        if isinstance(conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
            return

        with conn:
            with conn.cursor() as cur:
                cur.execute("SELECT slot_name FROM pg_replication_slots WHERE slot_name=%s", (PHYSICAL_REPLICATION_SLOT,))
                slot_exists = cur.fetchone()
                if not slot_exists:
                    cur.execute("SELECT pg_create_physical_replication_slot(%s)", (PHYSICAL_REPLICATION_SLOT,))
                    conn.commit()

        conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
        if isinstance(conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
            return

        with conn:
            with conn.cursor() as cur:
                cur.execute("SELECT slot_name FROM pg_replication_slots WHERE slot_name=%s", (PHYSICAL_REPLICATION_SLOT,))
                slot_exists = cur.fetchone()
                if slot_exists:
                    cur.execute("SELECT pg_drop_physical_replication_slot(%s)", (PHYSICAL_REPLICATION_SLOT,))
                    conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'removing data directory of node'
    duration_step_start = time.perf_counter()

    if OperatingSystem.CENTOS.value in distributor:
        commands = [
            f"sudo systemctl stop postgresql-{postgres_version}_{node['port']}.service",
            f"sudo rm -rf '{data_directory}'"
        ]
    else:
        commands = [
            f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{data_directory}' stop -l /dev/null",
            f"sudo rm -rf '{data_directory}'"
        ]

    try:
        if repmgr_exists:
            commands.insert(0, stop_repmgr_command)
            jsn = json.loads(remote_command(host, check_node_command, True), strict=False)

            if "is primary" in jsn['stdout'] or (node["cluster_role"] == ClusterRoles.PRIMARY.value and 'is the server running' in jsn['stderr'].lower()):
                commands.insert(0, primary_command)
            elif "is standby" in jsn['stdout'] or (node["cluster_role"] == ClusterRoles.STANDBY.value and 'is the server running' in jsn['stderr'].lower()):
                commands.insert(0, standby_command)

        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)

            err_list = ('is the server running', 'info', f'directory "{data_directory}" does not exist', 'service not loaded', 'no such file or directory')
            if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_deleted='true' WHERE node_id=%s""", (Severity.NOT_OK.value, Status.DELETED.value, node["id"],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    if haproxy_exists:
        step = 'configuring HAProxy'
        duration_step_start = time.perf_counter()

        for host in haproxy_hosts:
            commands = [
                "sudo test ! -f '/etc/haproxy/haproxy.cfg.bak' && sudo rsync -a '/etc/haproxy/haproxy.cfg' '/etc/haproxy/haproxy.cfg.bak'",
                f"sudo sed -i '/    server pg{node['id']}/d' '/etc/haproxy/haproxy.cfg'",
                "sudo systemctl restart haproxy",
                "sudo systemctl status haproxy"
            ]
            for command in commands:
                try:
                    jsn = json.loads(remote_command(host, command, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                except Exception as error:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                    return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def start_node(host, postgres_version, data_directory, restart=False) -> str:
    """start up postgres server function"""

    postgres_bin = find_postgres_bin(host, postgres_version, OperatingSystem.UBUNTU.value)
    client = create_ssh_connection(host['name'], host['username'], host['password'], host['private_key'], int(host['ssh_port']))
    if isinstance(client, tuple):
        stderr = str(client[1])
        return stderr
    
    try:
        option = 'restart' if restart else 'start'
        command = f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{data_directory}' {option} -l /dev/null"
        stdin, stdout, stderr = client.exec_command(command)
        exit_status = stdout.channel.recv_exit_status()# Blocking call
        stdin.channel.shutdown_write()
        try:
            stderr = stderr.read().decode('ascii').strip()
            stdout = stdout.read().decode('ascii').strip()
        except UnicodeDecodeError:
            stderr = stderr.read().decode('utf-8').strip()
            stdout = stdout.read().decode('utf-8').strip()

        if stderr != '' and 'error' in str(stderr).lower():
            return str(stderr)
    except Exception as error:
        return str(error)
    
    return ''

# def restart_node(host, postgres_version, data_directory) -> str:
#     """restart postgres server function"""

#     postgres_bin = find_postgres_bin(host, postgres_version)
#     client = create_ssh_connection(host['name'], host['username'], host['password'], host['private_key'], int(host['ssh_port']))
#     if isinstance(client, tuple):
#         stderr = str(client[1])
#         return stderr
    
#     try:
#         stdin, stdout, stderr = client.exec_command(f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{data_directory}' restart")
#         exit_status = stdout.channel.recv_exit_status()# Blocking call
#         stdin.channel.shutdown_write()
#         try:
#             stderr = stderr.read().decode('ascii').strip()
#             stdout = stdout.read().decode('ascii').strip()
#         except UnicodeDecodeError:
#             stderr = stderr.read().decode('utf-8').strip()
#             stdout = stdout.read().decode('utf-8').strip()

#         if stderr != '' and 'error' in str(stderr).lower():
#             return str(stderr)
#     except Exception as error:
#         return str(error)
    
#     return ''

def status_pg_service(client, node_port, postgres_version) -> bool:
    """verify error in service (fits on CentOS)"""
    
    try:
        command = f'sudo systemctl status postgresql-{postgres_version}_{node_port}.service'
        stdin, stdout, stderr = client.exec_command(command)
        exit_status = stdout.channel.recv_exit_status()# Blocking call
        stdin.channel.shutdown_write()
        try:
            stderr = stderr.read().decode('utf-8').strip()
            stdout = stdout.read().decode('utf-8').strip()
        except UnicodeDecodeError:
            stderr = stderr.read().decode('ascii').strip()
            stdout = stdout.read().decode('ascii').strip()
        except Exception as error:
            logging.error(f'status_pg_service: {str(error)}')

        if stdout != '' and 'lock file "postmaster.pid" already exists' in str(stdout).lower():
            # means node is already up so need to sop and start node
            return True
    except Exception as error:
        return False
    
    return False

def start_pg_node_task(data, task_id):
    """start PostgreSQL using pg_ctl"""
    # CentOS: Default location of pg_ctl is "/usr/pgsql-12/bin"

    task = 'start pg node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        postgres_version = int(node['version'])
        data_directory = node['dir_123Cluster']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'starting node'
    duration_step_start = time.perf_counter()

    stderr = ''
    if OperatingSystem.CENTOS.value in distributor:
        try:
            command = f"sudo systemctl start postgresql-{postgres_version}_{node['port']}.service"
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                stderr = str(jsn['stderr'])
            else:
                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    else:
        stderr = start_node(host, postgres_version, data_directory) 
        if stderr != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(stderr),1)
        else:
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                if stderr != '' :
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false' WHERE id=%s""", (Severity.NOT_OK.value, Status.OK.value, node['id'],))
                else:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                    cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def stop_pg_node_task(data, task_id):
    """stop PostgreSQL using pg_ctl"""

    task = 'stop pg node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        postgres_version = int(node['version'])
        data_directory = node['dir_123Cluster']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        if data_directory.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_directory = data_directory.replace('~', tilda_path)

        postgres_bin = find_postgres_bin(host, postgres_version, distributor)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'stopping node'
    duration_step_start = time.perf_counter()

    if OperatingSystem.CENTOS.value in distributor:
        command = f"sudo systemctl stop postgresql-{postgres_version}_{node['port']}.service"
    else:
        command = f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{data_directory}' stop -l /dev/null"

    try:
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        if jsn['stderr'] != '' and 'is the server running' not in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status='stopped', ind_started='false' WHERE id=%s""", (Severity.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='false' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def build_repmgr_config(host, node, distributor, target_pg_version, postgres_bin_target) -> dict:

    try:
        node_id = node['id']
        node_port = node['port']
        data_directory = node['dir_123Cluster']
        centos_dist = OperatingSystem.CENTOS.value in distributor
        postgres_home_dir = '/var/lib/pgsql' if centos_dist else '/var/lib/postgresql'
        pgpass_path = f'{postgres_home_dir}/.pgpass'
        repmgr_path = f'/etc/repmgr/{target_pg_version}'
        repmgr_config_file = 'repmgr.conf'
        hostname = host['name']
        ssh_port = int(host['ssh_port'])
        append_lines = f"node_id={node_id}\nnode_name='pgnode{node_id}'\n\
conninfo='host={hostname} port={node_port} dbname=repmgr user=repmgr passfile={pgpass_path}'\n\
data_directory='{data_directory}'\n\
pg_bindir='{postgres_bin_target}'\n\
use_replication_slots=yes\n\
replication_type='physical'\n\
failover='automatic'\n\
passfile='{pgpass_path}'\n\
use_primary_conninfo_password=false\n\
promote_command='{postgres_bin_target}/repmgr standby promote -f {repmgr_path}/{repmgr_config_file} --log-to-file'\n\
follow_command='{postgres_bin_target}/repmgr standby follow -f {repmgr_path}/{repmgr_config_file} --log-to-file --upstream-node-id=%n'\n"

        if centos_dist:
            append_lines += f"service_start_command='sudo systemctl start postgresql-{target_pg_version}_{node_port}'\n\
service_stop_command='sudo systemctl stop postgresql-{target_pg_version}_{node_port}'\n\
service_restart_command='sudo systemctl restart postgresql-{target_pg_version}_{node_port}'\n\
service_reload_command='sudo systemctl reload postgresql-{target_pg_version}_{node_port}'\n"

        elif OperatingSystem.UBUNTU.value in distributor:
            append_lines += f"service_start_command='{postgres_bin_target}/pg_ctl -D {data_directory} start -l /dev/null '\n\
service_stop_command='{postgres_bin_target}/pg_ctl -D {data_directory} stop -m fast -l /dev/null '\n\
service_restart_command='{postgres_bin_target}/pg_ctl -D {data_directory} restart -m fast -l /dev/null '\n\
service_reload_command='{postgres_bin_target}/pg_ctl -D {data_directory} reload -l /dev/null'\n\
repmgrd_service_start_command='sudo systemctl start repmgrd'\n\
repmgrd_service_stop_command='sudo systemctl stop repmgrd'\n"

        remove_append_lines = "/node_id/d;/node_name/d;/conninfo/d;/data_directory/d;/pg_bindir/d;/use_replication_slots/d;/replication_type/d;/failover/d;/promote_command/d;/repmgrd_service_start_command/d;/repmgrd_service_stop_command/d"

        if ssh_port != 22:
            remove_append_lines = remove_append_lines + ";/ssh_options/d"
            append_lines = append_lines + f"ssh_options='-p {ssh_port} -q -o ConnectTimeout=10'\n"
        
        repmgr_config_path = f'{repmgr_path}/{repmgr_config_file}'
        jsn = check_remote_directory_exists(host, repmgr_config_path, 'f', is_dir_empty=False)
        if jsn['stderr'] != '':
            return jsn
        
        repmgr_conf_exists = 'true' in jsn['stdout']
        
        commands = [
            f"""echo "{append_lines}" | sudo tee {'-a' if repmgr_conf_exists else ''} '{repmgr_config_path}'"""
        ]
        if repmgr_conf_exists:
            commands = [
                f"sudo rsync -a '{repmgr_config_path}' '{repmgr_config_path}.bak'",
                f"sudo sed -i '{remove_append_lines}' '{repmgr_config_file}'",
            ] + commands
        else:
            commands.insert(0, f"sudo mkdir -p '{repmgr_path}'")
        
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                return jsn

        return jsn
    except Exception as e:
        return {'stdout': '', 'stderr': str(e), 'excode': 9}

def recreate_repmgr(node, hostname) -> str:
    "drop repmgr extension and recreate repmgr (repmgr primary register helper)"
    
    try:
        target_pg_version = int(node['version'])
        repmgr_pwd = node['ex_params'].get('repmgr_pwd').strip()
        conn = connect_to_postgres_database(hostname, 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
        if isinstance(conn, tuple):
            return str(conn[1])
        
        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        # create database could not be run in multi-statement transaction,
        # so dont use with ... as - that way psycopg2 will create single transaction for each statement
        cur = conn.cursor()
        recovery_in_progress = True
        # check when we would be able to execute dml queries on primary
        #TODO choose appropiate interval to break from loop if primary stuck in recovery
        while recovery_in_progress:
            time.sleep(1)
            cur.execute('SELECT pg_is_in_recovery()')
            recovery_in_progress = cur.fetchone()[0]
        
        cur.execute("SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname='repmgr'")
        cur.execute(f'''DROP DATABASE IF EXISTS repmgr {'WITH (FORCE)' if int(target_pg_version) > 12 else ''}''')
        cur.execute("DROP USER IF EXISTS repmgr")
        cur.execute("CREATE USER repmgr WITH LOGIN SUPERUSER PASSWORD %s", (repmgr_pwd,))
        cur.execute("CREATE DATABASE repmgr WITH OWNER repmgr")
        cur.execute("""ALTER USER repmgr SET search_path TO repmgr, "$user", public""")
        cur.close()
        conn.close()

        conn = connect_to_postgres_database(hostname, 'repmgr', node['database_user'], node['database_pwd'].strip(), node['port'])
        if isinstance(conn, tuple):
            return str(conn[1])
        
        with conn:
            with conn.cursor() as cur:
                conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                cur.execute("DROP EXTENSION IF EXISTS repmgr CASCADE")
                conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
        
        return ''
    except Exception as error:
        return str(error)

def pg_upgrade_node_task(data, task_id):
    """upgrade PostgreSQL version using pg_upgrade"""

    task = 'pg upgrade node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        postgres_version = int(node['version'])
        data_directory = node['dir_123Cluster']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        if data_directory.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_directory = data_directory.replace('~', tilda_path)

        postgres_bin = find_postgres_bin(host, postgres_version, distributor)
        target_pg_version = int(data['target_version'])
        target_version_exists = check_pg_version_exists(host, target_pg_version, distributor)
        offline = True if Repository.LOCAL.value in node['ex_params']['repository'] else False
        app_host = json.loads(data['app_host'])
        date = datetime.now().strftime('%d%m%Y_%H%M')
        conf_backup_path = f"{TASK_TMP_FILES_DIR}/pg_conf/{node['name'].strip()}_{date}"
        use_ssl = node['ex_params'].get('ssl')
        certificate_authentication = node['ex_params'].get('certificate_authentication')
        standby_data = data.get('standbys')
        repmgr_executable = f"{postgres_bin}/repmgr"
        repmgr_path = f'/etc/repmgr/{postgres_version}'
        repmgr_config_file = 'repmgr.conf'
        jsn = check_remote_directory_exists(host, repmgr_executable, 'f', is_dir_empty=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        repmgr_primary_exists = 'true' in jsn['stdout']
        has_standby = isinstance(standby_data, list) and len(standby_data) > 0
        if has_standby:
            standbys = [{'host': json.loads(i['host']), 'node': json.loads(i['node'])} for i in standby_data]

        datadir_tmp = f"{TASK_TMP_FILES_DIR}/{os.path.basename(data_directory)}_{host['name'].strip()}_{node['port']}_{date}"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if not target_version_exists:
        step = f'installing postgresql {target_pg_version}'
        duration_step_start = time.perf_counter()

        if offline:
            jsn = install_offline_remote_to_remote(distributor, release, host, target_pg_version, DBType.POSTGRES.value, app_host)
        else:
            jsn = install_postgres_online(distributor, release, host, target_pg_version)
        
        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration'])
        if filter:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        elif 'warning' in error_lower:
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        else:
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'backing up the node config files'
    duration_step_start = time.perf_counter()

    try:
        postgres_bin_target_version = find_postgres_bin(host, target_pg_version, distributor)
        commands = [
            f"sudo mkdir -p '{conf_backup_path}'",
            f"sudo rsync -a '{data_directory}/postgresql.conf' '{conf_backup_path}' ",
            f"sudo rsync -a '{data_directory}/postgresql.auto.conf' '{conf_backup_path}' ",
            f"sudo rsync -a '{data_directory}/pg_hba.conf' '{conf_backup_path}' "
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if 'error' in jsn['stderr'] or 'error' in jsn['stdout']:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    
        # backup the ssl certificates, since we entirely destroy primary data folder
        if use_ssl:
            jsn = transfer_ssl_certificates(node, host, host, mode='backup')
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    if has_standby:
        step = 'deconstructing cluster'
        duration_step_start = time.perf_counter()
        
        last_known_err = ''
        
        try:
            last_known_err = None
            for standby in standbys:
                host_standby = standby['host']
                node_standby = standby['node']

                jsn = check_remote_directory_exists(host_standby, repmgr_executable, 'f', is_dir_empty=False)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                repmgr_node_exists = 'true' in jsn['stdout']
                if repmgr_node_exists:
                    repmgr_pwd = node['ex_params'].get('repmgr_pwd').strip()
                    commands = [
                        f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_path}/{repmgr_config_file}' standby unregister --force",
                        f"sudo systemctl stop repmgr-{postgres_version}" if OperatingSystem.CENTOS.value in distributor else "sudo systemctl stop repmgrd",
                        "sudo pkill -9 -f repmgr",
                        f"sudo systemctl stop postgresql-{postgres_version}_{node_standby['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{node_standby['dir_123Cluster']}' stop -l /dev/null",
                    ]
                    for command in commands:
                        jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                        if jsn['stdout'] != '':
                            print(jsn['stdout'])

                        err_list = ('info', 'notice', 'warning')
                        if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list) and not ('logfile stop' in command and 'is the server running' in jsn['stderr'].lower()):
                            last_known_err = jsn
                            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

            if repmgr_primary_exists:
                repmgr_pwd = node['ex_params'].get('repmgr_pwd').strip()
                commands = [
                    f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_path}/{repmgr_config_file}' primary unregister --force",
                    f"sudo systemctl stop repmgr-{postgres_version}" if OperatingSystem.CENTOS.value in distributor else "sudo systemctl stop repmgrd",
                    "sudo pkill -9 -f repmgr",
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host, command, True), strict=False)
                    if jsn['stdout'] != '':
                        print(jsn['stdout'])

                    err_list = ('info', 'notice', 'warning')
                    if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
        except Exception as error:
            logging.error(f"{task}@{step}: {task_id}: {str(error)}")
               
        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])
    
    if repmgr_primary_exists:
        step = 'upgrading repmgr on primary'
        duration_step_start = time.perf_counter()

        try:
            jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node['port']}'", True), strict=False)
            if str(node['port']) not in jsn['stdout']:
                if OperatingSystem.CENTOS.value in distributor:
                    command = f"sudo systemctl start postgresql-{postgres_version}_{node['port']}.service"
                else:
                    command = f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{data_directory}' start -l /dev/null"

                jsn = json.loads(remote_command(host, command, True), strict=False)
                if jsn['stdout'] != '':
                    print(jsn['stdout'])

                if jsn['stderr'] != '' and 'is the server running' not in jsn['stderr'].lower():
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return

            error = recreate_repmgr(node, host['name'])
            if error != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
                return
            
            # temporary, until getting the repmgr for version 17
            if target_pg_version < 17:
                
                repmgr_executable_target = f"{postgres_bin_target_version}/repmgr"
                centos_dist = OperatingSystem.CENTOS.value in distributor
                if centos_dist:
                    if int(release) < 8:
                        repmgr_version = RepmgrVersions.REDHAT7.value
                    else:
                        repmgr_version = RepmgrVersions.REDHAT_PG_LEGACY.value if postgres_version < 13 else RepmgrVersions.LATEST.value

                    if postgres_version < 11:
                        repmgr_version = RepmgrVersions.REDHAT_PG_10_DEFAULT.value
                        
                    package_manager = 'dnf' if int(release) > 7 else 'yum'
                    json_c = "http://vault.centos.org/centos/8-stream/BaseOS/x86_64/os/Packages/json-c-0.13.1-3.el8.x86_64.rpm" if int(release) == 8 else "json-c"
                    repmgr_install = f"sudo {package_manager} -y install {json_c if int(release) > 7 else ''} {f'repmgr_{postgres_version}-{repmgr_version}' if not (postgres_version == 10 and int(release) < 9) else f'https://dl.2ndquadrant.com/default/release/browse/rpm/packages/rhel/{release}/x86_64/10/repmgr10-{RepmgrVersions.REDHAT7.value}-1.el{release}.x86_64.rpm'} "
                elif OperatingSystem.UBUNTU.value in distributor:
                    repmgr_install = f"sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y postgresql-{target_pg_version}-repmgr"
                
                jsn = check_remote_directory_exists(host, repmgr_executable_target, 'f', is_dir_empty=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return

                repmgr_target_exists = 'true' in jsn['stdout']
                if not repmgr_target_exists:
                    if offline:
                        jsn = install_offline_remote_to_remote(distributor, release, host, target_pg_version, DBType.POSTGRES.value, app_host, Packages.REPMGR.value)
                        error_lower = jsn['stderr'].lower()
                        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
                        if filter:
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return
                    else: 
                        jsn = json.loads(remote_command(host, repmgr_install, True), strict=False)
                        error_lower = jsn['stderr'].lower()
                        if jsn['stderr'] != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return
                
                jsn = build_repmgr_config(host, node, distributor, target_pg_version, postgres_bin_target_version)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
         
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    step = 'stopping node'
    duration_step_start = time.perf_counter()

    if OperatingSystem.CENTOS.value in distributor:
        command = f"sudo systemctl stop postgresql-{postgres_version}_{node['port']}.service"
    else:
        command = f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{data_directory}' stop -l /dev/null"

    try:
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        if jsn['stderr'] != '' and 'is the server running' not in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    step = f'upgrading from postgresql {postgres_version} to postgresql {target_pg_version}'
    duration_step_start = time.perf_counter()

    try:
        commands = [
            f"sudo mkdir -p '{datadir_tmp}'",
            f"sudo chown -R postgres:postgres '{datadir_tmp}'",
            f"sudo chmod -R 700 '{datadir_tmp}'",
            f"sudo -iu postgres {postgres_bin_target_version}/initdb -D '{datadir_tmp}' ",
            f"sudo -iu postgres {postgres_bin_target_version}/pg_upgrade -b '{postgres_bin}' -B '{postgres_bin_target_version}' -d '{data_directory}' -D '{datadir_tmp}' ",
            f"sudo rm -rf '{data_directory}' ",
            f"sudo mv '{datadir_tmp}' '{data_directory}'",
        ]
        
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if 'error' in jsn['stderr'].lower() and 'error: nothing to do' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                if 'exists but is not empty' not in jsn['stderr'].lower():
                    remote_command(host, f"sudo rm -rf '{datadir_tmp}'", True)
                return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    step = 'restoring node config files'
    duration_step_start = time.perf_counter()

    try:
        # restore the ssl first, if present
        if use_ssl:
            jsn = transfer_ssl_certificates(node, host, host, mode='restore')
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        commands = [
            f"sudo rsync -a '{conf_backup_path}/postgresql.conf' '{data_directory}/postgresql.conf' ",
            f"sudo rsync -a '{conf_backup_path}/postgresql.auto.conf' '{data_directory}/postgresql.auto.conf' ",
            f"sudo rsync -a '{conf_backup_path}/pg_hba.conf' '{data_directory}/pg_hba.conf'",
        ]
        
        if target_pg_version > 11:
            commands.append(f"sudo sed -i 's/clientcert=1/clientcert=verify-full/g' '{data_directory}/pg_hba.conf'")
            
        commands += [
            f"sudo chown -R postgres:postgres '{data_directory}'",
            f"sudo chmod -R 700 '{data_directory}'"
        ]
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if 'error' in jsn['stderr'] or 'error' in jsn['stdout']:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    step = 'starting node'
    duration_step_start = time.perf_counter()

    try:
        commands = [
            f"sudo systemctl start postgresql-{target_pg_version}_{node['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin_target_version}/pg_ctl -D '{data_directory}' start -l /dev/null",
            f"sudo -iu postgres {postgres_bin_target_version}/vacuumdb --all --analyze-in-stages"
        ]
        if OperatingSystem.CENTOS.value in distributor:
            centos_commands = [
                f"sudo sed -i '/^PGDATA[^ ]*/c PGDATA={data_directory}' '/var/lib/pgsql/.bash_profile'",
                f"sudo rsync -a '/usr/lib/systemd/system/postgresql-{target_pg_version}.service' '/usr/lib/systemd/system/postgresql-{target_pg_version}_{node['port']}.service'",
                f"sudo sed -i '/Environment=PGDATA/c Environment=PGDATA={data_directory}' '/usr/lib/systemd/system/postgresql-{target_pg_version}_{node['port']}.service'",
                "sudo systemctl daemon-reload",
                f"sudo systemctl enable postgresql-{target_pg_version}_{node['port']}.service",
            ]
            if int(release) == 8:
                centos_commands.insert(2, f"sudo sed -i '/^KillSignal.*/a ExecStopPost=+systemctl daemon-reload' '/usr/lib/systemd/system/postgresql-{target_pg_version}_{node['port']}.service'")
            
            commands = centos_commands + commands
        
        for command in commands:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    if has_standby:
        step = 'cloning standbys'
        duration_step_start = time.perf_counter()
        
        last_known_err = None
        try:
            repmgr_path_target = f'/etc/repmgr/{target_pg_version}'
            commands = [
                f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin_target_version}/repmgr -f '{repmgr_path_target}/{repmgr_config_file}' primary register --force",
                # repmgr successfully rebuilt - restart service
                f"sudo systemctl restart repmgr-{target_pg_version}" if OperatingSystem.CENTOS.value in distributor else "sudo systemctl restart repmgrd"
            ]
            for command in commands:
                jsn = json.loads(remote_command(host, command, True), strict=False)
                if jsn['stdout'] != '':
                    print(jsn['stdout'])

                err_list = ('info', 'notice', 'warning')
                if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

            for standby in standbys:
                host_standby = standby['host']
                node_standby = standby['node']
                standby_directory = node_standby['dir_123Cluster']
                conf_backup_path = f"{TASK_TMP_FILES_DIR}/pg_conf/{node_standby['name']}_{date}"

                # means sslmode = verify-full are requested
                if use_ssl and certificate_authentication:
                    # then ssl certificates are unique per host - so keep copy of existing certificates
                    jsn = transfer_ssl_certificates(node_standby, host_standby, host_standby, mode='backup')
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                target_version_exists = check_pg_version_exists(host_standby, target_pg_version, distributor)
                if not target_version_exists:
                    if offline:
                        jsn = install_offline_remote_to_remote(distributor, release, host_standby, target_pg_version, DBType.POSTGRES.value, app_host)
                    else:
                        jsn = install_postgres_online(distributor, release, host_standby, target_pg_version)
        
                error_lower = jsn['stderr'].lower()
                filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration'])
                if filter:
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                jsn = check_remote_directory_exists(host_standby, repmgr_executable_target, 'f', is_dir_empty=False)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

                repmgr_target_exists = 'true' in jsn['stdout']
                if not repmgr_target_exists:
                    if offline:
                        jsn = install_offline_remote_to_remote(distributor, release, host_standby, target_pg_version, DBType.POSTGRES.value, app_host, Packages.REPMGR.value)
                        error_lower = jsn['stderr'].lower()
                        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
                        if filter:
                            last_known_err = jsn
                            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                    else: 
                        jsn = json.loads(remote_command(host_standby, repmgr_install, True), strict=False)
                        error_lower = jsn['stderr'].lower()
                        if jsn['stderr'] != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                            last_known_err = jsn
                            logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                jsn = build_repmgr_config(host_standby, node_standby, distributor, target_pg_version, postgres_bin_target_version)
                if jsn['stderr'] != '':
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                commands = [
                    f"sudo mkdir -p '{conf_backup_path}'",
                    f"sudo rsync -a '{standby_directory}/pg_hba.conf' '{conf_backup_path}/pg_hba.conf'",
                    f"sudo rsync -a '{standby_directory}/postgresql.conf' '{conf_backup_path}/postgresql.conf'",
                    # 'repmgr standby clone' should be made to existing empty directory
                    f"sudo rm -rf '{standby_directory}'",
                    f"sudo mkdir -p '{standby_directory}'",
                    f"sudo chown -R postgres:postgres '{standby_directory}'",
                    f"sudo chmod -R 700 '{standby_directory}'",
                    f"""sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin_target_version}/repmgr -f '{repmgr_path_target}/{repmgr_config_file}' standby clone -h '{host['name']}' -p '{node['port']}' -U repmgr -d repmgr --fast-checkpoint --force""",
                    f"sudo rsync -a  '{conf_backup_path}/pg_hba.conf' '{standby_directory}/pg_hba.conf'",
                    f"sudo rsync -a  '{conf_backup_path}/postgresql.conf' '{standby_directory}/postgresql.conf'",
                ]

                if target_pg_version > 11:
                    commands.append(f"sudo sed -i 's/clientcert=1/clientcert=verify-full/g' '{standby_directory}/pg_hba.conf'")
                
                if OperatingSystem.CENTOS.value in distributor:
                    centos_commands = [
                        f"sudo sed -i '/^PGDATA[^ ]*/c PGDATA={standby_directory}' '/var/lib/pgsql/.bash_profile'",
                        f"sudo rsync -a '/usr/lib/systemd/system/postgresql-{target_pg_version}.service' '/usr/lib/systemd/system/postgresql-{target_pg_version}_{node_standby['port']}.service'",
                        f"sudo sed -i '/Environment=PGDATA/c Environment=PGDATA={standby_directory}' '/usr/lib/systemd/system/postgresql-{target_pg_version}_{node_standby['port']}.service'",
                        "sudo systemctl daemon-reload",
                        f"sudo systemctl enable postgresql-{target_pg_version}_{node_standby['port']}.service",
                    ]
                    if int(release) == 8:
                        centos_commands.insert(2, f"sudo sed -i '/^KillSignal.*/a ExecStopPost=+systemctl daemon-reload' '/usr/lib/systemd/system/postgresql-{target_pg_version}_{node_standby['port']}.service'")
                    
                    commands += centos_commands
                
                commands += [
                    f"sudo systemctl start postgresql-{target_pg_version}_{node_standby['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin_target_version}/pg_ctl -D '{standby_directory}' start -l /dev/null",
                    f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin_target_version}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby register --force --wait-sync",
                    f"sudo systemctl start repmgr-{target_pg_version}" if OperatingSystem.CENTOS.value in distributor else "sudo systemctl restart repmgrd",
                    f"sudo rm -rf '{conf_backup_path}'"
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host_standby, command, True), strict=False)
                    err_list = ('info', 'notice', 'warning', 'no such file or directory')
                    if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                    
                # means sslmode = verify-full are requested
                if use_ssl and certificate_authentication:
                    # then ssl certificates are unique per host - so put back existing certificates
                    jsn = transfer_ssl_certificates(node_standby, host_standby, host_standby, mode='restore')
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                        
                    cmd = f"sudo systemctl restart postgresql-{target_pg_version}_{node_standby['port']}.service" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin_target_version}/pg_ctl -D '{standby_directory}' restart -l /dev/null"
                    jsn = json.loads(remote_command(host_standby, cmd, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
        except Exception as error:
            logging.error(f"{task}@{step}: {task_id}: {str(error)}")
        
        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, target_pg_version, node['id'],))
                if has_standby:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, version=%s, ind_started='true' WHERE cluster_id=%s""", (Severity.OK.value, Status.OK.value, target_pg_version, node['cluster_id'],))
                
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restart_pg_node_task(data, task_id):
    """restart PostgreSQL using pg_ctl"""
    
    task = 'restart pg node task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        postgres_version = int(node['version'])
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        data_directory = node['dir_123Cluster']

        if data_directory.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_directory = data_directory.replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'restarting node'
    duration_step_start = time.perf_counter()

    stderr = ''
    
    if OperatingSystem.CENTOS.value in distributor:
        try:
            command = f"sudo systemctl restart postgresql-{postgres_version}_{node['port']}.service"
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                stderr = str(jsn['stderr'])
            else:
                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    else:
        stderr = start_node(host, postgres_version, data_directory, restart=True) 
        if stderr != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(stderr),1)
        else:
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)



def pg_parameters_apply_task(data, task_id):
    """edit parameters of PostgreSQL cluster using ALTER SYSTEM"""
    
    task = 'pg parameters apply task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        parameters = data['parameters']
        has_postmaster = any(p for p in parameters if 'postmaster' in p['context'])
        cluster_nodes = [{"host": json.loads(n['host']), "node": json.loads(n['node'])} for n in data['cluster_data']]
        # we dont know who is primary, so do search
        primary_data = None
        primary_search_error = None
        for node_data in cluster_nodes:
            host = node_data['host']
            node = node_data['node']
            node_conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
            if isinstance(node_conn, tuple):
                 primary_search_error = str(node_conn[1])
                         
            with node_conn:
                with node_conn.cursor() as node_cur:
                    node_cur.execute("SELECT pg_is_in_recovery()")
                    is_in_recovery = node_cur.fetchone()[0]
                    if not is_in_recovery:
                        primary_data = node_data
                        break
        
        if not primary_data:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(primary_search_error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None, str(primary_search_error),1)
            return
        
        host_primary = primary_data['host']
        primary = primary_data['node']
        standbys = [n for n in cluster_nodes if n['node']['name'] != primary['name']]
        has_standby = len(standbys) > 0
        postgres_version = int(primary['version'])
        primary_datadir = primary['dir_123Cluster'].strip()
        distributor, release = check_os_version(host_primary)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        postgres_bin = find_postgres_bin(host_primary, postgres_version, distributor)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    if has_standby and has_postmaster:
        step = 'stopping all standbys'
        duration_step_start = time.perf_counter()
        
        last_known_err = None
        try:
            for standby in standbys:
                host = standby['host']
                node = standby['node']

                stop_cmd = f"sudo systemctl stop postgresql-{postgres_version}_{node['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{node['dir_123Cluster']}' stop -l /dev/null"
                jsn = json.loads(remote_command(host, stop_cmd, True), strict=False)
                if jsn['stdout'] != '':
                    print(jsn['stdout'])

                err_list = ('info', 'notice', 'warning')
                if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list) and not ('logfile stop' in stop_cmd and 'is the server running' in jsn['stderr'].lower()):
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

        except Exception as error:
            logging.error(f"{task}@{step}: {task_id}: {str(error)}")
        
        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'applying parameters'
    duration_step_start = time.perf_counter()

    try:
        primary_conn = connect_to_postgres_database(host_primary['name'], 'postgres', primary['database_user'], primary['database_pwd'].strip(), primary['port'])
        if isinstance(primary_conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(primary_conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(primary_conn[1]),1)
            return
        
        primary_conn.autocommit = True
        primary_cur = primary_conn.cursor()
        
        for parameter in parameters:
            primary_cur.execute(f"ALTER SYSTEM SET {parameter['name']} = '{parameter['setting']}'") 
            primary_cur.execute('SELECT pg_reload_conf()')
        
        primary_cur.close()
        primary_conn.close()

        if has_postmaster:
            if OperatingSystem.CENTOS.value in distributor.lower():
                restart_cmd = f"sudo systemctl restart postgresql-{postgres_version}_{primary['port']}.service"
            elif OperatingSystem.UBUNTU.value in distributor.lower():
                restart_cmd = f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{primary_datadir}' restart -l /dev/null"
            
            jsn = json.loads(remote_command(host_primary, restart_cmd, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

    
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

       
    if has_standby and has_postmaster:
        step = 'starting all standbys'
        duration_step_start = time.perf_counter()
        
        last_known_err = None
        try:
            for standby in standbys:
                host = standby['host']
                node = standby['node']

                start_cmd = f"sudo systemctl start postgresql-{postgres_version}_{node['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{node['dir_123Cluster']}' start -l /dev/null"
                jsn = json.loads(remote_command(host, start_cmd, True), strict=False)
                if jsn['stdout'] != '':
                    print(jsn['stdout'])

                err_list = ('info', 'notice', 'warning')
                if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list) and not ('logfile stop' in stop_cmd and 'is the server running' in jsn['stderr'].lower()):
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                else:
                    with get_postgres_connection() as conn_app:
                        with conn_app.cursor() as cur_app:
                            cur_app.execute("""UPDATE "clusterCore_c123_node" SET ind_started = 'true', severity=%s, status=%s, WHERE id=%s""", (Severity.OK.value, Status.OK.value,  node['id'],))

        except Exception as error:
            logging.error(f"{task}@{step}: {task_id}: {str(error)}")
        
        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)



def pg_split_brain_recovery_task(data, task_id):
    """resolve PostgreSQL split brain issue task"""
    
    task = 'pg split brain recovery task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        cluster_data = [{"host": json.loads(n['host']), "node": json.loads(n['node'])} for n in data['cluster_data']]
        pg_error_filter = ['error', 'fail', 'fatal']
        running_as_primary = []
        primaries = []
        standbys = []
        repmgr_config_file = "repmgr.conf"
        # parameters to evaluate from some node
        postgres_version = None
        postgres_bin = None
        repmgr_executable = None
        repmgr_path = None
        repmgr_pwd = None

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    step = 'searching for the best primary candidate'
    duration_step_start = time.perf_counter()

    try:
        nodes_lsn = []
        # categorize the nodes based on repmgr info
        for node_data in cluster_data:
            host = node_data['host']
            node = node_data['node']
            # if we still did not evaluate runtime parameters - evaluate it now
            if not repmgr_executable:
                distributor_node, release_node = check_os_version(host)
                if OperatingSystem.INVALID.value in distributor_node:
                   logging.error(f"{task}@{step}: {task_id}: {release_node}")
                   continue
               
                postgres_version = int(node['version'])
                postgres_bin = find_postgres_bin(host, postgres_version, distributor_node)
                repmgr_executable = f"{postgres_bin}/repmgr"
                repmgr_path = f'/etc/repmgr/{postgres_version}'
                repmgr_pwd = node["ex_params"].get("repmgr_pwd").strip()
            
            jsn = json.loads(remote_command(host, f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} node check -f '{repmgr_path}/{repmgr_config_file}'", True), strict=False)
            if jsn['stderr'] != '' and any(err in jsn['stderr'].lower() for err in pg_error_filter):
                logging.error(f"{task}@{step}: {task_id}: {jsn}")
            
            if 'is primary' in jsn['stdout']:
                if 'running as primary' in jsn['stdout']:
                    running_as_primary.append(node_data)
                else:
                    primaries.append(node_data)
                    node_conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
                    if not isinstance(node_conn, tuple):              
                        with node_conn:
                            with node_conn.cursor() as node_cur:
                                # find out what function to use for lsn obtain
                                node_cur.execute('SELECT pg_is_in_recovery()')
                                is_in_recovery = node_cur.fetchone()[0]
                                last_lsn_query = 'SELECT pg_last_wal_receive_lsn()' if is_in_recovery else 'SELECT pg_current_wal_lsn()'
                                node_cur.execute(last_lsn_query)
                                lsn = str(node_cur.fetchone()[0])
                                nodes_lsn.append({"id": node['id'], "lsn": lsn})
            else:
                standbys.append(node_data)
        
        if len(primaries) > 0: 
            last_lsn_node = max(nodes_lsn, key=lambda x: x["lsn"], default=None) if len(nodes_lsn) > 0 else primaries[0]
            new_primary_data = next(n for n in primaries if n['node']['id'] == last_lsn_node['id'])
            host_new_primary = new_primary_data['host']
            new_primary = new_primary_data['node']
            distributor, release = check_os_version(host_new_primary)
            if OperatingSystem.INVALID.value in distributor:
                jsn = release
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        else:
            #split brain and no primary are completely different situations, if task for split brain trigger for no primary - something wrong
            error = "no primary candidates are found in cluster"
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None, error,1)
            return  
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    step = 'fixing cluster split brain'
    duration_step_start = time.perf_counter()

    try:
        if len(primaries) > 1:
            other_primaries = [n for n in primaries if n['node']['id'] == new_primary['id']]
            for node_data in other_primaries:
                host = node_data['host']
                node = node_data['node']
                datadir = node['dir_123Cluster']
                clone_commands = [
                    f"sudo systemctl stop postgresql-{postgres_version}_{node['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{datadir}' stop -l /dev/null",  
                    f"sudo rm -rf '{datadir}'",
                    f"sudo mkdir -p '{datadir}'",
                    f"sudo chown -R postgres:postgres '{datadir}'",
                    f"sudo chmod 0700 '{datadir}'",
                    f"""sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_path}/{repmgr_config_file}' standby clone -h '{host_new_primary['name']}' -p '{new_primary['port']}' -U repmgr -d repmgr --fast-checkpoint --force""",
                    f"sudo systemctl start postgresql-{postgres_version}_{node['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{datadir}' start -l /dev/null",
                    f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_path}/{repmgr_config_file}' standby register --force --wait-sync",     
                ]
                for command in clone_commands:
                    jsn = json.loads(remote_command(host, command, True), strict=False)
                    if jsn['stderr'] != '' and any(err in jsn['stderr'].lower() for err in pg_error_filter):
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
        else:
            if len(running_as_primary) > 0:
                 for node_data in running_as_primary:
                    host = node_data['host']
                    node = node_data['node']
                    datadir = node['dir_123Cluster']
                    clone_commands = [
                        f"sudo systemctl stop postgresql-{postgres_version}_{node['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{datadir}' stop -l /dev/null",  
                        f"sudo rm -rf '{datadir}'",
                        f"sudo mkdir -p '{datadir}'",
                        f"sudo chown -R postgres:postgres '{datadir}'",
                        f"sudo chmod 0700 '{datadir}'",
                        f"""sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_path}/{repmgr_config_file}' standby clone -h '{host_new_primary['name']}' -p '{new_primary['port']}' -U repmgr -d repmgr --fast-checkpoint --force""",
                        f"sudo systemctl start postgresql-{postgres_version}_{node['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{datadir}' start -l /dev/null",
                        f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_path}/{repmgr_config_file}' standby register --force --wait-sync",     
                    ]
                    for command in clone_commands:
                        jsn = json.loads(remote_command(host, command, True), strict=False)
                        if jsn['stderr'] != '' and any(err in jsn['stderr'].lower() for err in pg_error_filter):
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return


        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)


def pg_cluster_primary_search_task(data, task_id):
    """pick up new primary for failed cluster"""
    
    task = 'pg cluster primary search task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        old_primary_id = data['old_primary_id']
        repmgr_config_file = "repmgr.conf"
        cluster_data = [{"host": json.loads(n['host']), "node": json.loads(n['node'])} for n in data['cluster_data']]
        # we dont know which machines are working and which are failed, so other parameters will be evaluated after we found primary candidate
        postgres_version = None
        postgres_bin = None
        repmgr_executable = None
        repmgr_path = None
        repmgr_pwd = None
        
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'searching for the best primary candidate'
    duration_step_start = time.perf_counter()

    try:
        nodes_lsn = []
        can_connect_to_repmgr = False
        for node_data in cluster_data:
            host = node_data['host']
            node = node_data['node']
            # if we still did not evaluate runtime parameters - evaluate it now
            if not repmgr_executable:
                distributor_node, release_node = check_os_version(host)
                if OperatingSystem.INVALID.value in distributor_node:
                   logging.error(f"{task}@{step}: {task_id}: {release_node}")
                   continue
                
                postgres_version = int(node['version'])
                postgres_bin = find_postgres_bin(host, postgres_version, distributor_node)
                repmgr_executable = f"{postgres_bin}/repmgr"
                repmgr_path = f'/etc/repmgr/{postgres_version}'
                repmgr_pwd = node["ex_params"].get("repmgr_pwd").strip()

            jsn = json.loads(remote_command(host, f"sudo -iu postgres {repmgr_executable} -f '{repmgr_path}/{repmgr_config_file}' cluster show", True), strict=False)
            if 'timeline' in jsn['stdout'].lower():
                can_connect_to_repmgr = True
            has_registered_primary = next((i for i in jsn['stdout'].splitlines() if '|' in i and 'primary' in i and not any(s in i for s in ['running as standby', 'unreachable', 'failed'] )), None)
            if has_registered_primary:
                error = f"""Cluster already has reachable primary with repmgr entry:
{has_registered_primary}
Check "repmgr cluster show" of machine {host['name']}.
"""
                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                step = 'null'
                status = Severity.OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None, error,1)
                return

            node_conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
            if isinstance(node_conn, tuple):
                continue
                            
            with node_conn:
                with node_conn.cursor() as node_cur:
                    # find out what function to use for lsn obtain
                    node_cur.execute('SELECT pg_is_in_recovery()')
                    is_in_recovery = node_cur.fetchone()[0]
                    last_lsn_query = 'SELECT pg_last_wal_receive_lsn()' if is_in_recovery else 'SELECT pg_current_wal_lsn()'
                    node_cur.execute(last_lsn_query)
                    lsn = str(node_cur.fetchone()[0])
                    nodes_lsn.append({"id": node['id'], "lsn": lsn})
        
        if not can_connect_to_repmgr:
            error = "could not check repmgr status of cluster"
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None, error,1)
            return

        
        last_lsn_node = max(nodes_lsn, key=lambda x: x["lsn"], default=None)
        if not last_lsn_node:
            error = "no primary candidates are found in cluster"
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None, error,1)
            return
        
        new_primary_data = next(n for n in cluster_data if n['node']['id'] == last_lsn_node['id'])
        host_new_primary = new_primary_data['host']
        new_primary = new_primary_data['node']
       
        distributor, release = check_os_version(host_new_primary)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    step = 'trying to promote new primary by repmgr'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host_new_primary, f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_path}/{repmgr_config_file}' standby promote --siblings-follow", True), strict=False)
        if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower() or 'error' in jsn['stdout'].lower():
            status = Severity.WARNING.value
        else:
            status = Severity.OK.value

        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    # if previous step was successful
    if status == Severity.OK.value:
        step = 'cloning old primary as standby'
        duration_step_start = time.perf_counter()

        try:

            old_primary_data = next(n for n in cluster_data if n['node']['id'] == old_primary_id)
            old_primary = old_primary_data['node']
            host_old_primary = old_primary_data['host']
            # may be a chance that same primary will be picked up by lsn after failure and promoted?Then are check for "do nothing" case.
            if host_old_primary['name'] != host_new_primary['name']:
                old_primary_datadir = old_primary['dir_123Cluster']
                pg_error_filter = ['error', 'fail', 'fatal']
                clone_commands = [  
                    f"sudo systemctl stop postgresql-{postgres_version}_{old_primary['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{old_primary_datadir}' stop -l /dev/null", 
                    f"sudo rm -rf '{old_primary_datadir}'",
                    f"sudo mkdir -p '{old_primary_datadir}'",
                    f"sudo chown -R postgres:postgres '{old_primary_datadir}'",
                    f"sudo chmod 0700 '{old_primary_datadir}'",
                    f"""sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby clone -h '{host_new_primary['name']}' -p '{new_primary['port']}' -U repmgr -d repmgr --fast-checkpoint --force""",
                    f"sudo systemctl start postgresql-{postgres_version}_{old_primary['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{old_primary_datadir}' start -l /dev/null",
                    f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby register --force --wait-sync",
                ]
                for command in clone_commands:
                    jsn = json.loads(remote_command(host_old_primary, command, True), strict=False)
                    if jsn['stderr'] != '' and any(err in jsn['stderr'].lower() for err in pg_error_filter):
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return

            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    else:
        step = 'rebuilding cluster around new primary'
        duration_step_start = time.perf_counter()

        try:
            last_known_err = None
            # next assume new primary node are still up after failed promote, but maybe need to add some crash recovery beforehand
            standbys = [n for n in cluster_data if n['node']['id'] != new_primary['id']]
            repmgr_pwd = new_primary['ex_params'].get('repmgr_pwd')
            pg_error_filter = ['error', 'fail', 'fatal']
            
            #since we recreating repmgr db, should ensure it not accessed
            stop_repmgr_commands = [
                f'sudo systemctl stop repmgrd' if OperatingSystem.UBUNTU.value in distributor else f"sudo systemctl stop repmgr-{postgres_version}",
                "sudo pkill -9 -f repmgr",
            ]
            for command in stop_repmgr_commands:
                jsn = json.loads(remote_command(host_new_primary, command, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            
            #after killing repmgr, primary goes into crash recovery mode for a little while, need to wait before connecting with db driver
            time.sleep(1)
            
            conn = connect_to_postgres_database(host_new_primary['name'], 'postgres', new_primary['database_user'], new_primary['database_pwd'].strip(), new_primary['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return
            
            conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
            #create database could not be run in multi-statement transaction,
            # so dont use with ... as - that way psycopg2 will create single transaction for each statement
            cur = conn.cursor()
            recovery_in_progress = True
            # check when we would be able to execute dml queries on primary 
            #TODO choose appropiate interval to break from loop if primary stuck in recovery 
            while recovery_in_progress:
                time.sleep(1)
                cur.execute('SELECT pg_is_in_recovery()')
                recovery_in_progress = cur.fetchone()[0]
            
            cur.execute("SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname='repmgr'")
            cur.execute(f'''DROP DATABASE IF EXISTS repmgr {'WITH (FORCE)' if postgres_version > 12 else ''}''')
            cur.execute("DROP USER IF EXISTS repmgr")
            cur.execute("CREATE USER repmgr WITH LOGIN SUPERUSER PASSWORD %s", (repmgr_pwd,))
            cur.execute("CREATE DATABASE repmgr WITH OWNER repmgr")
            cur.execute("""ALTER USER repmgr SET search_path TO repmgr, "$user", public""")
            cur.close()
            conn.close()
            
            primary_register_commands = [
                f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' primary register --force",
                # repmgr successfully rebuilt - restart service
                f"sudo systemctl restart repmgr-{postgres_version}" if OperatingSystem.CENTOS.value in distributor else "sudo systemctl restart repmgrd"
            ]
            for command in primary_register_commands:
                jsn = json.loads(remote_command(host_new_primary, command, True), strict=False)
                if jsn['stderr'] != '' and any(err in jsn['stderr'].lower() for err in pg_error_filter):
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return

            for standby in standbys:
                host = standby['host']
                node = standby['node']
                standby_datadir = node['dir_123Cluster']
                clone_commands = [  
                    f"sudo systemctl stop postgresql-{postgres_version}_{node['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{standby_datadir}' stop -l /dev/null",  
                    f"sudo rm -rf '{standby_datadir}'",
                    f"sudo mkdir -p '{standby_datadir}'",
                    f"sudo chown -R postgres:postgres '{standby_datadir}'",
                    f"sudo chmod 0700 '{standby_datadir}'"
                    f"""sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby clone -h '{host_new_primary['name']}' -p '{new_primary['port']}' -U repmgr -d repmgr --fast-checkpoint --force""",
                    f"sudo systemctl start postgresql-{postgres_version}_{standby['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{standby_datadir}' start -l /dev/null",
                    f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby register --force --wait-sync",
                ]
                for command in clone_commands:
                    jsn = json.loads(remote_command(host, command, True), strict=False)
                    if jsn['stderr'] != '' and any(err in jsn['stderr'].lower() for err in pg_error_filter):
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                       

            status = Severity.WARNING.value if last_known_err else Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])
        
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def pg_basebackup_task(data, task_id):
    """pg_basebackup to backup the data directory"""
    
    task = 'pg_basebackup task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        postgres_version = int(node['version'])
        backup = json.loads(data['backup_data'])
        data_directory = node['dir_123Cluster']
        compress = backup['ex_params'].get('compress')
        incremental = backup['ex_params'].get('incremental')
        incremental_physical = backup['ex_params'].get('incremental_physical')
        log_archive_path = backup['ex_params'].get('log_archive_path')
        PG_PASSWORD = node['database_pwd'].strip()
        # handle wrong input - if postgres version < 17 - disable physical incremental backup no matter what
        if postgres_version < 17:
            incremental_physical = False
        
        is_sched = backup['is_sched']
        backup_host = data['backup_host']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        postgres_bin = find_postgres_bin(host, postgres_version, distributor)
        
        backup_host_distributor, backup_host_release = check_local_os_version()
        if OperatingSystem.INVALID.value in backup_host_distributor:
            jsn = backup_host_release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        postgres_bin_backup_host = find_local_postgres_bin(postgres_version, backup_host_distributor)
        
        backup_name = backup['name']
        global PHYSICAL_REPLICATION_SLOT

        if data_directory.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_directory = data_directory.replace('~', tilda_path)

        path = backup['ex_params'].get('backup_path')
        if path.startswith('~'):
            jsn = find_local_tilda()
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            path = path.replace('~', tilda_path)

        if incremental_physical:
            conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return
               
            with conn:
                with conn.cursor() as cur:
                    cur.execute("SELECT system_identifier FROM pg_control_system()")
                    system_identifier = cur.fetchone()[0]
                    path = f"{path}/{system_identifier}"
        
        new_basebackup = True
        with get_postgres_connection() as conn:
            with conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as cur:
                cur.execute("""SELECT * FROM "clusterCore_c123_cluster" WHERE id =%s""", (node["cluster_id"],))
                cluster = cur.fetchone()
                new_basebackup = cluster["ex_params"].get("new_basebackup")

        if isinstance(incremental, str) and len(incremental) > 0 and not new_basebackup:
            path = incremental

        # d=directory, f=file
        jsn = check_local_directory_exists(path, 'd', is_dir_empty=True)
        if jsn['stderr'] != '' and 'no such file' not in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        backup_folder_empty = True if 'true' in jsn['stdout'] else False
        
        if incremental_physical:
            new_basebackup = True
            # if backup completed at least once
            if isinstance(incremental_physical, str):
                # search if some incremental path exists on primary machine
                jsn = json.loads(local_command(f"sudo find '{path}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n | tail -1 | awk '{{print $2}}'", True), strict=False)
                if jsn['stdout'] != '':
                    incremental_physical = jsn['stdout']
                    jsn = check_local_directory_exists(incremental_physical, 'd', is_dir_empty=True)
                    if jsn['stderr'] != '' and 'no such file' not in jsn['stderr'].lower():
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
            
                    incremental_folder_empty = True if 'true' in jsn['stdout'] else False
                else:
                    incremental_folder_empty = True
        
        if is_sched and (((incremental or incremental_physical) and (backup_folder_empty or new_basebackup)) or not incremental):
            subfolder = datetime.utcnow().strftime('%d%m%Y%H%M%S')
            path = f"{path}/{subfolder}"
            # since we generating new subfolder - its empty
            backup_folder_empty = True
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'verifying backup mode'
    duration_step_start = time.perf_counter()
    
    try:
        conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], PG_PASSWORD, node['port'])
        if isinstance(conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
            return
        
        cur = conn.cursor()
        
        if postgres_version < 15:
            cur.execute("SELECT pg_is_in_backup()")
            is_in_backup = cur.fetchone()[0]
            if is_in_backup:
                stdout = f"The database instance {host['name']}:{node['port']} is already in backup mode"
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,None,2)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,None,2)
                return
        
        cur.execute("SELECT pg_is_in_recovery()")
        is_in_recovery = cur.fetchone()[0]
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    if not is_in_recovery:

        try:
            cur.execute("SHOW wal_level")
            row = cur.fetchone()[0]
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        if 'minimal' in str(row).strip().lower():
            try:
                #when 'minimal' - basebackup does not store wal, so we change
                conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                cur.execute("ALTER SYSTEM SET wal_level=replica")
                conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

        try:
            cur.execute("SHOW archive_mode")
            row = cur.fetchone()[0]
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        if str(row).strip().lower() != 'on':
            try:
                conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                cur.execute("ALTER SYSTEM SET archive_mode=on")
                conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
        
        jsn = check_local_directory_exists(log_archive_path, dir_or_file='d')
        if jsn['stderr'] != '' and 'no such file' not in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        log_archive_exists = True if 'true' in jsn['stdout'] else False
        if not log_archive_exists:
            jsn = json.loads(local_command(f"sudo mkdir -p '{log_archive_path}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        commands = [
            f"sudo chown -R postgres:postgres '{log_archive_path}'",
            f"sudo chmod -R 777 '{log_archive_path}'",
        ]
        for command in commands:
            jsn = json.loads(local_command(command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        try:
            archive_command = f'''rsync -az --ignore-existing -e "ssh -p {backup_host['ssh_port']} -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null" %p postgres@{backup_host['name']}:{log_archive_path}'''
            conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
            cur.execute(f"ALTER SYSTEM SET archive_command='{archive_command}'")
            conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        try:
            cur.execute("SELECT slot_name FROM pg_replication_slots WHERE slot_name=%s", (PHYSICAL_REPLICATION_SLOT,))
            slot_exists = cur.fetchone()
            if not slot_exists:
                cur.execute("SELECT pg_create_physical_replication_slot(%s)", (PHYSICAL_REPLICATION_SLOT,))
                conn.commit()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        if incremental_physical:
            try:
                cur.execute("SHOW summarize_wal")
                row = cur.fetchone()[0]
                if str(row).strip().lower() != 'on':
                    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                    cur.execute("ALTER SYSTEM SET summarize_wal=on")
                    cur.execute("SELECT pg_reload_conf()")
                    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'setting up replication config'
    duration_step_start = time.perf_counter()
    
    try:
        last_lsn_query = 'SELECT pg_current_wal_lsn()' if not is_in_recovery else "SELECT pg_last_wal_receive_lsn()"
        cur.execute(last_lsn_query)
        pg_current_wal_lsn = cur.fetchone()[0]
        cur.close()
        conn.close()

        cmd = f"sudo systemctl restart postgresql-{postgres_version}_{node['port']}.service" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{data_directory}' restart -l /dev/null"
        jsn = json.loads(remote_command(host, cmd, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if backup_folder_empty:
        step = 'verifying space for backup'
        duration_step_start = time.perf_counter()
        
        try:
            jsn = json.loads(local_command(f"sudo mkdir -p '{path}'", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
            
        enough_space = check_space(host, backup_host, data_directory, path)
        if not enough_space:
            error = f'There is insufficient space available on the destination {path} for the backup'
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return
        elif isinstance(enough_space, tuple):
            error = str(enough_space[1])
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'starting backup'
    duration_step_start = time.perf_counter()

    chown_backup_dir = f"sudo chown -R postgres:postgres '{path}'"
    use_custom_replication_slot = f'-S {PHYSICAL_REPLICATION_SLOT}' if not is_in_recovery else ''
    use_end_lsn = f'-E {pg_current_wal_lsn}' if postgres_version > 10 else ''
    
    if is_sched and not backup_folder_empty and incremental:
        wal_location = f"{path}/pg_wal"
        commands = [
            chown_backup_dir,
            f"sudo -iu postgres PGPASSWORD='{PG_PASSWORD}' {postgres_bin_backup_host}/pg_receivewal -h '{host['name']}' -U '{node['database_user']}' -p '{node['port']}' -D '{wal_location}' -n {use_custom_replication_slot} {use_end_lsn}"
        ]
    else:
        use_method_stream = '-Xs' if not is_in_recovery else ''
        backup_format = '-Fp' if not compress else '-Ft -z'
        if isinstance(incremental_physical, str) and not incremental_folder_empty:
            physical_incremental = f"-i '{incremental_physical}/backup_manifest'"
        else:
            physical_incremental = ''
        
        basebackup = f"sudo -iu postgres PGPASSWORD='{PG_PASSWORD}' {postgres_bin_backup_host}/pg_basebackup -h '{host['name']}' -U '{node['database_user']}' -p '{node['port']}' -D '{path}' {backup_format} --checkpoint=fast {use_method_stream} {use_custom_replication_slot} {physical_incremental}"
        commands = [
            chown_backup_dir,
            basebackup
        ]

        if compress:
            commands += [
                f"sudo mv '{path}/base.tar.gz' '{path}/{backup_name}.tar.gz'",
                f"sudo chown postgres:postgres '{path}/{backup_name}.tar.gz'",
                f"sudo mv '{path}/pg_wal.tar.gz' '{path}/{backup_name}_pg_wal.tar.gz'",
                f"sudo chown postgres:postgres '{path}/{backup_name}_pg_wal.tar.gz'"
            ]

    modified_date = get_date_modified_folder(folder_path=path)
    if isinstance(modified_date, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
        return
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                command = """UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}' WHERE id=%s""" % (modified_date, backup['id'])
                cur.execute(command)
                conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    for command in commands:
        try:

            jsn = json.loads(local_command(command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    status = Severity.OK.value if status != Severity.WARNING.value else status
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking the backup size'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(local_command(f"sudo du -bs '{path}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        size = jsn['stdout'].strip()
        backup_size = float(size)/1024
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    # special case: need duration_task_end for clusterCore_c123_inventory
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity='%s', status='%s', ind_started='true', ind_deleted='false' WHERE id=%s""" % (Severity.OK.value, Status.OK.value, node['id']))
                cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"new_basebackup": false}' WHERE id=%s""", (node['cluster_id'],))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,backup['id']))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{duration}}', '{0}', true) WHERE id={1}""".format(duration_task_end,backup['id']))
                if incremental or incremental_physical:
                    if incremental_physical:
                        query = """UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{incremental_physical}}', '"{0}"', true) WHERE id={1}"""
                    else:
                        query = """UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{incremental}}', '"{0}"', true) WHERE id={1}"""

                    cur.execute(query.format(path, backup['id']))
                
                conn.commit()
                
                step = 'null'
                status = Severity.OK.value
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

def pg_dump_backup_task(data, task_id):
    """pg_dump and pg_dumpall"""
    
    task = 'pg_dump backup task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        # backup_type = dumpall or dump
        backup_type = data['backup_type']
        dumpall = 'dumpall' in backup_type
        if not dumpall:
            db_data = json.loads(data['db_data'])

        host = data['host_data']
        backup_host = data['backup_host']

        node = json.loads(data['node_data'])
        backup = json.loads(data['backup_data'])
        postgres_version = int(node['version'])
        path = backup['ex_params'].get('backup_path')
        if path.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            path = path.replace('~', tilda_path)

        data_directory = node['dir_123Cluster']
        if data_directory.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            data_directory = data_directory.replace('~', tilda_path)

        compress = backup['ex_params'].get('compress')
        plain = backup['ex_params'].get('plain')
        include_data = data['include_data']
        include_publications = data['include_publications']
        include_subscriptions = data['include_subscriptions']
        use_inserts = data['use_inserts']
        exclude_filter = []
        if not include_data:
            exclude_filter.append("--schema-only")
        if not include_publications:
            exclude_filter.append("--no-publications")
        if not include_subscriptions:
            exclude_filter.append("--no-subscriptions")
        
        backup_host_distributor, backup_host_release = check_local_os_version()
        if OperatingSystem.INVALID.value in backup_host_distributor:
            jsn = backup_host_release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        postgres_bin_backup_host = find_local_postgres_bin(postgres_version, backup_host_distributor)

        date = datetime.now().strftime('%d%m%Y_%H%M%S')
        # for -Fc use .dump
        if compress:
            file = f'{path}/{backup["name"]}_{date}.gz'
        else:
            if dumpall:
                file = f'{path}/{backup["name"]}_{date}.out'
            else:
                file = f'{path}/{backup["name"]}_{date}.dump'
    
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
     
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
   

    step = 'running dump command'
    duration_step_start = time.perf_counter()

    # -Fc = format custom
    if plain:
        dump_format = "-Fp"
    else:
        dump_format = "-Ft" if compress else "-Fc"

    try:
        if not dumpall:
            # Example command: "export PGPASSFILE='/var/lib/pgsql/.pgpass';pg_dump -h localhost -U user123cluster -d cluster123 > /root/123cluster/postgres/dumps/cluster123_dump_file.dump"
            dump_command = f"sudo -iu postgres PGPASSWORD='{node['database_pwd'].strip()}' {postgres_bin_backup_host}/pg_dump {dump_format} -h '{host['name']}' -U '{node['database_user']}' -d '{db_data['name']}' -p '{node['port']}' --serializable-deferrable --quote-all-identifiers -c --if-exists -C -f '{file}' {'--inserts' if use_inserts else ''} {' '.join(exclude_filter)}"
        else:
            # Example command: sudo -iu postgres /usr/lib/postgresql/13/bin/pg_dumpall > /root/123cluster/postgres/dumps/cluster123.out"
            parts = [
                "sudo -iu postgres",
                f"PGPASSWORD='{node['database_pwd'].strip()}'",
                f"{postgres_bin_backup_host}/pg_dumpall",
                f"-h '{host['name']}'",
                f"-U '{node['database_user']}'",
                f"-p '{node['port']}'",
                "-c --if-exists --quote-all-identifiers"
            ]

            if use_inserts:
                parts.append('--inserts')

            if postgres_version > 11:
                parts.append("--exclude-database='repmgr'")

            parts.append(f"| sudo gzip -c | sudo tee '{file}'" if compress else f"-f '{file}'")
            parts += exclude_filter

            dump_command = " ".join(parts)
        
        commands = [
            f"sudo mkdir -p '{path}'",
            f"sudo chown -R postgres:postgres '{path}'",
            dump_command,
        ]
        for command in commands:
            jsn = json.loads(local_command(command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)


    step = 'checking the backup size'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(local_command(f"sudo du -b '{file}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        size = jsn['stdout'].strip()
        backup_size = float(size)/1024
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        modified_date = get_date_modified_folder(folder_path=path)
        if isinstance(modified_date, tuple):
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return

        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                if dumpall:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                else:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_data['id'],))
                
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,backup['id']))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{backup_path}}', '"{0}"', true) WHERE id={1}""".format(file,backup['id']))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}', status='true' WHERE id=%s""" % (modified_date, backup['id']))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def pg_dumpall_restore_task(data, task_id):
    """restore from pg_dumpall"""

    task = 'pg_dumpall restore task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_data']
        node = json.loads(data['node_data'])
        postgres_version = int(node['version'])
        backup = json.loads(data['export'])
        compress = backup['ex_params'].get('compress')
        backup_host = data['backup_host']
        is_backup_host_ip = is_ip(backup_host['name'])
        if not is_backup_host_ip:
            backup_host_ip = socket.gethostbyname(backup_host['name'])
        
        PG_PASSWORD = node['database_pwd']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        path_file = backup['ex_params'].get('backup_path')
        if path_file.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            path_file = path_file.replace('~', tilda_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'preparing target for restore'
    duration_step_start = time.perf_counter()

    jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep '{node['port']}'", True), strict=False)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    if jsn['stdout'] == '':
        if OperatingSystem.CENTOS.value in distributor:
            command = f"sudo systemctl start postgresql-{postgres_version}_{node['port']}.service"
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        else:
            stderr = start_node(host, postgres_version, node['dir_123Cluster']) 
            if stderr != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,stderr,1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,stderr,1)
                return

        
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        
    step = 'running restore command'
    duration_step_start = time.perf_counter()

    try:
        # Example command: "export PGPASSFILE='/var/lib/pgsql/.pgpass';psql -h localhost -U user123cluster -d cluster123 < /root/123cluster/postgres/dumps/cluster123_dump_file.txt"
        command = f"sudo -iu postgres PGPASSWORD='{PG_PASSWORD}' psql -h '{host['name']}' -U '{node['database_user']}' -d postgres -p '{node['port']}' < '{path_file}'" if not compress else f"sudo gunzip < '{path_file}' | sudo -iu postgres PGPASSWORD='{PG_PASSWORD}' psql -h '{host['name']}' -U '{node['database_user']}' -d postgres -p '{node['port']}'"
        jsn = json.loads(local_command(command, True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        if jsn['stderr'] != '':
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        else:
            status = Severity.OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    step = 'updating app database'
    duration_step_start = time.perf_counter()
   
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                conn.commit()
                db_set_from_app = {}
                # for clarity - call dict_cur
                with conn.cursor(cursor_factory = psycopg2.extras.RealDictCursor) as dict_cur:
                    dict_cur.execute("""SELECT DISTINCT ON (name) id, name FROM "clusterCore_c123_database" WHERE node_id=%s""", (node['id'],))
                    db_list_from_app = dict_cur.fetchall()
                    db_set_from_app = {db['name'].strip() for db in db_list_from_app}

                conn_node = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
                if isinstance(conn_node, tuple):
                    error = str(conn_node[1])
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                    return
                
                with conn_node:
                    with conn_node.cursor() as cur_node:
                        cur_node.execute("SELECT datname FROM pg_database WHERE datistemplate = false AND datname NOT IN ('postgres', 'repmgr')")
                        db_set_from_restore = {db[0].strip() for db in cur_node.fetchall()}
                        common_db = db_set_from_app.intersection(db_set_from_restore)
                        missing_db = db_set_from_restore - db_set_from_app
                        extra_db = db_set_from_app - db_set_from_restore
                        for db in common_db:
                            cur_node.execute("SELECT pg_database_size(%s)", (db,))
                            temp = cur_node.fetchone()[0]
                            db_size = 0 if temp is None else float(temp)/1024
                            cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, "size"=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, Status.OK.value, db_size, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                            conn.commit()
                        
                        # db in app but not in restore
                        for db in extra_db:
                            cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", [Severity.NOT_OK.value, Status.OK.value, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                            conn.commit()
                        
                        for db in missing_db:
                            cur_node.execute("SELECT pg_database_size(%s)", (db,))
                            temp = cur_node.fetchone()[0]
                            db_size = 0 if temp is None else float(temp)/1024
                            
                            cur.execute("""INSERT INTO "clusterCore_c123_database" 
                            ("name", "type", created_date, last_update_date, size, ind_deployed, ind_started, ind_autostartup, ind_deleted, severity, sizing, database_user, database_pwd, created_by_id, last_update_by_id, node_id) 
                            VALUES (%s, %s, now(), now(), %s, 'false', 'true', 'false', 'false', %s, 'MEDIUM', %s, %s, %s, %s, %s)""", [db, DBType.POSTGRES.value, db_size, Severity.OK.value, node['database_user'], node['database_pwd'].strip(), created_by, created_by, node['id']])
                            conn.commit()
                    
                        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
                        if jsn['stderr'] != '':
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return

                        node_size = float(jsn['stdout'].strip())
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def pg_node_restore_task(data, task_id):
    """restore for pg node from basebackup"""

    task = 'pg node restore task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        primary_host = data['host_data']
        distributor, release = check_os_version(primary_host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        primary_node = json.loads(data['node_data'])
        use_ssl = primary_node['ex_params'].get('ssl')
        certificate_authentication = primary_node['ex_params'].get('certificate_authentication')
        backup = json.loads(data['backup'])
        is_sched = backup['is_sched']
        incremental = backup['ex_params'].get('incremental')
        incremental_physical = backup['ex_params'].get('incremental_physical')
        use_pg_combinebackup = False
        standbys_id = data.get('standbys_id')
        standby_data = data.get('standbys')
        app_db_timezone = data['app_db_timezone']
        has_standby = isinstance(standby_data, list) and len(standby_data) > 0
        if has_standby:
            standbys = [{'host': json.loads(i['host']), 'node': json.loads(i['node'])} for i in standby_data]
    
        autostart = data['autostart']
        backup_host = data['backup_host']
        # restore_option = 'this'/'latest'/'custom'
        restore_option = RestoreOptions[data['restore_option'].upper()]
        # example date = '2021-12-30 09:08:33'
        date = data['date'].strip()
        compress = backup['ex_params'].get('compress')
        postgres_version = int(primary_node['version'])
        repmgr_path = f'/etc/repmgr/{postgres_version}'
        repmgr_config_file = "repmgr.conf"
        data_directory = primary_node['dir_123Cluster']
        if data_directory.startswith('~'):
            tilda_path = find_tilda(primary_host)
            data_directory = data_directory.replace('~', tilda_path)

        path_file = backup['ex_params'].get('backup_path')
        if path_file.startswith('~'):
            jsn = find_tilda(primary_host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            path_file = path_file.replace('~', tilda_path)

        postgres_bin = find_postgres_bin(primary_host, postgres_version, distributor)
        repmgr_executable = f"{postgres_bin}/repmgr"
        jsn = check_remote_directory_exists(primary_host, repmgr_executable, 'f', is_dir_empty=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        repmgr_exists = 'true' in jsn['stdout']
        if repmgr_exists:
            repmgr_pwd = primary_node['ex_params'].get('repmgr_pwd').strip()
        
        log_archive_path = backup['ex_params'].get('log_archive_path')
        restore_command = f'''rsync -az --ignore-existing -e \\"ssh -p {backup_host['ssh_port']} -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null\\" \\"postgres@{backup_host['name']}:{log_archive_path}/%f\\" %p || rsync -a --ignore-existing {data_directory}/pg_wal/%f %p'''
        timezone_cmd = 'cat /etc/timezone'
        
        jsn = json.loads(local_command(timezone_cmd, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        jsn2 = json.loads(remote_command(primary_host, timezone_cmd, True), strict=False)
        if jsn2['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn2['stdout'],jsn2['stderr'],jsn2['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn2['stdout'],jsn2['stderr'],jsn2['excode'])
            return

        try:
            backup_host_timezone = pytz.timezone(jsn['stdout'])
            node_host_timezone = pytz.timezone(jsn2['stdout'])
            
            if date:
                # our restore time supplied without timezone, so need to know exact timezone of app
                restore_datetime_raw = datetime.strptime(date, "%Y-%m-%d %H:%M:%S").replace(tzinfo=pytz.timezone(app_db_timezone)) if restore_option == RestoreOptions.THIS else datetime.strptime(data['date'], "%Y-%m-%d %H:%M:%S")
                target_timestamp = restore_datetime_raw.astimezone(backup_host_timezone).timestamp()
                date = restore_datetime_raw.astimezone(node_host_timezone).strftime("%Y-%m-%d %H:%M:%S")
        except Exception as e:
            error = f'Invalid timezone input: backup host: {jsn["stdout"]} node host: {jsn2["stdout"]} error: {str(e)}'
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return
    
        if is_sched:
            if restore_option in [RestoreOptions.LATEST, RestoreOptions.CUSTOM]:
                if restore_option == RestoreOptions.CUSTOM:
                    obtain_closest_subfolder = f"sudo find '{path_file}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n"
                else:
                    obtain_closest_subfolder = f"sudo find '{path_file}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n | tail -1 | awk '{{print $2}}'"
            else:
                obtain_closest_subfolder = f"sudo find '{path_file}' -mindepth 1 -maxdepth 1 -type d -exec stat -c '%Y %n' {{}} \; | sort -n | head -1 | awk '{{print $2}}'"

            jsn = json.loads(local_command(obtain_closest_subfolder, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            if jsn['stdout'] == '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                return
            else:
                if restore_option in [RestoreOptions.THIS, RestoreOptions.LATEST]:
                    if (incremental_physical and restore_option != RestoreOptions.LATEST):
                        closest_subfolder = jsn['stdout']
                    
                    if not incremental_physical:
                        path_file = jsn['stdout']
                else:
                    output_lines = [line.strip() for line in jsn['stdout'].split("\n") if line.strip()]
                    closest_subfolder = None
                    min_time_difference = positive_infinity

                    for line in output_lines:
                        subfolder_timestamp_str, subfolder = line.split(" ", 1)
                        subfolder_timestamp = int(subfolder_timestamp_str)
                        time_difference = abs(target_timestamp - subfolder_timestamp)
                        if time_difference < min_time_difference:
                            min_time_difference = time_difference
                            closest_subfolder = subfolder

                    if not closest_subfolder:
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'could not find the closest subfolder to restore',1)
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'could not find the closest subfolder to restore',1)
                        return

                    if not incremental_physical:
                        path_file = closest_subfolder
        
        if restore_option in [RestoreOptions.THIS, RestoreOptions.CUSTOM]:
            recovery_conf = f"restore_command = '{restore_command}'\nrecovery_target_time = '{date}'\nrecovery_target_action = 'pause'\n"
        elif restore_option == RestoreOptions.LATEST:
            recovery_conf = f"restore_command = '{restore_command}'\nrecovery_target_timeline = 'latest'\nrecovery_target_action = 'pause'\n"
        
        if incremental_physical:
            if isinstance(incremental_physical, str):
                # searching for non-empty incremental directories under root directory of backup
                jsn = json.loads(local_command(f'''sudo find '{os.path.dirname(incremental_physical)}' -mindepth 1 -maxdepth 1 -type d {f" ! -newer '{closest_subfolder}'" if restore_option in [RestoreOptions.THIS, RestoreOptions.CUSTOM] else ''} -exec sh -c 'ls -A "{{}}" | grep -q .' \; -print''', True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
                
                if jsn['stdout'].strip() == '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,'backup folder is empty',1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,'backup folder is empty',1)
                    return
                
                # filter out possible empty lines and sort in the date ascending
                dir_list = sorted([l for l in jsn['stdout'].splitlines() if l.strip()])
                # we can use pg_combine_backup only when folders are more than one
                if len(dir_list) > 1:
                    use_pg_combinebackup = postgres_version > 16
                    path_file = incremental_physical
                else:
                    path_file = dir_list[0]
            else:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,'', 'incremental backup has never run successfully' , 2)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,'', 'incremental backup has never run successfully' , 2)
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
       
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if primary_host['name'] != backup_host['name']:
        step = 'copying backup files of node to new host'
        duration_step_start = time.perf_counter()
            
        try:
            node_backups = os.path.dirname(path_file)
            jsn = transfer_files_rsync(backup_host, primary_host, node_backups, [f"{node_backups}/" if use_pg_combinebackup else path_file])
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            jsn = json.loads(remote_command(primary_host, f"sudo chown -R postgres:postgres '{path_file}'", True), strict=False)
            if jsn['stderr'] !='' and 'error' in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
   
    if has_standby:
        step = 'stopping all standbys'
        duration_step_start = time.perf_counter()
        
        last_known_err = None
        try:
            for standby in standbys:
                host = standby['host']
                node = standby['node']
                repmgr_pwd = node['ex_params'].get('repmgr_pwd').strip()

                commands = [
                    f"sudo systemctl stop repmgr-{postgres_version}" if OperatingSystem.CENTOS.value in distributor else "sudo systemctl stop repmgrd",
                    f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby unregister --force",
                    f"sudo systemctl stop postgresql-{postgres_version}_{node['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{node['dir_123Cluster']}' stop -l /dev/null",
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host, command, True), strict=False)
                    if jsn['stdout'] != '':
                        print(jsn['stdout'])

                    err_list = ('info', 'notice', 'warning')
                    if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list) and not ('logfile stop' in command and 'is the server running' in jsn['stderr'].lower()):
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

            commands = [
                f"sudo systemctl stop repmgr-{postgres_version}" if OperatingSystem.CENTOS.value in distributor else "sudo systemctl stop repmgrd",
                f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' primary unregister --force"
            ]
            for command in commands:
                jsn = json.loads(remote_command(primary_host, command, True), strict=False)
                if jsn['stdout'] != '':
                    print(jsn['stdout'])

                err_list = ('info', 'notice', 'warning')
                if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
        except Exception as error:
            logging.error(f"{task}@{step}: {task_id}: {str(error)}")
        
        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])
    
    step = 'running restore commands'
    duration_step_start = time.perf_counter()

    date = datetime.now().strftime('%d%m%Y_%H%M')
    conf_backup_path = f"{TASK_TMP_FILES_DIR}/pg_conf/{primary_node['name']}_{date}"
    
    # means sslmode = verify-full are requested
    if use_ssl and certificate_authentication:
        # then ssl certificates are unique per host - so keep copy of existing certificates
        jsn = transfer_ssl_certificates(primary_node, primary_host, primary_host, mode='backup')
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    
    if compress:
        # extract backup archive name (last modified "non-wal" archive in chosen backup folder)
        backup_filename_cmd = "sudo ls -lt --file-type '{}' | grep '\.gz$' | grep -v 'pg_wal' | head -n 1 | awk '{{print $9}}'"
        # extract pg_wal archive name (last modified "wal" archive file in chosen backup folder)
        pg_wal_filename_cmd = "sudo ls -lt --file-type '{}' | grep '\.gz$' | grep 'pg_wal' | head -n 1 | awk '{{print $9}}'"
        if not use_pg_combinebackup:
            jsn = json.loads(remote_command(primary_host, backup_filename_cmd.format(path_file), True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
 
            backup_archive_file = jsn['stdout']

            jsn = json.loads(remote_command(primary_host, pg_wal_filename_cmd.format(path_file), True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    
            wal_archive_file = jsn['stdout']
        else:
            # for each incremental directory - find and unpack backup archives
            for d in dir_list:
                jsn = json.loads(remote_command(primary_host, backup_filename_cmd.format(d), True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
    
                backup_archive_file = jsn['stdout']

                jsn = json.loads(remote_command(primary_host, pg_wal_filename_cmd.format(d), True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
        
                wal_archive_file = jsn['stdout']
                commands = [
                    f"sudo tar -xzf '{d}/{backup_archive_file}' -C '{d}'",
                    f"sudo tar -xzf '{d}/{wal_archive_file}' -C '{d}/pg_wal'",
                ]
                for command in commands:
                    jsn = json.loads(remote_command(primary_host, command, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
    
    if postgres_version in [10,11]:
        commands = [
            # f"sudo mkdir -p '{wal_backup_path}'",
            # f"sudo chown postgres:postgres '{wal_backup_path}'",
            # f"sudo mv '{data_directory}/pg_wal' '{wal_backup_path}'",
            f"sudo mkdir -p '{conf_backup_path}'",
            f"sudo rsync -a '{data_directory}/pg_hba.conf' '{conf_backup_path}/pg_hba.conf'",
            f"sudo rsync -a '{data_directory}/postgresql.conf' '{conf_backup_path}/postgresql.conf'",
            f"sudo rm -rf '{data_directory}'",
            f"sudo mkdir -p '{data_directory}'"
        ]
        
        if compress and not use_pg_combinebackup:
            commands += [
                f"sudo tar -xzf '{data_directory}/{backup_archive_file}' -C '{data_directory}'",
                f"sudo tar -xzf '{data_directory}/({wal_archive_file}' -C '{data_directory}/pg_wal'",
                f"sudo find '{data_directory}' -mindepth 1 -maxdepth 1 -type f -name '*.gz' -exec rm -f {{}} \;"
            ]

        commands += [
            # moving config files one by one (since if some of it will not move - we are in trouble)
            f"sudo rsync -a '{conf_backup_path}/pg_hba.conf' '{data_directory}'",
            f"sudo rsync -a '{conf_backup_path}/postgresql.conf' '{data_directory}'",
            f"sudo touch '{data_directory}/recovery.conf'",
            f"""echo "{recovery_conf}" | sudo tee '{data_directory}/recovery.conf'""",
            f"sudo chown -R postgres:postgres '{data_directory}'",
            f"sudo chmod 700 '{data_directory}'",
            f"sudo rm -rf '{conf_backup_path}'"
        ]
    elif postgres_version >= 12:
        commands = [
            f"sudo mkdir -p '{conf_backup_path}'",
            f"sudo rsync -a '{data_directory}/pg_hba.conf' '{conf_backup_path}/pg_hba.conf'",
            f"sudo rsync -a '{data_directory}/postgresql.conf' '{conf_backup_path}/postgresql.conf'",
            f"sudo rm -rf '{data_directory}'",
            f"sudo mkdir -p '{data_directory}'",
        ]

        if use_pg_combinebackup:
            commands += [
                f"sudo chown -R postgres:postgres '{data_directory}'",
                f"sudo chmod 700 '{data_directory}'",
            ]
        
        commands += [
            f"""sudo -iu postgres {postgres_bin}/pg_combinebackup -o '{data_directory}' {' '.join([f"'{d}'" for d in dir_list])} """ if use_pg_combinebackup else f"sudo rsync -a --recursive '{path_file}/' '{data_directory}'",
            f"sudo touch '{data_directory}/recovery.signal'",
        ]

        if compress and not use_pg_combinebackup:
            commands += [
                f"sudo tar -xzf '{data_directory}/{backup_archive_file}' -C '{data_directory}'",
                f"sudo tar -xzf '{data_directory}/{wal_archive_file}' -C '{data_directory}/pg_wal'",
                f"sudo find '{data_directory}' -mindepth 1 -maxdepth 1 -type f -name '*.gz' -exec rm -f {{}} \;"
            ]
        
        commands += [
            # moving config files one by one (since if some of it will not move - we are in trouble)
            f"sudo rsync -a '{conf_backup_path}/pg_hba.conf' '{data_directory}/pg_hba.conf'",
            f"sudo rsync -a '{conf_backup_path}/postgresql.conf' '{data_directory}/postgresql.conf'",
            f"sudo sed -i '/restore_command/d' '{data_directory}/postgresql.conf'", 
            f"sudo sed -i '/recovery_target_time/d' '{data_directory}/postgresql.conf'",
            f"sudo sed -i '/recovery_target_timeline/d' '{data_directory}/postgresql.conf'",
            f"sudo sed -i '/recovery_target_action/d' '{data_directory}/postgresql.conf'",
            f"sudo sed -i '/recovery_target_inclusive/d' '{data_directory}/postgresql.conf'",
            f"""echo "{recovery_conf}" | sudo tee -a '{data_directory}/postgresql.conf'""",
            f"sudo rm -f '{data_directory}/standby.signal'",
            f"sudo chown -R postgres:postgres '{data_directory}'",
            f"sudo chmod 700 '{data_directory}'",
            f"sudo rm -rf '{conf_backup_path}'",
        ]

    try:
        # check if instance is up
        command = f"sudo pgrep -u postgres -fa -- -D | grep '{data_directory}'"
        jsn = json.loads(remote_command(primary_host, command, True), strict=False)
        if jsn['stdout'] != '':
            if OperatingSystem.CENTOS.value in distributor.lower():
                commands.insert(0, f"sudo systemctl stop postgresql-{postgres_version}_{primary_node['port']}.service")
            elif OperatingSystem.UBUNTU.value in distributor.lower():
                commands.insert(0, f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{data_directory}' stop -l /dev/null")

        err_list = ('info', 'notice', 'warning')
        if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        for command in commands:
            jsn = json.loads(remote_command(primary_host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            err_list = ('info', 'notice', 'warning')
            error_lower = jsn['stderr'].lower()
            if jsn['stderr'] != '' and not any(err in error_lower for err in err_list):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        # means sslmode = verify-full are requested
        if use_ssl and certificate_authentication:
            # then ssl certificates are unique per host - so put back existing certificates
            jsn = transfer_ssl_certificates(primary_node, primary_host, primary_host, mode='restore')
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            change_ownership_commands = [
                f"sudo chown -R postgres:postgres '{data_directory}'",
                f"sudo chmod 700 '{data_directory}'",
            ]
            for command in change_ownership_commands:
                jsn = json.loads(remote_command(primary_host, command, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
        
        if OperatingSystem.CENTOS.value in distributor.lower():
            command = f"sudo systemctl start postgresql-{postgres_version}_{primary_node['port']}.service"
            jsn = json.loads(remote_command(primary_host, command, True), strict=False)
            if jsn['stdout'] != '':
                print(jsn['stdout'])

            if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        else:
            stderr = start_node(primary_host, postgres_version, data_directory)
            if stderr != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,stderr,1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,stderr,1)
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'promoting the primary'
    duration_step_start = time.perf_counter()
    
    error = pg_wal_replay_resume(primary_host, primary_node, distributor)
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    #if we dont have password for repmgr user - do not proceed with repmgr rebuild
    if repmgr_exists and repmgr_pwd:
        step = 'resetting repmgr tables'
        duration_step_start = time.perf_counter()
        
        try:
            #since we recreating repmgr db, should ensure it not accessed
            stop_repmgr_commands = [
                f'sudo systemctl stop repmgrd' if OperatingSystem.UBUNTU.value in distributor else f"sudo systemctl stop repmgr-{postgres_version}",
                "sudo pkill -9 -f repmgr",
            ]
            for command in stop_repmgr_commands:
                jsn = json.loads(remote_command(primary_host, command, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            
            #after killing repmgr, primary goes into crash recovery mode for a little while, need to wait before connecting with db driver
            time.sleep(1)
            
            conn = connect_to_postgres_database(primary_host['name'], 'postgres', primary_node['database_user'], primary_node['database_pwd'].strip(), primary_node['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return
            
            conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
            #create database could not be run in multi-statement transaction,
            # so dont use with ... as - that way psycopg2 will create single transaction for each statement
            cur = conn.cursor()
            recovery_in_progress = True
            # check when we would be able to execute dml queries on primary 
            #TODO choose appropiate interval to break from loop if primary stuck in recovery 
            while recovery_in_progress:
                time.sleep(1)
                cur.execute('SELECT pg_is_in_recovery()')
                recovery_in_progress = cur.fetchone()[0]
            
            cur.execute("SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname='repmgr'")
            cur.execute(f'''DROP DATABASE IF EXISTS repmgr {'WITH (FORCE)' if postgres_version > 12 else ''}''')
            cur.execute("DROP USER IF EXISTS repmgr")
            cur.execute("CREATE USER repmgr WITH LOGIN SUPERUSER PASSWORD %s", (repmgr_pwd,))
            cur.execute("CREATE DATABASE repmgr WITH OWNER repmgr")
            cur.execute("""ALTER USER repmgr SET search_path TO repmgr, "$user", public""")
            cur.close()
            conn.close()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        conn = connect_to_postgres_database(primary_host['name'], 'repmgr', primary_node['database_user'], primary_node['database_pwd'].strip(), primary_node['port'])
        if isinstance(conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
            return
        
        try:
            with conn:
                with conn.cursor() as cur:
                    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                    cur.execute("DROP EXTENSION IF EXISTS repmgr CASCADE")
                    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    if has_standby:
        step = 'cloning standbys'
        duration_step_start = time.perf_counter()
        
        last_known_err = None
        try:
            commands = [
                f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' primary register --force",
                # repmgr successfully rebuilt - restart service
                f"sudo systemctl restart repmgr-{postgres_version}" if OperatingSystem.CENTOS.value in distributor else "sudo systemctl restart repmgrd"
            ]
            for command in commands:
                jsn = json.loads(remote_command(primary_host, command, True), strict=False)
                if jsn['stdout'] != '':
                    print(jsn['stdout'])

                err_list = ('info', 'notice', 'warning')
                if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                    last_known_err = jsn
                    logging.error(f"{task}@{step}: {task_id}: {last_known_err}")

            for standby in standbys:
                host = standby['host']
                node = standby['node']
                standby_directory = node['dir_123Cluster']
                conf_backup_path = f"{TASK_TMP_FILES_DIR}/pg_conf/{node['name']}_{date}"
                repmgr_pwd = node['ex_params'].get('repmgr_pwd').strip()

                # means sslmode = verify-full are requested
                if use_ssl and certificate_authentication:
                    # then ssl certificates are unique per host - so keep copy of existing certificates
                    jsn = transfer_ssl_certificates(node, host, host, mode = 'backup')
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                commands = [
                    f"sudo mkdir -p '{conf_backup_path}'",
                    f"sudo rsync -a '{standby_directory}/pg_hba.conf' '{conf_backup_path}/pg_hba.conf'",
                    f"sudo rsync -a '{standby_directory}/postgresql.conf' '{conf_backup_path}/postgresql.conf'",
                    # 'repmgr standby clone' should be made to existing empty directory
                    f"sudo rm -rf '{standby_directory}'",
                    f"sudo mkdir -p '{standby_directory}'",
                    f"sudo chown -R postgres:postgres '{standby_directory}'",
                    f"sudo chmod 0700 '{standby_directory}'",
                    f"""sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby clone -h '{primary_host['name']}' -p '{primary_node['port']}' -U repmgr -d repmgr --fast-checkpoint --force""",
                    f"sudo rsync -a  '{conf_backup_path}/pg_hba.conf' '{standby_directory}/pg_hba.conf'",
                    f"sudo rsync -a  '{conf_backup_path}/postgresql.conf' '{standby_directory}/postgresql.conf'",
                    f"sudo rm -f '{standby_directory}/recovery.done'" if postgres_version in [10,11] else '',
                    f"sudo systemctl start postgresql-{postgres_version}_{node['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{standby_directory}' start -l /dev/null",
                    f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '{repmgr_path}/{repmgr_config_file}' standby register --force --wait-sync",
                    f"sudo systemctl start repmgr-{postgres_version}" if OperatingSystem.CENTOS.value in distributor else "sudo systemctl restart repmgrd",
                    f"sudo rm -rf '{conf_backup_path}'"
                ]
                for command in commands:
                    jsn = json.loads(remote_command(host, command, True), strict=False)
                    err_list = ('info', 'notice', 'warning', 'no such file or directory')
                    if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                    
                # means sslmode = verify-full are requested
                if use_ssl and certificate_authentication:
                    # then ssl certificates are unique per host - so put back existing certificates
                    jsn = transfer_ssl_certificates(node, host, host, mode='restore')
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                        
                    cmd = f"sudo systemctl restart postgresql-{postgres_version}_{node['port']}.service" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{standby_directory}' restart -l /dev/null"
                    jsn = json.loads(remote_command(host, cmd, True), strict=False)
                    if jsn['stderr'] != '':
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
                
                if not autostart:
                    command = f"sudo systemctl stop postgresql-{postgres_version}_{node['port']}" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{standby_directory}' stop -l /dev/null"
                    jsn = json.loads(remote_command(host, command, True), strict=False)
                    if jsn['stdout'] != '':
                        print(jsn['stdout'])

                    err_list = ('info', 'notice', 'warning')
                    if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
                        last_known_err = jsn
                        logging.error(f"{task}@{step}: {task_id}: {last_known_err}")
        except Exception as error:
            logging.error(f"{task}@{step}: {task_id}: {str(error)}")
        
        status = Severity.OK.value if not last_known_err else Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == Severity.OK.value else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'getting node size'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(primary_host, f"sudo du -sb '{primary_node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024

        remote_command(primary_host, f"sudo rm -rf '{node_backups if use_pg_combinebackup else path_file}'", True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
     
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'updating app database'
    duration_step_start = time.perf_counter()
   
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                if is_sched and incremental:
                    cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"new_basebackup": true}' WHERE id=%s""", (primary_node['cluster_id'],))
                
                if autostart:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity='%s', status='%s', ind_started='true', ind_deleted='false', size=%s WHERE id=%s""" % (Severity.OK.value, Status.OK.value, node_size, primary_node['id'],))
                    if isinstance(standbys_id, list) and len(standbys_id) > 0:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=ANY(%s)""", (Severity.OK.value, Status.OK.value, standbys_id,))
                else:
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity='%s', status='%s', ind_started='false', ind_deleted='false', size=%s WHERE id=%s""" % (Severity.OK.value, Status.OK.value, node_size, primary_node['id'],))
                    if isinstance(standbys_id, list) and len(standbys_id) > 0:
                        cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='false' WHERE id=ANY(%s)""", (Severity.OK.value, Status.OK.value, standbys_id,))
                
                conn.commit()
                db_set_from_app = {}
                # for clarity - call dict_cur
                with conn.cursor(cursor_factory = psycopg2.extras.RealDictCursor) as dict_cur:
                    dict_cur.execute("""SELECT DISTINCT ON (name) id, name FROM "clusterCore_c123_database" WHERE node_id=%s""", (primary_node['id'],))
                    db_list_from_app = dict_cur.fetchall()
                    db_set_from_app = {db['name'].strip() for db in db_list_from_app}

                conn_node = connect_to_postgres_database(primary_host['name'], 'postgres', primary_node['database_user'], primary_node['database_pwd'].strip(), primary_node['port'])
                if isinstance(conn_node, tuple):
                    error = str(conn_node[1])
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                    return
                
                with conn_node:
                    with conn_node.cursor() as cur_node:
                        cur_node.execute("SELECT datname FROM pg_database WHERE datistemplate = false AND datname NOT IN ('postgres', 'repmgr')")
                        db_set_from_restore = {db[0].strip() for db in cur_node.fetchall()}
                        common_db = db_set_from_app.intersection(db_set_from_restore)
                        missing_db = db_set_from_restore - db_set_from_app
                        extra_db = db_set_from_app - db_set_from_restore
                        for db in common_db:
                            cur_node.execute("SELECT pg_database_size(%s)", (db,))
                            temp = cur_node.fetchone()[0]
                            db_size = 0 if temp is None else float(temp)/1024
                            cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, "size"=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, Status.OK.value, db_size, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                            conn.commit()
                        
                        # db in app but not in restore
                        for db in extra_db:
                            cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", [Severity.NOT_OK.value, Status.OK.value, next((row['id'] for row in db_list_from_app if row['name'] == db), None)])
                            conn.commit()
                        
                        for db in missing_db:
                            cur_node.execute("SELECT pg_database_size(%s)", (db,))
                            temp = cur_node.fetchone()[0]
                            db_size = 0 if temp is None else float(temp)/1024
                            
                            cur.execute("""INSERT INTO "clusterCore_c123_database"
                            ("name", "type", created_date, last_update_date, size, ind_deployed, ind_started, ind_autostartup, ind_deleted, severity, sizing, database_user, database_pwd, created_by_id, last_update_by_id, node_id)
                            VALUES (%s, %s, now(), now(), %s, 'false', 'true', 'false', 'false', %s, 'MEDIUM', %s, %s, %s, %s, %s)""", [db, DBType.POSTGRES.value, db_size, Severity.OK.value, primary_node['database_user'], primary_node['database_pwd'].strip(), created_by, created_by, primary_node['id']])
                            conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'clearing recovery config'
    duration_step_start = time.perf_counter()

    if postgres_version in [10,11]:
        commands = [f"sudo rm -f '{data_directory}/recovery.conf'"]
    else:
        commands = [
            f"sudo sed -i '/restore_command/d' '{data_directory}/postgresql.conf'",
            f"sudo sed -i '/recovery_target_time /d' '{data_directory}/postgresql.conf'",
            f"sudo sed -i '/recovery_target_timeline /d' '{data_directory}/postgresql.conf'",
            f"sudo sed -i '/recovery_target_action /d' '{data_directory}/postgresql.conf'",
            f"sudo sed -i '/recovery_target_inclusive /d' '{data_directory}/postgresql.conf'"
        ]
    
    if OperatingSystem.CENTOS.value in distributor.lower():
        commands.append(f"sudo systemctl restart postgresql-{postgres_version}_{primary_node['port']}.service")
    elif OperatingSystem.UBUNTU.value in distributor.lower():
        commands.append(f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{data_directory}' restart -l /dev/null")

    client = create_ssh_connection(primary_host['name'], primary_host['username'], primary_host['password'], primary_host['private_key'], int(primary_host['ssh_port']))
    if isinstance(client, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(client[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(client[1]),1)
        return
    
    try:
        for command in commands: 
            stdin, stdout, stderr = client.exec_command(command)
            exit_status = stdout.channel.recv_exit_status()# Blocking call
            stdin.channel.shutdown_write()
            try:
                stderr = stderr.read().decode('ascii').strip()
                stdout = stdout.read().decode('ascii').strip()
            except UnicodeDecodeError:
                stderr = stderr.read().decode('utf-8').strip()
                stdout = stdout.read().decode('utf-8').strip()
            
            if stderr != '' and 'error' in stderr:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,stderr,exit_status)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,stderr,exit_status)
                return
        
        client.close()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def create_database(host, node, db, template = ''):
    """create postgresql database (if not exist)"""  
    
    try:
        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
        if isinstance(conn, tuple):
            jsn['stderr'] = str(conn[1])
            jsn['excode'] = 1
            return jsn
        
        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT) 
        cur = conn.cursor()
        cur.execute('SELECT 1 FROM pg_catalog.pg_database WHERE datname = %s', (db,))
        db_exists = cur.fetchone()
        if not db_exists:
            if isinstance(template, str) and template.strip() != '':
                cur.execute("SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname=%s", (template,))
                cur.execute("CREATE DATABASE %s TEMPLATE %s" % (db,template,))
            else:
                cur.execute("CREATE DATABASE %s" % (db,))
        
        cur.close()
        conn.close()
        return jsn
    except Exception as e:
        jsn['stderr'] = str(e)
        jsn['excode'] = 9
        return jsn

def drop_database(host, node, db, pg_version, recreate = False):
    """drop/recreate postgresql database (if exists)"""
    
    try:
        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
        if isinstance(conn, tuple):
            jsn['stderr'] = str(conn[1])
            jsn['excode'] = 1
            return jsn
        
        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        cur = conn.cursor()
        cur.execute("SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname=%s", (db,))
        if pg_version < 13:
            cur.execute("DROP DATABASE IF EXISTS %s" % (db,))
        else:
            cur.execute("DROP DATABASE IF EXISTS %s WITH (FORCE)" % (db,))
        
        if recreate:
            cur.execute("CREATE DATABASE %s" % db)
        
        cur.close()
        conn.close()
        return jsn
    except Exception as e:
        jsn['stderr'] = str(e)
        jsn['excode'] = 9
        return jsn

def pg_restore_task(data, task_id):
    """pg_restore for db"""
    
    task = 'pg_restore task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host = data['host_primary']
        node = json.loads(data['node_data'])
        db_data = json.loads(data['db_data'])
        backup_host = data['backup_host']
        is_backup_host_ip = is_ip(backup_host['name'])
        if not is_backup_host_ip:
            backup_host_ip = socket.gethostbyname(backup_host['name'])
        
        backup = json.loads(data['backup'])
        compress = backup['ex_params'].get('compress')
        plain = backup['ex_params'].get('plain')
        postgres_version = int(node['version'])
        backup_host_distributor, backup_host_release = check_local_os_version()
        if OperatingSystem.INVALID.value in backup_host_distributor:
            jsn = backup_host_release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        postgres_bin_backup_host = find_local_postgres_bin(postgres_version, backup_host_distributor)
        PG_PASSWORD = node['database_pwd'].strip()
        db_name = db_data['name']
        distributor, release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            update_failed_object(node['id'])
            return

        path_file = backup['ex_params'].get('backup_path')
        if path_file.startswith('~'):
            jsn = find_tilda(host)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            tilda_path = jsn['stdout']
            path_file = path_file.replace('~', tilda_path)
        
        include_data = data['include_data']
        include_publications = data['include_publications']
        include_subscriptions = data['include_subscriptions']
        exclude_filter = []
        if not include_data:
            exclude_filter.append("--schema-only")
        if not include_publications:
            exclude_filter.append("--no-publications")
        if not include_subscriptions:
            exclude_filter.append("--no-subscriptions")
    except Exception as error: 
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking node availability'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(host, f"sudo netstat -tunlp | grep {node['port']}", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        if jsn['stdout'] == '':
            if OperatingSystem.CENTOS.value in distributor:
                command = f"sudo systemctl start postgresql-{postgres_version}_{node['port']}.service"
                jsn = json.loads(remote_command(host, command, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
            else:
                stderr = start_node(host, postgres_version, node['dir_123Cluster']) 
                if stderr != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,stderr,1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,stderr,1)
                    return
        
        jsn =  drop_database(host, node, db_name, postgres_version)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'preparing target for restore'
    duration_step_start = time.perf_counter()
    
    try:
       create_users_query = f'''
DO
$$
BEGIN
   IF EXISTS (
      SELECT FROM pg_catalog.pg_user
      WHERE  usename = '{db_name}_owner_user') THEN

      NULL;
   ELSE
      BEGIN
         CREATE USER {db_name}_owner_user WITH LOGIN PASSWORD '{db_name}_owner_user';
      EXCEPTION
         WHEN duplicate_object THEN
            NULL;
      END;
   END IF;
   IF EXISTS (
      SELECT FROM pg_catalog.pg_user
      WHERE  usename = '{db_name}_read_user') THEN
        NULL;
   ELSE
      BEGIN
         CREATE USER {db_name}_read_user WITH LOGIN PASSWORD '{db_name}_read_user';
      EXCEPTION
         WHEN duplicate_object THEN
            NULL;
      END;
   END IF;
   IF EXISTS (
      SELECT FROM pg_catalog.pg_user
      WHERE  usename = '{db_name}_write_user') THEN
        NULL;
   ELSE
      BEGIN
         CREATE USER {db_name}_write_user WITH LOGIN PASSWORD '{db_name}_write_user';
      EXCEPTION
         WHEN duplicate_object THEN
            NULL;
      END;
   END IF;
END
$$;
'''
       conn = connect_to_postgres_database(host['name'], 'postgres', node['database_user'], node['database_pwd'].strip(), node['port'])
       if isinstance(conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
            return
       
       with conn:
            with conn.cursor() as cur:
                cur.execute(create_users_query)
    
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'restoring the database'
    duration_step_start = time.perf_counter()

    if plain:
        command = f"sudo -iu postgres PGPASSWORD='{PG_PASSWORD}' psql -h '{host['name']}' -U '{node['database_user']}' -d postgres -p '{node['port']}' < '{path_file}'"
    else:
        dump_format = "-Ft" if compress else "-Fc"
        # Example command: "export PGPASSFILE='/var/lib/pgsql/.pgpass';psql -h localhost -U user123cluster -d cluster123 < /root/123cluster/postgres/dumps/cluster123_dump_file.txt"
        command = f"sudo -iu postgres PGPASSWORD='{PG_PASSWORD}' {postgres_bin_backup_host}/pg_restore -c --if-exists -C {dump_format} -h '{host['name']}' -U '{node['database_user']}' -d postgres -p '{node['port']}' '{path_file}' {' '.join(exclude_filter)}"
    
    jsn = json.loads(local_command(command, True), strict=False)
    if jsn['stdout'] != '':
        print(jsn['stdout'])

    if jsn['stderr'] != '':
        status = Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
    else:
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        jsn = json.loads(remote_command(host, f"sudo du -sb '{node['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())
        
        with  get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_data['id'],))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.WARNING.value if status == Severity.WARNING.value else Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.WARNING.value if status == Severity.WARNING.value else Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def copy_pg_database_task(data, task_id):
    """copy database from some instance into another"""

    task = 'copy pg database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        host_source = json.loads(data['host_source'])
        host_target = json.loads(data['host_target'])
        backup_host = json.loads(data['backup_host'])
        is_target_host_ip = is_ip(host_target['name'])
        if not is_target_host_ip:
            target_host_ip = socket.gethostbyname(host_target['name'])
        
        is_backup_host_ip = is_ip(backup_host['name'])
        if not is_backup_host_ip:
            backup_host_ip = socket.gethostbyname(backup_host['name'])
        
        node_source = json.loads(data['node_source'])
        node_target = json.loads(data['node_target'])
        source_version = int(node_source['version'])
        target_version = int(node_target['version'])
        db_source = json.loads(data['db_source'])
        db_target = json.loads(data['db_target'])
        # db_option = 'override' or 'copy'
        db_option = data['db_option']
        db_exists = data['db_exists']
        copy_option = data['copy_option']
        copy_name = data.get('copy_name')
        new_db_name = copy_name if isinstance(copy_name, str) and len(copy_name.strip()) > 0  else db_target["name"]
        override = db_exists and db_option == 'override'
        local = host_source['name'] == host_target['name']
        local_override = override and local
        tmp_folder = f"{TASK_TMP_FILES_DIR}/{node_source['name'].replace(':', '_')}"
        dump_path = f"{tmp_folder}/{db_source['name']}.dump"
        
        source_distributor, source_release = check_os_version(host_source)
        if OperatingSystem.INVALID.value in source_distributor:
            jsn = source_release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        backup_host_distributor, backup_host_release = check_local_os_version()
        if OperatingSystem.INVALID.value in backup_host_distributor:
            jsn = backup_host_release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        postgres_bin_source_backup_host = find_local_postgres_bin(source_version, backup_host_distributor)
        postgres_bin_target_backup_host = find_local_postgres_bin(target_version, backup_host_distributor)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if not local_override:
        step = 'fetching users data of source database'
        duration_step_start = time.perf_counter()

        try:
            conn = connect_to_postgres_database(host_source['name'], 'postgres', node_source['database_user'], node_source['database_pwd'].strip(), node_source['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return
            
            with conn:
                with conn.cursor() as cur:
                    # this returns empty query
                    # cur.execute(f"WITH c AS (SELECT rolname FROM pg_roles WHERE (has_database_privilege(rolname, '{db_source['name']}', 'CONNECT'))) SELECT 'DROP ROLE IF EXISTS ' || rolname || '; CREATE ROLE ' || rolname || ' WITH ' || (CASE a.rolcanlogin::int WHEN 1 THEN 'LOGIN' ELSE '' END) || ' ' || (CASE a.rolcreaterole::int WHEN 1 THEN 'CREATEROLE' ELSE 'NOCREATEROLE' END) || ' ' || (CASE a.rolreplication::int WHEN 1 THEN 'REPLICATION' ELSE 'NOREPLICATION' END) || ' ' || ' PASSWORD \'\'\' || p.passwd || \'\'\';' FROM pg_roles a INNER JOIN pg_shadow p ON a.rolname = p.usename WHERE a.rolname IN (SELECT c.rolname FROM c)")
                    # also
                    # cur.execute(f"WITH c AS (SELECT rolname FROM pg_roles WHERE (has_database_privilege(rolname, '{db_source['name']}', 'CONNECT')) AND (rolcanlogin)) SELECT 'DROP ROLE IF EXISTS ' || rolname || '; CREATE ROLE ' || rolname || ' WITH ' || (CASE a.rolcanlogin::int WHEN 1 THEN 'LOGIN' ELSE '' END) || ' ' || (CASE a.rolcreaterole::int WHEN 1 THEN 'CREATEROLE' ELSE 'NOCREATEROLE' END) || ' ' || (CASE a.rolreplication::int WHEN 1 THEN 'REPLICATION' ELSE 'NOREPLICATION' END) || ' ' || ' PASSWORD \'\'\' || p.passwd || \'\'\';' FROM pg_roles a INNER JOIN pg_shadow p ON a.rolname = p.usename WHERE a.rolname IN (SELECT c.rolname FROM c)")
                    # this works
                    cur.execute(f"WITH c AS (SELECT rolname FROM pg_roles WHERE (has_database_privilege(rolname, '{db_source['name']}', 'CONNECT')) AND (rolcanlogin) AND rolsuper = 'false'::boolean) SELECT 'DROP ROLE IF EXISTS ' || rolname || '; CREATE ROLE ' || rolname || ' WITH ' || (CASE a.rolcanlogin::int WHEN 1 THEN 'LOGIN' ELSE '' END) || ' ' || (CASE a.rolcreaterole::int WHEN 1 THEN 'CREATEROLE' ELSE 'NOCREATEROLE' END) || ' ' || (CASE a.rolreplication::int WHEN 1 THEN 'REPLICATION' ELSE 'NOREPLICATION' END) || ' ' || ' PASSWORD \'\'\' || p.passwd || \'\'\';' FROM pg_roles a INNER JOIN pg_shadow p ON a.rolname = p.usename WHERE a.rolname IN (SELECT c.rolname FROM c)")
                    user_migration_queries = [c[0] for c in cur.fetchall()]
                    # this returns empty query
                    # cur.execute(f"SELECT 'GRANT CONNECT ON DATABASE {new_db_name} TO ' || rolname || ';' AS ddl FROM pg_roles WHERE (has_database_privilege(rolname, '{db_source['name']}', 'CONNECT')) AND (rolname != 'postgres')")
                    # also
                    # cur.execute(f"SELECT 'GRANT CONNECT ON DATABASE {new_db_name} TO ' || rolname || ';' AS ddl FROM pg_roles WHERE (has_database_privilege(rolname, '{db_source['name']}', 'CONNECT')) AND (rolcanlogin) AND (rolname != 'postgres')")
                    # this works
                    cur.execute(f"SELECT 'GRANT CONNECT ON DATABASE {new_db_name} TO ' || rolname || ';' AS ddl FROM pg_roles WHERE (has_database_privilege(rolname, '{db_source['name']}', 'CONNECT')) AND (rolcanlogin) AND (rolname != 'postgres') AND rolsuper = 'false'::boolean")
                  
                    user_migration_queries += [c[0] for c in cur.fetchall()]
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if copy_option != 'one_time':
        step = 'setting up source instance'
        duration_step_start = time.perf_counter()
        
        try:
            conn = connect_to_postgres_database(host_source['name'], 'postgres', node_source['database_user'], node_source['database_pwd'].strip(), node_source['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return
            
            with conn:
                with conn.cursor() as cur:
                    cur.execute("SHOW config_file")
                    config_file = cur.fetchone()[0]
                    if config_file.endswith('/'):
                        config_file = config_file[:-1]
                    
                    config_dir = os.path.dirname(config_file)
                
                    # for logical replication connections from target host
                    commands = [
                        f"sudo sed -i '/host\s*all\s*{node_target['database_user']}\s*{host_target['name']}{chr(92) + '/32' if is_target_host_ip else ''}\s*md5/d' '{config_dir}/pg_hba.conf'",
                        f"echo -e 'host\tall\t{node_target['database_user']}\t{host_target['name']}{'/32' if is_target_host_ip else ''}\tmd5' | sudo tee -a '{config_dir}/pg_hba.conf'",
                    ]
                    if not is_target_host_ip:
                        commands += [
                            f"sudo sed -i '/host\s*all\s*{node_target['database_user']}\s*{target_host_ip}\/32\s*md5/d' '{config_dir}/pg_hba.conf'",
                            f"echo -e 'host\tall\t{node_target['database_user']}\t{target_host_ip}/32\tmd5' | sudo tee -a '{config_dir}/pg_hba.conf'",
                        ]
                    
                    for command in commands:
                        jsn = json.loads(remote_command(host_source, command, True), strict=False)
                        error_lower = jsn['stderr'].lower()
                        if error_lower != '':
                            status = Severity.NOT_OK.value
                            duration_step_end = time.perf_counter() - duration_step_start
                            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            step = 'null'
                            status = Severity.NOT_OK.value
                            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                            return
                    
                    cur.execute('SELECT pg_reload_conf()')
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
   
    step = 'setting up target instance'
    duration_step_start = time.perf_counter()
    
    try:
        drop_user_errors = ''
        jsn = create_database(host_target, node_target, new_db_name, template=db_source['name'] if local and not override else '')
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        conn = connect_to_postgres_database(host_target['name'], 'postgres', node_target['database_user'], node_target['database_pwd'].strip(), node_target['port'])
        if isinstance(conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
            return
            
        with conn:
            with conn.cursor() as cur:
                if  not local_override:
                    for query in user_migration_queries:
                        cur.execute(query)
    except Exception as error:
        strerror = str(error)
        if 'cannot be dropped' in strerror.lower():
            drop_user_errors = f"{drop_user_errors}\n{jsn['stderr']}"
        else:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,strerror,9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,strerror,9)
            return
    
    status = Severity.OK.value if drop_user_errors == '' else Severity.WARNING.value
    duration_step_end = time.perf_counter() - duration_step_start
    if status == Severity.OK.value:
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    else:
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,drop_user_errors,1)
    
    if copy_option == 'one_time':
        if local_override or not local:
            step = 'copying the database'
            duration_step_start = time.perf_counter()
            
            # warning! dump/restore does not fail if database does not exist
            try:
                # whether to recreate db
                use_create = '-C' if override else ''
                # dump from local source to local file
                dump_command = f"sudo -iu postgres PGPASSWORD='{node_source['database_pwd'].strip()}' {postgres_bin_source_backup_host}/pg_dump -Fc -h '{host_source['name']}' -U '{node_source['database_user']}' -d '{db_source['name']}' -p '{node_source['port']}' --serializable-deferrable --quote-all-identifiers -c --if-exists {use_create} -f '{dump_path}'"
                # name of database to connect on restore
                restore_connect_db = 'postgres' if override else new_db_name
                #restore from local dump to remote instance
                restore_command = f"sudo -iu postgres PGPASSWORD='{node_target['database_pwd'].strip()}' {postgres_bin_target_backup_host}/pg_restore -c --if-exists {use_create} -Fc -h '{host_target['name']}' -U '{node_target['database_user']}' -d {restore_connect_db} -p '{node_target['port']}' '{dump_path}'"
                commands = [
                    f"sudo mkdir -p '{tmp_folder}'",
                    f"sudo chown -R postgres:postgres '{tmp_folder}'",
                    dump_command,
                ]
                for command in commands:
                    jsn = json.loads(local_command(command, True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                
                if override:
                    jsn =  drop_database(host_target, node_target, new_db_name, target_version)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by, jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                
                jsn = json.loads(local_command(restore_command, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.WARNING.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                else:
                    status = Severity.OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
                return
    else:
        step = 'copying tables structures'
        duration_step_start = time.perf_counter()

        try:
            jsn = json.loads(local_command(f"""sudo -iu postgres PGPASSWORD='{node_source['database_pwd'].strip()}' {postgres_bin_source_backup_host}/pg_dump -c --if-exists --no-publications --no-subscriptions --quote-all-identifiers -h '{host_source['name']}' -p '{node_source['port']}' -U '{node_source['database_user']}' -d '{db_source['name']}' --schema-only | PGPASSWORD='{node_target['database_pwd'].strip()}' psql -h '{host_target['name']}' -p '{node_target['port']}' -d '{new_db_name}' -U '{node_target['database_user']}'""", True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.WARNING.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            else:
                status = Severity.OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return

        step = 'creating publication and subscription'
        duration_step_start = time.perf_counter()

        try:
            now = datetime.now()
            dt_string = now.strftime("%d%m%Y%H%M")
            pub_name = f'pub_{dt_string}'
            
            conn = connect_to_postgres_database(host_source['name'],db_source['name'], node_source['database_user'], node_source['database_pwd'].strip(), node_source['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return
            
            #for compatibilty with versions less than 12 - ALTER SYSTEM not in with...as
            cur = conn.cursor()
            cur.execute("SHOW wal_level")
            row = cur.fetchone()[0]
            conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
            if 'logical' not in str(row).strip().lower():
                #for publication/subscription - wal level should be >= 'logical'
                cur.execute("ALTER SYSTEM SET wal_level=logical")
    
            cur.execute('CREATE PUBLICATION %s FOR ALL TABLES' % (pub_name,))
            cur.close()
            conn.close()

            # need to reload logical replication worker of source instance before creation of very first subscription, so restart the source instance
            if node_source['dir_123Cluster'].startswith('~'):
                jsn = find_tilda(host_source)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return

                tilda_path = jsn['stdout']
                node_source['dir_123Cluster'] = node_source['dir_123Cluster'].replace('~', tilda_path)

            client = create_ssh_connection(host_source['name'], host_source['username'], host_source['password'], host_source['private_key'], int(host_source['ssh_port']))
            if isinstance(client, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(client[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(client[1]),1)
                return

            postgres_bin_source = find_postgres_bin(host_source, source_version, source_distributor)

            if OperatingSystem.CENTOS.value in source_distributor:
                command = f"sudo systemctl restart postgresql-{source_version}_{node_source['port']}.service"
            else:
                command = f"sudo -iu postgres {postgres_bin_source}/pg_ctl -D '{node_source['dir_123Cluster']}' restart -l /dev/null"
            
            try:
                stdin, stdout, stderr = client.exec_command(command)
                exit_status = stdout.channel.recv_exit_status()# Blocking call
                stdin.channel.shutdown_write()
                try:
                    stderr = stderr.read().decode('ascii').strip()
                    stdout = stdout.read().decode('ascii').strip()
                except UnicodeDecodeError:
                    stderr = stderr.read().decode('utf-8').strip()
                    stdout = stdout.read().decode('utf-8').strip()

                if stderr != '' and 'error' in stderr:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,stdout,stderr,exit_status)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,stdout,stderr,exit_status)
                    return
            except Exception as error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
                return
            
            sub_name = f'sub_{dt_string}'
            conn = connect_to_postgres_database(host_target['name'],db_target['name'], node_target['database_user'], node_target['database_pwd'].strip(), node_target['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return
            
            sub_name = f'sub_{dt_string}'
            conn = connect_to_postgres_database(host_target['name'],db_target['name'], node_target['database_user'], node_target['database_pwd'].strip(), node_target['port'])
            if isinstance(conn, tuple):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
                return
            
            cur = conn.cursor()
            cur.execute("SHOW wal_level")
            row = cur.fetchone()[0]
            conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)

            if 'logical' not in str(row).strip().lower():
                cur.execute("ALTER SYSTEM SET wal_level=logical")
            # seems like fine to pass url-unsafe passwords if libpq connection string are in SQL statement - postgres internally handle the necessary escapes
            cur.execute(f'''CREATE SUBSCRIPTION {sub_name} CONNECTION 'host={host_source['name']} port={node_source['port']} user={node_source['database_user']} password={node_source['database_pwd'].strip()} dbname={db_source['name']}' PUBLICATION {pub_name}''')
            cur.close()
            conn.close()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        jsn = json.loads(remote_command(host_target, f"sudo du -sb '{node_target['dir_123Cluster']}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        node_size = float(jsn['stdout'].strip())/1024
        remote_command(host_target, f"sudo rm -rf '{tmp_folder}'", True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
        
    try:
        ex_params_source = db_source.get('ex_params')
        ex_params_target = db_target.get('ex_params')
        if copy_option == 'live':
            if not ex_params_source:
                ex_params_source = {f'pub_{db_target["name"]}': pub_name}
            elif not ex_params_source.get(f'pub_{db_target["name"]}'):
                ex_params_source[f'pub_{db_target["name"]}'] = pub_name
            if not ex_params_target:
                ex_params_target = {f'sub_{db_target["name"]}': {'name': sub_name, 'publisher_db_id': db_source['id']}}
            elif not ex_params_source.get(f'sub_{db_target["name"]}'):
                ex_params_target[f'sub_{db_target["name"]}'] =  {'name': sub_name, 'publisher_db_id': db_source['id']}
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', ex_params = %s WHERE id=%s""", (Severity.OK.value, Status.OK.value, json.dumps(ex_params_source), db_source['id'],))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', ex_params = %s WHERE id=%s""", (Severity.OK.value, Status.OK.value, json.dumps(ex_params_target), db_target['id'],))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, node_size, node_target['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def switchover_repmgr(standby_host, postgres_version, distributor, repmgr_pwd, siblings = None):
  
    try:
        jsn = {'stdout': '', 'stderr': '', 'excode': 0}
        repmgr_path = f'/etc/repmgr/{postgres_version}'
        repmgr_conf_file = f'{repmgr_path}/repmgr.conf'
        postgres_bin = find_postgres_bin(standby_host, postgres_version, distributor)
        repmgr_executable = f'{postgres_bin}/repmgr'
        
        jsn = check_remote_directory_exists(standby_host, repmgr_executable, 'f', is_dir_empty=False)
        if jsn['stderr'] != '':
            return jsn
        
        repmgr_exists = 'true' in jsn['stdout']
        if not repmgr_exists:
            jsn['stderr'] = 'error: repmgr does not exist'
            jsn['stdout'] = None
            jsn['excode'] = 1
            return jsn
        
        # must never use --force (switchover are too complex to debug if something goes wrong, so strongly unrecommended by repmgr docs)
        # for other repmgr commands --force are preferred for us, other commands simpler to debug, and we could not predict every situation of old config leftover)
        command = f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_conf_file}' standby switchover --siblings-follow"
        jsn = json.loads(remote_command(standby_host, command, True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])
        
        #we know thats crashing case for sure, proceed to workaround
        if OperatingSystem.UBUNTU.value in distributor and postgres_version < 13 and siblings and len(siblings) > 0:
            for s in siblings:
                host = s['host']
                node = s['node']
                command = f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{node['dir_123Cluster']}' restart -l /dev/null"
                jsn = json.loads(remote_command(host, command, True), strict=False)
                if jsn['stdout'] != '':
                    print(jsn['stdout'])
                
                if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                    return jsn
        
        return jsn
    except Exception as error:
        return {'stdout': None, 'stderr': str(error), 'excode': 9}

def switchover_repmgr_task(data, task_id):
    """switchover standby to primary"""
    
    task = 'switchover repmgr task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
        global PHYSICAL_REPLICATION_SLOT
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        standby = json.loads(data['standby_data'])
        primary = json.loads(data['primary_data'])
        primary_hostname = data['primary_hostname']
        old_primary_id = standby['ex_params'].get('pg_primary_id')
        standby_host = json.loads(data['standby_host'])
        siblings = [{'host': json.loads(s['host']), 'node':json.loads(s['node'])} for s in data['siblings_data']]
        postgres_version = int(standby['version'])
        repmgr_pwd = standby['ex_params'].get('repmgr_pwd').strip()
        distributor, release = check_os_version(standby_host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        global PHYSICAL_REPLICATION_SLOT
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'switching over repmgr'
    duration_step_start = time.perf_counter()

    jsn = switchover_repmgr(standby_host, postgres_version, distributor, repmgr_pwd, siblings)
    if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower() or 'error' in jsn['stdout'].lower():
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    conn = connect_to_postgres_database(standby_host['name'], 'postgres', standby['database_user'], standby['database_pwd'].strip(), standby['port'])
    if isinstance(conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
        return

    with conn:
        with conn.cursor() as cur:
            cur.execute("SELECT slot_name FROM pg_replication_slots WHERE slot_name=%s", (PHYSICAL_REPLICATION_SLOT,))
            slot_exists = cur.fetchone()
            if not slot_exists:
                cur.execute("SELECT pg_create_physical_replication_slot(%s)", (PHYSICAL_REPLICATION_SLOT,))
                conn.commit()

    conn = connect_to_postgres_database(primary_hostname, 'postgres', primary['database_user'], primary['database_pwd'].strip(), primary['port'])
    if isinstance(conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(conn[1]),1)
        return

    with conn:
        with conn.cursor() as cur:
            cur.execute("SELECT slot_name FROM pg_replication_slots WHERE slot_name=%s", (PHYSICAL_REPLICATION_SLOT,))
            slot_exists = cur.fetchone()
            if slot_exists:
                cur.execute("SELECT pg_drop_replication_slot(%s)", (PHYSICAL_REPLICATION_SLOT,))
                conn.commit()

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ex_params = ex_params || '{"pg_primary_id":%s}', cluster_role=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, standby['id'], ClusterRoles.PRIMARY.value, standby['id']))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ex_params = ex_params || '{"pg_primary_id":%s}', cluster_role=%s WHERE cluster_id=%s AND id!=%s""", (Severity.OK.value, Status.OK.value, standby['id'], ClusterRoles.STANDBY.value, standby['cluster_id'], standby['id']))
                cur.execute("""UPDATE "clusterCore_c123_database" SET node_id=%s WHERE node_id=%s""", (standby['id'], old_primary_id))
                cur.execute("""UPDATE "clusterCore_c123_cluster" SET ex_params = ex_params || '{"new_basebackup": true}' WHERE id=%s""", (standby['cluster_id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
            
def pause_repmgr_task(data, task_id):
    """pause/unpause repmgrd service"""

    task = 'pause repmgr task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        primary = json.loads(data['primary_data'])
        host = json.loads(data['host_data'])
        toggle_pause = data['toggle_pause']
        standbys_id = data['standbys_id']
        postgres_version = int(primary['version'])

        distributor,release = check_os_version(host)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        postgres_bin = find_postgres_bin(host, postgres_version, distributor)
        repmgr_path = f'/etc/repmgr/{postgres_version}'
        repmgr_config_file = "repmgr.conf"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'un/pausing repmgr'
    duration_step_start = time.perf_counter()

    repmgr_executable = f"{postgres_bin}/repmgr"
    jsn = check_remote_directory_exists(host, repmgr_executable, 'f', is_dir_empty=False)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return
    
    repmgr_exists = 'true' in jsn['stdout']
    if not repmgr_exists:
        error = 'repmgr does not exist'
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    if toggle_pause:
        command = f"sudo -iu postgres {repmgr_executable} -f '{repmgr_path}/{repmgr_config_file}' service pause"
    else:
        command = f"sudo -iu postgres {repmgr_executable} -f '{repmgr_path}/{repmgr_config_file}' service unpause"

    try:          
        jsn = json.loads(remote_command(host, command, True), strict=False)
        if jsn['stdout'] != '':
            print(jsn['stdout'])

        err_list = ('info', 'notice')
        if jsn['stderr'] != '' and not any(err in jsn['stderr'].lower() for err in err_list):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET ex_params = ex_params || '{"repmgr_paused":%s}' WHERE cluster_id = %s AND type != %s""", (toggle_pause, primary['cluster_id'], DBType.HAPROXY.value))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def clone_standby_repmgr_task(data, task_id):
    """clone standby from basebackup of primary via repmgr"""

    task = 'clone standby repmgr task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        standby_data = json.loads(data['standby_data'])
        host_standby = json.loads(data['host_standby'])
        primary_data = json.loads(data['primary_data'])
        host_primary = json.loads(data['host_primary'])
        use_ssl = primary_data['ex_params'].get('ssl')
        certificate_authentication = primary_data['ex_params'].get('certificate_authentication')
        repmgr_pwd = primary_data['ex_params'].get('repmgr_pwd').strip()
        postgres_version = int(standby_data['version'])
        distributor, release = check_os_version(host_standby)
        if OperatingSystem.INVALID.value in distributor:
            jsn = release
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        postgres_bin = find_postgres_bin(host_standby, postgres_version, distributor)
        repmgr_path = f'/etc/repmgr/{postgres_version}'
        repmgr_config_file = f"{repmgr_path}/repmgr.conf"
        repmgr_executable = f"{postgres_bin}/repmgr"
        
        clone_cmd = f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_config_file}' standby clone -h '{host_primary['name']}' -p '{primary_data['port']}' -U repmgr -d repmgr --fast-checkpoint --force"
        register_cmd = f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_config_file}' standby register --force --wait-sync"
        pg_error_filter = ['error', 'fail', 'fatal']
        standby_data_directory = standby_data['dir_123Cluster'].strip()
        if standby_data_directory.startswith('~'):
            jsn = find_tilda(host_primary)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return 
        if OperatingSystem.CENTOS.value in distributor:
            stop_cmd = f"sudo systemctl stop postgresql-{postgres_version}_{standby_data['port']}"
            start_cmd =  f"sudo systemctl start postgresql-{postgres_version}_{standby_data['port']}"
        else:
            stop_cmd = f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{standby_data_directory}' stop -l /dev/null"
            start_cmd = f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{standby_data_directory}' start -l /dev/null"
        
        jsn = check_remote_directory_exists(host_standby, repmgr_executable, 'f', is_dir_empty=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    
        repmgr_exists = 'true' in jsn['stdout']
        if not repmgr_exists:
            error = 'repmgr does not exist'
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return
        
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking cloning prerequisites'
    duration_step_start = time.perf_counter()

    try:
        # counts number of primary records in repmgr cluster metadata, returns true if count == 1 else - false
        jsn = json.loads(remote_command(host_standby, f"""sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable}  -f '{repmgr_config_file}' cluster show | grep -c ' primary ' | awk '{{print ($1 == 1) ? "true" : "false"}}'""", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        has_single_primary = 'true' in jsn['stdout']
        if not has_single_primary:
            # either no primary or split brain
            # (actually, we can clone from any cluster node https://www.repmgr.org/docs/current/repmgr-standby-clone.html,
            # but currently desingned to clone from primary)
            error = f'Cluster does not have a single primary'
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by, '', error,2)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by, '', error,2)
            return

        jsn = json.loads(remote_command(host_standby, f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_config_file}' node check", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        is_standby = 'is standby' in jsn['stdout'].lower()
        if not is_standby:
            #could be either primary or failed node
            error = f'Target node is not standby.Aborting.'
            # we not doing any harm by such checks, so fail with yellow step
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by, '', error,3)
            step = 'null'
            status = Severity.WARNING.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by, '', error, 3)
            return
        
        jsn = json.loads(remote_command(host_primary, f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {repmgr_executable} -f '{repmgr_config_file}' node check", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        # clone could be performed from any node, but now using "user choose which standby to clone from primary" concept
        # and anyway both source and target should have reachable repmgr
        is_primary = 'is primary' in jsn['stdout'].lower()
        if not is_primary:
            #could be either standby or failed node
            error = f'Source node is not primary.Aborting.'
            status = Severity.WARNING.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by, '', error,4)
            step = 'null'
            status = Severity.WARNING.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by, '', error, 4)
            return
        
        # target data directory must not contain running postgres instance even for --dry-run of clone (because clone are restore of basebackup)
        jsn = json.loads(remote_command(host_standby, f'sudo netstat -tunlp | grep {standby_data["port"]}', True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        if jsn['stdout'] != '':
            jsn = json.loads(remote_command(host_standby, stop_cmd, True), strict=False)
            #if node are not running already - thats fine
            if jsn['stderr'] != ''  and 'is the server running' not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

        jsn = json.loads(remote_command(host_standby, f"{clone_cmd} --dry-run", True), strict=False)
        # errors means prerequisites are not met for clone, but node itself are ok, try to bring it back (since it standby - should be no harm)
        if jsn['stderr'] != '' and any(err in jsn['stderr'].lower() for err in pg_error_filter):
            jsn_inner = json.loads(remote_command(host_standby, start_cmd, True), strict=False)
            if jsn_inner['stderr'] != '' and any(err in jsn_inner['stderr'].lower() for err in pg_error_filter):
                # we are not able to bring standby back after stop for --dry-run of clone  - fail with red step
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn_inner['stdout'],jsn_inner['stderr'],jsn_inner['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn_inner['stdout'],jsn_inner['stderr'],jsn_inner['excode'])
                return
            else:
                # node are healthy standby but prerequsites for clone does not met on primary - fail with yellow step
                status = Severity.WARNING.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.WARNING.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'cloning standby'
    duration_step_start = time.perf_counter()

    # means sslmode = verify-full are requested
    if use_ssl and certificate_authentication:
        # then ssl certificates are unique per host - so keep copy of existing certificates
        jsn = transfer_ssl_certificates(standby_data, host_standby, host_standby, mode = 'backup')
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    
    # recommended way are to run clone over empty data directory, and since passed --dry-run - we are justified data directory wipe
    commands = [
        f"sudo rm -rf '{standby_data_directory}'",
        f"sudo mkdir -p '{standby_data_directory}'",
        f"sudo chown -R postgres:postgres '{standby_data_directory}'",
        f"sudo chmod -R 0700 '{standby_data_directory}'",
        clone_cmd,
        start_cmd,
        register_cmd
    ]
    
    for command in commands:
        jsn = json.loads(remote_command(host_standby, command, True), strict=False)
        # return from here - errors means prerequisites are not met for clone, so need examine logs of instance
        if jsn['stderr'] != '' and any(err in jsn['stderr'].lower() for err in pg_error_filter):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

    # means sslmode = verify-full are requested
    if use_ssl and certificate_authentication:
        # then ssl certificates are unique per host - so put back existing certificates
        jsn = transfer_ssl_certificates(standby_data, host_standby, host_standby, mode='restore')
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        cmd = f"sudo systemctl restart postgresql-{postgres_version}_{standby_data['port']}.service" if OperatingSystem.CENTOS.value in distributor else f"sudo -iu postgres {postgres_bin}/pg_ctl -D '{standby_data_directory}' restart -l /dev/null"
        jsn = json.loads(remote_command(host_standby, cmd, True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, cluster_role=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, ClusterRoles.STANDBY.value, standby_data['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def generate_haproxy_conf(node_list, haproxy_port = 5000):
    config = f"""global
    log         127.0.0.1 local2
    chroot      /var/lib/haproxy
    pidfile     /var/run/haproxy.pid
    maxconn     40000
    user        haproxy
    group       haproxy
    daemon
    stats socket /var/lib/haproxy/stats
    ssl-default-bind-ciphers PROFILE=SYSTEM
    ssl-default-server-ciphers PROFILE=SYSTEM

defaults
    mode tcp
    log global
    timeout client 30m
    timeout server 30m
    timeout check 5s
    timeout connect 5s

listen cluster_{haproxy_port}
    bind *:{haproxy_port}
    mode tcp
    balance leastconn
    option httpchk
    http-check expect status 200
"""
    for node in node_list:
        config += f"    server pg{node['node_id']} {node['host_ip']}:{node['node_port']} check port 80\n"

    return config

def append_haproxy_conf(node_list, haproxy_port = 5000):
    appendix = f"""listen cluster_{haproxy_port}
    bind *:{haproxy_port}
    mode tcp
    balance leastconn
    option httpchk
    http-check expect status 200
"""
    for node in node_list:
        appendix += f"    server pg{node['node_id']} {node['host_ip']}:{node['node_port']} check port 80\n"

    return appendix

def install_healthcheck(distributor, release, host, node, app_host, offline):
    """creating health check script for haproxy on host of node"""

    healthcheck_path = '/root/pg-healthcheck'
    pg_ha_service_path = '/lib/systemd/system/postgresql-ha.service' if OperatingSystem.UBUNTU.value in distributor.lower() else '/usr/lib/systemd/system/postgresql-ha.service'
    if not offline:
        try:
            if OperatingSystem.UBUNTU.value in distributor.lower():
                jsn = json.loads(remote_command(host, 'sudo apt-get update && sudo DEBIAN_FRONTEND=noninteractive apt-get install -qq -y python3 python3-psycopg2 python3-pip', True), strict=False)
                error_lower = jsn['stderr'].lower()
                if error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['error', 'fail', 'err', 'errno'] if 'nothing to do' not in error_lower)):
                    return jsn
            elif OperatingSystem.CENTOS.value in distributor.lower() or 'red hat' in distributor.lower():
                jsn = json.loads(remote_command(host, f'sudo {"dnf" if int(release) > 7 else "yum"} --enablerepo=appstream install -q -y epel-release python3 python3-psycopg2 python3-pip', True), strict=False)
                error_lower = jsn['stderr'].lower()
                if jsn['stderr'] != '' and any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno']) and not any(err in error_lower for err in ['already installed', 'nothing to do', 'is listed more than once in the configuration']):
                    return jsn
            else:
                jsn = {
                    'stderr': 'Distributor is not Ubuntu or CentOS',
                    'stdout': None,
                    'excode': 99
                }
                return jsn

            jsn = json.loads(remote_command(host, 'sudo python3 -m pip install cherrypy', True), strict=False)
            if 'error' in jsn['stderr'].lower() and 'error: nothing to do' not in jsn['stderr'].lower() and int(jsn['excode']) != 0:
                return jsn
        except Exception as error:
            return {'stdout': None, 'stderr': str(error), 'excode': 9}
    else:
        jsn = install_offline_remote_to_remote(distributor, release, host, None, DBType.POSTGRES.value, app_host, Packages.PYTHON3.value)
        error_lower = jsn['stderr'].lower()
        filter = error_lower != '' and ('E:' in jsn['stderr'] or any(err in error_lower for err in ['err', 'error', 'exception', 'fail', 'errno'])) and not any(err in error_lower for err in ['already installed', 'nothing to do'])
        if filter:
            return jsn 
    
    healthcheck_script = generate_healthcheck_script(node)
    pg_ha_service_content = generate_postgresql_ha_service(node)
    commands = [
        f"echo '{healthcheck_script}' | sudo tee '{healthcheck_path}'",
        f"echo '{pg_ha_service_content}' | sudo tee '{pg_ha_service_path}'",
        f"sudo chmod +x '{healthcheck_path}'",
        "sudo systemctl daemon-reload",
        "sudo systemctl restart postgresql-ha.service",
        "sudo systemctl enable postgresql-ha.service"
    ]
    for command in commands:
        try:
            jsn = json.loads(remote_command(host, command, True), strict=False)
            if 'error' in jsn['stderr'].lower() and int(jsn['excode']) != 0:
                return jsn
        except Exception as error:
            return {'stdout': None, 'stderr': str(error), 'excode': 9}
    
    return jsn
        
def generate_healthcheck_script(node):
    update_params = '''{"environment": "production","log.screen": False,"server.socket_host": "0.0.0.0","server.socket_port": int(SYS_WEBPORT)}'''
    script = f"""#!/usr/bin/python3
import sys,psycopg2,cherrypy
SYS_PGPORT={node['port']}
SYS_CHECKPEERS=0
SYS_WEBPORT=80
class WebHandler(object):
    @cherrypy.expose
    def index(self):
        global SYS_PGPORT
        con = None
        try:
            con = psycopg2.connect("dbname=postgres user={node['database_user']} password={node['database_pwd'].strip()} port={node['port']}")
        except:
            cherrypy.response.status = 500
            return ''
        rs = con.cursor()
        rs.execute("SELECT pg_is_in_recovery()::int;")
        row = rs.fetchone()
        if (int(row[0]) == 1):
            cherrypy.response.status = 500
            return ''
        cherrypy.response.status = 200
        return ''
    
if __name__ == "__main__":
    for ar in sys.argv:
        if (ar.lower().startswith("--pgport=")):
            SYS_PGPORT=ar.replace("--pgport=","")
        if (ar.lower().startswith("--webport=")):
            SYS_WEBPORT=ar.replace("--webport=","")
        if (ar.lower() == "--with-peers"):
            SYS_CHECKPEERS=1
    if (str(SYS_PGPORT) == ""):
        print("  ERR: Invalid/missing PG port.")
        sys.exit(2)
    if (str(SYS_WEBPORT) == ""):
        print("  ERR: Invalid/missing WEB port.")
        sys.exit(2)
    cherrypy.config.update({update_params})
    cherrypy.quickstart(WebHandler())
"""
    return script

def generate_postgresql_ha_service(node):
    service_content = f"""[Unit]
Description=PostgreSQL {int(node['version'])} NLB
After=network.target

[Service]
ExecStart=/usr/bin/python3 /root/pg-healthcheck

[Install]
WantedBy=multi-user.target
"""
    return service_content

# TODO:
def change_node_size_task(data, task_id):

    task = 'change node size task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by) 

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)